// Emacs style mode select   -*- C++ -*- 
//-----------------------------------------------------------------------------
//
// $Id: info.c,v 1.2 2000-08-12 21:29:25 fraggle Exp $
//
// Copyright (C) 1993-1996 by id Software, Inc.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// DESCRIPTION:
//  Thing frame/state LUT,
//  generated by multigen utilitiy.
//  This one is the original DOOM version, preserved.
//  BOOM changes include commenting and addition of predefined lumps
//  for providing things that aren't in the IWAD without sending a
//  separate must-use wad file around with the EXE.
//
//-----------------------------------------------------------------------------

static const char
rcsid[] = "$Id: info.c,v 1.2 2000-08-12 21:29:25 fraggle Exp $";

#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "doomdef.h"
#include "sounds.h"
#include "m_fixed.h"
#include "p_mobj.h"
#include "w_wad.h"
#include "info.h"


// ********************************************************************
// Sprite names
// ********************************************************************
// This is the list of sprite 4-character prefixes.  They are searched
// through, with a NULL entry terminating the list.  In DOOM originally
// this NULL entry was missing, and coincidentally the next thing in
// memory was the dummy state_t[] entry that started with zero bytes.
// killough 1/17/98: add an explicit NULL entry.
// NUMSPRITES is an enum from info.h where all these are listed
// as SPR_xxxx

char *sprnames[NUMSPRITES+1] = {
  "TROO","SHTG","PUNG","PISG","PISF","SHTF","SHT2","CHGG","CHGF","MISG",
  "MISF","SAWG","PLSG","PLSF","BFGG","BFGF","BLUD","PUFF","BAL1","BAL2",
  "PLSS","PLSE","MISL","BFS1","BFE1","BFE2","TFOG","IFOG","PLAY","POSS",
  "SPOS","VILE","FIRE","FATB","FBXP","SKEL","MANF","FATT","CPOS","SARG",
  "HEAD","BAL7","BOSS","BOS2","SKUL","SPID","BSPI","APLS","APBX","CYBR",
  "PAIN","SSWV","KEEN","BBRN","BOSF","ARM1","ARM2","BAR1","BEXP","FCAN",
  "BON1","BON2","BKEY","RKEY","YKEY","BSKU","RSKU","YSKU","STIM","MEDI",
  "SOUL","PINV","PSTR","PINS","MEGA","SUIT","PMAP","PVIS","CLIP","AMMO",
  "ROCK","BROK","CELL","CELP","SHEL","SBOX","BPAK","BFUG","MGUN","CSAW",
  "LAUN","PLAS","SHOT","SGN2","COLU","SMT2","GOR1","POL2","POL5","POL4",
  "POL3","POL1","POL6","GOR2","GOR3","GOR4","GOR5","SMIT","COL1","COL2",
  "COL3","COL4","CAND","CBRA","COL6","TRE1","TRE2","ELEC","CEYE","FSKU",
  "COL5","TBLU","TGRN","TRED","SMBT","SMGT","SMRT","HDB1","HDB2","HDB3",
  "HDB4","HDB5","HDB6","POB1","POB2","BRS1","TLMP","TLP2",
  "TNT1", // invisible sprite                                 phares 3/9/98

#ifdef DOGS
  "DOGS", // killough 7/19/98: Marine's best friend :)
#endif

#ifdef BETA
  "PLS1", // killough 7/19/98: first  of two plasma fireballs in the beta
  "PLS2", // killough 7/19/98: second of two plasma fireballs in the beta
  "BON3", // killough 7/11/98: evil sceptre in the beta version
  "BON4", // killough 7/11/98: unholy bible in the beta version
#endif

  NULL
};

// ********************************************************************
// Function addresses or Code Pointers
// ********************************************************************
// These function addresses are the Code Pointers that have been
// modified for years by Dehacked enthusiasts.  The new BEX format
// allows more extensive changes (see d_deh.c)

// Doesn't work with g++, needs actionf_p1
void A_Light0();
void A_WeaponReady();
void A_Lower();
void A_Raise();
void A_Punch();
void A_ReFire();
void A_FirePistol();
void A_Light1();
void A_FireShotgun();
void A_Light2();
void A_FireShotgun2();
void A_CheckReload();
void A_OpenShotgun2();
void A_LoadShotgun2();
void A_CloseShotgun2();
void A_FireCGun();
void A_GunFlash();
void A_FireMissile();
void A_Saw();
void A_FirePlasma();
void A_BFGsound();
void A_FireBFG();
void A_BFGSpray();
void A_Explode();
void A_Pain();
void A_PlayerScream();
void A_Fall();
void A_Stop();
void A_XScream();
void A_Look();
void A_Chase();
void A_FaceTarget();
void A_PosAttack();
void A_Scream();
void A_Die();
void A_SPosAttack();
void A_VileChase();
void A_VileStart();
void A_VileTarget();
void A_VileAttack();
void A_StartFire();
void A_Fire();
void A_FireCrackle();
void A_Tracer();
void A_SkelWhoosh();
void A_SkelFist();
void A_SkelMissile();
void A_FatRaise();
void A_FatAttack1();
void A_FatAttack2();
void A_FatAttack3();
void A_BossDeath();
void A_CPosAttack();
void A_CPosRefire();
void A_TroopAttack();
void A_SargAttack();
void A_HeadAttack();
void A_BruisAttack();
void A_SkullAttack();
void A_Metal();
void A_SpidRefire();
void A_BabyMetal();
void A_BspiAttack();
void A_Hoof();
void A_CyberAttack();
void A_PainAttack();
void A_PainDie();
void A_KeenDie();
void A_BrainPain();
void A_BrainScream();
void A_BrainDie();
void A_BrainAwake();
void A_BrainSpit();
void A_SpawnSound();
void A_SpawnFly();
void A_BrainExplode();
void A_FireOldBFG();      // killough 7/19/98: classic BFG firing function
void A_Detonate();        // killough 8/9/98: detonate a bomb or other device
void A_Mushroom();        // killough 10/98: mushroom effect
void A_BetaSkullAttack(); // killough 10/98: beta lost souls attacked different

// ********************************************************************
// State or "frame" information
// ********************************************************************
// Each of the states, otherwise known as "frames", is outlined
// here.  The data in each element of the array is the way it is
// initialized, with sprite names identified by their enumerator
// value such as SPR_SHTG.  These correlate to the above sprite
// array so don't change them around unless you understand what
// you're doing.
//
// The commented name beginning with S_ at the end of each line
// is there to help figure out where the next-frame pointer is
// pointing.  These are also additionally identified in info.h
// as enumerated values.  From a change-and-recompile point of
// view this is fairly workable, but it adds a lot to the effort
// when trying to change things externally.  See also the d_deh.c
// parts where frame rewiring is done for more details and the
// extended way a BEX file can handle this.
//
// killough 11/98: cleaned up, since we don't need to accomodate C++

state_t states[NUMSTATES] = {
  {SPR_TROO,0,-1,NULL,S_NULL},  // S_NULL
  {SPR_SHTG,4,0,A_Light0,S_NULL}, // S_LIGHTDONE
  {SPR_PUNG,0,1,A_WeaponReady,S_PUNCH}, // S_PUNCH
  {SPR_PUNG,0,1,A_Lower,S_PUNCHDOWN}, // S_PUNCHDOWN
  {SPR_PUNG,0,1,A_Raise,S_PUNCHUP}, // S_PUNCHUP
  {SPR_PUNG,1,4,NULL,S_PUNCH2},   // S_PUNCH1
  {SPR_PUNG,2,4,A_Punch,S_PUNCH3},  // S_PUNCH2
  {SPR_PUNG,3,5,NULL,S_PUNCH4},   // S_PUNCH3
  {SPR_PUNG,2,4,NULL,S_PUNCH5},   // S_PUNCH4
  {SPR_PUNG,1,5,A_ReFire,S_PUNCH},  // S_PUNCH5
  {SPR_PISG,0,1,A_WeaponReady,S_PISTOL},// S_PISTOL
  {SPR_PISG,0,1,A_Lower,S_PISTOLDOWN},  // S_PISTOLDOWN
  {SPR_PISG,0,1,A_Raise,S_PISTOLUP},  // S_PISTOLUP
  {SPR_PISG,0,4,NULL,S_PISTOL2},  // S_PISTOL1
  {SPR_PISG,1,6,A_FirePistol,S_PISTOL3},// S_PISTOL2
  {SPR_PISG,2,4,NULL,S_PISTOL4},  // S_PISTOL3
  {SPR_PISG,1,5,A_ReFire,S_PISTOL}, // S_PISTOL4
  {SPR_PISF,32768,7,A_Light1,S_LIGHTDONE},  // S_PISTOLFLASH
  {SPR_SHTG,0,1,A_WeaponReady,S_SGUN},  // S_SGUN
  {SPR_SHTG,0,1,A_Lower,S_SGUNDOWN},  // S_SGUNDOWN
  {SPR_SHTG,0,1,A_Raise,S_SGUNUP},  // S_SGUNUP
  {SPR_SHTG,0,3,NULL,S_SGUN2},  // S_SGUN1
  {SPR_SHTG,0,7,A_FireShotgun,S_SGUN3}, // S_SGUN2
  {SPR_SHTG,1,5,NULL,S_SGUN4},  // S_SGUN3
  {SPR_SHTG,2,5,NULL,S_SGUN5},  // S_SGUN4
  {SPR_SHTG,3,4,NULL,S_SGUN6},  // S_SGUN5
  {SPR_SHTG,2,5,NULL,S_SGUN7},  // S_SGUN6
  {SPR_SHTG,1,5,NULL,S_SGUN8},  // S_SGUN7
  {SPR_SHTG,0,3,NULL,S_SGUN9},  // S_SGUN8
  {SPR_SHTG,0,7,A_ReFire,S_SGUN}, // S_SGUN9
  {SPR_SHTF,32768,4,A_Light1,S_SGUNFLASH2}, // S_SGUNFLASH1
  {SPR_SHTF,32769,3,A_Light2,S_LIGHTDONE},  // S_SGUNFLASH2
  {SPR_SHT2,0,1,A_WeaponReady,S_DSGUN}, // S_DSGUN
  {SPR_SHT2,0,1,A_Lower,S_DSGUNDOWN}, // S_DSGUNDOWN
  {SPR_SHT2,0,1,A_Raise,S_DSGUNUP}, // S_DSGUNUP
  {SPR_SHT2,0,3,NULL,S_DSGUN2}, // S_DSGUN1

  // killough 9/5/98: make SSG lighting flash more uniform along super shotgun:

  {SPR_SHT2, 32768 /* killough */,7,A_FireShotgun2,S_DSGUN3}, // S_DSGUN2
  {SPR_SHT2,1,7,NULL,S_DSGUN4}, // S_DSGUN3
  {SPR_SHT2,2,7,A_CheckReload,S_DSGUN5},  // S_DSGUN4
  {SPR_SHT2,3,7,A_OpenShotgun2,S_DSGUN6}, // S_DSGUN5
  {SPR_SHT2,4,7,NULL,S_DSGUN7}, // S_DSGUN6
  {SPR_SHT2,5,7,A_LoadShotgun2,S_DSGUN8}, // S_DSGUN7
  {SPR_SHT2,6,6,NULL,S_DSGUN9}, // S_DSGUN8
  {SPR_SHT2,7,6,A_CloseShotgun2,S_DSGUN10}, // S_DSGUN9
  {SPR_SHT2,0,5,A_ReFire,S_DSGUN},  // S_DSGUN10
  {SPR_SHT2,1,7,NULL,S_DSNR2},  // S_DSNR1
  {SPR_SHT2,0,3,NULL,S_DSGUNDOWN},  // S_DSNR2

  // killough 8/20/98: reduce first SSG flash frame one tic, to fix
  // Doom II SSG flash bug, in which SSG raises before flash finishes

  {SPR_SHT2,32776,4/*killough*/,A_Light1,S_DSGUNFLASH2}, // S_DSGUNFLASH1
  {SPR_SHT2,32777,4,A_Light2,S_LIGHTDONE},  // S_DSGUNFLASH2
  {SPR_CHGG,0,1,A_WeaponReady,S_CHAIN}, // S_CHAIN
  {SPR_CHGG,0,1,A_Lower,S_CHAINDOWN}, // S_CHAINDOWN
  {SPR_CHGG,0,1,A_Raise,S_CHAINUP}, // S_CHAINUP
  {SPR_CHGG,0,4,A_FireCGun,S_CHAIN2}, // S_CHAIN1
  {SPR_CHGG,1,4,A_FireCGun,S_CHAIN3}, // S_CHAIN2
  {SPR_CHGG,1,0,A_ReFire,S_CHAIN},  // S_CHAIN3
  {SPR_CHGF,32768,5,A_Light1,S_LIGHTDONE},  // S_CHAINFLASH1
  {SPR_CHGF,32769,5,A_Light2,S_LIGHTDONE},  // S_CHAINFLASH2
  {SPR_MISG,0,1,A_WeaponReady,S_MISSILE}, // S_MISSILE
  {SPR_MISG,0,1,A_Lower,S_MISSILEDOWN}, // S_MISSILEDOWN
  {SPR_MISG,0,1,A_Raise,S_MISSILEUP}, // S_MISSILEUP
  {SPR_MISG,1,8,A_GunFlash,S_MISSILE2}, // S_MISSILE1
  {SPR_MISG,1,12,A_FireMissile,S_MISSILE3}, // S_MISSILE2
  {SPR_MISG,1,0,A_ReFire,S_MISSILE},  // S_MISSILE3
  {SPR_MISF,32768,3,A_Light1,S_MISSILEFLASH2},  // S_MISSILEFLASH1
  {SPR_MISF,32769,4,NULL,S_MISSILEFLASH3},  // S_MISSILEFLASH2
  {SPR_MISF,32770,4,A_Light2,S_MISSILEFLASH4},  // S_MISSILEFLASH3
  {SPR_MISF,32771,4,A_Light2,S_LIGHTDONE},  // S_MISSILEFLASH4
  {SPR_SAWG,2,4,A_WeaponReady,S_SAWB},  // S_SAW
  {SPR_SAWG,3,4,A_WeaponReady,S_SAW}, // S_SAWB
  {SPR_SAWG,2,1,A_Lower,S_SAWDOWN}, // S_SAWDOWN
  {SPR_SAWG,2,1,A_Raise,S_SAWUP}, // S_SAWUP
  {SPR_SAWG,0,4,A_Saw,S_SAW2}, // S_SAW1
  {SPR_SAWG,1,4,A_Saw,S_SAW3},  // S_SAW2
  {SPR_SAWG,1,0,A_ReFire,S_SAW},  // S_SAW3
  {SPR_PLSG,0,1,A_WeaponReady,S_PLASMA},  // S_PLASMA
  {SPR_PLSG,0,1,A_Lower,S_PLASMADOWN},  // S_PLASMADOWN
  {SPR_PLSG,0,1,A_Raise,S_PLASMAUP},  // S_PLASMAUP
  {SPR_PLSG,0,3,A_FirePlasma,S_PLASMA2},  // S_PLASMA1
  {SPR_PLSG,1,20,A_ReFire,S_PLASMA},  // S_PLASMA2
  {SPR_PLSF,32768,4,A_Light1,S_LIGHTDONE},  // S_PLASMAFLASH1
  {SPR_PLSF,32769,4,A_Light1,S_LIGHTDONE},  // S_PLASMAFLASH2
  {SPR_BFGG,0,1,A_WeaponReady,S_BFG}, // S_BFG
  {SPR_BFGG,0,1,A_Lower,S_BFGDOWN}, // S_BFGDOWN
  {SPR_BFGG,0,1,A_Raise,S_BFGUP}, // S_BFGUP
  {SPR_BFGG,0,20,A_BFGsound,S_BFG2},  // S_BFG1
  {SPR_BFGG,1,10,A_GunFlash,S_BFG3},  // S_BFG2
  {SPR_BFGG,1,10,A_FireBFG,S_BFG4}, // S_BFG3
  {SPR_BFGG,1,20,A_ReFire,S_BFG}, // S_BFG4
  {SPR_BFGF,32768,11,A_Light1,S_BFGFLASH2}, // S_BFGFLASH1
  {SPR_BFGF,32769,6,A_Light2,S_LIGHTDONE},  // S_BFGFLASH2
  {SPR_BLUD,2,8,NULL,S_BLOOD2}, // S_BLOOD1
  {SPR_BLUD,1,8,NULL,S_BLOOD3}, // S_BLOOD2
  {SPR_BLUD,0,8,NULL,S_NULL}, // S_BLOOD3
  {SPR_PUFF,32768,4,NULL,S_PUFF2},  // S_PUFF1
  {SPR_PUFF,1,4,NULL,S_PUFF3},  // S_PUFF2
  {SPR_PUFF,2,4,NULL,S_PUFF4},  // S_PUFF3
  {SPR_PUFF,3,4,NULL,S_NULL}, // S_PUFF4
  {SPR_BAL1,32768,4,NULL,S_TBALL2}, // S_TBALL1
  {SPR_BAL1,32769,4,NULL,S_TBALL1}, // S_TBALL2
  {SPR_BAL1,32770,6,NULL,S_TBALLX2},  // S_TBALLX1
  {SPR_BAL1,32771,6,NULL,S_TBALLX3},  // S_TBALLX2
  {SPR_BAL1,32772,6,NULL,S_NULL}, // S_TBALLX3
  {SPR_BAL2,32768,4,NULL,S_RBALL2}, // S_RBALL1
  {SPR_BAL2,32769,4,NULL,S_RBALL1}, // S_RBALL2
  {SPR_BAL2,32770,6,NULL,S_RBALLX2},  // S_RBALLX1
  {SPR_BAL2,32771,6,NULL,S_RBALLX3},  // S_RBALLX2
  {SPR_BAL2,32772,6,NULL,S_NULL}, // S_RBALLX3
  {SPR_PLSS,32768,6,NULL,S_PLASBALL2},  // S_PLASBALL
  {SPR_PLSS,32769,6,NULL,S_PLASBALL}, // S_PLASBALL2
  {SPR_PLSE,32768,4,NULL,S_PLASEXP2}, // S_PLASEXP
  {SPR_PLSE,32769,4,NULL,S_PLASEXP3}, // S_PLASEXP2
  {SPR_PLSE,32770,4,NULL,S_PLASEXP4}, // S_PLASEXP3
  {SPR_PLSE,32771,4,NULL,S_PLASEXP5}, // S_PLASEXP4
  {SPR_PLSE,32772,4,NULL,S_NULL}, // S_PLASEXP5
  {SPR_MISL,32768,1,NULL,S_ROCKET}, // S_ROCKET
  {SPR_BFS1,32768,4,NULL,S_BFGSHOT2}, // S_BFGSHOT
  {SPR_BFS1,32769,4,NULL,S_BFGSHOT},  // S_BFGSHOT2
  {SPR_BFE1,32768,8,NULL,S_BFGLAND2}, // S_BFGLAND
  {SPR_BFE1,32769,8,NULL,S_BFGLAND3}, // S_BFGLAND2
  {SPR_BFE1,32770,8,A_BFGSpray,S_BFGLAND4}, // S_BFGLAND3
  {SPR_BFE1,32771,8,NULL,S_BFGLAND5}, // S_BFGLAND4
  {SPR_BFE1,32772,8,NULL,S_BFGLAND6}, // S_BFGLAND5
  {SPR_BFE1,32773,8,NULL,S_NULL}, // S_BFGLAND6
  {SPR_BFE2,32768,8,NULL,S_BFGEXP2},  // S_BFGEXP
  {SPR_BFE2,32769,8,NULL,S_BFGEXP3},  // S_BFGEXP2
  {SPR_BFE2,32770,8,NULL,S_BFGEXP4},  // S_BFGEXP3
  {SPR_BFE2,32771,8,NULL,S_NULL}, // S_BFGEXP4
  {SPR_MISL,32769,8,A_Explode,S_EXPLODE2},  // S_EXPLODE1
  {SPR_MISL,32770,6,NULL,S_EXPLODE3}, // S_EXPLODE2
  {SPR_MISL,32771,4,NULL,S_NULL}, // S_EXPLODE3
  {SPR_TFOG,32768,6,NULL,S_TFOG01}, // S_TFOG
  {SPR_TFOG,32769,6,NULL,S_TFOG02}, // S_TFOG01
  {SPR_TFOG,32768,6,NULL,S_TFOG2},  // S_TFOG02
  {SPR_TFOG,32769,6,NULL,S_TFOG3},  // S_TFOG2
  {SPR_TFOG,32770,6,NULL,S_TFOG4},  // S_TFOG3
  {SPR_TFOG,32771,6,NULL,S_TFOG5},  // S_TFOG4
  {SPR_TFOG,32772,6,NULL,S_TFOG6},  // S_TFOG5
  {SPR_TFOG,32773,6,NULL,S_TFOG7},  // S_TFOG6
  {SPR_TFOG,32774,6,NULL,S_TFOG8},  // S_TFOG7
  {SPR_TFOG,32775,6,NULL,S_TFOG9},  // S_TFOG8
  {SPR_TFOG,32776,6,NULL,S_TFOG10}, // S_TFOG9
  {SPR_TFOG,32777,6,NULL,S_NULL}, // S_TFOG10
  {SPR_IFOG,32768,6,NULL,S_IFOG01}, // S_IFOG
  {SPR_IFOG,32769,6,NULL,S_IFOG02}, // S_IFOG01
  {SPR_IFOG,32768,6,NULL,S_IFOG2},  // S_IFOG02
  {SPR_IFOG,32769,6,NULL,S_IFOG3},  // S_IFOG2
  {SPR_IFOG,32770,6,NULL,S_IFOG4},  // S_IFOG3
  {SPR_IFOG,32771,6,NULL,S_IFOG5},  // S_IFOG4
  {SPR_IFOG,32772,6,NULL,S_NULL}, // S_IFOG5
  {SPR_PLAY,0,-1,NULL,S_NULL},  // S_PLAY
  {SPR_PLAY,0,4,NULL,S_PLAY_RUN2},  // S_PLAY_RUN1
  {SPR_PLAY,1,4,NULL,S_PLAY_RUN3},  // S_PLAY_RUN2
  {SPR_PLAY,2,4,NULL,S_PLAY_RUN4},  // S_PLAY_RUN3
  {SPR_PLAY,3,4,NULL,S_PLAY_RUN1},  // S_PLAY_RUN4
  {SPR_PLAY,4,12,NULL,S_PLAY},  // S_PLAY_ATK1
  {SPR_PLAY,32773,6,NULL,S_PLAY_ATK1},  // S_PLAY_ATK2
  {SPR_PLAY,6,4,NULL,S_PLAY_PAIN2}, // S_PLAY_PAIN
  {SPR_PLAY,6,4,A_Pain,S_PLAY}, // S_PLAY_PAIN2
  {SPR_PLAY,7,10,NULL,S_PLAY_DIE2}, // S_PLAY_DIE1
  {SPR_PLAY,8,10,A_PlayerScream,S_PLAY_DIE3}, // S_PLAY_DIE2
  {SPR_PLAY,9,10,A_Fall,S_PLAY_DIE4}, // S_PLAY_DIE3
  {SPR_PLAY,10,10,NULL,S_PLAY_DIE5},  // S_PLAY_DIE4
  {SPR_PLAY,11,10,NULL,S_PLAY_DIE6},  // S_PLAY_DIE5
  {SPR_PLAY,12,10,NULL,S_PLAY_DIE7},  // S_PLAY_DIE6
  {SPR_PLAY,13,-1,NULL,S_NULL}, // S_PLAY_DIE7
  {SPR_PLAY,14,5,NULL,S_PLAY_XDIE2},  // S_PLAY_XDIE1
  {SPR_PLAY,15,5,A_XScream,S_PLAY_XDIE3}, // S_PLAY_XDIE2
  {SPR_PLAY,16,5,A_Fall,S_PLAY_XDIE4},  // S_PLAY_XDIE3
  {SPR_PLAY,17,5,NULL,S_PLAY_XDIE5},  // S_PLAY_XDIE4
  {SPR_PLAY,18,5,NULL,S_PLAY_XDIE6},  // S_PLAY_XDIE5
  {SPR_PLAY,19,5,NULL,S_PLAY_XDIE7},  // S_PLAY_XDIE6
  {SPR_PLAY,20,5,NULL,S_PLAY_XDIE8},  // S_PLAY_XDIE7
  {SPR_PLAY,21,5,NULL,S_PLAY_XDIE9},  // S_PLAY_XDIE8
  {SPR_PLAY,22,-1,NULL,S_NULL}, // S_PLAY_XDIE9
  {SPR_POSS,0,10,A_Look,S_POSS_STND2},  // S_POSS_STND
  {SPR_POSS,1,10,A_Look,S_POSS_STND}, // S_POSS_STND2
  {SPR_POSS,0,4,A_Chase,S_POSS_RUN2}, // S_POSS_RUN1
  {SPR_POSS,0,4,A_Chase,S_POSS_RUN3}, // S_POSS_RUN2
  {SPR_POSS,1,4,A_Chase,S_POSS_RUN4}, // S_POSS_RUN3
  {SPR_POSS,1,4,A_Chase,S_POSS_RUN5}, // S_POSS_RUN4
  {SPR_POSS,2,4,A_Chase,S_POSS_RUN6}, // S_POSS_RUN5
  {SPR_POSS,2,4,A_Chase,S_POSS_RUN7}, // S_POSS_RUN6
  {SPR_POSS,3,4,A_Chase,S_POSS_RUN8}, // S_POSS_RUN7
  {SPR_POSS,3,4,A_Chase,S_POSS_RUN1}, // S_POSS_RUN8
  {SPR_POSS,4,10,A_FaceTarget,S_POSS_ATK2}, // S_POSS_ATK1
  {SPR_POSS,5,8,A_PosAttack,S_POSS_ATK3}, // S_POSS_ATK2
  {SPR_POSS,4,8,NULL,S_POSS_RUN1},  // S_POSS_ATK3
  {SPR_POSS,6,3,NULL,S_POSS_PAIN2}, // S_POSS_PAIN
  {SPR_POSS,6,3,A_Pain,S_POSS_RUN1},  // S_POSS_PAIN2
  {SPR_POSS,7,5,NULL,S_POSS_DIE2},  // S_POSS_DIE1
  {SPR_POSS,8,5,A_Scream,S_POSS_DIE3},  // S_POSS_DIE2
  {SPR_POSS,9,5,A_Fall,S_POSS_DIE4},  // S_POSS_DIE3
  {SPR_POSS,10,5,NULL,S_POSS_DIE5}, // S_POSS_DIE4
  {SPR_POSS,11,-1,NULL,S_NULL}, // S_POSS_DIE5
  {SPR_POSS,12,5,NULL,S_POSS_XDIE2},  // S_POSS_XDIE1
  {SPR_POSS,13,5,A_XScream,S_POSS_XDIE3}, // S_POSS_XDIE2
  {SPR_POSS,14,5,A_Fall,S_POSS_XDIE4},  // S_POSS_XDIE3
  {SPR_POSS,15,5,NULL,S_POSS_XDIE5},  // S_POSS_XDIE4
  {SPR_POSS,16,5,NULL,S_POSS_XDIE6},  // S_POSS_XDIE5
  {SPR_POSS,17,5,NULL,S_POSS_XDIE7},  // S_POSS_XDIE6
  {SPR_POSS,18,5,NULL,S_POSS_XDIE8},  // S_POSS_XDIE7
  {SPR_POSS,19,5,NULL,S_POSS_XDIE9},  // S_POSS_XDIE8
  {SPR_POSS,20,-1,NULL,S_NULL}, // S_POSS_XDIE9
  {SPR_POSS,10,5,NULL,S_POSS_RAISE2}, // S_POSS_RAISE1
  {SPR_POSS,9,5,NULL,S_POSS_RAISE3},  // S_POSS_RAISE2
  {SPR_POSS,8,5,NULL,S_POSS_RAISE4},  // S_POSS_RAISE3
  {SPR_POSS,7,5,NULL,S_POSS_RUN1},  // S_POSS_RAISE4
  {SPR_SPOS,0,10,A_Look,S_SPOS_STND2},  // S_SPOS_STND
  {SPR_SPOS,1,10,A_Look,S_SPOS_STND}, // S_SPOS_STND2
  {SPR_SPOS,0,3,A_Chase,S_SPOS_RUN2}, // S_SPOS_RUN1
  {SPR_SPOS,0,3,A_Chase,S_SPOS_RUN3}, // S_SPOS_RUN2
  {SPR_SPOS,1,3,A_Chase,S_SPOS_RUN4}, // S_SPOS_RUN3
  {SPR_SPOS,1,3,A_Chase,S_SPOS_RUN5}, // S_SPOS_RUN4
  {SPR_SPOS,2,3,A_Chase,S_SPOS_RUN6}, // S_SPOS_RUN5
  {SPR_SPOS,2,3,A_Chase,S_SPOS_RUN7}, // S_SPOS_RUN6
  {SPR_SPOS,3,3,A_Chase,S_SPOS_RUN8}, // S_SPOS_RUN7
  {SPR_SPOS,3,3,A_Chase,S_SPOS_RUN1}, // S_SPOS_RUN8
  {SPR_SPOS,4,10,A_FaceTarget,S_SPOS_ATK2}, // S_SPOS_ATK1
  {SPR_SPOS,32773,10,A_SPosAttack,S_SPOS_ATK3}, // S_SPOS_ATK2
  {SPR_SPOS,4,10,NULL,S_SPOS_RUN1}, // S_SPOS_ATK3
  {SPR_SPOS,6,3,NULL,S_SPOS_PAIN2}, // S_SPOS_PAIN
  {SPR_SPOS,6,3,A_Pain,S_SPOS_RUN1},  // S_SPOS_PAIN2
  {SPR_SPOS,7,5,NULL,S_SPOS_DIE2},  // S_SPOS_DIE1
  {SPR_SPOS,8,5,A_Scream,S_SPOS_DIE3},  // S_SPOS_DIE2
  {SPR_SPOS,9,5,A_Fall,S_SPOS_DIE4},  // S_SPOS_DIE3
  {SPR_SPOS,10,5,NULL,S_SPOS_DIE5}, // S_SPOS_DIE4
  {SPR_SPOS,11,-1,NULL,S_NULL}, // S_SPOS_DIE5
  {SPR_SPOS,12,5,NULL,S_SPOS_XDIE2},  // S_SPOS_XDIE1
  {SPR_SPOS,13,5,A_XScream,S_SPOS_XDIE3}, // S_SPOS_XDIE2
  {SPR_SPOS,14,5,A_Fall,S_SPOS_XDIE4},  // S_SPOS_XDIE3
  {SPR_SPOS,15,5,NULL,S_SPOS_XDIE5},  // S_SPOS_XDIE4
  {SPR_SPOS,16,5,NULL,S_SPOS_XDIE6},  // S_SPOS_XDIE5
  {SPR_SPOS,17,5,NULL,S_SPOS_XDIE7},  // S_SPOS_XDIE6
  {SPR_SPOS,18,5,NULL,S_SPOS_XDIE8},  // S_SPOS_XDIE7
  {SPR_SPOS,19,5,NULL,S_SPOS_XDIE9},  // S_SPOS_XDIE8
  {SPR_SPOS,20,-1,NULL,S_NULL}, // S_SPOS_XDIE9
  {SPR_SPOS,11,5,NULL,S_SPOS_RAISE2}, // S_SPOS_RAISE1
  {SPR_SPOS,10,5,NULL,S_SPOS_RAISE3}, // S_SPOS_RAISE2
  {SPR_SPOS,9,5,NULL,S_SPOS_RAISE4},  // S_SPOS_RAISE3
  {SPR_SPOS,8,5,NULL,S_SPOS_RAISE5},  // S_SPOS_RAISE4
  {SPR_SPOS,7,5,NULL,S_SPOS_RUN1},  // S_SPOS_RAISE5
  {SPR_VILE,0,10,A_Look,S_VILE_STND2},  // S_VILE_STND
  {SPR_VILE,1,10,A_Look,S_VILE_STND}, // S_VILE_STND2
  {SPR_VILE,0,2,A_VileChase,S_VILE_RUN2}, // S_VILE_RUN1
  {SPR_VILE,0,2,A_VileChase,S_VILE_RUN3}, // S_VILE_RUN2
  {SPR_VILE,1,2,A_VileChase,S_VILE_RUN4}, // S_VILE_RUN3
  {SPR_VILE,1,2,A_VileChase,S_VILE_RUN5}, // S_VILE_RUN4
  {SPR_VILE,2,2,A_VileChase,S_VILE_RUN6}, // S_VILE_RUN5
  {SPR_VILE,2,2,A_VileChase,S_VILE_RUN7}, // S_VILE_RUN6
  {SPR_VILE,3,2,A_VileChase,S_VILE_RUN8}, // S_VILE_RUN7
  {SPR_VILE,3,2,A_VileChase,S_VILE_RUN9}, // S_VILE_RUN8
  {SPR_VILE,4,2,A_VileChase,S_VILE_RUN10},  // S_VILE_RUN9
  {SPR_VILE,4,2,A_VileChase,S_VILE_RUN11},  // S_VILE_RUN10
  {SPR_VILE,5,2,A_VileChase,S_VILE_RUN12},  // S_VILE_RUN11
  {SPR_VILE,5,2,A_VileChase,S_VILE_RUN1}, // S_VILE_RUN12
  {SPR_VILE,32774,0,A_VileStart,S_VILE_ATK2}, // S_VILE_ATK1
  {SPR_VILE,32774,10,A_FaceTarget,S_VILE_ATK3}, // S_VILE_ATK2
  {SPR_VILE,32775,8,A_VileTarget,S_VILE_ATK4},  // S_VILE_ATK3
  {SPR_VILE,32776,8,A_FaceTarget,S_VILE_ATK5},  // S_VILE_ATK4
  {SPR_VILE,32777,8,A_FaceTarget,S_VILE_ATK6},  // S_VILE_ATK5
  {SPR_VILE,32778,8,A_FaceTarget,S_VILE_ATK7},  // S_VILE_ATK6
  {SPR_VILE,32779,8,A_FaceTarget,S_VILE_ATK8},  // S_VILE_ATK7
  {SPR_VILE,32780,8,A_FaceTarget,S_VILE_ATK9},  // S_VILE_ATK8
  {SPR_VILE,32781,8,A_FaceTarget,S_VILE_ATK10}, // S_VILE_ATK9
  {SPR_VILE,32782,8,A_VileAttack,S_VILE_ATK11}, // S_VILE_ATK10
  {SPR_VILE,32783,20,NULL,S_VILE_RUN1}, // S_VILE_ATK11
  {SPR_VILE,32794,10,NULL,S_VILE_HEAL2},  // S_VILE_HEAL1
  {SPR_VILE,32795,10,NULL,S_VILE_HEAL3},  // S_VILE_HEAL2
  {SPR_VILE,32796,10,NULL,S_VILE_RUN1}, // S_VILE_HEAL3
  {SPR_VILE,16,5,NULL,S_VILE_PAIN2},  // S_VILE_PAIN
  {SPR_VILE,16,5,A_Pain,S_VILE_RUN1}, // S_VILE_PAIN2
  {SPR_VILE,16,7,NULL,S_VILE_DIE2}, // S_VILE_DIE1
  {SPR_VILE,17,7,A_Scream,S_VILE_DIE3},  // S_VILE_DIE2
  {SPR_VILE,18,7,A_Fall,S_VILE_DIE4}, // S_VILE_DIE3
  {SPR_VILE,19,7,NULL,S_VILE_DIE5}, // S_VILE_DIE4
  {SPR_VILE,20,7,NULL,S_VILE_DIE6}, // S_VILE_DIE5
  {SPR_VILE,21,7,NULL,S_VILE_DIE7}, // S_VILE_DIE6
  {SPR_VILE,22,7,NULL,S_VILE_DIE8}, // S_VILE_DIE7
  {SPR_VILE,23,5,NULL,S_VILE_DIE9}, // S_VILE_DIE8
  {SPR_VILE,24,5,NULL,S_VILE_DIE10},  // S_VILE_DIE9
  {SPR_VILE,25,-1,NULL,S_NULL}, // S_VILE_DIE10
  {SPR_FIRE,32768,2,A_StartFire,S_FIRE2}, // S_FIRE1
  {SPR_FIRE,32769,2,A_Fire,S_FIRE3},  // S_FIRE2
  {SPR_FIRE,32768,2,A_Fire,S_FIRE4},  // S_FIRE3
  {SPR_FIRE,32769,2,A_Fire,S_FIRE5},  // S_FIRE4
  {SPR_FIRE,32770,2,A_FireCrackle,S_FIRE6}, // S_FIRE5
  {SPR_FIRE,32769,2,A_Fire,S_FIRE7},  // S_FIRE6
  {SPR_FIRE,32770,2,A_Fire,S_FIRE8},  // S_FIRE7
  {SPR_FIRE,32769,2,A_Fire,S_FIRE9},  // S_FIRE8
  {SPR_FIRE,32770,2,A_Fire,S_FIRE10}, // S_FIRE9
  {SPR_FIRE,32771,2,A_Fire,S_FIRE11}, // S_FIRE10
  {SPR_FIRE,32770,2,A_Fire,S_FIRE12}, // S_FIRE11
  {SPR_FIRE,32771,2,A_Fire,S_FIRE13}, // S_FIRE12
  {SPR_FIRE,32770,2,A_Fire,S_FIRE14}, // S_FIRE13
  {SPR_FIRE,32771,2,A_Fire,S_FIRE15}, // S_FIRE14
  {SPR_FIRE,32772,2,A_Fire,S_FIRE16}, // S_FIRE15
  {SPR_FIRE,32771,2,A_Fire,S_FIRE17}, // S_FIRE16
  {SPR_FIRE,32772,2,A_Fire,S_FIRE18}, // S_FIRE17
  {SPR_FIRE,32771,2,A_Fire,S_FIRE19}, // S_FIRE18
  {SPR_FIRE,32772,2,A_FireCrackle,S_FIRE20},  // S_FIRE19
  {SPR_FIRE,32773,2,A_Fire,S_FIRE21}, // S_FIRE20
  {SPR_FIRE,32772,2,A_Fire,S_FIRE22}, // S_FIRE21
  {SPR_FIRE,32773,2,A_Fire,S_FIRE23}, // S_FIRE22
  {SPR_FIRE,32772,2,A_Fire,S_FIRE24}, // S_FIRE23
  {SPR_FIRE,32773,2,A_Fire,S_FIRE25}, // S_FIRE24
  {SPR_FIRE,32774,2,A_Fire,S_FIRE26}, // S_FIRE25
  {SPR_FIRE,32775,2,A_Fire,S_FIRE27}, // S_FIRE26
  {SPR_FIRE,32774,2,A_Fire,S_FIRE28}, // S_FIRE27
  {SPR_FIRE,32775,2,A_Fire,S_FIRE29}, // S_FIRE28
  {SPR_FIRE,32774,2,A_Fire,S_FIRE30}, // S_FIRE29
  {SPR_FIRE,32775,2,A_Fire,S_NULL}, // S_FIRE30
  {SPR_PUFF,1,4,NULL,S_SMOKE2}, // S_SMOKE1
  {SPR_PUFF,2,4,NULL,S_SMOKE3}, // S_SMOKE2
  {SPR_PUFF,1,4,NULL,S_SMOKE4}, // S_SMOKE3
  {SPR_PUFF,2,4,NULL,S_SMOKE5}, // S_SMOKE4
  {SPR_PUFF,3,4,NULL,S_NULL}, // S_SMOKE5
  {SPR_FATB,32768,2,A_Tracer,S_TRACER2},  // S_TRACER
  {SPR_FATB,32769,2,A_Tracer,S_TRACER}, // S_TRACER2
  {SPR_FBXP,32768,8,NULL,S_TRACEEXP2},  // S_TRACEEXP1
  {SPR_FBXP,32769,6,NULL,S_TRACEEXP3},  // S_TRACEEXP2
  {SPR_FBXP,32770,4,NULL,S_NULL}, // S_TRACEEXP3
  {SPR_SKEL,0,10,A_Look,S_SKEL_STND2},  // S_SKEL_STND
  {SPR_SKEL,1,10,A_Look,S_SKEL_STND}, // S_SKEL_STND2
  {SPR_SKEL,0,2,A_Chase,S_SKEL_RUN2}, // S_SKEL_RUN1
  {SPR_SKEL,0,2,A_Chase,S_SKEL_RUN3}, // S_SKEL_RUN2
  {SPR_SKEL,1,2,A_Chase,S_SKEL_RUN4}, // S_SKEL_RUN3
  {SPR_SKEL,1,2,A_Chase,S_SKEL_RUN5}, // S_SKEL_RUN4
  {SPR_SKEL,2,2,A_Chase,S_SKEL_RUN6}, // S_SKEL_RUN5
  {SPR_SKEL,2,2,A_Chase,S_SKEL_RUN7}, // S_SKEL_RUN6
  {SPR_SKEL,3,2,A_Chase,S_SKEL_RUN8}, // S_SKEL_RUN7
  {SPR_SKEL,3,2,A_Chase,S_SKEL_RUN9}, // S_SKEL_RUN8
  {SPR_SKEL,4,2,A_Chase,S_SKEL_RUN10},  // S_SKEL_RUN9
  {SPR_SKEL,4,2,A_Chase,S_SKEL_RUN11},  // S_SKEL_RUN10
  {SPR_SKEL,5,2,A_Chase,S_SKEL_RUN12},  // S_SKEL_RUN11
  {SPR_SKEL,5,2,A_Chase,S_SKEL_RUN1}, // S_SKEL_RUN12
  {SPR_SKEL,6,0,A_FaceTarget,S_SKEL_FIST2}, // S_SKEL_FIST1
  {SPR_SKEL,6,6,A_SkelWhoosh,S_SKEL_FIST3}, // S_SKEL_FIST2
  {SPR_SKEL,7,6,A_FaceTarget,S_SKEL_FIST4}, // S_SKEL_FIST3
  {SPR_SKEL,8,6,A_SkelFist,S_SKEL_RUN1},  // S_SKEL_FIST4
  {SPR_SKEL,32777,0,A_FaceTarget,S_SKEL_MISS2}, // S_SKEL_MISS1
  {SPR_SKEL,32777,10,A_FaceTarget,S_SKEL_MISS3},  // S_SKEL_MISS2
  {SPR_SKEL,10,10,A_SkelMissile,S_SKEL_MISS4},  // S_SKEL_MISS3
  {SPR_SKEL,10,10,A_FaceTarget,S_SKEL_RUN1},  // S_SKEL_MISS4
  {SPR_SKEL,11,5,NULL,S_SKEL_PAIN2},  // S_SKEL_PAIN
  {SPR_SKEL,11,5,A_Pain,S_SKEL_RUN1}, // S_SKEL_PAIN2
  {SPR_SKEL,11,7,NULL,S_SKEL_DIE2}, // S_SKEL_DIE1
  {SPR_SKEL,12,7,NULL,S_SKEL_DIE3}, // S_SKEL_DIE2
  {SPR_SKEL,13,7,A_Scream,S_SKEL_DIE4}, // S_SKEL_DIE3
  {SPR_SKEL,14,7,A_Fall,S_SKEL_DIE5}, // S_SKEL_DIE4
  {SPR_SKEL,15,7,NULL,S_SKEL_DIE6}, // S_SKEL_DIE5
  {SPR_SKEL,16,-1,NULL,S_NULL}, // S_SKEL_DIE6
  {SPR_SKEL,16,5,NULL,S_SKEL_RAISE2}, // S_SKEL_RAISE1
  {SPR_SKEL,15,5,NULL,S_SKEL_RAISE3}, // S_SKEL_RAISE2
  {SPR_SKEL,14,5,NULL,S_SKEL_RAISE4}, // S_SKEL_RAISE3
  {SPR_SKEL,13,5,NULL,S_SKEL_RAISE5}, // S_SKEL_RAISE4
  {SPR_SKEL,12,5,NULL,S_SKEL_RAISE6}, // S_SKEL_RAISE5
  {SPR_SKEL,11,5,NULL,S_SKEL_RUN1}, // S_SKEL_RAISE6
  {SPR_MANF,32768,4,NULL,S_FATSHOT2}, // S_FATSHOT1
  {SPR_MANF,32769,4,NULL,S_FATSHOT1}, // S_FATSHOT2
  {SPR_MISL,32769,8,NULL,S_FATSHOTX2},  // S_FATSHOTX1
  {SPR_MISL,32770,6,NULL,S_FATSHOTX3},  // S_FATSHOTX2
  {SPR_MISL,32771,4,NULL,S_NULL}, // S_FATSHOTX3
  {SPR_FATT,0,15,A_Look,S_FATT_STND2},  // S_FATT_STND
  {SPR_FATT,1,15,A_Look,S_FATT_STND}, // S_FATT_STND2
  {SPR_FATT,0,4,A_Chase,S_FATT_RUN2}, // S_FATT_RUN1
  {SPR_FATT,0,4,A_Chase,S_FATT_RUN3}, // S_FATT_RUN2
  {SPR_FATT,1,4,A_Chase,S_FATT_RUN4}, // S_FATT_RUN3
  {SPR_FATT,1,4,A_Chase,S_FATT_RUN5}, // S_FATT_RUN4
  {SPR_FATT,2,4,A_Chase,S_FATT_RUN6}, // S_FATT_RUN5
  {SPR_FATT,2,4,A_Chase,S_FATT_RUN7}, // S_FATT_RUN6
  {SPR_FATT,3,4,A_Chase,S_FATT_RUN8}, // S_FATT_RUN7
  {SPR_FATT,3,4,A_Chase,S_FATT_RUN9}, // S_FATT_RUN8
  {SPR_FATT,4,4,A_Chase,S_FATT_RUN10},  // S_FATT_RUN9
  {SPR_FATT,4,4,A_Chase,S_FATT_RUN11},  // S_FATT_RUN10
  {SPR_FATT,5,4,A_Chase,S_FATT_RUN12},  // S_FATT_RUN11
  {SPR_FATT,5,4,A_Chase,S_FATT_RUN1}, // S_FATT_RUN12
  {SPR_FATT,6,20,A_FatRaise,S_FATT_ATK2}, // S_FATT_ATK1
  {SPR_FATT,32775,10,A_FatAttack1,S_FATT_ATK3}, // S_FATT_ATK2
  {SPR_FATT,8,5,A_FaceTarget,S_FATT_ATK4},  // S_FATT_ATK3
  {SPR_FATT,6,5,A_FaceTarget,S_FATT_ATK5},  // S_FATT_ATK4
  {SPR_FATT,32775,10,A_FatAttack2,S_FATT_ATK6}, // S_FATT_ATK5
  {SPR_FATT,8,5,A_FaceTarget,S_FATT_ATK7},  // S_FATT_ATK6
  {SPR_FATT,6,5,A_FaceTarget,S_FATT_ATK8},  // S_FATT_ATK7
  {SPR_FATT,32775,10,A_FatAttack3,S_FATT_ATK9}, // S_FATT_ATK8
  {SPR_FATT,8,5,A_FaceTarget,S_FATT_ATK10}, // S_FATT_ATK9
  {SPR_FATT,6,5,A_FaceTarget,S_FATT_RUN1},  // S_FATT_ATK10
  {SPR_FATT,9,3,NULL,S_FATT_PAIN2}, // S_FATT_PAIN
  {SPR_FATT,9,3,A_Pain,S_FATT_RUN1},  // S_FATT_PAIN2
  {SPR_FATT,10,6,NULL,S_FATT_DIE2}, // S_FATT_DIE1
  {SPR_FATT,11,6,A_Scream,S_FATT_DIE3}, // S_FATT_DIE2
  {SPR_FATT,12,6,A_Fall,S_FATT_DIE4}, // S_FATT_DIE3
  {SPR_FATT,13,6,NULL,S_FATT_DIE5}, // S_FATT_DIE4
  {SPR_FATT,14,6,NULL,S_FATT_DIE6}, // S_FATT_DIE5
  {SPR_FATT,15,6,NULL,S_FATT_DIE7}, // S_FATT_DIE6
  {SPR_FATT,16,6,NULL,S_FATT_DIE8}, // S_FATT_DIE7
  {SPR_FATT,17,6,NULL,S_FATT_DIE9}, // S_FATT_DIE8
  {SPR_FATT,18,6,NULL,S_FATT_DIE10},  // S_FATT_DIE9
  {SPR_FATT,19,-1,A_BossDeath,S_NULL},  // S_FATT_DIE10
  {SPR_FATT,17,5,NULL,S_FATT_RAISE2}, // S_FATT_RAISE1
  {SPR_FATT,16,5,NULL,S_FATT_RAISE3}, // S_FATT_RAISE2
  {SPR_FATT,15,5,NULL,S_FATT_RAISE4}, // S_FATT_RAISE3
  {SPR_FATT,14,5,NULL,S_FATT_RAISE5}, // S_FATT_RAISE4
  {SPR_FATT,13,5,NULL,S_FATT_RAISE6}, // S_FATT_RAISE5
  {SPR_FATT,12,5,NULL,S_FATT_RAISE7}, // S_FATT_RAISE6
  {SPR_FATT,11,5,NULL,S_FATT_RAISE8}, // S_FATT_RAISE7
  {SPR_FATT,10,5,NULL,S_FATT_RUN1}, // S_FATT_RAISE8
  {SPR_CPOS,0,10,A_Look,S_CPOS_STND2},  // S_CPOS_STND
  {SPR_CPOS,1,10,A_Look,S_CPOS_STND}, // S_CPOS_STND2
  {SPR_CPOS,0,3,A_Chase,S_CPOS_RUN2}, // S_CPOS_RUN1
  {SPR_CPOS,0,3,A_Chase,S_CPOS_RUN3}, // S_CPOS_RUN2
  {SPR_CPOS,1,3,A_Chase,S_CPOS_RUN4}, // S_CPOS_RUN3
  {SPR_CPOS,1,3,A_Chase,S_CPOS_RUN5}, // S_CPOS_RUN4
  {SPR_CPOS,2,3,A_Chase,S_CPOS_RUN6}, // S_CPOS_RUN5
  {SPR_CPOS,2,3,A_Chase,S_CPOS_RUN7}, // S_CPOS_RUN6
  {SPR_CPOS,3,3,A_Chase,S_CPOS_RUN8}, // S_CPOS_RUN7
  {SPR_CPOS,3,3,A_Chase,S_CPOS_RUN1}, // S_CPOS_RUN8
  {SPR_CPOS,4,10,A_FaceTarget,S_CPOS_ATK2}, // S_CPOS_ATK1
  {SPR_CPOS,32773,4,A_CPosAttack,S_CPOS_ATK3},  // S_CPOS_ATK2
  {SPR_CPOS,32772,4,A_CPosAttack,S_CPOS_ATK4},  // S_CPOS_ATK3
  {SPR_CPOS,5,1,A_CPosRefire,S_CPOS_ATK2},  // S_CPOS_ATK4
  {SPR_CPOS,6,3,NULL,S_CPOS_PAIN2}, // S_CPOS_PAIN
  {SPR_CPOS,6,3,A_Pain,S_CPOS_RUN1},  // S_CPOS_PAIN2
  {SPR_CPOS,7,5,NULL,S_CPOS_DIE2},  // S_CPOS_DIE1
  {SPR_CPOS,8,5,A_Scream,S_CPOS_DIE3},  // S_CPOS_DIE2
  {SPR_CPOS,9,5,A_Fall,S_CPOS_DIE4},  // S_CPOS_DIE3
  {SPR_CPOS,10,5,NULL,S_CPOS_DIE5}, // S_CPOS_DIE4
  {SPR_CPOS,11,5,NULL,S_CPOS_DIE6}, // S_CPOS_DIE5
  {SPR_CPOS,12,5,NULL,S_CPOS_DIE7}, // S_CPOS_DIE6
  {SPR_CPOS,13,-1,NULL,S_NULL}, // S_CPOS_DIE7
  {SPR_CPOS,14,5,NULL,S_CPOS_XDIE2},  // S_CPOS_XDIE1
  {SPR_CPOS,15,5,A_XScream,S_CPOS_XDIE3}, // S_CPOS_XDIE2
  {SPR_CPOS,16,5,A_Fall,S_CPOS_XDIE4},  // S_CPOS_XDIE3
  {SPR_CPOS,17,5,NULL,S_CPOS_XDIE5},  // S_CPOS_XDIE4
  {SPR_CPOS,18,5,NULL,S_CPOS_XDIE6},  // S_CPOS_XDIE5
  {SPR_CPOS,19,-1,NULL,S_NULL}, // S_CPOS_XDIE6
  {SPR_CPOS,13,5,NULL,S_CPOS_RAISE2}, // S_CPOS_RAISE1
  {SPR_CPOS,12,5,NULL,S_CPOS_RAISE3}, // S_CPOS_RAISE2
  {SPR_CPOS,11,5,NULL,S_CPOS_RAISE4}, // S_CPOS_RAISE3
  {SPR_CPOS,10,5,NULL,S_CPOS_RAISE5}, // S_CPOS_RAISE4
  {SPR_CPOS,9,5,NULL,S_CPOS_RAISE6},  // S_CPOS_RAISE5
  {SPR_CPOS,8,5,NULL,S_CPOS_RAISE7},  // S_CPOS_RAISE6
  {SPR_CPOS,7,5,NULL,S_CPOS_RUN1},  // S_CPOS_RAISE7
  {SPR_TROO,0,10,A_Look,S_TROO_STND2},  // S_TROO_STND
  {SPR_TROO,1,10,A_Look,S_TROO_STND}, // S_TROO_STND2
  {SPR_TROO,0,3,A_Chase,S_TROO_RUN2}, // S_TROO_RUN1
  {SPR_TROO,0,3,A_Chase,S_TROO_RUN3}, // S_TROO_RUN2
  {SPR_TROO,1,3,A_Chase,S_TROO_RUN4}, // S_TROO_RUN3
  {SPR_TROO,1,3,A_Chase,S_TROO_RUN5}, // S_TROO_RUN4
  {SPR_TROO,2,3,A_Chase,S_TROO_RUN6}, // S_TROO_RUN5
  {SPR_TROO,2,3,A_Chase,S_TROO_RUN7}, // S_TROO_RUN6
  {SPR_TROO,3,3,A_Chase,S_TROO_RUN8}, // S_TROO_RUN7
  {SPR_TROO,3,3,A_Chase,S_TROO_RUN1}, // S_TROO_RUN8
  {SPR_TROO,4,8,A_FaceTarget,S_TROO_ATK2},  // S_TROO_ATK1
  {SPR_TROO,5,8,A_FaceTarget,S_TROO_ATK3},  // S_TROO_ATK2
  {SPR_TROO,6,6,A_TroopAttack,S_TROO_RUN1}, // S_TROO_ATK3
  {SPR_TROO,7,2,NULL,S_TROO_PAIN2}, // S_TROO_PAIN
  {SPR_TROO,7,2,A_Pain,S_TROO_RUN1},  // S_TROO_PAIN2
  {SPR_TROO,8,8,NULL,S_TROO_DIE2},  // S_TROO_DIE1
  {SPR_TROO,9,8,A_Scream,S_TROO_DIE3},  // S_TROO_DIE2
  {SPR_TROO,10,6,NULL,S_TROO_DIE4}, // S_TROO_DIE3
  {SPR_TROO,11,6,A_Fall,S_TROO_DIE5}, // S_TROO_DIE4
  {SPR_TROO,12,-1,NULL,S_NULL}, // S_TROO_DIE5
  {SPR_TROO,13,5,NULL,S_TROO_XDIE2},  // S_TROO_XDIE1
  {SPR_TROO,14,5,A_XScream,S_TROO_XDIE3}, // S_TROO_XDIE2
  {SPR_TROO,15,5,NULL,S_TROO_XDIE4},  // S_TROO_XDIE3
  {SPR_TROO,16,5,A_Fall,S_TROO_XDIE5},  // S_TROO_XDIE4
  {SPR_TROO,17,5,NULL,S_TROO_XDIE6},  // S_TROO_XDIE5
  {SPR_TROO,18,5,NULL,S_TROO_XDIE7},  // S_TROO_XDIE6
  {SPR_TROO,19,5,NULL,S_TROO_XDIE8},  // S_TROO_XDIE7
  {SPR_TROO,20,-1,NULL,S_NULL}, // S_TROO_XDIE8
  {SPR_TROO,12,8,NULL,S_TROO_RAISE2}, // S_TROO_RAISE1
  {SPR_TROO,11,8,NULL,S_TROO_RAISE3}, // S_TROO_RAISE2
  {SPR_TROO,10,6,NULL,S_TROO_RAISE4}, // S_TROO_RAISE3
  {SPR_TROO,9,6,NULL,S_TROO_RAISE5},  // S_TROO_RAISE4
  {SPR_TROO,8,6,NULL,S_TROO_RUN1},  // S_TROO_RAISE5
  {SPR_SARG,0,10,A_Look,S_SARG_STND2},  // S_SARG_STND
  {SPR_SARG,1,10,A_Look,S_SARG_STND}, // S_SARG_STND2
  {SPR_SARG,0,2,A_Chase,S_SARG_RUN2}, // S_SARG_RUN1
  {SPR_SARG,0,2,A_Chase,S_SARG_RUN3}, // S_SARG_RUN2
  {SPR_SARG,1,2,A_Chase,S_SARG_RUN4}, // S_SARG_RUN3
  {SPR_SARG,1,2,A_Chase,S_SARG_RUN5}, // S_SARG_RUN4
  {SPR_SARG,2,2,A_Chase,S_SARG_RUN6}, // S_SARG_RUN5
  {SPR_SARG,2,2,A_Chase,S_SARG_RUN7}, // S_SARG_RUN6
  {SPR_SARG,3,2,A_Chase,S_SARG_RUN8}, // S_SARG_RUN7
  {SPR_SARG,3,2,A_Chase,S_SARG_RUN1}, // S_SARG_RUN8
  {SPR_SARG,4,8,A_FaceTarget,S_SARG_ATK2},  // S_SARG_ATK1
  {SPR_SARG,5,8,A_FaceTarget,S_SARG_ATK3},  // S_SARG_ATK2
  {SPR_SARG,6,8,A_SargAttack,S_SARG_RUN1},  // S_SARG_ATK3
  {SPR_SARG,7,2,NULL,S_SARG_PAIN2}, // S_SARG_PAIN
  {SPR_SARG,7,2,A_Pain,S_SARG_RUN1},  // S_SARG_PAIN2
  {SPR_SARG,8,8,NULL,S_SARG_DIE2},  // S_SARG_DIE1
  {SPR_SARG,9,8,A_Scream,S_SARG_DIE3},  // S_SARG_DIE2
  {SPR_SARG,10,4,NULL,S_SARG_DIE4}, // S_SARG_DIE3
  {SPR_SARG,11,4,A_Fall,S_SARG_DIE5}, // S_SARG_DIE4
  {SPR_SARG,12,4,NULL,S_SARG_DIE6}, // S_SARG_DIE5
  {SPR_SARG,13,-1,NULL,S_NULL}, // S_SARG_DIE6
  {SPR_SARG,13,5,NULL,S_SARG_RAISE2}, // S_SARG_RAISE1
  {SPR_SARG,12,5,NULL,S_SARG_RAISE3}, // S_SARG_RAISE2
  {SPR_SARG,11,5,NULL,S_SARG_RAISE4}, // S_SARG_RAISE3
  {SPR_SARG,10,5,NULL,S_SARG_RAISE5}, // S_SARG_RAISE4
  {SPR_SARG,9,5,NULL,S_SARG_RAISE6},  // S_SARG_RAISE5
  {SPR_SARG,8,5,NULL,S_SARG_RUN1},  // S_SARG_RAISE6
  {SPR_HEAD,0,10,A_Look,S_HEAD_STND}, // S_HEAD_STND
  {SPR_HEAD,0,3,A_Chase,S_HEAD_RUN1}, // S_HEAD_RUN1
  {SPR_HEAD,1,5,A_FaceTarget,S_HEAD_ATK2},  // S_HEAD_ATK1
  {SPR_HEAD,2,5,A_FaceTarget,S_HEAD_ATK3},  // S_HEAD_ATK2
  {SPR_HEAD,32771,5,A_HeadAttack,S_HEAD_RUN1},  // S_HEAD_ATK3
  {SPR_HEAD,4,3,NULL,S_HEAD_PAIN2}, // S_HEAD_PAIN
  {SPR_HEAD,4,3,A_Pain,S_HEAD_PAIN3}, // S_HEAD_PAIN2
  {SPR_HEAD,5,6,NULL,S_HEAD_RUN1},  // S_HEAD_PAIN3
  {SPR_HEAD,6,8,NULL,S_HEAD_DIE2},  // S_HEAD_DIE1
  {SPR_HEAD,7,8,A_Scream,S_HEAD_DIE3},  // S_HEAD_DIE2
  {SPR_HEAD,8,8,NULL,S_HEAD_DIE4},  // S_HEAD_DIE3
  {SPR_HEAD,9,8,NULL,S_HEAD_DIE5},  // S_HEAD_DIE4
  {SPR_HEAD,10,8,A_Fall,S_HEAD_DIE6}, // S_HEAD_DIE5
  {SPR_HEAD,11,-1,NULL,S_NULL}, // S_HEAD_DIE6
  {SPR_HEAD,11,8,NULL,S_HEAD_RAISE2}, // S_HEAD_RAISE1
  {SPR_HEAD,10,8,NULL,S_HEAD_RAISE3}, // S_HEAD_RAISE2
  {SPR_HEAD,9,8,NULL,S_HEAD_RAISE4},  // S_HEAD_RAISE3
  {SPR_HEAD,8,8,NULL,S_HEAD_RAISE5},  // S_HEAD_RAISE4
  {SPR_HEAD,7,8,NULL,S_HEAD_RAISE6},  // S_HEAD_RAISE5
  {SPR_HEAD,6,8,NULL,S_HEAD_RUN1},  // S_HEAD_RAISE6
  {SPR_BAL7,32768,4,NULL,S_BRBALL2},  // S_BRBALL1
  {SPR_BAL7,32769,4,NULL,S_BRBALL1},  // S_BRBALL2
  {SPR_BAL7,32770,6,NULL,S_BRBALLX2}, // S_BRBALLX1
  {SPR_BAL7,32771,6,NULL,S_BRBALLX3}, // S_BRBALLX2
  {SPR_BAL7,32772,6,NULL,S_NULL}, // S_BRBALLX3
  {SPR_BOSS,0,10,A_Look,S_BOSS_STND2},  // S_BOSS_STND
  {SPR_BOSS,1,10,A_Look,S_BOSS_STND}, // S_BOSS_STND2
  {SPR_BOSS,0,3,A_Chase,S_BOSS_RUN2}, // S_BOSS_RUN1
  {SPR_BOSS,0,3,A_Chase,S_BOSS_RUN3}, // S_BOSS_RUN2
  {SPR_BOSS,1,3,A_Chase,S_BOSS_RUN4}, // S_BOSS_RUN3
  {SPR_BOSS,1,3,A_Chase,S_BOSS_RUN5}, // S_BOSS_RUN4
  {SPR_BOSS,2,3,A_Chase,S_BOSS_RUN6}, // S_BOSS_RUN5
  {SPR_BOSS,2,3,A_Chase,S_BOSS_RUN7}, // S_BOSS_RUN6
  {SPR_BOSS,3,3,A_Chase,S_BOSS_RUN8}, // S_BOSS_RUN7
  {SPR_BOSS,3,3,A_Chase,S_BOSS_RUN1}, // S_BOSS_RUN8
  {SPR_BOSS,4,8,A_FaceTarget,S_BOSS_ATK2},  // S_BOSS_ATK1
  {SPR_BOSS,5,8,A_FaceTarget,S_BOSS_ATK3},  // S_BOSS_ATK2
  {SPR_BOSS,6,8,A_BruisAttack,S_BOSS_RUN1}, // S_BOSS_ATK3
  {SPR_BOSS,7,2,NULL,S_BOSS_PAIN2}, // S_BOSS_PAIN
  {SPR_BOSS,7,2,A_Pain,S_BOSS_RUN1},  // S_BOSS_PAIN2
  {SPR_BOSS,8,8,NULL,S_BOSS_DIE2},  // S_BOSS_DIE1
  {SPR_BOSS,9,8,A_Scream,S_BOSS_DIE3},  // S_BOSS_DIE2
  {SPR_BOSS,10,8,NULL,S_BOSS_DIE4}, // S_BOSS_DIE3
  {SPR_BOSS,11,8,A_Fall,S_BOSS_DIE5}, // S_BOSS_DIE4
  {SPR_BOSS,12,8,NULL,S_BOSS_DIE6}, // S_BOSS_DIE5
  {SPR_BOSS,13,8,NULL,S_BOSS_DIE7}, // S_BOSS_DIE6
  {SPR_BOSS,14,-1,A_BossDeath,S_NULL},  // S_BOSS_DIE7
  {SPR_BOSS,14,8,NULL,S_BOSS_RAISE2}, // S_BOSS_RAISE1
  {SPR_BOSS,13,8,NULL,S_BOSS_RAISE3}, // S_BOSS_RAISE2
  {SPR_BOSS,12,8,NULL,S_BOSS_RAISE4}, // S_BOSS_RAISE3
  {SPR_BOSS,11,8,NULL,S_BOSS_RAISE5}, // S_BOSS_RAISE4
  {SPR_BOSS,10,8,NULL,S_BOSS_RAISE6}, // S_BOSS_RAISE5
  {SPR_BOSS,9,8,NULL,S_BOSS_RAISE7},  // S_BOSS_RAISE6
  {SPR_BOSS,8,8,NULL,S_BOSS_RUN1},  // S_BOSS_RAISE7
  {SPR_BOS2,0,10,A_Look,S_BOS2_STND2},  // S_BOS2_STND
  {SPR_BOS2,1,10,A_Look,S_BOS2_STND}, // S_BOS2_STND2
  {SPR_BOS2,0,3,A_Chase,S_BOS2_RUN2}, // S_BOS2_RUN1
  {SPR_BOS2,0,3,A_Chase,S_BOS2_RUN3}, // S_BOS2_RUN2
  {SPR_BOS2,1,3,A_Chase,S_BOS2_RUN4}, // S_BOS2_RUN3
  {SPR_BOS2,1,3,A_Chase,S_BOS2_RUN5}, // S_BOS2_RUN4
  {SPR_BOS2,2,3,A_Chase,S_BOS2_RUN6}, // S_BOS2_RUN5
  {SPR_BOS2,2,3,A_Chase,S_BOS2_RUN7}, // S_BOS2_RUN6
  {SPR_BOS2,3,3,A_Chase,S_BOS2_RUN8}, // S_BOS2_RUN7
  {SPR_BOS2,3,3,A_Chase,S_BOS2_RUN1}, // S_BOS2_RUN8
  {SPR_BOS2,4,8,A_FaceTarget,S_BOS2_ATK2},  // S_BOS2_ATK1
  {SPR_BOS2,5,8,A_FaceTarget,S_BOS2_ATK3},  // S_BOS2_ATK2
  {SPR_BOS2,6,8,A_BruisAttack,S_BOS2_RUN1}, // S_BOS2_ATK3
  {SPR_BOS2,7,2,NULL,S_BOS2_PAIN2}, // S_BOS2_PAIN
  {SPR_BOS2,7,2,A_Pain,S_BOS2_RUN1},  // S_BOS2_PAIN2
  {SPR_BOS2,8,8,NULL,S_BOS2_DIE2},  // S_BOS2_DIE1
  {SPR_BOS2,9,8,A_Scream,S_BOS2_DIE3},  // S_BOS2_DIE2
  {SPR_BOS2,10,8,NULL,S_BOS2_DIE4}, // S_BOS2_DIE3
  {SPR_BOS2,11,8,A_Fall,S_BOS2_DIE5}, // S_BOS2_DIE4
  {SPR_BOS2,12,8,NULL,S_BOS2_DIE6}, // S_BOS2_DIE5
  {SPR_BOS2,13,8,NULL,S_BOS2_DIE7}, // S_BOS2_DIE6
  {SPR_BOS2,14,-1,NULL,S_NULL}, // S_BOS2_DIE7
  {SPR_BOS2,14,8,NULL,S_BOS2_RAISE2}, // S_BOS2_RAISE1
  {SPR_BOS2,13,8,NULL,S_BOS2_RAISE3}, // S_BOS2_RAISE2
  {SPR_BOS2,12,8,NULL,S_BOS2_RAISE4}, // S_BOS2_RAISE3
  {SPR_BOS2,11,8,NULL,S_BOS2_RAISE5}, // S_BOS2_RAISE4
  {SPR_BOS2,10,8,NULL,S_BOS2_RAISE6}, // S_BOS2_RAISE5
  {SPR_BOS2,9,8,NULL,S_BOS2_RAISE7},  // S_BOS2_RAISE6
  {SPR_BOS2,8,8,NULL,S_BOS2_RUN1},  // S_BOS2_RAISE7
  {SPR_SKUL,32768,10,A_Look,S_SKULL_STND2}, // S_SKULL_STND
  {SPR_SKUL,32769,10,A_Look,S_SKULL_STND},  // S_SKULL_STND2
  {SPR_SKUL,32768,6,A_Chase,S_SKULL_RUN2},  // S_SKULL_RUN1
  {SPR_SKUL,32769,6,A_Chase,S_SKULL_RUN1},  // S_SKULL_RUN2
  {SPR_SKUL,32770,10,A_FaceTarget,S_SKULL_ATK2},  // S_SKULL_ATK1
  {SPR_SKUL,32771,4,A_SkullAttack,S_SKULL_ATK3},  // S_SKULL_ATK2
  {SPR_SKUL,32770,4,NULL,S_SKULL_ATK4}, // S_SKULL_ATK3
  {SPR_SKUL,32771,4,NULL,S_SKULL_ATK3}, // S_SKULL_ATK4
  {SPR_SKUL,32772,3,NULL,S_SKULL_PAIN2},  // S_SKULL_PAIN
  {SPR_SKUL,32772,3,A_Pain,S_SKULL_RUN1}, // S_SKULL_PAIN2
  {SPR_SKUL,32773,6,NULL,S_SKULL_DIE2}, // S_SKULL_DIE1
  {SPR_SKUL,32774,6,A_Scream,S_SKULL_DIE3}, // S_SKULL_DIE2
  {SPR_SKUL,32775,6,NULL,S_SKULL_DIE4}, // S_SKULL_DIE3
  {SPR_SKUL,32776,6,A_Fall,S_SKULL_DIE5}, // S_SKULL_DIE4
  {SPR_SKUL,9,6,NULL,S_SKULL_DIE6}, // S_SKULL_DIE5
  {SPR_SKUL,10,6,NULL,S_NULL},  // S_SKULL_DIE6
  {SPR_SPID,0,10,A_Look,S_SPID_STND2},  // S_SPID_STND
  {SPR_SPID,1,10,A_Look,S_SPID_STND}, // S_SPID_STND2
  {SPR_SPID,0,3,A_Metal,S_SPID_RUN2}, // S_SPID_RUN1
  {SPR_SPID,0,3,A_Chase,S_SPID_RUN3}, // S_SPID_RUN2
  {SPR_SPID,1,3,A_Chase,S_SPID_RUN4}, // S_SPID_RUN3
  {SPR_SPID,1,3,A_Chase,S_SPID_RUN5}, // S_SPID_RUN4
  {SPR_SPID,2,3,A_Metal,S_SPID_RUN6}, // S_SPID_RUN5
  {SPR_SPID,2,3,A_Chase,S_SPID_RUN7}, // S_SPID_RUN6
  {SPR_SPID,3,3,A_Chase,S_SPID_RUN8}, // S_SPID_RUN7
  {SPR_SPID,3,3,A_Chase,S_SPID_RUN9}, // S_SPID_RUN8
  {SPR_SPID,4,3,A_Metal,S_SPID_RUN10},  // S_SPID_RUN9
  {SPR_SPID,4,3,A_Chase,S_SPID_RUN11},  // S_SPID_RUN10
  {SPR_SPID,5,3,A_Chase,S_SPID_RUN12},  // S_SPID_RUN11
  {SPR_SPID,5,3,A_Chase,S_SPID_RUN1}, // S_SPID_RUN12
  {SPR_SPID,32768,20,A_FaceTarget,S_SPID_ATK2}, // S_SPID_ATK1
  {SPR_SPID,32774,4,A_SPosAttack,S_SPID_ATK3},  // S_SPID_ATK2
  {SPR_SPID,32775,4,A_SPosAttack,S_SPID_ATK4},  // S_SPID_ATK3
  {SPR_SPID,32775,1,A_SpidRefire,S_SPID_ATK2},  // S_SPID_ATK4
  {SPR_SPID,8,3,NULL,S_SPID_PAIN2}, // S_SPID_PAIN
  {SPR_SPID,8,3,A_Pain,S_SPID_RUN1},  // S_SPID_PAIN2
  {SPR_SPID,9,20,A_Scream,S_SPID_DIE2}, // S_SPID_DIE1
  {SPR_SPID,10,10,A_Fall,S_SPID_DIE3},  // S_SPID_DIE2
  {SPR_SPID,11,10,NULL,S_SPID_DIE4},  // S_SPID_DIE3
  {SPR_SPID,12,10,NULL,S_SPID_DIE5},  // S_SPID_DIE4
  {SPR_SPID,13,10,NULL,S_SPID_DIE6},  // S_SPID_DIE5
  {SPR_SPID,14,10,NULL,S_SPID_DIE7},  // S_SPID_DIE6
  {SPR_SPID,15,10,NULL,S_SPID_DIE8},  // S_SPID_DIE7
  {SPR_SPID,16,10,NULL,S_SPID_DIE9},  // S_SPID_DIE8
  {SPR_SPID,17,10,NULL,S_SPID_DIE10}, // S_SPID_DIE9
  {SPR_SPID,18,30,NULL,S_SPID_DIE11}, // S_SPID_DIE10
  {SPR_SPID,18,-1,A_BossDeath,S_NULL},  // S_SPID_DIE11
  {SPR_BSPI,0,10,A_Look,S_BSPI_STND2},  // S_BSPI_STND
  {SPR_BSPI,1,10,A_Look,S_BSPI_STND}, // S_BSPI_STND2
  {SPR_BSPI,0,20,NULL,S_BSPI_RUN1}, // S_BSPI_SIGHT
  {SPR_BSPI,0,3,A_BabyMetal,S_BSPI_RUN2}, // S_BSPI_RUN1
  {SPR_BSPI,0,3,A_Chase,S_BSPI_RUN3}, // S_BSPI_RUN2
  {SPR_BSPI,1,3,A_Chase,S_BSPI_RUN4}, // S_BSPI_RUN3
  {SPR_BSPI,1,3,A_Chase,S_BSPI_RUN5}, // S_BSPI_RUN4
  {SPR_BSPI,2,3,A_Chase,S_BSPI_RUN6}, // S_BSPI_RUN5
  {SPR_BSPI,2,3,A_Chase,S_BSPI_RUN7}, // S_BSPI_RUN6
  {SPR_BSPI,3,3,A_BabyMetal,S_BSPI_RUN8}, // S_BSPI_RUN7
  {SPR_BSPI,3,3,A_Chase,S_BSPI_RUN9}, // S_BSPI_RUN8
  {SPR_BSPI,4,3,A_Chase,S_BSPI_RUN10},  // S_BSPI_RUN9
  {SPR_BSPI,4,3,A_Chase,S_BSPI_RUN11},  // S_BSPI_RUN10
  {SPR_BSPI,5,3,A_Chase,S_BSPI_RUN12},  // S_BSPI_RUN11
  {SPR_BSPI,5,3,A_Chase,S_BSPI_RUN1}, // S_BSPI_RUN12
  {SPR_BSPI,32768,20,A_FaceTarget,S_BSPI_ATK2}, // S_BSPI_ATK1
  {SPR_BSPI,32774,4,A_BspiAttack,S_BSPI_ATK3},  // S_BSPI_ATK2
  {SPR_BSPI,32775,4,NULL,S_BSPI_ATK4},  // S_BSPI_ATK3
  {SPR_BSPI,32775,1,A_SpidRefire,S_BSPI_ATK2},  // S_BSPI_ATK4
  {SPR_BSPI,8,3,NULL,S_BSPI_PAIN2}, // S_BSPI_PAIN
  {SPR_BSPI,8,3,A_Pain,S_BSPI_RUN1},  // S_BSPI_PAIN2
  {SPR_BSPI,9,20,A_Scream,S_BSPI_DIE2}, // S_BSPI_DIE1
  {SPR_BSPI,10,7,A_Fall,S_BSPI_DIE3}, // S_BSPI_DIE2
  {SPR_BSPI,11,7,NULL,S_BSPI_DIE4}, // S_BSPI_DIE3
  {SPR_BSPI,12,7,NULL,S_BSPI_DIE5}, // S_BSPI_DIE4
  {SPR_BSPI,13,7,NULL,S_BSPI_DIE6}, // S_BSPI_DIE5
  {SPR_BSPI,14,7,NULL,S_BSPI_DIE7}, // S_BSPI_DIE6
  {SPR_BSPI,15,-1,A_BossDeath,S_NULL},  // S_BSPI_DIE7
  {SPR_BSPI,15,5,NULL,S_BSPI_RAISE2}, // S_BSPI_RAISE1
  {SPR_BSPI,14,5,NULL,S_BSPI_RAISE3}, // S_BSPI_RAISE2
  {SPR_BSPI,13,5,NULL,S_BSPI_RAISE4}, // S_BSPI_RAISE3
  {SPR_BSPI,12,5,NULL,S_BSPI_RAISE5}, // S_BSPI_RAISE4
  {SPR_BSPI,11,5,NULL,S_BSPI_RAISE6}, // S_BSPI_RAISE5
  {SPR_BSPI,10,5,NULL,S_BSPI_RAISE7}, // S_BSPI_RAISE6
  {SPR_BSPI,9,5,NULL,S_BSPI_RUN1},  // S_BSPI_RAISE7
  {SPR_APLS,32768,5,NULL,S_ARACH_PLAZ2},  // S_ARACH_PLAZ
  {SPR_APLS,32769,5,NULL,S_ARACH_PLAZ}, // S_ARACH_PLAZ2
  {SPR_APBX,32768,5,NULL,S_ARACH_PLEX2},  // S_ARACH_PLEX
  {SPR_APBX,32769,5,NULL,S_ARACH_PLEX3},  // S_ARACH_PLEX2
  {SPR_APBX,32770,5,NULL,S_ARACH_PLEX4},  // S_ARACH_PLEX3
  {SPR_APBX,32771,5,NULL,S_ARACH_PLEX5},  // S_ARACH_PLEX4
  {SPR_APBX,32772,5,NULL,S_NULL}, // S_ARACH_PLEX5
  {SPR_CYBR,0,10,A_Look,S_CYBER_STND2}, // S_CYBER_STND
  {SPR_CYBR,1,10,A_Look,S_CYBER_STND},  // S_CYBER_STND2
  {SPR_CYBR,0,3,A_Hoof,S_CYBER_RUN2}, // S_CYBER_RUN1
  {SPR_CYBR,0,3,A_Chase,S_CYBER_RUN3},  // S_CYBER_RUN2
  {SPR_CYBR,1,3,A_Chase,S_CYBER_RUN4},  // S_CYBER_RUN3
  {SPR_CYBR,1,3,A_Chase,S_CYBER_RUN5},  // S_CYBER_RUN4
  {SPR_CYBR,2,3,A_Chase,S_CYBER_RUN6},  // S_CYBER_RUN5
  {SPR_CYBR,2,3,A_Chase,S_CYBER_RUN7},  // S_CYBER_RUN6
  {SPR_CYBR,3,3,A_Metal,S_CYBER_RUN8},  // S_CYBER_RUN7
  {SPR_CYBR,3,3,A_Chase,S_CYBER_RUN1},  // S_CYBER_RUN8
  {SPR_CYBR,4,6,A_FaceTarget,S_CYBER_ATK2}, // S_CYBER_ATK1
  {SPR_CYBR,5,12,A_CyberAttack,S_CYBER_ATK3}, // S_CYBER_ATK2
  {SPR_CYBR,4,12,A_FaceTarget,S_CYBER_ATK4},  // S_CYBER_ATK3
  {SPR_CYBR,5,12,A_CyberAttack,S_CYBER_ATK5}, // S_CYBER_ATK4
  {SPR_CYBR,4,12,A_FaceTarget,S_CYBER_ATK6},  // S_CYBER_ATK5
  {SPR_CYBR,5,12,A_CyberAttack,S_CYBER_RUN1}, // S_CYBER_ATK6
  {SPR_CYBR,6,10,A_Pain,S_CYBER_RUN1},  // S_CYBER_PAIN
  {SPR_CYBR,7,10,NULL,S_CYBER_DIE2},  // S_CYBER_DIE1
  {SPR_CYBR,8,10,A_Scream,S_CYBER_DIE3},  // S_CYBER_DIE2
  {SPR_CYBR,9,10,NULL,S_CYBER_DIE4},  // S_CYBER_DIE3
  {SPR_CYBR,10,10,NULL,S_CYBER_DIE5}, // S_CYBER_DIE4
  {SPR_CYBR,11,10,NULL,S_CYBER_DIE6}, // S_CYBER_DIE5
  {SPR_CYBR,12,10,A_Fall,S_CYBER_DIE7}, // S_CYBER_DIE6
  {SPR_CYBR,13,10,NULL,S_CYBER_DIE8}, // S_CYBER_DIE7
  {SPR_CYBR,14,10,NULL,S_CYBER_DIE9}, // S_CYBER_DIE8
  {SPR_CYBR,15,30,NULL,S_CYBER_DIE10},  // S_CYBER_DIE9
  {SPR_CYBR,15,-1,A_BossDeath,S_NULL},  // S_CYBER_DIE10
  {SPR_PAIN,0,10,A_Look,S_PAIN_STND}, // S_PAIN_STND
  {SPR_PAIN,0,3,A_Chase,S_PAIN_RUN2}, // S_PAIN_RUN1
  {SPR_PAIN,0,3,A_Chase,S_PAIN_RUN3}, // S_PAIN_RUN2
  {SPR_PAIN,1,3,A_Chase,S_PAIN_RUN4}, // S_PAIN_RUN3
  {SPR_PAIN,1,3,A_Chase,S_PAIN_RUN5}, // S_PAIN_RUN4
  {SPR_PAIN,2,3,A_Chase,S_PAIN_RUN6}, // S_PAIN_RUN5
  {SPR_PAIN,2,3,A_Chase,S_PAIN_RUN1}, // S_PAIN_RUN6
  {SPR_PAIN,3,5,A_FaceTarget,S_PAIN_ATK2},  // S_PAIN_ATK1
  {SPR_PAIN,4,5,A_FaceTarget,S_PAIN_ATK3},  // S_PAIN_ATK2
  {SPR_PAIN,32773,5,A_FaceTarget,S_PAIN_ATK4},  // S_PAIN_ATK3
  {SPR_PAIN,32773,0,A_PainAttack,S_PAIN_RUN1},  // S_PAIN_ATK4
  {SPR_PAIN,6,6,NULL,S_PAIN_PAIN2}, // S_PAIN_PAIN
  {SPR_PAIN,6,6,A_Pain,S_PAIN_RUN1},  // S_PAIN_PAIN2
  {SPR_PAIN,32775,8,NULL,S_PAIN_DIE2},  // S_PAIN_DIE1
  {SPR_PAIN,32776,8,A_Scream,S_PAIN_DIE3},  // S_PAIN_DIE2
  {SPR_PAIN,32777,8,NULL,S_PAIN_DIE4},  // S_PAIN_DIE3
  {SPR_PAIN,32778,8,NULL,S_PAIN_DIE5},  // S_PAIN_DIE4
  {SPR_PAIN,32779,8,A_PainDie,S_PAIN_DIE6}, // S_PAIN_DIE5
  {SPR_PAIN,32780,8,NULL,S_NULL}, // S_PAIN_DIE6
  {SPR_PAIN,12,8,NULL,S_PAIN_RAISE2}, // S_PAIN_RAISE1
  {SPR_PAIN,11,8,NULL,S_PAIN_RAISE3}, // S_PAIN_RAISE2
  {SPR_PAIN,10,8,NULL,S_PAIN_RAISE4}, // S_PAIN_RAISE3
  {SPR_PAIN,9,8,NULL,S_PAIN_RAISE5},  // S_PAIN_RAISE4
  {SPR_PAIN,8,8,NULL,S_PAIN_RAISE6},  // S_PAIN_RAISE5
  {SPR_PAIN,7,8,NULL,S_PAIN_RUN1},  // S_PAIN_RAISE6
  {SPR_SSWV,0,10,A_Look,S_SSWV_STND2},  // S_SSWV_STND
  {SPR_SSWV,1,10,A_Look,S_SSWV_STND}, // S_SSWV_STND2
  {SPR_SSWV,0,3,A_Chase,S_SSWV_RUN2}, // S_SSWV_RUN1
  {SPR_SSWV,0,3,A_Chase,S_SSWV_RUN3}, // S_SSWV_RUN2
  {SPR_SSWV,1,3,A_Chase,S_SSWV_RUN4}, // S_SSWV_RUN3
  {SPR_SSWV,1,3,A_Chase,S_SSWV_RUN5}, // S_SSWV_RUN4
  {SPR_SSWV,2,3,A_Chase,S_SSWV_RUN6}, // S_SSWV_RUN5
  {SPR_SSWV,2,3,A_Chase,S_SSWV_RUN7}, // S_SSWV_RUN6
  {SPR_SSWV,3,3,A_Chase,S_SSWV_RUN8}, // S_SSWV_RUN7
  {SPR_SSWV,3,3,A_Chase,S_SSWV_RUN1}, // S_SSWV_RUN8
  {SPR_SSWV,4,10,A_FaceTarget,S_SSWV_ATK2}, // S_SSWV_ATK1
  {SPR_SSWV,5,10,A_FaceTarget,S_SSWV_ATK3}, // S_SSWV_ATK2
  {SPR_SSWV,32774,4,A_CPosAttack,S_SSWV_ATK4},  // S_SSWV_ATK3
  {SPR_SSWV,5,6,A_FaceTarget,S_SSWV_ATK5},  // S_SSWV_ATK4
  {SPR_SSWV,32774,4,A_CPosAttack,S_SSWV_ATK6},  // S_SSWV_ATK5
  {SPR_SSWV,5,1,A_CPosRefire,S_SSWV_ATK2},  // S_SSWV_ATK6
  {SPR_SSWV,7,3,NULL,S_SSWV_PAIN2}, // S_SSWV_PAIN
  {SPR_SSWV,7,3,A_Pain,S_SSWV_RUN1},  // S_SSWV_PAIN2
  {SPR_SSWV,8,5,NULL,S_SSWV_DIE2},  // S_SSWV_DIE1
  {SPR_SSWV,9,5,A_Scream,S_SSWV_DIE3},  // S_SSWV_DIE2
  {SPR_SSWV,10,5,A_Fall,S_SSWV_DIE4}, // S_SSWV_DIE3
  {SPR_SSWV,11,5,NULL,S_SSWV_DIE5}, // S_SSWV_DIE4
  {SPR_SSWV,12,-1,NULL,S_NULL}, // S_SSWV_DIE5
  {SPR_SSWV,13,5,NULL,S_SSWV_XDIE2},  // S_SSWV_XDIE1
  {SPR_SSWV,14,5,A_XScream,S_SSWV_XDIE3}, // S_SSWV_XDIE2
  {SPR_SSWV,15,5,A_Fall,S_SSWV_XDIE4},  // S_SSWV_XDIE3
  {SPR_SSWV,16,5,NULL,S_SSWV_XDIE5},  // S_SSWV_XDIE4
  {SPR_SSWV,17,5,NULL,S_SSWV_XDIE6},  // S_SSWV_XDIE5
  {SPR_SSWV,18,5,NULL,S_SSWV_XDIE7},  // S_SSWV_XDIE6
  {SPR_SSWV,19,5,NULL,S_SSWV_XDIE8},  // S_SSWV_XDIE7
  {SPR_SSWV,20,5,NULL,S_SSWV_XDIE9},  // S_SSWV_XDIE8
  {SPR_SSWV,21,-1,NULL,S_NULL}, // S_SSWV_XDIE9
  {SPR_SSWV,12,5,NULL,S_SSWV_RAISE2}, // S_SSWV_RAISE1
  {SPR_SSWV,11,5,NULL,S_SSWV_RAISE3}, // S_SSWV_RAISE2
  {SPR_SSWV,10,5,NULL,S_SSWV_RAISE4}, // S_SSWV_RAISE3
  {SPR_SSWV,9,5,NULL,S_SSWV_RAISE5},  // S_SSWV_RAISE4
  {SPR_SSWV,8,5,NULL,S_SSWV_RUN1},  // S_SSWV_RAISE5
  {SPR_KEEN,0,-1,NULL,S_KEENSTND},  // S_KEENSTND
  {SPR_KEEN,0,6,NULL,S_COMMKEEN2},  // S_COMMKEEN
  {SPR_KEEN,1,6,NULL,S_COMMKEEN3},  // S_COMMKEEN2
  {SPR_KEEN,2,6,A_Scream,S_COMMKEEN4},  // S_COMMKEEN3
  {SPR_KEEN,3,6,NULL,S_COMMKEEN5},  // S_COMMKEEN4
  {SPR_KEEN,4,6,NULL,S_COMMKEEN6},  // S_COMMKEEN5
  {SPR_KEEN,5,6,NULL,S_COMMKEEN7},  // S_COMMKEEN6
  {SPR_KEEN,6,6,NULL,S_COMMKEEN8},  // S_COMMKEEN7
  {SPR_KEEN,7,6,NULL,S_COMMKEEN9},  // S_COMMKEEN8
  {SPR_KEEN,8,6,NULL,S_COMMKEEN10}, // S_COMMKEEN9
  {SPR_KEEN,9,6,NULL,S_COMMKEEN11}, // S_COMMKEEN10
  {SPR_KEEN,10,6,A_KeenDie,S_COMMKEEN12},// S_COMMKEEN11
  {SPR_KEEN,11,-1,NULL,S_NULL},   // S_COMMKEEN12
  {SPR_KEEN,12,4,NULL,S_KEENPAIN2}, // S_KEENPAIN
  {SPR_KEEN,12,8,A_Pain,S_KEENSTND},  // S_KEENPAIN2
  {SPR_BBRN,0,-1,NULL,S_NULL},    // S_BRAIN
  {SPR_BBRN,1,36,A_BrainPain,S_BRAIN},  // S_BRAIN_PAIN
  {SPR_BBRN,0,100,A_BrainScream,S_BRAIN_DIE2},  // S_BRAIN_DIE1
  {SPR_BBRN,0,10,NULL,S_BRAIN_DIE3},  // S_BRAIN_DIE2
  {SPR_BBRN,0,10,NULL,S_BRAIN_DIE4},  // S_BRAIN_DIE3
  {SPR_BBRN,0,-1,A_BrainDie,S_NULL},  // S_BRAIN_DIE4
  {SPR_SSWV,0,10,A_Look,S_BRAINEYE},  // S_BRAINEYE
  {SPR_SSWV,0,181,A_BrainAwake,S_BRAINEYE1},  // S_BRAINEYESEE
  {SPR_SSWV,0,150,A_BrainSpit,S_BRAINEYE1}, // S_BRAINEYE1
  {SPR_BOSF,32768,3,A_SpawnSound,S_SPAWN2}, // S_SPAWN1
  {SPR_BOSF,32769,3,A_SpawnFly,S_SPAWN3}, // S_SPAWN2
  {SPR_BOSF,32770,3,A_SpawnFly,S_SPAWN4}, // S_SPAWN3
  {SPR_BOSF,32771,3,A_SpawnFly,S_SPAWN1}, // S_SPAWN4
  {SPR_FIRE,32768,4,A_Fire,S_SPAWNFIRE2}, // S_SPAWNFIRE1
  {SPR_FIRE,32769,4,A_Fire,S_SPAWNFIRE3}, // S_SPAWNFIRE2
  {SPR_FIRE,32770,4,A_Fire,S_SPAWNFIRE4}, // S_SPAWNFIRE3
  {SPR_FIRE,32771,4,A_Fire,S_SPAWNFIRE5}, // S_SPAWNFIRE4
  {SPR_FIRE,32772,4,A_Fire,S_SPAWNFIRE6}, // S_SPAWNFIRE5
  {SPR_FIRE,32773,4,A_Fire,S_SPAWNFIRE7}, // S_SPAWNFIRE6
  {SPR_FIRE,32774,4,A_Fire,S_SPAWNFIRE8}, // S_SPAWNFIRE7
  {SPR_FIRE,32775,4,A_Fire,S_NULL},   // S_SPAWNFIRE8
  {SPR_MISL,32769,10,NULL,S_BRAINEXPLODE2}, // S_BRAINEXPLODE1
  {SPR_MISL,32770,10,NULL,S_BRAINEXPLODE3}, // S_BRAINEXPLODE2
  {SPR_MISL,32771,10,A_BrainExplode,S_NULL},  // S_BRAINEXPLODE3
  {SPR_ARM1,0,6,NULL,S_ARM1A},  // S_ARM1
  {SPR_ARM1,32769,7,NULL,S_ARM1}, // S_ARM1A
  {SPR_ARM2,0,6,NULL,S_ARM2A},  // S_ARM2
  {SPR_ARM2,32769,6,NULL,S_ARM2}, // S_ARM2A
  {SPR_BAR1,0,6,NULL,S_BAR2}, // S_BAR1
  {SPR_BAR1,1,6,NULL,S_BAR1}, // S_BAR2
  {SPR_BEXP,32768,5,NULL,S_BEXP2},  // S_BEXP
  {SPR_BEXP,32769,5,A_Scream,S_BEXP3},  // S_BEXP2
  {SPR_BEXP,32770,5,NULL,S_BEXP4},  // S_BEXP3
  {SPR_BEXP,32771,10,A_Explode,S_BEXP5},  // S_BEXP4
  {SPR_BEXP,32772,10,NULL,S_NULL},  // S_BEXP5
  {SPR_FCAN,32768,4,NULL,S_BBAR2},  // S_BBAR1
  {SPR_FCAN,32769,4,NULL,S_BBAR3},  // S_BBAR2
  {SPR_FCAN,32770,4,NULL,S_BBAR1},  // S_BBAR3
  {SPR_BON1,0,6,NULL,S_BON1A},  // S_BON1
  {SPR_BON1,1,6,NULL,S_BON1B},  // S_BON1A
  {SPR_BON1,2,6,NULL,S_BON1C},  // S_BON1B
  {SPR_BON1,3,6,NULL,S_BON1D},  // S_BON1C
  {SPR_BON1,2,6,NULL,S_BON1E},  // S_BON1D
  {SPR_BON1,1,6,NULL,S_BON1}, // S_BON1E
  {SPR_BON2,0,6,NULL,S_BON2A},  // S_BON2
  {SPR_BON2,1,6,NULL,S_BON2B},  // S_BON2A
  {SPR_BON2,2,6,NULL,S_BON2C},  // S_BON2B
  {SPR_BON2,3,6,NULL,S_BON2D},  // S_BON2C
  {SPR_BON2,2,6,NULL,S_BON2E},  // S_BON2D
  {SPR_BON2,1,6,NULL,S_BON2}, // S_BON2E
  {SPR_BKEY,0,10,NULL,S_BKEY2}, // S_BKEY
  {SPR_BKEY,32769,10,NULL,S_BKEY},  // S_BKEY2
  {SPR_RKEY,0,10,NULL,S_RKEY2}, // S_RKEY
  {SPR_RKEY,32769,10,NULL,S_RKEY},  // S_RKEY2
  {SPR_YKEY,0,10,NULL,S_YKEY2}, // S_YKEY
  {SPR_YKEY,32769,10,NULL,S_YKEY},  // S_YKEY2
  {SPR_BSKU,0,10,NULL,S_BSKULL2}, // S_BSKULL
  {SPR_BSKU,32769,10,NULL,S_BSKULL},  // S_BSKULL2
  {SPR_RSKU,0,10,NULL,S_RSKULL2}, // S_RSKULL
  {SPR_RSKU,32769,10,NULL,S_RSKULL},  // S_RSKULL2
  {SPR_YSKU,0,10,NULL,S_YSKULL2}, // S_YSKULL
  {SPR_YSKU,32769,10,NULL,S_YSKULL},  // S_YSKULL2
  {SPR_STIM,0,-1,NULL,S_NULL},  // S_STIM
  {SPR_MEDI,0,-1,NULL,S_NULL},  // S_MEDI
  {SPR_SOUL,32768,6,NULL,S_SOUL2},  // S_SOUL
  {SPR_SOUL,32769,6,NULL,S_SOUL3},  // S_SOUL2
  {SPR_SOUL,32770,6,NULL,S_SOUL4},  // S_SOUL3
  {SPR_SOUL,32771,6,NULL,S_SOUL5},  // S_SOUL4
  {SPR_SOUL,32770,6,NULL,S_SOUL6},  // S_SOUL5
  {SPR_SOUL,32769,6,NULL,S_SOUL}, // S_SOUL6
  {SPR_PINV,32768,6,NULL,S_PINV2},  // S_PINV
  {SPR_PINV,32769,6,NULL,S_PINV3},  // S_PINV2
  {SPR_PINV,32770,6,NULL,S_PINV4},  // S_PINV3
  {SPR_PINV,32771,6,NULL,S_PINV}, // S_PINV4
  {SPR_PSTR,32768,-1,NULL,S_NULL},  // S_PSTR
  {SPR_PINS,32768,6,NULL,S_PINS2},  // S_PINS
  {SPR_PINS,32769,6,NULL,S_PINS3},  // S_PINS2
  {SPR_PINS,32770,6,NULL,S_PINS4},  // S_PINS3
  {SPR_PINS,32771,6,NULL,S_PINS}, // S_PINS4
  {SPR_MEGA,32768,6,NULL,S_MEGA2},  // S_MEGA
  {SPR_MEGA,32769,6,NULL,S_MEGA3},  // S_MEGA2
  {SPR_MEGA,32770,6,NULL,S_MEGA4},  // S_MEGA3
  {SPR_MEGA,32771,6,NULL,S_MEGA}, // S_MEGA4
  {SPR_SUIT,32768,-1,NULL,S_NULL},  // S_SUIT
  {SPR_PMAP,32768,6,NULL,S_PMAP2},  // S_PMAP
  {SPR_PMAP,32769,6,NULL,S_PMAP3},  // S_PMAP2
  {SPR_PMAP,32770,6,NULL,S_PMAP4},  // S_PMAP3
  {SPR_PMAP,32771,6,NULL,S_PMAP5},  // S_PMAP4
  {SPR_PMAP,32770,6,NULL,S_PMAP6},  // S_PMAP5
  {SPR_PMAP,32769,6,NULL,S_PMAP}, // S_PMAP6
  {SPR_PVIS,32768,6,NULL,S_PVIS2},  // S_PVIS
  {SPR_PVIS,1,6,NULL,S_PVIS}, // S_PVIS2
  {SPR_CLIP,0,-1,NULL,S_NULL},  // S_CLIP
  {SPR_AMMO,0,-1,NULL,S_NULL},  // S_AMMO
  {SPR_ROCK,0,-1,NULL,S_NULL},  // S_ROCK
  {SPR_BROK,0,-1,NULL,S_NULL},  // S_BROK
  {SPR_CELL,0,-1,NULL,S_NULL},  // S_CELL
  {SPR_CELP,0,-1,NULL,S_NULL},  // S_CELP
  {SPR_SHEL,0,-1,NULL,S_NULL},  // S_SHEL
  {SPR_SBOX,0,-1,NULL,S_NULL},  // S_SBOX
  {SPR_BPAK,0,-1,NULL,S_NULL},  // S_BPAK
  {SPR_BFUG,0,-1,NULL,S_NULL},  // S_BFUG
  {SPR_MGUN,0,-1,NULL,S_NULL},  // S_MGUN
  {SPR_CSAW,0,-1,NULL,S_NULL},  // S_CSAW
  {SPR_LAUN,0,-1,NULL,S_NULL},  // S_LAUN
  {SPR_PLAS,0,-1,NULL,S_NULL},  // S_PLAS
  {SPR_SHOT,0,-1,NULL,S_NULL},  // S_SHOT
  {SPR_SGN2,0,-1,NULL,S_NULL},  // S_SHOT2
  {SPR_COLU,32768,-1,NULL,S_NULL},  // S_COLU
  {SPR_SMT2,0,-1,NULL,S_NULL},  // S_STALAG
  {SPR_GOR1,0,10,NULL,S_BLOODYTWITCH2}, // S_BLOODYTWITCH
  {SPR_GOR1,1,15,NULL,S_BLOODYTWITCH3}, // S_BLOODYTWITCH2
  {SPR_GOR1,2,8,NULL,S_BLOODYTWITCH4},  // S_BLOODYTWITCH3
  {SPR_GOR1,1,6,NULL,S_BLOODYTWITCH}, // S_BLOODYTWITCH4
  {SPR_PLAY,13,-1,NULL,S_NULL}, // S_DEADTORSO
  {SPR_PLAY,18,-1,NULL,S_NULL}, // S_DEADBOTTOM
  {SPR_POL2,0,-1,NULL,S_NULL},  // S_HEADSONSTICK
  {SPR_POL5,0,-1,NULL,S_NULL},  // S_GIBS
  {SPR_POL4,0,-1,NULL,S_NULL},  // S_HEADONASTICK
  {SPR_POL3,32768,6,NULL,S_HEADCANDLES2}, // S_HEADCANDLES
  {SPR_POL3,32769,6,NULL,S_HEADCANDLES},  // S_HEADCANDLES2
  {SPR_POL1,0,-1,NULL,S_NULL},  // S_DEADSTICK
  {SPR_POL6,0,6,NULL,S_LIVESTICK2}, // S_LIVESTICK
  {SPR_POL6,1,8,NULL,S_LIVESTICK},  // S_LIVESTICK2
  {SPR_GOR2,0,-1,NULL,S_NULL},  // S_MEAT2
  {SPR_GOR3,0,-1,NULL,S_NULL},  // S_MEAT3
  {SPR_GOR4,0,-1,NULL,S_NULL},  // S_MEAT4
  {SPR_GOR5,0,-1,NULL,S_NULL},  // S_MEAT5
  {SPR_SMIT,0,-1,NULL,S_NULL},  // S_STALAGTITE
  {SPR_COL1,0,-1,NULL,S_NULL},  // S_TALLGRNCOL
  {SPR_COL2,0,-1,NULL,S_NULL},  // S_SHRTGRNCOL
  {SPR_COL3,0,-1,NULL,S_NULL},  // S_TALLREDCOL
  {SPR_COL4,0,-1,NULL,S_NULL},  // S_SHRTREDCOL
  {SPR_CAND,32768,-1,NULL,S_NULL},  // S_CANDLESTIK
  {SPR_CBRA,32768,-1,NULL,S_NULL},  // S_CANDELABRA
  {SPR_COL6,0,-1,NULL,S_NULL},  // S_SKULLCOL
  {SPR_TRE1,0,-1,NULL,S_NULL},  // S_TORCHTREE
  {SPR_TRE2,0,-1,NULL,S_NULL},  // S_BIGTREE
  {SPR_ELEC,0,-1,NULL,S_NULL},  // S_TECHPILLAR
  {SPR_CEYE,32768,6,NULL,S_EVILEYE2}, // S_EVILEYE
  {SPR_CEYE,32769,6,NULL,S_EVILEYE3}, // S_EVILEYE2
  {SPR_CEYE,32770,6,NULL,S_EVILEYE4}, // S_EVILEYE3
  {SPR_CEYE,32769,6,NULL,S_EVILEYE},  // S_EVILEYE4
  {SPR_FSKU,32768,6,NULL,S_FLOATSKULL2},  // S_FLOATSKULL
  {SPR_FSKU,32769,6,NULL,S_FLOATSKULL3},  // S_FLOATSKULL2
  {SPR_FSKU,32770,6,NULL,S_FLOATSKULL}, // S_FLOATSKULL3
  {SPR_COL5,0,14,NULL,S_HEARTCOL2}, // S_HEARTCOL
  {SPR_COL5,1,14,NULL,S_HEARTCOL},  // S_HEARTCOL2
  {SPR_TBLU,32768,4,NULL,S_BLUETORCH2}, // S_BLUETORCH
  {SPR_TBLU,32769,4,NULL,S_BLUETORCH3}, // S_BLUETORCH2
  {SPR_TBLU,32770,4,NULL,S_BLUETORCH4}, // S_BLUETORCH3
  {SPR_TBLU,32771,4,NULL,S_BLUETORCH},  // S_BLUETORCH4
  {SPR_TGRN,32768,4,NULL,S_GREENTORCH2},  // S_GREENTORCH
  {SPR_TGRN,32769,4,NULL,S_GREENTORCH3},  // S_GREENTORCH2
  {SPR_TGRN,32770,4,NULL,S_GREENTORCH4},  // S_GREENTORCH3
  {SPR_TGRN,32771,4,NULL,S_GREENTORCH}, // S_GREENTORCH4
  {SPR_TRED,32768,4,NULL,S_REDTORCH2},  // S_REDTORCH
  {SPR_TRED,32769,4,NULL,S_REDTORCH3},  // S_REDTORCH2
  {SPR_TRED,32770,4,NULL,S_REDTORCH4},  // S_REDTORCH3
  {SPR_TRED,32771,4,NULL,S_REDTORCH}, // S_REDTORCH4
  {SPR_SMBT,32768,4,NULL,S_BTORCHSHRT2},  // S_BTORCHSHRT
  {SPR_SMBT,32769,4,NULL,S_BTORCHSHRT3},  // S_BTORCHSHRT2
  {SPR_SMBT,32770,4,NULL,S_BTORCHSHRT4},  // S_BTORCHSHRT3
  {SPR_SMBT,32771,4,NULL,S_BTORCHSHRT}, // S_BTORCHSHRT4
  {SPR_SMGT,32768,4,NULL,S_GTORCHSHRT2},  // S_GTORCHSHRT
  {SPR_SMGT,32769,4,NULL,S_GTORCHSHRT3},  // S_GTORCHSHRT2
  {SPR_SMGT,32770,4,NULL,S_GTORCHSHRT4},  // S_GTORCHSHRT3
  {SPR_SMGT,32771,4,NULL,S_GTORCHSHRT}, // S_GTORCHSHRT4
  {SPR_SMRT,32768,4,NULL,S_RTORCHSHRT2},  // S_RTORCHSHRT
  {SPR_SMRT,32769,4,NULL,S_RTORCHSHRT3},  // S_RTORCHSHRT2
  {SPR_SMRT,32770,4,NULL,S_RTORCHSHRT4},  // S_RTORCHSHRT3
  {SPR_SMRT,32771,4,NULL,S_RTORCHSHRT}, // S_RTORCHSHRT4
  {SPR_HDB1,0,-1,NULL,S_NULL},  // S_HANGNOGUTS
  {SPR_HDB2,0,-1,NULL,S_NULL},  // S_HANGBNOBRAIN
  {SPR_HDB3,0,-1,NULL,S_NULL},  // S_HANGTLOOKDN
  {SPR_HDB4,0,-1,NULL,S_NULL},  // S_HANGTSKULL
  {SPR_HDB5,0,-1,NULL,S_NULL},  // S_HANGTLOOKUP
  {SPR_HDB6,0,-1,NULL,S_NULL},  // S_HANGTNOBRAIN
  {SPR_POB1,0,-1,NULL,S_NULL},  // S_COLONGIBS
  {SPR_POB2,0,-1,NULL,S_NULL},  // S_SMALLPOOL
  {SPR_BRS1,0,-1,NULL,S_NULL},    // S_BRAINSTEM
  {SPR_TLMP,32768,4,NULL,S_TECHLAMP2},  // S_TECHLAMP
  {SPR_TLMP,32769,4,NULL,S_TECHLAMP3},  // S_TECHLAMP2
  {SPR_TLMP,32770,4,NULL,S_TECHLAMP4},  // S_TECHLAMP3
  {SPR_TLMP,32771,4,NULL,S_TECHLAMP},   // S_TECHLAMP4
  {SPR_TLP2,32768,4,NULL,S_TECH2LAMP2}, // S_TECH2LAMP
  {SPR_TLP2,32769,4,NULL,S_TECH2LAMP3}, // S_TECH2LAMP2
  {SPR_TLP2,32770,4,NULL,S_TECH2LAMP4}, // S_TECH2LAMP3
  {SPR_TLP2,32771,4,NULL,S_TECH2LAMP},  // S_TECH2LAMP4

  {SPR_TNT1,0,-1,NULL,S_TNT1},          // S_TNT1    // phares 3/8/98

  // killough 8/9/98: grenade
  {SPR_MISL,32768,1000,A_Die,S_GRENADE},      // S_GRENADE

  // killough 8/10/98: variable damage explosion
  {SPR_MISL,32769,4,A_Scream,S_DETONATE2},    // S_DETONATE
  {SPR_MISL,32770,6,A_Detonate,S_DETONATE3},  // S_DETONATE2
  {SPR_MISL,32771,10,NULL,S_NULL},            // S_DETONATE3

#ifdef DOGS
  // killough 7/19/98: Marine's best friend :)
  {SPR_DOGS,0,10,A_Look,S_DOGS_STND2},  // S_DOGS_STND
  {SPR_DOGS,1,10,A_Look,S_DOGS_STND}, // S_DOGS_STND2
  {SPR_DOGS,0,2,A_Chase,S_DOGS_RUN2}, // S_DOGS_RUN1
  {SPR_DOGS,0,2,A_Chase,S_DOGS_RUN3}, // S_DOGS_RUN2
  {SPR_DOGS,1,2,A_Chase,S_DOGS_RUN4}, // S_DOGS_RUN3
  {SPR_DOGS,1,2,A_Chase,S_DOGS_RUN5}, // S_DOGS_RUN4
  {SPR_DOGS,2,2,A_Chase,S_DOGS_RUN6}, // S_DOGS_RUN5
  {SPR_DOGS,2,2,A_Chase,S_DOGS_RUN7}, // S_DOGS_RUN6
  {SPR_DOGS,3,2,A_Chase,S_DOGS_RUN8}, // S_DOGS_RUN7
  {SPR_DOGS,3,2,A_Chase,S_DOGS_RUN1}, // S_DOGS_RUN8
  {SPR_DOGS,4,8,A_FaceTarget,S_DOGS_ATK2},  // S_DOGS_ATK1
  {SPR_DOGS,5,8,A_FaceTarget,S_DOGS_ATK3},  // S_DOGS_ATK2
  {SPR_DOGS,6,8,A_SargAttack,S_DOGS_RUN1},  // S_DOGS_ATK3
  {SPR_DOGS,7,2,NULL,S_DOGS_PAIN2}, // S_DOGS_PAIN
  {SPR_DOGS,7,2,A_Pain,S_DOGS_RUN1},  // S_DOGS_PAIN2
  {SPR_DOGS,8,8,NULL,S_DOGS_DIE2},  // S_DOGS_DIE1
  {SPR_DOGS,9,8,A_Scream,S_DOGS_DIE3},  // S_DOGS_DIE2
  {SPR_DOGS,10,4,NULL,S_DOGS_DIE4}, // S_DOGS_DIE3
  {SPR_DOGS,11,4,A_Fall,S_DOGS_DIE5}, // S_DOGS_DIE4
  {SPR_DOGS,12,4,NULL,S_DOGS_DIE6}, // S_DOGS_DIE5
  {SPR_DOGS,13,-1,NULL,S_NULL}, // S_DOGS_DIE6
  {SPR_DOGS,13,5,NULL,S_DOGS_RAISE2}, // S_DOGS_RAISE1
  {SPR_DOGS,12,5,NULL,S_DOGS_RAISE3}, // S_DOGS_RAISE2
  {SPR_DOGS,11,5,NULL,S_DOGS_RAISE4}, // S_DOGS_RAISE3
  {SPR_DOGS,10,5,NULL,S_DOGS_RAISE5}, // S_DOGS_RAISE4
  {SPR_DOGS,9,5,NULL,S_DOGS_RAISE6},  // S_DOGS_RAISE5
  {SPR_DOGS,8,5,NULL,S_DOGS_RUN1},  // S_DOGS_RAISE6
#endif

#ifdef BETA
  // killough 7/11/98: beta BFG begins here
  // S_OLDBFG1

#define BFGDELAY 1
#define OLDBFG_1FRAMES(x) {SPR_BFGG,1,BFGDELAY,A_FireOldBFG,x+S_OLDBFG1+2},
#define OLDBFG_2FRAMES(x) OLDBFG_1FRAMES(x) OLDBFG_1FRAMES(x+1)
#define OLDBFG_4FRAMES(x) OLDBFG_2FRAMES(x) OLDBFG_2FRAMES(x+2)
#define OLDBFG_8FRAMES(x) OLDBFG_4FRAMES(x) OLDBFG_4FRAMES(x+4)
  {SPR_BFGG,0,10,A_BFGsound,S_OLDBFG1+1},  // S_OLDBFG1

  OLDBFG_8FRAMES(0)
  OLDBFG_8FRAMES(8)
  OLDBFG_8FRAMES(16) 
  OLDBFG_8FRAMES(24)
  OLDBFG_8FRAMES(32)

  {SPR_BFGG,1,0,A_Light0,S_OLDBFG43}, // S_OLDBFG42
  {SPR_BFGG,1,20,A_ReFire,S_BFG},   // S_OLDBFG43

  // killough 7/11/98: end of beta BFG

  // killough 7/19/98: First plasma fireball in the beta:
  {SPR_PLS1,32768,6,NULL,S_PLS1BALL2},  // S_PLS1BALL
  {SPR_PLS1,32769,6,NULL,S_PLS1BALL}, // S_PLS1BALL2
  {SPR_PLS1,32770,4,NULL,S_PLS1EXP2}, // S_PLS1EXP
  {SPR_PLS1,32771,4,NULL,S_PLS1EXP3}, // S_PLS1EXP2
  {SPR_PLS1,32772,4,NULL,S_PLS1EXP4}, // S_PLS1EXP3
  {SPR_PLS1,32773,4,NULL,S_PLS1EXP5}, // S_PLS1EXP4
  {SPR_PLS1,32774,4,NULL,S_NULL}, // S_PLS1EXP5

  // killough 7/19/98: Second plasma fireball in the beta:
  {SPR_PLS2,32768,4,NULL,S_PLS2BALL2}, // S_PLS2BALL
  {SPR_PLS2,32769,4,NULL,S_PLS2BALL},  // S_PLS2BALL2
  {SPR_PLS2,32770,6,NULL,S_PLS2BALLX2},  // S_PLS2BALLX1
  {SPR_PLS2,32771,6,NULL,S_PLS2BALLX3},  // S_PLS2BALLX2
  {SPR_PLS2,32772,6,NULL,S_NULL}, // S_PLS2BALLX3

  {SPR_BON3,0,6,NULL,S_BON3},           // S_BON3  // killough 7/11/98:
  {SPR_BON4,0,6,NULL,S_BON4},           // S_BON4  // beta bonus items

  // killough 10/98: beta lost souls attacked from a distance, 
  // animated with colors, and stayed in the air when killed.
  // This is an approximation, but I'm sure it can be improved.

  // spawnstate
  {SPR_SKUL,0,10,A_Look,S_BSKUL_STND},  // S_BSKUL_STND

  // chasestate
  {SPR_SKUL,1,5,A_Chase,S_BSKUL_RUN2},  // S_BSKUL_RUN1
  {SPR_SKUL,2,5,A_Chase,S_BSKUL_RUN3},  // S_BSKUL_RUN2
  {SPR_SKUL,3,5,A_Chase,S_BSKUL_RUN4},  // S_BSKUL_RUN3
  {SPR_SKUL,0,5,A_Chase,S_BSKUL_RUN1},  // S_BSKUL_RUN4

  // missilestate
  {SPR_SKUL,4,4,A_FaceTarget,S_BSKUL_ATK2},     // S_BSKUL_ATK1
  {SPR_SKUL,5,5,A_BetaSkullAttack,S_BSKUL_ATK3}, // S_BSKUL_ATK2
  {SPR_SKUL,5,4,NULL,S_BSKUL_RUN1},              // S_BSKUL_ATK3

  // painstate
  {SPR_SKUL,6,4,NULL,S_BSKUL_PAIN2},     // S_BSKUL_PAIN1
  {SPR_SKUL,7,2,A_Pain,S_BSKUL_RUN1},   // S_BSKUL_PAIN2
  {SPR_SKUL,8,4,NULL,S_BSKUL_RUN1},      // S_BSKUL_PAIN3

  // deathstate
  {SPR_SKUL, 9,5,NULL,S_BSKUL_DIE2},     // S_BSKUL_DIE1
  {SPR_SKUL,10,5,NULL,S_BSKUL_DIE3},     // S_BSKUL_DIE2
  {SPR_SKUL,11,5,NULL,S_BSKUL_DIE4},     // S_BSKUL_DIE3
  {SPR_SKUL,12,5,NULL,S_BSKUL_DIE5},     // S_BSKUL_DIE4
  {SPR_SKUL,13,5,A_Scream,S_BSKUL_DIE6}, // S_BSKUL_DIE5
  {SPR_SKUL,14,5,NULL,S_BSKUL_DIE7},     // S_BSKUL_DIE6
  {SPR_SKUL,15,5,A_Fall,S_BSKUL_DIE8},   // S_BSKUL_DIE7
  {SPR_SKUL,16,5,A_Stop,S_BSKUL_DIE8},   // S_BSKUL_DIE8

#endif

  // killough 10/98: mushroom effect
  {SPR_MISL,32769,8,A_Mushroom,S_EXPLODE2},  // S_MUSHROOM
};

// ********************************************************************
// Object "Thing" definitions
// ********************************************************************
// Now we get to the actual objects and their characteristics.  If
// you've seen Dehacked, much of this is where the Bits are set,
// commented below as "flags", as well as where you wire in which
// frames are the beginning frames for near and far attack, death,
// and such.  Sounds are hooked in here too, as well as how much
// mass, speed and so forth a Thing has.  Everything you ever wanted
// to know...
//
// Like all this other stuff, the MT_* entries are enumerated in info.h
//
// Note that these are all just indices of the elements involved, and
// not real pointers to them.  For example, the player's death sequence
// is S_PLAY_DIE1, which just evaluates to the index in the states[]
// array above, which actually knows what happens then and what the
// sprite looks like, if it makes noise or not, etc.
//
// Additional comments about each of the entries are located in info.h
// next to the mobjinfo_t structure definition.
//
// This goes on for the next 3000+ lines...

mobjinfo_t mobjinfo[NUMMOBJTYPES] = {
  {   // MT_PLAYER
    -1,   // doomednum
    S_PLAY,   // spawnstate
    100,    // spawnhealth
    S_PLAY_RUN1,    // seestate
    sfx_None,   // seesound
    0,    // reactiontime
    sfx_None,   // attacksound
    S_PLAY_PAIN,    // painstate
    255,    // painchance
    sfx_plpain,   // painsound
    S_NULL,   // meleestate
    S_PLAY_ATK1,    // missilestate
    S_PLAY_DIE1,    // deathstate
    S_PLAY_XDIE1,   // xdeathstate
    sfx_pldeth,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_DROPOFF|MF_PICKUP|MF_NOTDMATCH,  // flags // killough 7/18/98
    S_NULL    // raisestate
  },
  
  {   // MT_POSSESSED
    3004,   // doomednum
    S_POSS_STND,    // spawnstate
    20,   // spawnhealth
    S_POSS_RUN1,    // seestate
    sfx_posit1,   // seesound
    8,    // reactiontime
    sfx_pistol,   // attacksound
    S_POSS_PAIN,    // painstate
    200,    // painchance
    sfx_popain,   // painsound
    0,    // meleestate
    S_POSS_ATK1,    // missilestate
    S_POSS_DIE1,    // deathstate
    S_POSS_XDIE1,   // xdeathstate
    sfx_podth1,   // deathsound
    8,    // speed
    20*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_posact,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL, // flags
    S_POSS_RAISE1   // raisestate
  },
  
  {   // MT_SHOTGUY
    9,    // doomednum
    S_SPOS_STND,    // spawnstate
    30,   // spawnhealth
    S_SPOS_RUN1,    // seestate
    sfx_posit2,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_SPOS_PAIN,    // painstate
    170,    // painchance
    sfx_popain,   // painsound
    0,    // meleestate
    S_SPOS_ATK1,    // missilestate
    S_SPOS_DIE1,    // deathstate
    S_SPOS_XDIE1,   // xdeathstate
    sfx_podth2,   // deathsound
    8,    // speed
    20*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_posact,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL, // flags
    S_SPOS_RAISE1   // raisestate
  },
  
  {   // MT_VILE
    64,   // doomednum
    S_VILE_STND,    // spawnstate
    700,    // spawnhealth
    S_VILE_RUN1,    // seestate
    sfx_vilsit,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_VILE_PAIN,    // painstate
    10,   // painchance
    sfx_vipain,   // painsound
    0,    // meleestate
    S_VILE_ATK1,    // missilestate
    S_VILE_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_vildth,   // deathsound
    15,   // speed
    20*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    500,    // mass
    0,    // damage
    sfx_vilact,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_FIRE
    -1,   // doomednum
    S_FIRE1,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOGRAVITY|MF_TRANSLUCENT,   // flags  // killough 2/21/98
    S_NULL    // raisestate
  },
  
  {   // MT_UNDEAD
    66,   // doomednum
    S_SKEL_STND,    // spawnstate
    300,    // spawnhealth
    S_SKEL_RUN1,    // seestate
    sfx_skesit,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_SKEL_PAIN,    // painstate
    100,    // painchance
    sfx_popain,   // painsound
    S_SKEL_FIST1,   // meleestate
    S_SKEL_MISS1,   // missilestate
    S_SKEL_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_skedth,   // deathsound
    10,   // speed
    20*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    500,    // mass
    0,    // damage
    sfx_skeact,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,  // flags
    S_SKEL_RAISE1   // raisestate
  },
  
  {   // MT_TRACER
    -1,   // doomednum
    S_TRACER,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_skeatk,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_TRACEEXP1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_barexp,   // deathsound
    10*FRACUNIT,    // speed
    11*FRACUNIT,    // radius
    8*FRACUNIT,   // height
    100,    // mass
    10,   // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_SMOKE
    -1,   // doomednum
    S_SMOKE1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOGRAVITY|MF_TRANSLUCENT,   // flags             // phares
    S_NULL    // raisestate
  },
  
  {   // MT_FATSO
    67,   // doomednum
    S_FATT_STND,    // spawnstate
    600,    // spawnhealth
    S_FATT_RUN1,    // seestate
    sfx_mansit,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_FATT_PAIN,    // painstate
    80,   // painchance
    sfx_mnpain,   // painsound
    0,    // meleestate
    S_FATT_ATK1,    // missilestate
    S_FATT_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_mandth,   // deathsound
    8,    // speed
    48*FRACUNIT,    // radius
    64*FRACUNIT,    // height
    1000,   // mass
    0,    // damage
    sfx_posact,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_FATT_RAISE1   // raisestate
  },
  
  {   // MT_FATSHOT
    -1,   // doomednum
    S_FATSHOT1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_firsht,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_FATSHOTX1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    20*FRACUNIT,    // speed
    6*FRACUNIT,   // radius
    8*FRACUNIT,   // height
    100,    // mass
    8,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_TRANSLUCENT,   // flags \\ killough 2/21/98
    S_NULL    // raisestate
  },
  
  {   // MT_CHAINGUY
    65,   // doomednum
    S_CPOS_STND,    // spawnstate
    70,   // spawnhealth
    S_CPOS_RUN1,    // seestate
    sfx_posit2,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_CPOS_PAIN,    // painstate
    170,    // painchance
    sfx_popain,   // painsound
    0,    // meleestate
    S_CPOS_ATK1,    // missilestate
    S_CPOS_DIE1,    // deathstate
    S_CPOS_XDIE1,   // xdeathstate
    sfx_podth2,   // deathsound
    8,    // speed
    20*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_posact,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_CPOS_RAISE1   // raisestate
  },
  
  {   // MT_TROOP
    3001,   // doomednum
    S_TROO_STND,    // spawnstate
    60,   // spawnhealth
    S_TROO_RUN1,    // seestate
    sfx_bgsit1,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_TROO_PAIN,    // painstate
    200,    // painchance
    sfx_popain,   // painsound
    S_TROO_ATK1,    // meleestate
    S_TROO_ATK1,    // missilestate
    S_TROO_DIE1,    // deathstate
    S_TROO_XDIE1,   // xdeathstate
    sfx_bgdth1,   // deathsound
    8,    // speed
    20*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_bgact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL, // flags
    S_TROO_RAISE1   // raisestate
  },
  
  {   // MT_SERGEANT
    3002,   // doomednum
    S_SARG_STND,    // spawnstate
    150,    // spawnhealth
    S_SARG_RUN1,    // seestate
    sfx_sgtsit,   // seesound
    8,    // reactiontime
    sfx_sgtatk,   // attacksound
    S_SARG_PAIN,    // painstate
    180,    // painchance
    sfx_dmpain,   // painsound
    S_SARG_ATK1,    // meleestate
    0,    // missilestate
    S_SARG_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_sgtdth,   // deathsound
    10,   // speed
    30*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    400,    // mass
    0,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL, // flags
    S_SARG_RAISE1   // raisestate
  },
  
  {   // MT_SHADOWS
    58,   // doomednum
    S_SARG_STND,    // spawnstate
    150,    // spawnhealth
    S_SARG_RUN1,    // seestate
    sfx_sgtsit,   // seesound
    8,    // reactiontime
    sfx_sgtatk,   // attacksound
    S_SARG_PAIN,    // painstate
    180,    // painchance
    sfx_dmpain,   // painsound
    S_SARG_ATK1,    // meleestate
    0,    // missilestate
    S_SARG_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_sgtdth,   // deathsound
    10,   // speed
    30*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    400,    // mass
    0,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_SHADOW|MF_COUNTKILL,   // flags
    S_SARG_RAISE1   // raisestate
  },
  
  {   // MT_HEAD
    3005,   // doomednum
    S_HEAD_STND,    // spawnstate
    400,    // spawnhealth
    S_HEAD_RUN1,    // seestate
    sfx_cacsit,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_HEAD_PAIN,    // painstate
    128,    // painchance
    sfx_dmpain,   // painsound
    0,    // meleestate
    S_HEAD_ATK1,    // missilestate
    S_HEAD_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_cacdth,   // deathsound
    8,    // speed
    31*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    400,    // mass
    0,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_FLOAT|MF_NOGRAVITY|MF_COUNTKILL,   // flags
    S_HEAD_RAISE1   // raisestate
  },
  
  {   // MT_BRUISER
    3003,   // doomednum
    S_BOSS_STND,    // spawnstate
    1000,   // spawnhealth
    S_BOSS_RUN1,    // seestate
    sfx_brssit,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_BOSS_PAIN,    // painstate
    50,   // painchance
    sfx_dmpain,   // painsound
    S_BOSS_ATK1,    // meleestate
    S_BOSS_ATK1,    // missilestate
    S_BOSS_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_brsdth,   // deathsound
    8,    // speed
    24*FRACUNIT,    // radius
    64*FRACUNIT,    // height
    1000,   // mass
    0,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_BOSS_RAISE1   // raisestate
  },
  
  {   // MT_BRUISERSHOT
    -1,   // doomednum
    S_BRBALL1,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_firsht,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_BRBALLX1,   // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    15*FRACUNIT,   // speed
    6*FRACUNIT,   // radius
    8*FRACUNIT,   // height
    100,    // mass
    8,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_TRANSLUCENT,   // flags  killough 2/21/98
    S_NULL    // raisestate
  },
  
  {   // MT_KNIGHT
    69,   // doomednum
    S_BOS2_STND,    // spawnstate
    500,    // spawnhealth
    S_BOS2_RUN1,    // seestate
    sfx_kntsit,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_BOS2_PAIN,    // painstate
    50,   // painchance
    sfx_dmpain,   // painsound
    S_BOS2_ATK1,    // meleestate
    S_BOS2_ATK1,    // missilestate
    S_BOS2_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_kntdth,   // deathsound
    8,    // speed
    24*FRACUNIT,    // radius
    64*FRACUNIT,    // height
    1000,   // mass
    0,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_BOS2_RAISE1   // raisestate
  },

  {   // MT_SKULL
    3006,   // doomednum
    S_SKULL_STND,   // spawnstate
    100,    // spawnhealth
    S_SKULL_RUN1,   // seestate
    0,    // seesound
    8,    // reactiontime
    sfx_sklatk,   // attacksound
    S_SKULL_PAIN,   // painstate
    256,    // painchance
    sfx_dmpain,   // painsound
    0,    // meleestate
    S_SKULL_ATK1,   // missilestate
    S_SKULL_DIE1,   // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    8,    // speed
    16*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    50,   // mass
    3,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_FLOAT|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_SPIDER
    7,    // doomednum
    S_SPID_STND,    // spawnstate
    3000,   // spawnhealth
    S_SPID_RUN1,    // seestate
    sfx_spisit,   // seesound
    8,    // reactiontime
    sfx_shotgn,   // attacksound
    S_SPID_PAIN,    // painstate
    40,   // painchance
    sfx_dmpain,   // painsound
    0,    // meleestate
    S_SPID_ATK1,    // missilestate
    S_SPID_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_spidth,   // deathsound
    12,   // speed
    128*FRACUNIT,   // radius
    100*FRACUNIT,   // height
    1000,   // mass
    0,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_BABY
    68,   // doomednum
    S_BSPI_STND,    // spawnstate
    500,    // spawnhealth
    S_BSPI_SIGHT,   // seestate
    sfx_bspsit,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_BSPI_PAIN,    // painstate
    128,    // painchance
    sfx_dmpain,   // painsound
    0,    // meleestate
    S_BSPI_ATK1,    // missilestate
    S_BSPI_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_bspdth,   // deathsound
    12,   // speed
    64*FRACUNIT,    // radius
    64*FRACUNIT,    // height
    600,    // mass
    0,    // damage
    sfx_bspact,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_BSPI_RAISE1   // raisestate
  },
  
  {   // MT_CYBORG
    16,   // doomednum
    S_CYBER_STND,   // spawnstate
    4000,   // spawnhealth
    S_CYBER_RUN1,   // seestate
    sfx_cybsit,   // seesound
    8,    // reactiontime
    0,    // attacksound
    S_CYBER_PAIN,   // painstate
    20,   // painchance
    sfx_dmpain,   // painsound
    0,    // meleestate
    S_CYBER_ATK1,   // missilestate
    S_CYBER_DIE1,   // deathstate
    S_NULL,   // xdeathstate
    sfx_cybdth,   // deathsound
    16,   // speed
    40*FRACUNIT,    // radius
    110*FRACUNIT,   // height
    1000,   // mass
    0,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_PAIN
    71,   // doomednum
    S_PAIN_STND,    // spawnstate
    400,    // spawnhealth
    S_PAIN_RUN1,    // seestate
    sfx_pesit,    // seesound
    8,    // reactiontime
    0,    // attacksound
    S_PAIN_PAIN,    // painstate
    128,    // painchance
    sfx_pepain,   // painsound
    0,    // meleestate
    S_PAIN_ATK1,    // missilestate
    S_PAIN_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_pedth,    // deathsound
    8,    // speed
    31*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    400,    // mass
    0,    // damage
    sfx_dmact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_FLOAT|MF_NOGRAVITY|MF_COUNTKILL,   // flags
    S_PAIN_RAISE1   // raisestate
  },
  
  {   // MT_WOLFSS
    84,   // doomednum
    S_SSWV_STND,    // spawnstate
    50,   // spawnhealth
    S_SSWV_RUN1,    // seestate
    sfx_sssit,    // seesound
    8,    // reactiontime
    0,    // attacksound
    S_SSWV_PAIN,    // painstate
    170,    // painchance
    sfx_popain,   // painsound
    0,    // meleestate
    S_SSWV_ATK1,    // missilestate
    S_SSWV_DIE1,    // deathstate
    S_SSWV_XDIE1,   // xdeathstate
    sfx_ssdth,    // deathsound
    8,    // speed
    20*FRACUNIT,    // radius
    56*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_posact,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL,   // flags
    S_SSWV_RAISE1   // raisestate
  },
  
  {   // MT_KEEN
    72,   // doomednum
    S_KEENSTND,   // spawnstate
    100,    // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_KEENPAIN,   // painstate
    256,    // painchance
    sfx_keenpn,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_COMMKEEN,   // deathstate
    S_NULL,   // xdeathstate
    sfx_keendt,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    72*FRACUNIT,    // height
    10000000,   // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY|MF_SHOOTABLE|MF_COUNTKILL,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_BOSSBRAIN
    88,   // doomednum
    S_BRAIN,    // spawnstate
    250,    // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_BRAIN_PAIN,   // painstate
    255,    // painchance
    sfx_bospn,    // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_BRAIN_DIE1,   // deathstate
    S_NULL,   // xdeathstate
    sfx_bosdth,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    10000000,   // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SHOOTABLE,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_BOSSSPIT
    89,   // doomednum
    S_BRAINEYE,   // spawnstate
    1000,   // spawnhealth
    S_BRAINEYESEE,    // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    32*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOSECTOR,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_BOSSTARGET
    87,   // doomednum
    S_NULL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    32*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOSECTOR,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_SPAWNSHOT
    -1,   // doomednum
    S_SPAWN1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_bospit,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    10*FRACUNIT,    // speed
    6*FRACUNIT,   // radius
    32*FRACUNIT,    // height
    100,    // mass
    3,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_NOCLIP,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_SPAWNFIRE
    -1,   // doomednum
    S_SPAWNFIRE1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOGRAVITY|MF_TRANSLUCENT,   // flags             // phares
    S_NULL    // raisestate
  },
  
  {   // MT_BARREL
    2035,   // doomednum
    S_BAR1,   // spawnstate
    20,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_BEXP,   // deathstate
    S_NULL,   // xdeathstate
    sfx_barexp,   // deathsound
    0,    // speed
    10*FRACUNIT,    // radius
    42*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SHOOTABLE|MF_NOBLOOD,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_TROOPSHOT
    -1,   // doomednum
    S_TBALL1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_firsht,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_TBALLX1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    10*FRACUNIT,    // speed
    6*FRACUNIT,   // radius
    8*FRACUNIT,   // height
    100,    // mass
    3,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_TRANSLUCENT, // flags // phares
    S_NULL    // raisestate
  },
  
  {   // MT_HEADSHOT
    -1,   // doomednum
    S_RBALL1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_firsht,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_RBALLX1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    10*FRACUNIT,    // speed
    6*FRACUNIT,   // radius
    8*FRACUNIT,   // height
    100,    // mass
    5,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_TRANSLUCENT, // flags // phares,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_ROCKET
    -1,   // doomednum
    S_ROCKET,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_rlaunc,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_EXPLODE1,   // deathstate
    S_NULL,   // xdeathstate
    sfx_barexp,   // deathsound
    20*FRACUNIT,    // speed
    11*FRACUNIT,    // radius
    8*FRACUNIT,   // height
    100,    // mass
    20,   // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_PLASMA
    -1,   // doomednum
    S_PLASBALL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_plasma,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_PLASEXP,    // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    25*FRACUNIT,    // speed
    13*FRACUNIT,    // radius
    8*FRACUNIT,   // height
    100,    // mass
    5,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_TRANSLUCENT, // flags // phares
    S_NULL    // raisestate
  },
  
  {   // MT_BFG
    -1,   // doomednum
    S_BFGSHOT,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    0,    // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_BFGLAND,    // deathstate
    S_NULL,   // xdeathstate
    sfx_rxplod,   // deathsound
    25*FRACUNIT,    // speed
    13*FRACUNIT,    // radius
    8*FRACUNIT,   // height
    100,    // mass
    100,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_TRANSLUCENT, // flags // phares
    S_NULL    // raisestate
  },
  
  {   // MT_ARACHPLAZ
    -1,   // doomednum
    S_ARACH_PLAZ,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_plasma,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_ARACH_PLEX,   // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    25*FRACUNIT,    // speed
    13*FRACUNIT,    // radius
    8*FRACUNIT,   // height
    100,    // mass
    5,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_TRANSLUCENT, // flags // phares
    S_NULL    // raisestate
  },
  
  {   // MT_PUFF
    -1,   // doomednum
    S_PUFF1,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOGRAVITY|MF_TRANSLUCENT, // flags // phares
    S_NULL    // raisestate
  },
  
  {   // MT_BLOOD
    -1,   // doomednum
    S_BLOOD1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_TFOG
    -1,   // doomednum
    S_TFOG,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOGRAVITY|MF_TRANSLUCENT, // flags // phares
    S_NULL    // raisestate
  },
  
  {   // MT_IFOG
    -1,   // doomednum
    S_IFOG,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOGRAVITY|MF_TRANSLUCENT, // flags // phares
    S_NULL    // raisestate
  },
  
  {   // MT_TELEPORTMAN
    14,   // doomednum
    S_NULL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOSECTOR,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_EXTRABFG
    -1,   // doomednum
    S_BFGEXP,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_NOGRAVITY,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC0
    2018,   // doomednum
    S_ARM1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC1
    2019,   // doomednum
    S_ARM2,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC2
    2014,   // doomednum
    S_BON1,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC3
    2015,   // doomednum
    S_BON2,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC4
    5,    // doomednum
    S_BKEY,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_NOTDMATCH,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC5
    13,   // doomednum
    S_RKEY,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_NOTDMATCH,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC6
    6,    // doomednum
    S_YKEY,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_NOTDMATCH,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC7
    39,   // doomednum
    S_YSKULL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_NOTDMATCH,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC8
    38,   // doomednum
    S_RSKULL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_NOTDMATCH,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC9
    40,   // doomednum
    S_BSKULL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_NOTDMATCH,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC10
    2011,   // doomednum
    S_STIM,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC11
    2012,   // doomednum
    S_MEDI,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC12
    2013,   // doomednum
    S_SOUL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM|MF_TRANSLUCENT,    // flags   // killough 2/21/98
    S_NULL    // raisestate
  },
  
  {   // MT_INV
    2022,   // doomednum
    S_PINV,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM|MF_TRANSLUCENT,    // flags // killough 2/21/98
    S_NULL    // raisestate
  },
  
  {   // MT_MISC13
    2023,   // doomednum
    S_PSTR,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_INS
    2024,   // doomednum
    S_PINS,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM|MF_TRANSLUCENT,    // flags // killough 2/21/98
    S_NULL    // raisestate
  },
  
  {   // MT_MISC14
    2025,   // doomednum
    S_SUIT,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC15
    2026,   // doomednum
    S_PMAP,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC16
    2045,   // doomednum
    S_PVIS,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MEGA
    83,   // doomednum
    S_MEGA,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM|MF_TRANSLUCENT,    // flags // killough 2/21/98
    S_NULL    // raisestate
  },
  
  {   // MT_CLIP
    2007,   // doomednum
    S_CLIP,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC17
    2048,   // doomednum
    S_AMMO,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC18
    2010,   // doomednum
    S_ROCK,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC19
    2046,   // doomednum
    S_BROK,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC20
    2047,   // doomednum
    S_CELL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC21
    17,   // doomednum
    S_CELP,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC22
    2008,   // doomednum
    S_SHEL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC23
    2049,   // doomednum
    S_SBOX,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC24
    8,    // doomednum
    S_BPAK,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC25
    2006,   // doomednum
    S_BFUG,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_CHAINGUN
    2002,   // doomednum
    S_MGUN,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC26
    2005,   // doomednum
    S_CSAW,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC27
    2003,   // doomednum
    S_LAUN,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC28
    2004,   // doomednum
    S_PLAS,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_SHOTGUN
    2001,   // doomednum
    S_SHOT,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_SUPERSHOTGUN
    82,   // doomednum
    S_SHOT2,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC29
    85,   // doomednum
    S_TECHLAMP,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC30
    86,   // doomednum
    S_TECH2LAMP,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC31
    2028,   // doomednum
    S_COLU,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC32
    30,   // doomednum
    S_TALLGRNCOL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC33
    31,   // doomednum
    S_SHRTGRNCOL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC34
    32,   // doomednum
    S_TALLREDCOL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC35
    33,   // doomednum
    S_SHRTREDCOL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC36
    37,   // doomednum
    S_SKULLCOL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC37
    36,   // doomednum
    S_HEARTCOL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC38
    41,   // doomednum
    S_EVILEYE,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC39
    42,   // doomednum
    S_FLOATSKULL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC40
    43,   // doomednum
    S_TORCHTREE,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC41
    44,   // doomednum
    S_BLUETORCH,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC42
    45,   // doomednum
    S_GREENTORCH,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC43
    46,   // doomednum
    S_REDTORCH,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC44
    55,   // doomednum
    S_BTORCHSHRT,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC45
    56,   // doomednum
    S_GTORCHSHRT,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC46
    57,   // doomednum
    S_RTORCHSHRT,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC47
    47,   // doomednum
    S_STALAGTITE,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC48
    48,   // doomednum
    S_TECHPILLAR,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC49
    34,   // doomednum
    S_CANDLESTIK,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC50
    35,   // doomednum
    S_CANDELABRA,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC51
    49,   // doomednum
    S_BLOODYTWITCH,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    68*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC52
    50,   // doomednum
    S_MEAT2,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    84*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC53
    51,   // doomednum
    S_MEAT3,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    84*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC54
    52,   // doomednum
    S_MEAT4,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    68*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC55
    53,   // doomednum
    S_MEAT5,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    52*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC56
    59,   // doomednum
    S_MEAT2,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    84*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPAWNCEILING|MF_NOGRAVITY,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC57
    60,   // doomednum
    S_MEAT4,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    68*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPAWNCEILING|MF_NOGRAVITY,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC58
    61,   // doomednum
    S_MEAT3,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    52*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPAWNCEILING|MF_NOGRAVITY,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC59
    62,   // doomednum
    S_MEAT5,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    52*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPAWNCEILING|MF_NOGRAVITY,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC60
    63,   // doomednum
    S_BLOODYTWITCH,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    68*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPAWNCEILING|MF_NOGRAVITY,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC61
    22,   // doomednum
    S_HEAD_DIE6,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC62
    15,   // doomednum
    S_PLAY_DIE7,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC63
    18,   // doomednum
    S_POSS_DIE5,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC64
    21,   // doomednum
    S_SARG_DIE6,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC65
    23,   // doomednum
    S_SKULL_DIE6,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC66
    20,   // doomednum
    S_TROO_DIE5,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC67
    19,   // doomednum
    S_SPOS_DIE5,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC68
    10,   // doomednum
    S_PLAY_XDIE9,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC69
    12,   // doomednum
    S_PLAY_XDIE9,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC70
    28,   // doomednum
    S_HEADSONSTICK,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC71
    24,   // doomednum
    S_GIBS,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    0,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC72
    27,   // doomednum
    S_HEADONASTICK,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC73
    29,   // doomednum
    S_HEADCANDLES,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC74
    25,   // doomednum
    S_DEADSTICK,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC75
    26,   // doomednum
    S_LIVESTICK,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC76
    54,   // doomednum
    S_BIGTREE,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    32*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC77
    70,   // doomednum
    S_BBAR1,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID,   // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC78
    73,   // doomednum
    S_HANGNOGUTS,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    88*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC79
    74,   // doomednum
    S_HANGBNOBRAIN,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    88*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC80
    75,   // doomednum
    S_HANGTLOOKDN,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    64*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC81
    76,   // doomednum
    S_HANGTSKULL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    64*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC82
    77,   // doomednum
    S_HANGTLOOKUP,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    64*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC83
    78,   // doomednum
    S_HANGTNOBRAIN,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    16*FRACUNIT,    // radius
    64*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SOLID|MF_SPAWNCEILING|MF_NOGRAVITY,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC84
    79,   // doomednum
    S_COLONGIBS,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC85
    80,   // doomednum
    S_SMALLPOOL,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP,    // flags
    S_NULL    // raisestate
  },
  
  {   // MT_MISC86
    81,   // doomednum
    S_BRAINSTEM,    // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP,    // flags
    S_NULL    // raisestate
  },

  // For use with wind and current effects
  {   // MT_PUSH                       // phares
    5001,           // doomednum       //   |      //jff 5/11/98 deconflict
    S_TNT1,         // spawnstate      //   V      // with DOSDoom        
    1000,           // spawnhealth                              
    S_NULL,         // seestate                                 
    sfx_None,       // seesound                                 
    8,              // reactiontime                             
    sfx_None,       // attacksound                              
    S_NULL,         // painstate                                
    0,              // painchance                               
    sfx_None,       // painsound                                
    S_NULL,         // meleestate                               
    S_NULL,         // missilestate                             
    S_NULL,         // deathstate                               
    S_NULL,         // xdeathstate                              
    sfx_None,       // deathsound                               
    0,              // speed                                    
    8,              // radius                                   
    8,              // height                                   
    10,             // mass                                     
    0,              // damage                                   
    sfx_None,       // activesound                              
    MF_NOBLOCKMAP,  // flags
    S_NULL          // raisestate                                   
  },

  // For use with wind and current effects
  {   // MT_PULL
    5002,           // doomednum                   //jff 5/11/98 deconflict            
    S_TNT1,         // spawnstate                  // with DOSDoom        
    1000,           // spawnhealth                              
    S_NULL,         // seestate                                 
    sfx_None,       // seesound                                 
    8,              // reactiontime                             
    sfx_None,       // attacksound                              
    S_NULL,         // painstate                                
    0,              // painchance                               
    sfx_None,       // painsound                                
    S_NULL,         // meleestate                               
    S_NULL,         // missilestate                             
    S_NULL,         // deathstate                               
    S_NULL,         // xdeathstate                              
    sfx_None,       // deathsound                               
    0,              // speed                                    
    8,              // radius                                   
    8,              // height                                   
    10,             // mass                                     
    0,              // damage                                   
    sfx_None,       // activesound                              
    MF_NOBLOCKMAP,  // flags
    S_NULL          // raisestate                                   
  },

#ifdef DOGS
  // Marine's best friend :)      // killough 7/19/98
  {   // MT_DOGS
    888,   // doomednum
    S_DOGS_STND,    // spawnstate
    500,    // spawnhealth
    S_DOGS_RUN1,    // seestate
    sfx_dgsit,   // seesound
    8,    // reactiontime
    sfx_dgatk,   // attacksound
    S_DOGS_PAIN,    // painstate
    180,    // painchance
    sfx_dgpain,   // painsound
    S_DOGS_ATK1,    // meleestate
    0,    // missilestate
    S_DOGS_DIE1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_dgdth,   // deathsound
    10,   // speed
    12*FRACUNIT,    // radius
    28*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_dgact,    // activesound
    MF_SOLID|MF_SHOOTABLE|MF_COUNTKILL, // flags
    S_DOGS_RAISE1   // raisestate
  },
#endif

#ifdef BETA
  // killough 7/11/98: this is the first of two plasma fireballs in the beta
  {   // MT_PLASMA1
    -1,   // doomednum
    S_PLS1BALL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_plasma,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_PLS1EXP,    // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    25*FRACUNIT,    // speed
    13*FRACUNIT,    // radius
    8*FRACUNIT,   // height
    100,    // mass
    4,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_BOUNCES,
    S_NULL    // raisestate
  },
  
  // killough 7/11/98: this is the second of two plasma fireballs in the beta
  {   // MT_PLASMA2
    -1,   // doomednum
    S_PLS2BALL,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_plasma,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_PLS2BALLX1,    // deathstate
    S_NULL,   // xdeathstate
    sfx_firxpl,   // deathsound
    25*FRACUNIT,    // speed
    6*FRACUNIT,    // radius
    8*FRACUNIT,   // height
    100,    // mass
    4,    // damage
    sfx_None,   // activesound
    MF_NOBLOCKMAP|MF_MISSILE|MF_DROPOFF|MF_NOGRAVITY|MF_BOUNCES,
    S_NULL    // raisestate
  },

  // killough 7/11/98: this is the evil sceptre in the beta version
  {   // MT_SCEPTRE
    2016,   // doomednum
    S_BON3,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    10*FRACUNIT,    // radius
    16*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM,    // flags
    S_NULL    // raisestate
  },

  // killough 7/11/98: this is the unholy bible in the beta version
  {   // MT_BIBLE
    2017,   // doomednum
    S_BON4,   // spawnstate
    1000,   // spawnhealth
    S_NULL,   // seestate
    sfx_None,   // seesound
    8,    // reactiontime
    sfx_None,   // attacksound
    S_NULL,   // painstate
    0,    // painchance
    sfx_None,   // painsound
    S_NULL,   // meleestate
    S_NULL,   // missilestate
    S_NULL,   // deathstate
    S_NULL,   // xdeathstate
    sfx_None,   // deathsound
    0,    // speed
    20*FRACUNIT,    // radius
    10*FRACUNIT,    // height
    100,    // mass
    0,    // damage
    sfx_None,   // activesound
    MF_SPECIAL|MF_COUNTITEM,    // flags
    S_NULL    // raisestate
  },
#endif /* BETA */

};


// ********************************************************************
// Predefined lumps
// ********************************************************************
// These are internally defined lumps, which can be overwritten in 
// a PWAD.  There were several things that needed these, such 
// as new menus, font characters, color maps, etc. and this
// is an easy way of incorporating them directly into the EXE rather
// than requiring that an ancillary file be carried around with it.
//
// Near the end of this section there is the actual list where the lumps
// are incorporated, and there it's easy to see explanations in context.

// killough 1/31/98: predefined lumps used to make up for wad deficiencies

// Horizontal Sensitivity menu
static const char m_horsen[]={ // 2023/12/02 -- Harmony font version
  112,0,12,0,0,0,0,0,200,1,0,0,217,1,0,0,234,1,0,0,251,1,0,0,12,2,0,0,22,2,0,0,
  32,2,0,0,42,2,0,0,52,2,0,0,69,2,0,0,86,2,0,0,103,2,0,0,120,2,0,0,135,2,0,0,
  152,2,0,0,169,2,0,0,186,2,0,0,204,2,0,0,222,2,0,0,240,2,0,0,2,3,0,0,19,3,0,0,
  36,3,0,0,53,3,0,0,68,3,0,0,85,3,0,0,102,3,0,0,119,3,0,0,136,3,0,0,156,3,0,0,
  173,3,0,0,190,3,0,0,210,3,0,0,227,3,0,0,244,3,0,0,4,4,0,0,21,4,0,0,38,4,0,0,
  55,4,0,0,72,4,0,0,89,4,0,0,103,4,0,0,121,4,0,0,140,4,0,0,160,4,0,0,177,4,0,0,
  194,4,0,0,211,4,0,0,228,4,0,0,245,4,0,0,6,5,0,0,26,5,0,0,46,5,0,0,65,5,0,0,
  80,5,0,0,97,5,0,0,114,5,0,0,131,5,0,0,149,5,0,0,167,5,0,0,185,5,0,0,203,5,0,
  0,220,5,0,0,237,5,0,0,254,5,0,0,13,6,0,0,30,6,0,0,47,6,0,0,64,6,0,0,81,6,0,0,
  95,6,0,0,105,6,0,0,115,6,0,0,130,6,0,0,147,6,0,0,164,6,0,0,181,6,0,0,198,6,0,
  0,207,6,0,0,216,6,0,0,225,6,0,0,242,6,0,0,3,7,0,0,20,7,0,0,37,7,0,0,57,7,0,0,
  66,7,0,0,75,7,0,0,84,7,0,0,99,7,0,0,116,7,0,0,133,7,0,0,150,7,0,0,167,7,0,0,
  184,7,0,0,201,7,0,0,218,7,0,0,235,7,0,0,252,7,0,0,13,8,0,0,28,8,0,0,45,8,0,0,
  62,8,0,0,79,8,0,0,96,8,0,0,106,8,0,0,116,8,0,0,126,8,0,0,136,8,0,0,147,8,0,0,
  158,8,0,0,169,8,0,0,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  159,159,159,159,159,242,241,242,159,159,159,159,255,5,5,159,159,242,241,242,
  159,159,255,5,5,159,159,242,241,242,159,159,255,5,5,159,159,242,241,242,159,
  159,255,5,5,159,159,242,241,242,159,159,255,0,12,159,159,159,159,159,159,159,
  242,241,242,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,1,
  10,159,159,159,159,159,159,246,245,246,159,159,159,255,0,12,159,159,246,245,
  244,244,243,242,241,242,243,246,159,159,255,0,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,159,246,241,242,
  243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,243,244,159,
  159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,3,
  159,159,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,3,159,159,244,
  159,159,6,6,246,246,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,
  159,246,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,246,245,244,244,243,242,241,242,243,246,
  159,159,255,1,10,159,159,159,159,159,159,246,245,246,159,159,159,255,0,12,
  159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,243,242,241,
  242,243,244,159,159,255,0,3,159,159,244,159,159,4,8,159,159,243,242,241,246,
  159,159,159,159,255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,
  255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,255,0,3,159,159,
  244,159,159,4,8,159,159,243,242,241,246,159,159,159,159,255,0,12,159,159,244,
  159,159,159,243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,
  243,242,241,242,243,244,159,159,255,1,11,159,159,246,244,245,159,246,241,242,
  243,244,159,159,255,2,3,159,159,159,159,159,7,5,245,245,246,159,159,159,159,
  255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,245,244,
  243,242,241,242,243,244,159,159,255,3,9,159,159,159,159,246,245,246,159,159,
  159,159,255,0,4,159,159,159,159,159,159,7,5,245,245,246,159,159,159,159,255,
  0,4,159,159,244,245,159,159,6,6,246,246,242,242,243,244,159,159,255,0,4,159,
  159,244,245,159,159,5,7,159,159,244,241,242,243,244,159,159,255,0,12,159,159,
  244,159,159,159,244,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,
  245,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,244,243,243,
  245,242,243,244,159,159,255,0,12,159,159,244,159,245,244,243,246,246,242,243,
  244,159,159,255,0,12,159,159,244,159,244,244,246,246,245,242,243,244,159,159,
  255,0,12,159,159,244,245,244,245,159,246,241,242,243,244,159,159,255,0,5,159,
  159,244,245,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,5,159,159,
  244,245,245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,159,
  159,159,159,6,6,246,246,245,246,159,159,159,159,255,1,10,159,159,159,159,159,
  159,246,245,246,159,159,159,255,0,12,159,159,246,245,244,244,243,242,241,242,
  243,246,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,244,159,159,159,159,246,241,242,243,244,159,159,255,0,3,
  159,159,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,3,159,159,244,
  159,159,6,6,246,246,241,242,243,244,159,159,255,0,3,159,159,244,159,159,6,6,
  246,246,241,242,243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,
  242,243,244,159,159,255,0,12,159,159,244,159,159,159,159,246,241,242,243,244,
  159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  0,12,159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,1,10,159,
  159,159,159,159,159,246,245,246,159,159,159,255,0,12,159,159,159,159,159,159,
  159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,
  159,159,255,0,12,159,159,159,159,245,244,243,242,245,242,243,244,159,159,255,
  3,9,159,159,245,243,242,245,246,159,159,159,159,255,4,5,159,159,243,242,241,
  246,246,255,4,5,159,159,245,242,241,246,246,255,0,10,159,159,159,159,159,159,
  159,242,241,242,159,159,255,0,12,159,159,244,245,244,244,246,242,241,242,246,
  159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,
  159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,4,159,159,159,
  159,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,244,245,159,159,
  255,0,12,159,159,244,245,159,159,159,246,245,246,159,159,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,159,159,
  244,242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,
  246,245,246,159,159,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,
  244,245,159,159,255,0,4,159,159,159,159,159,159,255,2,10,159,159,159,159,159,
  246,245,246,159,159,159,159,255,0,12,159,159,246,245,244,244,243,242,241,242,
  243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,159,255,0,3,
  159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,
  159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,
  242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,
  159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  246,245,244,244,243,242,241,242,243,244,159,159,255,2,10,159,159,159,159,159,
  246,245,246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,
  159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,159,159,159,159,159,246,245,242,243,244,159,159,255,7,5,246,246,
  242,243,244,159,159,255,7,5,246,246,242,243,244,159,159,255,7,5,246,246,242,
  243,244,159,159,255,7,5,246,246,242,243,244,159,159,255,6,6,246,246,245,242,
  243,244,159,159,255,6,6,246,246,241,242,243,244,159,159,255,6,6,246,246,241,
  242,243,244,159,159,255,6,6,246,246,245,246,159,159,159,159,255};

// Vertical Sensitivity menu
static const char m_versen[]={ // 2023/12/02 -- Harmony font version
  88,0,12,0,0,0,0,0,104,1,0,0,116,1,0,0,130,1,0,0,145,1,0,0,161,1,0,0,173,1,0,
  0,184,1,0,0,195,1,0,0,207,1,0,0,223,1,0,0,238,1,0,0,252,1,0,0,8,2,0,0,25,2,0,
  0,42,2,0,0,59,2,0,0,76,2,0,0,96,2,0,0,116,2,0,0,136,2,0,0,156,2,0,0,175,2,0,
  0,194,2,0,0,213,2,0,0,232,2,0,0,249,2,0,0,10,3,0,0,27,3,0,0,44,3,0,0,64,3,0,
  0,81,3,0,0,98,3,0,0,118,3,0,0,135,3,0,0,152,3,0,0,168,3,0,0,185,3,0,0,194,3,
  0,0,203,3,0,0,212,3,0,0,229,3,0,0,246,3,0,0,7,4,0,0,24,4,0,0,44,4,0,0,53,4,0,
  0,62,4,0,0,71,4,0,0,88,4,0,0,105,4,0,0,122,4,0,0,139,4,0,0,153,4,0,0,168,4,0,
  0,185,4,0,0,202,4,0,0,219,4,0,0,236,4,0,0,253,4,0,0,14,5,0,0,31,5,0,0,50,5,0,
  0,69,5,0,0,88,5,0,0,105,5,0,0,120,5,0,0,137,5,0,0,154,5,0,0,171,5,0,0,188,5,
  0,0,205,5,0,0,222,5,0,0,239,5,0,0,0,6,0,0,17,6,0,0,34,6,0,0,49,6,0,0,66,6,0,
  0,83,6,0,0,100,6,0,0,117,6,0,0,127,6,0,0,137,6,0,0,147,6,0,0,157,6,0,0,168,6,
  0,0,179,6,0,0,190,6,0,0,0,7,159,159,159,159,159,159,159,246,246,255,0,9,159,
  159,244,245,244,244,243,244,245,246,246,255,0,10,159,159,244,245,244,244,243,
  242,241,244,159,159,255,0,11,159,159,159,159,159,159,243,242,241,242,244,159,
  159,255,5,7,159,159,242,241,242,243,246,159,159,255,6,6,246,246,241,242,243,
  244,159,159,255,6,6,246,246,241,242,243,244,159,159,255,5,7,159,159,242,241,
  242,243,246,159,159,255,0,11,159,159,159,159,159,159,243,242,241,242,244,159,
  159,255,0,10,159,159,246,245,244,244,243,242,241,244,159,159,255,0,9,159,159,
  246,245,244,244,243,244,245,246,246,255,0,7,159,159,159,159,159,159,159,246,
  246,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,
  244,159,246,245,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,
  7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,
  246,246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,
  246,242,243,244,159,159,255,0,6,159,159,244,159,159,159,159,159,7,5,246,246,
  242,243,244,159,159,255,0,4,159,159,244,159,159,159,6,6,246,246,245,242,243,
  244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,
  159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,
  0,4,159,159,159,159,159,159,6,6,246,246,245,246,159,159,159,159,255,0,12,159,
  159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,243,242,241,242,
  243,244,159,159,255,0,3,159,159,244,159,159,4,8,159,159,243,242,241,246,159,
  159,159,159,255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,255,
  0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,255,0,3,159,159,244,
  159,159,4,8,159,159,243,242,241,246,159,159,159,159,255,0,12,159,159,244,159,
  159,159,243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,243,
  242,241,242,243,244,159,159,255,1,11,159,159,246,244,245,159,246,241,242,243,
  244,159,159,255,2,3,159,159,159,159,159,7,5,245,245,246,159,159,159,159,255,
  0,4,159,159,159,159,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,
  244,245,159,159,255,0,12,159,159,244,245,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,159,159,244,242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,
  159,5,7,159,159,246,245,246,159,159,159,159,255,0,4,159,159,244,245,159,159,
  255,0,4,159,159,244,245,159,159,255,0,4,159,159,159,159,159,159,255,0,12,159,
  159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,159,159,245,244,243,242,241,242,
  243,244,159,159,255,3,9,159,159,159,159,246,245,246,159,159,159,159,255,1,10,
  159,159,159,159,159,159,246,245,246,159,159,159,255,0,12,159,159,246,245,244,
  244,243,242,241,242,243,246,159,159,255,0,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,159,246,245,242,243,
  244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,
  0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,
  159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,
  246,242,243,244,159,159,255,0,4,159,159,244,159,159,159,6,6,246,246,245,242,
  243,244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,
  159,159,255,0,4,159,159,246,245,159,159,6,6,246,246,241,242,243,246,159,159,
  255,1,3,159,159,159,159,159,6,5,246,246,245,246,159,159,159,255,2,10,159,159,
  159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,246,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,
  159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,
  159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,
  5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,
  241,242,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,2,10,159,159,
  159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,
  246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,159,159,159,159,159,246,245,242,243,244,159,159,255,7,5,
  246,246,242,243,244,159,159,255,7,5,246,246,242,243,244,159,159,255,7,5,246,
  246,242,243,244,159,159,255,7,5,246,246,242,243,244,159,159,255,6,6,246,246,
  245,242,243,244,159,159,255,6,6,246,246,241,242,243,244,159,159,255,6,6,246,
  246,241,242,243,244,159,159,255,6,6,246,246,245,246,159,159,159,159,255};

static const char m_generl[] = { // 2023/12/02 -- Harmony font version
  84,0,12,0,0,0,0,0,88,1,0,0,103,1,0,0,120,1,0,0,137,1,0,0,154,1,0,0,171,1,0,0,
  188,1,0,0,205,1,0,0,228,1,0,0,248,1,0,0,12,2,0,0,25,2,0,0,37,2,0,0,54,2,0,0,
  71,2,0,0,88,2,0,0,105,2,0,0,125,2,0,0,145,2,0,0,165,2,0,0,185,2,0,0,204,2,0,
  0,223,2,0,0,242,2,0,0,5,3,0,0,22,3,0,0,39,3,0,0,56,3,0,0,73,3,0,0,87,3,0,0,
  97,3,0,0,107,3,0,0,122,3,0,0,139,3,0,0,156,3,0,0,173,3,0,0,190,3,0,0,207,3,0,
  0,224,3,0,0,241,3,0,0,2,4,0,0,22,4,0,0,42,4,0,0,62,4,0,0,82,4,0,0,101,4,0,0,
  120,4,0,0,139,4,0,0,158,4,0,0,175,4,0,0,192,4,0,0,209,4,0,0,226,4,0,0,246,4,
  0,0,7,5,0,0,24,5,0,0,44,5,0,0,61,5,0,0,78,5,0,0,94,5,0,0,111,5,0,0,126,5,0,0,
  143,5,0,0,160,5,0,0,177,5,0,0,194,5,0,0,211,5,0,0,228,5,0,0,245,5,0,0,6,6,0,
  0,23,6,0,0,40,6,0,0,55,6,0,0,72,6,0,0,89,6,0,0,106,6,0,0,123,6,0,0,133,6,0,0,
  143,6,0,0,153,6,0,0,163,6,0,0,174,6,0,0,185,6,0,0,196,6,0,0,1,10,159,159,159,
  159,159,159,246,245,246,159,159,159,255,0,12,159,159,246,245,244,244,243,242,
  241,242,243,246,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,159,159,159,159,159,245,242,243,244,159,159,
  255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,
  244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,
  246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,4,2,159,159,159,159,
  7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,4,8,159,159,243,
  159,245,242,243,244,159,159,255,0,3,159,159,159,159,159,4,8,159,159,243,242,
  241,242,243,244,159,159,255,4,8,159,159,243,242,241,242,243,246,159,159,255,
  4,7,159,159,159,246,245,246,159,159,159,255,0,12,159,159,159,159,159,159,159,
  246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,244,159,159,244,159,246,245,242,243,244,159,159,255,0,6,
  159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,
  159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,
  244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,
  159,159,159,159,159,7,5,246,246,242,243,244,159,159,255,0,4,159,159,244,159,
  159,159,6,6,246,246,245,242,243,244,159,159,255,0,4,159,159,244,245,159,159,
  6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,
  246,241,242,243,244,159,159,255,0,4,159,159,159,159,159,159,6,6,246,246,245,
  246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,
  159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,
  159,159,159,245,244,243,242,245,242,243,244,159,159,255,3,9,159,159,245,243,
  242,245,246,159,159,159,159,255,4,5,159,159,243,242,241,246,246,255,4,5,159,
  159,245,242,241,246,246,255,0,10,159,159,159,159,159,159,159,242,241,242,159,
  159,255,0,12,159,159,244,245,244,244,246,242,241,242,246,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,
  159,159,246,245,246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,159,159,244,159,246,245,242,243,244,159,159,255,0,6,159,
  159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,
  244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,
  159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,159,
  159,159,159,159,7,5,246,246,242,243,244,159,159,255,0,4,159,159,244,159,159,
  159,6,6,246,246,245,242,243,244,159,159,255,0,4,159,159,244,245,159,159,6,6,
  246,246,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,246,
  241,242,243,244,159,159,255,0,4,159,159,159,159,159,159,6,6,246,246,245,246,
  159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,159,159,159,243,242,241,242,243,244,159,159,255,0,3,159,159,244,159,159,
  4,8,159,159,243,242,241,246,159,159,159,159,255,0,3,159,159,244,159,159,4,5,
  159,159,243,242,241,246,246,255,0,3,159,159,244,159,159,4,5,159,159,243,242,
  241,246,246,255,0,3,159,159,244,159,159,4,8,159,159,243,242,241,246,159,159,
  159,159,255,0,12,159,159,244,159,159,159,243,242,241,242,243,244,159,159,255,
  0,12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,1,11,159,
  159,246,244,245,159,246,241,242,243,244,159,159,255,2,3,159,159,159,159,159,
  7,5,245,245,246,159,159,159,159,255,2,10,159,159,159,159,159,246,245,246,159,
  159,159,159,255,0,12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,
  159,159,244,159,159,159,159,242,241,242,159,159,159,159,255,0,3,159,159,244,
  159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,
  159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,
  159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,12,
  159,159,244,159,159,159,159,242,241,242,159,159,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,
  243,242,241,242,243,244,159,159,255,2,10,159,159,159,159,159,246,245,246,159,
  159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,
  255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,
  159,159,159,159,246,245,242,243,244,159,159,255,7,5,246,246,242,243,244,159,
  159,255,7,5,246,246,242,243,244,159,159,255,7,5,246,246,242,243,244,159,159,
  255,7,5,246,246,242,243,244,159,159,255,6,6,246,246,245,242,243,244,159,159,
  255,6,6,246,246,241,242,243,244,159,159,255,6,6,246,246,241,242,243,244,159,
  159,255,6,6,246,246,245,246,159,159,159,159,255,
};

static const char m_compat[] = { // 2023/12/02 -- Harmony font version
  135,0,12,0,0,0,0,0,36,2,0,0,51,2,0,0,68,2,0,0,85,2,0,0,102,2,0,0,119,2,0,0,
  136,2,0,0,153,2,0,0,170,2,0,0,189,2,0,0,208,2,0,0,227,2,0,0,244,2,0,0,3,3,0,
  0,20,3,0,0,37,3,0,0,54,3,0,0,72,3,0,0,90,3,0,0,108,3,0,0,126,3,0,0,143,3,0,0,
  160,3,0,0,177,3,0,0,192,3,0,0,209,3,0,0,226,3,0,0,243,3,0,0,4,4,0,0,20,4,0,0,
  30,4,0,0,40,4,0,0,50,4,0,0,60,4,0,0,76,4,0,0,93,4,0,0,110,4,0,0,127,4,0,0,
  144,4,0,0,161,4,0,0,178,4,0,0,195,4,0,0,212,4,0,0,232,4,0,0,249,4,0,0,10,5,0,
  0,27,5,0,0,41,5,0,0,55,5,0,0,69,5,0,0,81,5,0,0,96,5,0,0,113,5,0,0,130,5,0,0,
  147,5,0,0,164,5,0,0,181,5,0,0,198,5,0,0,215,5,0,0,232,5,0,0,249,5,0,0,10,6,0,
  0,25,6,0,0,34,6,0,0,43,6,0,0,52,6,0,0,69,6,0,0,86,6,0,0,103,6,0,0,120,6,0,0,
  140,6,0,0,149,6,0,0,158,6,0,0,167,6,0,0,184,6,0,0,201,6,0,0,218,6,0,0,235,6,
  0,0,249,6,0,0,10,7,0,0,27,7,0,0,44,7,0,0,61,7,0,0,81,7,0,0,101,7,0,0,121,7,0,
  0,141,7,0,0,158,7,0,0,175,7,0,0,194,7,0,0,203,7,0,0,220,7,0,0,237,7,0,0,254,
  7,0,0,15,8,0,0,29,8,0,0,46,8,0,0,63,8,0,0,80,8,0,0,97,8,0,0,107,8,0,0,117,8,
  0,0,127,8,0,0,137,8,0,0,148,8,0,0,159,8,0,0,170,8,0,0,181,8,0,0,198,8,0,0,
  215,8,0,0,232,8,0,0,249,8,0,0,7,9,0,0,16,9,0,0,25,9,0,0,34,9,0,0,51,9,0,0,68,
  9,0,0,85,9,0,0,102,9,0,0,122,9,0,0,131,9,0,0,140,9,0,0,149,9,0,0,168,9,0,0,
  188,9,0,0,208,9,0,0,228,9,0,0,245,9,0,0,6,10,0,0,23,10,0,0,40,10,0,0,57,10,0,
  0,74,10,0,0,91,10,0,0,1,10,159,159,159,159,159,159,246,245,246,159,159,159,
  255,0,12,159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,
  159,159,159,159,246,245,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,
  246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,
  244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,
  0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,4,159,159,244,
  159,159,159,6,6,246,246,245,242,243,244,159,159,255,0,4,159,159,244,245,159,
  159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,246,245,159,159,6,6,
  246,246,241,242,243,246,159,159,255,1,3,159,159,159,159,159,6,5,246,246,245,
  246,159,159,159,255,1,10,159,159,159,159,159,159,246,245,246,159,159,159,255,
  0,12,159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,
  159,159,159,246,241,242,243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,
  246,241,242,243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,
  243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,243,244,159,
  159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,12,
  159,159,244,159,159,159,159,246,241,242,243,244,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,
  243,242,241,242,243,246,159,159,255,1,10,159,159,159,159,159,159,246,245,246,
  159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,
  255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,245,
  245,244,159,246,242,241,242,243,244,159,159,255,1,11,159,159,245,244,244,159,
  246,245,246,159,159,159,159,255,2,5,159,159,245,244,243,246,246,255,3,5,159,
  159,244,243,243,246,246,255,3,5,159,159,244,243,243,246,246,255,2,5,159,159,
  245,244,243,246,246,255,1,11,159,159,245,244,244,159,246,245,246,159,159,159,
  159,255,0,12,159,159,245,245,244,159,246,242,241,242,243,244,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,
  159,159,246,245,246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,159,159,159,243,242,241,242,243,244,159,159,255,0,3,159,
  159,244,159,159,4,8,159,159,243,242,241,246,159,159,159,159,255,0,3,159,159,
  244,159,159,4,5,159,159,243,242,241,246,246,255,0,3,159,159,244,159,159,4,5,
  159,159,243,242,241,246,246,255,0,3,159,159,244,159,159,4,5,159,159,243,242,
  241,246,246,255,0,9,159,159,244,159,159,159,243,242,241,246,246,255,0,9,159,
  159,244,245,244,244,243,242,241,246,246,255,0,9,159,159,246,245,244,244,243,
  242,244,246,246,255,1,7,159,159,159,159,159,159,246,245,245,255,2,10,159,159,
  159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,246,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,
  159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,
  159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,
  5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,
  241,242,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,2,10,159,159,
  159,159,159,246,245,246,159,159,159,159,255,0,4,159,159,159,159,159,159,255,
  0,4,159,159,244,245,159,159,255,0,4,159,159,244,245,159,159,255,0,12,159,159,
  244,245,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,244,245,159,159,244,242,241,242,243,
  244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,246,245,246,159,159,
  159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,244,245,159,159,255,
  0,4,159,159,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,
  159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,159,159,245,244,243,242,241,242,243,244,159,159,255,3,9,159,159,
  159,159,246,245,246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,159,159,244,159,159,245,242,243,244,159,159,255,0,6,159,
  159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,
  244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,
  159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,159,
  159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,12,159,159,244,245,244,
  244,243,159,245,242,243,244,159,159,255,0,12,159,159,244,245,244,159,243,242,
  241,242,243,244,159,159,255,1,3,159,159,159,159,159,5,7,159,159,246,242,242,
  243,245,159,159,255,7,4,245,245,246,159,159,159,255,0,12,159,159,159,159,159,
  159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,159,159,245,244,243,242,241,242,243,244,159,159,
  255,3,9,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,159,159,
  159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,0,12,159,159,159,159,159,159,159,246,245,242,243,244,159,
  159,255,7,5,246,246,242,243,244,159,159,255,7,5,246,246,242,243,244,159,159,
  255,7,5,246,246,242,243,244,159,159,255,7,5,246,246,242,243,244,159,159,255,
  6,6,246,246,245,242,243,244,159,159,255,6,6,246,246,241,242,243,244,159,159,
  255,6,6,246,246,241,242,243,244,159,159,255,6,6,246,246,245,246,159,159,159,
  159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,245,
  244,243,242,241,242,243,244,159,159,255,3,9,159,159,159,159,246,245,246,159,
  159,159,159,255,0,4,159,159,159,159,159,159,255,0,4,159,159,244,245,159,159,
  255,0,4,159,159,244,245,159,159,255,0,12,159,159,244,245,159,159,159,246,245,
  246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,
  159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  0,12,159,159,244,245,159,159,244,242,241,242,243,244,159,159,255,0,4,159,159,
  244,245,159,159,5,7,159,159,246,245,246,159,159,159,159,255,0,4,159,159,244,
  245,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,159,159,159,159,
  255,0,5,159,159,159,159,159,159,159,6,5,246,246,245,246,159,159,159,255,0,5,
  159,159,244,245,244,159,159,6,6,246,246,241,242,243,245,159,159,255,0,5,159,
  159,244,245,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,5,159,159,
  159,159,244,159,159,6,6,246,246,245,242,243,244,159,159,255,2,3,159,159,244,
  159,159,7,5,246,246,242,243,244,159,159,255,2,3,159,159,244,159,159,7,5,246,
  246,242,243,244,159,159,255,2,3,159,159,244,159,159,7,5,246,246,242,243,244,
  159,159,255,2,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,12,
  159,159,159,159,244,159,159,246,245,242,243,244,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,245,159,159,255,0,11,159,159,159,159,159,159,159,246,245,
  246,159,159,159,255,
};

//jff 02/16/98 HUD bargraph patches
// The HUD uses vertical stripe characters to make up the bars in
// view during play, and those are made up here at various stages
// of "fullness".  Note that all HUD elements can be recolored
// dynamically in the code, so only the shapes are defined here.

static const char stbr123[]=  // full
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 52,  0,  0,  0, 53,  0,  0,  0, 65,  0,  0,  0,
   66,  0,  0,  0, 78,  0,  0,  0, 79,  0,  0,  0, 91,  0,  0,  0,
    0,  7,  0, 47,185,180,176,180,185, 47,  0,255,255,  0,  7,  0,
   47,185,180,176,180,185, 47,  0,255,255,  0,  7,  0, 47,185,180,
  176,180,185, 47,  0,255,255,  0,  7,  0, 47,185,180,176,180,185,
   47,  0,255,255,
};

static const char stbr124[]=  // 3/4 full
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 52,  0,  0,  0, 53,  0,  0,  0, 65,  0,  0,  0,
   66,  0,  0,  0, 78,  0,  0,  0, 79,  0,  0,  0, 80,  0,  0,  0,
    0,  7,  0, 47,185,180,176,180,185, 47,  0,255,255,  0,  7,  0,
   47,185,180,176,180,185, 47,  0,255,255,  0,  7,  0, 47,185,180,
  176,180,185, 47,  0,255,255,255,255,
};

static const char stbr125[]=  // half full
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 52,  0,  0,  0, 53,  0,  0,  0, 65,  0,  0,  0,
   66,  0,  0,  0, 67,  0,  0,  0, 68,  0,  0,  0, 69,  0,  0,  0,
    0,  7,  0, 47,185,180,176,180,185, 47,  0,255,255,  0,  7,  0,
   47,185,180,176,180,185, 47,  0,255,255,255,255,255,255,
};

static const char stbr126[]=  // quarter full
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 52,  0,  0,  0, 53,  0,  0,  0, 54,  0,  0,  0,
   55,  0,  0,  0, 56,  0,  0,  0, 57,  0,  0,  0, 58,  0,  0,  0,
    0,  7,  0, 47,185,180,176,180,185, 47,  0,255,255,255,255,255,
  255,255,255,
};

static const char stbr127[]=  // empty block
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 41,  0,  0,  0, 42,  0,  0,  0, 43,  0,  0,  0,
   44,  0,  0,  0, 45,  0,  0,  0, 46,  0,  0,  0, 47,  0,  0,  0,
  255,255,255,255,255,255,255,255,
};

// The digits themselves that are used in the HUD

static const char dig0[]=  // zero
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 62,  0,  0,  0,
   73,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,  0,  0,176,  0,  0,255,  1,  2,  0,176,  0,  0,  5,
    2,  0,176,  0,  0,255,  1,  6,  0,176,176,176,176,176,  0,  0,
  255,  2,  5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char dig1[]=  // one
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 34,  0,  0,  0, 46,  0,  0,  0, 57,  0,  0,  0,
   67,  0,  0,  0,  5,  1,  0,176,  0,255,  1,  1,  0,176,  0,  5,
    2,  0,176,  0,  0,255,  1,  6,  0,176,176,176,176,176,  0,  0,
  255,  2,  5,  0,  0,  0,  0,176,  0,  0,255,  4,  3,  0,  0,  0,
    0,  0,255,
};

static const char dig2[]=  // two
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 41,  0,  0,  0, 52,  0,  0,  0, 63,  0,  0,  0,
   74,  0,  0,  0,  1,  1,  0,176,  0,  3,  3,  0,176,176,176,  0,
  255,  1,  6,  0,176,  0,176,  0,176,  0,  0,255,  1,  6,  0,176,
    0,176,  0,176,  0,  0,255,  1,  6,  0,176,176,176,  0,176,  0,
    0,255,  2,  3,  0,  0,  0,  0,  0,  6,  1,  0,  0,  0,255,
};

static const char dig3[]=  // three
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 41,  0,  0,  0, 55,  0,  0,  0, 69,  0,  0,  0,
   80,  0,  0,  0,  1,  2,  0,176,176,  0,  4,  2,  0,176,176,  0,
  255,  1,  3,  0,176,  0,  0,  0,  5,  2,  0,176,  0,  0,255,  1,
    3,  0,176,  0,176,  0,  5,  2,  0,176,  0,  0,255,  1,  6,  0,
  176,176,176,176,176,  0,  0,255,  2,  5,  0,  0,  0,  0,  0,  0,
    0,255,
};

static const char dig4[]=  // four
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 37,  0,  0,  0, 46,  0,  0,  0, 57,  0,  0,  0,
   67,  0,  0,  0,  1,  4,  0,176,176,176,176,  0,255,  2,  4,  0,
    0,  0,176,  0,  0,255,  1,  6,  0,176,176,176,176,176,  0,  0,
  255,  2,  5,  0,  0,  0,176,  0,  0,  0,255,  4,  2,  0,  0,  0,
    0,255,
};

static const char dig45[]=  // minus '-'
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 29,  0,  0,  0, 35,  0,  0,  0, 42,  0,  0,  0,
   49,  0,  0,  0,255,  3,  1,  0,176,  0,255,  3,  2,  0,176,  0,
    0,255,  3,  2,  0,176,  0,  0,255,  4,  1,  0,  0,  0,255,
};


static const char dig47[]=  // slash '/'
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 29,  0,  0,  0, 36,  0,  0,  0, 43,  0,  0,  0,
   50,  0,  0,  0,255,  4,  2,  0,176,  0,  0,255,  3,  2,  0,176,
    0,  0,255,  2,  2,  0,176,  0,  0,255,255,
};

static const char dig5[]=  // five
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 41,  0,  0,  0, 52,  0,  0,  0, 63,  0,  0,  0,
   74,  0,  0,  0,  1,  3,  0,176,176,176,  0,  5,  1,  0,176,  0,
  255,  1,  6,  0,176,  0,176,  0,176,  0,  0,255,  1,  6,  0,176,
    0,176,  0,176,  0,  0,255,  1,  6,  0,176,  0,176,176,176,  0,
    0,255,  2,  5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char dig58[]=  // colon ':'
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 29,  0,  0,  0, 30,  0,  0,  0, 41,  0,  0,  0,
   52,  0,  0,  0,255,255,  2,  1,  0,176,  0,  4,  1,  0,176,  0,
  255,  3,  1,  0,  0,  0,  5,  1,  0,  0,  0,255,255,
};

static const char dig6[]=  // six
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 60,  0,  0,  0,
   71,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,176,  0,176,  0,  0,255,  1,  6,  0,176,  0,176,  0,
  176,  0,  0,255,  1,  6,  0,176,  0,176,176,176,  0,  0,255,  2,
    5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char dig7[]=  // seven
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 34,  0,  0,  0, 41,  0,  0,  0, 51,  0,  0,  0,
   62,  0,  0,  0,  1,  1,  0,176,  0,255,  1,  2,  0,176,  0,  0,
  255,  1,  5,  0,176,  0,176,176,176,  0,255,  1,  6,  0,176,176,
    0,  0,  0,  0,  0,255,  2,  2,  0,  0,  0,  0,255,
};

static const char dig8[]=  // eight
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 60,  0,  0,  0,
   71,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,176,  0,176,  0,  0,255,  1,  6,  0,176,  0,176,  0,
  176,  0,  0,255,  1,  6,  0,176,176,176,176,176,  0,  0,255,  2,
    5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char dig9[]=  // nine
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 41,  0,  0,  0, 52,  0,  0,  0, 63,  0,  0,  0,
   74,  0,  0,  0,  1,  3,  0,176,176,176,  0,  5,  1,  0,176,  0,
  255,  1,  6,  0,176,  0,176,  0,176,  0,  0,255,  1,  6,  0,176,
    0,176,  0,176,  0,  0,255,  1,  6,  0,176,176,176,176,176,  0,
    0,255,  2,  5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char dig91[]=  // open bracket '['
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 29,  0,  0,  0, 39,  0,  0,  0, 50,  0,  0,  0,
   61,  0,  0,  0,255,  1,  5,  0,176,176,176,176,176,  0,255,  1,
    6,  0,176,  0,  0,  0,176,  0,  0,255,  2,  1,  0,  0,  0,  6,
    1,  0,  0,  0,255,255,
};

static const char dig93[]=  // close bracket ']'
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 29,  0,  0,  0, 40,  0,  0,  0, 51,  0,  0,  0,
   61,  0,  0,  0,255,  1,  1,  0,176,  0,  5,  1,  0,176,  0,255,
    1,  6,  0,176,176,176,176,176,  0,  0,255,  2,  5,  0,  0,  0,
    0,  0,  0,  0,255,255,
};

static const char diga[]=  // these are the letters A-Z
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 58,  0,  0,  0,
   68,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,176,  0,  0,  0,  0,255,  1,  4,  0,176,  0,176,  0,
    0,255,  1,  5,  0,176,176,176,176,176,  0,255,  2,  5,  0,  0,
    0,  0,  0,  0,  0,255,
};

static const char digb[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 60,  0,  0,  0,
   70,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,176,  0,176,  0,  0,255,  1,  6,  0,176,176,176,  0,
  176,  0,  0,255,  2,  5,  0,  0,176,176,176,  0,  0,255,  3,  4,
    0,  0,  0,  0,  0,  0,255,
};

static const char digc[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 62,  0,  0,  0,
   75,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,  0,  0,176,  0,  0,255,  1,  2,  0,176,  0,  0,  5,
    2,  0,176,  0,  0,255,  1,  2,  0,176,  0,  0,  5,  2,  0,176,
    0,  0,255,  2,  1,  0,  0,  0,  6,  1,  0,  0,  0,255,
};

static const char digd[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 62,  0,  0,  0,
   72,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,  0,  0,176,  0,  0,255,  1,  2,  0,176,  0,  0,  5,
    2,  0,176,  0,  0,255,  2,  5,  0,176,176,176,  0,  0,  0,255,
    3,  3,  0,  0,  0,  0,  0,255,
};

static const char dige[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 60,  0,  0,  0,
   74,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,176,  0,176,  0,  0,255,  1,  6,  0,176,  0,176,  0,
  176,  0,  0,255,  1,  2,  0,176,  0,  0,  4,  3,  0,  0,176,  0,
    0,255,  2,  1,  0,  0,  0,  6,  1,  0,  0,  0,255,
};

static const char digf[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 58,  0,  0,  0,
   70,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,176,  0,  0,  0,  0,255,  1,  4,  0,176,  0,176,  0,
    0,255,  1,  2,  0,176,  0,  0,  4,  1,  0,  0,  0,255,  2,  1,
    0,  0,  0,255,
};

static const char digg[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 63,  0,  0,  0,
   74,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,  0,  0,176,  0,  0,255,  1,  3,  0,176,  0,176,  0,
    5,  2,  0,176,  0,  0,255,  1,  6,  0,176,  0,176,176,176,  0,
    0,255,  2,  1,  0,  0,  0,  4,  3,  0,  0,  0,  0,  0,255,
};

static const char digh[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 48,  0,  0,  0, 55,  0,  0,  0,
   65,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  2,  5,
    0,  0,176,  0,  0,  0,  0,255,  3,  2,  0,176,  0,  0,255,  1,
    5,  0,176,176,176,176,176,  0,255,  2,  5,  0,  0,  0,  0,  0,
    0,  0,255,
};

static const char digi[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 29,  0,  0,  0, 40,  0,  0,  0, 51,  0,  0,  0,
   62,  0,  0,  0,255,  1,  1,  0,176,  0,  5,  1,  0,176,  0,255,
    1,  6,  0,176,176,176,176,176,  0,  0,255,  1,  6,  0,176,  0,
    0,  0,176,  0,  0,255,  5,  2,  0,  0,  0,  0,255,
};

static const char digj[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 40,  0,  0,  0, 53,  0,  0,  0, 64,  0,  0,  0,
   75,  0,  0,  0,  1,  1,  0,176,  0,  4,  2,  0,176,176,  0,255,
    1,  2,  0,176,  0,  0,  5,  2,  0,176,  0,  0,255,  1,  6,  0,
  176,176,176,176,176,  0,  0,255,  1,  6,  0,176,  0,  0,  0,  0,
    0,  0,255,  2,  1,  0,  0,  0,255,
};

static const char digk[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 48,  0,  0,  0, 57,  0,  0,  0,
   68,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  2,  5,
    0,  0,176,  0,  0,  0,  0,255,  1,  4,  0,176,176,176,  0,  0,
  255,  1,  6,  0,176,  0,176,176,176,  0,  0,255,  2,  1,  0,  0,
    0,  4,  3,  0,  0,  0,  0,  0,255,
};

static const char digl[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 48,  0,  0,  0, 55,  0,  0,  0,
   62,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  2,  5,
    0,  0,  0,  0,176,  0,  0,255,  5,  2,  0,176,  0,  0,255,  5,
    2,  0,176,  0,  0,255,  6,  1,  0,  0,  0,255,
};

static const char digm[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 48,  0,  0,  0, 56,  0,  0,  0,
   66,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  2,  5,
    0,176,176,  0,  0,  0,  0,255,  2,  3,  0,176,176,  0,  0,255,
    1,  5,  0,176,176,176,176,176,  0,255,  2,  5,  0,  0,  0,  0,
    0,  0,  0,255,
};

static const char dign[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 48,  0,  0,  0, 55,  0,  0,  0,
   65,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  2,  5,
    0,176,176,  0,  0,  0,  0,255,  3,  2,  0,176,176,  0,255,  1,
    5,  0,176,176,176,176,176,  0,255,  2,  5,  0,  0,  0,  0,  0,
    0,  0,255,
};

static const char digo[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 62,  0,  0,  0,
   73,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,  0,  0,176,  0,  0,255,  1,  2,  0,176,  0,  0,  5,
    2,  0,176,  0,  0,255,  1,  6,  0,176,176,176,176,176,  0,  0,
  255,  2,  5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char digp[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 58,  0,  0,  0,
   67,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,176,  0,  0,  0,  0,255,  1,  4,  0,176,  0,176,  0,
    0,255,  1,  4,  0,176,176,176,  0,  0,255,  2,  3,  0,  0,  0,
    0,  0,255,
};

static const char digq[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 63,  0,  0,  0,
   74,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,  0,176,176,  0,  0,255,  1,  2,  0,176,  0,  0,  4,
    3,  0,  0,176,176,  0,255,  1,  6,  0,176,176,176,176,176,  0,
    0,255,  2,  5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char digr[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 49,  0,  0,  0, 58,  0,  0,  0,
   67,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,
    0,176,  0,176,  0,  0,  0,  0,255,  1,  4,  0,176,176,176,  0,
    0,255,  2,  4,  0,  0,176,176,176,  0,255,  3,  4,  0,  0,  0,
    0,  0,  0,255,
};

static const char digs[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 41,  0,  0,  0, 52,  0,  0,  0, 63,  0,  0,  0,
   74,  0,  0,  0,  1,  3,  0,176,176,176,  0,  5,  1,  0,176,  0,
  255,  1,  6,  0,176,  0,176,  0,176,  0,  0,255,  1,  6,  0,176,
    0,176,  0,176,  0,  0,255,  1,  6,  0,176,  0,176,176,176,  0,
    0,255,  2,  5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char digt[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 34,  0,  0,  0, 41,  0,  0,  0, 51,  0,  0,  0,
   62,  0,  0,  0,  1,  1,  0,176,  0,255,  1,  2,  0,176,  0,  0,
  255,  1,  5,  0,176,176,176,176,176,  0,255,  1,  6,  0,176,  0,
    0,  0,  0,  0,  0,255,  2,  1,  0,  0,  0,255,
};

static const char digu[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 48,  0,  0,  0, 55,  0,  0,  0,
   66,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  2,  5,
    0,  0,  0,  0,176,  0,  0,255,  5,  2,  0,176,  0,  0,255,  1,
    6,  0,176,176,176,176,176,  0,  0,255,  2,  5,  0,  0,  0,  0,
    0,  0,  0,255,
};

static const char digv[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 36,  0,  0,  0, 45,  0,  0,  0, 53,  0,  0,  0,
   64,  0,  0,  0,  1,  3,  0,176,176,176,  0,255,  2,  4,  0,  0,
    0,176,176,  0,255,  4,  3,  0,176,176,  0,  0,255,  1,  6,  0,
  176,176,176,  0,  0,  0,  0,255,  2,  3,  0,  0,  0,  0,  0,255,
};

static const char digw[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 38,  0,  0,  0, 47,  0,  0,  0, 56,  0,  0,  0,
   67,  0,  0,  0,  1,  5,  0,176,176,176,176,176,  0,255,  3,  4,
    0,176,176,  0,  0,  0,255,  3,  4,  0,176,176,  0,  0,  0,255,
    1,  6,  0,176,176,176,176,176,  0,  0,255,  2,  5,  0,  0,  0,
    0,  0,  0,  0,255,
};

static const char digx[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 41,  0,  0,  0, 51,  0,  0,  0, 60,  0,  0,  0,
   71,  0,  0,  0,  1,  2,  0,176,176,  0,  4,  2,  0,176,176,  0,
  255,  2,  5,  0,176,176,176,  0,  0,  0,255,  2,  4,  0,176,176,
  176,  0,  0,255,  1,  6,  0,176,176,  0,176,176,  0,  0,255,  2,
    5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char digy[]=
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 42,  0,  0,  0, 52,  0,  0,  0, 61,  0,  0,  0,
   72,  0,  0,  0,  1,  3,  0,176,176,176,  0,  5,  2,  0,176,  0,
    0,255,  2,  5,  0,  0,176,  0,176,  0,  0,255,  3,  4,  0,176,
    0,176,  0,  0,255,  1,  6,  0,176,176,176,176,176,  0,  0,255,
    2,  5,  0,  0,  0,  0,  0,  0,  0,255,
};

static const char digz[]=  // end of letters A-Z
{
    5,  0,  7,  0,  0,  0,  0,  0,
   28,  0,  0,  0, 40,  0,  0,  0, 51,  0,  0,  0, 62,  0,  0,  0,
   73,  0,  0,  0,  1,  1,  0,176,  0,  4,  2,  0,176,176,  0,255,
    1,  6,  0,176,  0,176,176,176,  0,  0,255,  1,  6,  0,176,176,
  176,  0,176,  0,  0,255,  1,  6,  0,176,176,  0,  0,176,  0,  0,
  255,  2,  2,  0,  0,  0,  0,  6,  1,  0,  0,  0,255,
};

static const char stkeys6[]=  // show both blue keys on stat bar
{
    7,  0,  7,  0,  0,  0,  0,  0,
   36,  0,  0,  0, 43,  0,  0,  0, 51,  0,  0,  0, 62,  0,  0,  0,
   74,  0,  0,  0, 86,  0,  0,  0, 96,  0,  0,  0,  1,  2,  0,198,
  200,  0,255,  0,  3,  0,198,202,203,  0,255,  0,  6,  0,198,  0,
  202,204,199,197,  0,255,  0,  7,  0,198,202,202,198,207,199,203,
    0,255,  0,  7,  0,200,  0,201,199,200,207,202,  0,255,  2,  5,
    0,202,199,207,199,202,  0,255,  3,  3,  0,204,199,197,  0,255,
};

static const char stkeys7[]=  // show both yellow keys on stat bar
{
    7,  0,  7,  0,  0,  0,  0,  0,
   36,  0,  0,  0, 43,  0,  0,  0, 51,  0,  0,  0, 62,  0,  0,  0,
   74,  0,  0,  0, 86,  0,  0,  0, 96,  0,  0,  0,  1,  2,  0,161,
  161,  0,255,  0,  3,  0,161,163,163,  0,255,  0,  6,  0,161,  0,
  163,220,162,235,  0,255,  0,  7,  0,161,163,217,162,246,214,221,
    0,255,  0,  7,  0,161,  0,214,213,214,246,218,  0,255,  2,  5,
    0,217,162,246,214,221,  0,255,  3,  3,  0,220,162,235,  0,255,
};

static const char stkeys8[]=  // show both red keys on stat bar
{
    7,  0,  7,  0,  0,  0,  0,  0,
   36,  0,  0,  0, 43,  0,  0,  0, 51,  0,  0,  0, 62,  0,  0,  0,
   74,  0,  0,  0, 86,  0,  0,  0, 96,  0,  0,  0,  1,  2,  0,174,
  176,  0,255,  0,  3,  0,174,179,179,  0,255,  0,  6,  0,176,  0,
  179,184,174,174,  0,255,  0,  7,  0,176,179,185,175,246,185,182,
    0,255,  0,  7,  0,176,  0,183,177,176,246,179,  0,255,  2,  5,
    0,185,177,246,185,182,  0,255,  3,  3,  0,185,175,174,  0,255,
};

// Background characters for the message review when enabled
// Makes a general purpose background box, could be used for other
// purposes as well
//
//  ul uc ur
//  cl cc cr
//  ll lc lr
//

static const char boxcc[]=  // message background characters
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0,151,151,151, 77, 77, 77, 78, 77,  0,255,  0,  8,  0,
   77, 77, 78, 77, 77, 77, 77, 78,  0,255,  0,  8,  0, 78,151, 77,
   78, 78, 77, 78, 77,  0,255,  0,  8,  0, 77, 77, 78, 77, 77, 77,
   77,151,  0,255,  0,  8,  0,151, 77, 77,151, 77, 78, 77, 77,  0,
  255,  0,  8,  0, 77, 78, 77, 77, 77, 77,151, 78,  0,255,  0,  8,
    0, 78, 77, 77,151, 77, 78, 77, 77,  0,255,  0,  8,  0, 77, 78,
   77, 77, 77,151,151,151,  0,255,
};

static const char boxcl[]=
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0, 10, 10,  9, 10,  9,  9, 10,  9,  0,255,  0,  8,  0,
   11,126,126,126,126, 11,126,126,  0,255,  0,  8,  0,238,238,238,
  238,238,238,238,238,  0,255,  0,  8,  0,151,151, 76, 77,151,151,
  151,151,  0,255,  0,  8,  0,151, 76,151,151,151, 77,151,151,  0,
  255,  0,  8,  0,151,151,151, 77, 77,151, 78,151,  0,255,  0,  8,
    0, 77, 78, 77, 77, 78, 77,151, 77,  0,255,  0,  8,  0,151,151,
   77, 77, 77, 77, 78, 78,  0,255,
};

static const char boxcr[]=
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0, 78, 78, 77,151,151, 77,151,151,  0,255,  0,  8,  0,
   77,151, 77,151, 77, 77, 78, 77,  0,255,  0,  8,  0,151, 78,151,
   76,151,151,151,151,  0,255,  0,  8,  0,151,151, 77, 76, 76,151,
   76,151,  0,255,  0,  8,  0,151,151,151,151, 76, 76,151,151,  0,
  255,  0,  8,  0,238,238,238,238,238,238,238,238,  0,255,  0,  8,
    0, 11,126,126,126,126, 11,126,126,  0,255,  0,  8,  0, 10, 10,
    9, 10,  9,  9, 10,  9,  0,255,
};

static const char boxlc[]=
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0, 77, 77, 77, 78, 77,  9,  9, 11,  0,255,  0,  8,  0,
   77, 77, 77, 77, 78,159, 10,126,  0,255,  0,  8,  0, 78, 78, 77,
   78, 77,  9,  9,126,  0,255,  0,  8,  0, 77, 77, 77, 77,151,159,
   10, 11,  0,255,  0,  8,  0,151, 77, 78, 77, 77,  9, 10,126,  0,
  255,  0,  8,  0, 77, 77, 77,151, 78,  9,  9,126,  0,255,  0,  8,
    0,151, 77, 78, 77, 77,159, 10,126,  0,255,  0,  8,  0, 77, 77,
  151,151,151,159,  9,126,  0,255,
};

static const char boxll[]=
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0, 10,  9,  9, 10,  9, 10,  9, 10,  0,255,  0,  8,  0,
  126,126, 11,126,126,126, 10, 11,  0,255,  0,  8,  0,238,238,238,
  238,238, 10, 10,126,  0,255,  0,  8,  0, 77,151,151,151,151,159,
   10, 11,  0,255,  0,  8,  0,151,151, 77,151,151,  9, 10,126,  0,
  255,  0,  8,  0, 77, 77,151, 78,151,  9,  9,126,  0,255,  0,  8,
    0, 77, 78, 77,151, 77,159, 10,126,  0,255,  0,  8,  0, 77, 77,
   77, 78, 78,159,  9,126,  0,255,
};

static const char boxlr[]=
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0,151,151, 77,151,151,  9,  9, 11,  0,255,  0,  8,  0,
  151, 77, 77, 78, 77,159, 10,126,  0,255,  0,  8,  0, 76,151,151,
  151,151,  9,  9,126,  0,255,  0,  8,  0, 76, 76,151, 76,151,159,
   10,126,  0,255,  0,  8,  0,151, 76, 76,151,151,  9,  9,126,  0,
  255,  0,  8,  0,238,238,238,238,238,  9, 10,126,  0,255,  0,  8,
    0,126,126, 11,126,126,126, 10, 11,  0,255,  0,  8,  0, 10,  9,
    9, 10,  9, 10,  9, 10,  0,255,
};

static const char boxuc[]=
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0,  9, 11,239,151,151,151, 77, 77,  0,255,  0,  8,  0,
   10,126,239, 77, 77, 78, 77, 77,  0,255,  0,  8,  0,  9,126,239,
   78,151, 77, 78, 78,  0,255,  0,  8,  0, 10, 11,239, 77, 77, 78,
   77, 77,  0,255,  0,  8,  0, 10,126,239,151, 77, 77,151, 77,  0,
  255,  0,  8,  0,  9,126,239, 77, 78, 77, 77, 77,  0,255,  0,  8,
    0, 10,126,239, 78, 77, 77,151, 77,  0,255,  0,  8,  0,  9,126,
  239, 77, 78, 77, 77, 77,  0,255,
};

static const char boxul[]=
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0,159, 10, 10, 10, 10,  9, 10,  9,  0,255,  0,  8,  0,
   10, 11,126, 11,126,126,126,126,  0,255,  0,  8,  0, 10,126,239,
  238,238,238,238,238,  0,255,  0,  8,  0, 10, 11,239,151,151, 76,
   77,151,  0,255,  0,  8,  0, 10,126,239,151, 76,151,151,151,  0,
  255,  0,  8,  0,  9,126,239,151,151,151, 77, 77,  0,255,  0,  8,
    0, 10,126,239, 77, 78, 77, 77, 78,  0,255,  0,  8,  0,  9,126,
  239,151,151, 77, 77, 77,  0,255,
};

static const char boxur[]=  // end message background characters
{
    8,  0,  8,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 53,  0,  0,  0, 66,  0,  0,  0, 79,  0,  0,  0,
   92,  0,  0,  0,105,  0,  0,  0,118,  0,  0,  0,131,  0,  0,  0,
    0,  8,  0,  9, 11,239, 78, 78, 77,151,151,  0,255,  0,  8,  0,
   10,126,239, 77,151, 77,151, 77,  0,255,  0,  8,  0,  9,126,239,
  151, 78,151, 76,151,  0,255,  0,  8,  0, 10,126,239,151,151, 77,
   76, 76,  0,255,  0,  8,  0,  9,126,239,151,151,151,151, 76,  0,
  255,  0,  8,  0, 10,126,239,238,238,238,238,238,  0,255,  0,  8,
    0, 10, 11,126, 11,126,126,126,126,  0,255,  0,  8,  0,159, 10,
   10, 10, 10,  9, 10,  9,  0,255,
};

static const char sttminus[]=  // minus on status bar
{
    8,  0,  6,  0,  0,  0,  0,  0,
   40,  0,  0,  0, 50,  0,  0,  0, 61,  0,  0,  0, 72,  0,  0,  0,
   83,  0,  0,  0, 94,  0,  0,  0,105,  0,  0,  0,116,  0,  0,  0,
    0,  5,  0,191,191,191,191,191,  0,255,  0,  6,  0,191,180,182,
  186,191,109,  0,255,  0,  6,  0,191,177,183,186,191,109,  0,255,
    0,  6,  0,191,177,181,186,191,109,  0,255,  0,  6,  0,191,177,
  183,187,191,109,  0,255,  0,  6,  0,191,182,184,185,191,109,  0,
  255,  0,  6,  0,191,191,191,191,191,109,  0,255,  1,  5,  0,109,
  109,109,109,109,  0,255,
};

// An invisible sprite, used for the TNT1 sprite for point wind
static const char invis_sprite[]=
{
    1,  0,  1,  0,  0,  0,  0,  0,
   12,  0,  0,  0,255,
};

// A new setup menu was added in BOOM, and the various menu graphic
// text strings etc. are included here.

// phares 3/31/98
//
// Lumps for menus

static const char m_keybnd[]= // "Key Bindings"  // 2023/12/02 -- Harmony font version
{
  126,0,12,0,0,0,0,0,0,2,0,0,17,2,0,0,34,2,0,0,51,2,0,0,68,2,0,0,85,2,0,0,102,
  2,0,0,112,2,0,0,122,2,0,0,133,2,0,0,150,2,0,0,167,2,0,0,184,2,0,0,201,2,0,0,
  221,2,0,0,231,2,0,0,240,2,0,0,1,3,0,0,18,3,0,0,35,3,0,0,52,3,0,0,72,3,0,0,92,
  3,0,0,112,3,0,0,132,3,0,0,151,3,0,0,170,3,0,0,189,3,0,0,208,3,0,0,227,3,0,0,
  247,3,0,0,11,4,0,0,31,4,0,0,48,4,0,0,65,4,0,0,82,4,0,0,99,4,0,0,116,4,0,0,
  133,4,0,0,150,4,0,0,166,4,0,0,167,4,0,0,168,4,0,0,169,4,0,0,170,4,0,0,187,4,
  0,0,204,4,0,0,221,4,0,0,238,4,0,0,2,5,0,0,22,5,0,0,42,5,0,0,62,5,0,0,79,5,0,
  0,96,5,0,0,115,5,0,0,124,5,0,0,141,5,0,0,158,5,0,0,175,5,0,0,192,5,0,0,206,5,
  0,0,223,5,0,0,240,5,0,0,1,6,0,0,18,6,0,0,32,6,0,0,42,6,0,0,52,6,0,0,67,6,0,0,
  84,6,0,0,101,6,0,0,118,6,0,0,135,6,0,0,152,6,0,0,169,6,0,0,186,6,0,0,203,6,0,
  0,220,6,0,0,237,6,0,0,254,6,0,0,15,7,0,0,32,7,0,0,49,7,0,0,64,7,0,0,76,7,0,0,
  93,7,0,0,110,7,0,0,127,7,0,0,144,7,0,0,158,7,0,0,175,7,0,0,192,7,0,0,209,7,0,
  0,226,7,0,0,240,7,0,0,250,7,0,0,4,8,0,0,19,8,0,0,36,8,0,0,53,8,0,0,70,8,0,0,
  87,8,0,0,102,8,0,0,119,8,0,0,136,8,0,0,153,8,0,0,170,8,0,0,187,8,0,0,204,8,0,
  0,227,8,0,0,247,8,0,0,11,9,0,0,24,9,0,0,36,9,0,0,53,9,0,0,73,9,0,0,93,9,0,0,
  110,9,0,0,127,9,0,0,144,9,0,0,161,9,0,0,178,9,0,0,195,9,0,0,215,9,0,0,234,9,
  0,0,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,
  159,159,245,245,245,245,244,243,242,243,244,245,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,245,245,245,244,243,242,245,
  245,245,245,159,159,255,0,12,159,159,159,159,159,244,243,242,245,159,159,159,
  159,159,255,3,5,159,159,244,243,242,159,159,255,3,5,159,159,244,243,242,159,
  159,255,2,6,159,159,159,244,243,242,159,159,255,0,12,159,159,159,246,244,244,
  243,242,245,159,159,159,159,159,255,0,12,159,159,245,246,244,244,243,242,243,
  245,245,245,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,
  159,159,255,0,12,159,159,244,245,246,159,243,242,241,242,243,244,159,159,255,
  0,4,159,159,159,159,159,159,5,7,159,159,246,242,242,243,244,159,159,255,7,5,
  245,245,245,245,245,159,159,255,8,4,246,246,159,159,159,159,255,0,12,159,159,
  159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,244,159,159,244,159,246,245,242,243,
  244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,
  159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,
  159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,
  255,0,6,159,159,244,159,159,159,159,159,7,5,246,246,242,243,244,159,159,255,
  0,4,159,159,244,159,159,159,6,6,246,246,245,242,243,244,159,159,255,0,4,159,
  159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,244,
  245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,159,159,159,
  159,6,6,246,246,245,246,159,159,159,159,255,0,5,159,159,159,159,159,159,159,
  6,5,246,246,245,246,159,159,159,255,0,5,159,159,244,245,244,159,159,6,6,246,
  246,241,242,243,245,159,159,255,0,5,159,159,244,245,244,159,159,6,6,246,246,
  241,242,243,244,159,159,255,0,5,159,159,159,159,244,159,159,6,6,246,246,245,
  242,243,244,159,159,255,2,3,159,159,244,159,159,7,5,246,246,242,243,244,159,
  159,255,2,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,2,3,159,
  159,244,159,159,7,5,246,246,242,243,244,159,159,255,2,3,159,159,244,159,159,
  7,5,246,246,242,243,244,159,159,255,0,12,159,159,159,159,244,159,159,246,245,
  242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,
  159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,245,159,159,255,
  0,11,159,159,159,159,159,159,159,246,245,246,159,159,159,255,255,255,255,255,
  0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,244,159,
  159,245,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,
  246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,
  242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,
  243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,159,245,242,243,244,159,159,
  255,0,12,159,159,244,245,244,159,243,242,241,242,243,244,159,159,255,1,3,159,
  159,159,159,159,5,7,159,159,246,242,242,243,245,159,159,255,7,4,245,245,246,
  159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,
  255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,
  159,245,244,243,242,241,242,243,244,159,159,255,3,9,159,159,159,159,246,245,
  246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,
  159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,
  159,159,159,245,244,243,242,245,242,243,244,159,159,255,3,9,159,159,245,243,
  242,245,246,159,159,159,159,255,4,5,159,159,243,242,241,246,246,255,4,5,159,
  159,245,242,241,246,246,255,0,10,159,159,159,159,159,159,159,242,241,242,159,
  159,255,0,12,159,159,244,245,244,244,246,242,241,242,246,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,
  159,159,246,245,246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,159,159,159,159,159,245,242,243,244,159,159,255,0,3,159,
  159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,
  7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,
  243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,
  255,0,12,159,159,244,246,159,159,159,159,245,242,243,246,159,159,255,0,12,
  159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,1,10,159,159,159,
  246,244,243,242,241,242,246,159,159,255,3,7,159,159,159,159,246,245,246,159,
  159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,245,
  244,243,242,241,242,243,244,159,159,255,3,9,159,159,159,159,246,245,246,159,
  159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,
  255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,
  159,245,244,243,242,245,242,243,244,159,159,255,3,9,159,159,245,243,242,245,
  246,159,159,159,159,255,4,5,159,159,243,242,241,246,246,255,4,5,159,159,245,
  242,241,246,246,255,0,10,159,159,159,159,159,159,159,242,241,242,159,159,255,
  0,12,159,159,244,245,244,244,246,242,241,242,246,159,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,159,159,
  246,245,246,159,159,159,159,255,1,10,159,159,159,159,159,159,246,245,246,159,
  159,159,255,0,12,159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,
  0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,
  159,244,159,159,159,159,159,245,242,243,244,159,159,255,0,3,159,159,244,159,
  159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,
  242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,
  159,255,0,3,159,159,244,159,159,4,2,159,159,159,159,7,5,246,246,242,243,244,
  159,159,255,0,3,159,159,244,159,159,4,8,159,159,243,159,245,242,243,244,159,
  159,255,0,3,159,159,159,159,159,4,8,159,159,243,242,241,242,243,244,159,159,
  255,4,8,159,159,243,242,241,242,243,246,159,159,255,4,7,159,159,159,246,245,
  246,159,159,159,255,1,3,159,159,159,159,159,6,5,246,246,245,246,159,159,159,
  255,0,5,159,159,245,245,245,159,159,6,6,246,246,241,242,243,246,159,159,255,
  0,5,159,159,244,245,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,12,
  159,159,244,245,244,244,159,246,241,242,243,244,159,159,255,0,12,159,159,244,
  159,244,244,246,246,241,242,243,244,159,159,255,0,12,159,159,244,159,246,244,
  243,246,245,242,243,244,159,159,255,0,12,159,159,244,159,159,244,243,242,245,
  242,243,244,159,159,255,0,12,159,159,244,159,159,246,243,242,241,242,243,244,
  159,159,255,0,12,159,159,244,246,159,159,243,242,241,242,243,244,159,159,255,
  0,4,159,159,244,245,159,159,5,7,159,159,242,241,242,243,244,159,159,255,0,4,
  159,159,245,245,159,159,6,6,246,246,241,242,243,246,159,159,255,1,3,159,159,
  159,159,159,7,4,245,245,246,159,159,159,255,
};

static const char m_setup[]=  // "Setup"  // 2023/12/02 -- Harmony font version
{
  59,0,12,0,0,0,0,0,244,0,0,0,5,1,0,0,25,1,0,0,45,1,0,0,62,1,0,0,79,1,0,0,96,1,
  0,0,113,1,0,0,130,1,0,0,147,1,0,0,167,1,0,0,186,1,0,0,202,1,0,0,219,1,0,0,
  236,1,0,0,253,1,0,0,14,2,0,0,34,2,0,0,54,2,0,0,74,2,0,0,94,2,0,0,113,2,0,0,
  132,2,0,0,151,2,0,0,170,2,0,0,179,2,0,0,188,2,0,0,197,2,0,0,214,2,0,0,231,2,
  0,0,248,2,0,0,9,3,0,0,29,3,0,0,38,3,0,0,47,3,0,0,56,3,0,0,71,3,0,0,87,3,0,0,
  104,3,0,0,121,3,0,0,132,3,0,0,143,3,0,0,154,3,0,0,165,3,0,0,182,3,0,0,199,3,
  0,0,215,3,0,0,230,3,0,0,247,3,0,0,8,4,0,0,25,4,0,0,42,4,0,0,62,4,0,0,79,4,0,
  0,96,4,0,0,113,4,0,0,127,4,0,0,141,4,0,0,155,4,0,0,1,3,159,159,159,159,159,6,
  5,246,246,245,246,159,159,159,255,0,5,159,159,245,245,245,159,159,6,6,246,
  246,241,242,243,246,159,159,255,0,5,159,159,244,245,244,159,159,6,6,246,246,
  241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,159,246,241,242,243,
  244,159,159,255,0,12,159,159,244,159,244,244,246,246,241,242,243,244,159,159,
  255,0,12,159,159,244,159,246,244,243,246,245,242,243,244,159,159,255,0,12,
  159,159,244,159,159,244,243,242,245,242,243,244,159,159,255,0,12,159,159,244,
  159,159,246,243,242,241,242,243,244,159,159,255,0,12,159,159,244,246,159,159,
  243,242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,
  242,241,242,243,244,159,159,255,0,4,159,159,245,245,159,159,6,6,246,246,241,
  242,243,246,159,159,255,1,3,159,159,159,159,159,7,4,245,245,246,159,159,159,
  255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,244,
  159,246,245,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,
  246,246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,
  246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,
  242,243,244,159,159,255,0,6,159,159,244,159,159,159,159,159,7,5,246,246,242,
  243,244,159,159,255,0,4,159,159,244,159,159,159,6,6,246,246,245,242,243,244,
  159,159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,
  255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,
  159,159,159,159,159,159,6,6,246,246,245,246,159,159,159,159,255,0,4,159,159,
  159,159,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,244,245,159,
  159,255,0,12,159,159,244,245,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,159,
  159,244,242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,
  159,246,245,246,159,159,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,
  159,244,245,159,159,255,0,4,159,159,159,159,159,159,255,0,10,159,159,159,159,
  159,159,159,246,245,246,159,159,255,0,11,159,159,244,245,244,244,243,242,241,
  242,244,246,246,255,0,12,159,159,244,245,244,244,243,242,241,242,243,246,159,
  159,255,0,12,159,159,159,159,159,159,159,246,241,242,243,244,159,159,255,6,6,
  159,159,244,242,243,244,159,159,255,6,6,159,159,244,242,243,244,159,159,255,
  6,6,159,159,244,242,243,244,159,159,255,6,6,159,159,244,242,243,244,159,159,
  255,0,12,159,159,159,159,159,159,159,246,241,242,243,244,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,246,159,159,255,0,11,159,159,244,
  245,244,244,243,242,241,242,244,246,246,255,0,10,159,159,159,159,159,159,159,
  246,245,246,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,
  159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,
  159,244,159,159,159,243,242,241,242,243,244,159,159,255,0,3,159,159,244,159,
  159,4,8,159,159,243,242,241,246,159,159,159,159,255,0,3,159,159,244,159,159,
  4,5,159,159,243,242,241,246,246,255,0,3,159,159,244,159,159,4,5,159,159,243,
  242,241,246,246,255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,
  255,0,9,159,159,244,159,159,159,243,242,241,246,246,255,0,9,159,159,244,245,
  244,244,243,242,241,246,246,255,0,9,159,159,246,245,244,244,243,242,244,246,
  246,255,1,7,159,159,159,159,159,159,246,245,245,255,
};

static const char m_auto[]=  // "Automap"  // 2023/12/02 -- Harmony font version
{
  85,0,12,0,0,0,0,0,92,1,0,0,107,1,0,0,124,1,0,0,141,1,0,0,158,1,0,0,175,1,0,0,
  192,1,0,0,209,1,0,0,226,1,0,0,243,1,0,0,4,2,0,0,21,2,0,0,36,2,0,0,51,2,0,0,
  67,2,0,0,84,2,0,0,101,2,0,0,112,2,0,0,123,2,0,0,134,2,0,0,145,2,0,0,162,2,0,
  0,179,2,0,0,195,2,0,0,210,2,0,0,219,2,0,0,228,2,0,0,237,2,0,0,254,2,0,0,15,3,
  0,0,32,3,0,0,49,3,0,0,69,3,0,0,78,3,0,0,87,3,0,0,96,3,0,0,111,3,0,0,128,3,0,
  0,145,3,0,0,162,3,0,0,180,3,0,0,198,3,0,0,216,3,0,0,234,3,0,0,251,3,0,0,12,4,
  0,0,29,4,0,0,44,4,0,0,61,4,0,0,78,4,0,0,95,4,0,0,112,4,0,0,128,4,0,0,138,4,0,
  0,148,4,0,0,158,4,0,0,168,4,0,0,184,4,0,0,201,4,0,0,218,4,0,0,235,4,0,0,252,
  4,0,0,11,5,0,0,28,5,0,0,45,5,0,0,62,5,0,0,79,5,0,0,96,5,0,0,113,5,0,0,130,5,
  0,0,147,5,0,0,164,5,0,0,181,5,0,0,196,5,0,0,213,5,0,0,230,5,0,0,247,5,0,0,8,
  6,0,0,28,6,0,0,45,6,0,0,62,6,0,0,79,6,0,0,93,6,0,0,107,6,0,0,121,6,0,0,2,10,
  159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,246,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,
  159,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,
  0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,
  159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,
  159,242,241,242,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,
  159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,2,10,
  159,159,159,159,159,246,245,246,159,159,159,159,255,0,10,159,159,159,159,159,
  159,159,246,245,246,159,159,255,0,11,159,159,244,245,244,244,243,242,241,242,
  244,246,246,255,0,12,159,159,244,245,244,244,243,242,241,242,243,246,159,159,
  255,0,12,159,159,159,159,159,159,159,246,241,242,243,244,159,159,255,6,6,159,
  159,244,242,243,244,159,159,255,6,6,159,159,244,242,243,244,159,159,255,6,6,
  159,159,244,242,243,244,159,159,255,6,6,159,159,244,242,243,244,159,159,255,
  0,12,159,159,159,159,159,159,159,246,241,242,243,244,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,246,159,159,255,0,11,159,159,244,245,
  244,244,243,242,241,242,244,246,246,255,0,10,159,159,159,159,159,159,159,246,
  245,246,159,159,255,0,4,159,159,159,159,159,159,255,0,4,159,159,244,245,159,
  159,255,0,4,159,159,244,245,159,159,255,0,12,159,159,244,245,159,159,159,246,
  245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,245,159,159,244,242,241,242,243,244,159,159,255,0,4,159,
  159,244,245,159,159,5,7,159,159,246,245,246,159,159,159,159,255,0,4,159,159,
  244,245,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,159,159,159,
  159,255,1,10,159,159,159,159,159,159,246,245,246,159,159,159,255,0,12,159,
  159,246,245,244,244,243,242,241,242,243,246,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,159,
  246,241,242,243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,
  243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,243,244,159,
  159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,3,
  159,159,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,12,159,159,244,
  159,159,159,159,246,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,243,242,241,
  242,243,246,159,159,255,1,10,159,159,159,159,159,159,246,245,246,159,159,159,
  255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,245,245,244,159,
  246,242,241,242,243,244,159,159,255,1,11,159,159,245,244,244,159,246,245,246,
  159,159,159,159,255,2,5,159,159,245,244,243,246,246,255,3,5,159,159,244,243,
  243,246,246,255,3,5,159,159,244,243,243,246,246,255,2,5,159,159,245,244,243,
  246,246,255,1,11,159,159,245,244,244,159,246,245,246,159,159,159,159,255,0,
  12,159,159,245,245,244,159,246,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,2,10,159,159,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,159,159,159,159,242,241,242,159,159,159,159,255,0,3,159,159,244,159,159,
  5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,
  241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,
  255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,12,159,159,
  244,159,159,159,159,242,241,242,159,159,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,243,242,
  241,242,243,244,159,159,255,2,10,159,159,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,
  159,243,242,241,242,243,244,159,159,255,0,3,159,159,244,159,159,4,8,159,159,
  243,242,241,246,159,159,159,159,255,0,3,159,159,244,159,159,4,5,159,159,243,
  242,241,246,246,255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,
  255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,255,0,9,159,159,
  244,159,159,159,243,242,241,246,246,255,0,9,159,159,244,245,244,244,243,242,
  241,246,246,255,0,9,159,159,246,245,244,244,243,242,244,246,246,255,1,7,159,
  159,159,159,159,159,246,245,245,255,
};

static const char m_chat[]=  // "Chat Strings"  // 2023/12/02 -- Harmony font version
{
  127,0,12,0,0,0,0,0,4,2,0,0,19,2,0,0,36,2,0,0,53,2,0,0,70,2,0,0,87,2,0,0,104,
  2,0,0,121,2,0,0,138,2,0,0,157,2,0,0,176,2,0,0,195,2,0,0,212,2,0,0,229,2,0,0,
  246,2,0,0,7,3,0,0,24,3,0,0,34,3,0,0,44,3,0,0,54,3,0,0,64,3,0,0,81,3,0,0,98,3,
  0,0,115,3,0,0,132,3,0,0,147,3,0,0,164,3,0,0,181,3,0,0,198,3,0,0,215,3,0,0,
  232,3,0,0,249,3,0,0,10,4,0,0,27,4,0,0,44,4,0,0,61,4,0,0,76,4,0,0,85,4,0,0,94,
  4,0,0,103,4,0,0,120,4,0,0,137,4,0,0,154,4,0,0,171,4,0,0,191,4,0,0,200,4,0,0,
  209,4,0,0,218,4,0,0,219,4,0,0,220,4,0,0,221,4,0,0,222,4,0,0,239,4,0,0,3,5,0,
  0,23,5,0,0,40,5,0,0,57,5,0,0,74,5,0,0,91,5,0,0,108,5,0,0,125,5,0,0,145,5,0,0,
  164,5,0,0,180,5,0,0,189,5,0,0,198,5,0,0,207,5,0,0,224,5,0,0,241,5,0,0,2,6,0,
  0,19,6,0,0,39,6,0,0,48,6,0,0,57,6,0,0,66,6,0,0,83,6,0,0,100,6,0,0,117,6,0,0,
  134,6,0,0,154,6,0,0,171,6,0,0,188,6,0,0,208,6,0,0,225,6,0,0,242,6,0,0,2,7,0,
  0,19,7,0,0,36,7,0,0,53,7,0,0,70,7,0,0,87,7,0,0,101,7,0,0,118,7,0,0,135,7,0,0,
  152,7,0,0,169,7,0,0,183,7,0,0,193,7,0,0,203,7,0,0,218,7,0,0,235,7,0,0,252,7,
  0,0,13,8,0,0,30,8,0,0,45,8,0,0,62,8,0,0,79,8,0,0,96,8,0,0,113,8,0,0,130,8,0,
  0,147,8,0,0,170,8,0,0,190,8,0,0,210,8,0,0,223,8,0,0,235,8,0,0,252,8,0,0,16,9,
  0,0,36,9,0,0,53,9,0,0,70,9,0,0,87,9,0,0,104,9,0,0,121,9,0,0,138,9,0,0,158,9,
  0,0,177,9,0,0,1,10,159,159,159,159,159,159,246,245,246,159,159,159,255,0,12,
  159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,
  159,246,245,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,
  243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,
  255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,
  244,159,159,7,5,246,246,242,243,244,159,159,255,0,4,159,159,244,159,159,159,
  6,6,246,246,245,242,243,244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,
  246,241,242,243,244,159,159,255,0,4,159,159,246,245,159,159,6,6,246,246,241,
  242,243,246,159,159,255,1,3,159,159,159,159,159,6,5,246,246,245,246,159,159,
  159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,
  159,159,242,241,242,159,159,159,159,255,5,5,159,159,242,241,242,159,159,255,
  5,5,159,159,242,241,242,159,159,255,5,5,159,159,242,241,242,159,159,255,5,5,
  159,159,242,241,242,159,159,255,0,12,159,159,159,159,159,159,159,242,241,242,
  159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,2,10,159,159,
  159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,246,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,
  159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,
  159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,
  5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,
  241,242,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,2,10,159,159,
  159,159,159,246,245,246,159,159,159,159,255,0,4,159,159,159,159,159,159,255,
  0,4,159,159,244,245,159,159,255,0,4,159,159,244,245,159,159,255,0,12,159,159,
  244,245,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,244,245,159,159,244,242,241,242,243,
  244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,246,245,246,159,159,
  159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,244,245,159,159,255,
  0,4,159,159,159,159,159,159,255,255,255,255,255,1,3,159,159,159,159,159,6,5,
  246,246,245,246,159,159,159,255,0,5,159,159,245,245,245,159,159,6,6,246,246,
  241,242,243,246,159,159,255,0,5,159,159,244,245,244,159,159,6,6,246,246,241,
  242,243,244,159,159,255,0,12,159,159,244,245,244,244,159,246,241,242,243,244,
  159,159,255,0,12,159,159,244,159,244,244,246,246,241,242,243,244,159,159,255,
  0,12,159,159,244,159,246,244,243,246,245,242,243,244,159,159,255,0,12,159,
  159,244,159,159,244,243,242,245,242,243,244,159,159,255,0,12,159,159,244,159,
  159,246,243,242,241,242,243,244,159,159,255,0,12,159,159,244,246,159,159,243,
  242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,242,
  241,242,243,244,159,159,255,0,4,159,159,245,245,159,159,6,6,246,246,241,242,
  243,246,159,159,255,1,3,159,159,159,159,159,7,4,245,245,246,159,159,159,255,
  0,4,159,159,159,159,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,
  244,245,159,159,255,0,12,159,159,244,245,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,159,159,244,242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,
  159,5,7,159,159,246,245,246,159,159,159,159,255,0,4,159,159,244,245,159,159,
  255,0,4,159,159,244,245,159,159,255,0,4,159,159,159,159,159,159,255,0,12,159,
  159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,243,242,241,242,
  243,244,159,159,255,0,3,159,159,244,159,159,4,8,159,159,243,242,241,246,159,
  159,159,159,255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,255,
  0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,255,0,3,159,159,244,
  159,159,4,8,159,159,243,242,241,246,159,159,159,159,255,0,12,159,159,244,159,
  159,159,243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,243,
  242,241,242,243,244,159,159,255,1,11,159,159,246,244,245,159,246,241,242,243,
  244,159,159,255,2,3,159,159,159,159,159,7,5,245,245,246,159,159,159,159,255,
  0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,245,244,243,
  242,241,242,243,244,159,159,255,3,9,159,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,245,
  244,243,242,245,242,243,244,159,159,255,3,9,159,159,245,243,242,245,246,159,
  159,159,159,255,4,5,159,159,243,242,241,246,246,255,4,5,159,159,245,242,241,
  246,246,255,0,10,159,159,159,159,159,159,159,242,241,242,159,159,255,0,12,
  159,159,244,245,244,244,246,242,241,242,246,159,159,159,255,0,12,159,159,244,
  245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,159,159,246,245,
  246,159,159,159,159,255,1,10,159,159,159,159,159,159,246,245,246,159,159,159,
  255,0,12,159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,
  159,159,159,159,159,245,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,
  246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,
  244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,
  0,3,159,159,244,159,159,4,2,159,159,159,159,7,5,246,246,242,243,244,159,159,
  255,0,3,159,159,244,159,159,4,8,159,159,243,159,245,242,243,244,159,159,255,
  0,3,159,159,159,159,159,4,8,159,159,243,242,241,242,243,244,159,159,255,4,8,
  159,159,243,242,241,242,243,246,159,159,255,4,7,159,159,159,246,245,246,159,
  159,159,255,1,3,159,159,159,159,159,6,5,246,246,245,246,159,159,159,255,0,5,
  159,159,245,245,245,159,159,6,6,246,246,241,242,243,246,159,159,255,0,5,159,
  159,244,245,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,159,246,241,242,243,244,159,159,255,0,12,159,159,244,159,244,
  244,246,246,241,242,243,244,159,159,255,0,12,159,159,244,159,246,244,243,246,
  245,242,243,244,159,159,255,0,12,159,159,244,159,159,244,243,242,245,242,243,
  244,159,159,255,0,12,159,159,244,159,159,246,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,246,159,159,243,242,241,242,243,244,159,159,255,0,4,159,
  159,244,245,159,159,5,7,159,159,242,241,242,243,244,159,159,255,0,4,159,159,
  245,245,159,159,6,6,246,246,241,242,243,246,159,159,255,1,3,159,159,159,159,
  159,7,4,245,245,246,159,159,159,255,
};

static const char m_enem[]=  // "Enemies"  // 2023/12/02 -- Harmony font version
{
  79,0,12,0,0,0,0,0,68,1,0,0,85,1,0,0,102,1,0,0,119,1,0,0,136,1,0,0,156,1,0,0,
  176,1,0,0,196,1,0,0,216,1,0,0,235,1,0,0,254,1,0,0,17,2,0,0,36,2,0,0,53,2,0,0,
  70,2,0,0,87,2,0,0,104,2,0,0,118,2,0,0,128,2,0,0,138,2,0,0,153,2,0,0,170,2,0,
  0,187,2,0,0,204,2,0,0,221,2,0,0,238,2,0,0,255,2,0,0,16,3,0,0,33,3,0,0,53,3,0,
  0,73,3,0,0,93,3,0,0,113,3,0,0,132,3,0,0,151,3,0,0,170,3,0,0,189,3,0,0,206,3,
  0,0,223,3,0,0,240,3,0,0,1,4,0,0,17,4,0,0,27,4,0,0,37,4,0,0,47,4,0,0,57,4,0,0,
  73,4,0,0,90,4,0,0,107,4,0,0,124,4,0,0,141,4,0,0,158,4,0,0,175,4,0,0,192,4,0,
  0,209,4,0,0,223,4,0,0,240,4,0,0,1,5,0,0,18,5,0,0,35,5,0,0,55,5,0,0,75,5,0,0,
  95,5,0,0,115,5,0,0,134,5,0,0,153,5,0,0,172,5,0,0,191,5,0,0,208,5,0,0,228,5,0,
  0,248,5,0,0,9,6,0,0,26,6,0,0,43,6,0,0,60,6,0,0,77,6,0,0,94,6,0,0,114,6,0,0,
  133,6,0,0,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,
  244,159,246,245,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,
  7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,
  246,246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,
  246,242,243,244,159,159,255,0,6,159,159,244,159,159,159,159,159,7,5,246,246,
  242,243,244,159,159,255,0,4,159,159,244,159,159,159,6,6,246,246,245,242,243,
  244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,
  159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,
  0,4,159,159,159,159,159,159,6,6,246,246,245,246,159,159,159,159,255,0,12,159,
  159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,159,159,245,244,243,242,245,242,
  243,244,159,159,255,3,9,159,159,245,243,242,245,246,159,159,159,159,255,4,5,
  159,159,243,242,241,246,246,255,4,5,159,159,245,242,241,246,246,255,0,10,159,
  159,159,159,159,159,159,242,241,242,159,159,255,0,12,159,159,244,245,244,244,
  246,242,241,242,246,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,
  159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,
  0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,244,159,
  246,245,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,
  246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,
  242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,
  243,244,159,159,255,0,6,159,159,244,159,159,159,159,159,7,5,246,246,242,243,
  244,159,159,255,0,4,159,159,244,159,159,159,6,6,246,246,245,242,243,244,159,
  159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,
  0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,
  159,159,159,159,159,6,6,246,246,245,246,159,159,159,159,255,0,12,159,159,159,
  159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,245,245,244,159,246,242,241,242,243,244,
  159,159,255,1,11,159,159,245,244,244,159,246,245,246,159,159,159,159,255,2,5,
  159,159,245,244,243,246,246,255,3,5,159,159,244,243,243,246,246,255,3,5,159,
  159,244,243,243,246,246,255,2,5,159,159,245,244,243,246,246,255,1,11,159,159,
  245,244,244,159,246,245,246,159,159,159,159,255,0,12,159,159,245,245,244,159,
  246,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,
  159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,
  0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,245,244,243,
  242,241,242,243,244,159,159,255,3,9,159,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,
  244,159,246,245,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,
  7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,
  246,246,242,243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,
  246,242,243,244,159,159,255,0,6,159,159,244,159,159,159,159,159,7,5,246,246,
  242,243,244,159,159,255,0,4,159,159,244,159,159,159,6,6,246,246,245,242,243,
  244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,
  159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,
  0,4,159,159,159,159,159,159,6,6,246,246,245,246,159,159,159,159,255,1,3,159,
  159,159,159,159,6,5,246,246,245,246,159,159,159,255,0,5,159,159,245,245,245,
  159,159,6,6,246,246,241,242,243,246,159,159,255,0,5,159,159,244,245,244,159,
  159,6,6,246,246,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,159,
  246,241,242,243,244,159,159,255,0,12,159,159,244,159,244,244,246,246,241,242,
  243,244,159,159,255,0,12,159,159,244,159,246,244,243,246,245,242,243,244,159,
  159,255,0,12,159,159,244,159,159,244,243,242,245,242,243,244,159,159,255,0,
  12,159,159,244,159,159,246,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,246,159,159,243,242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,
  159,5,7,159,159,242,241,242,243,244,159,159,255,0,4,159,159,245,245,159,159,
  6,6,246,246,241,242,243,246,159,159,255,1,3,159,159,159,159,159,7,4,245,245,
  246,159,159,159,255,
};

static const char m_stat[]=  // "Status Bar / HUD"  // 2023/12/02 -- Harmony font version
{
  158,0,12,0,0,0,0,0,128,2,0,0,145,2,0,0,165,2,0,0,185,2,0,0,202,2,0,0,219,2,0,
  0,236,2,0,0,253,2,0,0,14,3,0,0,31,3,0,0,51,3,0,0,70,3,0,0,86,3,0,0,95,3,0,0,
  104,3,0,0,113,3,0,0,130,3,0,0,147,3,0,0,164,3,0,0,181,3,0,0,201,3,0,0,210,3,
  0,0,219,3,0,0,228,3,0,0,243,3,0,0,4,4,0,0,21,4,0,0,38,4,0,0,55,4,0,0,72,4,0,
  0,89,4,0,0,106,4,0,0,123,4,0,0,140,4,0,0,157,4,0,0,172,4,0,0,181,4,0,0,190,4,
  0,0,199,4,0,0,216,4,0,0,233,4,0,0,250,4,0,0,11,5,0,0,31,5,0,0,40,5,0,0,49,5,
  0,0,58,5,0,0,73,5,0,0,89,5,0,0,106,5,0,0,123,5,0,0,134,5,0,0,145,5,0,0,156,5,
  0,0,167,5,0,0,184,5,0,0,201,5,0,0,217,5,0,0,232,5,0,0,249,5,0,0,13,6,0,0,33,
  6,0,0,50,6,0,0,67,6,0,0,84,6,0,0,101,6,0,0,118,6,0,0,135,6,0,0,155,6,0,0,174,
  6,0,0,190,6,0,0,191,6,0,0,192,6,0,0,193,6,0,0,194,6,0,0,211,6,0,0,228,6,0,0,
  245,6,0,0,6,7,0,0,26,7,0,0,46,7,0,0,66,7,0,0,86,7,0,0,103,7,0,0,120,7,0,0,
  139,7,0,0,148,7,0,0,163,7,0,0,180,7,0,0,197,7,0,0,214,7,0,0,231,7,0,0,248,7,
  0,0,9,8,0,0,26,8,0,0,43,8,0,0,60,8,0,0,77,8,0,0,92,8,0,0,109,8,0,0,126,8,0,0,
  143,8,0,0,160,8,0,0,180,8,0,0,197,8,0,0,214,8,0,0,234,8,0,0,251,8,0,0,12,9,0,
  0,28,9,0,0,45,9,0,0,55,9,0,0,66,9,0,0,78,9,0,0,89,9,0,0,100,9,0,0,111,9,0,0,
  122,9,0,0,132,9,0,0,142,9,0,0,152,9,0,0,161,9,0,0,169,9,0,0,186,9,0,0,203,9,
  0,0,220,9,0,0,237,9,0,0,247,9,0,0,1,10,0,0,11,10,0,0,21,10,0,0,38,10,0,0,55,
  10,0,0,72,10,0,0,89,10,0,0,104,10,0,0,120,10,0,0,137,10,0,0,154,10,0,0,165,
  10,0,0,176,10,0,0,187,10,0,0,198,10,0,0,215,10,0,0,232,10,0,0,248,10,0,0,7,
  11,0,0,24,11,0,0,41,11,0,0,58,11,0,0,75,11,0,0,92,11,0,0,109,11,0,0,126,11,0,
  0,143,11,0,0,160,11,0,0,177,11,0,0,192,11,0,0,1,3,159,159,159,159,159,6,5,
  246,246,245,246,159,159,159,255,0,5,159,159,245,245,245,159,159,6,6,246,246,
  241,242,243,246,159,159,255,0,5,159,159,244,245,244,159,159,6,6,246,246,241,
  242,243,244,159,159,255,0,12,159,159,244,245,244,244,159,246,241,242,243,244,
  159,159,255,0,12,159,159,244,159,244,244,246,246,241,242,243,244,159,159,255,
  0,12,159,159,244,159,246,244,243,246,245,242,243,244,159,159,255,0,12,159,
  159,244,159,159,244,243,242,245,242,243,244,159,159,255,0,12,159,159,244,159,
  159,246,243,242,241,242,243,244,159,159,255,0,12,159,159,244,246,159,159,243,
  242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,242,
  241,242,243,244,159,159,255,0,4,159,159,245,245,159,159,6,6,246,246,241,242,
  243,246,159,159,255,1,3,159,159,159,159,159,7,4,245,245,246,159,159,159,255,
  0,4,159,159,159,159,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,
  244,245,159,159,255,0,12,159,159,244,245,159,159,159,246,245,246,159,159,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,159,159,244,242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,
  159,5,7,159,159,246,245,246,159,159,159,159,255,0,4,159,159,244,245,159,159,
  255,0,4,159,159,244,245,159,159,255,0,4,159,159,159,159,159,159,255,2,10,159,
  159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,246,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,
  159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,
  159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,
  159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,
  242,241,242,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,
  159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  0,12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,2,10,159,
  159,159,159,159,246,245,246,159,159,159,159,255,0,4,159,159,159,159,159,159,
  255,0,4,159,159,244,245,159,159,255,0,4,159,159,244,245,159,159,255,0,12,159,
  159,244,245,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,245,159,159,244,242,241,242,
  243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,246,245,246,159,
  159,159,159,255,0,4,159,159,244,245,159,159,255,0,4,159,159,244,245,159,159,
  255,0,4,159,159,159,159,159,159,255,0,10,159,159,159,159,159,159,159,246,245,
  246,159,159,255,0,11,159,159,244,245,244,244,243,242,241,242,244,246,246,255,
  0,12,159,159,244,245,244,244,243,242,241,242,243,246,159,159,255,0,12,159,
  159,159,159,159,159,159,246,241,242,243,244,159,159,255,6,6,159,159,244,242,
  243,244,159,159,255,6,6,159,159,244,242,243,244,159,159,255,6,6,159,159,244,
  242,243,244,159,159,255,6,6,159,159,244,242,243,244,159,159,255,0,12,159,159,
  159,159,159,159,159,246,241,242,243,244,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,246,159,159,255,0,11,159,159,244,245,244,244,243,242,
  241,242,244,246,246,255,0,10,159,159,159,159,159,159,159,246,245,246,159,159,
  255,1,3,159,159,159,159,159,6,5,246,246,245,246,159,159,159,255,0,5,159,159,
  245,245,245,159,159,6,6,246,246,241,242,243,246,159,159,255,0,5,159,159,244,
  245,244,159,159,6,6,246,246,241,242,243,244,159,159,255,0,12,159,159,244,245,
  244,244,159,246,241,242,243,244,159,159,255,0,12,159,159,244,159,244,244,246,
  246,241,242,243,244,159,159,255,0,12,159,159,244,159,246,244,243,246,245,242,
  243,244,159,159,255,0,12,159,159,244,159,159,244,243,242,245,242,243,244,159,
  159,255,0,12,159,159,244,159,159,246,243,242,241,242,243,244,159,159,255,0,
  12,159,159,244,246,159,159,243,242,241,242,243,244,159,159,255,0,4,159,159,
  244,245,159,159,5,7,159,159,242,241,242,243,244,159,159,255,0,4,159,159,245,
  245,159,159,6,6,246,246,241,242,243,246,159,159,255,1,3,159,159,159,159,159,
  7,4,245,245,246,159,159,159,255,255,255,255,255,0,12,159,159,159,159,159,159,
  159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,
  159,159,255,0,12,159,159,244,159,159,244,159,159,245,242,243,244,159,159,255,
  0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,
  159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,
  159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,
  244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,12,159,159,244,
  245,244,244,243,159,245,242,243,244,159,159,255,0,12,159,159,244,245,244,159,
  243,242,241,242,243,244,159,159,255,1,3,159,159,159,159,159,5,7,159,159,246,
  242,242,243,245,159,159,255,7,4,245,245,246,159,159,159,255,2,10,159,159,159,
  159,159,246,245,246,159,159,159,159,255,0,12,159,159,246,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,159,
  255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,
  244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,
  159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,
  242,159,159,255,0,12,159,159,244,159,159,159,159,242,241,242,159,159,159,159,
  255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,
  159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,2,10,159,159,159,
  159,159,246,245,246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,244,159,159,159,243,242,241,242,243,244,159,159,255,0,3,159,
  159,244,159,159,4,8,159,159,243,242,241,246,159,159,159,159,255,0,3,159,159,
  244,159,159,4,5,159,159,243,242,241,246,246,255,0,3,159,159,244,159,159,4,5,
  159,159,243,242,241,246,246,255,0,3,159,159,244,159,159,4,8,159,159,243,242,
  241,246,159,159,159,159,255,0,12,159,159,244,159,159,159,243,242,241,242,243,
  244,159,159,255,0,12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,
  255,1,11,159,159,246,244,245,159,246,241,242,243,244,159,159,255,2,3,159,159,
  159,159,159,7,5,245,245,246,159,159,159,159,255,7,5,246,246,246,159,159,159,
  159,255,6,6,246,246,242,242,243,244,159,159,255,5,7,159,159,243,241,242,243,
  244,159,159,255,5,6,244,244,242,241,242,244,159,159,255,4,6,159,159,243,242,
  241,244,159,159,255,3,6,159,159,245,243,242,242,246,246,255,2,6,159,159,245,
  244,243,244,246,246,255,2,5,159,159,244,244,245,246,246,255,1,5,159,159,245,
  244,245,159,159,255,0,5,159,159,244,245,245,159,159,255,0,4,159,159,244,245,
  159,159,255,0,3,159,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,159,159,159,159,159,242,241,242,159,159,159,159,255,5,5,159,
  159,242,241,242,159,159,255,5,5,159,159,242,241,242,159,159,255,5,5,159,159,
  242,241,242,159,159,255,5,5,159,159,242,241,242,159,159,255,0,12,159,159,159,
  159,159,159,159,242,241,242,159,159,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,
  159,159,255,0,10,159,159,159,159,159,159,159,246,245,246,159,159,255,0,11,
  159,159,244,245,244,244,243,242,241,242,244,246,246,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,246,159,159,255,0,12,159,159,159,159,159,159,159,
  246,241,242,243,244,159,159,255,6,6,159,159,244,242,243,244,159,159,255,6,6,
  159,159,244,242,243,244,159,159,255,6,6,159,159,244,242,243,244,159,159,255,
  6,6,159,159,244,242,243,244,159,159,255,0,12,159,159,159,159,159,159,159,246,
  241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  246,159,159,255,0,11,159,159,244,245,244,244,243,242,241,242,244,246,246,255,
  0,10,159,159,159,159,159,159,159,246,245,246,159,159,255,0,12,159,159,159,
  159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,0,12,159,159,244,159,159,159,159,159,245,242,243,244,
  159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,
  159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,
  159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,
  242,243,244,159,159,255,0,12,159,159,244,246,159,159,159,159,245,242,243,246,
  159,159,255,0,12,159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,
  1,10,159,159,159,246,244,243,242,241,242,246,159,159,255,3,7,159,159,159,159,
  246,245,246,159,159,255,
};



static const char m_weap[]=  // "Weapons"  // 2023/12/02 -- Harmony font version
{
  86,0,12,0,0,0,0,0,96,1,0,0,113,1,0,0,130,1,0,0,147,1,0,0,164,1,0,0,180,1,0,0,
  190,1,0,0,200,1,0,0,210,1,0,0,220,1,0,0,236,1,0,0,253,1,0,0,14,2,0,0,31,2,0,
  0,48,2,0,0,65,2,0,0,82,2,0,0,99,2,0,0,116,2,0,0,136,2,0,0,156,2,0,0,176,2,0,
  0,196,2,0,0,215,2,0,0,234,2,0,0,253,2,0,0,16,3,0,0,31,3,0,0,48,3,0,0,65,3,0,
  0,82,3,0,0,99,3,0,0,116,3,0,0,133,3,0,0,150,3,0,0,167,3,0,0,184,3,0,0,201,3,
  0,0,216,3,0,0,233,3,0,0,250,3,0,0,11,4,0,0,28,4,0,0,48,4,0,0,65,4,0,0,82,4,0,
  0,99,4,0,0,113,4,0,0,127,4,0,0,141,4,0,0,153,4,0,0,168,4,0,0,185,4,0,0,202,4,
  0,0,219,4,0,0,237,4,0,0,255,4,0,0,17,5,0,0,35,5,0,0,52,5,0,0,69,5,0,0,86,5,0,
  0,101,5,0,0,118,5,0,0,135,5,0,0,152,5,0,0,169,5,0,0,183,5,0,0,193,5,0,0,203,
  5,0,0,218,5,0,0,235,5,0,0,252,5,0,0,13,6,0,0,30,6,0,0,47,6,0,0,67,6,0,0,87,6,
  0,0,104,6,0,0,121,6,0,0,138,6,0,0,155,6,0,0,172,6,0,0,189,6,0,0,209,6,0,0,
  228,6,0,0,0,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,
  244,243,246,245,242,243,244,159,159,255,0,11,159,159,159,159,159,159,159,246,
  241,242,244,159,159,255,5,5,159,159,244,241,244,159,159,255,4,5,159,159,244,
  242,241,159,159,255,4,5,159,159,244,242,241,159,159,255,5,5,159,159,244,241,
  244,159,159,255,0,11,159,159,159,159,159,159,159,246,241,242,244,159,159,255,
  0,12,159,159,244,245,244,244,243,246,245,242,243,244,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,159,159,
  246,245,246,159,159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,
  159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,244,159,159,244,159,246,245,242,243,244,159,159,255,0,6,159,159,
  244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,
  159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,159,
  159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,6,159,159,244,159,159,
  159,159,159,7,5,246,246,242,243,244,159,159,255,0,4,159,159,244,159,159,159,
  6,6,246,246,245,242,243,244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,
  246,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,6,6,246,246,241,
  242,243,244,159,159,255,0,4,159,159,159,159,159,159,6,6,246,246,245,246,159,
  159,159,159,255,2,10,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,
  159,159,242,241,242,159,159,159,159,255,0,3,159,159,244,159,159,5,5,159,159,
  242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,
  159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,3,159,
  159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,12,159,159,244,159,159,
  159,159,242,241,242,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,243,242,241,242,243,
  244,159,159,255,2,10,159,159,159,159,159,246,245,246,159,159,159,159,255,0,
  12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,243,242,
  241,242,243,244,159,159,255,0,3,159,159,244,159,159,4,8,159,159,243,242,241,
  246,159,159,159,159,255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,
  246,255,0,3,159,159,244,159,159,4,5,159,159,243,242,241,246,246,255,0,3,159,
  159,244,159,159,4,5,159,159,243,242,241,246,246,255,0,9,159,159,244,159,159,
  159,243,242,241,246,246,255,0,9,159,159,244,245,244,244,243,242,241,246,246,
  255,0,9,159,159,246,245,244,244,243,242,244,246,246,255,1,7,159,159,159,159,
  159,159,246,245,245,255,1,10,159,159,159,159,159,159,246,245,246,159,159,159,
  255,0,12,159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,
  159,159,159,159,246,241,242,243,244,159,159,255,0,3,159,159,244,159,159,6,6,
  246,246,241,242,243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,
  242,243,244,159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,243,244,
  159,159,255,0,3,159,159,244,159,159,6,6,246,246,241,242,243,244,159,159,255,
  0,12,159,159,244,159,159,159,159,246,241,242,243,244,159,159,255,0,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,
  244,244,243,242,241,242,243,246,159,159,255,1,10,159,159,159,159,159,159,246,
  245,246,159,159,159,255,0,12,159,159,159,159,159,159,159,246,245,246,159,159,
  159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,
  159,159,159,245,244,243,242,245,242,243,244,159,159,255,3,9,159,159,245,243,
  242,245,246,159,159,159,159,255,4,5,159,159,243,242,241,246,246,255,4,5,159,
  159,245,242,241,246,246,255,0,10,159,159,159,159,159,159,159,242,241,242,159,
  159,255,0,12,159,159,244,245,244,244,246,242,241,242,246,159,159,159,255,0,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,159,159,159,
  159,159,246,245,246,159,159,159,159,255,1,3,159,159,159,159,159,6,5,246,246,
  245,246,159,159,159,255,0,5,159,159,245,245,245,159,159,6,6,246,246,241,242,
  243,246,159,159,255,0,5,159,159,244,245,244,159,159,6,6,246,246,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,159,246,241,242,243,244,159,159,
  255,0,12,159,159,244,159,244,244,246,246,241,242,243,244,159,159,255,0,12,
  159,159,244,159,246,244,243,246,245,242,243,244,159,159,255,0,12,159,159,244,
  159,159,244,243,242,245,242,243,244,159,159,255,0,12,159,159,244,159,159,246,
  243,242,241,242,243,244,159,159,255,0,12,159,159,244,246,159,159,243,242,241,
  242,243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,242,241,242,
  243,244,159,159,255,0,4,159,159,245,245,159,159,6,6,246,246,241,242,243,246,
  159,159,255,1,3,159,159,159,159,159,7,4,245,245,246,159,159,159,255,
};

static const char m_mess[]=  // "Messages"  // 2023/12/02 -- Harmony font version
{
  98,0,12,0,0,0,0,0,144,1,0,0,161,1,0,0,178,1,0,0,195,1,0,0,212,1,0,0,228,1,0,
  0,238,1,0,0,248,1,0,0,2,2,0,0,12,2,0,0,28,2,0,0,45,2,0,0,62,2,0,0,79,2,0,0,
  96,2,0,0,113,2,0,0,130,2,0,0,147,2,0,0,164,2,0,0,184,2,0,0,204,2,0,0,224,2,0,
  0,244,2,0,0,7,3,0,0,26,3,0,0,45,3,0,0,64,3,0,0,81,3,0,0,101,3,0,0,121,3,0,0,
  138,3,0,0,155,3,0,0,172,3,0,0,189,3,0,0,206,3,0,0,223,3,0,0,243,3,0,0,6,4,0,
  0,22,4,0,0,39,4,0,0,59,4,0,0,79,4,0,0,96,4,0,0,113,4,0,0,130,4,0,0,147,4,0,0,
  164,4,0,0,181,4,0,0,201,4,0,0,220,4,0,0,236,4,0,0,251,4,0,0,12,5,0,0,29,5,0,
  0,46,5,0,0,63,5,0,0,80,5,0,0,97,5,0,0,114,5,0,0,131,5,0,0,148,5,0,0,165,5,0,
  0,180,5,0,0,195,5,0,0,212,5,0,0,229,5,0,0,246,5,0,0,7,6,0,0,24,6,0,0,41,6,0,
  0,64,6,0,0,84,6,0,0,104,6,0,0,117,6,0,0,129,6,0,0,146,6,0,0,163,6,0,0,180,6,
  0,0,197,6,0,0,217,6,0,0,237,6,0,0,1,7,0,0,21,7,0,0,40,7,0,0,59,7,0,0,78,7,0,
  0,97,7,0,0,114,7,0,0,134,7,0,0,154,7,0,0,171,7,0,0,188,7,0,0,205,7,0,0,222,7,
  0,0,239,7,0,0,0,8,0,0,20,8,0,0,39,8,0,0,0,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,0,12,159,159,245,245,244,159,246,242,241,242,243,244,159,159,255,1,11,
  159,159,245,244,244,159,246,245,246,159,159,159,159,255,2,5,159,159,245,244,
  243,246,246,255,3,5,159,159,244,243,243,246,246,255,3,5,159,159,244,243,243,
  246,246,255,2,5,159,159,245,244,243,246,246,255,1,11,159,159,245,244,244,159,
  246,245,246,159,159,159,159,255,0,12,159,159,245,245,244,159,246,242,241,242,
  243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,0,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,
  12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,
  159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,0,12,159,159,244,159,159,244,159,246,245,242,243,
  244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,
  159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,
  159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,159,
  255,0,6,159,159,244,159,159,159,159,159,7,5,246,246,242,243,244,159,159,255,
  0,4,159,159,244,159,159,159,6,6,246,246,245,242,243,244,159,159,255,0,4,159,
  159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,244,
  245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,159,159,159,
  159,6,6,246,246,245,246,159,159,159,159,255,1,3,159,159,159,159,159,6,5,246,
  246,245,246,159,159,159,255,0,5,159,159,245,245,245,159,159,6,6,246,246,241,
  242,243,246,159,159,255,0,5,159,159,244,245,244,159,159,6,6,246,246,241,242,
  243,244,159,159,255,0,12,159,159,244,245,244,244,159,246,241,242,243,244,159,
  159,255,0,12,159,159,244,159,244,244,246,246,241,242,243,244,159,159,255,0,
  12,159,159,244,159,246,244,243,246,245,242,243,244,159,159,255,0,12,159,159,
  244,159,159,244,243,242,245,242,243,244,159,159,255,0,12,159,159,244,159,159,
  246,243,242,241,242,243,244,159,159,255,0,12,159,159,244,246,159,159,243,242,
  241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,242,241,
  242,243,244,159,159,255,0,4,159,159,245,245,159,159,6,6,246,246,241,242,243,
  246,159,159,255,1,3,159,159,159,159,159,7,4,245,245,246,159,159,159,255,1,3,
  159,159,159,159,159,6,5,246,246,245,246,159,159,159,255,0,5,159,159,245,245,
  245,159,159,6,6,246,246,241,242,243,246,159,159,255,0,5,159,159,244,245,244,
  159,159,6,6,246,246,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,
  159,246,241,242,243,244,159,159,255,0,12,159,159,244,159,244,244,246,246,241,
  242,243,244,159,159,255,0,12,159,159,244,159,246,244,243,246,245,242,243,244,
  159,159,255,0,12,159,159,244,159,159,244,243,242,245,242,243,244,159,159,255,
  0,12,159,159,244,159,159,246,243,242,241,242,243,244,159,159,255,0,12,159,
  159,244,246,159,159,243,242,241,242,243,244,159,159,255,0,4,159,159,244,245,
  159,159,5,7,159,159,242,241,242,243,244,159,159,255,0,4,159,159,245,245,159,
  159,6,6,246,246,241,242,243,246,159,159,255,1,3,159,159,159,159,159,7,4,245,
  245,246,159,159,159,255,2,10,159,159,159,159,159,246,245,246,159,159,159,159,
  255,0,12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,0,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,
  159,159,159,159,242,241,242,159,159,159,159,255,0,3,159,159,244,159,159,5,5,
  159,159,242,241,242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,
  242,159,159,255,0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,
  0,3,159,159,244,159,159,5,5,159,159,242,241,242,159,159,255,0,12,159,159,244,
  159,159,159,159,242,241,242,159,159,159,159,255,0,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,0,12,159,159,246,245,244,244,243,242,241,
  242,243,244,159,159,255,2,10,159,159,159,159,159,246,245,246,159,159,159,159,
  255,1,10,159,159,159,159,159,159,246,245,246,159,159,159,255,0,12,159,159,
  246,245,244,244,243,242,241,242,243,246,159,159,255,0,12,159,159,244,245,244,
  244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,159,159,159,
  245,242,243,244,159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,
  159,159,255,0,3,159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,
  159,159,244,159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,159,
  159,4,2,159,159,159,159,7,5,246,246,242,243,244,159,159,255,0,3,159,159,244,
  159,159,4,8,159,159,243,159,245,242,243,244,159,159,255,0,3,159,159,159,159,
  159,4,8,159,159,243,242,241,242,243,244,159,159,255,4,8,159,159,243,242,241,
  242,243,246,159,159,255,4,7,159,159,159,246,245,246,159,159,159,255,0,12,159,
  159,159,159,159,159,159,246,245,246,159,159,159,159,255,0,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,0,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,0,12,159,159,244,159,159,244,159,246,245,242,
  243,244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,
  244,159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,
  159,159,255,0,6,159,159,244,159,159,244,159,159,7,5,246,246,242,243,244,159,
  159,255,0,6,159,159,244,159,159,159,159,159,7,5,246,246,242,243,244,159,159,
  255,0,4,159,159,244,159,159,159,6,6,246,246,245,242,243,244,159,159,255,0,4,
  159,159,244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,
  244,245,159,159,6,6,246,246,241,242,243,244,159,159,255,0,4,159,159,159,159,
  159,159,6,6,246,246,245,246,159,159,159,159,255,1,3,159,159,159,159,159,6,5,
  246,246,245,246,159,159,159,255,0,5,159,159,245,245,245,159,159,6,6,246,246,
  241,242,243,246,159,159,255,0,5,159,159,244,245,244,159,159,6,6,246,246,241,
  242,243,244,159,159,255,0,12,159,159,244,245,244,244,159,246,241,242,243,244,
  159,159,255,0,12,159,159,244,159,244,244,246,246,241,242,243,244,159,159,255,
  0,12,159,159,244,159,246,244,243,246,245,242,243,244,159,159,255,0,12,159,
  159,244,159,159,244,243,242,245,242,243,244,159,159,255,0,12,159,159,244,159,
  159,246,243,242,241,242,243,244,159,159,255,0,12,159,159,244,246,159,159,243,
  242,241,242,243,244,159,159,255,0,4,159,159,244,245,159,159,5,7,159,159,242,
  241,242,243,244,159,159,255,0,4,159,159,245,245,159,159,6,6,246,246,241,242,
  243,246,159,159,255,1,3,159,159,159,159,159,7,4,245,245,246,159,159,159,255,
};


static const char m_colors[]=  // Palette for color selection
{
  137,  0,137,  0,  0,  0,  0,  0,
   44,  2,  0,  0,186,  2,  0,  0, 72,  3,  0,  0,214,  3,  0,  0,
  100,  4,  0,  0,242,  4,  0,  0,128,  5,  0,  0, 14,  6,  0,  0,
  156,  6,  0,  0, 42,  7,  0,  0,184,  7,  0,  0, 70,  8,  0,  0,
  212,  8,  0,  0, 98,  9,  0,  0,240,  9,  0,  0,126, 10,  0,  0,
   12, 11,  0,  0,154, 11,  0,  0, 40, 12,  0,  0,182, 12,  0,  0,
   68, 13,  0,  0,210, 13,  0,  0, 96, 14,  0,  0,238, 14,  0,  0,
  124, 15,  0,  0, 10, 16,  0,  0,152, 16,  0,  0, 38, 17,  0,  0,
  180, 17,  0,  0, 66, 18,  0,  0,208, 18,  0,  0, 94, 19,  0,  0,
  236, 19,  0,  0,122, 20,  0,  0,  8, 21,  0,  0,150, 21,  0,  0,
   36, 22,  0,  0,178, 22,  0,  0, 64, 23,  0,  0,206, 23,  0,  0,
   92, 24,  0,  0,234, 24,  0,  0,120, 25,  0,  0,  6, 26,  0,  0,
  148, 26,  0,  0, 34, 27,  0,  0,176, 27,  0,  0, 62, 28,  0,  0,
  204, 28,  0,  0, 90, 29,  0,  0,232, 29,  0,  0,118, 30,  0,  0,
    4, 31,  0,  0,146, 31,  0,  0, 32, 32,  0,  0,174, 32,  0,  0,
   60, 33,  0,  0,202, 33,  0,  0, 88, 34,  0,  0,230, 34,  0,  0,
  116, 35,  0,  0,  2, 36,  0,  0,144, 36,  0,  0, 30, 37,  0,  0,
  172, 37,  0,  0, 58, 38,  0,  0,200, 38,  0,  0, 86, 39,  0,  0,
  228, 39,  0,  0,114, 40,  0,  0,  0, 41,  0,  0,142, 41,  0,  0,
   28, 42,  0,  0,170, 42,  0,  0, 56, 43,  0,  0,198, 43,  0,  0,
   84, 44,  0,  0,226, 44,  0,  0,112, 45,  0,  0,254, 45,  0,  0,
  140, 46,  0,  0, 26, 47,  0,  0,168, 47,  0,  0, 54, 48,  0,  0,
  196, 48,  0,  0, 82, 49,  0,  0,224, 49,  0,  0,110, 50,  0,  0,
  252, 50,  0,  0,138, 51,  0,  0, 24, 52,  0,  0,166, 52,  0,  0,
   52, 53,  0,  0,194, 53,  0,  0, 80, 54,  0,  0,222, 54,  0,  0,
  108, 55,  0,  0,250, 55,  0,  0,136, 56,  0,  0, 22, 57,  0,  0,
  164, 57,  0,  0, 50, 58,  0,  0,192, 58,  0,  0, 78, 59,  0,  0,
  220, 59,  0,  0,106, 60,  0,  0,248, 60,  0,  0,134, 61,  0,  0,
   20, 62,  0,  0,162, 62,  0,  0, 48, 63,  0,  0,190, 63,  0,  0,
   76, 64,  0,  0,218, 64,  0,  0,104, 65,  0,  0,246, 65,  0,  0,
  132, 66,  0,  0, 18, 67,  0,  0,160, 67,  0,  0, 46, 68,  0,  0,
  188, 68,  0,  0, 74, 69,  0,  0,216, 69,  0,  0,102, 70,  0,  0,
  244, 70,  0,  0,130, 71,  0,  0, 16, 72,  0,  0,158, 72,  0,  0,
   44, 73,  0,  0,186, 73,  0,  0, 72, 74,  0,  0,214, 74,  0,  0,
  100, 75,  0,  0,242, 75,  0,  0,128, 76,  0,  0, 14, 77,  0,  0,
  156, 77,  0,  0,  0,137,  0, 91, 91, 91, 94, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 98,
    0,255,  0,137,  0, 91, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,105,  0,255,
    0,137,  0, 91, 98,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105, 94, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  4,  0,  0,  0,  0,  0,  4,  0, 16, 16, 16, 16, 16, 16,
   16,  0, 32, 32, 32, 32, 32, 32, 32,  0, 48, 48, 48, 48, 48, 48,
   48,  0, 64, 64, 64, 64, 64, 64, 64,  0, 80, 80, 80, 80, 80, 80,
   80,  0, 96, 96, 96, 96, 96, 96, 96,  0,112,112,112,112,112,112,
  112,  0,128,128,128,128,128,128,128,  0,144,144,144,144,144,144,
  144,  0,160,160,160,160,160,160,160,  0,176,176,176,176,176,176,
  176,  0,192,192,192,192,192,192,192,  0,255,255,255,255,255,255,
  255,  0,255,255,255,255,255,255,255,  0,207,207,207,207,207,207,
  207,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    0,  4,  0,  0,  0,  4,  0,  0, 16, 16, 16, 16, 16, 16, 16,  0,
   32, 32, 32, 32, 32, 32, 32,  0, 48, 48, 48, 48, 48, 48, 48,  0,
   64, 64, 64, 64, 64, 64, 64,  0, 80, 80, 80, 80, 80, 80, 80,  0,
   96, 96, 96, 96, 96, 96, 96,  0,112,112,112,112,112,112,112,  0,
  128,128,128,128,128,128,128,  0,144,144,144,144,144,144,144,  0,
  160,160,160,160,160,160,160,  0,176,176,176,176,176,176,176,  0,
  192,192,192,192,192,192,192,  0,255,255,255,255,255,255,255,  0,
  255,255,255,255,255,255,255,  0,207,207,207,207,207,207,207,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  0,  0,
    4,  0,  4,  0,  0,  0, 16, 16, 16, 16, 16, 16, 16,  0, 32, 32,
   32, 32, 32, 32, 32,  0, 48, 48, 48, 48, 48, 48, 48,  0, 64, 64,
   64, 64, 64, 64, 64,  0, 80, 80, 80, 80, 80, 80, 80,  0, 96, 96,
   96, 96, 96, 96, 96,  0,112,112,112,112,112,112,112,  0,128,128,
  128,128,128,128,128,  0,144,144,144,144,144,144,144,  0,160,160,
  160,160,160,160,160,  0,176,176,176,176,176,176,176,  0,192,192,
  192,192,192,192,192,  0,255,255,255,255,255,255,255,  0,255,255,
  255,255,255,255,255,  0,207,207,207,207,207,207,207,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  0,  0,  0,  4,
    0,  0,  0,  0, 16, 16, 16, 16, 16, 16, 16,  0, 32, 32, 32, 32,
   32, 32, 32,  0, 48, 48, 48, 48, 48, 48, 48,  0, 64, 64, 64, 64,
   64, 64, 64,  0, 80, 80, 80, 80, 80, 80, 80,  0, 96, 96, 96, 96,
   96, 96, 96,  0,112,112,112,112,112,112,112,  0,128,128,128,128,
  128,128,128,  0,144,144,144,144,144,144,144,  0,160,160,160,160,
  160,160,160,  0,176,176,176,176,176,176,176,  0,192,192,192,192,
  192,192,192,  0,255,255,255,255,255,255,255,  0,255,255,255,255,
  255,255,255,  0,207,207,207,207,207,207,207,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  0,  0,  4,  0,  4,  0,
    0,  0, 16, 16, 16, 16, 16, 16, 16,  0, 32, 32, 32, 32, 32, 32,
   32,  0, 48, 48, 48, 48, 48, 48, 48,  0, 64, 64, 64, 64, 64, 64,
   64,  0, 80, 80, 80, 80, 80, 80, 80,  0, 96, 96, 96, 96, 96, 96,
   96,  0,112,112,112,112,112,112,112,  0,128,128,128,128,128,128,
  128,  0,144,144,144,144,144,144,144,  0,160,160,160,160,160,160,
  160,  0,176,176,176,176,176,176,176,  0,192,192,192,192,192,192,
  192,  0,255,255,255,255,255,255,255,  0,255,255,255,255,255,255,
  255,  0,207,207,207,207,207,207,207,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  0,  4,  0,  0,  0,  4,  0,  0,
   16, 16, 16, 16, 16, 16, 16,  0, 32, 32, 32, 32, 32, 32, 32,  0,
   48, 48, 48, 48, 48, 48, 48,  0, 64, 64, 64, 64, 64, 64, 64,  0,
   80, 80, 80, 80, 80, 80, 80,  0, 96, 96, 96, 96, 96, 96, 96,  0,
  112,112,112,112,112,112,112,  0,128,128,128,128,128,128,128,  0,
  144,144,144,144,144,144,144,  0,160,160,160,160,160,160,160,  0,
  176,176,176,176,176,176,176,  0,192,192,192,192,192,192,192,  0,
  255,255,255,255,255,255,255,  0,255,255,255,255,255,255,255,  0,
  207,207,207,207,207,207,207,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  4,  0,  0,  0,  0,  0,  4,  0, 16, 16,
   16, 16, 16, 16, 16,  0, 32, 32, 32, 32, 32, 32, 32,  0, 48, 48,
   48, 48, 48, 48, 48,  0, 64, 64, 64, 64, 64, 64, 64,  0, 80, 80,
   80, 80, 80, 80, 80,  0, 96, 96, 96, 96, 96, 96, 96,  0,112,112,
  112,112,112,112,112,  0,128,128,128,128,128,128,128,  0,144,144,
  144,144,144,144,144,  0,160,160,160,160,160,160,160,  0,176,176,
  176,176,176,176,176,  0,192,192,192,192,192,192,192,  0,255,255,
  255,255,255,255,255,  0,255,255,255,255,255,255,255,  0,207,207,
  207,207,207,207,207,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  1,  1,  1,  1,  1,  1,  1,  0, 17, 17, 17, 17, 17, 17,
   17,  0, 33, 33, 33, 33, 33, 33, 33,  0, 49, 49, 49, 49, 49, 49,
   49,  0, 65, 65, 65, 65, 65, 65, 65,  0, 81, 81, 81, 81, 81, 81,
   81,  0, 97, 97, 97, 97, 97, 97, 97,  0,113,113,113,113,113,113,
  113,  0,129,129,129,129,129,129,129,  0,145,145,145,145,145,145,
  145,  0,161,161,161,161,161,161,161,  0,177,177,177,177,177,177,
  177,  0,193,193,193,193,193,193,193,  0,209,209,209,209,209,209,
  209,  0,225,225,225,225,225,225,225,  0,241,241,241,241,241,241,
  241,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    1,  1,  1,  1,  1,  1,  1,  0, 17, 17, 17, 17, 17, 17, 17,  0,
   33, 33, 33, 33, 33, 33, 33,  0, 49, 49, 49, 49, 49, 49, 49,  0,
   65, 65, 65, 65, 65, 65, 65,  0, 81, 81, 81, 81, 81, 81, 81,  0,
   97, 97, 97, 97, 97, 97, 97,  0,113,113,113,113,113,113,113,  0,
  129,129,129,129,129,129,129,  0,145,145,145,145,145,145,145,  0,
  161,161,161,161,161,161,161,  0,177,177,177,177,177,177,177,  0,
  193,193,193,193,193,193,193,  0,209,209,209,209,209,209,209,  0,
  225,225,225,225,225,225,225,  0,241,241,241,241,241,241,241,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  1,  1,
    1,  1,  1,  1,  1,  0, 17, 17, 17, 17, 17, 17, 17,  0, 33, 33,
   33, 33, 33, 33, 33,  0, 49, 49, 49, 49, 49, 49, 49,  0, 65, 65,
   65, 65, 65, 65, 65,  0, 81, 81, 81, 81, 81, 81, 81,  0, 97, 97,
   97, 97, 97, 97, 97,  0,113,113,113,113,113,113,113,  0,129,129,
  129,129,129,129,129,  0,145,145,145,145,145,145,145,  0,161,161,
  161,161,161,161,161,  0,177,177,177,177,177,177,177,  0,193,193,
  193,193,193,193,193,  0,209,209,209,209,209,209,209,  0,225,225,
  225,225,225,225,225,  0,241,241,241,241,241,241,241,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  1,  1,  1,  1,
    1,  1,  1,  0, 17, 17, 17, 17, 17, 17, 17,  0, 33, 33, 33, 33,
   33, 33, 33,  0, 49, 49, 49, 49, 49, 49, 49,  0, 65, 65, 65, 65,
   65, 65, 65,  0, 81, 81, 81, 81, 81, 81, 81,  0, 97, 97, 97, 97,
   97, 97, 97,  0,113,113,113,113,113,113,113,  0,129,129,129,129,
  129,129,129,  0,145,145,145,145,145,145,145,  0,161,161,161,161,
  161,161,161,  0,177,177,177,177,177,177,177,  0,193,193,193,193,
  193,193,193,  0,209,209,209,209,209,209,209,  0,225,225,225,225,
  225,225,225,  0,241,241,241,241,241,241,241,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  1,  1,  1,  1,  1,  1,
    1,  0, 17, 17, 17, 17, 17, 17, 17,  0, 33, 33, 33, 33, 33, 33,
   33,  0, 49, 49, 49, 49, 49, 49, 49,  0, 65, 65, 65, 65, 65, 65,
   65,  0, 81, 81, 81, 81, 81, 81, 81,  0, 97, 97, 97, 97, 97, 97,
   97,  0,113,113,113,113,113,113,113,  0,129,129,129,129,129,129,
  129,  0,145,145,145,145,145,145,145,  0,161,161,161,161,161,161,
  161,  0,177,177,177,177,177,177,177,  0,193,193,193,193,193,193,
  193,  0,209,209,209,209,209,209,209,  0,225,225,225,225,225,225,
  225,  0,241,241,241,241,241,241,241,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  1,  1,  1,  1,  1,  1,  1,  0,
   17, 17, 17, 17, 17, 17, 17,  0, 33, 33, 33, 33, 33, 33, 33,  0,
   49, 49, 49, 49, 49, 49, 49,  0, 65, 65, 65, 65, 65, 65, 65,  0,
   81, 81, 81, 81, 81, 81, 81,  0, 97, 97, 97, 97, 97, 97, 97,  0,
  113,113,113,113,113,113,113,  0,129,129,129,129,129,129,129,  0,
  145,145,145,145,145,145,145,  0,161,161,161,161,161,161,161,  0,
  177,177,177,177,177,177,177,  0,193,193,193,193,193,193,193,  0,
  209,209,209,209,209,209,209,  0,225,225,225,225,225,225,225,  0,
  241,241,241,241,241,241,241,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  1,  1,  1,  1,  1,  1,  1,  0, 17, 17,
   17, 17, 17, 17, 17,  0, 33, 33, 33, 33, 33, 33, 33,  0, 49, 49,
   49, 49, 49, 49, 49,  0, 65, 65, 65, 65, 65, 65, 65,  0, 81, 81,
   81, 81, 81, 81, 81,  0, 97, 97, 97, 97, 97, 97, 97,  0,113,113,
  113,113,113,113,113,  0,129,129,129,129,129,129,129,  0,145,145,
  145,145,145,145,145,  0,161,161,161,161,161,161,161,  0,177,177,
  177,177,177,177,177,  0,193,193,193,193,193,193,193,  0,209,209,
  209,209,209,209,209,  0,225,225,225,225,225,225,225,  0,241,241,
  241,241,241,241,241,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  2,  2,  2,  2,  2,  2,  2,  0, 18, 18, 18, 18, 18, 18,
   18,  0, 34, 34, 34, 34, 34, 34, 34,  0, 50, 50, 50, 50, 50, 50,
   50,  0, 66, 66, 66, 66, 66, 66, 66,  0, 82, 82, 82, 82, 82, 82,
   82,  0, 98, 98, 98, 98, 98, 98, 98,  0,114,114,114,114,114,114,
  114,  0,130,130,130,130,130,130,130,  0,146,146,146,146,146,146,
  146,  0,162,162,162,162,162,162,162,  0,178,178,178,178,178,178,
  178,  0,194,194,194,194,194,194,194,  0,210,210,210,210,210,210,
  210,  0,226,226,226,226,226,226,226,  0,242,242,242,242,242,242,
  242,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    2,  2,  2,  2,  2,  2,  2,  0, 18, 18, 18, 18, 18, 18, 18,  0,
   34, 34, 34, 34, 34, 34, 34,  0, 50, 50, 50, 50, 50, 50, 50,  0,
   66, 66, 66, 66, 66, 66, 66,  0, 82, 82, 82, 82, 82, 82, 82,  0,
   98, 98, 98, 98, 98, 98, 98,  0,114,114,114,114,114,114,114,  0,
  130,130,130,130,130,130,130,  0,146,146,146,146,146,146,146,  0,
  162,162,162,162,162,162,162,  0,178,178,178,178,178,178,178,  0,
  194,194,194,194,194,194,194,  0,210,210,210,210,210,210,210,  0,
  226,226,226,226,226,226,226,  0,242,242,242,242,242,242,242,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  2,  2,
    2,  2,  2,  2,  2,  0, 18, 18, 18, 18, 18, 18, 18,  0, 34, 34,
   34, 34, 34, 34, 34,  0, 50, 50, 50, 50, 50, 50, 50,  0, 66, 66,
   66, 66, 66, 66, 66,  0, 82, 82, 82, 82, 82, 82, 82,  0, 98, 98,
   98, 98, 98, 98, 98,  0,114,114,114,114,114,114,114,  0,130,130,
  130,130,130,130,130,  0,146,146,146,146,146,146,146,  0,162,162,
  162,162,162,162,162,  0,178,178,178,178,178,178,178,  0,194,194,
  194,194,194,194,194,  0,210,210,210,210,210,210,210,  0,226,226,
  226,226,226,226,226,  0,242,242,242,242,242,242,242,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  2,  2,  2,  2,
    2,  2,  2,  0, 18, 18, 18, 18, 18, 18, 18,  0, 34, 34, 34, 34,
   34, 34, 34,  0, 50, 50, 50, 50, 50, 50, 50,  0, 66, 66, 66, 66,
   66, 66, 66,  0, 82, 82, 82, 82, 82, 82, 82,  0, 98, 98, 98, 98,
   98, 98, 98,  0,114,114,114,114,114,114,114,  0,130,130,130,130,
  130,130,130,  0,146,146,146,146,146,146,146,  0,162,162,162,162,
  162,162,162,  0,178,178,178,178,178,178,178,  0,194,194,194,194,
  194,194,194,  0,210,210,210,210,210,210,210,  0,226,226,226,226,
  226,226,226,  0,242,242,242,242,242,242,242,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  2,  2,  2,  2,  2,  2,
    2,  0, 18, 18, 18, 18, 18, 18, 18,  0, 34, 34, 34, 34, 34, 34,
   34,  0, 50, 50, 50, 50, 50, 50, 50,  0, 66, 66, 66, 66, 66, 66,
   66,  0, 82, 82, 82, 82, 82, 82, 82,  0, 98, 98, 98, 98, 98, 98,
   98,  0,114,114,114,114,114,114,114,  0,130,130,130,130,130,130,
  130,  0,146,146,146,146,146,146,146,  0,162,162,162,162,162,162,
  162,  0,178,178,178,178,178,178,178,  0,194,194,194,194,194,194,
  194,  0,210,210,210,210,210,210,210,  0,226,226,226,226,226,226,
  226,  0,242,242,242,242,242,242,242,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  2,  2,  2,  2,  2,  2,  2,  0,
   18, 18, 18, 18, 18, 18, 18,  0, 34, 34, 34, 34, 34, 34, 34,  0,
   50, 50, 50, 50, 50, 50, 50,  0, 66, 66, 66, 66, 66, 66, 66,  0,
   82, 82, 82, 82, 82, 82, 82,  0, 98, 98, 98, 98, 98, 98, 98,  0,
  114,114,114,114,114,114,114,  0,130,130,130,130,130,130,130,  0,
  146,146,146,146,146,146,146,  0,162,162,162,162,162,162,162,  0,
  178,178,178,178,178,178,178,  0,194,194,194,194,194,194,194,  0,
  210,210,210,210,210,210,210,  0,226,226,226,226,226,226,226,  0,
  242,242,242,242,242,242,242,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  2,  2,  2,  2,  2,  2,  2,  0, 18, 18,
   18, 18, 18, 18, 18,  0, 34, 34, 34, 34, 34, 34, 34,  0, 50, 50,
   50, 50, 50, 50, 50,  0, 66, 66, 66, 66, 66, 66, 66,  0, 82, 82,
   82, 82, 82, 82, 82,  0, 98, 98, 98, 98, 98, 98, 98,  0,114,114,
  114,114,114,114,114,  0,130,130,130,130,130,130,130,  0,146,146,
  146,146,146,146,146,  0,162,162,162,162,162,162,162,  0,178,178,
  178,178,178,178,178,  0,194,194,194,194,194,194,194,  0,210,210,
  210,210,210,210,210,  0,226,226,226,226,226,226,226,  0,242,242,
  242,242,242,242,242,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  3,  3,  3,  3,  3,  3,  3,  0, 19, 19, 19, 19, 19, 19,
   19,  0, 35, 35, 35, 35, 35, 35, 35,  0, 51, 51, 51, 51, 51, 51,
   51,  0, 67, 67, 67, 67, 67, 67, 67,  0, 83, 83, 83, 83, 83, 83,
   83,  0, 99, 99, 99, 99, 99, 99, 99,  0,115,115,115,115,115,115,
  115,  0,131,131,131,131,131,131,131,  0,147,147,147,147,147,147,
  147,  0,163,163,163,163,163,163,163,  0,179,179,179,179,179,179,
  179,  0,195,195,195,195,195,195,195,  0,211,211,211,211,211,211,
  211,  0,227,227,227,227,227,227,227,  0,243,243,243,243,243,243,
  243,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    3,  3,  3,  3,  3,  3,  3,  0, 19, 19, 19, 19, 19, 19, 19,  0,
   35, 35, 35, 35, 35, 35, 35,  0, 51, 51, 51, 51, 51, 51, 51,  0,
   67, 67, 67, 67, 67, 67, 67,  0, 83, 83, 83, 83, 83, 83, 83,  0,
   99, 99, 99, 99, 99, 99, 99,  0,115,115,115,115,115,115,115,  0,
  131,131,131,131,131,131,131,  0,147,147,147,147,147,147,147,  0,
  163,163,163,163,163,163,163,  0,179,179,179,179,179,179,179,  0,
  195,195,195,195,195,195,195,  0,211,211,211,211,211,211,211,  0,
  227,227,227,227,227,227,227,  0,243,243,243,243,243,243,243,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  3,  3,
    3,  3,  3,  3,  3,  0, 19, 19, 19, 19, 19, 19, 19,  0, 35, 35,
   35, 35, 35, 35, 35,  0, 51, 51, 51, 51, 51, 51, 51,  0, 67, 67,
   67, 67, 67, 67, 67,  0, 83, 83, 83, 83, 83, 83, 83,  0, 99, 99,
   99, 99, 99, 99, 99,  0,115,115,115,115,115,115,115,  0,131,131,
  131,131,131,131,131,  0,147,147,147,147,147,147,147,  0,163,163,
  163,163,163,163,163,  0,179,179,179,179,179,179,179,  0,195,195,
  195,195,195,195,195,  0,211,211,211,211,211,211,211,  0,227,227,
  227,227,227,227,227,  0,243,243,243,243,243,243,243,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  3,  3,  3,  3,
    3,  3,  3,  0, 19, 19, 19, 19, 19, 19, 19,  0, 35, 35, 35, 35,
   35, 35, 35,  0, 51, 51, 51, 51, 51, 51, 51,  0, 67, 67, 67, 67,
   67, 67, 67,  0, 83, 83, 83, 83, 83, 83, 83,  0, 99, 99, 99, 99,
   99, 99, 99,  0,115,115,115,115,115,115,115,  0,131,131,131,131,
  131,131,131,  0,147,147,147,147,147,147,147,  0,163,163,163,163,
  163,163,163,  0,179,179,179,179,179,179,179,  0,195,195,195,195,
  195,195,195,  0,211,211,211,211,211,211,211,  0,227,227,227,227,
  227,227,227,  0,243,243,243,243,243,243,243,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  3,  3,  3,  3,  3,  3,
    3,  0, 19, 19, 19, 19, 19, 19, 19,  0, 35, 35, 35, 35, 35, 35,
   35,  0, 51, 51, 51, 51, 51, 51, 51,  0, 67, 67, 67, 67, 67, 67,
   67,  0, 83, 83, 83, 83, 83, 83, 83,  0, 99, 99, 99, 99, 99, 99,
   99,  0,115,115,115,115,115,115,115,  0,131,131,131,131,131,131,
  131,  0,147,147,147,147,147,147,147,  0,163,163,163,163,163,163,
  163,  0,179,179,179,179,179,179,179,  0,195,195,195,195,195,195,
  195,  0,211,211,211,211,211,211,211,  0,227,227,227,227,227,227,
  227,  0,243,243,243,243,243,243,243,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  3,  3,  3,  3,  3,  3,  3,  0,
   19, 19, 19, 19, 19, 19, 19,  0, 35, 35, 35, 35, 35, 35, 35,  0,
   51, 51, 51, 51, 51, 51, 51,  0, 67, 67, 67, 67, 67, 67, 67,  0,
   83, 83, 83, 83, 83, 83, 83,  0, 99, 99, 99, 99, 99, 99, 99,  0,
  115,115,115,115,115,115,115,  0,131,131,131,131,131,131,131,  0,
  147,147,147,147,147,147,147,  0,163,163,163,163,163,163,163,  0,
  179,179,179,179,179,179,179,  0,195,195,195,195,195,195,195,  0,
  211,211,211,211,211,211,211,  0,227,227,227,227,227,227,227,  0,
  243,243,243,243,243,243,243,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  3,  3,  3,  3,  3,  3,  3,  0, 19, 19,
   19, 19, 19, 19, 19,  0, 35, 35, 35, 35, 35, 35, 35,  0, 51, 51,
   51, 51, 51, 51, 51,  0, 67, 67, 67, 67, 67, 67, 67,  0, 83, 83,
   83, 83, 83, 83, 83,  0, 99, 99, 99, 99, 99, 99, 99,  0,115,115,
  115,115,115,115,115,  0,131,131,131,131,131,131,131,  0,147,147,
  147,147,147,147,147,  0,163,163,163,163,163,163,163,  0,179,179,
  179,179,179,179,179,  0,195,195,195,195,195,195,195,  0,211,211,
  211,211,211,211,211,  0,227,227,227,227,227,227,227,  0,243,243,
  243,243,243,243,243,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  4,  4,  4,  4,  4,  4,  4,  0, 20, 20, 20, 20, 20, 20,
   20,  0, 36, 36, 36, 36, 36, 36, 36,  0, 52, 52, 52, 52, 52, 52,
   52,  0, 68, 68, 68, 68, 68, 68, 68,  0, 84, 84, 84, 84, 84, 84,
   84,  0,100,100,100,100,100,100,100,  0,116,116,116,116,116,116,
  116,  0,132,132,132,132,132,132,132,  0,148,148,148,148,148,148,
  148,  0,164,164,164,164,164,164,164,  0,180,180,180,180,180,180,
  180,  0,196,196,196,196,196,196,196,  0,212,212,212,212,212,212,
  212,  0,228,228,228,228,228,228,228,  0,244,244,244,244,244,244,
  244,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    4,  4,  4,  4,  4,  4,  4,  0, 20, 20, 20, 20, 20, 20, 20,  0,
   36, 36, 36, 36, 36, 36, 36,  0, 52, 52, 52, 52, 52, 52, 52,  0,
   68, 68, 68, 68, 68, 68, 68,  0, 84, 84, 84, 84, 84, 84, 84,  0,
  100,100,100,100,100,100,100,  0,116,116,116,116,116,116,116,  0,
  132,132,132,132,132,132,132,  0,148,148,148,148,148,148,148,  0,
  164,164,164,164,164,164,164,  0,180,180,180,180,180,180,180,  0,
  196,196,196,196,196,196,196,  0,212,212,212,212,212,212,212,  0,
  228,228,228,228,228,228,228,  0,244,244,244,244,244,244,244,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  4,  4,
    4,  4,  4,  4,  4,  0, 20, 20, 20, 20, 20, 20, 20,  0, 36, 36,
   36, 36, 36, 36, 36,  0, 52, 52, 52, 52, 52, 52, 52,  0, 68, 68,
   68, 68, 68, 68, 68,  0, 84, 84, 84, 84, 84, 84, 84,  0,100,100,
  100,100,100,100,100,  0,116,116,116,116,116,116,116,  0,132,132,
  132,132,132,132,132,  0,148,148,148,148,148,148,148,  0,164,164,
  164,164,164,164,164,  0,180,180,180,180,180,180,180,  0,196,196,
  196,196,196,196,196,  0,212,212,212,212,212,212,212,  0,228,228,
  228,228,228,228,228,  0,244,244,244,244,244,244,244,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  4,  4,  4,  4,
    4,  4,  4,  0, 20, 20, 20, 20, 20, 20, 20,  0, 36, 36, 36, 36,
   36, 36, 36,  0, 52, 52, 52, 52, 52, 52, 52,  0, 68, 68, 68, 68,
   68, 68, 68,  0, 84, 84, 84, 84, 84, 84, 84,  0,100,100,100,100,
  100,100,100,  0,116,116,116,116,116,116,116,  0,132,132,132,132,
  132,132,132,  0,148,148,148,148,148,148,148,  0,164,164,164,164,
  164,164,164,  0,180,180,180,180,180,180,180,  0,196,196,196,196,
  196,196,196,  0,212,212,212,212,212,212,212,  0,228,228,228,228,
  228,228,228,  0,244,244,244,244,244,244,244,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  4,  4,  4,  4,  4,  4,
    4,  0, 20, 20, 20, 20, 20, 20, 20,  0, 36, 36, 36, 36, 36, 36,
   36,  0, 52, 52, 52, 52, 52, 52, 52,  0, 68, 68, 68, 68, 68, 68,
   68,  0, 84, 84, 84, 84, 84, 84, 84,  0,100,100,100,100,100,100,
  100,  0,116,116,116,116,116,116,116,  0,132,132,132,132,132,132,
  132,  0,148,148,148,148,148,148,148,  0,164,164,164,164,164,164,
  164,  0,180,180,180,180,180,180,180,  0,196,196,196,196,196,196,
  196,  0,212,212,212,212,212,212,212,  0,228,228,228,228,228,228,
  228,  0,244,244,244,244,244,244,244,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  4,  4,  4,  4,  4,  4,  4,  0,
   20, 20, 20, 20, 20, 20, 20,  0, 36, 36, 36, 36, 36, 36, 36,  0,
   52, 52, 52, 52, 52, 52, 52,  0, 68, 68, 68, 68, 68, 68, 68,  0,
   84, 84, 84, 84, 84, 84, 84,  0,100,100,100,100,100,100,100,  0,
  116,116,116,116,116,116,116,  0,132,132,132,132,132,132,132,  0,
  148,148,148,148,148,148,148,  0,164,164,164,164,164,164,164,  0,
  180,180,180,180,180,180,180,  0,196,196,196,196,196,196,196,  0,
  212,212,212,212,212,212,212,  0,228,228,228,228,228,228,228,  0,
  244,244,244,244,244,244,244,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  4,  4,  4,  4,  4,  4,  4,  0, 20, 20,
   20, 20, 20, 20, 20,  0, 36, 36, 36, 36, 36, 36, 36,  0, 52, 52,
   52, 52, 52, 52, 52,  0, 68, 68, 68, 68, 68, 68, 68,  0, 84, 84,
   84, 84, 84, 84, 84,  0,100,100,100,100,100,100,100,  0,116,116,
  116,116,116,116,116,  0,132,132,132,132,132,132,132,  0,148,148,
  148,148,148,148,148,  0,164,164,164,164,164,164,164,  0,180,180,
  180,180,180,180,180,  0,196,196,196,196,196,196,196,  0,212,212,
  212,212,212,212,212,  0,228,228,228,228,228,228,228,  0,244,244,
  244,244,244,244,244,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  5,  5,  5,  5,  5,  5,  5,  0, 21, 21, 21, 21, 21, 21,
   21,  0, 37, 37, 37, 37, 37, 37, 37,  0, 53, 53, 53, 53, 53, 53,
   53,  0, 69, 69, 69, 69, 69, 69, 69,  0, 85, 85, 85, 85, 85, 85,
   85,  0,101,101,101,101,101,101,101,  0,117,117,117,117,117,117,
  117,  0,133,133,133,133,133,133,133,  0,149,149,149,149,149,149,
  149,  0,165,165,165,165,165,165,165,  0,181,181,181,181,181,181,
  181,  0,197,197,197,197,197,197,197,  0,213,213,213,213,213,213,
  213,  0,229,229,229,229,229,229,229,  0,245,245,245,245,245,245,
  245,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    5,  5,  5,  5,  5,  5,  5,  0, 21, 21, 21, 21, 21, 21, 21,  0,
   37, 37, 37, 37, 37, 37, 37,  0, 53, 53, 53, 53, 53, 53, 53,  0,
   69, 69, 69, 69, 69, 69, 69,  0, 85, 85, 85, 85, 85, 85, 85,  0,
  101,101,101,101,101,101,101,  0,117,117,117,117,117,117,117,  0,
  133,133,133,133,133,133,133,  0,149,149,149,149,149,149,149,  0,
  165,165,165,165,165,165,165,  0,181,181,181,181,181,181,181,  0,
  197,197,197,197,197,197,197,  0,213,213,213,213,213,213,213,  0,
  229,229,229,229,229,229,229,  0,245,245,245,245,245,245,245,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  5,  5,
    5,  5,  5,  5,  5,  0, 21, 21, 21, 21, 21, 21, 21,  0, 37, 37,
   37, 37, 37, 37, 37,  0, 53, 53, 53, 53, 53, 53, 53,  0, 69, 69,
   69, 69, 69, 69, 69,  0, 85, 85, 85, 85, 85, 85, 85,  0,101,101,
  101,101,101,101,101,  0,117,117,117,117,117,117,117,  0,133,133,
  133,133,133,133,133,  0,149,149,149,149,149,149,149,  0,165,165,
  165,165,165,165,165,  0,181,181,181,181,181,181,181,  0,197,197,
  197,197,197,197,197,  0,213,213,213,213,213,213,213,  0,229,229,
  229,229,229,229,229,  0,245,245,245,245,245,245,245,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  5,  5,  5,  5,
    5,  5,  5,  0, 21, 21, 21, 21, 21, 21, 21,  0, 37, 37, 37, 37,
   37, 37, 37,  0, 53, 53, 53, 53, 53, 53, 53,  0, 69, 69, 69, 69,
   69, 69, 69,  0, 85, 85, 85, 85, 85, 85, 85,  0,101,101,101,101,
  101,101,101,  0,117,117,117,117,117,117,117,  0,133,133,133,133,
  133,133,133,  0,149,149,149,149,149,149,149,  0,165,165,165,165,
  165,165,165,  0,181,181,181,181,181,181,181,  0,197,197,197,197,
  197,197,197,  0,213,213,213,213,213,213,213,  0,229,229,229,229,
  229,229,229,  0,245,245,245,245,245,245,245,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  5,  5,  5,  5,  5,  5,
    5,  0, 21, 21, 21, 21, 21, 21, 21,  0, 37, 37, 37, 37, 37, 37,
   37,  0, 53, 53, 53, 53, 53, 53, 53,  0, 69, 69, 69, 69, 69, 69,
   69,  0, 85, 85, 85, 85, 85, 85, 85,  0,101,101,101,101,101,101,
  101,  0,117,117,117,117,117,117,117,  0,133,133,133,133,133,133,
  133,  0,149,149,149,149,149,149,149,  0,165,165,165,165,165,165,
  165,  0,181,181,181,181,181,181,181,  0,197,197,197,197,197,197,
  197,  0,213,213,213,213,213,213,213,  0,229,229,229,229,229,229,
  229,  0,245,245,245,245,245,245,245,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  5,  5,  5,  5,  5,  5,  5,  0,
   21, 21, 21, 21, 21, 21, 21,  0, 37, 37, 37, 37, 37, 37, 37,  0,
   53, 53, 53, 53, 53, 53, 53,  0, 69, 69, 69, 69, 69, 69, 69,  0,
   85, 85, 85, 85, 85, 85, 85,  0,101,101,101,101,101,101,101,  0,
  117,117,117,117,117,117,117,  0,133,133,133,133,133,133,133,  0,
  149,149,149,149,149,149,149,  0,165,165,165,165,165,165,165,  0,
  181,181,181,181,181,181,181,  0,197,197,197,197,197,197,197,  0,
  213,213,213,213,213,213,213,  0,229,229,229,229,229,229,229,  0,
  245,245,245,245,245,245,245,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  5,  5,  5,  5,  5,  5,  5,  0, 21, 21,
   21, 21, 21, 21, 21,  0, 37, 37, 37, 37, 37, 37, 37,  0, 53, 53,
   53, 53, 53, 53, 53,  0, 69, 69, 69, 69, 69, 69, 69,  0, 85, 85,
   85, 85, 85, 85, 85,  0,101,101,101,101,101,101,101,  0,117,117,
  117,117,117,117,117,  0,133,133,133,133,133,133,133,  0,149,149,
  149,149,149,149,149,  0,165,165,165,165,165,165,165,  0,181,181,
  181,181,181,181,181,  0,197,197,197,197,197,197,197,  0,213,213,
  213,213,213,213,213,  0,229,229,229,229,229,229,229,  0,245,245,
  245,245,245,245,245,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  6,  6,  6,  6,  6,  6,  6,  0, 22, 22, 22, 22, 22, 22,
   22,  0, 38, 38, 38, 38, 38, 38, 38,  0, 54, 54, 54, 54, 54, 54,
   54,  0, 70, 70, 70, 70, 70, 70, 70,  0, 86, 86, 86, 86, 86, 86,
   86,  0,102,102,102,102,102,102,102,  0,118,118,118,118,118,118,
  118,  0,134,134,134,134,134,134,134,  0,150,150,150,150,150,150,
  150,  0,166,166,166,166,166,166,166,  0,182,182,182,182,182,182,
  182,  0,198,198,198,198,198,198,198,  0,214,214,214,214,214,214,
  214,  0,230,230,230,230,230,230,230,  0,246,246,246,246,246,246,
  246,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    6,  6,  6,  6,  6,  6,  6,  0, 22, 22, 22, 22, 22, 22, 22,  0,
   38, 38, 38, 38, 38, 38, 38,  0, 54, 54, 54, 54, 54, 54, 54,  0,
   70, 70, 70, 70, 70, 70, 70,  0, 86, 86, 86, 86, 86, 86, 86,  0,
  102,102,102,102,102,102,102,  0,118,118,118,118,118,118,118,  0,
  134,134,134,134,134,134,134,  0,150,150,150,150,150,150,150,  0,
  166,166,166,166,166,166,166,  0,182,182,182,182,182,182,182,  0,
  198,198,198,198,198,198,198,  0,214,214,214,214,214,214,214,  0,
  230,230,230,230,230,230,230,  0,246,246,246,246,246,246,246,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  6,  6,
    6,  6,  6,  6,  6,  0, 22, 22, 22, 22, 22, 22, 22,  0, 38, 38,
   38, 38, 38, 38, 38,  0, 54, 54, 54, 54, 54, 54, 54,  0, 70, 70,
   70, 70, 70, 70, 70,  0, 86, 86, 86, 86, 86, 86, 86,  0,102,102,
  102,102,102,102,102,  0,118,118,118,118,118,118,118,  0,134,134,
  134,134,134,134,134,  0,150,150,150,150,150,150,150,  0,166,166,
  166,166,166,166,166,  0,182,182,182,182,182,182,182,  0,198,198,
  198,198,198,198,198,  0,214,214,214,214,214,214,214,  0,230,230,
  230,230,230,230,230,  0,246,246,246,246,246,246,246,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  6,  6,  6,  6,
    6,  6,  6,  0, 22, 22, 22, 22, 22, 22, 22,  0, 38, 38, 38, 38,
   38, 38, 38,  0, 54, 54, 54, 54, 54, 54, 54,  0, 70, 70, 70, 70,
   70, 70, 70,  0, 86, 86, 86, 86, 86, 86, 86,  0,102,102,102,102,
  102,102,102,  0,118,118,118,118,118,118,118,  0,134,134,134,134,
  134,134,134,  0,150,150,150,150,150,150,150,  0,166,166,166,166,
  166,166,166,  0,182,182,182,182,182,182,182,  0,198,198,198,198,
  198,198,198,  0,214,214,214,214,214,214,214,  0,230,230,230,230,
  230,230,230,  0,246,246,246,246,246,246,246,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  6,  6,  6,  6,  6,  6,
    6,  0, 22, 22, 22, 22, 22, 22, 22,  0, 38, 38, 38, 38, 38, 38,
   38,  0, 54, 54, 54, 54, 54, 54, 54,  0, 70, 70, 70, 70, 70, 70,
   70,  0, 86, 86, 86, 86, 86, 86, 86,  0,102,102,102,102,102,102,
  102,  0,118,118,118,118,118,118,118,  0,134,134,134,134,134,134,
  134,  0,150,150,150,150,150,150,150,  0,166,166,166,166,166,166,
  166,  0,182,182,182,182,182,182,182,  0,198,198,198,198,198,198,
  198,  0,214,214,214,214,214,214,214,  0,230,230,230,230,230,230,
  230,  0,246,246,246,246,246,246,246,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  6,  6,  6,  6,  6,  6,  6,  0,
   22, 22, 22, 22, 22, 22, 22,  0, 38, 38, 38, 38, 38, 38, 38,  0,
   54, 54, 54, 54, 54, 54, 54,  0, 70, 70, 70, 70, 70, 70, 70,  0,
   86, 86, 86, 86, 86, 86, 86,  0,102,102,102,102,102,102,102,  0,
  118,118,118,118,118,118,118,  0,134,134,134,134,134,134,134,  0,
  150,150,150,150,150,150,150,  0,166,166,166,166,166,166,166,  0,
  182,182,182,182,182,182,182,  0,198,198,198,198,198,198,198,  0,
  214,214,214,214,214,214,214,  0,230,230,230,230,230,230,230,  0,
  246,246,246,246,246,246,246,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  6,  6,  6,  6,  6,  6,  6,  0, 22, 22,
   22, 22, 22, 22, 22,  0, 38, 38, 38, 38, 38, 38, 38,  0, 54, 54,
   54, 54, 54, 54, 54,  0, 70, 70, 70, 70, 70, 70, 70,  0, 86, 86,
   86, 86, 86, 86, 86,  0,102,102,102,102,102,102,102,  0,118,118,
  118,118,118,118,118,  0,134,134,134,134,134,134,134,  0,150,150,
  150,150,150,150,150,  0,166,166,166,166,166,166,166,  0,182,182,
  182,182,182,182,182,  0,198,198,198,198,198,198,198,  0,214,214,
  214,214,214,214,214,  0,230,230,230,230,230,230,230,  0,246,246,
  246,246,246,246,246,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  7,  7,  7,  7,  7,  7,  7,  0, 23, 23, 23, 23, 23, 23,
   23,  0, 39, 39, 39, 39, 39, 39, 39,  0, 55, 55, 55, 55, 55, 55,
   55,  0, 71, 71, 71, 71, 71, 71, 71,  0, 87, 87, 87, 87, 87, 87,
   87,  0,103,103,103,103,103,103,103,  0,119,119,119,119,119,119,
  119,  0,135,135,135,135,135,135,135,  0,151,151,151,151,151,151,
  151,  0,167,167,167,167,167,167,167,  0,183,183,183,183,183,183,
  183,  0,199,199,199,199,199,199,199,  0,215,215,215,215,215,215,
  215,  0,231,231,231,231,231,231,231,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    7,  7,  7,  7,  7,  7,  7,  0, 23, 23, 23, 23, 23, 23, 23,  0,
   39, 39, 39, 39, 39, 39, 39,  0, 55, 55, 55, 55, 55, 55, 55,  0,
   71, 71, 71, 71, 71, 71, 71,  0, 87, 87, 87, 87, 87, 87, 87,  0,
  103,103,103,103,103,103,103,  0,119,119,119,119,119,119,119,  0,
  135,135,135,135,135,135,135,  0,151,151,151,151,151,151,151,  0,
  167,167,167,167,167,167,167,  0,183,183,183,183,183,183,183,  0,
  199,199,199,199,199,199,199,  0,215,215,215,215,215,215,215,  0,
  231,231,231,231,231,231,231,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  7,  7,
    7,  7,  7,  7,  7,  0, 23, 23, 23, 23, 23, 23, 23,  0, 39, 39,
   39, 39, 39, 39, 39,  0, 55, 55, 55, 55, 55, 55, 55,  0, 71, 71,
   71, 71, 71, 71, 71,  0, 87, 87, 87, 87, 87, 87, 87,  0,103,103,
  103,103,103,103,103,  0,119,119,119,119,119,119,119,  0,135,135,
  135,135,135,135,135,  0,151,151,151,151,151,151,151,  0,167,167,
  167,167,167,167,167,  0,183,183,183,183,183,183,183,  0,199,199,
  199,199,199,199,199,  0,215,215,215,215,215,215,215,  0,231,231,
  231,231,231,231,231,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  7,  7,  7,  7,
    7,  7,  7,  0, 23, 23, 23, 23, 23, 23, 23,  0, 39, 39, 39, 39,
   39, 39, 39,  0, 55, 55, 55, 55, 55, 55, 55,  0, 71, 71, 71, 71,
   71, 71, 71,  0, 87, 87, 87, 87, 87, 87, 87,  0,103,103,103,103,
  103,103,103,  0,119,119,119,119,119,119,119,  0,135,135,135,135,
  135,135,135,  0,151,151,151,151,151,151,151,  0,167,167,167,167,
  167,167,167,  0,183,183,183,183,183,183,183,  0,199,199,199,199,
  199,199,199,  0,215,215,215,215,215,215,215,  0,231,231,231,231,
  231,231,231,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  7,  7,  7,  7,  7,  7,
    7,  0, 23, 23, 23, 23, 23, 23, 23,  0, 39, 39, 39, 39, 39, 39,
   39,  0, 55, 55, 55, 55, 55, 55, 55,  0, 71, 71, 71, 71, 71, 71,
   71,  0, 87, 87, 87, 87, 87, 87, 87,  0,103,103,103,103,103,103,
  103,  0,119,119,119,119,119,119,119,  0,135,135,135,135,135,135,
  135,  0,151,151,151,151,151,151,151,  0,167,167,167,167,167,167,
  167,  0,183,183,183,183,183,183,183,  0,199,199,199,199,199,199,
  199,  0,215,215,215,215,215,215,215,  0,231,231,231,231,231,231,
  231,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  7,  7,  7,  7,  7,  7,  7,  0,
   23, 23, 23, 23, 23, 23, 23,  0, 39, 39, 39, 39, 39, 39, 39,  0,
   55, 55, 55, 55, 55, 55, 55,  0, 71, 71, 71, 71, 71, 71, 71,  0,
   87, 87, 87, 87, 87, 87, 87,  0,103,103,103,103,103,103,103,  0,
  119,119,119,119,119,119,119,  0,135,135,135,135,135,135,135,  0,
  151,151,151,151,151,151,151,  0,167,167,167,167,167,167,167,  0,
  183,183,183,183,183,183,183,  0,199,199,199,199,199,199,199,  0,
  215,215,215,215,215,215,215,  0,231,231,231,231,231,231,231,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  7,  7,  7,  7,  7,  7,  7,  0, 23, 23,
   23, 23, 23, 23, 23,  0, 39, 39, 39, 39, 39, 39, 39,  0, 55, 55,
   55, 55, 55, 55, 55,  0, 71, 71, 71, 71, 71, 71, 71,  0, 87, 87,
   87, 87, 87, 87, 87,  0,103,103,103,103,103,103,103,  0,119,119,
  119,119,119,119,119,  0,135,135,135,135,135,135,135,  0,151,151,
  151,151,151,151,151,  0,167,167,167,167,167,167,167,  0,183,183,
  183,183,183,183,183,  0,199,199,199,199,199,199,199,  0,215,215,
  215,215,215,215,215,  0,231,231,231,231,231,231,231,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  8,  8,  8,  8,  8,  8,  8,  0, 24, 24, 24, 24, 24, 24,
   24,  0, 40, 40, 40, 40, 40, 40, 40,  0, 56, 56, 56, 56, 56, 56,
   56,  0, 72, 72, 72, 72, 72, 72, 72,  0, 88, 88, 88, 88, 88, 88,
   88,  0,104,104,104,104,104,104,104,  0,120,120,120,120,120,120,
  120,  0,136,136,136,136,136,136,136,  0,152,152,152,152,152,152,
  152,  0,255,255,255,255,255,255,255,  0,184,184,184,184,184,184,
  184,  0,200,200,200,200,200,200,200,  0,216,216,216,216,216,216,
  216,  0,232,232,232,232,232,232,232,  0,248,248,248,248,248,248,
  248,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    8,  8,  8,  8,  8,  8,  8,  0, 24, 24, 24, 24, 24, 24, 24,  0,
   40, 40, 40, 40, 40, 40, 40,  0, 56, 56, 56, 56, 56, 56, 56,  0,
   72, 72, 72, 72, 72, 72, 72,  0, 88, 88, 88, 88, 88, 88, 88,  0,
  104,104,104,104,104,104,104,  0,120,120,120,120,120,120,120,  0,
  136,136,136,136,136,136,136,  0,152,152,152,152,152,152,152,  0,
  255,255,255,255,255,255,255,  0,184,184,184,184,184,184,184,  0,
  200,200,200,200,200,200,200,  0,216,216,216,216,216,216,216,  0,
  232,232,232,232,232,232,232,  0,248,248,248,248,248,248,248,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  8,  8,
    8,  8,  8,  8,  8,  0, 24, 24, 24, 24, 24, 24, 24,  0, 40, 40,
   40, 40, 40, 40, 40,  0, 56, 56, 56, 56, 56, 56, 56,  0, 72, 72,
   72, 72, 72, 72, 72,  0, 88, 88, 88, 88, 88, 88, 88,  0,104,104,
  104,104,104,104,104,  0,120,120,120,120,120,120,120,  0,136,136,
  136,136,136,136,136,  0,152,152,152,152,152,152,152,  0,255,255,
  255,255,255,255,255,  0,184,184,184,184,184,184,184,  0,200,200,
  200,200,200,200,200,  0,216,216,216,216,216,216,216,  0,232,232,
  232,232,232,232,232,  0,248,248,248,248,248,248,248,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  8,  8,  8,  8,
    8,  8,  8,  0, 24, 24, 24, 24, 24, 24, 24,  0, 40, 40, 40, 40,
   40, 40, 40,  0, 56, 56, 56, 56, 56, 56, 56,  0, 72, 72, 72, 72,
   72, 72, 72,  0, 88, 88, 88, 88, 88, 88, 88,  0,104,104,104,104,
  104,104,104,  0,120,120,120,120,120,120,120,  0,136,136,136,136,
  136,136,136,  0,152,152,152,152,152,152,152,  0,255,255,255,255,
  255,255,255,  0,184,184,184,184,184,184,184,  0,200,200,200,200,
  200,200,200,  0,216,216,216,216,216,216,216,  0,232,232,232,232,
  232,232,232,  0,248,248,248,248,248,248,248,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  8,  8,  8,  8,  8,  8,
    8,  0, 24, 24, 24, 24, 24, 24, 24,  0, 40, 40, 40, 40, 40, 40,
   40,  0, 56, 56, 56, 56, 56, 56, 56,  0, 72, 72, 72, 72, 72, 72,
   72,  0, 88, 88, 88, 88, 88, 88, 88,  0,104,104,104,104,104,104,
  104,  0,120,120,120,120,120,120,120,  0,136,136,136,136,136,136,
  136,  0,152,152,152,152,152,152,152,  0,255,255,255,255,255,255,
  255,  0,184,184,184,184,184,184,184,  0,200,200,200,200,200,200,
  200,  0,216,216,216,216,216,216,216,  0,232,232,232,232,232,232,
  232,  0,248,248,248,248,248,248,248,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  8,  8,  8,  8,  8,  8,  8,  0,
   24, 24, 24, 24, 24, 24, 24,  0, 40, 40, 40, 40, 40, 40, 40,  0,
   56, 56, 56, 56, 56, 56, 56,  0, 72, 72, 72, 72, 72, 72, 72,  0,
   88, 88, 88, 88, 88, 88, 88,  0,104,104,104,104,104,104,104,  0,
  120,120,120,120,120,120,120,  0,136,136,136,136,136,136,136,  0,
  152,152,152,152,152,152,152,  0,255,255,255,255,255,255,255,  0,
  184,184,184,184,184,184,184,  0,200,200,200,200,200,200,200,  0,
  216,216,216,216,216,216,216,  0,232,232,232,232,232,232,232,  0,
  248,248,248,248,248,248,248,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  8,  8,  8,  8,  8,  8,  8,  0, 24, 24,
   24, 24, 24, 24, 24,  0, 40, 40, 40, 40, 40, 40, 40,  0, 56, 56,
   56, 56, 56, 56, 56,  0, 72, 72, 72, 72, 72, 72, 72,  0, 88, 88,
   88, 88, 88, 88, 88,  0,104,104,104,104,104,104,104,  0,120,120,
  120,120,120,120,120,  0,136,136,136,136,136,136,136,  0,152,152,
  152,152,152,152,152,  0,255,255,255,255,255,255,255,  0,184,184,
  184,184,184,184,184,  0,200,200,200,200,200,200,200,  0,216,216,
  216,216,216,216,216,  0,232,232,232,232,232,232,232,  0,248,248,
  248,248,248,248,248,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  9,  9,  9,  9,  9,  9,  9,  0, 25, 25, 25, 25, 25, 25,
   25,  0, 41, 41, 41, 41, 41, 41, 41,  0, 57, 57, 57, 57, 57, 57,
   57,  0, 73, 73, 73, 73, 73, 73, 73,  0, 89, 89, 89, 89, 89, 89,
   89,  0,105,105,105,105,105,105,105,  0,121,121,121,121,121,121,
  121,  0,137,137,137,137,137,137,137,  0,153,153,153,153,153,153,
  153,  0,169,169,169,169,169,169,169,  0,185,185,185,185,185,185,
  185,  0,201,201,201,201,201,201,201,  0,217,217,217,217,217,217,
  217,  0,233,233,233,233,233,233,233,  0,249,249,249,249,249,249,
  249,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
    9,  9,  9,  9,  9,  9,  9,  0, 25, 25, 25, 25, 25, 25, 25,  0,
   41, 41, 41, 41, 41, 41, 41,  0, 57, 57, 57, 57, 57, 57, 57,  0,
   73, 73, 73, 73, 73, 73, 73,  0, 89, 89, 89, 89, 89, 89, 89,  0,
  105,105,105,105,105,105,105,  0,121,121,121,121,121,121,121,  0,
  137,137,137,137,137,137,137,  0,153,153,153,153,153,153,153,  0,
  169,169,169,169,169,169,169,  0,185,185,185,185,185,185,185,  0,
  201,201,201,201,201,201,201,  0,217,217,217,217,217,217,217,  0,
  233,233,233,233,233,233,233,  0,249,249,249,249,249,249,249,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  9,  9,
    9,  9,  9,  9,  9,  0, 25, 25, 25, 25, 25, 25, 25,  0, 41, 41,
   41, 41, 41, 41, 41,  0, 57, 57, 57, 57, 57, 57, 57,  0, 73, 73,
   73, 73, 73, 73, 73,  0, 89, 89, 89, 89, 89, 89, 89,  0,105,105,
  105,105,105,105,105,  0,121,121,121,121,121,121,121,  0,137,137,
  137,137,137,137,137,  0,153,153,153,153,153,153,153,  0,169,169,
  169,169,169,169,169,  0,185,185,185,185,185,185,185,  0,201,201,
  201,201,201,201,201,  0,217,217,217,217,217,217,217,  0,233,233,
  233,233,233,233,233,  0,249,249,249,249,249,249,249,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,  9,  9,  9,  9,
    9,  9,  9,  0, 25, 25, 25, 25, 25, 25, 25,  0, 41, 41, 41, 41,
   41, 41, 41,  0, 57, 57, 57, 57, 57, 57, 57,  0, 73, 73, 73, 73,
   73, 73, 73,  0, 89, 89, 89, 89, 89, 89, 89,  0,105,105,105,105,
  105,105,105,  0,121,121,121,121,121,121,121,  0,137,137,137,137,
  137,137,137,  0,153,153,153,153,153,153,153,  0,169,169,169,169,
  169,169,169,  0,185,185,185,185,185,185,185,  0,201,201,201,201,
  201,201,201,  0,217,217,217,217,217,217,217,  0,233,233,233,233,
  233,233,233,  0,249,249,249,249,249,249,249,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0,  9,  9,  9,  9,  9,  9,
    9,  0, 25, 25, 25, 25, 25, 25, 25,  0, 41, 41, 41, 41, 41, 41,
   41,  0, 57, 57, 57, 57, 57, 57, 57,  0, 73, 73, 73, 73, 73, 73,
   73,  0, 89, 89, 89, 89, 89, 89, 89,  0,105,105,105,105,105,105,
  105,  0,121,121,121,121,121,121,121,  0,137,137,137,137,137,137,
  137,  0,153,153,153,153,153,153,153,  0,169,169,169,169,169,169,
  169,  0,185,185,185,185,185,185,185,  0,201,201,201,201,201,201,
  201,  0,217,217,217,217,217,217,217,  0,233,233,233,233,233,233,
  233,  0,249,249,249,249,249,249,249,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0,  9,  9,  9,  9,  9,  9,  9,  0,
   25, 25, 25, 25, 25, 25, 25,  0, 41, 41, 41, 41, 41, 41, 41,  0,
   57, 57, 57, 57, 57, 57, 57,  0, 73, 73, 73, 73, 73, 73, 73,  0,
   89, 89, 89, 89, 89, 89, 89,  0,105,105,105,105,105,105,105,  0,
  121,121,121,121,121,121,121,  0,137,137,137,137,137,137,137,  0,
  153,153,153,153,153,153,153,  0,169,169,169,169,169,169,169,  0,
  185,185,185,185,185,185,185,  0,201,201,201,201,201,201,201,  0,
  217,217,217,217,217,217,217,  0,233,233,233,233,233,233,233,  0,
  249,249,249,249,249,249,249,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0,  9,  9,  9,  9,  9,  9,  9,  0, 25, 25,
   25, 25, 25, 25, 25,  0, 41, 41, 41, 41, 41, 41, 41,  0, 57, 57,
   57, 57, 57, 57, 57,  0, 73, 73, 73, 73, 73, 73, 73,  0, 89, 89,
   89, 89, 89, 89, 89,  0,105,105,105,105,105,105,105,  0,121,121,
  121,121,121,121,121,  0,137,137,137,137,137,137,137,  0,153,153,
  153,153,153,153,153,  0,169,169,169,169,169,169,169,  0,185,185,
  185,185,185,185,185,  0,201,201,201,201,201,201,201,  0,217,217,
  217,217,217,217,217,  0,233,233,233,233,233,233,233,  0,249,249,
  249,249,249,249,249,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0, 10, 10, 10, 10, 10, 10, 10,  0, 26, 26, 26, 26, 26, 26,
   26,  0, 42, 42, 42, 42, 42, 42, 42,  0, 58, 58, 58, 58, 58, 58,
   58,  0, 74, 74, 74, 74, 74, 74, 74,  0, 90, 90, 90, 90, 90, 90,
   90,  0,106,106,106,106,106,106,106,  0,122,122,122,122,122,122,
  122,  0,138,138,138,138,138,138,138,  0,154,154,154,154,154,154,
  154,  0,170,170,170,170,170,170,170,  0,186,186,186,186,186,186,
  186,  0,202,202,202,202,202,202,202,  0,218,218,218,218,218,218,
  218,  0,234,234,234,234,234,234,234,  0,250,250,250,250,250,250,
  250,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
   10, 10, 10, 10, 10, 10, 10,  0, 26, 26, 26, 26, 26, 26, 26,  0,
   42, 42, 42, 42, 42, 42, 42,  0, 58, 58, 58, 58, 58, 58, 58,  0,
   74, 74, 74, 74, 74, 74, 74,  0, 90, 90, 90, 90, 90, 90, 90,  0,
  106,106,106,106,106,106,106,  0,122,122,122,122,122,122,122,  0,
  138,138,138,138,138,138,138,  0,154,154,154,154,154,154,154,  0,
  170,170,170,170,170,170,170,  0,186,186,186,186,186,186,186,  0,
  202,202,202,202,202,202,202,  0,218,218,218,218,218,218,218,  0,
  234,234,234,234,234,234,234,  0,250,250,250,250,250,250,250,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 10, 10,
   10, 10, 10, 10, 10,  0, 26, 26, 26, 26, 26, 26, 26,  0, 42, 42,
   42, 42, 42, 42, 42,  0, 58, 58, 58, 58, 58, 58, 58,  0, 74, 74,
   74, 74, 74, 74, 74,  0, 90, 90, 90, 90, 90, 90, 90,  0,106,106,
  106,106,106,106,106,  0,122,122,122,122,122,122,122,  0,138,138,
  138,138,138,138,138,  0,154,154,154,154,154,154,154,  0,170,170,
  170,170,170,170,170,  0,186,186,186,186,186,186,186,  0,202,202,
  202,202,202,202,202,  0,218,218,218,218,218,218,218,  0,234,234,
  234,234,234,234,234,  0,250,250,250,250,250,250,250,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 10, 10, 10, 10,
   10, 10, 10,  0, 26, 26, 26, 26, 26, 26, 26,  0, 42, 42, 42, 42,
   42, 42, 42,  0, 58, 58, 58, 58, 58, 58, 58,  0, 74, 74, 74, 74,
   74, 74, 74,  0, 90, 90, 90, 90, 90, 90, 90,  0,106,106,106,106,
  106,106,106,  0,122,122,122,122,122,122,122,  0,138,138,138,138,
  138,138,138,  0,154,154,154,154,154,154,154,  0,170,170,170,170,
  170,170,170,  0,186,186,186,186,186,186,186,  0,202,202,202,202,
  202,202,202,  0,218,218,218,218,218,218,218,  0,234,234,234,234,
  234,234,234,  0,250,250,250,250,250,250,250,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0, 10, 10, 10, 10, 10, 10,
   10,  0, 26, 26, 26, 26, 26, 26, 26,  0, 42, 42, 42, 42, 42, 42,
   42,  0, 58, 58, 58, 58, 58, 58, 58,  0, 74, 74, 74, 74, 74, 74,
   74,  0, 90, 90, 90, 90, 90, 90, 90,  0,106,106,106,106,106,106,
  106,  0,122,122,122,122,122,122,122,  0,138,138,138,138,138,138,
  138,  0,154,154,154,154,154,154,154,  0,170,170,170,170,170,170,
  170,  0,186,186,186,186,186,186,186,  0,202,202,202,202,202,202,
  202,  0,218,218,218,218,218,218,218,  0,234,234,234,234,234,234,
  234,  0,250,250,250,250,250,250,250,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0, 10, 10, 10, 10, 10, 10, 10,  0,
   26, 26, 26, 26, 26, 26, 26,  0, 42, 42, 42, 42, 42, 42, 42,  0,
   58, 58, 58, 58, 58, 58, 58,  0, 74, 74, 74, 74, 74, 74, 74,  0,
   90, 90, 90, 90, 90, 90, 90,  0,106,106,106,106,106,106,106,  0,
  122,122,122,122,122,122,122,  0,138,138,138,138,138,138,138,  0,
  154,154,154,154,154,154,154,  0,170,170,170,170,170,170,170,  0,
  186,186,186,186,186,186,186,  0,202,202,202,202,202,202,202,  0,
  218,218,218,218,218,218,218,  0,234,234,234,234,234,234,234,  0,
  250,250,250,250,250,250,250,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0, 10, 10, 10, 10, 10, 10, 10,  0, 26, 26,
   26, 26, 26, 26, 26,  0, 42, 42, 42, 42, 42, 42, 42,  0, 58, 58,
   58, 58, 58, 58, 58,  0, 74, 74, 74, 74, 74, 74, 74,  0, 90, 90,
   90, 90, 90, 90, 90,  0,106,106,106,106,106,106,106,  0,122,122,
  122,122,122,122,122,  0,138,138,138,138,138,138,138,  0,154,154,
  154,154,154,154,154,  0,170,170,170,170,170,170,170,  0,186,186,
  186,186,186,186,186,  0,202,202,202,202,202,202,202,  0,218,218,
  218,218,218,218,218,  0,234,234,234,234,234,234,234,  0,250,250,
  250,250,250,250,250,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0, 11, 11, 11, 11, 11, 11, 11,  0, 27, 27, 27, 27, 27, 27,
   27,  0, 43, 43, 43, 43, 43, 43, 43,  0, 59, 59, 59, 59, 59, 59,
   59,  0, 75, 75, 75, 75, 75, 75, 75,  0, 91, 91, 91, 91, 91, 91,
   91,  0,107,107,107,107,107,107,107,  0,123,123,123,123,123,123,
  123,  0,139,139,139,139,139,139,139,  0,155,155,155,155,155,155,
  155,  0,171,171,171,171,171,171,171,  0,187,187,187,187,187,187,
  187,  0,203,203,203,203,203,203,203,  0,219,219,219,219,219,219,
  219,  0,235,235,235,235,235,235,235,  0,251,251,251,251,251,251,
  251,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
   11, 11, 11, 11, 11, 11, 11,  0, 27, 27, 27, 27, 27, 27, 27,  0,
   43, 43, 43, 43, 43, 43, 43,  0, 59, 59, 59, 59, 59, 59, 59,  0,
   75, 75, 75, 75, 75, 75, 75,  0, 91, 91, 91, 91, 91, 91, 91,  0,
  107,107,107,107,107,107,107,  0,123,123,123,123,123,123,123,  0,
  139,139,139,139,139,139,139,  0,155,155,155,155,155,155,155,  0,
  171,171,171,171,171,171,171,  0,187,187,187,187,187,187,187,  0,
  203,203,203,203,203,203,203,  0,219,219,219,219,219,219,219,  0,
  235,235,235,235,235,235,235,  0,251,251,251,251,251,251,251,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 11, 11,
   11, 11, 11, 11, 11,  0, 27, 27, 27, 27, 27, 27, 27,  0, 43, 43,
   43, 43, 43, 43, 43,  0, 59, 59, 59, 59, 59, 59, 59,  0, 75, 75,
   75, 75, 75, 75, 75,  0, 91, 91, 91, 91, 91, 91, 91,  0,107,107,
  107,107,107,107,107,  0,123,123,123,123,123,123,123,  0,139,139,
  139,139,139,139,139,  0,155,155,155,155,155,155,155,  0,171,171,
  171,171,171,171,171,  0,187,187,187,187,187,187,187,  0,203,203,
  203,203,203,203,203,  0,219,219,219,219,219,219,219,  0,235,235,
  235,235,235,235,235,  0,251,251,251,251,251,251,251,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 11, 11, 11, 11,
   11, 11, 11,  0, 27, 27, 27, 27, 27, 27, 27,  0, 43, 43, 43, 43,
   43, 43, 43,  0, 59, 59, 59, 59, 59, 59, 59,  0, 75, 75, 75, 75,
   75, 75, 75,  0, 91, 91, 91, 91, 91, 91, 91,  0,107,107,107,107,
  107,107,107,  0,123,123,123,123,123,123,123,  0,139,139,139,139,
  139,139,139,  0,155,155,155,155,155,155,155,  0,171,171,171,171,
  171,171,171,  0,187,187,187,187,187,187,187,  0,203,203,203,203,
  203,203,203,  0,219,219,219,219,219,219,219,  0,235,235,235,235,
  235,235,235,  0,251,251,251,251,251,251,251,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0, 11, 11, 11, 11, 11, 11,
   11,  0, 27, 27, 27, 27, 27, 27, 27,  0, 43, 43, 43, 43, 43, 43,
   43,  0, 59, 59, 59, 59, 59, 59, 59,  0, 75, 75, 75, 75, 75, 75,
   75,  0, 91, 91, 91, 91, 91, 91, 91,  0,107,107,107,107,107,107,
  107,  0,123,123,123,123,123,123,123,  0,139,139,139,139,139,139,
  139,  0,155,155,155,155,155,155,155,  0,171,171,171,171,171,171,
  171,  0,187,187,187,187,187,187,187,  0,203,203,203,203,203,203,
  203,  0,219,219,219,219,219,219,219,  0,235,235,235,235,235,235,
  235,  0,251,251,251,251,251,251,251,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0, 11, 11, 11, 11, 11, 11, 11,  0,
   27, 27, 27, 27, 27, 27, 27,  0, 43, 43, 43, 43, 43, 43, 43,  0,
   59, 59, 59, 59, 59, 59, 59,  0, 75, 75, 75, 75, 75, 75, 75,  0,
   91, 91, 91, 91, 91, 91, 91,  0,107,107,107,107,107,107,107,  0,
  123,123,123,123,123,123,123,  0,139,139,139,139,139,139,139,  0,
  155,155,155,155,155,155,155,  0,171,171,171,171,171,171,171,  0,
  187,187,187,187,187,187,187,  0,203,203,203,203,203,203,203,  0,
  219,219,219,219,219,219,219,  0,235,235,235,235,235,235,235,  0,
  251,251,251,251,251,251,251,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0, 11, 11, 11, 11, 11, 11, 11,  0, 27, 27,
   27, 27, 27, 27, 27,  0, 43, 43, 43, 43, 43, 43, 43,  0, 59, 59,
   59, 59, 59, 59, 59,  0, 75, 75, 75, 75, 75, 75, 75,  0, 91, 91,

   91, 91, 91, 91, 91,  0,107,107,107,107,107,107,107,  0,123,123,
  123,123,123,123,123,  0,139,139,139,139,139,139,139,  0,155,155,
  155,155,155,155,155,  0,171,171,171,171,171,171,171,  0,187,187,
  187,187,187,187,187,  0,203,203,203,203,203,203,203,  0,219,219,
  219,219,219,219,219,  0,235,235,235,235,235,235,235,  0,251,251,
  251,251,251,251,251,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0, 12, 12, 12, 12, 12, 12, 12,  0, 28, 28, 28, 28, 28, 28,
   28,  0, 44, 44, 44, 44, 44, 44, 44,  0, 60, 60, 60, 60, 60, 60,
   60,  0, 76, 76, 76, 76, 76, 76, 76,  0, 92, 92, 92, 92, 92, 92,
   92,  0,108,108,108,108,108,108,108,  0,124,124,124,124,124,124,
  124,  0,140,140,140,140,140,140,140,  0,156,156,156,156,156,156,
  156,  0,172,172,172,172,172,172,172,  0,188,188,188,188,188,188,
  188,  0,204,204,204,204,204,204,204,  0,220,220,220,220,220,220,
  220,  0,236,236,236,236,236,236,236,  0,252,252,252,252,252,252,
  252,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
   12, 12, 12, 12, 12, 12, 12,  0, 28, 28, 28, 28, 28, 28, 28,  0,
   44, 44, 44, 44, 44, 44, 44,  0, 60, 60, 60, 60, 60, 60, 60,  0,
   76, 76, 76, 76, 76, 76, 76,  0, 92, 92, 92, 92, 92, 92, 92,  0,
  108,108,108,108,108,108,108,  0,124,124,124,124,124,124,124,  0,
  140,140,140,140,140,140,140,  0,156,156,156,156,156,156,156,  0,
  172,172,172,172,172,172,172,  0,188,188,188,188,188,188,188,  0,
  204,204,204,204,204,204,204,  0,220,220,220,220,220,220,220,  0,
  236,236,236,236,236,236,236,  0,252,252,252,252,252,252,252,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 12, 12,
   12, 12, 12, 12, 12,  0, 28, 28, 28, 28, 28, 28, 28,  0, 44, 44,
   44, 44, 44, 44, 44,  0, 60, 60, 60, 60, 60, 60, 60,  0, 76, 76,
   76, 76, 76, 76, 76,  0, 92, 92, 92, 92, 92, 92, 92,  0,108,108,
  108,108,108,108,108,  0,124,124,124,124,124,124,124,  0,140,140,
  140,140,140,140,140,  0,156,156,156,156,156,156,156,  0,172,172,
  172,172,172,172,172,  0,188,188,188,188,188,188,188,  0,204,204,
  204,204,204,204,204,  0,220,220,220,220,220,220,220,  0,236,236,
  236,236,236,236,236,  0,252,252,252,252,252,252,252,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 12, 12, 12, 12,
   12, 12, 12,  0, 28, 28, 28, 28, 28, 28, 28,  0, 44, 44, 44, 44,
   44, 44, 44,  0, 60, 60, 60, 60, 60, 60, 60,  0, 76, 76, 76, 76,
   76, 76, 76,  0, 92, 92, 92, 92, 92, 92, 92,  0,108,108,108,108,
  108,108,108,  0,124,124,124,124,124,124,124,  0,140,140,140,140,
  140,140,140,  0,156,156,156,156,156,156,156,  0,172,172,172,172,
  172,172,172,  0,188,188,188,188,188,188,188,  0,204,204,204,204,
  204,204,204,  0,220,220,220,220,220,220,220,  0,236,236,236,236,
  236,236,236,  0,252,252,252,252,252,252,252,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0, 12, 12, 12, 12, 12, 12,
   12,  0, 28, 28, 28, 28, 28, 28, 28,  0, 44, 44, 44, 44, 44, 44,
   44,  0, 60, 60, 60, 60, 60, 60, 60,  0, 76, 76, 76, 76, 76, 76,
   76,  0, 92, 92, 92, 92, 92, 92, 92,  0,108,108,108,108,108,108,
  108,  0,124,124,124,124,124,124,124,  0,140,140,140,140,140,140,
  140,  0,156,156,156,156,156,156,156,  0,172,172,172,172,172,172,
  172,  0,188,188,188,188,188,188,188,  0,204,204,204,204,204,204,
  204,  0,220,220,220,220,220,220,220,  0,236,236,236,236,236,236,
  236,  0,252,252,252,252,252,252,252,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0, 12, 12, 12, 12, 12, 12, 12,  0,
   28, 28, 28, 28, 28, 28, 28,  0, 44, 44, 44, 44, 44, 44, 44,  0,
   60, 60, 60, 60, 60, 60, 60,  0, 76, 76, 76, 76, 76, 76, 76,  0,
   92, 92, 92, 92, 92, 92, 92,  0,108,108,108,108,108,108,108,  0,
  124,124,124,124,124,124,124,  0,140,140,140,140,140,140,140,  0,
  156,156,156,156,156,156,156,  0,172,172,172,172,172,172,172,  0,
  188,188,188,188,188,188,188,  0,204,204,204,204,204,204,204,  0,
  220,220,220,220,220,220,220,  0,236,236,236,236,236,236,236,  0,
  252,252,252,252,252,252,252,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0, 12, 12, 12, 12, 12, 12, 12,  0, 28, 28,
   28, 28, 28, 28, 28,  0, 44, 44, 44, 44, 44, 44, 44,  0, 60, 60,
   60, 60, 60, 60, 60,  0, 76, 76, 76, 76, 76, 76, 76,  0, 92, 92,
   92, 92, 92, 92, 92,  0,108,108,108,108,108,108,108,  0,124,124,
  124,124,124,124,124,  0,140,140,140,140,140,140,140,  0,156,156,
  156,156,156,156,156,  0,172,172,172,172,172,172,172,  0,188,188,
  188,188,188,188,188,  0,204,204,204,204,204,204,204,  0,220,220,
  220,220,220,220,220,  0,236,236,236,236,236,236,236,  0,252,252,
  252,252,252,252,252,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0, 13, 13, 13, 13, 13, 13, 13,  0, 29, 29, 29, 29, 29, 29,
   29,  0, 45, 45, 45, 45, 45, 45, 45,  0, 61, 61, 61, 61, 61, 61,
   61,  0, 77, 77, 77, 77, 77, 77, 77,  0, 93, 93, 93, 93, 93, 93,
   93,  0,109,109,109,109,109,109,109,  0,125,125,125,125,125,125,
  125,  0,141,141,141,141,141,141,141,  0,157,157,157,157,157,157,
  157,  0,173,173,173,173,173,173,173,  0,189,189,189,189,189,189,
  189,  0,205,205,205,205,205,205,205,  0,221,221,221,221,221,221,
  221,  0,237,237,237,237,237,237,237,  0,253,253,253,253,253,253,
  253,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
   13, 13, 13, 13, 13, 13, 13,  0, 29, 29, 29, 29, 29, 29, 29,  0,
   45, 45, 45, 45, 45, 45, 45,  0, 61, 61, 61, 61, 61, 61, 61,  0,
   77, 77, 77, 77, 77, 77, 77,  0, 93, 93, 93, 93, 93, 93, 93,  0,
  109,109,109,109,109,109,109,  0,125,125,125,125,125,125,125,  0,
  141,141,141,141,141,141,141,  0,157,157,157,157,157,157,157,  0,
  173,173,173,173,173,173,173,  0,189,189,189,189,189,189,189,  0,
  205,205,205,205,205,205,205,  0,221,221,221,221,221,221,221,  0,
  237,237,237,237,237,237,237,  0,253,253,253,253,253,253,253,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 13, 13,
   13, 13, 13, 13, 13,  0, 29, 29, 29, 29, 29, 29, 29,  0, 45, 45,
   45, 45, 45, 45, 45,  0, 61, 61, 61, 61, 61, 61, 61,  0, 77, 77,
   77, 77, 77, 77, 77,  0, 93, 93, 93, 93, 93, 93, 93,  0,109,109,
  109,109,109,109,109,  0,125,125,125,125,125,125,125,  0,141,141,
  141,141,141,141,141,  0,157,157,157,157,157,157,157,  0,173,173,
  173,173,173,173,173,  0,189,189,189,189,189,189,189,  0,205,205,
  205,205,205,205,205,  0,221,221,221,221,221,221,221,  0,237,237,
  237,237,237,237,237,  0,253,253,253,253,253,253,253,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 13, 13, 13, 13,
   13, 13, 13,  0, 29, 29, 29, 29, 29, 29, 29,  0, 45, 45, 45, 45,
   45, 45, 45,  0, 61, 61, 61, 61, 61, 61, 61,  0, 77, 77, 77, 77,
   77, 77, 77,  0, 93, 93, 93, 93, 93, 93, 93,  0,109,109,109,109,
  109,109,109,  0,125,125,125,125,125,125,125,  0,141,141,141,141,
  141,141,141,  0,157,157,157,157,157,157,157,  0,173,173,173,173,
  173,173,173,  0,189,189,189,189,189,189,189,  0,205,205,205,205,
  205,205,205,  0,221,221,221,221,221,221,221,  0,237,237,237,237,
  237,237,237,  0,253,253,253,253,253,253,253,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0, 13, 13, 13, 13, 13, 13,
   13,  0, 29, 29, 29, 29, 29, 29, 29,  0, 45, 45, 45, 45, 45, 45,
   45,  0, 61, 61, 61, 61, 61, 61, 61,  0, 77, 77, 77, 77, 77, 77,
   77,  0, 93, 93, 93, 93, 93, 93, 93,  0,109,109,109,109,109,109,
  109,  0,125,125,125,125,125,125,125,  0,141,141,141,141,141,141,
  141,  0,157,157,157,157,157,157,157,  0,173,173,173,173,173,173,
  173,  0,189,189,189,189,189,189,189,  0,205,205,205,205,205,205,
  205,  0,221,221,221,221,221,221,221,  0,237,237,237,237,237,237,
  237,  0,253,253,253,253,253,253,253,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0, 13, 13, 13, 13, 13, 13, 13,  0,
   29, 29, 29, 29, 29, 29, 29,  0, 45, 45, 45, 45, 45, 45, 45,  0,
   61, 61, 61, 61, 61, 61, 61,  0, 77, 77, 77, 77, 77, 77, 77,  0,
   93, 93, 93, 93, 93, 93, 93,  0,109,109,109,109,109,109,109,  0,
  125,125,125,125,125,125,125,  0,141,141,141,141,141,141,141,  0,
  157,157,157,157,157,157,157,  0,173,173,173,173,173,173,173,  0,
  189,189,189,189,189,189,189,  0,205,205,205,205,205,205,205,  0,
  221,221,221,221,221,221,221,  0,237,237,237,237,237,237,237,  0,
  253,253,253,253,253,253,253,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0, 13, 13, 13, 13, 13, 13, 13,  0, 29, 29,
   29, 29, 29, 29, 29,  0, 45, 45, 45, 45, 45, 45, 45,  0, 61, 61,
   61, 61, 61, 61, 61,  0, 77, 77, 77, 77, 77, 77, 77,  0, 93, 93,
   93, 93, 93, 93, 93,  0,109,109,109,109,109,109,109,  0,125,125,
  125,125,125,125,125,  0,141,141,141,141,141,141,141,  0,157,157,
  157,157,157,157,157,  0,173,173,173,173,173,173,173,  0,189,189,
  189,189,189,189,189,  0,205,205,205,205,205,205,205,  0,221,221,
  221,221,221,221,221,  0,237,237,237,237,237,237,237,  0,253,253,
  253,253,253,253,253,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0, 14, 14, 14, 14, 14, 14, 14,  0, 30, 30, 30, 30, 30, 30,
   30,  0, 46, 46, 46, 46, 46, 46, 46,  0, 62, 62, 62, 62, 62, 62,
   62,  0, 78, 78, 78, 78, 78, 78, 78,  0, 94, 94, 94, 94, 94, 94,
   94,  0,110,110,110,110,110,110,110,  0,126,126,126,126,126,126,
  126,  0,142,142,142,142,142,142,142,  0,158,158,158,158,158,158,
  158,  0,174,174,174,174,174,174,174,  0, 45, 45, 45, 45, 45, 45,
   45,  0,206,206,206,206,206,206,206,  0,222,222,222,222,222,222,
  222,  0,238,238,238,238,238,238,238,  0,254,254,254,254,254,254,
  254,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
   14, 14, 14, 14, 14, 14, 14,  0, 30, 30, 30, 30, 30, 30, 30,  0,
   46, 46, 46, 46, 46, 46, 46,  0, 62, 62, 62, 62, 62, 62, 62,  0,
   78, 78, 78, 78, 78, 78, 78,  0, 94, 94, 94, 94, 94, 94, 94,  0,
  110,110,110,110,110,110,110,  0,126,126,126,126,126,126,126,  0,
  142,142,142,142,142,142,142,  0,158,158,158,158,158,158,158,  0,
  174,174,174,174,174,174,174,  0, 45, 45, 45, 45, 45, 45, 45,  0,
  206,206,206,206,206,206,206,  0,222,222,222,222,222,222,222,  0,
  238,238,238,238,238,238,238,  0,254,254,254,254,254,254,254,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 14, 14,
   14, 14, 14, 14, 14,  0, 30, 30, 30, 30, 30, 30, 30,  0, 46, 46,
   46, 46, 46, 46, 46,  0, 62, 62, 62, 62, 62, 62, 62,  0, 78, 78,
   78, 78, 78, 78, 78,  0, 94, 94, 94, 94, 94, 94, 94,  0,110,110,
  110,110,110,110,110,  0,126,126,126,126,126,126,126,  0,142,142,
  142,142,142,142,142,  0,158,158,158,158,158,158,158,  0,174,174,
  174,174,174,174,174,  0, 45, 45, 45, 45, 45, 45, 45,  0,206,206,
  206,206,206,206,206,  0,222,222,222,222,222,222,222,  0,238,238,
  238,238,238,238,238,  0,254,254,254,254,254,254,254,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 14, 14, 14, 14,
   14, 14, 14,  0, 30, 30, 30, 30, 30, 30, 30,  0, 46, 46, 46, 46,
   46, 46, 46,  0, 62, 62, 62, 62, 62, 62, 62,  0, 78, 78, 78, 78,
   78, 78, 78,  0, 94, 94, 94, 94, 94, 94, 94,  0,110,110,110,110,
  110,110,110,  0,126,126,126,126,126,126,126,  0,142,142,142,142,
  142,142,142,  0,158,158,158,158,158,158,158,  0,174,174,174,174,
  174,174,174,  0, 45, 45, 45, 45, 45, 45, 45,  0,206,206,206,206,
  206,206,206,  0,222,222,222,222,222,222,222,  0,238,238,238,238,
  238,238,238,  0,254,254,254,254,254,254,254,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0, 14, 14, 14, 14, 14, 14,
   14,  0, 30, 30, 30, 30, 30, 30, 30,  0, 46, 46, 46, 46, 46, 46,
   46,  0, 62, 62, 62, 62, 62, 62, 62,  0, 78, 78, 78, 78, 78, 78,
   78,  0, 94, 94, 94, 94, 94, 94, 94,  0,110,110,110,110,110,110,
  110,  0,126,126,126,126,126,126,126,  0,142,142,142,142,142,142,
  142,  0,158,158,158,158,158,158,158,  0,174,174,174,174,174,174,
  174,  0, 45, 45, 45, 45, 45, 45, 45,  0,206,206,206,206,206,206,
  206,  0,222,222,222,222,222,222,222,  0,238,238,238,238,238,238,
  238,  0,254,254,254,254,254,254,254,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0, 14, 14, 14, 14, 14, 14, 14,  0,
   30, 30, 30, 30, 30, 30, 30,  0, 46, 46, 46, 46, 46, 46, 46,  0,
   62, 62, 62, 62, 62, 62, 62,  0, 78, 78, 78, 78, 78, 78, 78,  0,
   94, 94, 94, 94, 94, 94, 94,  0,110,110,110,110,110,110,110,  0,
  126,126,126,126,126,126,126,  0,142,142,142,142,142,142,142,  0,
  158,158,158,158,158,158,158,  0,174,174,174,174,174,174,174,  0,
   45, 45, 45, 45, 45, 45, 45,  0,206,206,206,206,206,206,206,  0,
  222,222,222,222,222,222,222,  0,238,238,238,238,238,238,238,  0,
  254,254,254,254,254,254,254,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0, 14, 14, 14, 14, 14, 14, 14,  0, 30, 30,
   30, 30, 30, 30, 30,  0, 46, 46, 46, 46, 46, 46, 46,  0, 62, 62,
   62, 62, 62, 62, 62,  0, 78, 78, 78, 78, 78, 78, 78,  0, 94, 94,
   94, 94, 94, 94, 94,  0,110,110,110,110,110,110,110,  0,126,126,
  126,126,126,126,126,  0,142,142,142,142,142,142,142,  0,158,158,
  158,158,158,158,158,  0,174,174,174,174,174,174,174,  0, 45, 45,
   45, 45, 45, 45, 45,  0,206,206,206,206,206,206,206,  0,222,222,
  222,222,222,222,222,  0,238,238,238,238,238,238,238,  0,254,254,
  254,254,254,254,254,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0, 15, 15, 15, 15, 15, 15, 15,  0, 31, 31, 31, 31, 31, 31,
   31,  0, 47, 47, 47, 47, 47, 47, 47,  0, 63, 63, 63, 63, 63, 63,
   63,  0, 79, 79, 79, 79, 79, 79, 79,  0, 95, 95, 95, 95, 95, 95,
   95,  0,111,111,111,111,111,111,111,  0,127,127,127,127,127,127,
  127,  0,143,143,143,143,143,143,143,  0,159,159,159,159,159,159,
  159,  0,175,175,175,175,175,175,175,  0, 47, 47, 47, 47, 47, 47,
   47,  0,207,207,207,207,207,207,207,  0,223,223,223,223,223,223,
  223,  0,239,239,239,239,239,239,239,  0,255,255,255,255,255,255,
  255,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0,
   15, 15, 15, 15, 15, 15, 15,  0, 31, 31, 31, 31, 31, 31, 31,  0,
   47, 47, 47, 47, 47, 47, 47,  0, 63, 63, 63, 63, 63, 63, 63,  0,
   79, 79, 79, 79, 79, 79, 79,  0, 95, 95, 95, 95, 95, 95, 95,  0,
  111,111,111,111,111,111,111,  0,127,127,127,127,127,127,127,  0,
  143,143,143,143,143,143,143,  0,159,159,159,159,159,159,159,  0,
  175,175,175,175,175,175,175,  0, 47, 47, 47, 47, 47, 47, 47,  0,
  207,207,207,207,207,207,207,  0,223,223,223,223,223,223,223,  0,
  239,239,239,239,239,239,239,  0,255,255,255,255,255,255,255,  0,
    0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 15, 15,
   15, 15, 15, 15, 15,  0, 31, 31, 31, 31, 31, 31, 31,  0, 47, 47,
   47, 47, 47, 47, 47,  0, 63, 63, 63, 63, 63, 63, 63,  0, 79, 79,
   79, 79, 79, 79, 79,  0, 95, 95, 95, 95, 95, 95, 95,  0,111,111,
  111,111,111,111,111,  0,127,127,127,127,127,127,127,  0,143,143,
  143,143,143,143,143,  0,159,159,159,159,159,159,159,  0,175,175,
  175,175,175,175,175,  0, 47, 47, 47, 47, 47, 47, 47,  0,207,207,
  207,207,207,207,207,  0,223,223,223,223,223,223,223,  0,239,239,
  239,239,239,239,239,  0,255,255,255,255,255,255,255,  0,  0, 91,
   98,105,  0,255,  0,137,  0, 91, 98,105,  0,  0, 15, 15, 15, 15,
   15, 15, 15,  0, 31, 31, 31, 31, 31, 31, 31,  0, 47, 47, 47, 47,
   47, 47, 47,  0, 63, 63, 63, 63, 63, 63, 63,  0, 79, 79, 79, 79,
   79, 79, 79,  0, 95, 95, 95, 95, 95, 95, 95,  0,111,111,111,111,
  111,111,111,  0,127,127,127,127,127,127,127,  0,143,143,143,143,
  143,143,143,  0,159,159,159,159,159,159,159,  0,175,175,175,175,
  175,175,175,  0, 47, 47, 47, 47, 47, 47, 47,  0,207,207,207,207,
  207,207,207,  0,223,223,223,223,223,223,223,  0,239,239,239,239,
  239,239,239,  0,255,255,255,255,255,255,255,  0,  0, 91, 98,105,
    0,255,  0,137,  0, 91, 98,105,  0,  0, 15, 15, 15, 15, 15, 15,
   15,  0, 31, 31, 31, 31, 31, 31, 31,  0, 47, 47, 47, 47, 47, 47,
   47,  0, 63, 63, 63, 63, 63, 63, 63,  0, 79, 79, 79, 79, 79, 79,
   79,  0, 95, 95, 95, 95, 95, 95, 95,  0,111,111,111,111,111,111,
  111,  0,127,127,127,127,127,127,127,  0,143,143,143,143,143,143,
  143,  0,159,159,159,159,159,159,159,  0,175,175,175,175,175,175,
  175,  0, 47, 47, 47, 47, 47, 47, 47,  0,207,207,207,207,207,207,
  207,  0,223,223,223,223,223,223,223,  0,239,239,239,239,239,239,
  239,  0,255,255,255,255,255,255,255,  0,  0, 91, 98,105,  0,255,
    0,137,  0, 91, 98,105,  0,  0, 15, 15, 15, 15, 15, 15, 15,  0,
   31, 31, 31, 31, 31, 31, 31,  0, 47, 47, 47, 47, 47, 47, 47,  0,
   63, 63, 63, 63, 63, 63, 63,  0, 79, 79, 79, 79, 79, 79, 79,  0,
   95, 95, 95, 95, 95, 95, 95,  0,111,111,111,111,111,111,111,  0,
  127,127,127,127,127,127,127,  0,143,143,143,143,143,143,143,  0,
  159,159,159,159,159,159,159,  0,175,175,175,175,175,175,175,  0,
   47, 47, 47, 47, 47, 47, 47,  0,207,207,207,207,207,207,207,  0,
  223,223,223,223,223,223,223,  0,239,239,239,239,239,239,239,  0,
  255,255,255,255,255,255,255,  0,  0, 91, 98,105,  0,255,  0,137,
    0, 91, 98,105,  0,  0, 15, 15, 15, 15, 15, 15, 15,  0, 31, 31,
   31, 31, 31, 31, 31,  0, 47, 47, 47, 47, 47, 47, 47,  0, 63, 63,
   63, 63, 63, 63, 63,  0, 79, 79, 79, 79, 79, 79, 79,  0, 95, 95,
   95, 95, 95, 95, 95,  0,111,111,111,111,111,111,111,  0,127,127,
  127,127,127,127,127,  0,143,143,143,143,143,143,143,  0,159,159,
  159,159,159,159,159,  0,175,175,175,175,175,175,175,  0, 47, 47,
   47, 47, 47, 47, 47,  0,207,207,207,207,207,207,207,  0,223,223,
  223,223,223,223,223,  0,239,239,239,239,239,239,239,  0,255,255,
  255,255,255,255,255,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91,
   98,105,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98,105,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0, 91, 98,105,  0,255,  0,137,  0, 91, 98, 94, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91, 91,
   91, 91, 98,105,  0,255,  0,137,  0, 91, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98, 98,
   98,105,  0,255,  0,137,  0, 98,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
  105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,105,
    0,255,
};

static const char m_palno[]=  // Marker to not show this color in automap
{
    7,  0,  7,  0,  0,  0,  0,  0,
   36,  0,  0,  0, 48,  0,  0,  0, 60,  0,  0,  0, 72,  0,  0,  0,
   84,  0,  0,  0, 96,  0,  0,  0,108,  0,  0,  0,  0,  7,  0,  4,
    0,  0,  0,  0,  0,  4,  0,255,  0,  7,  0,  0,  4,  0,  0,  0,
    4,  0,  0,255,  0,  7,  0,  0,  0,  4,  0,  4,  0,  0,  0,255,
    0,  7,  0,  0,  0,  0,  4,  0,  0,  0,  0,255,  0,  7,  0,  0,
    0,  4,  0,  4,  0,  0,  0,255,  0,  7,  0,  0,  4,  0,  0,  0,
    4,  0,  0,255,  0,  7,  0,  4,  0,  0,  0,  0,  0,  4,  0,255,
};

static const char stcfn096[]= // backquote character, never existed // 2023/12/03 version made with Harmony font
{
  3,0,3,0,0,0,0,0,20,0,0,0,28,0,0,0,36,0,0,0,0,3,159,159,159,159,159,255,0,3,
  246,246,243,159,159,255,0,2,246,246,159,159,255,
};

static const char m_butt1[]=  // Options reset button unlit  // 2023/12/02 version from HarmonyC.wad
{
  15,0,15,0,0,0,0,0,68,0,0,0,88,0,0,0,108,0,0,0,128,0,0,0,148,0,0,0,168,0,0,0,
  188,0,0,0,208,0,0,0,228,0,0,0,248,0,0,0,12,1,0,0,32,1,0,0,52,1,0,0,72,1,0,0,
  92,1,0,0,0,15,0,155,155,155,155,155,155,155,155,155,155,155,155,155,155,154,
  0,255,0,15,0,155,158,158,158,158,158,158,158,111,111,158,111,111,154,154,0,
  255,0,15,0,155,158,111,111,111,111,111,157,111,111,157,157,111,151,154,0,255,
  0,15,0,155,158,111,111,111,111,111,157,111,157,111,157,111,151,154,0,255,0,
  15,0,155,158,111,157,111,157,157,157,111,157,111,111,111,151,154,0,255,0,15,
  0,155,158,111,111,111,111,111,157,157,157,157,157,157,151,154,0,255,0,15,0,
  155,158,111,111,157,111,111,157,111,111,111,111,111,151,154,0,255,0,15,0,155,
  158,157,157,157,157,157,157,111,111,111,111,111,151,154,0,255,0,15,0,155,158,
  111,111,111,111,111,157,111,157,111,157,111,151,154,0,255,0,15,0,154,158,111,
  111,111,111,111,157,157,157,157,157,111,152,155,0,255,0,15,0,154,158,111,157,
  111,157,111,157,111,157,157,157,157,152,155,0,255,0,15,0,154,158,111,157,157,
  157,111,157,111,111,111,111,111,152,155,0,255,0,15,0,154,158,157,157,157,157,
  111,157,111,158,111,111,111,152,155,0,255,0,15,0,155,158,158,158,158,158,158,
  158,111,158,158,158,158,154,155,0,255,0,15,0,155,155,155,155,155,155,155,155,
  155,155,155,155,155,155,155,0,255,
};


static const char m_butt2[]=  // Options reset button lit  // 2023/12/02 version from HarmonyC.wad
{
  15,0,15,0,0,0,0,0,68,0,0,0,88,0,0,0,108,0,0,0,128,0,0,0,148,0,0,0,168,0,0,0,
  188,0,0,0,208,0,0,0,228,0,0,0,248,0,0,0,12,1,0,0,32,1,0,0,52,1,0,0,72,1,0,0,
  92,1,0,0,0,15,0,155,155,155,155,155,155,155,155,155,155,155,155,155,155,154,
  0,255,0,15,0,155,158,158,158,158,158,158,158,9,9,158,9,9,154,154,0,255,0,15,
  0,155,158,9,9,9,9,9,157,9,9,157,157,9,151,154,0,255,0,15,0,155,158,9,9,9,9,9,
  157,9,157,9,157,9,151,154,0,255,0,15,0,155,158,9,157,9,157,157,157,9,157,9,9,
  9,151,154,0,255,0,15,0,155,158,9,9,9,9,9,157,157,157,157,157,157,151,154,0,
  255,0,15,0,155,158,9,9,157,9,9,157,9,9,9,9,9,151,154,0,255,0,15,0,155,158,
  157,157,157,157,157,157,9,9,9,9,9,151,154,0,255,0,15,0,155,158,9,9,9,9,9,157,
  9,157,9,157,9,151,154,0,255,0,15,0,154,158,9,9,9,9,9,157,157,157,157,157,9,
  152,155,0,255,0,15,0,154,158,9,157,9,157,9,157,9,157,157,157,157,152,155,0,
  255,0,15,0,154,158,9,157,157,157,9,157,9,9,9,9,9,152,155,0,255,0,15,0,154,
  158,157,157,157,157,9,157,9,158,9,9,9,152,155,0,255,0,15,0,155,158,158,158,
  158,158,158,158,9,158,158,158,158,154,155,0,255,0,15,0,155,155,155,155,155,
  155,155,155,155,155,155,155,155,155,155,0,255,
};

// Animated textures definition lump -- see SWANTBLS.EXE docs
static const char animated[]=
{
    0, 78, 85, 75, 65, 71, 69, 51,  0,  0, 78, 85, 75, 65, 71, 69,
   49,  0,  0,  8,  0,  0,  0,  0, 70, 87, 65, 84, 69, 82, 52,  0,
    0, 70, 87, 65, 84, 69, 82, 49,  0,  0,  8,  0,  0,  0,  0, 83,
   87, 65, 84, 69, 82, 52,  0,  0, 83, 87, 65, 84, 69, 82, 49,  0,
    0,  8,  0,  0,  0,  0, 76, 65, 86, 65, 52,  0,  0,  0,  0, 76,
   65, 86, 65, 49,  0,  0,  0,  0,  8,  0,  0,  0,  0, 66, 76, 79,
   79, 68, 51,  0,  0,  0, 66, 76, 79, 79, 68, 49,  0,  0,  0,  8,
    0,  0,  0,  0, 82, 82, 79, 67, 75, 48, 56,  0,  0, 82, 82, 79,
   67, 75, 48, 53,  0,  0,  8,  0,  0,  0,  0, 83, 76, 73, 77, 69,
   48, 52,  0,  0, 83, 76, 73, 77, 69, 48, 49,  0,  0,  8,  0,  0,
    0,  0, 83, 76, 73, 77, 69, 48, 56,  0,  0, 83, 76, 73, 77, 69,
   48, 53,  0,  0,  8,  0,  0,  0,  0, 83, 76, 73, 77, 69, 49, 50,
    0,  0, 83, 76, 73, 77, 69, 48, 57,  0,  0,  8,  0,  0,  0,  1,
   66, 76, 79, 68, 71, 82, 52,  0,  0, 66, 76, 79, 68, 71, 82, 49,
    0,  0,  8,  0,  0,  0,  1, 83, 76, 65, 68, 82, 73, 80, 51,  0,
   83, 76, 65, 68, 82, 73, 80, 49,  0,  8,  0,  0,  0,  1, 66, 76,
   79, 68, 82, 73, 80, 52,  0, 66, 76, 79, 68, 82, 73, 80, 49,  0,
    8,  0,  0,  0,  1, 70, 73, 82, 69, 87, 65, 76, 76,  0, 70, 73,
   82, 69, 87, 65, 76, 65,  0,  8,  0,  0,  0,  1, 71, 83, 84, 70,
   79, 78, 84, 51,  0, 71, 83, 84, 70, 79, 78, 84, 49,  0,  8,  0,
    0,  0,  1, 70, 73, 82, 69, 76, 65, 86, 65,  0, 70, 73, 82, 69,
   76, 65, 86, 51,  0,  8,  0,  0,  0,  1, 70, 73, 82, 69, 77, 65,
   71, 51,  0, 70, 73, 82, 69, 77, 65, 71, 49,  0,  8,  0,  0,  0,
    1, 70, 73, 82, 69, 66, 76, 85, 50,  0, 70, 73, 82, 69, 66, 76,
   85, 49,  0,  8,  0,  0,  0,  1, 82, 79, 67, 75, 82, 69, 68, 51,
    0, 82, 79, 67, 75, 82, 69, 68, 49,  0,  8,  0,  0,  0,  1, 66,
   70, 65, 76, 76, 52,  0,  0,  0, 66, 70, 65, 76, 76, 49,  0,  0,
    0,  8,  0,  0,  0,  1, 83, 70, 65, 76, 76, 52,  0,  0,  0, 83,
   70, 65, 76, 76, 49,  0,  0,  0,  8,  0,  0,  0,  1, 87, 70, 65,
   76, 76, 52,  0,  0,  0, 87, 70, 65, 76, 76, 49,  0,  0,  0,  8,
    0,  0,  0,  1, 68, 66, 82, 65, 73, 78, 52,  0,  0, 68, 66, 82,
   65, 73, 78, 49,  0,  0,  8,  0,  0,  0,255,255,255,255,
};

// Switch textures definition lump -- see SWANTBLS.EXE docs
static const char switches[]=
{
   83, 87, 49, 66, 82, 67, 79, 77,  0, 83, 87, 50, 66, 82, 67, 79,
   77,  0,  1,  0, 83, 87, 49, 66, 82, 78, 49,  0,  0, 83, 87, 50,
   66, 82, 78, 49,  0,  0,  1,  0, 83, 87, 49, 66, 82, 78, 50,  0,
    0, 83, 87, 50, 66, 82, 78, 50,  0,  0,  1,  0, 83, 87, 49, 66,
   82, 78, 71, 78,  0, 83, 87, 50, 66, 82, 78, 71, 78,  0,  1,  0,
   83, 87, 49, 66, 82, 79, 87, 78,  0, 83, 87, 50, 66, 82, 79, 87,
   78,  0,  1,  0, 83, 87, 49, 67, 79, 77, 77,  0,  0, 83, 87, 50,
   67, 79, 77, 77,  0,  0,  1,  0, 83, 87, 49, 67, 79, 77, 80,  0,
    0, 83, 87, 50, 67, 79, 77, 80,  0,  0,  1,  0, 83, 87, 49, 68,
   73, 82, 84,  0,  0, 83, 87, 50, 68, 73, 82, 84,  0,  0,  1,  0,
   83, 87, 49, 69, 88, 73, 84,  0,  0, 83, 87, 50, 69, 88, 73, 84,
    0,  0,  1,  0, 83, 87, 49, 71, 82, 65, 89,  0,  0, 83, 87, 50,
   71, 82, 65, 89,  0,  0,  1,  0, 83, 87, 49, 71, 82, 65, 89, 49,
    0, 83, 87, 50, 71, 82, 65, 89, 49,  0,  1,  0, 83, 87, 49, 77,
   69, 84, 65, 76,  0, 83, 87, 50, 77, 69, 84, 65, 76,  0,  1,  0,
   83, 87, 49, 80, 73, 80, 69,  0,  0, 83, 87, 50, 80, 73, 80, 69,
    0,  0,  1,  0, 83, 87, 49, 83, 76, 65, 68,  0,  0, 83, 87, 50,
   83, 76, 65, 68,  0,  0,  1,  0, 83, 87, 49, 83, 84, 65, 82, 71,
    0, 83, 87, 50, 83, 84, 65, 82, 71,  0,  1,  0, 83, 87, 49, 83,
   84, 79, 78, 49,  0, 83, 87, 50, 83, 84, 79, 78, 49,  0,  1,  0,
   83, 87, 49, 83, 84, 79, 78, 50,  0, 83, 87, 50, 83, 84, 79, 78,
   50,  0,  1,  0, 83, 87, 49, 83, 84, 79, 78, 69,  0, 83, 87, 50,
   83, 84, 79, 78, 69,  0,  1,  0, 83, 87, 49, 83, 84, 82, 84, 78,
    0, 83, 87, 50, 83, 84, 82, 84, 78,  0,  1,  0, 83, 87, 49, 66,
   76, 85, 69,  0,  0, 83, 87, 50, 66, 76, 85, 69,  0,  0,  2,  0,
   83, 87, 49, 67, 77, 84,  0,  0,  0, 83, 87, 50, 67, 77, 84,  0,
    0,  0,  2,  0, 83, 87, 49, 71, 65, 82, 71,  0,  0, 83, 87, 50,
   71, 65, 82, 71,  0,  0,  2,  0, 83, 87, 49, 71, 83, 84, 79, 78,
    0, 83, 87, 50, 71, 83, 84, 79, 78,  0,  2,  0, 83, 87, 49, 72,
   79, 84,  0,  0,  0, 83, 87, 50, 72, 79, 84,  0,  0,  0,  2,  0,
   83, 87, 49, 76, 73, 79, 78,  0,  0, 83, 87, 50, 76, 73, 79, 78,
    0,  0,  2,  0, 83, 87, 49, 83, 65, 84, 89, 82,  0, 83, 87, 50,
   83, 65, 84, 89, 82,  0,  2,  0, 83, 87, 49, 83, 75, 73, 78,  0,
    0, 83, 87, 50, 83, 75, 73, 78,  0,  0,  2,  0, 83, 87, 49, 86,
   73, 78, 69,  0,  0, 83, 87, 50, 86, 73, 78, 69,  0,  0,  2,  0,
   83, 87, 49, 87, 79, 79, 68,  0,  0, 83, 87, 50, 87, 79, 79, 68,
    0,  0,  2,  0, 83, 87, 49, 80, 65, 78, 69, 76,  0, 83, 87, 50,
   80, 65, 78, 69, 76,  0,  3,  0, 83, 87, 49, 82, 79, 67, 75,  0,
    0, 83, 87, 50, 82, 79, 67, 75,  0,  0,  3,  0, 83, 87, 49, 77,
   69, 84, 50,  0,  0, 83, 87, 50, 77, 69, 84, 50,  0,  0,  3,  0,
   83, 87, 49, 87, 68, 77, 69, 84,  0, 83, 87, 50, 87, 68, 77, 69,
   84,  0,  3,  0, 83, 87, 49, 66, 82, 73, 75,  0,  0, 83, 87, 50,
   66, 82, 73, 75,  0,  0,  3,  0, 83, 87, 49, 77, 79, 68, 49,  0,
    0, 83, 87, 50, 77, 79, 68, 49,  0,  0,  3,  0, 83, 87, 49, 90,
   73, 77,  0,  0,  0, 83, 87, 50, 90, 73, 77,  0,  0,  0,  3,  0,
   83, 87, 49, 83, 84, 79, 78, 54,  0, 83, 87, 50, 83, 84, 79, 78,
   54,  0,  3,  0, 83, 87, 49, 84, 69, 75,  0,  0,  0, 83, 87, 50,
   84, 69, 75,  0,  0,  0,  3,  0, 83, 87, 49, 77, 65, 82, 66,  0,
    0, 83, 87, 50, 77, 65, 82, 66,  0,  0,  3,  0, 83, 87, 49, 83,
   75, 85, 76, 76,  0, 83, 87, 50, 83, 75, 85, 76, 76,  0,  3,  0,
    0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,
    0,  0,  0,  0,
};

// colormap for underwater, original default greenish
static const unsigned char watermap[] = {
  0,126,127,103,193,5,6,7,8,123,124,126,12,122,123,124,193,193,194,194,88,90,
  92,93,95,96,98,98,100,101,102,103,3,105,106,107,14,15,15,238,238,239,239,239,
  1,2,2,2,193,193,193,193,193,193,193,86,112,112,112,112,113,114,115,115,115,
  116,117,117,118,118,119,120,121,121,122,122,123,124,124,125,193,193,194,194,
  194,194,194,194,194,194,194,194,194,195,195,195,93,94,95,96,97,98,100,101,
  102,104,104,106,106,108,109,110,112,112,113,113,113,114,115,115,117,119,120,
  121,123,124,125,127,112,112,112,113,114,115,115,116,117,118,118,119,119,121,
  121,122,115,116,117,118,120,121,122,123,115,116,117,118,119,121,121,122,112,
  112,112,114,117,120,122,78,193,193,193,194,91,95,136,34,186,187,187,188,188,
  189,189,45,45,46,47,47,47,47,2,2,193,194,194,195,196,197,198,199,200,200,200,
  200,201,202,204,205,193,193,193,193,112,112,114,116,117,118,118,120,120,120,
  74,75,193,193,193,86,112,112,113,114,75,75,76,77,122,124,125,126,205,206,207,
  241,243,244,246,0,113,112,195,198,198,199,204,96,0,126,127,103,193,5,6,7,8,
  123,124,126,12,122,123,124,193,194,194,88,90,90,93,95,95,96,98,98,100,101,
  103,103,105,105,106,107,14,15,15,238,238,239,239,239,1,2,2,2,193,193,193,193,
  193,193,86,86,112,112,112,113,114,115,115,115,115,117,117,117,118,118,120,
  120,121,121,122,122,123,124,124,125,193,194,194,194,194,194,194,194,194,194,
  194,194,195,195,195,195,94,94,95,97,97,98,101,101,103,104,104,106,106,108,
  109,110,112,112,113,113,113,114,115,115,117,119,120,121,123,124,125,127,112,
  112,113,114,115,115,115,117,117,118,118,119,119,121,122,122,115,116,117,118,
  120,121,122,123,115,116,117,118,119,121,121,122,112,112,112,114,117,120,122,
  78,193,193,193,194,91,95,136,34,186,187,188,188,188,189,189,45,45,46,47,47,
  47,47,2,2,193,194,194,195,196,197,198,199,200,200,200,200,201,202,204,205,
  193,193,193,193,112,112,114,116,118,118,120,120,120,120,75,75,193,193,193,86,
  112,112,113,114,75,75,76,77,122,124,125,126,205,206,207,241,243,244,246,0,
  113,112,195,198,198,199,204,96,0,126,127,104,193,5,6,7,8,123,124,126,12,122,
  123,124,193,194,194,90,90,92,93,95,96,98,98,100,101,102,103,3,105,106,107,14,
  15,15,238,238,239,239,239,1,1,2,2,2,193,193,193,193,193,193,86,112,112,112,
  112,113,114,115,115,115,116,117,117,118,118,119,120,121,121,122,122,123,123,
  124,124,125,194,194,194,194,194,194,194,194,194,194,194,194,195,195,195,93,
  94,95,96,97,98,100,101,102,103,104,104,106,106,108,109,110,112,113,113,113,
  114,115,115,117,119,119,120,121,123,124,125,127,112,112,113,114,115,115,116,
  117,118,118,119,119,121,121,122,122,115,117,118,119,121,121,123,124,116,117,
  118,118,119,121,122,122,112,112,112,114,117,120,122,78,193,193,193,88,91,95,
  136,34,187,187,188,188,189,189,45,45,46,47,47,47,47,2,2,2,194,194,194,195,
  196,197,198,199,200,200,200,200,201,202,204,205,193,193,193,86,112,112,114,
  117,118,118,120,120,120,74,75,75,193,193,193,86,112,112,113,114,75,76,76,77,
  123,124,125,126,205,206,207,241,243,244,246,0,114,112,195,198,198,199,204,96,
  0,126,127,104,193,5,6,7,8,123,124,126,12,123,124,125,194,194,88,90,92,93,95,
  96,98,98,100,100,101,102,3,3,106,106,107,14,15,15,238,238,239,239,1,1,1,2,2,
  2,194,193,193,193,193,194,112,112,112,112,113,114,115,115,115,116,117,117,
  118,118,118,119,121,121,121,122,122,123,123,124,125,125,194,194,194,194,194,
  194,194,194,194,194,194,195,195,195,93,94,95,95,97,98,98,100,102,102,104,104,
  106,106,108,108,110,110,112,113,113,113,114,115,115,117,119,120,121,123,123,
  124,125,127,112,113,114,115,115,116,117,118,118,119,119,119,121,121,122,122,
  116,117,118,119,121,122,123,124,116,117,118,119,121,121,122,123,112,112,114,
  117,117,120,122,78,193,194,194,88,92,95,136,34,187,188,188,189,189,189,45,45,
  46,47,47,47,47,2,2,2,194,194,195,195,196,197,198,199,200,200,200,201,202,202,
  204,206,193,194,193,112,112,113,114,118,118,120,120,120,74,75,75,75,193,193,
  193,86,112,112,113,114,75,76,77,78,123,124,125,126,206,207,207,241,243,244,
  246,0,114,112,195,198,198,199,204,96,0,126,127,104,194,5,6,7,8,123,124,126,
  12,123,124,125,194,88,90,90,93,93,95,96,98,100,100,101,102,103,3,105,106,107,
  14,15,15,15,238,238,239,239,1,1,2,2,2,2,194,194,193,193,194,194,194,112,112,
  113,114,115,115,115,116,116,117,118,118,118,119,120,121,121,121,122,122,123,
  124,124,125,125,194,194,194,194,194,194,194,194,194,194,195,195,195,93,93,95,
  95,96,97,98,100,101,102,103,104,106,106,106,108,109,110,5,113,113,113,114,
  114,115,115,117,119,120,121,123,123,124,125,127,113,113,114,115,116,116,117,
  118,118,119,119,121,121,122,122,122,116,117,118,120,121,122,123,124,116,118,
  118,119,121,121,122,123,112,112,114,117,120,122,78,78,194,194,194,90,93,96,
  136,34,187,188,188,189,189,45,45,46,46,47,47,47,47,2,2,2,194,194,195,196,196,
  197,198,199,200,200,200,201,202,204,205,206,194,194,193,112,112,114,115,118,
  120,120,120,120,74,75,75,75,194,194,193,112,112,112,113,114,76,76,77,78,123,
  124,125,126,206,207,241,241,243,244,246,0,115,112,195,198,199,199,204,117,0,
  127,127,104,194,5,6,7,8,123,124,126,12,123,124,125,194,90,90,93,93,95,96,98,
  98,100,101,102,103,3,105,106,106,14,14,15,15,238,238,239,239,239,1,1,2,2,2,2,
  194,194,194,112,112,112,88,112,113,114,114,115,115,116,117,117,117,118,118,
  119,119,121,121,121,122,122,122,123,124,124,125,126,194,194,194,194,194,194,
  194,194,195,195,195,195,93,94,94,95,96,97,98,100,100,101,103,103,104,106,106,
  108,108,109,110,5,113,113,113,114,115,115,117,117,119,120,121,123,124,125,
  125,127,113,114,115,115,116,117,118,118,119,119,119,121,121,122,122,123,117,
  118,119,120,121,122,123,124,117,118,119,119,121,121,122,123,112,112,114,117,
  120,122,78,78,194,194,194,90,95,98,136,188,188,188,189,189,45,45,45,46,47,47,
  47,47,47,2,2,2,194,194,195,196,197,197,198,200,200,200,200,201,202,204,205,
  206,194,194,112,112,113,115,115,118,120,120,120,74,75,75,75,76,194,194,193,
  112,112,112,112,114,76,77,78,78,123,125,125,126,206,207,241,243,243,244,246,
  0,115,112,195,198,199,204,204,117,0,127,127,106,194,6,6,7,8,124,126,126,12,
  123,124,125,112,90,92,93,95,96,96,98,100,100,102,102,103,3,106,106,14,14,15,
  15,238,238,238,239,239,1,1,2,2,2,2,2,194,194,112,112,112,112,112,113,113,114,
  115,115,116,116,117,117,118,118,119,119,120,121,121,121,122,122,123,123,124,
  124,125,126,194,194,194,194,194,194,194,194,195,195,195,195,93,94,95,96,97,
  97,98,100,101,102,103,104,104,106,106,108,108,109,110,5,113,113,114,114,115,
  115,117,119,120,120,121,123,124,125,127,127,114,115,115,116,117,117,118,118,
  119,119,121,121,121,122,123,123,117,118,119,121,122,122,124,124,117,118,119,
  121,121,122,123,123,112,112,114,117,120,122,78,78,194,194,112,92,96,98,101,
  188,188,189,189,189,45,45,46,46,47,47,47,47,2,2,2,2,194,194,195,196,197,198,
  198,200,200,200,201,201,202,204,205,206,194,194,112,112,114,115,115,120,120,
  120,74,74,75,75,75,76,194,194,112,112,112,112,112,114,76,77,78,78,124,125,
  126,126,206,207,241,243,244,244,246,0,115,112,195,198,199,204,204,118,0,127,
  127,106,194,6,6,7,8,124,126,12,12,124,124,125,112,112,93,96,96,96,96,100,100,
  101,102,103,3,105,106,106,14,14,15,15,238,238,239,239,1,1,2,2,2,2,2,2,194,
  194,112,112,112,112,112,112,114,115,115,115,116,117,117,118,118,118,119,120,
  121,121,121,122,122,122,123,124,124,124,125,126,194,194,194,194,194,194,194,
  195,195,195,93,93,94,95,95,96,98,98,100,101,101,103,104,104,106,106,106,108,
  108,109,5,5,113,113,114,115,115,115,117,119,120,120,121,123,124,125,127,127,
  115,115,116,117,117,118,118,119,119,121,121,121,122,122,123,123,118,119,119,
  121,122,123,124,124,118,118,119,121,121,122,123,123,112,114,114,117,120,122,
  78,78,194,194,112,93,96,100,102,188,188,189,189,45,45,45,46,47,47,47,47,47,2,
  2,2,2,194,194,194,196,197,198,198,200,200,200,201,202,202,204,205,206,194,
  194,112,112,114,115,117,120,120,120,74,75,75,75,76,76,194,194,112,112,112,
  112,112,114,77,78,78,78,124,125,126,126,206,207,241,243,244,244,246,0,115,
  112,90,198,199,204,204,118,0,127,127,106,194,6,7,8,8,124,126,12,12,124,125,
  125,112,112,113,96,96,96,96,100,101,102,103,103,3,106,106,14,14,15,15,238,
  238,238,239,1,1,1,2,2,2,2,2,2,194,112,112,112,112,112,112,112,115,115,115,
  116,117,117,118,118,118,119,120,121,121,121,122,122,122,123,123,124,124,125,
  125,126,194,194,194,194,194,195,195,195,195,93,94,94,95,96,96,97,98,98,100,
  101,102,103,104,104,106,106,108,108,109,110,5,5,113,114,114,115,115,117,119,
  119,120,121,123,123,124,125,127,127,115,115,116,117,118,118,119,119,119,121,
  121,122,122,123,123,124,118,119,121,121,122,123,124,124,118,119,119,121,121,
  122,123,123,112,114,117,118,120,122,78,1,194,112,112,95,98,101,103,189,189,
  189,45,45,45,46,46,47,47,47,47,47,2,2,2,2,194,194,195,196,197,198,200,200,
  200,200,201,202,204,204,205,207,194,194,112,112,115,115,117,120,120,74,75,75,
  75,76,76,77,194,194,112,112,112,112,114,114,78,78,78,239,124,125,126,126,207,
  207,241,243,244,244,246,0,116,114,90,198,199,204,204,118,0,127,127,106,194,6,
  7,8,8,124,126,12,12,124,125,125,112,113,96,96,96,96,100,100,118,103,119,120,
  106,106,14,14,15,15,15,238,238,239,1,1,1,1,2,2,2,2,2,2,194,112,112,112,112,
  112,113,113,115,115,116,117,117,118,118,118,119,120,120,121,121,121,122,122,
  122,123,124,124,124,125,125,126,194,194,194,194,195,195,195,195,93,93,94,95,
  96,96,97,98,100,100,101,102,103,104,104,104,106,106,108,109,109,110,5,5,113,
  114,115,115,115,117,119,120,120,121,123,123,124,125,127,127,115,116,117,118,
  118,118,119,119,121,121,121,122,122,123,123,124,118,119,121,121,122,123,124,
  124,118,119,121,121,122,123,123,123,112,114,117,118,121,122,78,1,194,112,112,
  96,98,101,103,189,189,45,45,45,46,46,47,47,47,47,47,2,2,2,2,2,194,194,195,93,
  197,198,200,200,200,201,201,202,204,205,205,207,194,112,112,113,115,116,117,
  120,120,75,75,75,122,76,76,78,194,112,112,112,112,114,114,114,78,78,78,239,
  124,125,126,126,207,241,241,243,244,244,246,0,117,114,195,198,199,204,204,
  119,0,127,7,106,194,6,7,8,8,124,126,12,12,124,125,125,113,96,96,96,96,96,100,
  118,119,119,120,120,106,106,14,14,15,15,238,238,238,239,1,1,1,2,2,2,2,2,2,2,
  194,112,112,112,113,113,114,115,115,116,117,117,118,118,118,119,119,120,121,
  121,121,122,122,122,123,123,124,124,125,125,126,126,194,194,195,195,195,195,
  195,93,94,94,95,95,96,97,98,98,100,101,102,103,103,104,104,106,106,108,108,
  109,109,110,5,5,114,115,115,115,117,119,119,120,121,121,123,124,124,125,127,
  127,116,117,118,118,118,119,119,119,121,121,121,122,122,123,124,124,119,119,
  121,122,123,123,124,125,119,121,121,121,122,123,123,110,114,114,117,118,121,
  78,78,1,194,194,113,96,100,101,3,189,189,45,45,45,46,47,47,47,47,47,47,2,2,2,
  2,2,194,195,195,94,96,198,200,200,200,201,202,202,204,205,206,207,194,194,
  112,113,116,117,117,120,120,75,122,122,122,76,78,78,194,112,112,112,114,114,
  114,114,78,78,78,239,124,125,126,126,207,241,241,243,244,246,246,0,117,114,
  96,199,199,204,204,119,0,127,7,108,194,6,7,8,8,124,126,127,12,124,125,125,96,
  96,96,96,96,117,118,119,119,120,121,121,121,121,14,15,15,15,238,238,1,1,1,1,
  1,2,2,2,2,2,2,2,194,112,112,113,113,114,115,115,116,117,117,118,118,118,119,
  119,120,121,121,121,121,122,122,122,123,124,124,124,125,125,126,126,194,195,
  195,195,195,93,93,94,95,95,96,96,97,98,98,100,101,101,102,103,104,104,106,
  106,106,108,108,109,110,5,5,5,114,115,115,117,117,119,120,120,121,121,123,
  124,124,125,127,7,117,117,118,118,119,119,119,121,121,122,122,122,123,123,
  124,124,119,121,121,122,123,124,124,125,119,121,121,121,122,123,123,110,114,
  116,118,120,121,78,78,1,194,195,114,96,100,102,105,45,45,45,45,46,46,47,47,
  47,47,47,2,2,2,2,2,2,194,195,93,95,97,200,200,200,200,201,202,204,204,205,
  206,207,194,194,113,115,116,117,120,120,120,75,122,122,122,78,78,78,194,112,
  112,113,114,114,114,117,78,78,78,239,124,125,126,126,207,241,243,243,244,246,
  246,0,117,114,96,199,199,204,204,119,0,127,7,108,194,6,7,8,8,5,126,12,8,124,
  125,125,96,96,96,117,117,118,118,119,120,120,121,121,121,121,15,15,15,238,
  238,239,1,1,1,1,2,2,2,2,2,2,2,2,195,113,113,114,114,115,115,115,116,117,118,
  118,118,119,120,120,121,121,121,121,122,122,122,123,123,124,124,124,125,125,
  126,126,195,195,195,195,93,94,94,95,96,96,97,97,98,100,100,101,102,102,103,
  104,104,106,106,106,108,108,109,109,110,5,5,6,115,115,115,117,119,119,120,
  120,121,123,123,124,125,125,127,8,118,118,118,119,119,119,121,121,121,122,
  122,123,123,124,124,124,119,121,122,122,123,124,124,125,119,121,121,122,123,
  123,124,124,115,116,118,121,121,122,123,1,194,195,115,96,101,103,105,45,45,
  45,46,46,47,47,47,47,47,47,2,2,2,2,2,2,195,195,94,96,98,201,201,201,201,202,
  202,204,204,205,206,241,194,195,114,115,117,118,120,120,122,122,122,122,122,
  78,78,78,194,112,113,115,114,114,117,117,78,78,239,1,124,125,126,127,241,241,
  243,243,244,246,246,0,118,117,96,199,204,204,108,119,0,127,7,108,195,6,7,8,8,
  5,126,12,8,125,125,126,96,116,117,117,118,118,119,119,119,121,121,121,121,
  122,15,15,238,238,238,1,1,1,1,2,2,2,2,2,2,2,2,127,195,114,114,115,115,115,
  116,116,116,118,118,118,119,120,120,121,121,121,121,122,122,122,123,123,124,
  124,125,124,125,126,126,126,195,195,93,93,94,95,95,96,96,97,98,98,100,100,
  101,101,102,103,104,104,104,106,106,106,108,109,109,110,110,5,5,6,115,115,
  117,117,119,120,120,121,121,123,123,124,125,125,127,8,118,118,119,119,119,
  121,121,121,122,122,122,123,123,124,124,124,121,121,122,123,124,124,125,125,
  121,121,121,122,123,123,124,124,116,116,118,121,122,123,124,126,195,195,115,
  117,102,3,106,45,45,46,46,47,47,47,47,47,47,2,2,2,2,2,2,127,195,93,95,97,98,
  201,201,201,201,202,202,204,205,205,207,241,195,195,115,115,117,118,120,120,
  122,122,122,122,78,78,78,78,195,195,114,115,116,117,117,117,78,78,239,1,125,
  125,126,127,241,241,243,244,244,246,246,0,118,117,96,199,204,204,108,121,0,
  127,8,109,195,6,7,8,8,125,126,127,8,125,125,126,116,117,117,118,118,118,119,
  119,121,121,121,121,122,122,123,123,124,238,1,1,1,1,1,2,2,2,2,2,2,2,127,127,
  93,115,115,115,115,116,116,117,117,118,118,119,120,121,121,121,121,122,122,
  122,122,123,123,124,124,124,125,125,125,126,126,127,93,93,94,94,95,96,96,97,
  97,98,98,100,100,101,101,102,103,104,104,106,106,106,106,108,108,109,109,110,
  5,5,6,6,115,117,117,119,119,120,121,121,123,123,124,124,125,127,127,8,118,
  119,119,119,121,121,121,122,122,122,123,123,124,124,124,125,121,122,122,123,
  124,124,125,125,121,121,122,122,123,109,124,5,116,117,120,121,122,123,124,
  126,195,93,116,117,103,105,107,15,46,46,47,47,47,47,47,47,47,2,2,2,2,2,2,127,
  93,95,96,98,100,201,201,201,201,202,204,204,205,206,207,241,195,93,115,116,
  117,118,119,122,122,122,122,78,78,78,78,78,195,195,115,116,116,117,117,117,
  78,239,1,1,125,126,126,127,241,243,243,244,244,246,246,0,119,117,96,199,204,
  204,109,121,0,7,8,109,195,7,7,8,8,126,127,127,8,125,125,126,117,117,118,118,
  119,119,119,121,121,121,122,122,122,122,123,123,124,124,1,1,1,1,2,2,2,2,2,2,
  2,2,127,127,94,95,115,116,116,117,117,117,117,118,118,120,121,121,121,121,
  122,122,122,122,122,123,124,124,124,125,125,125,125,126,126,127,94,94,95,95,
  96,97,97,98,98,98,100,100,101,102,102,103,104,104,104,106,106,106,108,108,
  109,109,110,5,5,5,6,6,117,117,119,119,120,120,121,121,123,123,124,124,125,
  127,127,8,119,119,119,121,121,121,122,122,122,123,123,124,124,124,124,125,
  121,122,123,123,124,124,125,126,121,122,122,123,123,110,124,5,116,118,121,
  121,122,123,125,126,195,95,117,118,119,106,107,15,46,47,47,47,47,47,47,47,2,
  2,2,2,2,2,2,127,94,95,97,98,101,202,202,202,202,202,204,204,205,206,207,241,
  195,94,116,117,118,119,120,122,122,122,78,78,78,78,78,78,195,94,115,116,116,
  117,117,120,239,1,1,1,125,126,126,127,241,243,243,244,244,246,246,0,120,118,
  96,199,204,204,109,121,0,7,8,109,94,7,7,8,8,126,127,127,8,125,126,126,118,
  118,118,119,119,119,121,121,121,121,122,122,122,123,123,124,124,124,125,1,1,
  2,2,2,2,2,2,2,127,127,127,127,95,115,117,117,117,118,118,118,118,118,120,121,
  121,121,121,121,122,122,122,122,123,124,124,124,124,125,125,125,125,126,126,
  127,95,95,96,96,97,98,98,98,100,100,101,101,102,103,103,104,104,104,106,106,
  106,108,108,108,109,110,110,5,5,5,6,6,117,119,119,120,120,121,121,123,123,
  123,124,125,125,127,127,8,119,119,121,121,121,122,122,122,123,123,123,124,
  124,124,125,125,122,122,123,124,124,125,125,126,121,122,122,123,109,110,124,
  5,117,118,121,122,122,123,125,126,94,95,117,118,120,106,14,15,47,47,47,47,47,
  47,47,2,2,2,2,2,2,2,127,127,95,96,98,100,102,202,202,202,202,204,204,205,205,
  206,207,241,94,95,117,118,118,120,121,122,122,78,78,78,78,78,78,239,94,115,
  115,116,117,118,120,120,1,1,1,2,125,126,126,127,241,243,243,244,244,246,246,
  0,121,118,97,204,204,204,110,122,0,7,8,110,95,7,7,8,0,6,127,7,8,125,126,126,
  118,119,119,119,119,121,121,121,122,122,122,122,123,123,124,124,124,125,125,
  125,126,126,126,2,2,2,2,127,127,127,127,127,96,97,117,118,118,118,118,118,
  119,119,120,121,121,121,122,122,122,122,123,123,124,124,124,124,125,125,125,
  125,126,126,127,127,96,97,97,97,98,98,100,100,101,101,102,102,103,104,104,
  104,106,106,106,106,108,108,108,109,109,110,5,5,5,6,6,6,119,119,120,120,120,
  121,121,123,123,124,124,125,125,127,127,8,121,121,121,121,122,122,122,123,
  123,124,124,124,124,125,125,125,122,123,123,124,124,125,126,126,122,122,123,
  123,110,124,5,5,118,120,121,122,123,124,125,126,95,97,118,119,121,14,15,238,
  47,47,47,47,47,47,2,2,2,2,2,2,2,2,127,127,96,98,100,101,103,202,202,202,202,
  204,204,205,206,207,241,243,95,96,118,118,118,121,121,122,78,78,78,78,78,78,
  1,1,95,115,116,117,118,118,120,120,1,1,1,2,125,126,127,127,243,243,244,244,
  246,246,0,0,121,120,98,204,204,204,110,122,0,7,8,110,96,7,8,8,0,6,127,7,8,
  126,126,126,119,119,119,121,121,121,121,122,122,122,123,123,123,124,124,124,
  125,125,125,126,126,126,126,126,126,126,127,127,127,127,127,127,97,98,118,
  118,118,119,119,119,119,120,121,121,122,122,122,122,122,123,123,124,124,124,
  125,125,125,125,125,126,126,126,127,127,97,98,98,98,100,100,101,101,102,102,
  103,103,104,104,104,106,106,106,106,108,108,108,109,109,110,5,5,5,5,6,6,6,
  119,119,120,120,121,121,123,123,124,124,124,125,127,127,7,8,121,121,121,122,
  122,122,123,123,124,124,124,124,110,125,125,5,122,123,124,124,125,125,126,
  126,122,123,123,109,110,5,5,126,118,120,122,122,123,124,126,126,96,98,119,
  119,121,14,15,238,47,47,47,47,47,2,2,2,2,2,2,2,2,127,127,127,97,98,101,102,
  104,204,204,204,204,204,205,205,206,207,241,243,96,97,118,119,120,121,121,78,
  78,78,78,78,78,1,1,1,96,116,116,118,118,121,121,122,1,1,126,2,126,126,127,
  127,243,243,244,244,246,246,0,0,121,121,100,204,204,108,110,123,0,8,8,5,97,7,
  8,8,0,6,127,8,8,126,126,127,119,119,121,121,121,121,122,122,123,123,123,123,
  124,124,125,125,125,125,126,126,126,126,126,126,126,127,127,127,127,127,127,
  127,98,100,119,119,119,119,119,119,121,121,121,121,122,122,122,122,123,124,
  124,124,124,125,125,125,125,125,126,126,126,127,127,127,98,100,100,100,101,
  102,102,102,103,103,104,104,104,106,106,106,106,106,108,108,108,109,109,110,
  110,5,5,5,6,6,6,7,120,120,120,121,121,123,123,123,124,124,125,125,127,127,8,
  8,121,122,122,122,123,123,123,124,124,109,124,125,125,125,5,126,123,124,124,
  124,125,125,126,126,108,123,109,110,5,5,126,6,120,120,122,123,124,125,126,
  126,97,100,119,121,121,15,238,239,47,47,47,47,2,2,2,2,2,2,2,2,2,127,127,127,
  98,101,102,103,104,204,204,204,204,205,205,206,207,207,241,243,97,98,119,119,
  120,121,122,78,78,78,78,78,1,1,1,1,97,117,117,118,120,120,121,122,126,126,
  126,2,126,126,127,127,243,243,244,244,246,246,0,0,122,120,101,204,204,109,5,
  123,0,8,8,5,98,7,8,8,0,6,7,8,8,126,126,127,121,121,121,121,122,122,122,123,
  123,123,124,124,124,125,125,125,126,126,126,126,126,126,126,126,127,127,127,
  127,127,127,127,127,100,101,119,119,119,121,121,121,121,121,122,122,122,122,
  123,123,124,124,124,124,125,125,125,125,125,126,126,126,126,127,127,127,100,
  101,101,102,102,103,103,104,104,104,104,104,106,106,106,106,108,108,108,109,
  109,109,110,110,5,5,5,6,6,6,6,7,120,120,121,121,123,123,123,124,124,124,125,
  125,127,127,8,8,122,122,106,106,123,108,108,124,109,109,110,125,125,5,126,
  126,123,124,124,125,125,126,126,127,123,109,110,110,5,5,126,6,120,121,122,
  123,124,125,126,126,98,101,121,121,122,122,238,239,47,47,2,2,2,2,2,2,2,2,2,2,
  127,127,127,127,100,102,103,104,106,204,204,204,204,205,206,206,207,241,243,
  243,98,101,119,121,121,122,122,78,78,78,123,124,125,1,126,126,98,118,118,120,
  120,120,122,122,126,126,126,126,126,127,127,127,243,244,244,244,246,246,0,0,
  122,122,102,204,204,109,5,124,0,8,8,5,101,7,8,8,0,127,7,8,8,126,127,127,121,
  121,122,122,122,122,123,123,124,124,124,124,125,125,125,126,126,126,126,126,
  126,126,126,126,127,127,127,127,127,127,127,127,102,102,103,121,121,121,121,
  121,122,122,122,122,123,123,124,124,124,124,125,125,125,125,125,125,126,126,
  126,126,127,127,127,7,102,102,103,103,103,104,104,104,104,106,106,106,106,
  106,106,108,108,108,109,109,109,110,5,5,5,5,5,6,6,6,7,7,121,121,121,123,123,
  123,124,124,124,125,125,127,127,127,8,8,106,106,106,108,108,108,109,109,109,
  110,110,125,5,126,126,126,124,124,124,125,126,126,126,127,109,110,110,5,5,5,
  6,6,121,122,123,124,125,126,126,126,101,102,121,122,122,123,238,239,47,2,2,2,
  2,2,2,2,2,2,2,127,127,127,127,127,102,103,104,106,106,205,205,205,205,205,
  206,207,207,241,243,243,101,102,121,121,122,122,123,123,123,124,124,125,125,
  126,126,126,101,118,119,119,121,122,122,122,126,126,126,126,126,127,127,7,
  243,244,244,246,246,246,0,0,122,122,104,204,204,110,5,124,0,8,8,5,102,8,8,0,
  0,7,7,8,8,127,127,127,122,122,122,123,123,123,124,124,124,124,125,125,125,
  125,126,126,126,126,126,126,126,126,127,127,127,127,127,127,127,127,127,127,
  103,104,104,121,121,122,122,122,122,122,123,123,124,124,124,124,124,125,125,
  125,125,125,126,126,126,126,126,126,127,127,127,7,103,104,104,104,104,104,
  106,106,106,106,106,106,108,108,108,108,109,109,109,110,110,5,5,5,5,6,6,6,6,
  6,7,7,121,121,123,123,123,124,124,124,125,125,125,127,127,127,8,8,106,108,
  108,108,109,109,109,110,110,110,5,5,5,126,126,126,124,124,125,125,126,126,
  127,127,110,110,5,5,5,6,6,6,121,122,123,124,125,126,126,127,102,104,122,122,
  123,124,1,1,2,2,2,2,2,2,2,2,2,2,127,127,127,127,127,127,103,104,106,106,108,
  205,205,205,205,206,207,207,241,241,243,244,102,103,121,122,122,123,123,124,
  124,125,125,125,126,126,126,126,102,103,119,121,122,122,122,122,126,126,126,
  126,127,127,127,8,244,244,244,246,246,246,0,0,123,122,104,204,108,110,6,125,
  0,8,8,6,104,8,8,0,0,7,8,8,0,127,127,127,122,123,123,123,123,124,124,124,125,
  125,125,125,126,126,126,126,126,126,126,126,126,127,127,127,127,127,127,127,
  127,127,8,8,104,104,104,122,122,122,122,123,123,123,124,124,124,124,125,125,
  125,125,125,125,126,126,126,126,126,126,126,127,127,127,7,8,104,104,104,106,
  106,106,106,106,106,108,108,108,108,109,109,109,109,110,110,5,5,5,5,5,6,6,6,
  6,6,7,7,7,123,123,123,123,124,124,124,125,125,125,127,127,127,8,8,8,108,108,
  109,109,109,110,110,110,5,5,5,126,126,126,6,127,124,125,125,126,126,126,127,
  127,110,5,5,5,6,6,6,127,122,123,124,124,125,126,126,127,104,104,122,123,124,
  124,1,1,2,2,2,2,2,2,2,2,2,127,127,127,127,127,127,8,104,106,106,108,108,206,
  206,206,206,207,207,241,241,243,243,244,104,104,122,122,123,124,124,125,125,
  125,126,126,126,126,126,126,104,104,121,121,122,122,123,123,126,126,126,127,
  127,127,7,8,244,244,244,246,246,246,0,0,124,123,106,204,109,5,6,110,0,8,8,6,
  104,8,8,0,0,7,8,8,0,127,127,7,123,123,124,124,124,124,125,125,125,125,126,
  126,126,126,126,126,126,126,126,127,127,127,127,127,127,127,127,127,127,8,8,
  8,106,106,106,106,123,123,123,123,124,124,124,124,125,125,125,125,125,125,
  125,126,126,126,126,126,126,127,127,127,127,7,7,8,106,106,106,106,106,106,
  108,108,108,108,108,109,109,109,109,110,110,110,5,5,5,5,5,6,6,6,6,6,7,7,7,7,
  123,123,124,124,124,124,125,125,125,125,127,127,127,8,8,8,109,109,109,110,
  110,110,5,5,5,5,126,126,126,6,127,127,125,125,126,126,126,127,127,127,5,5,5,
  6,6,6,127,7,123,124,124,125,126,126,126,127,104,106,123,124,124,125,126,2,2,
  2,2,2,2,2,2,2,127,127,127,127,127,127,8,8,106,106,108,108,109,207,207,207,
  207,207,241,241,243,243,244,244,104,106,106,123,124,124,125,125,126,126,126,
  126,126,126,126,126,104,106,121,122,123,123,124,124,126,126,127,127,127,127,
  7,8,244,244,246,246,246,246,0,0,124,123,106,109,110,5,6,110,0,8,0,6,106,8,8,
  0,0,7,8,8,0,127,7,7,108,124,109,109,125,125,125,126,126,126,126,126,126,126,
  126,126,126,127,127,127,127,127,127,127,127,127,127,127,8,8,8,8,106,108,108,
  108,108,124,124,124,124,125,125,125,125,125,125,126,126,126,126,126,126,126,
  126,126,127,127,127,127,7,7,8,8,106,106,108,108,108,108,108,109,109,109,109,
  109,110,110,110,5,5,5,5,5,5,6,6,6,6,6,6,7,7,7,7,8,124,124,124,124,125,125,
  125,125,127,127,127,127,7,8,8,0,110,110,110,110,5,5,5,5,126,126,126,6,6,127,
  127,127,5,126,126,126,127,127,127,7,5,6,6,6,6,7,7,7,123,124,124,125,126,126,
  127,127,106,108,108,124,125,125,126,2,2,2,2,2,2,2,127,127,127,127,127,127,
  127,8,8,8,106,108,108,109,110,207,207,207,207,241,241,243,243,243,244,244,
  106,106,108,124,124,125,125,126,126,126,126,126,126,126,126,126,106,106,122,
  123,124,124,124,124,126,127,127,127,127,7,8,8,244,246,246,246,246,0,0,0,125,
  124,108,109,5,5,6,5,0,8,0,7,108,8,8,0,0,8,8,8,0,7,7,8,109,110,110,110,110,
  125,126,126,126,126,126,126,126,126,127,127,127,127,127,127,127,127,127,127,
  127,8,8,8,8,8,8,8,108,108,109,109,109,109,125,125,125,125,125,125,125,126,
  126,126,126,126,126,126,126,127,127,127,127,127,127,7,7,8,8,8,108,108,109,
  109,109,109,109,110,110,110,110,5,5,5,5,5,5,5,6,6,6,6,6,6,6,7,7,7,7,7,8,8,
  124,124,125,125,125,125,125,127,127,127,127,127,8,8,8,0,110,5,5,5,5,5,126,
  126,6,6,6,127,127,127,7,7,126,126,126,127,127,127,7,7,6,6,6,6,7,7,7,7,124,
  125,125,126,126,127,127,127,108,108,109,125,126,126,126,2,2,2,2,127,127,127,
  127,127,127,127,127,127,8,8,8,8,108,109,109,110,5,241,241,241,241,241,243,
  243,243,244,244,246,108,108,109,124,125,125,126,126,126,126,126,126,126,126,
  127,127,108,108,123,123,124,124,124,124,127,127,127,127,7,7,8,8,246,246,246,
  246,246,0,0,0,126,124,109,110,5,6,7,126,0,8,0,7,109,8,0,0,0,8,8,0,0,7,8,8,
  110,5,5,5,5,126,126,126,126,126,126,127,127,127,127,127,127,127,127,127,127,
  127,127,8,8,8,8,8,8,8,8,0,109,110,110,110,110,110,125,125,125,125,126,126,
  126,126,126,126,126,126,127,127,127,127,127,127,127,127,7,7,8,8,8,8,109,110,
  110,110,110,110,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,7,7,7,7,7,7,7,8,8,8,125,125,
  125,125,125,127,127,127,127,127,7,8,8,8,0,0,5,5,5,6,6,6,6,6,6,127,127,127,7,
  7,7,7,6,6,127,127,7,7,8,8,6,6,6,7,7,7,7,8,124,125,126,126,127,127,127,127,
  109,110,110,5,126,126,126,127,127,127,127,127,127,127,127,127,127,127,8,8,8,
  8,8,0,109,110,110,5,5,243,243,243,243,243,243,244,244,244,244,246,109,109,
  110,125,125,126,126,126,126,126,126,127,127,127,127,127,109,109,110,124,124,
  124,124,124,127,127,127,127,7,8,8,8,246,246,246,246,0,0,0,0,126,125,110,5,6,
  6,7,6,0,0,0,7,110,0,0,0,0,8,8,0,0,8,8,8,5,5,126,126,126,126,127,127,127,127,
  127,127,127,127,127,127,127,127,7,8,8,8,8,8,8,8,8,8,8,0,0,0,5,5,5,5,5,5,126,
  126,126,126,126,126,126,127,127,127,127,127,127,127,127,127,127,7,7,7,8,8,8,
  8,8,8,5,5,5,5,5,5,5,5,5,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,8,8,8,8,8,8,125,
  125,127,127,127,127,127,127,7,8,8,8,8,8,0,0,6,6,6,6,6,6,127,7,7,7,7,7,7,7,8,
  8,127,127,7,7,7,8,8,8,7,7,7,7,7,8,8,8,126,126,126,127,127,127,127,8,110,5,5,
  126,126,127,127,127,127,127,127,127,127,127,127,127,8,8,8,8,8,8,0,0,5,5,5,6,
  6,243,243,243,243,244,244,244,244,246,246,246,110,5,5,126,126,126,126,127,
  127,127,127,127,127,127,127,127,110,5,5,126,126,126,126,126,127,127,127,8,8,
  8,8,0,246,246,246,246,0,0,0,0,126,126,5,6,6,7,7,127,0,0,0,8,5,0,0,0,0,8,0,0,
  0,8,8,8,6,6,6,127,127,127,127,127,127,127,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,
  8,0,0,0,0,6,6,6,6,6,6,6,6,127,127,127,127,127,127,127,127,127,127,7,7,7,7,8,
  8,8,8,8,8,8,8,8,0,6,6,6,6,6,6,6,6,6,6,6,6,6,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,
  8,8,0,127,127,127,127,127,7,7,8,8,8,8,8,8,0,0,0,6,6,7,7,7,7,7,7,7,7,7,8,8,8,
  8,8,7,7,7,8,8,8,8,8,7,7,7,8,8,8,8,8,126,127,127,127,127,8,8,8,5,6,6,6,127,
  127,127,127,127,127,127,8,8,8,8,8,8,8,8,8,0,0,0,0,6,6,6,6,6,244,244,244,244,
  244,244,244,246,246,246,246,5,6,6,6,127,127,127,127,127,127,127,127,127,127,
  127,8,5,6,6,126,126,126,126,12,8,8,8,8,8,8,0,0,246,246,246,0,0,0,0,0,127,127,
  6,6,7,7,8,7,0,0,0,8,6,0,0,0,0,0,0,0,0,8,0,0,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,
  8,8,8,8,8,8,0,0,0,0,0,0,0,0,0,0,6,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,
  8,8,8,8,8,8,0,0,0,0,6,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,
  0,0,0,0,7,7,8,8,8,8,8,8,8,8,8,0,0,0,0,0,7,7,7,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
  8,8,8,8,8,0,8,8,8,8,8,8,8,0,127,7,7,8,8,8,8,0,6,7,7,7,7,7,8,8,8,8,8,8,8,8,8,
  0,0,0,0,0,0,0,0,0,7,7,7,7,7,246,246,246,246,246,246,246,246,246,246,0,6,6,7,
  7,7,7,7,8,8,8,8,8,8,8,8,8,6,6,7,127,127,127,12,12,8,8,8,0,8,0,0,0,0,0,0,0,0,
  0,0,0,7,127,7,7,8,8,8,8,0,0,0,0,8,0,0,0,0,0,0,0,0,0,0,0,8,8,8,8,8,8,8,8,8,8,
  8,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,
  8,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,8,0,0,0,
  0,0,0,0,0,0,0,0,0,0,8,8,8,8,8,0,0,0,0,0,0,0,0,0,0,0,8,8,8,8,8,8,8,8,0,0,0,0,
  0,0,0,0,8,8,0,0,0,0,0,0,8,0,0,0,0,0,0,0,8,8,8,8,0,0,0,0,8,8,8,8,8,8,8,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,8,8,8,8,8,246,246,246,246,246,246,246,0,0,0,0,8,
  8,8,8,8,8,8,8,8,0,0,0,0,0,0,0,8,8,8,8,8,8,8,8,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,8,8,8,8,8,0,0,8,193,193,193,194,0,193,193,193,193,194,194,193,193,194,194,
  194,108,106,104,103,102,100,98,97,95,94,93,195,195,195,194,194,194,194,194,
  194,194,194,194,194,194,194,194,194,194,193,193,193,7,6,6,5,110,109,108,108,
  106,104,104,102,101,98,97,97,95,94,93,93,195,195,194,194,194,194,194,194,194,
  194,194,194,7,6,5,5,109,108,108,106,104,104,102,101,100,98,97,96,95,94,93,
  195,195,195,194,194,194,194,194,194,194,194,194,194,106,104,102,100,97,95,93,
  195,195,194,194,194,194,194,193,193,102,101,98,97,96,95,94,93,195,195,195,
  194,194,194,194,194,95,93,195,195,194,194,194,194,93,195,195,194,194,194,194,
  194,7,108,104,98,94,195,194,194,0,6,109,104,100,96,93,194,194,194,194,194,
  194,194,194,194,194,194,194,194,194,193,193,193,7,109,104,100,95,195,194,194,
  194,194,194,193,193,193,193,193,0,7,5,109,106,104,101,98,96,95,94,93,195,195,
  194,194,0,0,8,7,7,6,5,5,194,194,194,194,194,194,194,194,193,193,193,193,193,
  193,193,193,103,110,104,195,194,194,194,94,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,0,0,0,
};

//jff 2/18/98 palette color ranges for translation

// 2023/12/05 brick is now stand-in for red to make the minimized HUD work
static const char cr_brick[]=
    {240,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,                          // 2023/12/05 minimized HUD font range
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,79,79,46,79,                  // 2023/12/05 minimized HUD font range
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,191,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,76,   // 2023/12/05 HUD font range, from Harmony v1.1
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     64,65,66,191,68,69,70,71,72,73,74,75,76,77,78,79,                 // 2023/12/05 minimized HUD font range
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,65,68,69,71,72,73,247,248,249,250,251,252,253,254,255};       // 2023/12/05 HUD font range, from Harmony v1.1

static const char cr_tan[]=
    {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,78,45,78,78,
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,78,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,139,  // 2023/12/05 range for HUD font, from HELP screen
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,171,173,133,134,135,136,247,248,249,250,251,252,253,254,255};  // 2023/12/05 range for HUD font, from HELP screen
    
static const char cr_gray[]=
    {240,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,                           // 2023/12/05 minimized HUD font range
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,110,110,46,110,
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,110,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,26,    // 2023/12/05 HUD font range, from Harmony v1.1
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     2,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,     // 2023/12/05 minimized HUD font range
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,1,7,10,15,18,19,247,248,249,250,251,252,253,254,255};          // 2023/12/05 HUD font range, from Harmony v1.1
    
static const char cr_green[]=
    {240,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,                           // 2023/12/05 minimized HUD font range
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,111,111,46,111,                // 2023/12/05 minimized HUD font range
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,127,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,108,   // 2023/12/05 range for HUD font, analogy from red range
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     97,98,99,100,101,102,103,104,105,106,107,108,109,110,110,111,      // 2023/12/05 minimized HUD font range
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,97,100,101,103,104,105,247,248,249,250,251,252,253,254,255};   // 2023/12/05 range for HUD font, analogy from red range
    
static const char cr_brown[]=
    {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,143,143,46,143,
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,143,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,176,178,180,182,184,186,247,248,249,250,251,252,253,254,255};  // 2023/12/02 adapt colour ranges to Harmony palette/font
    
static const char cr_gold[]=
    {240,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,                           // 2023/12/05 minimized HUD font range
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,92,92,46,92,                   // 2023/12/05 minimized HUD font range
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,167,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,88,    // 2023/12/05 HUD font range, from Harmony v1.1
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     225,225,226,226,227,227,228,228,229,229,230,230,231,231,232,232,   // 2023/12/05 minimized HUD font range
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,225,170,173,133,178,179,247,248,249,250,251,252,253,254,255};  // 2023/12/05 HUD font range, from Harmony v1.1
    
static const char cr_red[]=
    {240,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,                           // 2023/12/05 minimized HUD font range
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,79,79,46,79,                   // 2023/12/05 minimized HUD font range
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,191,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,76,    // 2023/12/05 HUD font range, from Harmony v1.1
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     64,65,66,191,68,69,70,71,72,73,74,75,76,77,78,79,                  // 2023/12/05 minimized HUD font range
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,65,68,69,71,72,73,247,248,249,250,251,252,253,254,255};        // 2023/12/05 HUD font range, from Harmony v1.1

static const char cr_blue[]=
    {240,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,                           // 2023/12/05 minimized HUD font range
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,246,246,46,246,                // 2023/12/05 minimized HUD font range
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,203,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     241,145,146,147,242,149,150,151,152,243,154,155,156,157,158,159,   // 2023/12/05 minimized HUD font range
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255};  // 2023/12/05 HUD font range

// killough 2/28/98: The blue translation table above looks bad on the
// status bar but a little better on the HUD. So I've created a separate
// translation table for the status bar, which is the same as last week's.

static const char cr_blue2[]=
    {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,207,207,46,207,
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,207,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,159,
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     200,200,201,201,202,202,203,203,204,204,205,205,206,206,207,207,
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255};

    
static const char cr_orange[]=
    {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,223,223,46,223,
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,223,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,190,  // 2023/12/05 range for HUD font, from CREDIT screen
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,37,40,43,48,88,89,247,248,249,250,251,252,253,254,255};       // 2023/12/05 range for HUD font, from CREDIT screen
    
static const char cr_yellow[]=
    {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,
     16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,
     32,33,34,35,36,37,38,39,40,41,42,43,231,231,46,231,
     48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,
     64,65,66,231,68,69,70,71,72,73,74,75,76,77,78,79,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,
     96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,
     112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,127,
     128,129,130,131,132,133,134,135,136,137,138,139,140,141,142,143,
     144,145,146,147,148,149,150,151,152,153,154,155,156,157,158,92,  // 2023/12/05 range for HUD font, analogy from red range
     160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,
     80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,                 // 2023/12/05 range for minimized HUD
     192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,
     208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,
     224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,
     240,81,84,85,87,88,89,247,248,249,250,251,252,253,254,255};      // 2023/12/05 range for HUD font, analogy from red range

// There are some lumps that need to be included in order to use old
// versions of the DOOM IWADs, notably versions 1.1 and 1.2.  These
// are included here.

static const char stbar[] = {  // default status bar  // 2023/12/02 Harmony version
  64,1,32,0,0,0,0,0,8,5,0,0,45,5,0,0,82,5,0,0,119,5,0,0,156,5,0,0,193,5,0,0,
  230,5,0,0,11,6,0,0,48,6,0,0,85,6,0,0,122,6,0,0,159,6,0,0,196,6,0,0,233,6,0,0,
  14,7,0,0,51,7,0,0,88,7,0,0,125,7,0,0,162,7,0,0,199,7,0,0,236,7,0,0,17,8,0,0,
  54,8,0,0,91,8,0,0,128,8,0,0,165,8,0,0,202,8,0,0,239,8,0,0,20,9,0,0,57,9,0,0,
  94,9,0,0,131,9,0,0,168,9,0,0,205,9,0,0,242,9,0,0,23,10,0,0,60,10,0,0,97,10,0,
  0,134,10,0,0,171,10,0,0,208,10,0,0,245,10,0,0,26,11,0,0,63,11,0,0,100,11,0,0,
  137,11,0,0,174,11,0,0,211,11,0,0,248,11,0,0,29,12,0,0,66,12,0,0,103,12,0,0,
  140,12,0,0,177,12,0,0,214,12,0,0,251,12,0,0,32,13,0,0,69,13,0,0,106,13,0,0,
  143,13,0,0,180,13,0,0,217,13,0,0,254,13,0,0,35,14,0,0,72,14,0,0,109,14,0,0,
  146,14,0,0,183,14,0,0,220,14,0,0,1,15,0,0,38,15,0,0,75,15,0,0,112,15,0,0,149,
  15,0,0,186,15,0,0,223,15,0,0,4,16,0,0,41,16,0,0,78,16,0,0,115,16,0,0,152,16,
  0,0,189,16,0,0,226,16,0,0,7,17,0,0,44,17,0,0,81,17,0,0,118,17,0,0,155,17,0,0,
  192,17,0,0,229,17,0,0,10,18,0,0,47,18,0,0,84,18,0,0,121,18,0,0,158,18,0,0,
  195,18,0,0,232,18,0,0,13,19,0,0,50,19,0,0,87,19,0,0,124,19,0,0,161,19,0,0,
  198,19,0,0,235,19,0,0,16,20,0,0,53,20,0,0,90,20,0,0,127,20,0,0,164,20,0,0,
  201,20,0,0,238,20,0,0,19,21,0,0,56,21,0,0,93,21,0,0,130,21,0,0,167,21,0,0,
  204,21,0,0,241,21,0,0,22,22,0,0,59,22,0,0,96,22,0,0,133,22,0,0,170,22,0,0,
  207,22,0,0,244,22,0,0,25,23,0,0,62,23,0,0,99,23,0,0,136,23,0,0,173,23,0,0,
  210,23,0,0,247,23,0,0,28,24,0,0,65,24,0,0,102,24,0,0,139,24,0,0,176,24,0,0,
  213,24,0,0,250,24,0,0,31,25,0,0,68,25,0,0,105,25,0,0,142,25,0,0,179,25,0,0,
  216,25,0,0,253,25,0,0,34,26,0,0,71,26,0,0,108,26,0,0,145,26,0,0,182,26,0,0,
  219,26,0,0,0,27,0,0,37,27,0,0,74,27,0,0,111,27,0,0,148,27,0,0,185,27,0,0,222,
  27,0,0,3,28,0,0,40,28,0,0,77,28,0,0,114,28,0,0,151,28,0,0,188,28,0,0,225,28,
  0,0,6,29,0,0,43,29,0,0,80,29,0,0,117,29,0,0,154,29,0,0,191,29,0,0,228,29,0,0,
  9,30,0,0,46,30,0,0,83,30,0,0,120,30,0,0,157,30,0,0,194,30,0,0,231,30,0,0,12,
  31,0,0,49,31,0,0,86,31,0,0,123,31,0,0,160,31,0,0,197,31,0,0,234,31,0,0,15,32,
  0,0,52,32,0,0,89,32,0,0,126,32,0,0,163,32,0,0,200,32,0,0,237,32,0,0,18,33,0,
  0,55,33,0,0,92,33,0,0,129,33,0,0,166,33,0,0,203,33,0,0,240,33,0,0,21,34,0,0,
  58,34,0,0,95,34,0,0,132,34,0,0,169,34,0,0,206,34,0,0,243,34,0,0,24,35,0,0,61,
  35,0,0,98,35,0,0,135,35,0,0,172,35,0,0,209,35,0,0,246,35,0,0,27,36,0,0,64,36,
  0,0,101,36,0,0,138,36,0,0,175,36,0,0,212,36,0,0,249,36,0,0,30,37,0,0,67,37,0,
  0,104,37,0,0,141,37,0,0,178,37,0,0,215,37,0,0,252,37,0,0,33,38,0,0,70,38,0,0,
  107,38,0,0,144,38,0,0,181,38,0,0,218,38,0,0,255,38,0,0,36,39,0,0,73,39,0,0,
  110,39,0,0,147,39,0,0,184,39,0,0,221,39,0,0,2,40,0,0,39,40,0,0,76,40,0,0,113,
  40,0,0,150,40,0,0,187,40,0,0,224,40,0,0,5,41,0,0,42,41,0,0,79,41,0,0,116,41,
  0,0,153,41,0,0,190,41,0,0,227,41,0,0,8,42,0,0,45,42,0,0,82,42,0,0,119,42,0,0,
  156,42,0,0,193,42,0,0,230,42,0,0,11,43,0,0,48,43,0,0,85,43,0,0,122,43,0,0,
  159,43,0,0,196,43,0,0,233,43,0,0,14,44,0,0,51,44,0,0,88,44,0,0,125,44,0,0,
  162,44,0,0,199,44,0,0,236,44,0,0,17,45,0,0,54,45,0,0,91,45,0,0,128,45,0,0,
  165,45,0,0,202,45,0,0,239,45,0,0,20,46,0,0,57,46,0,0,94,46,0,0,131,46,0,0,
  168,46,0,0,205,46,0,0,242,46,0,0,23,47,0,0,60,47,0,0,97,47,0,0,134,47,0,0,
  171,47,0,0,208,47,0,0,245,47,0,0,26,48,0,0,63,48,0,0,100,48,0,0,137,48,0,0,
  174,48,0,0,211,48,0,0,248,48,0,0,29,49,0,0,66,49,0,0,103,49,0,0,140,49,0,0,
  177,49,0,0,214,49,0,0,251,49,0,0,32,50,0,0,69,50,0,0,106,50,0,0,143,50,0,0,
  180,50,0,0,217,50,0,0,254,50,0,0,35,51,0,0,0,32,0,154,157,158,158,158,159,
  159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,
  159,159,159,159,158,159,159,0,255,0,32,0,155,156,156,157,157,158,158,158,158,
  158,158,158,158,158,158,158,158,158,158,158,158,158,158,158,158,158,158,158,
  158,158,158,158,0,255,0,32,0,155,156,155,155,156,156,157,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,
  158,0,255,0,32,0,154,157,156,155,155,155,155,156,156,157,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,158,0,255,0,
  32,0,154,156,157,156,156,156,156,157,157,157,158,158,158,158,158,158,157,157,
  157,157,157,157,157,157,158,158,158,158,158,157,157,158,0,255,0,32,0,154,156,
  157,157,158,158,158,158,158,158,158,158,158,158,158,158,158,157,157,157,157,
  157,157,158,159,158,158,158,158,158,158,158,0,255,0,32,0,154,155,156,158,158,
  246,245,245,246,246,246,246,245,245,244,244,245,154,156,157,156,155,155,154,
  151,153,153,153,153,154,158,158,0,255,0,32,0,155,156,157,156,245,244,245,245,
  246,246,246,246,245,244,244,244,245,152,154,157,152,149,150,150,150,153,153,
  153,153,153,158,158,0,255,0,32,0,155,158,157,245,244,244,245,245,246,246,246,
  246,245,244,244,244,245,152,153,157,157,157,156,156,151,149,148,148,149,152,
  158,159,0,255,0,32,0,155,158,156,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,151,154,157,157,157,157,157,157,156,156,156,156,157,158,159,0,
  255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,151,154,157,157,158,159,159,159,159,159,159,159,159,159,158,0,255,0,32,0,
  155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,154,
  157,157,157,158,158,158,158,158,157,156,157,159,158,0,255,0,32,0,155,158,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,154,157,155,155,
  155,156,156,156,156,155,152,154,158,158,0,255,0,32,0,155,158,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,151,155,157,154,153,153,153,153,
  153,154,153,150,152,157,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,151,155,157,157,157,157,157,157,157,157,157,
  156,156,157,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,151,155,157,157,157,157,157,157,157,157,157,157,157,157,
  157,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,151,154,157,157,157,157,157,157,157,157,157,157,157,158,158,0,255,0,
  32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,
  154,156,156,157,157,158,159,159,159,159,159,159,159,159,0,255,0,32,0,155,158,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,154,156,156,
  158,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,151,154,156,157,159,158,158,
  158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,151,154,156,157,159,158,158,158,158,158,
  158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,152,155,156,157,159,158,158,158,158,158,158,158,158,
  158,157,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,152,155,157,156,159,158,158,158,158,158,158,158,158,158,158,0,
  255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,152,155,157,157,159,159,159,159,158,158,158,158,159,158,158,0,255,0,32,0,
  154,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,152,155,
  159,159,159,158,159,159,156,155,157,159,158,158,158,0,255,0,32,0,154,158,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,152,156,157,156,158,
  158,158,159,111,30,156,157,158,158,158,0,255,0,32,0,155,158,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,152,155,157,157,159,158,158,159,
  111,31,156,157,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,151,155,157,157,159,158,158,158,159,159,159,
  159,159,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,152,155,157,157,159,158,158,158,159,158,158,158,158,159,
  159,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,152,155,157,157,159,158,157,158,159,157,158,158,158,158,158,0,255,0,
  32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,152,
  155,157,157,159,158,158,157,158,157,158,158,158,157,158,0,255,0,32,0,155,158,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,152,155,157,157,
  159,158,158,31,157,157,158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,152,155,157,157,159,158,159,
  158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,152,155,157,157,159,159,159,157,158,158,
  158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,152,155,157,158,159,158,158,158,158,158,158,158,158,
  158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,152,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,
  255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,152,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,
  155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,
  157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,159,
  158,158,158,158,158,158,158,157,158,158,0,255,0,32,0,155,158,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,151,154,157,157,159,158,158,158,
  156,157,158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,151,155,157,157,159,158,158,158,157,156,157,
  158,158,158,157,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,244,151,156,157,157,159,158,158,158,158,157,157,158,158,158,
  158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,152,152,157,157,157,159,158,158,158,158,158,158,158,158,158,157,0,255,0,
  32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,151,155,
  157,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,
  246,244,244,244,245,245,246,246,246,246,245,244,244,151,153,156,156,156,157,
  159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,246,245,244,
  245,245,246,246,246,246,245,244,152,152,154,156,156,155,156,156,157,158,158,
  158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,158,158,158,158,158,158,
  158,158,157,155,155,155,156,157,156,156,155,155,156,157,158,158,158,159,159,
  159,159,159,159,159,0,255,0,32,0,155,157,157,157,157,157,157,157,157,157,157,
  157,157,157,157,157,156,156,155,155,156,157,157,157,157,157,157,157,157,157,
  157,158,0,255,0,32,0,155,157,157,157,157,157,157,157,157,157,157,157,157,157,
  157,157,157,157,156,156,156,156,157,157,157,157,157,157,157,156,157,157,0,
  255,0,32,0,155,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,
  157,157,157,157,156,157,157,157,157,157,157,157,157,157,157,157,0,255,0,32,0,
  155,157,157,157,157,157,157,158,157,157,157,157,157,157,157,157,157,157,157,
  157,157,157,157,158,158,157,157,157,157,157,157,157,0,255,0,32,0,155,157,157,
  157,158,158,158,158,158,158,158,158,158,158,158,158,157,157,157,157,157,157,
  158,157,157,157,158,158,157,157,157,158,0,255,0,32,0,155,157,157,158,156,246,
  246,246,246,246,246,246,246,245,245,245,246,155,157,157,157,157,154,152,152,
  152,153,156,158,157,157,158,0,255,0,32,0,155,157,158,156,244,244,245,245,246,
  246,246,246,245,244,244,244,245,152,157,157,157,154,150,149,156,156,150,151,
  155,158,157,158,0,255,0,32,0,155,157,157,245,244,244,245,245,246,246,246,246,
  245,244,244,244,245,151,155,158,155,151,150,150,158,158,151,150,152,157,157,
  158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,150,154,157,153,151,157,158,158,158,158,155,152,156,158,158,0,255,0,
  32,0,156,159,245,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,
  154,158,154,151,157,158,158,158,158,156,152,157,158,158,0,255,0,32,0,156,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,158,155,
  151,150,150,158,158,151,149,151,157,157,158,0,255,0,32,0,155,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,151,155,158,157,154,150,149,
  156,156,150,150,155,157,157,158,0,255,0,32,0,155,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,151,155,157,157,157,155,152,152,152,152,
  154,157,157,157,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,151,155,157,157,157,158,157,156,156,156,157,157,157,
  158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,151,155,157,157,157,157,157,157,157,157,156,156,157,157,158,0,
  255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,151,155,157,157,157,157,157,157,157,157,157,157,157,158,158,0,255,0,32,0,
  156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,
  157,157,157,157,158,159,159,159,159,159,159,159,158,0,255,0,32,0,156,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,157,
  158,159,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,158,159,158,158,
  158,158,158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,150,155,157,157,158,158,158,158,158,158,158,
  158,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,150,155,157,157,159,158,158,158,158,158,158,158,158,158,
  158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,150,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,
  32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,
  155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,
  159,158,158,159,159,159,158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,159,158,159,
  159,159,159,159,158,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,150,155,157,157,159,158,159,159,159,159,
  159,159,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,150,155,157,157,159,158,159,159,159,159,159,159,158,
  158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,150,155,157,157,159,158,158,159,159,159,159,158,158,158,158,0,
  255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,150,155,157,157,159,158,158,159,159,158,158,158,158,158,158,0,255,0,32,0,
  156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,
  157,157,159,159,158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,159,
  159,159,158,158,158,159,159,159,159,158,0,255,0,32,0,156,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,159,159,159,159,
  159,159,159,159,159,159,159,0,255,0,32,0,156,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,150,155,157,157,159,159,159,159,158,159,159,
  159,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,150,155,157,157,159,158,158,158,158,158,159,158,158,158,
  158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,150,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,
  32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,
  155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,
  159,158,158,158,158,158,158,158,159,159,159,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,151,156,158,158,159,159,158,
  158,158,158,159,159,159,159,159,0,255,0,32,0,156,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,152,157,159,159,159,159,158,158,158,158,
  158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,152,157,159,159,159,159,158,158,158,158,158,158,158,
  158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,152,157,159,159,31,159,159,159,159,159,158,158,158,158,158,0,255,
  0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,
  152,157,159,159,31,159,159,159,159,159,159,159,159,159,159,0,255,0,32,0,155,
  158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,157,158,
  158,159,159,159,159,159,159,159,159,159,159,159,0,255,0,32,0,155,158,246,244,
  244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,159,158,
  158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,
  245,246,246,246,246,245,244,244,244,245,150,156,158,158,159,158,158,158,158,
  158,158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,
  246,246,245,244,244,244,245,151,157,159,159,31,159,159,159,159,158,158,158,
  158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,
  244,244,244,245,151,157,158,158,159,159,158,158,158,158,158,158,158,158,158,
  0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,150,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,
  156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,
  157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,157,157,159,
  158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,149,155,157,157,159,158,158,158,
  158,158,158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,149,155,157,157,159,158,158,158,158,158,158,
  158,158,158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,149,155,157,157,159,158,158,158,158,158,158,158,158,158,
  158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,244,148,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,
  32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,152,15,
  156,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,
  245,244,244,244,245,245,246,246,246,246,245,244,244,153,150,154,157,157,157,
  158,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,245,154,155,155,157,157,157,157,157,158,158,
  158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,158,158,158,158,158,158,
  158,158,158,158,158,157,157,157,157,157,157,157,157,157,157,158,159,159,159,
  159,159,159,159,158,0,255,0,32,0,155,157,157,157,157,157,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,158,
  158,158,0,255,0,32,0,155,157,157,157,157,157,157,157,157,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,158,158,0,
  255,0,32,0,155,157,157,157,157,157,157,158,157,157,157,158,158,158,158,158,
  158,158,158,158,157,157,157,157,157,158,158,157,157,157,158,158,0,255,0,32,0,
  155,157,157,158,158,158,158,158,158,158,158,158,158,158,158,158,158,158,158,
  158,157,156,157,157,158,158,158,158,158,157,158,158,0,255,0,32,0,155,157,158,
  158,246,245,246,246,246,246,246,246,246,245,245,245,246,246,246,246,245,151,
  156,155,154,155,155,156,158,158,158,158,0,255,0,32,0,155,158,158,246,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,150,153,154,152,
  157,157,154,158,159,157,158,0,255,0,32,0,155,159,157,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,245,150,152,151,152,158,157,152,
  152,153,153,156,0,255,0,32,0,154,158,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,245,245,245,245,151,150,151,152,157,154,150,153,156,152,
  156,0,255,0,32,0,156,158,245,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,245,245,245,245,151,149,149,151,152,153,154,154,156,151,156,0,255,0,
  32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,
  245,245,245,152,149,147,150,156,154,150,150,155,151,155,0,255,0,32,0,155,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,
  152,152,146,149,157,156,151,152,153,152,156,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,151,154,150,
  150,157,156,153,158,157,156,157,0,255,0,32,0,156,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,245,245,245,245,150,155,154,154,155,154,
  156,158,157,157,157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,245,245,245,245,150,156,157,157,156,157,157,157,157,
  157,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,245,245,245,245,150,156,157,157,157,157,157,157,157,157,158,0,
  255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,150,156,157,157,157,157,157,157,157,158,158,0,255,0,32,0,
  157,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,
  245,245,150,156,157,157,157,157,157,157,157,157,158,0,255,0,32,0,157,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,150,
  156,157,157,157,157,157,157,157,158,158,0,255,0,32,0,156,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,150,156,157,157,
  157,157,157,157,157,158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,245,150,156,157,157,157,157,157,
  157,157,158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,245,245,245,245,150,156,157,157,157,157,157,157,157,158,
  158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,245,245,245,245,150,156,157,157,157,157,157,157,157,157,158,0,255,0,
  32,0,157,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,
  245,245,245,150,156,157,157,156,156,156,156,157,158,158,0,255,0,32,0,156,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,
  150,156,157,157,156,156,156,156,157,157,158,0,255,0,32,0,155,158,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,150,156,157,
  157,157,156,156,156,157,157,158,0,255,0,32,0,155,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,245,245,245,245,150,156,157,157,157,157,
  156,156,156,157,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,245,245,245,245,150,156,157,157,157,157,157,156,156,
  157,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,245,245,245,245,150,156,157,157,157,157,157,157,157,157,158,0,
  255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,150,155,157,157,157,157,157,157,157,157,158,0,255,0,32,0,
  156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,
  245,245,150,155,156,157,157,157,157,157,157,157,158,0,255,0,32,0,155,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,150,
  155,156,156,157,157,157,157,157,157,158,0,255,0,32,0,155,158,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,244,149,155,156,156,
  156,157,157,157,157,157,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,244,149,155,156,156,156,156,157,
  157,157,158,158,0,255,0,32,0,156,159,245,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,245,245,244,152,153,156,157,156,156,156,156,157,157,158,
  158,0,255,0,32,0,155,159,246,245,244,245,246,246,246,246,246,246,246,245,245,
  245,245,245,153,153,154,156,157,157,157,156,156,156,157,157,158,159,0,255,0,
  32,0,155,158,158,157,157,157,157,157,158,157,156,156,156,156,156,156,157,155,
  155,155,156,157,157,157,157,157,156,157,157,157,159,157,0,255,0,32,0,155,157,
  157,157,157,157,157,157,157,156,156,156,156,156,156,156,156,156,157,156,156,
  157,157,157,157,157,157,157,157,158,159,246,0,255,0,32,0,156,157,157,157,157,
  157,157,157,157,157,157,157,157,156,157,157,157,157,157,157,157,157,157,157,
  157,157,157,158,158,159,157,246,0,255,0,32,0,155,157,157,157,158,159,159,159,
  159,159,159,159,159,158,157,158,159,159,159,159,159,159,159,159,159,159,159,
  159,158,246,246,246,0,255,0,32,0,155,157,157,158,246,244,245,245,246,246,246,
  246,245,244,244,244,245,246,245,245,245,245,246,246,246,246,246,246,246,246,
  246,246,0,255,0,32,0,155,158,158,246,244,244,245,245,246,246,246,246,245,244,
  244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,
  255,0,32,0,156,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,
  156,158,245,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,
  245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,156,246,245,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,
  245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,156,246,245,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,
  246,246,246,246,246,246,246,0,255,0,32,0,155,246,245,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,
  246,246,246,246,0,255,0,32,0,155,246,245,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,
  246,0,255,0,32,0,156,246,245,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,
  32,0,156,246,245,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,
  245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,155,246,
  245,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,
  245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,156,246,245,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,
  246,246,246,246,246,246,246,246,0,255,0,32,0,154,246,245,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,
  246,246,246,246,246,0,255,0,32,0,155,246,245,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,
  246,246,0,255,0,32,0,156,246,245,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,
  255,0,32,0,155,246,245,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,
  154,246,245,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,
  245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,155,246,245,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,
  245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,155,246,245,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,
  246,246,246,246,246,246,246,0,255,0,32,0,155,246,245,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,
  246,246,246,246,0,255,0,32,0,155,246,245,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,
  246,0,255,0,32,0,155,246,245,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,
  32,0,156,246,245,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,
  245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,154,246,
  245,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,
  245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,155,246,245,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,
  246,246,246,246,246,246,246,246,0,255,0,32,0,155,246,245,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,
  246,246,246,246,246,0,255,0,32,0,155,246,245,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,
  246,246,0,255,0,32,0,155,246,245,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,
  255,0,32,0,155,156,245,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,
  155,158,245,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,
  245,245,245,245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,155,158,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,
  245,246,246,246,246,246,246,246,246,246,0,255,0,32,0,155,157,158,246,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,
  246,246,246,246,246,246,246,0,255,0,32,0,155,157,158,158,156,245,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,
  246,246,246,246,0,255,0,32,0,155,157,157,157,158,158,156,155,155,154,153,153,
  153,153,152,152,152,152,151,151,151,152,153,154,156,156,158,158,158,156,246,
  246,0,255,0,32,0,155,157,156,157,157,157,157,156,156,156,157,157,157,157,157,
  156,156,157,157,157,157,157,156,156,157,157,157,157,158,158,246,246,0,255,0,
  32,0,155,157,157,157,157,157,157,156,156,156,156,157,157,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,157,157,158,157,246,0,255,0,32,0,155,157,
  157,157,157,157,157,157,156,156,156,156,156,156,157,157,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,158,156,0,255,0,32,0,155,157,157,157,157,
  157,158,158,157,157,157,157,157,157,157,157,157,157,156,157,157,157,157,157,
  157,157,158,157,157,157,158,158,0,255,0,32,0,155,157,157,157,158,159,158,158,
  158,158,158,158,158,158,158,158,158,157,157,156,157,157,159,159,159,159,159,
  159,158,157,157,158,0,255,0,32,0,156,157,157,158,156,246,245,245,246,246,246,
  246,245,244,244,244,245,153,157,157,157,155,156,156,156,156,156,157,158,158,
  157,158,0,255,0,32,0,155,157,158,246,244,244,245,245,246,246,246,246,245,244,
  244,244,245,152,156,157,156,153,156,156,156,156,156,156,156,158,158,158,0,
  255,0,32,0,156,158,156,245,244,244,245,245,246,246,246,246,245,244,244,244,
  245,151,155,157,156,153,154,155,155,155,155,155,155,157,159,158,0,255,0,32,0,
  155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,
  157,157,151,149,149,149,149,149,151,153,156,159,158,0,255,0,32,0,156,158,245,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,151,
  150,150,151,151,151,152,154,157,159,158,0,255,0,32,0,156,159,245,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,151,155,157,156,151,149,148,149,
  149,150,152,156,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,151,155,157,157,154,153,154,154,154,154,156,
  158,157,158,158,0,255,0,32,0,157,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,151,155,157,157,157,157,157,157,157,157,157,157,157,158,
  158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,151,155,157,157,157,157,157,157,157,157,158,157,157,157,158,0,255,0,
  32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,
  155,157,157,157,157,157,157,157,157,157,158,157,157,158,0,255,0,32,0,155,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,
  157,157,157,157,157,157,157,157,157,158,159,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,157,157,158,
  159,159,159,159,159,159,159,158,0,255,0,32,0,155,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,151,155,157,157,157,159,159,159,158,158,
  158,158,158,158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,151,155,157,157,158,159,158,158,158,158,158,158,158,
  158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,151,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,
  255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,151,155,157,158,159,158,158,158,158,157,157,157,157,158,158,0,255,0,32,0,
  155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,
  157,157,159,158,158,158,157,157,156,156,156,157,157,0,255,0,32,0,156,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,158,159,
  158,158,158,158,157,156,156,156,156,156,0,255,0,32,0,156,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,159,158,157,157,
  158,158,158,157,157,156,156,0,255,0,32,0,155,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,151,155,157,157,159,158,157,156,156,157,158,
  158,158,157,156,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,150,155,157,157,159,158,158,157,156,156,157,158,158,158,
  157,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,151,155,157,157,159,158,158,158,158,157,156,157,158,158,158,0,255,0,
  32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,
  155,157,157,159,158,158,158,158,158,157,157,158,158,158,0,255,0,32,0,157,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,
  159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,158,159,158,158,
  158,158,157,158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,150,154,157,158,159,158,158,158,158,158,
  158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,151,155,157,158,159,158,158,158,158,158,158,158,158,
  158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,151,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,
  255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,151,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,
  156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,
  157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,157,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,159,
  158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,157,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,159,158,158,158,
  158,158,158,158,158,158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,151,155,157,157,159,158,158,158,158,158,158,
  158,158,158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,151,155,157,157,159,158,158,158,158,158,158,158,158,158,
  158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,151,155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,
  32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,
  155,157,157,159,158,158,158,158,158,158,158,158,158,158,0,255,0,32,0,155,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,157,
  159,158,158,158,158,157,157,158,158,158,158,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,151,155,157,158,159,158,158,
  158,159,158,156,157,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,151,155,157,157,159,158,158,158,159,159,
  157,156,158,158,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,151,155,157,157,159,158,158,158,159,159,158,156,157,
  158,158,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,151,155,157,158,159,158,158,158,158,159,159,158,155,158,158,0,
  255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,151,155,159,158,157,158,158,158,158,158,159,159,156,156,158,0,255,0,32,0,
  155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,
  159,159,157,155,157,158,158,158,159,159,158,155,157,0,255,0,32,0,155,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,150,155,159,159,158,
  157,156,158,158,158,158,159,159,157,155,0,255,0,32,0,156,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,151,155,159,159,158,158,156,157,
  158,158,158,158,159,158,156,0,255,0,32,0,157,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,150,155,158,159,159,158,157,155,157,158,158,
  158,159,159,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,150,155,157,158,159,159,158,156,156,158,158,158,158,158,
  159,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,149,154,157,157,159,159,159,158,156,156,158,158,158,158,158,0,255,0,
  32,0,157,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,244,150,
  155,157,157,159,158,159,159,156,157,158,158,158,158,158,0,255,0,32,0,157,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,151,152,157,157,157,
  159,158,158,158,157,158,158,158,158,158,158,0,255,0,32,0,155,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,150,155,157,157,157,158,158,158,
  158,158,158,158,158,158,158,158,0,255,0,32,0,155,158,246,244,244,245,245,246,
  246,246,246,246,246,245,153,152,154,157,157,157,157,158,158,158,158,158,158,
  158,158,158,158,158,0,255,0,32,0,155,158,158,157,157,157,157,157,158,157,158,
  158,157,156,155,156,157,157,157,157,157,157,157,158,159,158,158,158,158,158,
  158,158,0,255,0,32,0,156,157,157,157,158,158,158,158,157,157,157,157,157,157,
  157,157,158,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,0,
  255,0,32,0,155,157,157,157,156,155,155,157,158,157,157,157,157,157,156,156,
  157,158,158,157,157,157,157,157,156,156,157,158,158,157,157,158,0,255,0,32,0,
  155,157,157,154,153,154,153,154,157,158,157,157,157,155,153,153,153,153,157,
  158,157,157,157,154,153,154,153,153,156,158,157,158,0,255,0,32,0,155,157,154,
  152,155,158,158,157,155,158,158,157,154,153,155,157,158,157,155,157,158,157,
  154,152,156,158,158,156,155,157,158,158,0,255,0,32,0,155,157,152,154,154,151,
  155,31,157,155,158,156,152,156,154,151,155,31,157,156,158,157,151,156,154,
  151,155,30,157,156,159,158,0,255,0,32,0,155,156,152,156,152,154,30,111,31,
  155,158,156,152,157,151,154,30,111,30,156,158,156,151,158,152,154,30,111,30,
  156,158,158,0,255,0,32,0,155,156,152,157,156,30,111,111,31,155,158,156,152,
  157,156,30,111,111,30,156,158,156,151,158,155,30,111,111,30,155,158,158,0,
  255,0,32,0,155,157,153,156,31,111,111,31,157,155,158,156,153,156,159,111,111,
  31,157,156,158,157,153,156,30,111,111,31,157,156,158,158,0,255,0,32,0,156,
  157,155,155,157,30,30,157,155,157,158,157,155,154,157,30,30,157,155,157,158,
  157,156,155,157,30,30,157,156,158,158,157,0,255,0,32,0,155,157,157,156,155,
  155,155,155,157,158,157,157,157,156,155,155,155,155,157,158,157,157,157,156,
  155,155,155,156,158,158,157,158,0,255,0,32,0,156,157,157,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,158,157,157,157,157,157,157,157,157,158,
  158,157,156,157,158,0,255,0,32,0,155,157,157,157,157,157,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,157,157,157,157,156,157,157,157,156,157,
  158,158,0,255,0,32,0,155,157,157,157,157,157,157,157,157,156,157,157,157,157,
  157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,158,0,
  255,0,32,0,155,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,
  157,157,157,158,159,159,159,159,159,159,159,159,159,159,159,159,0,255,0,32,0,
  155,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,158,158,158,
  157,157,157,157,157,157,158,157,157,158,158,158,158,0,255,0,32,0,155,157,157,
  157,157,157,157,157,157,157,157,157,157,157,158,158,157,155,155,156,156,157,
  157,157,157,157,157,157,157,157,157,158,0,255,0,32,0,155,157,157,157,157,157,
  157,157,157,157,157,157,157,158,159,156,155,156,156,157,157,157,157,157,157,
  157,157,157,157,157,157,158,0,255,0,32,0,155,157,157,157,157,157,157,157,157,
  157,157,157,158,159,156,155,156,156,157,157,156,156,157,157,157,157,157,157,
  157,157,158,158,0,255,0,32,0,156,157,157,157,157,157,157,157,157,157,157,158,
  159,157,155,156,157,157,157,157,157,157,157,157,157,157,157,157,157,157,158,
  158,0,255,0,32,0,155,157,157,157,157,157,157,157,157,157,157,159,158,155,155,
  156,157,157,157,157,157,157,157,157,157,157,157,157,157,157,158,158,0,255,0,
  32,0,155,157,157,157,157,157,157,157,157,157,158,158,155,155,156,156,156,156,
  156,156,156,156,157,157,157,157,157,157,157,157,158,158,0,255,0,32,0,155,157,
  157,157,157,157,157,157,157,157,159,157,155,156,156,156,156,156,156,156,156,
  157,157,157,157,157,157,157,157,157,158,158,0,255,0,32,0,155,157,157,157,157,
  157,157,157,157,159,158,155,156,156,156,157,157,157,157,157,157,157,157,157,
  157,157,157,157,157,157,158,158,0,255,0,32,0,156,158,158,158,158,158,158,158,
  158,158,156,156,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,
  157,157,157,158,158,0,255,0,32,0,157,157,158,158,158,158,157,157,157,156,156,
  157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,
  158,158,0,255,0,32,0,156,157,157,158,158,158,158,157,156,156,157,157,158,157,
  157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,157,158,158,0,
  255,0,32,0,155,157,157,157,157,157,158,158,158,158,157,158,158,158,159,159,
  158,157,157,158,159,158,157,157,157,157,157,157,156,157,158,158,0,255,0,32,0,
  152,155,157,157,157,155,151,151,153,158,157,156,152,153,156,159,157,157,153,
  149,149,153,158,157,157,157,157,157,156,156,158,158,0,255,0,32,0,153,155,157,
  157,157,156,149,153,155,158,158,157,151,152,155,159,157,155,149,147,148,150,
  155,158,157,157,157,157,157,156,157,158,0,255,0,32,0,155,156,156,157,157,156,
  149,153,154,158,158,157,153,155,157,159,157,150,149,145,147,149,151,158,157,
  157,156,157,158,157,157,158,0,255,0,32,0,155,157,156,156,157,156,148,153,154,
  158,158,157,154,155,157,159,157,151,149,145,145,146,149,158,158,156,149,152,
  157,157,158,158,0,255,0,32,0,155,157,157,156,157,156,147,153,155,158,158,157,
  153,155,157,159,157,155,150,146,147,151,155,156,156,157,151,151,156,157,158,
  158,0,255,0,32,0,155,157,157,157,157,156,146,150,155,158,158,157,154,155,157,
  159,157,157,154,150,151,154,156,149,151,157,157,156,157,157,158,158,0,255,0,
  32,0,155,157,157,157,157,158,150,148,155,158,158,157,153,155,157,159,157,157,
  157,156,156,157,157,152,149,156,156,157,158,157,158,158,0,255,0,32,0,154,157,
  157,157,157,157,153,150,156,157,158,157,153,155,157,159,157,157,157,157,157,
  157,157,156,156,156,149,151,157,157,158,158,0,255,0,32,0,155,157,157,157,157,
  157,155,150,156,157,157,157,154,155,157,159,157,157,157,157,157,157,157,157,
  157,156,151,150,156,157,158,158,0,255,0,32,0,155,157,156,157,157,157,157,152,
  157,157,157,157,156,156,158,158,157,157,157,157,157,157,157,157,157,157,157,
  156,157,157,157,158,0,255,0,32,0,155,157,157,157,158,158,158,158,158,159,159,
  159,158,158,159,159,158,158,158,158,158,158,158,159,159,159,159,159,158,158,
  157,158,0,255,0,32,0,155,157,158,158,157,246,245,246,246,246,246,246,246,245,
  245,245,246,246,246,246,246,246,246,246,246,246,246,246,246,246,153,158,0,
  255,0,32,0,155,158,159,156,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,245,245,246,246,246,246,246,246,245,150,157,0,255,0,32,0,
  155,159,158,245,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,
  245,245,245,245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,156,159,156,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,
  245,246,246,246,246,246,246,246,150,156,0,255,0,32,0,156,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,
  246,246,246,246,246,150,156,0,255,0,32,0,156,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,
  246,246,150,157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,150,
  157,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,150,157,0,255,0,
  32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,
  245,245,245,245,245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,156,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,
  245,245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,
  246,246,246,246,246,246,150,157,0,255,0,32,0,155,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,
  246,246,246,150,157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,
  150,157,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,150,157,0,
  255,0,32,0,155,158,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,245,245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,
  155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,
  245,245,245,245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,156,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,
  245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,155,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,
  246,246,246,246,246,150,157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,
  246,246,150,157,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,150,
  157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,157,246,246,245,244,245,
  157,245,245,245,245,246,156,245,246,246,246,246,157,246,246,150,157,0,255,0,
  32,0,155,159,246,244,244,244,245,246,157,246,246,246,245,245,156,245,245,245,
  245,246,156,246,245,246,246,246,157,246,246,246,150,157,0,255,0,32,0,155,159,
  246,244,244,244,245,157,246,246,246,246,246,156,245,244,245,245,246,156,246,
  245,245,246,246,157,246,246,246,246,150,157,0,255,0,32,0,154,158,246,244,244,
  245,157,246,246,246,246,246,156,245,244,244,245,246,156,246,245,245,245,246,
  157,246,246,246,246,246,150,157,0,255,0,32,0,155,158,246,244,244,156,246,245,
  246,246,246,157,246,244,244,244,245,156,246,245,245,245,246,157,246,246,246,
  246,246,246,150,157,0,255,0,32,0,155,159,246,244,244,245,245,245,246,246,246,
  246,245,244,244,244,245,246,245,245,245,245,246,246,246,246,246,246,246,246,
  150,157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,150,157,0,
  255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,245,245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,
  156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,
  245,245,245,245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,156,159,246,
  244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,
  245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,156,159,246,244,244,244,
  245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,
  246,246,246,246,246,150,157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,
  246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,
  246,246,150,157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,
  245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,150,
  157,0,255,0,32,0,155,159,246,244,244,244,245,245,246,246,246,246,245,244,244,
  244,245,245,245,245,245,245,245,246,246,246,246,246,246,246,149,157,0,255,0,
  32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,
  245,245,245,245,245,246,246,246,246,246,246,246,150,156,0,255,0,32,0,156,159,
  246,244,244,244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,
  245,245,246,246,246,246,246,246,246,150,157,0,255,0,32,0,156,159,246,244,244,
  244,245,245,246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,
  246,246,246,246,246,245,149,157,0,255,0,32,0,156,159,246,244,244,244,245,245,
  246,246,246,246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,
  246,246,153,149,157,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,
  246,245,244,244,244,245,245,245,245,245,245,245,246,246,246,246,246,246,152,
  151,158,0,255,0,32,0,156,159,246,244,244,244,245,245,246,246,246,246,245,244,
  244,244,245,245,245,245,245,245,245,246,246,246,246,246,245,151,154,158,0,
  255,0,32,0,156,159,245,244,244,244,245,245,246,246,246,246,245,244,244,244,
  245,245,245,245,245,245,245,246,246,246,246,246,152,153,157,158,0,255,0,32,0,
  155,159,246,245,244,245,246,246,246,246,246,246,246,245,245,245,246,246,246,
  246,246,246,246,246,246,246,154,153,153,156,157,158,0,255,0,32,0,155,158,157,
  157,157,157,157,157,158,158,158,158,157,157,157,157,157,158,157,157,157,157,
  157,158,157,157,156,156,157,157,157,158,0,255,0,32,0,155,157,157,157,157,157,
  157,157,157,157,157,157,157,157,157,157,157,158,158,157,157,157,157,157,157,
  157,157,157,157,157,157,158,0,255,0,32,0,155,158,159,158,158,158,158,158,158,
  158,158,158,159,158,158,159,159,159,159,159,159,158,158,158,159,159,159,158,
  158,158,159,159,0,255,0,32,0,155,159,159,159,159,159,159,159,159,159,159,159,
  159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,159,
  159,0,255,
};

static const char m_nmare[] = {  // nightmare mode menu item  // 2023/12/02 version from HarmonyC.wad
  74,0,12,0,0,0,0,0,48,1,0,0,58,1,0,0,69,1,0,0,81,1,0,0,89,1,0,0,97,1,0,0,109,
  1,0,0,120,1,0,0,132,1,0,0,144,1,0,0,156,1,0,0,168,1,0,0,180,1,0,0,192,1,0,0,
  204,1,0,0,216,1,0,0,228,1,0,0,240,1,0,0,250,1,0,0,6,2,0,0,18,2,0,0,30,2,0,0,
  44,2,0,0,59,2,0,0,71,2,0,0,83,2,0,0,94,2,0,0,105,2,0,0,117,2,0,0,129,2,0,0,
  141,2,0,0,142,2,0,0,159,2,0,0,176,2,0,0,193,2,0,0,210,2,0,0,220,2,0,0,230,2,
  0,0,240,2,0,0,1,3,0,0,18,3,0,0,35,3,0,0,52,3,0,0,66,3,0,0,83,3,0,0,100,3,0,0,
  117,3,0,0,134,3,0,0,151,3,0,0,168,3,0,0,185,3,0,0,199,3,0,0,214,3,0,0,231,3,
  0,0,248,3,0,0,9,4,0,0,26,4,0,0,46,4,0,0,63,4,0,0,83,4,0,0,100,4,0,0,117,4,0,
  0,133,4,0,0,150,4,0,0,167,4,0,0,184,4,0,0,201,4,0,0,218,4,0,0,235,4,0,0,252,
  4,0,0,13,5,0,0,30,5,0,0,47,5,0,0,62,5,0,0,5,5,0,159,159,159,159,246,0,255,5,
  6,0,246,243,243,242,241,246,0,255,5,7,0,159,159,159,159,243,242,246,0,255,9,
  3,0,246,242,246,0,255,9,3,0,246,242,246,0,255,5,7,0,159,159,159,159,243,242,
  246,0,255,5,6,0,246,243,243,242,241,246,0,255,5,7,0,159,159,159,159,245,159,
  159,0,255,5,7,0,246,243,243,242,241,242,245,0,255,5,7,0,246,243,245,243,245,
  242,245,0,255,5,7,0,246,243,159,243,159,242,245,0,255,5,7,0,246,243,159,243,
  159,242,245,0,255,5,7,0,246,243,159,159,246,242,245,0,255,5,7,0,246,243,245,
  159,241,242,245,0,255,5,7,0,159,159,159,159,245,159,159,0,255,5,7,0,246,243,
  243,242,241,242,245,0,255,5,7,0,246,243,244,242,246,246,159,0,255,5,5,0,246,
  243,159,242,246,0,255,5,7,0,246,243,159,242,246,246,159,0,255,5,7,0,246,243,
  244,242,241,242,245,0,255,5,7,0,159,244,243,244,246,243,245,0,255,6,3,0,159,
  159,159,0,10,2,0,246,159,0,255,5,3,0,159,159,159,0,9,3,0,159,159,159,0,255,5,
  7,0,246,243,244,159,244,242,246,0,255,5,7,0,159,246,243,159,246,242,246,0,
  255,6,6,0,246,243,159,159,242,246,0,255,6,6,0,246,243,159,159,242,246,0,255,
  5,7,0,159,159,243,246,246,242,246,0,255,5,7,0,246,243,243,243,241,242,246,0,
  255,5,7,0,159,159,159,159,246,246,159,0,255,255,0,12,0,159,159,159,159,159,
  159,246,245,246,159,159,159,0,255,0,12,0,159,244,245,244,244,243,242,241,242,
  243,244,159,0,255,0,12,0,159,244,245,244,244,243,242,241,242,243,244,159,0,
  255,0,12,0,159,159,159,159,159,159,242,241,242,159,159,159,0,255,5,5,0,159,
  242,241,242,159,0,255,5,5,0,159,242,241,242,159,0,255,5,5,0,159,242,241,242,
  159,0,255,0,12,0,159,159,159,159,159,159,242,241,242,159,159,159,0,255,0,12,
  0,159,244,245,244,244,243,242,241,242,243,244,159,0,255,0,12,0,159,244,245,
  246,246,245,245,245,244,243,244,159,0,255,0,12,0,159,159,159,245,245,244,243,
  243,246,159,159,159,0,255,1,9,0,246,245,244,244,243,242,241,244,159,0,255,0,
  12,0,159,246,245,244,159,159,246,241,246,159,159,159,0,255,0,12,0,159,244,
  245,244,159,159,246,244,242,243,242,159,0,255,0,12,0,159,244,245,244,246,246,
  243,242,244,243,244,159,0,255,0,12,0,159,244,245,244,244,243,242,246,246,243,
  159,159,0,255,0,12,0,159,244,245,244,246,246,243,242,244,243,244,159,0,255,0,
  12,0,159,244,245,244,159,159,246,244,242,243,242,159,0,255,0,12,0,159,246,
  245,244,159,159,246,241,246,159,159,159,0,255,1,9,0,246,245,244,244,243,242,
  241,244,159,0,255,2,10,0,159,245,245,244,243,243,246,159,159,159,0,255,0,12,
  0,159,159,159,159,159,159,246,245,246,159,159,159,0,255,0,12,0,159,244,245,
  246,246,245,245,245,244,243,244,159,0,255,0,12,0,159,244,245,244,244,243,242,
  241,242,243,244,159,0,255,0,12,0,159,244,159,159,159,243,242,241,242,243,244,
  159,0,255,0,3,0,159,244,159,0,4,8,0,159,243,242,241,246,159,159,159,0,255,0,
  3,0,159,244,159,0,4,5,0,159,243,242,241,246,0,255,0,3,0,159,244,159,0,4,8,0,
  159,243,242,241,246,159,159,159,0,255,0,12,0,159,244,159,159,159,243,242,241,
  242,243,244,159,0,255,0,12,0,159,246,245,244,244,243,242,241,242,243,244,159,
  0,255,1,11,0,159,246,244,245,159,246,241,242,243,244,159,0,255,2,3,0,159,159,
  159,0,7,5,0,245,246,159,159,159,0,255,0,12,0,159,159,159,159,159,159,246,245,
  246,159,159,159,0,255,0,12,0,159,244,245,244,244,243,242,241,242,243,244,159,
  0,255,0,12,0,159,244,245,244,244,243,242,241,242,243,244,159,0,255,0,12,0,
  159,244,159,159,159,159,159,245,242,243,244,159,0,255,0,3,0,159,244,159,0,7,
  5,0,246,242,243,244,159,0,255,0,3,0,159,244,159,0,7,5,0,246,242,243,244,159,
  0,255,0,3,0,159,244,159,0,7,5,0,246,242,243,244,159,0,255,0,12,0,159,244,246,
  159,159,159,159,245,242,243,246,159,0,255,0,12,0,159,246,245,244,244,243,242,
  241,242,243,246,159,0,255,1,10,0,159,159,246,244,243,242,241,242,246,159,0,
  255,3,7,0,159,159,159,246,245,246,159,0,255,
};

static const char wiminus[] = {   // status bar widget minus sign  // 2023/12/02 version from HarmonyC.wad
  8,0,5,0,0,0,251,255,40,0,0,0,50,0,0,0,60,0,0,0,70,0,0,0,80,0,0,0,90,0,0,0,
  100,0,0,0,110,0,0,0,0,5,0,58,50,69,50,58,0,255,0,5,0,58,40,37,40,58,0,255,0,
  5,0,58,40,37,40,58,0,255,0,5,0,58,40,37,40,58,0,255,0,5,0,58,40,37,40,58,0,
  255,0,5,0,58,40,37,40,58,0,255,0,5,0,58,40,37,40,58,0,255,0,5,0,58,50,69,50,
  58,0,255,
};

static const char dsgetpow[] = {  // sound for getting powerup  // 2023/12/02 version from HarmonyC.wad
  3,0,68,172,249,43,0,0,125,125,125,125,123,125,125,125,128,125,125,125,125,
  125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,
  125,125,125,125,125,125,125,125,128,125,125,125,125,125,125,125,125,125,127,
  125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,123,125,125,125,
  125,128,125,125,128,125,125,125,125,125,125,125,125,128,125,125,125,125,127,
  123,125,125,125,125,128,125,123,125,125,125,125,125,127,128,125,125,125,125,
  125,125,125,125,125,125,125,125,125,125,125,128,125,128,125,125,125,125,125,
  125,128,128,125,125,128,125,125,125,125,125,127,123,127,125,125,128,125,125,
  125,125,125,125,125,125,125,128,125,125,125,125,128,125,125,125,125,127,125,
  125,125,125,125,125,125,125,125,125,125,125,128,123,125,125,125,125,125,125,
  125,125,125,125,125,125,125,125,125,127,125,125,125,125,128,125,125,125,125,
  125,125,125,125,125,128,125,125,125,125,125,125,125,125,125,125,125,125,125,
  125,125,125,125,125,125,125,123,127,125,125,128,125,125,125,125,125,125,125,
  123,125,125,123,125,125,125,128,125,125,125,128,125,125,125,125,125,125,125,
  125,125,125,125,125,125,125,128,125,125,125,127,125,125,125,125,125,125,128,
  125,125,125,125,125,125,125,127,125,125,125,125,125,125,125,125,125,125,125,
  125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,128,125,
  125,125,125,125,125,125,125,125,125,125,127,123,127,123,125,125,125,128,125,
  125,125,125,125,128,125,125,125,125,125,125,125,125,125,125,125,128,125,125,
  125,125,125,125,128,125,125,125,125,125,125,125,125,125,128,125,125,125,125,
  128,123,127,125,125,128,125,125,125,125,125,125,125,125,125,125,125,125,125,
  125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,
  123,127,123,125,128,125,125,125,125,125,125,125,128,125,128,125,125,125,125,
  127,123,127,125,125,125,125,128,125,125,125,125,127,125,128,125,125,128,125,
  127,125,127,125,125,125,123,127,125,125,125,123,125,125,125,125,127,125,127,
  125,125,125,125,128,125,125,125,125,125,125,125,125,125,125,125,125,125,125,
  125,125,127,125,125,125,125,127,125,125,125,125,125,125,128,125,125,125,125,
  125,125,125,125,127,125,128,125,125,127,125,127,125,125,123,128,125,125,125,
  125,127,125,125,125,125,125,125,125,128,125,125,125,125,128,125,125,125,125,
  125,125,125,128,125,125,127,125,127,125,125,125,125,125,125,125,125,125,125,
  128,125,125,125,125,127,125,125,125,125,128,125,125,127,125,125,125,125,125,
  125,125,125,125,125,125,125,125,125,125,125,127,125,125,128,123,128,123,127,
  125,125,125,125,127,125,125,125,125,128,125,128,125,125,128,125,128,125,125,
  125,125,127,123,128,125,125,125,125,128,127,125,127,125,125,125,125,127,125,
  125,125,125,125,125,128,125,125,127,125,125,125,125,125,125,128,125,125,125,
  125,125,125,125,125,125,125,125,125,125,125,125,125,127,125,125,125,125,125,
  125,125,125,125,125,125,128,125,125,128,123,127,125,125,125,125,128,125,125,
  127,125,125,125,125,125,128,125,125,125,125,128,125,125,125,125,128,123,128,
  125,125,125,125,127,123,125,125,127,125,125,125,125,125,125,125,125,127,125,
  125,125,125,125,125,125,128,125,125,125,123,128,125,125,125,125,127,128,125,
  128,125,125,125,128,125,125,125,125,125,125,127,125,127,125,125,125,125,128,
  125,123,125,125,125,125,125,125,125,125,125,125,125,128,125,125,128,125,125,
  125,125,125,127,125,125,125,125,125,125,125,125,125,125,125,125,125,128,125,
  125,125,125,128,125,128,125,128,125,125,125,125,125,123,127,123,127,125,125,
  125,125,128,125,125,125,125,125,125,125,125,125,128,125,125,127,125,125,125,
  125,125,125,127,125,127,128,125,125,125,125,125,125,125,125,128,125,127,125,
  125,125,125,125,125,128,125,128,125,125,125,125,128,125,125,125,125,125,128,
  125,128,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,128,
  125,125,125,125,125,125,125,125,127,125,125,125,125,123,125,125,125,125,125,
  125,125,125,125,125,125,125,125,127,128,125,125,125,125,125,125,125,125,125,
  125,125,125,125,125,128,125,125,125,128,125,125,125,123,127,123,127,123,128,
  125,125,125,125,125,125,125,125,127,123,128,123,128,125,125,125,125,125,125,
  128,125,125,125,125,125,125,125,125,125,125,125,125,125,125,125,128,125,125,
  125,125,127,125,125,125,125,125,127,125,125,125,125,128,125,128,125,125,127,
  125,125,125,125,125,125,125,125,128,125,125,125,125,125,128,127,125,125,125,
  125,125,125,125,125,125,127,125,125,125,125,127,125,127,125,128,125,128,125,
  127,125,127,125,127,125,123,125,123,125,123,125,123,125,123,125,123,125,123,
  125,123,125,123,125,128,125,128,125,128,125,127,125,127,125,127,125,127,128,
  125,128,125,128,125,128,125,123,125,123,125,123,125,123,125,125,123,125,123,
  125,123,125,123,125,123,125,123,125,127,125,127,128,125,127,125,127,125,128,
  125,128,125,128,125,127,125,127,125,125,123,125,123,125,123,125,123,125,123,
  125,125,123,125,125,125,125,123,125,123,125,128,125,128,125,128,125,128,125,
  128,125,127,128,127,128,125,128,125,125,128,125,128,125,123,125,123,125,123,
  125,123,125,123,125,125,123,125,123,125,123,125,123,125,128,125,128,125,127,
  125,127,128,125,128,125,127,125,127,125,128,127,125,127,125,127,125,127,125,
  123,125,123,125,123,125,123,125,123,125,123,125,123,125,123,125,123,125,128,
  125,128,125,128,125,128,130,128,127,125,128,125,128,125,127,125,127,125,125,
  123,125,123,125,123,121,123,123,121,123,125,123,125,123,125,123,125,123,125,
  127,125,125,128,125,128,125,127,128,125,128,125,127,125,127,125,127,125,127,
  125,123,125,123,125,123,125,123,125,123,121,123,123,121,123,121,123,125,125,
  127,128,125,128,125,128,130,130,127,125,128,125,128,127,125,128,125,128,125,
  125,127,128,125,128,125,123,123,125,125,123,125,123,121,121,123,125,125,123,
  121,123,123,121,123,125,123,127,125,127,125,128,130,127,128,130,128,130,127,
  130,128,125,128,127,125,128,125,123,125,123,125,125,123,125,123,121,123,121,
  123,123,125,123,125,125,125,127,125,127,125,128,127,125,127,125,127,125,128,
  125,127,125,128,125,128,125,127,125,128,125,123,125,123,123,125,123,123,121,
  123,121,123,121,123,121,123,125,123,125,128,130,130,127,130,127,130,127,130,
  130,128,125,128,127,130,127,130,127,128,125,123,125,125,125,123,121,123,121,
  123,123,121,123,123,123,121,123,121,123,125,123,125,127,125,127,125,130,128,
  130,128,130,128,130,130,128,130,128,130,125,128,125,123,123,121,123,125,123,
  121,123,121,123,121,123,121,123,125,123,125,123,125,125,128,125,125,127,125,
  127,125,128,130,128,130,127,125,127,130,130,128,125,128,125,128,125,125,123,
  125,123,125,123,125,123,121,121,119,121,121,123,121,123,125,128,125,128,125,
  128,130,132,130,127,130,127,125,125,127,130,127,128,127,125,128,125,128,125,
  123,123,125,123,123,121,123,123,123,121,121,123,121,123,121,123,121,123,123,
  125,128,125,128,127,125,128,130,128,132,130,132,130,127,128,130,132,130,128,
  125,125,123,123,121,123,121,123,123,121,123,121,123,121,123,123,121,123,123,
  121,123,125,125,127,125,127,130,130,128,130,128,130,128,130,128,125,127,130,
  127,130,130,128,125,128,125,125,121,123,121,119,121,119,121,119,121,121,121,
  123,123,123,125,128,130,132,132,130,127,130,132,130,125,128,130,128,130,127,
  125,127,130,127,128,125,123,125,123,125,123,121,121,119,121,123,121,121,119,
  116,119,121,119,121,121,123,123,125,125,127,128,130,132,132,134,132,130,132,
  134,134,132,134,132,132,130,128,125,123,125,121,116,119,121,119,116,114,116,
  119,121,119,121,121,123,125,125,123,125,128,130,132,130,132,130,130,127,130,
  132,134,132,130,127,128,130,130,127,130,128,125,123,123,121,119,121,119,116,
  116,119,116,119,121,123,123,125,127,128,130,130,132,132,130,127,130,127,130,
  127,130,127,130,127,130,127,130,128,125,123,123,125,125,121,121,119,121,121,
  119,119,116,119,121,119,121,123,125,125,123,125,128,127,130,130,132,134,132,
  130,132,130,132,134,134,132,130,130,125,128,125,123,121,119,121,119,116,116,
  114,119,119,119,119,121,123,121,123,121,123,125,127,128,125,128,130,130,132,
  132,130,132,130,132,130,132,130,132,132,134,132,130,128,125,123,125,123,123,
  121,119,116,116,114,112,114,114,116,119,121,123,125,130,132,130,132,134,132,
  134,130,127,128,130,128,128,130,128,130,130,132,128,128,123,123,125,123,121,
  119,119,121,119,121,121,123,121,121,123,121,123,123,121,121,123,125,125,128,
  125,128,132,134,132,130,132,130,132,132,132,130,132,130,130,132,130,128,125,
  123,121,116,116,119,116,116,119,116,116,119,121,121,121,123,125,125,127,128,
  127,130,130,134,132,136,134,134,136,134,132,132,130,127,128,125,128,125,128,
  125,123,123,125,123,121,119,119,116,119,114,116,114,116,119,116,121,123,128,
  132,130,130,127,130,132,130,128,130,132,132,130,130,130,132,130,132,130,132,
  130,128,125,123,125,121,116,116,119,121,119,116,116,119,121,121,123,125,125,
  125,123,123,125,125,128,125,130,130,132,130,130,132,130,130,132,130,128,130,
  128,130,130,128,125,123,123,125,123,121,119,121,116,119,116,119,121,119,119,
  121,123,125,123,123,121,125,125,128,130,132,134,136,134,132,132,134,132,132,
  132,130,132,130,128,130,128,128,121,119,116,114,114,116,114,116,116,114,114,
  119,119,121,123,127,132,130,128,125,128,130,132,132,130,132,132,130,130,127,
  128,130,128,125,125,125,127,125,128,130,132,130,125,121,121,119,121,119,116,
  119,116,119,116,119,116,119,116,119,119,121,123,125,130,132,136,139,141,143,
  139,139,136,139,134,128,125,123,123,121,121,116,119,121,119,116,116,119,121,
  119,121,123,123,121,121,123,125,130,128,125,123,123,125,127,130,127,130,127,
  130,127,128,130,130,134,136,132,130,132,130,132,130,130,128,127,128,123,123,
  116,114,112,110,112,114,114,112,110,112,114,121,125,130,134,132,134,134,132,
  134,136,134,132,132,130,128,128,125,123,125,128,125,130,127,128,125,128,130,
  130,128,125,123,119,116,119,116,121,121,123,121,119,116,116,119,121,123,123,
  121,123,123,125,130,130,132,132,136,141,139,136,134,132,130,130,128,125,128,
  125,128,123,123,121,119,119,116,114,116,114,116,121,125,125,125,127,130,128,
  130,128,125,125,123,123,127,130,127,130,132,134,136,134,132,134,134,130,125,
  123,125,127,128,125,123,121,121,119,116,114,114,112,114,116,119,119,116,119,
  119,125,132,130,132,130,132,134,134,132,130,132,134,132,130,128,127,130,132,
  130,130,127,125,123,121,119,121,123,123,121,119,119,121,121,123,121,125,125,
  123,119,114,114,116,119,121,121,123,125,125,128,130,132,134,136,134,134,132,
  130,125,125,127,130,132,134,136,136,132,125,121,116,112,114,116,116,119,123,
  125,128,125,127,125,123,121,123,121,123,125,128,130,136,141,141,141,136,134,
  130,125,116,116,119,121,125,130,134,139,136,134,130,128,123,116,112,112,116,
  121,125,130,132,134,134,132,125,121,116,112,114,114,119,123,127,132,134,134,
  132,132,130,127,125,123,123,121,119,119,121,123,125,128,125,128,132,132,130,
  127,127,125,123,121,119,121,121,121,116,116,114,116,114,112,114,116,119,121,
  123,127,130,130,128,130,128,127,130,132,130,130,128,128,125,128,125,128,130,
  132,134,132,130,128,125,123,121,119,121,123,121,119,119,121,121,123,119,114,
  116,116,114,116,121,127,134,136,139,136,139,136,132,132,134,136,139,136,134,
  134,130,128,125,127,125,125,123,121,123,123,125,128,130,128,130,128,125,121,
  119,121,116,119,119,123,123,125,123,125,128,125,128,130,128,125,123,123,125,
  128,130,128,125,128,125,128,125,125,127,130,127,130,127,125,125,121,116,114,
  116,114,114,112,110,112,114,116,114,114,119,121,125,125,127,125,128,125,128,
  130,132,130,132,132,134,132,130,127,125,123,123,125,123,123,121,123,123,125,
  123,121,121,123,125,128,130,132,130,128,125,123,121,123,123,128,130,130,132,
  127,123,119,114,116,121,125,130,132,134,141,141,139,141,139,136,134,128,123,
  121,123,127,128,132,134,134,132,130,125,125,123,121,123,121,125,125,128,125,
  125,127,128,125,123,121,119,116,119,116,119,121,123,121,123,123,127,125,125,
  123,123,121,123,125,128,130,132,134,134,130,125,121,116,114,110,108,105,108,
  110,112,110,112,114,119,121,123,125,125,128,127,130,132,128,132,132,134,134,
  132,132,130,132,134,136,141,141,139,136,134,127,121,114,112,116,121,127,130,
  132,134,132,130,123,116,114,112,110,112,114,112,114,119,123,130,134,136,136,
  134,134,132,132,134,134,134,136,139,141,143,143,141,141,139,134,130,125,123,
  123,125,125,128,125,121,119,114,114,112,114,119,119,121,123,125,123,123,121,
  119,116,119,121,119,116,116,119,119,123,121,123,121,119,119,116,114,112,110,
  108,110,116,119,121,121,119,121,125,125,128,125,132,134,136,136,139,141,145,
  147,145,145,141,139,143,147,150,147,143,136,134,132,125,119,114,108,110,108,
  103,105,110,108,110,112,114,112,108,105,105,108,112,121,128,139,143,145,147,
  145,145,143,145,141,136,136,139,141,145,147,147,147,145,141,134,125,116,110,
  110,112,114,119,125,127,130,125,121,119,114,110,103,101,103,108,112,112,116,
  121,123,121,114,112,112,110,112,110,112,112,114,116,119,121,123,121,123,125,
  123,125,123,119,121,125,136,145,154,158,163,165,167,163,156,150,141,139,136,
  134,134,136,139,139,141,136,132,119,105,97,90,86,90,94,97,97,101,105,108,110,
  116,121,125,127,130,132,134,136,139,141,150,152,154,154,152,147,143,136,130,
  125,125,128,127,130,132,132,134,132,125,116,110,105,101,101,103,112,119,121,
  116,112,116,114,119,119,119,121,123,123,121,119,123,125,130,136,141,145,143,
  139,136,134,132,132,132,132,134,136,139,141,136,134,136,141,141,134,132,127,
  123,119,116,112,110,112,114,112,114,114,112,110,99,92,92,97,103,105,110,119,
  128,132,130,128,127,125,119,116,119,121,123,121,123,125,127,130,128,127,130,
  127,130,128,127,125,127,132,136,141,143,145,147,145,147,147,143,141,143,145,
  145,143,143,143,141,134,128,130,130,127,125,121,112,112,108,105,103,101,103,
  105,110,112,114,121,128,132,134,136,139,134,125,121,119,121,121,119,121,125,
  128,128,125,125,121,121,119,119,121,123,121,125,128,130,128,123,119,112,108,
  105,99,90,88,86,90,94,101,108,116,128,132,136,134,136,143,154,163,169,176,
  180,185,189,187,183,178,174,167,158,147,136,130,123,119,114,108,101,92,86,81,
  79,81,79,81,83,88,90,94,97,103,112,121,130,134,136,139,141,143,150,152,154,
  156,154,154,152,154,154,152,147,143,139,132,125,121,114,110,110,108,105,103,
  99,97,99,97,94,92,94,94,97,99,103,110,119,128,136,141,147,150,147,147,145,
  147,150,152,154,156,156,161,158,161,158,154,150,141,134,123,114,108,103,101,
  105,103,99,94,88,86,83,81,79,79,83,94,105,114,123,132,139,143,145,147,150,
  150,150,150,152,152,154,154,156,154,150,141,132,123,116,108,103,101,97,99,
  101,103,105,110,114,116,119,116,114,116,116,121,125,132,139,141,145,143,147,
  145,147,145,145,143,141,143,139,139,143,141,143,141,139,136,130,125,123,121,
  116,114,108,101,92,88,86,90,92,97,101,108,116,123,128,125,123,121,123,121,
  125,130,134,139,145,143,139,132,123,114,108,103,99,92,88,90,94,99,103,105,
  112,123,132,134,136,141,150,158,167,174,176,183,185,187,189,189,183,174,163,
  152,141,132,121,108,99,92,94,97,94,97,92,92,92,90,88,88,90,94,99,103,112,119,
  125,130,134,139,141,139,136,139,143,150,154,152,150,152,154,152,145,136,127,
  123,116,105,99,92,88,86,88,94,99,101,101,103,110,121,130,141,150,161,172,180,
  185,189,192,194,187,180,174,163,154,139,123,105,94,79,66,52,46,44,39,37,35,
  37,35,39,44,52,61,72,86,97,114,132,147,158,167,176,185,194,200,205,209,218,
  225,227,225,220,214,207,200,187,176,165,147,128,110,97,83,68,55,44,35,30,28,
  26,22,26,35,46,63,81,101,121,141,161,174,192,203,211,218,220,222,220,214,200,
  189,176,165,150,132,114,101,90,79,70,59,52,46,41,41,44,48,52,64,70,75,81,86,
  88,92,101,112,123,132,143,156,169,180,189,196,203,207,209,207,200,198,200,
  198,198,187,178,169,158,147,134,119,105,94,81,72,59,50,44,41,39,39,41,46,50,
  59,68,79,97,112,128,143,156,172,183,187,194,198,196,196,192,189,183,174,163,
  154,145,136,125,116,108,103,99,94,88,81,79,81,83,83,86,88,92,99,108,114,123,
  132,136,145,152,158,163,172,176,180,183,180,178,174,161,150,139,127,114,99,
  79,63,52,46,41,39,37,37,41,50,59,68,79,90,108,123,139,156,169,183,192,203,
  211,216,220,218,216,211,205,198,187,178,165,156,145,130,114,101,86,72,61,52,
  46,44,44,46,50,59,68,79,90,101,114,127,136,147,156,165,176,185,192,194,196,
  198,194,189,178,167,154,139,123,110,99,90,83,77,72,66,61,57,52,48,46,48,52,
  57,63,72,81,99,119,136,154,167,178,187,196,205,209,214,216,211,207,203,194,
  187,176,165,154,141,121,105,92,81,75,72,75,72,75,77,83,88,94,101,103,103,105,
  110,114,121,125,130,134,139,145,145,150,150,150,152,152,150,150,152,154,158,
  156,154,150,145,139,132,125,114,108,99,90,83,79,72,68,63,57,52,50,50,48,52,
  57,59,61,68,77,90,105,125,147,178,207,229,245,251,255,255,255,255,255,249,
  240,229,214,198,180,167,152,132,112,90,68,48,28,13,8,6,8,13,17,24,28,33,41,
  48,57,70,81,90,101,112,123,139,154,169,178,185,189,196,198,200,198,194,192,
  187,185,178,172,163,156,150,143,134,125,116,114,112,108,103,99,101,99,99,97,
  94,92,94,97,94,97,99,99,103,108,110,112,116,116,119,119,121,119,119,116,119,
  121,123,125,128,134,141,145,150,152,156,163,167,172,176,178,178,178,176,174,
  169,161,150,139,125,112,99,88,77,68,59,55,50,46,48,50,52,57,64,72,81,92,103,
  112,123,132,141,150,158,165,169,174,178,183,185,187,187,189,187,189,187,183,
  180,178,174,169,163,154,143,132,123,110,94,79,68,55,48,41,33,28,26,28,30,37,
  46,55,66,79,92,108,119,134,141,150,156,163,165,167,165,163,167,174,183,189,
  198,205,211,216,211,205,194,185,174,163,147,132,119,110,103,94,86,77,63,50,
  37,26,19,17,19,26,35,48,66,86,105,121,136,150,161,174,185,192,196,203,207,
  211,218,216,214,209,200,189,176,165,154,141,130,119,110,103,97,90,83,79,70,
  61,52,46,44,37,37,35,37,41,48,50,52,55,52,55,63,83,112,147,183,218,255,255,
  255,255,255,255,255,255,255,242,216,196,178,158,132,103,70,28,0,0,0,0,0,0,0,
  0,0,15,39,63,88,114,136,161,178,194,214,233,247,255,255,255,255,249,236,218,
  196,176,154,139,128,119,114,112,114,119,125,125,123,116,110,101,90,83,77,68,
  63,57,55,57,55,57,55,55,57,59,61,63,66,68,72,77,81,88,108,141,178,218,251,
  255,255,255,255,255,255,255,240,214,185,154,134,116,103,86,68,41,10,0,0,0,0,
  0,0,0,11,44,79,108,130,147,163,174,185,192,198,203,214,220,227,231,229,222,
  209,194,178,161,141,123,112,110,114,114,116,119,125,134,139,134,121,108,94,
  83,72,61,52,48,46,48,52,55,57,59,59,66,72,77,81,88,94,101,110,116,128,152,
  192,227,251,255,255,255,255,255,255,244,220,196,169,141,121,105,94,77,57,33,
  6,0,0,0,0,0,0,0,19,55,88,119,141,156,172,183,192,198,203,207,214,222,225,225,
  222,216,205,194,178,161,141,125,119,112,114,112,116,116,123,127,125,119,108,
  97,83,72,61,55,46,44,48,52,57,61,68,72,75,81,86,88,94,97,99,101,110,121,141,
  174,211,247,255,255,255,255,255,255,238,216,198,178,152,125,108,99,88,66,39,
  11,0,0,0,0,0,0,0,17,48,81,110,127,145,161,178,191,198,203,205,207,216,222,
  220,218,211,200,194,185,172,156,139,125,121,119,121,119,116,116,121,121,116,
  110,97,86,77,66,55,44,35,37,44,55,63,72,79,86,94,99,101,99,94,88,88,92,105,
  132,169,207,240,255,255,255,255,255,255,240,227,214,196,172,145,125,112,94,
  72,39,11,0,0,0,0,0,0,0,0,33,68,99,119,139,156,178,194,205,207,205,205,209,
  209,211,209,203,196,187,176,163,147,132,121,119,116,112,108,103,108,116,123,
  125,123,116,114,110,105,94,79,75,75,79,83,88,90,90,94,101,108,108,105,101,92,
  86,77,70,70,81,112,152,185,211,231,249,255,255,244,222,209,211,209,194,165,
  134,112,103,92,79,59,35,19,6,4,4,4,8,13,30,59,92,119,141,161,176,189,194,196,
  192,189,191,196,196,192,187,176,167,156,145,134,121,112,108,105,101,94,88,86,
  92,103,116,123,128,130,130,128,119,105,97,97,101,112,119,125,127,125,123,125,
  128,130,128,125,123,119,112,99,86,77,72,81,99,125,154,169,178,180,183,187,
  183,172,165,169,178,180,172,150,128,112,105,101,97,88,79,70,66,61,59,61,63,
  70,86,103,125,143,150,154,152,154,158,161,161,156,158,156,156,152,141,125,
  116,112,110,108,108,105,110,112,116,116,116,119,132,150,158,165,163,154,143,
  134,127,121,121,123,130,134,132,125,119,110,103,103,103,105,110,108,103,94,
  90,86,88,90,94,92,92,94,108,130,145,154,156,158,172,183,187,183,176,176,180,
  180,172,156,147,141,139,136,130,116,101,86,79,77,77,75,70,72,79,92,99,103,
  101,105,112,125,130,130,128,130,132,134,136,134,136,141,143,145,147,145,147,
  150,154,161,158,152,147,147,152,154,152,143,136,134,132,125,116,108,105,103,
  101,99,92,88,83,83,81,83,86,86,88,88,92,94,94,97,99,99,101,103,112,128,152,
  176,194,200,207,214,214,209,203,200,198,194,183,167,150,139,127,116,110,94,
  79,61,48,41,39,37,35,37,44,61,81,97,103,105,110,121,132,141,145,147,154,161,
  169,176,178,178,176,172,172,169,165,161,154,152,147,141,134,130,128,132,130,
  125,116,112,108,101,97,86,79,77,81,88,86,83,81,83,92,105,116,123,128,130,130,
  132,130,130,132,132,136,141,141,141,134,132,139,152,158,161,158,161,169,174,
  169,156,145,147,154,152,141,123,110,103,101,99,94,83,75,66,66,68,75,77,75,75,
  83,97,110,119,123,134,145,156,163,165,165,169,172,174,178,176,172,165,156,
  147,143,134,127,121,116,112,108,103,101,103,108,110,110,108,110,112,114,112,
  110,110,112,114,116,119,116,119,116,119,119,121,123,125,128,134,136,136,139,
  136,134,132,134,130,123,119,116,121,130,136,139,141,143,147,150,147,143,139,
  141,143,141,136,130,125,121,121,119,116,105,94,86,83,86,86,88,86,90,94,105,
  114,121,127,132,141,150,156,158,158,156,158,158,161,161,156,150,141,134,128,
  121,116,114,114,112,105,103,101,103,108,114,119,121,123,123,125,123,125,123,
  121,123,128,134,139,139,136,139,141,145,145,141,134,130,130,125,121,116,105,
  101,94,86,77,70,66,70,88,105,121,132,141,156,172,183,183,180,180,189,189,185,
  169,152,141,134,130,125,112,97,79,66,59,59,61,61,59,68,83,101,114,121,128,
  134,147,154,158,158,158,156,156,158,154,154,145,139,134,128,125,123,123,123,
  125,123,121,116,121,125,134,136,134,132,130,128,125,119,110,103,97,94,94,97,
  94,94,97,105,114,125,134,136,139,141,147,152,156,158,161,161,163,163,165,163,
  156,150,139,128,121,114,103,92,86,79,79,79,77,70,68,70,83,99,110,119,121,125,
  136,152,165,174,176,176,174,176,174,169,165,156,147,141,134,130,119,112,105,
  101,97,92,90,86,88,90,92,97,99,103,108,114,116,119,121,125,130,136,141,143,
  141,139,143,147,152,147,145,143,143,145,143,139,134,132,134,136,139,136,132,
  128,125,121,119,116,112,108,108,105,103,103,105,103,105,110,112,114,121,125,
  130,130,132,130,132,136,139,136,132,125,121,114,108,103,101,99,103,110,121,
  136,147,158,165,169,176,176,172,163,152,143,134,125,121,116,112,108,103,99,
  94,92,86,75,66,61,64,68,81,97,110,125,143,156,167,174,176,172,169,167,165,
  163,156,152,147,145,143,139,134,130,123,116,108,101,97,94,94,97,99,108,114,
  121,125,130,130,128,125,125,127,130,132,130,132,132,134,139,141,139,134,132,
  128,123,119,112,105,97,92,86,81,79,77,79,86,99,112,130,150,167,180,187,192,
  194,192,187,180,172,165,158,154,147,141,132,119,103,86,70,55,46,41,39,41,52,
  61,77,94,110,123,136,145,150,156,161,165,167,169,172,174,174,176,172,165,156,
  147,139,130,123,114,108,105,103,110,114,116,116,114,112,114,114,110,105,103,
  101,101,99,99,97,97,99,103,112,116,123,127,132,139,150,161,169,178,185,189,
  192,192,185,178,167,152,139,123,105,92,79,70,68,63,59,57,59,61,64,72,83,92,
  99,103,112,130,147,167,180,189,196,203,203,198,189,178,169,158,147,136,127,
  121,112,103,99,94,90,83,75,72,75,79,83,86,90,99,110,121,127,134,139,141,143,
  143,141,143,141,141,145,150,150,154,150,150,147,145,143,136,132,130,127,125,
  123,121,123,121,119,116,110,108,105,101,99,94,92,94,97,99,105,112,119,128,
  132,139,143,147,150,152,154,154,156,158,158,156,154,150,143,139,132,121,116,
  110,103,99,90,83,79,83,88,92,99,103,112,123,132,136,141,143,145,150,147,150,
  147,147,145,145,143,141,136,132,125,119,116,112,112,110,110,112,112,116,121,
  128,130,132,132,130,130,127,123,121,119,119,116,119,116,116,119,116,119,121,
  121,123,123,125,128,130,132,134,136,139,139,141,143,141,141,134,132,130,127,
  125,123,121,123,125,128,130,132,136,134,136,134,132,132,127,125,123,121,116,
  112,108,103,101,97,92,86,77,72,70,70,75,83,101,121,141,161,178,192,200,205,
  200,198,194,189,185,176,165,154,141,130,119,105,90,77,63,55,52,50,50,52,57,
  66,77,94,114,130,143,150,154,158,161,163,163,161,163,163,167,165,163,154,143,
  136,132,128,125,123,123,119,116,119,116,119,116,114,114,116,116,114,112,108,
  101,97,94,94,97,101,103,108,114,119,123,128,132,139,145,147,156,165,174,180,
  183,178,172,163,158,152,141,132,121,114,103,92,81,77,75,77,79,81,88,92,94,94,
  99,108,119,128,136,141,141,143,145,147,150,154,152,147,145,150,154,158,158,
  150,141,136,134,134,125,121,116,112,110,108,110,108,103,97,92,92,99,101,99,
  97,97,101,110,119,125,132,139,143,145,150,154,156,154,150,147,150,147,152,
  150,145,141,134,130,130,127,123,116,110,105,105,108,105,108,105,105,110,112,
  116,119,119,116,119,119,121,125,125,128,130,130,132,134,139,136,139,136,139,
  136,139,141,143,141,141,136,134,130,125,123,114,112,105,99,97,94,92,97,101,
  103,110,116,123,130,134,139,143,145,147,152,150,152,150,152,150,147,141,136,
  132,125,121,116,112,108,105,101,99,101,103,105,108,112,114,119,121,123,123,
  125,127,128,132,134,136,139,136,139,136,136,139,136,134,132,130,132,130,132,
  130,132,132,130,132,134,132,132,127,125,123,121,119,116,114,112,110,110,112,
  114,112,114,116,116,119,123,125,128,130,132,132,134,130,127,125,116,110,103,
  97,94,92,99,108,119,132,143,154,161,163,161,158,156,154,156,156,154,150,145,
  141,136,130,119,108,94,86,81,79,79,77,75,75,79,90,101,114,125,134,143,147,
  154,158,158,156,158,156,158,161,163,158,152,143,132,128,121,119,114,112,110,
  108,110,108,110,108,110,110,114,119,123,125,121,116,112,110,110,112,114,116,
  121,125,127,130,132,132,130,132,134,136,143,150,152,154,156,158,154,150,145,
  143,139,136,132,125,116,108,101,92,86,83,86,90,94,94,97,99,105,112,119,125,
  132,136,145,150,154,156,154,150,145,141,136,132,128,125,128,125,123,121,123,
  125,127,128,125,128,130,132,130,123,119,116,114,114,112,114,114,112,110,105,
  105,110,112,110,110,114,123,132,139,139,141,143,147,150,150,152,150,152,147,
  145,143,141,136,130,123,119,116,114,110,105,103,101,99,99,101,103,108,110,
  112,116,121,125,130,130,132,134,136,139,139,141,143,143,141,141,139,136,136,
  134,132,132,130,130,125,125,123,123,121,119,119,116,116,114,112,112,108,110,
  108,110,112,114,114,116,119,121,121,123,125,127,132,136,141,143,150,152,154,
  156,152,150,145,141,136,132,125,119,114,112,110,105,105,101,101,99,101,103,
  105,108,110,112,116,121,127,134,136,141,141,143,145,143,145,143,139,136,134,
  136,134,132,130,127,125,123,121,121,119,116,114,114,112,114,116,119,116,119,
  116,121,123,121,123,125,123,125,128,130,130,132,134,134,136,134,136,136,134,
  134,132,132,134,134,132,130,130,128,125,125,123,121,116,114,112,108,108,105,
  108,105,103,101,97,94,94,92,99,108,119,132,143,152,161,167,169,172,169,172,
  169,167,165,158,150,139,130,121,116,108,99,90,83,81,79,79,77,75,77,83,97,110,
  123,132,139,143,145,150,152,154,152,152,154,158,156,154,150,143,134,130,125,
  123,123,121,116,114,112,112,114,112,114,112,114,116,121,125,123,119,116,112,
  110,112,114,116,121,123,128,130,130,127,128,130,132,139,143,147,150,147,147,
  145,145,143,143,141,134,128,123,116,112,105,103,97,97,94,97,99,105,108,110,
  112,114,119,123,128,132,141,145,150,152,152,150,147,147,143,139,136,134,128,
  125,121,119,114,110,110,108,110,112,114,112,114,114,116,119,121,123,128,132,
  134,136,134,136,134,132,132,130,130,127,130,127,128,125,123,123,121,121,123,
  121,123,121,123,125,128,125,127,128,125,128,130,128,130,127,125,125,123,121,
  123,121,121,123,125,123,128,125,123,125,123,125,128,127,130,130,130,128,128,
  125,128,127,125,123,125,123,125,123,123,121,123,121,123,125,125,127,125,125,
  123,125,123,125,125,128,125,128,130,128,130,128,125,125,127,125,127,128,125,
  128,125,128,127,125,123,123,121,123,121,123,121,123,121,119,116,114,112,114,
  116,119,121,128,132,134,136,139,141,139,141,139,139,136,132,130,128,127,125,
  128,123,119,116,114,114,112,110,112,110,112,114,119,123,125,128,127,130,132,
  132,134,132,134,132,132,134,132,134,132,130,130,127,125,125,123,125,121,121,
  119,121,123,121,123,121,123,121,123,121,123,121,119,121,119,119,121,123,123,
  125,125,130,132,134,136,134,136,134,136,134,136,134,132,130,128,125,123,121,
  121,119,116,114,116,116,119,119,121,121,123,125,123,125,128,125,128,125,123,
  121,121,119,116,114,112,108,110,114,121,128,130,134,139,145,147,150,147,147,
  150,147,145,143,136,130,127,121,119,114,110,103,97,94,94,94,97,99,99,103,112,
  119,127,136,141,143,145,150,152,154,154,152,150,147,143,141,136,132,127,123,
  116,114,112,110,112,110,108,110,112,112,114,116,119,123,125,128,130,125,123,
  125,123,125,125,128,127,125,128,125,127,130,132,130,132,132,130,132,134,136,
  139,139,134,134,132,132,130,128,123,121,116,112,110,108,105,108,110,110,112,
  114,116,119,123,125,132,132,134,139,141,145,143,145,143,141,139,134,132,130,
  127,123,121,119,116,114,114,110,112,112,114,116,114,116,119,121,121,125,130,
  130,132,130,132,130,132,130,132,130,128,125,125,123,125,123,125,123,123,125,
  127,130,132,130,130,132,130,132,130,130,127,125,125,123,123,121,121,116,114,
  116,114,116,119,116,119,119,121,123,127,128,130,132,132,134,134,136,134,134,
  132,130,132,130,128,127,125,125,123,121,121,119,121,121,119,121,123,121,123,
  121,121,123,125,125,128,125,125,128,125,123,125,123,125,128,127,125,127,130,
  130,132,130,132,130,132,130,132,130,132,130,128,130,128,127,125,123,121,121,
  119,119,116,119,116,119,116,119,119,121,123,125,125,127,130,127,130,132,130,
  132,130,132,130,130,127,125,123,121,121,119,116,114,116,119,123,123,125,130,
  132,134,132,134,132,134,134,132,132,130,128,127,125,123,125,121,116,116,114,
  116,116,116,114,114,116,119,123,127,128,130,127,130,130,132,132,130,132,130,
  132,130,132,130,128,127,125,123,125,123,125,123,125,123,121,123,121,123,121,
  123,125,127,125,128,125,123,121,123,123,125,125,123,125,123,125,125,128,125,
  128,125,128,130,128,130,132,130,130,127,130,127,130,127,130,127,125,125,123,
  123,121,119,121,119,121,119,121,119,121,119,121,123,121,125,125,128,130,127,
  128,130,128,130,132,132,130,132,130,127,128,130,128,125,128,125,125,127,125,
  123,125,123,125,125,123,121,121,123,121,119,119,121,119,121,123,121,123,125,
  123,125,128,130,128,130,128,130,132,130,132,132,130,132,130,128,130,128,130,
  127,125,125,125,125,123,123,121,123,123,121,119,121,119,121,121,119,121,121,
  123,121,123,123,125,128,127,130,127,132,130,132,134,132,132,134,132,130,132,
  132,128,127,125,123,123,121,119,121,119,116,116,119,116,119,121,121,119,121,
  123,125,125,128,128,130,130,132,130,132,130,132,130,132,130,132,130,128,130,
  128,130,127,130,127,125,127,125,123,121,123,121,119,121,119,121,119,121,119,
  121,119,121,123,121,123,125,125,127,130,127,130,132,130,132,134,134,132,130,
  132,132,130,128,128,125,125,123,121,121,119,116,116,114,112,110,114,116,119,
  121,123,127,130,132,132,134,136,136,134,136,134,132,132,130,132,130,128,125,
  123,119,121,119,119,116,114,116,114,116,119,121,125,125,127,128,130,132,130,
  132,132,130,132,130,128,125,128,125,123,123,125,123,125,123,123,123,125,123,
  125,127,125,128,130,127,128,130,128,127,125,123,125,123,121,121,119,121,119,
  121,121,123,123,125,127,128,130,132,130,132,130,132,134,132,132,130,128,127,
  125,125,123,123,119,121,121,119,116,119,116,119,121,119,121,125,128,127,130,
  127,130,132,132,130,132,132,132,130,130,128,130,128,125,123,125,123,125,123,
  123,121,123,121,123,121,123,121,123,121,123,121,123,123,121,123,121,123,121,
  123,125,125,123,125,125,128,130,127,130,132,130,132,130,132,130,132,130,128,
  130,128,125,125,123,123,121,121,119,121,119,121,119,121,119,121,121,123,123,
  125,125,128,130,128,130,128,130,130,128,130,127,130,127,130,127,125,128,125,
  128,128,125,123,125,123,125,123,125,123,125,123,125,125,123,125,123,125,123,
  125,125,123,125,123,125,128,125,128,125,125,127,125,128,130,128,130,128,130,
  128,130,128,130,128,125,128,125,123,125,123,121,121,123,121,123,121,123,121,
  123,123,125,123,125,128,125,128,125,128,125,128,125,123,125,123,125,125,123,
  125,128,125,128,125,127,128,130,130,128,130,127,125,127,125,127,125,125,123,
  125,123,125,123,125,123,125,123,125,123,125,123,125,123,125,123,125,125,123,
  121,123,121,121,119,119,116,114,114,116,119,121,125,130,132,136,139,141,139,
  139,141,141,139,139,134,134,132,130,127,125,121,116,114,112,110,110,105,105,
  108,110,112,116,116,119,123,125,130,134,136,134,136,139,139,136,139,136,134,
  132,132,130,130,128,125,123,123,121,119,121,119,121,119,121,123,121,123,121,
  123,125,125,123,121,123,121,123,123,121,123,125,125,128,125,127,128,130,132,
  130,132,134,134,132,130,132,130,130,128,125,125,123,121,119,119,116,114,114,
  114,116,116,119,119,121,123,125,127,130,130,132,132,134,136,134,132,134,134,
  132,130,130,127,125,123,123,121,119,119,116,119,116,119,116,119,119,121,123,
  121,123,125,125,128,125,127,128,130,128,130,128,127,130,128,128,130,127,125,
  127,125,125,123,125,123,121,123,123,121,123,121,123,121,123,121,123,123,121,
  123,125,123,125,123,125,128,125,128,130,128,130,128,128,130,132,128,130,130,
  127,130,128,125,125,123,125,123,123,121,123,121,123,121,123,121,123,121,123,
  123,125,123,125,127,125,127,125,128,130,128,130,128,130,130,128,130,127,125,
  127,125,127,125,128,125,123,125,125,123,125,123,121,123,121,123,123,121,123,
  125,123,125,125,123,125,128,125,125,128,130,128,130,130,127,130,127,130,127,
  130,128,127,125,123,123,125,123,121,123,121,121,123,121,123,121,123,125,123,
  125,127,125,127,125,125,127,130,127,125,127,128,125,128,125,128,125,123,125,
  123,123,125,123,125,123,125,123,125,123,125,123,125,123,125,125,123,125,123,
  125,123,125,123,125,123,125,123,123,125,127,125,127,125,127,125,128,125,128,
  128,125,128,125,128,125,123,125,123,125,123,125,125,123,125,123,125,123,125,
  128,125,127,125,128,127,125,127,125,127,125,127,125,128,125,123,125,128,125,
  123,123,125,123,125,123,125,125,123,125,123,125,123,125,123,125,128,125,128,
  125,125,127,130,128,130,128,130,128,130,127,128,130,127,125,127,125,123,125,
  123,125,123,125,123,121,125,123,125,123,125,123,125,123,123,125,127,125,128,
  125,127,130,130,128,125,127,125,128,125,128,125,128,125,127,125,127,125,127,
  125,123,125,125,123,121,123,121,123,123,121,123,125,123,125,123,125,123,125,
  123,125,123,125,128,125,125,127,130,127,130,127,130,130,128,125,128,125,128,
  125,125,123,121,123,121,119,121,121,119,121,119,121,123,123,125,123,125,127,
  130,127,130,127,130,127,125,127,125,127,128,125,123,125,123,125,125,123,125,
  123,125,127,125,128,125,127,125,127,125,127,125,123,121,123,121,123,121,123,
  121,123,125,123,125,123,125,127,128,130,128,130,128,130,128,130,127,130,127,
  125,125,123,125,123,125,123,125,123,125,123,125,125,128,125,128,125,128,125,
  128,125,128,125,128,125,128,128,125,128,125,127,125,123,125,125,123,125,127,
  128,125,128,130,128,127,125,127,125,127,125,123,125,125,123,125,123,121,123,
  123,125,123,125,127,125,127,125,128,125,128,125,127,125,128,128,125,128,125,
  128,125,123,125,123,121,123,121,123,125,123,125,123,125,125,125,128,125,128,
  125,127,125,128,125,128,125,128,125,123,123,125,123,125,123,125,125,123,125,
  123,125,123,125,125,125,125,123,125,123,125,123,128,125,128,128,125,125,123,
  123,121,121,119,121,119,121,123,123,125,123,125,127,128,125,128,130,128,130,
  130,130,130,130,128,130,127,128,125,123,125,123,125,125,123,125,123,125,123,
  125,123,125,127,125,128,125,128,127,125,123,125,123,125,123,125,125,123,125,
  123,125,128,127,125,128,130,128,130,130,127,130,132,130,130,128,125,128,125,
  123,125,123,125,123,121,121,123,125,123,123,125,123,125,127,125,127,125,130,
  128,130,127,130,127,125,127,125,128,125,128,125,128,127,125,127,125,127,125,
  123,125,123,127,125,127,125,128,125,127,125,123,125,123,121,123,121,123,125,
  123,125,125,128,130,128,130,127,128,130,128,125,128,127,125,123,125,123,125,
  123,121,121,123,121,123,125,123,125,123,125,123,125,123,125,123,125,127,125,
  127,128,125,127,125,127,125,127,125,123,125,123,125,125,123,125,123,121,123,
  123,123,121,123,121,123,121,119,121,119,121,123,121,123,123,125,128,130,132,
  130,132,130,130,128,125,125,123,125,125,125,127,130,127,128,130,127,125,125,
  123,123,121,123,121,123,121,123,121,123,125,123,125,123,125,123,125,123,125,
  123,125,128,125,128,130,128,130,127,130,132,130,132,130,127,130,130,128,125,
  127,125,127,125,123,123,121,123,121,123,121,123,123,121,123,125,123,125,127,
  125,127,130,130,128,130,130,128,130,127,130,127,130,127,130,127,125,125,128,
  125,123,125,123,125,125,123,125,123,125,123,125,123,125,123,123,121,123,123,
  125,123,125,128,125,128,130,128,130,128,127,130,128,125,128,125,123,125,123,
  125,123,125,125,123,125,125,123,125,123,121,123,121,123,125,123,125,123,125,
  123,125,123,125,123,125,127,125,128,127,125,127,125,127,125,127,125,128,125,
  123,125,123,125,123,121,123,121,123,121,123,121,123,121,123,121,121,123,125,
  123,125,123,125,128,125,127,125,128,125,128,125,128,125,128,125,125,127,128,
  125,128,128,125,123,125,123,125,123,125,128,125,128,125,125,127,125,128,125,
  123,125,123,125,123,125,123,125,125,127,125,128,127,125,128,127,125,127,125,
  127,128,125,128,125,127,125,127,125,123,125,125,127,125,127,128,127,128,127,
  125,127,125,127,125,128,125,123,123,125,128,125,128,125,128,125,125,127,125,
  127,125,128,125,128,125,128,125,125,123,125,123,125,123,125,123,125,123,125,
  125,125,123,125,128,125,125,127,125,128,125,125,123,125,123,125,123,125,123,
  125,123,123,125,123,125,123,125,127,125,127,125,127,125,127,125,127,125,123,
  125,123,125,123,125,125,123,125,123,125,123,125,128,125,125,127,125,123,125,
  123,125,123,125,123,125,123,125,123,125,123,125,123,125,127,128,125,123,125,
  123,123,125,123,125,123,125,128,125,128,125,127,125,123,125,123,125,123,125,
  123,125,123,125,123,125,123,125,123,125,125,123,125,123,125,127,125,128,127,
  130,127,130,127,130,128,125,128,125,125,125,123,125,123,125,123,121,123,121,
  121,123,125,125,123,125,128,125,128,125,127,125,127,125,127,125,128,125,127,
  125,127,125,123,125,128,125,127,125,127,125,127,125,127,125,128,125,128,125,
  123,125,123,125,125,125,123,125,123,125,123,125,128,125,128,125,128,128,125,
  127,130,125,128,125,127,125,125,128,125,128,125,123,125,123,125,123,125,123,
  125,123,125,123,125,123,125,123,125,123,125,123,123,125,123,125,128,125,128,
  125,128,125,128,127,127,125,127,125,127,125,127,125,128,125,123,125,125,123,
  125,123,125,123,123,125,123,125,123,125,123,125,123,125,123,125,127,125,127,
  125,127,125,128,125,128,125,128,125,128,125,123,125,127,125,127,125,123,125,
  125,123,125,123,125,123,125,123,125,123,125,123,125,123,125,123,125,123,125,
  123,125,123,125,128,125,128,125,128,125,127,125,125,123,125,128,125,128,125,
  128,125,123,125,123,125,125,123,125,123,125,123,125,123,125,123,125,128,125,
  128,125,128,127,125,128,125,123,125,123,125,125,127,125,123,125,125,123,125,
  123,125,128,127,125,127,125,127,125,123,125,125,123,125,125,123,125,123,125,
  125,125,125,125,128,125,125,123,125,127,128,125,123,123,125,128,125,125,123,
  125,127,125,127,128,127,125,127,125,127,125,128,125,128,127,125,128,125,123,
  125,123,125,123,125,123,125,123,125,123,125,125,128,125,127,125,128,127,125,
  128,125,128,125,127,125,127,125,123,125,123,125,123,125,125,123,125,123,125,
  123,125,123,125,127,125,127,125,128,125,128,125,127,125,127,125,123,125,123,
  125,123,123,125,123,125,128,125,128,125,128,125,128,125,125,127,125,127,125,
  123,125,123,125,123,125,125,123,125,123,125,123,125,123,125,123,125,123,125,
  127,125,127,125,128,125,127,128,125,128,125,128,125,128,128,127,128,127,128,
  125,128,125,128,125,123,125,123,125,123,125,125,123,125,123,125,123,125,123,
  125,123,125,123,125,127,125,127,128,125,125,128,125,128,125,127,125,123,125,
  123,125,125,123,125,128,123,125,123,125,125,123,125,123,125,127,125,128,125,
  125,127,125,128,125,123,125,123,125,123,125,123,125,125,123,125,123,125,123,
  125,123,125,128,125,125,128,125,128,125,127,125,127,125,123,125,123,125,123,
  125,123,125,123,125,123,125,127,128,125,128,125,128,125,127,125,127,125,127,
  125,128,125,127,125,123,125,123,125,123,125,123,125,123,125,123,125,123,123,
  125,125,123,125,127,128,125,128,125,128,125,128,125,128,125,127,125,127,125,
  127,125,128,128,125,128,125,128,125,127,125,127,125,127,125,127,125,127,125,
  127,125,123,125,123,125,125,123,125,125,123,125,123,125,123,125,123,125,123,
  125,127,125,127,125,128,125,128,125,128,125,128,125,128,125,128,125,128,125,
  123,125,123,125,123,125,123,125,125,125,123,125,123,125,127,125,125,125,123,
  125,127,125,127,125,127,125,127,125,128,125,128,125,123,125,123,125,123,125,
  123,125,123,125,123,125,125,125,125,125,125,123,125,123,125,123,125,123,125,
  123,125,123,125,123,125,125,125,128,125,125,128,125,123,125,125,123,125,123,
  125,128,125,128,125,128,125,127,125,127,125,123,125,123,125,125,123,125,123,
  125,123,125,123,125,123,125,125,123,125,128,125,128,125,128,125,128,125,128,
  125,128,128,125,123,125,123,123,125,123,125,128,125,128,125,125,128,125,123,
  125,125,123,125,128,125,125,123,125,123,125,123,125,123,125,128,125,127,125,
  127,125,128,125,128,125,128,125,128,125,127,125,123,125,125,125,123,125,123,
  125,123,125,123,125,123,125,128,125,128,125,128,128,125,128,125,128,128,125,
  128,125,127,125,127,125,127,125,127,125,123,123,125,128,125,127,125,127,125,
  127,125,123,125,123,125,123,125,123,125,125,123,125,123,125,123,125,128,125,
  128,125,127,128,125,127,125,127,125,127,125,128,125,128,125,128,125,123,125,
  125,123,125,123,125,123,123,125,123,125,123,125,127,125,125,128,125,128,125,
  128,125,128,125,125,128,125,128,125,128,125,128,125,128,125,125,125,125,125,
  123,125,123,125,123,125,123,125,123,125,123,125,127,125,128,125,128,125,127,
  125,127,125,128,125,128,125,128,125,127,125,125,128,125,127,125,128,125,128,
  125,128,125,123,125,123,125,125,123,123,125,123,125,123,125,123,125,123,125,
  128,125,128,125,128,125,127,125,127,125,127,125,128,125,128,125,128,125,128,
  125,128,125,128,125,128,125,123,127,125,125,125,125,125,125,127,125,127,125,
  127,125,127,125,127,125,127,125,128,125,128,125,128,125,128,125,128,128,125,
  127,125,127,125,127,125,128,125,128,125,123,123,125,123,125,123,125,125,125,
  125,123,125,125,127,125,127,128,125,128,125,127,125,127,125,127,125,127,125,
  128,125,128,125,123,125,123,125,123,125,125,123,125,123,125,123,125,123,125,
  123,125,128,127,125,127,125,127,125,128,125,128,127,125,127,125,127,125,128,
  125,128,125,128,125,127,125,125,125,123,125,123,123,125,123,125,123,125,123,
  125,123,125,123,125,123,125,125,125,125,127,125,128,125,128,125,128,125,128,
  125,128,125,125,123,125,127,125,127,125,128,125,128,125,128,125,123,125,123,
  125,123,125,123,125,125,125,123,125,125,123,125,123,125,123,125,123,125,123,
  125,128,125,128,125,128,125,128,125,128,125,128,125,128,125,125,127,125,123,
  125,123,125,125,123,125,123,125,123,125,123,125,123,125,125,125,125,125,125,
  125,128,125,125,125,128,125,125,125,125,125,123,125,123,123,125,125,128,125,
  127,125,127,125,123,125,125,125,125,125,125,128,125,125,128,125,125,125,123,
  123,125,127,125,127,125,125,128,125,125,127,125,125,123,125,125,128,125,123,
  123,125,125,123,127,128,125,123,125,127,125,123,125,123,125,123,125,128,125,
  123,125,127,125,125,125,128,125,125,125,127,125,128,125,128,125,127,125,125,
  123,125,123,125,125,123,125,123,125,125,127,123,125,125,125,125,125,128,125,
  127,125,125,125,128,125,123,125,123,123,125,125,123,125,123,125,123,125,128,
  127,125,127,125,128,125,128,125,128,125,128,125,125,125,123,125,123,125,123,
  125,123,125,125,123,125,123,125,128,125,123,125,127,125,127,125,128,125,128,
  125,128,125,128,125,128,125,125,125,123,125,123,125,125,125,125,123,125,125,
  125,128,127,125,125,125,123,125,125,123,125,123,125,123,125,123,125,125,125,
  125,125,127,125,127,125,127,125,127,125,125,125,125,125,123,125,123,125,125,
  125,125,125,125,125,128,128,125,128,125,127,125,128,125,128,125,127,125,128,
  125,128,125,128,125,128,125,123,125,123,125,128,125,123,125,123,125,125,125,
  125,125,125,125,128,125,127,128,125,128,125,128,125,127,125,128,125,128,125,
  128,125,128,125,125,128,125,128,125,125,127,125,127,128,125,125,127,125,128,
  125,125,127,125,125,128,125,123,125,125,125,125,125,128,125,125,123,128,125,
  125,127,128,125,125,127,125,127,125,125,125,128,125,125,127,125,125,128,125,
  125,125,125,123,123,125,125,125,128,125,128,125,127,125,125,128,125,128,125,
  127,125,125,123,125,123,125,125,125,125,123,125,125,123,125,125,127,123,127,
  125,125,127,125,128,125,127,125,127,125,127,125,128,125,128,125,128,127,125,
  125,125,125,123,125,125,125,125,123,128,123,125,123,125,123,125,125,125,127,
  123,125,125,128,125,125,125,125,125,125,125,123,125,123,123,125,123,125,123,
  125,123,125,128,125,125,125,123,125,123,125,125,123,125,125,125,125,123,125,
  123,125,123,125,128,125,127,125,123,125,125,125,125,123,125,123,125,123,125,
  125,123,125,125,125,123,125,123,125,123,125,123,123,125,125,125,125,123,125,
  125,123,125,125,123,125,123,127,123,125,123,125,123,125,123,125,125,123,125,
  123,125,123,125,123,125,125,125,128,125,125,125,125,125,123,125,125,123,125,
  123,125,123,125,123,125,123,125,123,125,123,125,123,125,125,125,123,125,123,
  125,125,123,125,125,125,125,123,125,123,123,125,123,125,125,123,123,125,125,
  125,125,125,123,125,125,125,125,123,125,125,127,125,123,125,123,125,123,125,
  123,125,125,125,123,123,125,123,125,123,125,123,128,123,125,125,125,125,125,
  128,127,125,125,127,125,125,125,123,125,125,125,125,123,125,123,123,125,123,
  125,123,125,125,123,125,127,125,125,127,125,127,125,128,125,128,125,125,127,
  128,125,128,125,125,125,123,125,123,125,125,125,125,125,125,125,125,125,125,
  125,125,125,127,123,127,125,128,125,127,125,128,125,125,127,125,125,128,125,
  125,125,125,128,125,128,125,128,125,125,127,125,127,125,128,125,128,125,128,
  125,128,125,125,123,127,125,125,125,125,127,125,125,125,125,125,128,127,125,
  128,125,128,125,128,125,128,125,127,125,127,125,127,125,128,125,128,125,128,
  125,125,128,125,128,125,128,125,128,125,128,125,128,125,127,125,128,128,125,
  127,125,128,125,128,125,128,125,128,125,127,125,127,125,125,128,125,128,125,
  128,125,125,127,125,127,125,128,125,128,125,128,125,127,125,125,128,125,127,
  125,127,125,128,127,125,127,125,128,125,128,125,128,125,
};

static const char mbftext[] = { // 2023/12/03 Harmony font version
  242,0,14,0,0,0,0,0,208,3,0,0,209,3,0,0,210,3,0,0,211,3,0,0,212,3,0,0,213,3,0,
  0,214,3,0,0,215,3,0,0,216,3,0,0,217,3,0,0,218,3,0,0,219,3,0,0,220,3,0,0,221,
  3,0,0,222,3,0,0,223,3,0,0,224,3,0,0,225,3,0,0,226,3,0,0,243,3,0,0,4,4,0,0,21,
  4,0,0,38,4,0,0,54,4,0,0,64,4,0,0,74,4,0,0,84,4,0,0,94,4,0,0,110,4,0,0,127,4,
  0,0,144,4,0,0,161,4,0,0,178,4,0,0,193,4,0,0,210,4,0,0,227,4,0,0,244,4,0,0,5,
  5,0,0,22,5,0,0,39,5,0,0,56,5,0,0,73,5,0,0,90,5,0,0,107,5,0,0,122,5,0,0,139,5,
  0,0,156,5,0,0,173,5,0,0,190,5,0,0,210,5,0,0,227,5,0,0,244,5,0,0,8,6,0,0,25,6,
  0,0,42,6,0,0,58,6,0,0,75,6,0,0,92,6,0,0,109,6,0,0,126,6,0,0,143,6,0,0,157,6,
  0,0,174,6,0,0,191,6,0,0,208,6,0,0,225,6,0,0,239,6,0,0,249,6,0,0,3,7,0,0,18,7,
  0,0,35,7,0,0,52,7,0,0,69,7,0,0,86,7,0,0,103,7,0,0,120,7,0,0,137,7,0,0,154,7,
  0,0,174,7,0,0,194,7,0,0,214,7,0,0,234,7,0,0,253,7,0,0,16,8,0,0,35,8,0,0,54,8,
  0,0,62,8,0,0,71,8,0,0,80,8,0,0,88,8,0,0,105,8,0,0,125,8,0,0,145,8,0,0,162,8,
  0,0,179,8,0,0,196,8,0,0,213,8,0,0,230,8,0,0,247,8,0,0,11,9,0,0,30,9,0,0,46,9,
  0,0,47,9,0,0,48,9,0,0,49,9,0,0,50,9,0,0,51,9,0,0,52,9,0,0,69,9,0,0,86,9,0,0,
  103,9,0,0,120,9,0,0,140,9,0,0,160,9,0,0,180,9,0,0,200,9,0,0,217,9,0,0,234,9,
  0,0,253,9,0,0,6,10,0,0,23,10,0,0,40,10,0,0,57,10,0,0,74,10,0,0,94,10,0,0,114,
  10,0,0,134,10,0,0,154,10,0,0,173,10,0,0,192,10,0,0,211,10,0,0,230,10,0,0,247,
  10,0,0,11,11,0,0,31,11,0,0,48,11,0,0,65,11,0,0,82,11,0,0,99,11,0,0,116,11,0,
  0,133,11,0,0,153,11,0,0,172,11,0,0,188,11,0,0,197,11,0,0,206,11,0,0,215,11,0,
  0,232,11,0,0,249,11,0,0,10,12,0,0,27,12,0,0,47,12,0,0,56,12,0,0,65,12,0,0,74,
  12,0,0,75,12,0,0,76,12,0,0,77,12,0,0,78,12,0,0,95,12,0,0,112,12,0,0,129,12,0,
  0,146,12,0,0,163,12,0,0,174,12,0,0,185,12,0,0,196,12,0,0,205,12,0,0,214,12,0,
  0,223,12,0,0,232,12,0,0,249,12,0,0,10,13,0,0,27,13,0,0,44,13,0,0,64,13,0,0,
  81,13,0,0,98,13,0,0,118,13,0,0,135,13,0,0,152,13,0,0,168,13,0,0,185,13,0,0,
  202,13,0,0,219,13,0,0,236,13,0,0,253,13,0,0,11,14,0,0,28,14,0,0,45,14,0,0,62,
  14,0,0,79,14,0,0,99,14,0,0,119,14,0,0,139,14,0,0,159,14,0,0,178,14,0,0,197,
  14,0,0,216,14,0,0,235,14,0,0,252,14,0,0,13,15,0,0,30,15,0,0,47,15,0,0,61,15,
  0,0,71,15,0,0,81,15,0,0,96,15,0,0,113,15,0,0,130,15,0,0,147,15,0,0,164,15,0,
  0,181,15,0,0,198,15,0,0,215,15,0,0,232,15,0,0,249,15,0,0,10,16,0,0,27,16,0,0,
  44,16,0,0,61,16,0,0,78,16,0,0,93,16,0,0,105,16,0,0,106,16,0,0,107,16,0,0,108,
  16,0,0,109,16,0,0,110,16,0,0,111,16,0,0,112,16,0,0,113,16,0,0,114,16,0,0,115,
  16,0,0,116,16,0,0,117,16,0,0,118,16,0,0,119,16,0,0,120,16,0,0,121,16,0,0,122,
  16,0,0,123,16,0,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,1,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,
  255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,
  159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,245,
  245,244,159,246,242,241,242,243,244,159,159,255,2,11,159,159,245,244,244,159,
  246,245,246,159,159,159,159,255,3,5,159,159,245,244,243,246,246,255,4,5,159,
  159,244,243,243,246,246,255,4,5,159,159,244,243,243,246,246,255,3,5,159,159,
  245,244,243,246,246,255,2,11,159,159,245,244,244,159,246,245,246,159,159,159,
  159,255,1,12,159,159,245,245,244,159,246,242,241,242,243,244,159,159,255,1,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,159,159,159,
  159,159,246,245,246,159,159,159,159,255,3,10,159,159,159,159,159,246,245,246,
  159,159,159,159,255,1,12,159,159,246,245,244,244,243,242,241,242,243,244,159,
  159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,
  12,159,159,244,159,159,159,159,242,241,242,159,159,159,159,255,1,3,159,159,
  244,159,159,6,5,159,159,242,241,242,159,159,255,1,3,159,159,244,159,159,6,5,
  159,159,242,241,242,159,159,255,1,3,159,159,244,159,159,6,5,159,159,242,241,
  242,159,159,255,1,3,159,159,244,159,159,6,5,159,159,242,241,242,159,159,255,
  1,12,159,159,244,159,159,159,159,242,241,242,159,159,159,159,255,1,12,159,
  159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,246,245,
  244,244,243,242,241,242,243,244,159,159,255,3,10,159,159,159,159,159,246,245,
  246,159,159,159,159,255,1,12,159,159,159,159,159,159,159,246,245,246,159,159,
  159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,
  159,244,159,159,159,243,242,241,242,243,244,159,159,255,1,3,159,159,244,159,
  159,5,8,159,159,243,242,241,246,159,159,159,159,255,1,3,159,159,244,159,159,
  5,5,159,159,243,242,241,246,246,255,1,3,159,159,244,159,159,5,5,159,159,243,
  242,241,246,246,255,1,3,159,159,244,159,159,5,8,159,159,243,242,241,246,159,
  159,159,159,255,1,12,159,159,244,159,159,159,243,242,241,242,243,244,159,159,
  255,1,12,159,159,246,245,244,244,243,242,241,242,243,244,159,159,255,2,11,
  159,159,246,244,245,159,246,241,242,243,244,159,159,255,3,3,159,159,159,159,
  159,8,5,245,245,246,159,159,159,159,255,1,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,1,12,159,159,159,159,245,244,243,242,241,242,243,244,159,159,255,4,9,159,
  159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,159,159,159,159,159,
  246,245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,
  159,255,1,12,159,159,159,159,245,244,243,242,245,242,243,244,159,159,255,4,9,
  159,159,245,243,242,245,246,159,159,159,159,255,5,5,159,159,243,242,241,246,
  246,255,5,5,159,159,245,242,241,246,246,255,1,10,159,159,159,159,159,159,159,
  242,241,242,159,159,255,1,12,159,159,244,245,244,244,246,242,241,242,246,159,
  159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,
  1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,
  159,159,159,159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,159,159,
  159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,1,12,159,159,244,159,159,244,159,246,245,242,243,244,159,
  159,255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,
  255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,255,
  1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,255,1,6,
  159,159,244,159,159,159,159,159,8,5,246,246,242,243,244,159,159,255,1,4,159,
  159,244,159,159,159,7,6,246,246,245,242,243,244,159,159,255,1,4,159,159,244,
  245,159,159,7,6,246,246,241,242,243,244,159,159,255,1,4,159,159,244,245,159,
  159,7,6,246,246,241,242,243,244,159,159,255,1,4,159,159,159,159,159,159,7,6,
  246,246,245,246,159,159,159,159,255,2,3,159,159,159,159,159,255,1,4,159,159,
  244,245,159,159,255,1,4,159,159,244,245,159,159,255,1,3,159,159,159,159,159,
  255,2,3,159,159,159,159,159,7,5,246,246,245,246,159,159,159,255,1,5,159,159,
  245,245,245,159,159,7,6,246,246,241,242,243,246,159,159,255,1,5,159,159,244,
  245,244,159,159,7,6,246,246,241,242,243,244,159,159,255,1,12,159,159,244,245,
  244,244,159,246,241,242,243,244,159,159,255,1,12,159,159,244,159,244,244,246,
  246,241,242,243,244,159,159,255,1,12,159,159,244,159,246,244,243,246,245,242,
  243,244,159,159,255,1,12,159,159,244,159,159,244,243,242,245,242,243,244,159,
  159,255,1,12,159,159,244,159,159,246,243,242,241,242,243,244,159,159,255,1,
  12,159,159,244,246,159,159,243,242,241,242,243,244,159,159,255,1,4,159,159,
  244,245,159,159,6,7,159,159,242,241,242,243,244,159,159,255,1,4,159,159,245,
  245,159,159,7,6,246,246,241,242,243,246,159,159,255,2,3,159,159,159,159,159,
  8,4,245,245,246,159,159,159,255,255,255,255,255,255,255,1,12,159,159,159,159,
  159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,1,12,159,159,244,159,159,244,159,159,245,242,243,244,159,
  159,255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,
  255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,255,
  1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,255,1,6,
  159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,255,1,12,159,
  159,244,245,244,244,243,159,245,242,243,244,159,159,255,1,12,159,159,244,245,
  244,159,243,242,241,242,243,244,159,159,255,2,3,159,159,159,159,159,6,7,159,
  159,246,242,242,243,245,159,159,255,8,4,245,245,246,159,159,159,255,1,12,159,
  159,159,159,159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,244,245,
  244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,1,12,159,159,244,159,159,244,159,246,245,242,
  243,244,159,159,255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,
  244,159,159,255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,
  159,159,255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,
  159,255,1,6,159,159,244,159,159,159,159,159,8,5,246,246,242,243,244,159,159,
  255,1,4,159,159,244,159,159,159,7,6,246,246,245,242,243,244,159,159,255,1,4,
  159,159,244,245,159,159,7,6,246,246,241,242,243,244,159,159,255,1,4,159,159,
  244,245,159,159,7,6,246,246,241,242,243,244,159,159,255,1,4,159,159,159,159,
  159,159,7,6,246,246,245,246,159,159,159,159,255,2,3,159,159,159,159,159,7,5,
  246,246,245,246,159,159,159,255,1,5,159,159,245,245,245,159,159,7,6,246,246,
  241,242,243,246,159,159,255,1,5,159,159,244,245,244,159,159,7,6,246,246,241,
  242,243,244,159,159,255,1,12,159,159,244,245,244,244,159,246,241,242,243,244,
  159,159,255,1,12,159,159,244,159,244,244,246,246,241,242,243,244,159,159,255,
  1,12,159,159,244,159,246,244,243,246,245,242,243,244,159,159,255,1,12,159,
  159,244,159,159,244,243,242,245,242,243,244,159,159,255,1,12,159,159,244,159,
  159,246,243,242,241,242,243,244,159,159,255,1,12,159,159,244,246,159,159,243,
  242,241,242,243,244,159,159,255,1,4,159,159,244,245,159,159,6,7,159,159,242,
  241,242,243,244,159,159,255,1,4,159,159,245,245,159,159,7,6,246,246,241,242,
  243,246,159,159,255,2,3,159,159,159,159,159,8,4,245,245,246,159,159,159,255,
  1,4,159,159,159,159,159,159,255,1,4,159,159,244,245,159,159,255,1,4,159,159,
  244,245,159,159,255,1,12,159,159,244,245,159,159,159,246,245,246,159,159,159,
  159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,
  244,245,159,159,244,242,241,242,243,244,159,159,255,1,4,159,159,244,245,159,
  159,6,7,159,159,246,245,246,159,159,159,159,255,1,4,159,159,244,245,159,159,
  255,1,4,159,159,244,245,159,159,255,1,4,159,159,159,159,159,159,255,255,255,
  255,255,1,12,159,159,159,159,159,159,159,246,245,246,159,159,159,159,255,1,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,
  244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,244,159,159,
  244,159,246,241,242,243,244,159,159,255,1,12,159,159,244,159,159,244,159,246,
  245,246,159,159,159,159,255,1,6,159,159,244,159,159,244,159,159,255,1,6,159,
  159,244,159,159,244,159,159,255,1,6,159,159,244,159,159,159,159,159,255,1,4,
  159,159,244,159,159,159,255,1,4,159,159,244,245,159,159,255,1,4,159,159,244,
  245,159,159,255,1,4,159,159,159,159,159,159,255,1,12,159,159,159,159,159,159,
  159,246,245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,
  159,159,255,1,12,159,159,244,159,159,159,243,242,241,242,243,244,159,159,255,
  1,3,159,159,244,159,159,5,8,159,159,243,242,241,246,159,159,159,159,255,1,3,
  159,159,244,159,159,5,5,159,159,243,242,241,246,246,255,1,3,159,159,244,159,
  159,5,5,159,159,243,242,241,246,246,255,1,3,159,159,244,159,159,5,8,159,159,
  243,242,241,246,159,159,159,159,255,1,12,159,159,244,159,159,159,243,242,241,
  242,243,244,159,159,255,1,12,159,159,246,245,244,244,243,242,241,242,243,244,
  159,159,255,2,11,159,159,246,244,245,159,246,241,242,243,244,159,159,255,3,3,
  159,159,159,159,159,8,5,245,245,246,159,159,159,159,255,1,12,159,159,159,159,
  159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,243,
  242,241,242,243,244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,
  243,244,159,159,255,1,12,159,159,159,159,245,244,243,242,241,242,243,244,159,
  159,255,4,9,159,159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,159,
  159,159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,
  243,242,241,242,243,244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,
  242,243,244,159,159,255,1,12,159,159,244,159,159,244,159,246,245,242,243,244,
  159,159,255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,
  159,255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,
  255,1,6,159,159,244,159,159,244,159,159,8,5,246,246,242,243,244,159,159,255,
  1,6,159,159,244,159,159,159,159,159,8,5,246,246,242,243,244,159,159,255,1,4,
  159,159,244,159,159,159,7,6,246,246,245,242,243,244,159,159,255,1,4,159,159,
  244,245,159,159,7,6,246,246,241,242,243,244,159,159,255,1,4,159,159,244,245,
  159,159,7,6,246,246,241,242,243,244,159,159,255,1,4,159,159,159,159,159,159,
  7,6,246,246,245,246,159,159,159,159,255,1,12,159,159,159,159,159,159,159,246,
  245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,
  255,1,12,159,159,159,159,245,244,243,242,245,242,243,244,159,159,255,4,9,159,
  159,245,243,242,245,246,159,159,159,159,255,5,5,159,159,243,242,241,246,246,
  255,5,5,159,159,245,242,241,246,246,255,1,10,159,159,159,159,159,159,159,242,
  241,242,159,159,255,1,12,159,159,244,245,244,244,246,242,241,242,246,159,159,
  159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,
  12,159,159,244,245,244,244,243,242,241,242,243,244,159,159,255,1,12,159,159,
  159,159,159,159,159,246,245,246,159,159,159,159,255,1,12,159,159,159,159,159,
  159,159,246,245,246,159,159,159,159,255,1,12,159,159,244,245,244,244,243,242,
  241,242,243,244,159,159,255,1,12,159,159,244,245,244,244,243,242,241,242,243,
  244,159,159,255,1,12,159,159,244,159,159,159,159,159,245,242,243,244,159,159,
  255,1,3,159,159,244,159,159,8,5,246,246,242,243,244,159,159,255,1,3,159,159,
  244,159,159,8,5,246,246,242,243,244,159,159,255,1,3,159,159,244,159,159,8,5,
  246,246,242,243,244,159,159,255,1,3,159,159,244,159,159,8,5,246,246,242,243,
  244,159,159,255,1,12,159,159,244,246,159,159,159,159,245,242,243,246,159,159,
  255,1,12,159,159,246,245,244,244,243,242,241,242,243,246,159,159,255,2,10,
  159,159,159,246,244,243,242,241,242,246,159,159,255,4,7,159,159,159,159,246,
  245,246,159,159,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,
};

#ifdef DOGS  // 2023/11/19 replaced with CC-BY dogs by nash from GZDoom
static const char dogsd5[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,51,1,0,0,72,1,0,0,100,1,0,0,129,1,0,0,165,1,0,0,191,
  1,0,0,215,1,0,0,238,1,0,0,5,2,0,0,30,2,0,0,57,2,0,0,83,2,0,0,110,2,0,0,136,2,
  0,0,156,2,0,0,164,2,0,0,165,2,0,0,166,2,0,0,167,2,0,0,168,2,0,0,169,2,0,0,
  170,2,0,0,171,2,0,0,172,2,0,0,173,2,0,0,174,2,0,0,175,2,0,0,176,2,0,0,177,2,
  0,0,178,2,0,0,179,2,0,0,180,2,0,0,181,2,0,0,182,2,0,0,183,2,0,0,184,2,0,0,
  185,2,0,0,186,2,0,0,187,2,0,0,188,2,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,52,5,6,6,5,
  79,238,1,1,255,46,16,68,68,66,65,111,109,108,108,143,73,72,67,66,68,15,1,238,
  238,255,44,23,13,13,143,141,68,109,107,106,107,142,147,71,68,68,64,64,65,65,
  66,64,69,72,75,239,239,255,43,24,109,109,109,109,109,111,108,106,106,107,108,
  73,72,74,74,69,71,68,67,68,61,67,66,72,238,238,255,34,2,67,67,72,72,41,16,14,
  14,109,107,107,108,109,109,108,107,78,1,1,2,73,76,239,239,62,5,67,67,72,71,
  76,238,238,255,34,21,67,67,70,74,238,143,14,143,107,107,107,107,108,109,107,
  106,111,13,74,238,239,1,1,255,36,19,73,73,69,141,107,107,108,107,107,107,108,
  108,108,108,109,148,147,73,238,239,239,255,37,18,68,68,141,106,106,106,106,
  107,107,108,108,108,109,108,67,61,67,238,239,239,255,37,18,68,68,141,106,107,
  107,107,107,107,107,109,108,109,108,67,61,68,238,239,239,255,35,20,239,239,
  150,1,142,107,107,107,107,107,108,108,108,108,108,109,147,69,72,238,239,239,
  255,34,22,64,64,70,15,239,14,13,143,108,13,107,107,108,108,108,107,111,149,
  72,238,239,239,239,239,255,42,14,109,109,109,108,109,108,108,108,108,14,238,
  1,2,6,238,238,63,3,1,1,1,239,239,255,44,22,143,143,108,111,109,107,107,107,
  108,109,111,5,1,238,238,74,71,71,73,75,75,73,15,15,255,45,21,143,143,139,67,
  107,107,108,107,108,109,110,1,238,237,69,64,67,68,75,75,75,238,238,255,48,15,
  111,111,109,108,109,109,109,111,6,239,76,69,67,68,74,238,238,255,51,3,6,6,2,
  2,2,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,
};
static const char dogsc5[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,67,1,0,0,94,1,0,0,123,1,0,0,142,1,0,0,170,1,0,0,198,
  1,0,0,224,1,0,0,251,1,0,0,22,2,0,0,51,2,0,0,80,2,0,0,105,2,0,0,131,2,0,0,155,
  2,0,0,168,2,0,0,169,2,0,0,170,2,0,0,171,2,0,0,172,2,0,0,173,2,0,0,174,2,0,0,
  175,2,0,0,176,2,0,0,177,2,0,0,178,2,0,0,179,2,0,0,180,2,0,0,181,2,0,0,182,2,
  0,0,183,2,0,0,184,2,0,0,185,2,0,0,186,2,0,0,187,2,0,0,188,2,0,0,189,2,0,0,
  190,2,0,0,191,2,0,0,192,2,0,0,193,2,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,50,12,5,5,5,
  5,6,1,239,72,67,66,69,69,238,238,63,5,67,67,70,71,75,238,238,255,46,22,109,
  109,5,110,108,109,108,109,111,1,238,69,66,64,64,63,64,66,63,68,65,71,238,238,
  255,44,24,109,109,109,15,107,107,107,107,108,109,110,239,239,74,70,69,70,71,
  72,72,68,70,70,72,238,238,255,42,14,110,110,109,110,111,109,107,107,107,107,
  109,110,6,239,2,2,255,34,4,62,62,66,70,74,74,40,15,14,14,109,108,107,109,109,
  108,107,108,108,108,110,110,79,1,1,255,34,23,239,239,69,68,148,149,108,107,
  106,107,107,108,107,108,107,108,5,109,108,142,141,147,69,238,238,255,37,21,
  147,147,142,107,108,108,108,107,108,109,108,108,108,109,107,106,143,141,141,
  147,147,73,73,255,37,22,68,68,140,140,105,106,107,107,108,108,107,107,108,
  108,108,107,106,142,141,147,67,67,74,74,255,37,22,139,139,141,107,108,107,
  108,107,107,108,107,107,108,108,107,107,142,141,141,146,67,64,73,73,255,35,
  24,71,71,69,236,142,108,107,106,107,108,108,108,107,108,108,108,108,106,107,
  142,142,146,67,71,238,238,255,34,24,62,62,68,71,76,1,239,13,108,108,107,108,
  107,108,107,108,5,109,143,142,141,139,71,72,238,238,255,42,13,109,109,108,
  109,109,108,107,108,109,109,109,14,142,237,237,61,3,239,239,238,1,1,255,43,
  21,13,13,143,110,108,107,107,108,108,110,6,7,1,73,70,72,70,69,69,69,71,1,1,
  255,44,19,142,142,67,64,107,107,108,108,109,6,2,238,67,66,68,69,71,72,73,76,
  76,255,47,8,109,109,109,110,109,5,7,7,238,238,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,
};
static const char dogsb1[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,49,1,0,0,69,1,0,0,91,1,0,0,109,1,0,0,135,1,0,0,159,1,
  0,0,181,1,0,0,202,1,0,0,223,1,0,0,244,1,0,0,11,2,0,0,35,2,0,0,58,2,0,0,83,2,
  0,0,106,2,0,0,107,2,0,0,108,2,0,0,109,2,0,0,110,2,0,0,111,2,0,0,112,2,0,0,
  113,2,0,0,114,2,0,0,115,2,0,0,116,2,0,0,117,2,0,0,118,2,0,0,119,2,0,0,120,2,
  0,0,121,2,0,0,122,2,0,0,123,2,0,0,124,2,0,0,125,2,0,0,126,2,0,0,127,2,0,0,
  128,2,0,0,129,2,0,0,130,2,0,0,131,2,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,60,3,72,72,
  68,73,73,255,48,15,111,111,111,74,71,70,71,72,71,71,70,68,69,65,65,70,70,255,
  46,17,110,110,109,108,141,69,71,70,69,68,68,67,69,73,73,68,69,74,74,255,46,
  13,109,109,14,15,150,73,73,74,74,68,68,74,15,1,1,255,37,21,63,63,136,103,139,
  147,73,147,79,69,69,74,1,239,238,79,237,143,73,238,1,237,237,255,38,19,70,70,
  69,139,14,14,72,8,1,139,107,1,1,74,79,14,236,14,15,238,238,255,40,17,141,141,
  15,15,150,79,141,3,8,8,2,1,71,13,142,142,13,15,15,255,41,16,143,143,15,75,
  150,72,14,8,8,0,2,34,13,13,143,142,15,15,255,41,16,14,14,78,75,238,150,143,8,
  7,2,2,70,13,142,143,13,15,15,255,40,16,142,142,15,14,73,7,142,141,5,7,1,239,
  238,13,141,142,14,14,255,38,18,69,69,139,104,73,236,70,2,238,69,142,110,110,
  238,239,15,14,15,15,15,255,37,19,62,62,67,141,140,141,70,71,72,237,238,238,1,
  239,238,238,238,238,150,238,238,255,46,18,109,109,14,15,15,14,74,74,74,73,69,
  68,70,15,239,77,239,1,74,74,255,46,20,107,107,109,108,13,139,69,71,69,68,69,
  68,65,69,69,71,68,68,65,67,237,237,255,48,18,71,71,5,74,72,71,70,72,73,72,73,
  71,70,69,67,69,146,66,76,76,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsa5[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,65,1,0,0,91,1,0,0,121,1,0,0,138,1,0,0,156,1,0,0,180,
  1,0,0,203,1,0,0,224,1,0,0,245,1,0,0,10,2,0,0,33,2,0,0,57,2,0,0,73,2,0,0,90,2,
  0,0,107,2,0,0,120,2,0,0,121,2,0,0,122,2,0,0,123,2,0,0,124,2,0,0,125,2,0,0,
  126,2,0,0,127,2,0,0,128,2,0,0,129,2,0,0,130,2,0,0,131,2,0,0,132,2,0,0,133,2,
  0,0,134,2,0,0,135,2,0,0,136,2,0,0,137,2,0,0,138,2,0,0,139,2,0,0,140,2,0,0,
  141,2,0,0,142,2,0,0,143,2,0,0,144,2,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,50,19,5,5,14,
  149,73,71,72,70,67,65,64,65,66,65,66,65,64,75,238,1,1,255,48,21,108,108,108,
  107,141,68,70,67,67,65,68,64,63,64,63,64,63,62,64,74,15,239,239,255,47,18,
  108,108,107,107,107,141,69,68,68,69,69,70,69,69,68,71,74,239,1,1,66,3,76,76,
  238,239,239,255,46,12,108,108,107,108,108,108,142,69,68,71,238,2,238,238,255,
  45,13,238,238,239,239,1,109,110,14,237,238,1,239,239,239,239,255,38,19,63,63,
  68,73,75,239,239,147,72,238,239,238,1,2,79,238,239,239,239,239,239,255,39,18,
  70,70,69,73,141,69,61,71,238,238,238,239,1,239,239,239,238,238,238,238,255,
  41,16,148,148,140,69,61,71,238,238,238,239,1,238,238,15,15,14,15,15,255,41,
  16,69,69,140,141,69,69,238,238,239,239,239,238,79,237,14,14,14,14,255,41,16,
  141,141,141,107,72,74,238,1,238,1,239,239,15,15,14,15,15,15,255,39,18,71,71,
  71,149,149,108,108,150,79,1,2,2,237,238,239,239,238,238,238,238,255,38,19,62,
  62,67,70,75,239,239,14,110,107,106,66,72,74,239,238,239,1,1,239,239,255,46,
  11,108,108,106,65,67,74,238,239,72,68,69,66,66,255,46,12,107,107,138,67,67,
  69,71,62,62,63,63,64,15,15,255,46,12,109,109,139,69,66,65,68,62,65,64,64,66,
  238,238,255,48,8,141,141,68,76,74,238,1,2,2,2,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsc1[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,42,1,0,0,63,1,0,0,87,1,0,0,112,1,0,0,146,1,0,0,169,1,
  0,0,190,1,0,0,211,1,0,0,232,1,0,0,253,1,0,0,19,2,0,0,43,2,0,0,78,2,0,0,102,2,
  0,0,124,2,0,0,134,2,0,0,135,2,0,0,136,2,0,0,137,2,0,0,138,2,0,0,139,2,0,0,
  140,2,0,0,141,2,0,0,142,2,0,0,143,2,0,0,144,2,0,0,145,2,0,0,146,2,0,0,147,2,
  0,0,148,2,0,0,149,2,0,0,150,2,0,0,151,2,0,0,152,2,0,0,153,2,0,0,154,2,0,0,
  155,2,0,0,156,2,0,0,157,2,0,0,158,2,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,47,16,
  237,237,238,238,238,15,76,237,238,70,72,238,238,73,72,74,14,14,255,44,19,111,
  111,110,13,72,73,72,73,73,70,69,67,69,66,64,64,65,64,67,73,73,255,43,20,107,
  107,109,109,15,73,73,150,76,74,72,73,76,73,71,73,75,74,73,72,237,237,255,33,
  6,66,66,64,64,66,71,72,72,43,19,5,5,1,2,238,238,239,239,1,239,1,1,1,239,238,
  15,79,15,73,71,71,255,34,18,64,64,137,103,140,73,149,14,79,69,148,14,1,239,
  238,79,15,238,238,238,255,36,16,147,147,236,15,73,79,108,71,107,7,2,104,237,
  78,13,14,15,15,255,37,16,14,14,15,237,150,109,140,7,7,8,2,34,14,142,142,13,
  13,13,255,37,16,142,142,14,236,73,150,142,8,7,0,110,36,238,142,142,143,13,13,
  255,37,16,142,142,15,237,75,109,140,6,8,8,2,34,13,141,142,142,142,142,255,36,
  17,147,147,13,15,149,1,108,70,142,110,2,104,238,14,142,143,13,14,14,255,34,
  19,66,66,139,3,140,72,149,14,79,70,139,13,239,239,79,15,150,14,237,237,237,
  255,33,6,65,65,64,67,141,72,73,73,42,20,6,6,6,1,239,238,15,237,237,237,150,
  15,239,239,1,239,239,79,1,79,15,15,255,44,19,109,109,14,13,72,72,149,74,76,
  75,74,75,237,71,71,70,70,72,74,238,238,255,46,17,14,14,72,71,73,72,72,74,238,
  76,74,70,68,67,69,71,74,239,239,255,49,5,238,238,238,238,239,239,239,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,
};
static const char dogsd1[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,42,1,0,0,63,1,0,0,87,1,0,0,113,1,0,0,148,1,0,0,172,1,
  0,0,194,1,0,0,215,1,0,0,236,1,0,0,2,2,0,0,33,2,0,0,64,2,0,0,83,2,0,0,110,2,0,
  0,135,2,0,0,136,2,0,0,137,2,0,0,138,2,0,0,139,2,0,0,140,2,0,0,141,2,0,0,142,
  2,0,0,143,2,0,0,144,2,0,0,145,2,0,0,146,2,0,0,147,2,0,0,148,2,0,0,149,2,0,0,
  150,2,0,0,151,2,0,0,152,2,0,0,153,2,0,0,154,2,0,0,155,2,0,0,156,2,0,0,157,2,
  0,0,158,2,0,0,159,2,0,0,160,2,0,0,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,47,16,15,15,
  74,76,75,75,76,75,75,73,73,69,67,67,67,147,68,68,255,45,19,111,111,13,142,70,
  73,70,71,70,69,67,68,69,70,65,66,64,65,70,239,239,255,44,21,109,109,14,14,13,
  72,74,75,76,73,70,71,75,15,76,238,75,74,15,72,68,15,15,255,34,24,67,67,136,
  138,147,72,68,68,72,15,238,1,1,79,238,238,238,238,238,239,239,1,1,1,239,239,
  62,2,236,236,70,70,255,34,19,69,69,68,138,74,14,238,6,68,107,79,239,15,239,
  238,79,14,15,15,15,15,255,36,17,142,142,78,236,79,108,141,6,8,1,2,71,14,14,
  13,143,143,14,14,255,37,16,14,14,15,75,237,109,7,8,2,2,30,14,13,143,142,13,
  14,14,255,37,16,14,14,237,75,237,108,7,8,2,2,30,14,13,143,143,13,14,14,255,
  36,17,143,143,15,236,151,105,142,7,2,1,2,73,14,14,13,142,142,13,13,255,34,19,
  69,69,69,138,74,236,1,6,69,143,79,1,74,238,237,238,14,14,15,15,15,55,3,79,79,
  239,238,238,255,33,26,64,64,65,103,138,148,73,70,75,73,76,238,1,1,79,238,79,
  238,238,238,238,239,1,1,1,237,238,238,255,43,14,5,5,79,1,1,79,238,239,238,
  239,239,1,239,239,1,1,255,44,22,109,109,109,109,14,72,73,74,74,72,70,69,70,
  70,68,67,68,67,67,65,64,67,238,238,255,46,20,108,108,142,71,73,71,73,74,73,
  72,69,70,66,66,65,63,64,64,67,147,15,15,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsb5[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,50,1,0,0,74,1,0,0,99,1,0,0,118,1,0,0,134,1,0,0,157,1,
  0,0,179,1,0,0,200,1,0,0,221,1,0,0,243,1,0,0,12,2,0,0,30,2,0,0,55,2,0,0,80,2,
  0,0,104,2,0,0,116,2,0,0,117,2,0,0,118,2,0,0,119,2,0,0,120,2,0,0,121,2,0,0,
  122,2,0,0,123,2,0,0,124,2,0,0,125,2,0,0,126,2,0,0,127,2,0,0,128,2,0,0,129,2,
  0,0,130,2,0,0,131,2,0,0,132,2,0,0,133,2,0,0,134,2,0,0,135,2,0,0,136,2,0,0,
  137,2,0,0,138,2,0,0,139,2,0,0,140,2,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,51,4,6,6,5,2,
  2,2,255,48,19,111,111,110,109,109,109,111,1,239,76,70,65,65,65,67,67,67,66,
  69,76,76,255,47,20,109,109,107,107,107,107,109,110,1,238,238,73,70,68,68,70,
  71,71,69,71,238,238,255,46,14,110,110,108,107,107,107,108,110,5,2,238,2,239,
  238,238,238,255,44,11,110,110,110,109,108,15,238,1,2,2,5,1,1,255,37,18,61,61,
  67,69,74,13,14,13,108,109,109,108,73,76,239,238,1,1,239,239,255,38,17,73,73,
  69,138,141,108,107,109,109,108,69,64,74,238,239,239,239,239,239,255,40,16,68,
  68,141,106,107,109,108,108,62,64,73,238,238,239,238,238,15,15,255,40,16,68,
  68,139,142,107,109,109,108,66,64,73,15,238,238,238,15,15,15,255,39,17,72,72,
  70,141,107,107,108,109,108,148,69,76,238,239,238,239,239,238,238,255,37,20,
  65,65,69,72,75,142,109,13,108,109,107,106,74,238,239,239,1,239,239,239,239,
  239,255,44,13,15,15,109,111,108,1,1,1,2,76,238,239,1,239,239,255,46,12,109,
  109,108,106,107,107,13,69,69,74,239,239,238,238,62,4,70,70,73,15,238,238,255,
  47,20,110,110,106,106,106,142,68,70,70,66,66,66,65,64,64,64,65,70,73,238,239,
  239,255,48,19,109,109,107,107,142,69,73,66,64,62,63,64,64,65,64,66,73,75,238,
  239,239,255,50,7,6,6,111,14,74,74,238,238,238,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsa1[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,48,1,0,0,64,1,0,0,85,1,0,0,105,1,0,0,128,1,0,0,151,1,0,0,173,
  1,0,0,194,1,0,0,216,1,0,0,239,1,0,0,6,2,0,0,29,2,0,0,54,2,0,0,80,2,0,0,106,2,
  0,0,129,2,0,0,130,2,0,0,131,2,0,0,132,2,0,0,133,2,0,0,134,2,0,0,135,2,0,0,
  136,2,0,0,137,2,0,0,138,2,0,0,139,2,0,0,140,2,0,0,141,2,0,0,142,2,0,0,143,2,
  0,0,144,2,0,0,145,2,0,0,146,2,0,0,147,2,0,0,148,2,0,0,149,2,0,0,150,2,0,0,
  151,2,0,0,152,2,0,0,153,2,0,0,154,2,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,60,3,70,70,69,
  237,237,255,52,11,238,238,74,73,72,72,70,66,66,63,65,74,74,255,47,16,5,5,5,7,
  71,72,69,68,68,69,69,72,70,70,67,65,15,15,255,45,15,108,108,108,109,109,141,
  70,73,71,68,70,71,74,15,239,239,239,255,39,18,62,62,137,139,139,71,69,70,74,
  238,2,1,239,72,14,71,73,238,1,1,255,39,18,65,65,67,139,140,72,236,141,1,110,
  68,139,13,1,78,149,72,150,15,15,255,41,17,148,148,142,15,14,73,1,3,70,3,5,8,
  2,73,139,140,14,15,15,255,42,16,108,108,14,15,14,150,237,147,110,8,8,1,74,
  141,139,13,15,15,255,41,17,107,107,107,143,14,150,149,76,71,111,7,0,2,38,142,
  140,13,14,14,255,40,18,70,70,107,148,15,15,14,150,109,69,107,8,8,1,72,140,
  139,13,15,15,255,40,18,72,72,146,141,14,14,72,1,6,67,141,14,2,239,13,141,140,
  237,78,78,255,39,18,63,63,137,140,138,70,72,147,74,69,71,72,15,238,1,236,149,
  75,238,238,255,39,4,64,64,67,68,68,68,45,12,111,111,111,111,1,1,238,15,150,
  72,71,75,239,239,255,45,21,108,108,108,108,143,13,72,72,73,71,67,67,67,71,74,
  70,71,70,70,69,67,69,69,255,46,21,110,110,109,109,110,148,70,69,69,68,69,68,
  66,67,65,67,64,65,63,63,66,239,239,255,49,18,1,1,79,79,238,76,74,74,73,73,73,
  73,70,68,71,15,147,72,15,15,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsf1[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,42,1,0,0,43,1,0,0,68,1,0,0,102,1,0,0,134,1,0,0,165,1,
  0,0,193,1,0,0,221,1,0,0,249,1,0,0,23,2,0,0,55,2,0,0,88,2,0,0,122,2,0,0,152,2,
  0,0,180,2,0,0,200,2,0,0,213,2,0,0,222,2,0,0,223,2,0,0,224,2,0,0,225,2,0,0,
  226,2,0,0,227,2,0,0,228,2,0,0,229,2,0,0,230,2,0,0,231,2,0,0,232,2,0,0,233,2,
  0,0,234,2,0,0,235,2,0,0,236,2,0,0,237,2,0,0,238,2,0,0,239,2,0,0,240,2,0,0,
  241,2,0,0,242,2,0,0,243,2,0,0,244,2,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,45,
  20,238,238,239,238,15,76,74,73,73,71,15,238,238,238,239,74,70,69,66,66,238,
  238,255,28,2,66,66,65,65,42,23,6,6,13,70,71,68,70,69,69,66,66,66,67,67,68,65,
  74,239,76,236,237,70,67,238,238,255,28,6,66,66,137,138,72,73,74,74,41,17,9,9,
  9,149,70,71,69,70,67,64,67,72,70,69,69,66,66,74,74,255,29,26,69,69,138,150,
  73,2,15,72,238,239,1,1,1,238,79,15,150,73,73,73,68,75,239,239,238,239,239,
  239,255,30,23,149,149,15,150,7,14,108,111,1,2,102,239,15,238,15,150,150,149,
  147,73,238,79,238,238,238,255,30,23,14,14,15,75,79,141,8,8,8,106,238,102,14,
  15,13,139,150,148,73,149,238,15,14,15,15,255,30,23,14,14,15,74,76,110,7,8,
  102,2,44,36,1,14,142,137,137,147,142,13,13,14,14,14,14,255,29,25,13,13,15,14,
  14,109,108,6,8,102,2,40,34,1,142,141,137,136,135,139,137,13,13,13,13,14,14,
  255,27,27,68,68,147,138,74,14,1,107,141,5,2,106,2,38,102,14,139,137,137,136,
  136,134,138,13,13,13,13,14,14,255,26,28,65,65,65,102,138,139,141,73,79,68,
  142,79,110,237,102,237,13,138,135,134,133,134,132,139,143,13,13,13,15,15,255,
  34,29,238,238,239,238,238,1,79,74,14,141,134,145,147,147,72,236,14,14,14,237,
  238,1,1,1,239,238,239,1,1,239,239,255,39,25,238,238,75,143,142,71,69,146,67,
  70,73,238,238,238,238,239,238,238,239,238,238,74,73,143,70,67,67,255,41,23,
  15,15,150,69,68,69,65,73,67,68,239,239,239,239,238,238,239,1,239,238,238,74,
  67,67,67,255,42,15,13,13,68,69,69,67,69,68,66,238,239,74,75,238,238,239,239,
  255,44,8,70,70,68,70,71,73,74,238,239,239,255,47,4,239,239,239,239,239,239,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,
};
static const char dogsd2d8[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,39,1,0,0,49,1,0,0,60,1,0,0,73,1,0,0,86,1,0,0,105,1,0,0,128,1,0,0,
  152,1,0,0,177,1,0,0,202,1,0,0,229,1,0,0,6,2,0,0,29,2,0,0,50,2,0,0,68,2,0,0,
  93,2,0,0,121,2,0,0,149,2,0,0,177,2,0,0,203,2,0,0,225,2,0,0,245,2,0,0,7,3,0,0,
  20,3,0,0,34,3,0,0,49,3,0,0,65,3,0,0,82,3,0,0,99,3,0,0,116,3,0,0,134,3,0,0,
  155,3,0,0,175,3,0,0,191,3,0,0,199,3,0,0,200,3,0,0,201,3,0,0,202,3,0,0,203,3,
  0,0,204,3,0,0,205,3,0,0,206,3,0,0,207,3,0,0,208,3,0,0,209,3,0,0,210,3,0,0,
  211,3,0,0,212,3,0,0,213,3,0,0,214,3,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,41,3,8,8,8,0,0,255,41,5,6,6,8,8,2,238,238,255,41,6,
  106,106,13,79,2,40,15,15,255,39,8,1,1,109,140,140,142,104,30,15,15,255,38,8,
  15,15,150,15,66,66,71,8,68,68,255,38,8,15,15,1,107,64,65,15,32,238,238,60,2,
  69,69,1,1,255,35,11,238,238,13,14,236,14,238,62,66,70,70,239,239,59,3,64,64,
  65,237,237,255,34,12,69,69,138,139,78,142,63,64,67,67,72,15,239,239,58,3,64,
  64,63,72,72,255,34,13,75,75,237,239,14,66,65,66,147,148,74,238,238,238,238,
  57,3,67,67,67,73,73,255,36,13,14,14,68,65,68,67,71,13,74,238,238,238,237,78,
  78,56,3,71,71,70,15,15,255,35,15,141,141,139,67,148,237,236,71,73,70,76,238,
  15,14,13,143,143,56,3,70,70,71,239,239,255,33,5,67,67,64,63,68,238,238,39,19,
  142,142,141,148,71,72,149,150,75,15,13,142,13,238,15,72,72,73,71,74,74,255,
  39,18,149,149,142,147,71,142,13,149,72,73,15,13,143,13,238,238,76,74,76,76,
  255,40,16,142,142,141,143,13,142,142,72,71,13,143,14,143,14,239,239,238,238,
  255,41,13,141,141,13,108,143,140,69,68,69,149,238,15,14,2,2,255,41,14,13,13,
  108,108,142,13,148,69,69,149,76,238,15,239,239,239,62,2,67,67,76,76,255,42,
  14,13,13,107,13,13,143,140,69,71,74,79,238,238,239,239,239,61,5,75,75,69,74,
  238,72,72,255,43,23,13,13,142,141,143,141,72,70,73,239,239,239,1,79,238,2,
  238,75,14,239,71,62,146,72,72,255,43,23,108,108,142,106,106,141,69,68,71,74,
  76,238,1,1,238,15,76,71,64,62,66,67,71,238,238,255,44,21,142,142,106,106,139,
  67,65,64,69,71,70,68,70,73,65,65,63,62,64,72,238,238,238,255,44,17,110,110,
  106,106,69,67,67,69,70,72,69,69,69,65,66,66,65,69,69,255,45,15,111,111,13,71,
  69,68,70,70,71,70,70,69,67,72,73,75,75,255,45,13,110,110,110,78,73,73,73,73,
  73,76,15,238,238,239,239,255,45,8,109,109,109,110,110,15,237,238,239,239,255,
  45,9,109,109,109,108,108,108,143,150,239,1,1,255,45,10,108,108,108,107,107,
  109,109,13,72,75,239,239,255,46,11,109,109,107,106,107,108,143,71,69,73,239,
  1,1,255,46,12,13,13,107,109,109,107,143,139,69,69,76,238,1,1,255,46,12,109,
  109,13,108,110,108,142,141,148,143,15,238,239,239,255,45,12,108,108,107,141,
  239,110,110,109,109,14,14,150,1,1,255,45,5,108,108,142,13,237,2,2,53,4,69,69,
  67,78,1,1,255,45,5,108,108,142,13,238,239,239,53,7,66,66,71,1,1,79,239,238,
  238,255,45,5,108,108,106,109,79,239,239,54,6,71,71,71,72,70,72,74,74,255,45,
  4,142,142,140,13,2,2,56,3,15,15,238,239,239,255,45,3,139,139,69,79,79,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsg1[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,42,1,0,0,64,1,0,0,93,1,0,0,129,1,0,0,164,1,0,0,196,1,
  0,0,223,1,0,0,249,1,0,0,20,2,0,0,47,2,0,0,74,2,0,0,98,2,0,0,129,2,0,0,160,2,
  0,0,190,2,0,0,211,2,0,0,225,2,0,0,235,2,0,0,236,2,0,0,237,2,0,0,238,2,0,0,
  239,2,0,0,240,2,0,0,241,2,0,0,242,2,0,0,243,2,0,0,244,2,0,0,245,2,0,0,246,2,
  0,0,247,2,0,0,248,2,0,0,249,2,0,0,250,2,0,0,251,2,0,0,252,2,0,0,253,2,0,0,
  254,2,0,0,255,2,0,0,0,3,0,0,1,3,0,0,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,50,3,1,1,1,
  239,239,54,10,239,239,1,111,78,237,74,75,15,15,73,73,255,31,2,137,137,69,69,
  46,18,111,111,13,71,70,70,71,73,72,72,13,143,149,71,70,69,148,65,65,65,255,
  31,4,138,138,74,237,239,239,41,23,5,5,13,71,73,71,72,73,73,73,70,72,71,71,15,
  150,75,76,74,150,15,79,72,67,67,255,31,7,15,15,2,70,15,238,239,239,239,39,19,
  7,7,79,238,236,72,148,73,72,70,66,66,68,65,66,65,66,69,67,73,73,255,31,27,
  237,237,14,15,238,1,2,238,1,1,239,239,238,79,237,75,74,70,67,73,150,71,70,69,
  69,65,66,73,73,255,31,22,8,8,7,1,238,102,98,74,239,239,239,238,1,15,76,150,
  148,74,2,239,239,239,1,1,255,31,21,7,7,7,15,239,98,15,102,238,239,239,14,15,
  13,74,149,150,15,79,238,1,239,239,255,30,22,14,14,6,7,150,2,98,40,34,150,239,
  239,237,142,140,137,140,14,14,15,15,79,238,238,255,30,22,237,237,6,1,40,2,98,
  38,98,237,239,238,15,141,137,137,137,142,142,14,14,15,238,238,255,30,22,72,
  72,13,141,239,2,102,34,102,238,239,15,237,140,137,138,136,137,143,14,14,15,1,
  1,255,33,19,238,238,8,38,102,238,238,238,15,14,136,134,135,137,138,13,14,13,
  14,1,1,255,36,26,6,6,79,79,237,237,13,146,146,147,146,149,237,14,14,15,239,
  239,1,1,1,239,238,238,239,239,239,239,255,38,26,5,5,75,73,148,72,147,68,70,
  71,238,15,15,238,239,238,238,79,79,238,239,15,76,236,149,71,69,69,255,39,25,
  108,108,141,68,69,70,68,65,76,74,65,67,76,76,76,75,76,238,79,1,239,238,77,70,
  68,69,69,255,40,16,111,111,69,69,67,66,67,69,67,68,65,238,73,76,237,238,1,1,
  255,42,9,73,73,69,69,71,69,69,74,238,239,239,255,45,5,238,238,238,239,239,
  239,239,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,
};
static const char dogsh8[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,45,1,0,0,55,1,0,0,66,
  1,0,0,78,1,0,0,98,1,0,0,124,1,0,0,148,1,0,0,172,1,0,0,197,1,0,0,221,1,0,0,
  239,1,0,0,255,1,0,0,13,2,0,0,27,2,0,0,42,2,0,0,58,2,0,0,75,2,0,0,94,2,0,0,
  115,2,0,0,138,2,0,0,163,2,0,0,189,2,0,0,218,2,0,0,246,2,0,0,18,3,0,0,45,3,0,
  0,72,3,0,0,98,3,0,0,121,3,0,0,143,3,0,0,171,3,0,0,198,3,0,0,223,3,0,0,245,3,
  0,0,10,4,0,0,28,4,0,0,42,4,0,0,43,4,0,0,44,4,0,0,45,4,0,0,46,4,0,0,47,4,0,0,
  48,4,0,0,49,4,0,0,50,4,0,0,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,54,4,149,149,71,76,239,239,255,52,5,109,109,
  142,148,150,239,239,255,51,6,108,108,14,143,14,237,239,239,255,50,7,13,13,
  139,143,13,236,239,1,1,255,48,8,143,143,142,141,149,142,150,239,239,239,58,3,
  70,70,72,76,76,255,47,8,14,14,141,141,69,237,6,1,1,1,57,9,70,70,65,69,68,67,
  69,69,70,238,238,255,47,19,142,142,142,110,110,109,109,14,73,70,239,69,148,
  13,142,69,68,62,64,74,74,255,47,19,108,108,108,108,109,108,14,143,143,71,72,
  148,143,109,239,76,73,148,64,67,67,255,46,14,108,108,109,107,108,107,143,140,
  142,141,148,71,140,142,1,1,64,2,65,65,65,65,255,46,13,108,108,107,108,108,
  107,142,69,68,69,71,71,141,14,14,64,2,72,72,238,238,255,46,13,107,107,107,
  106,107,107,142,147,66,67,69,73,76,238,238,255,46,11,108,108,107,108,108,109,
  143,70,70,73,238,238,238,255,46,9,108,108,107,107,107,14,149,1,1,1,1,255,46,
  9,108,108,108,108,109,108,14,15,238,1,1,255,46,10,108,108,107,108,107,14,15,
  15,239,239,239,239,255,45,11,109,109,108,108,109,14,15,15,239,1,239,239,239,
  255,45,12,109,109,108,108,109,110,109,14,74,76,238,239,239,239,255,45,14,109,
  109,108,107,106,107,107,141,72,72,73,74,76,238,238,238,255,45,16,109,109,108,
  3,106,107,142,70,65,67,69,69,71,73,74,76,239,239,255,45,18,109,109,107,142,
  141,3,140,67,63,65,66,68,70,69,69,71,70,76,239,239,255,45,20,108,108,3,142,
  140,141,140,68,67,65,64,65,67,67,68,67,65,69,73,76,238,238,255,45,21,109,109,
  142,142,142,142,139,67,69,71,71,68,69,67,65,65,64,65,64,67,71,238,238,255,45,
  13,109,109,142,108,142,141,69,69,70,73,74,76,238,15,15,59,7,1,1,238,64,64,64,
  63,70,70,255,45,13,109,109,143,14,13,141,70,68,70,236,237,237,14,239,239,62,
  6,71,71,65,62,67,73,15,15,255,44,14,68,68,5,140,142,14,13,13,141,14,15,143,
  13,15,1,1,63,5,239,239,66,66,66,75,75,255,44,15,66,66,64,14,141,143,14,14,14,
  238,14,143,143,239,239,1,1,65,3,67,67,67,67,67,255,44,16,70,70,102,139,15,72,
  71,75,15,239,14,15,76,76,237,239,239,239,66,2,69,69,66,66,255,45,15,140,140,
  71,72,237,236,74,238,239,239,238,76,75,73,74,237,237,66,2,69,69,73,73,255,45,
  8,15,15,236,236,69,15,239,238,1,1,55,6,76,76,76,74,73,73,15,15,255,45,8,238,
  238,15,149,14,76,74,15,239,239,57,5,74,74,75,73,72,73,73,255,42,2,73,73,13,
  13,45,8,14,14,78,150,7,6,70,15,238,238,58,5,73,73,72,70,72,237,237,255,41,13,
  70,70,145,138,15,109,14,74,2,69,139,142,238,239,239,60,5,71,71,71,69,74,238,
  238,255,42,12,65,65,138,74,15,76,74,75,140,142,14,1,238,238,61,4,69,69,69,71,
  73,73,255,44,10,68,68,14,150,79,73,108,6,7,1,103,103,62,3,65,65,67,73,73,255,
  45,9,71,71,1,6,69,5,7,8,1,66,66,62,3,65,65,65,15,15,255,47,7,70,70,71,6,7,8,
  1,69,69,62,2,68,68,67,67,255,49,4,107,107,2,2,70,70,63,1,71,71,71,255,255,
  255,255,255,255,255,255,255,255,
};
static const char dogse8[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,37,1,0,0,45,1,0,0,54,1,0,0,63,1,0,0,73,1,0,0,83,1,0,0,92,1,0,0,102,1,0,0,
  121,1,0,0,144,1,0,0,168,1,0,0,194,1,0,0,218,1,0,0,242,1,0,0,2,2,0,0,19,2,0,0,
  44,2,0,0,72,2,0,0,101,2,0,0,132,2,0,0,166,2,0,0,204,2,0,0,241,2,0,0,19,3,0,0,
  51,3,0,0,79,3,0,0,108,3,0,0,136,3,0,0,157,3,0,0,178,3,0,0,198,3,0,0,215,3,0,
  0,231,3,0,0,244,3,0,0,251,3,0,0,252,3,0,0,253,3,0,0,254,3,0,0,255,3,0,0,0,4,
  0,0,1,4,0,0,2,4,0,0,3,4,0,0,4,4,0,0,5,4,0,0,6,4,0,0,7,4,0,0,8,4,0,0,9,4,0,0,
  10,4,0,0,11,4,0,0,255,255,255,255,255,255,255,255,255,255,255,255,255,255,53,
  2,148,148,72,72,255,53,3,67,67,71,239,239,255,52,4,109,109,147,73,239,239,
  255,52,4,143,143,141,143,238,238,255,51,5,111,111,108,13,14,1,1,255,51,5,143,
  143,142,14,15,239,239,255,51,4,13,13,143,143,238,238,255,50,5,109,109,13,236,
  237,239,239,255,49,5,111,111,5,5,1,238,238,56,5,69,69,68,69,73,15,15,255,47,
  18,110,110,109,109,108,14,14,73,73,74,69,143,13,148,69,69,69,74,239,239,255,
  46,19,108,108,108,108,107,140,140,140,141,141,71,142,14,238,148,68,66,64,64,
  237,237,255,46,12,108,108,107,107,142,146,67,67,69,71,140,143,15,15,60,5,2,2,
  239,14,62,68,68,255,44,13,111,111,110,107,108,109,143,69,67,68,69,72,72,73,
  73,63,2,64,64,68,68,255,43,13,5,5,111,111,110,15,236,15,74,74,76,15,238,239,
  239,63,2,76,76,239,239,255,42,11,5,5,6,74,74,15,238,239,238,238,2,79,79,255,
  41,12,6,6,108,143,71,68,68,70,71,73,237,238,1,1,255,41,20,105,105,107,141,68,
  64,66,68,70,71,71,73,239,1,239,238,237,76,237,238,239,239,255,40,23,13,13,
  143,3,141,68,65,65,66,68,68,69,69,68,72,79,79,237,75,75,150,14,14,238,238,
  255,39,24,14,14,13,141,142,141,68,68,70,71,67,66,66,64,65,70,70,72,238,1,237,
  74,14,148,15,15,255,38,20,109,109,109,107,143,13,147,67,67,71,73,74,72,67,69,
  65,63,64,65,68,238,238,61,2,71,71,69,69,255,36,24,14,14,13,14,14,142,13,71,
  70,69,71,69,149,75,79,238,239,15,68,64,64,62,68,73,77,77,62,1,237,237,237,
  255,28,2,238,238,77,77,35,16,13,13,13,141,14,13,143,71,67,66,69,148,73,149,
  15,15,79,79,54,7,239,239,67,61,64,67,73,238,238,255,28,4,72,72,64,63,238,238,
  34,16,236,236,14,73,71,141,142,141,147,68,70,147,147,74,14,14,143,143,57,4,
  64,64,64,65,68,68,255,29,22,238,238,140,66,74,239,79,236,148,70,71,142,148,
  68,70,148,138,145,138,142,14,13,8,8,59,3,146,146,147,73,73,255,31,21,140,140,
  64,68,69,69,143,148,73,74,73,69,70,139,136,135,141,142,14,79,1,2,2,60,2,70,
  70,239,239,255,30,23,72,72,15,66,65,66,147,148,74,237,150,76,74,148,138,140,
  142,143,13,15,237,76,239,239,239,255,30,24,147,147,15,143,64,64,65,67,73,15,
  15,238,150,236,143,142,13,79,76,73,73,71,73,74,239,239,255,31,13,143,143,236,
  150,73,64,66,69,72,239,238,237,79,1,1,49,6,73,73,68,71,70,74,239,239,255,32,
  8,15,15,1,108,65,64,76,69,239,239,51,4,69,69,69,69,74,74,255,32,8,15,15,75,
  237,67,136,70,239,238,238,52,4,68,68,68,72,238,238,255,33,7,238,238,108,140,
  140,142,106,15,15,53,4,66,66,68,72,238,238,255,35,5,106,106,143,2,2,74,74,54,
  3,66,66,68,238,238,255,35,4,6,6,7,8,2,2,54,3,67,67,64,74,74,255,36,2,8,8,0,0,
  55,2,63,63,72,72,255,55,2,66,66,73,73,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,
};
static const char dogsd4d6[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,40,1,0,0,48,1,0,0,59,1,0,0,71,1,0,0,86,
  1,0,0,108,1,0,0,131,1,0,0,162,1,0,0,193,1,0,0,227,1,0,0,1,2,0,0,25,2,0,0,48,
  2,0,0,73,2,0,0,93,2,0,0,113,2,0,0,134,2,0,0,155,2,0,0,182,2,0,0,208,2,0,0,
  236,2,0,0,4,3,0,0,26,3,0,0,46,3,0,0,67,3,0,0,91,3,0,0,112,3,0,0,131,3,0,0,
  147,3,0,0,157,3,0,0,168,3,0,0,179,3,0,0,190,3,0,0,201,3,0,0,210,3,0,0,219,3,
  0,0,228,3,0,0,236,3,0,0,244,3,0,0,245,3,0,0,246,3,0,0,247,3,0,0,248,3,0,0,
  249,3,0,0,250,3,0,0,251,3,0,0,252,3,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,41,1,5,5,5,255,40,3,69,69,73,104,104,255,
  38,6,73,73,72,72,73,15,238,238,255,37,7,238,238,71,70,70,236,238,238,238,255,
  34,10,65,65,66,71,76,71,139,148,71,74,239,239,255,34,11,66,66,73,237,239,139,
  141,141,70,74,79,239,239,62,2,239,239,239,239,255,37,10,148,148,140,141,141,
  142,109,109,13,74,238,238,60,4,73,73,72,72,238,238,255,36,16,141,141,139,141,
  105,140,143,108,107,13,14,14,74,73,238,239,1,1,57,6,74,74,72,68,69,68,238,
  238,255,36,26,71,71,79,140,106,141,142,142,143,106,107,139,66,69,69,70,71,72,
  73,74,72,72,74,74,72,76,238,238,255,34,4,69,69,70,15,1,1,39,21,106,106,106,
  106,13,107,141,140,138,64,61,64,64,65,66,67,66,68,68,71,76,239,239,255,34,2,
  64,64,238,238,39,19,108,108,106,106,141,107,105,105,141,69,65,65,67,65,66,67,
  68,70,73,239,239,255,40,14,108,108,107,106,106,106,106,143,141,142,72,72,71,
  76,239,239,57,1,237,237,237,255,40,12,1,1,108,107,106,107,107,107,142,142,
  108,14,238,238,56,2,15,15,238,238,255,41,13,7,7,107,107,105,107,106,107,108,
  108,110,1,239,1,1,55,3,239,239,238,238,238,255,42,15,108,108,107,106,106,107,
  108,108,108,111,111,111,79,79,238,15,15,255,43,15,107,107,107,107,107,106,
  108,110,110,109,109,13,236,237,238,239,239,255,43,16,109,109,107,107,107,107,
  108,108,109,109,108,13,13,14,14,79,239,239,255,44,16,108,108,108,107,107,106,
  106,106,107,107,143,141,148,71,13,14,1,1,255,44,16,109,109,107,108,107,106,
  106,108,107,106,142,142,66,66,69,141,79,79,62,2,238,238,239,239,255,45,15,
  109,109,109,107,107,106,105,107,106,143,142,70,72,66,141,1,1,62,2,79,79,15,
  15,255,45,15,5,5,109,106,106,106,107,107,106,141,72,72,67,64,148,6,6,61,4,
  238,238,238,76,237,237,255,46,19,108,108,106,106,105,107,108,107,71,73,239,
  64,64,69,79,15,15,14,73,238,238,255,46,17,107,107,107,105,142,143,108,14,74,
  1,239,64,62,64,70,79,74,74,74,255,47,15,108,108,109,107,109,15,15,15,1,239,
  76,62,63,65,72,239,239,255,47,16,109,109,108,141,15,1,239,239,239,238,239,72,
  63,61,66,72,15,15,255,48,7,109,109,142,15,2,239,1,1,1,58,8,238,238,63,63,66,
  71,15,238,238,238,255,48,6,108,108,142,110,1,7,7,7,60,6,63,63,64,66,67,71,74,
  74,255,48,5,109,109,109,13,1,2,2,61,5,63,63,66,72,15,239,239,255,48,5,109,
  109,108,109,111,1,1,61,2,70,70,239,239,255,48,5,109,109,107,108,14,1,1,255,
  48,6,108,108,141,141,13,79,1,1,255,48,6,109,109,140,147,71,5,1,1,255,48,6,
  109,109,140,68,70,237,239,239,255,48,6,13,13,140,66,72,15,239,239,255,49,4,
  139,139,64,70,238,238,255,49,4,69,69,65,69,238,238,255,49,4,69,69,68,70,239,
  239,255,49,3,67,67,67,72,72,255,49,3,64,64,64,238,238,255,255,255,255,255,
  255,255,255,255,255,
};
static const char dogsa2a8[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,35,1,0,0,42,1,0,
  0,56,1,0,0,72,1,0,0,89,1,0,0,109,1,0,0,130,1,0,0,155,1,0,0,181,1,0,0,214,1,0,
  0,245,1,0,0,19,2,0,0,50,2,0,0,77,2,0,0,103,2,0,0,128,2,0,0,151,2,0,0,173,2,0,
  0,194,2,0,0,216,2,0,0,237,2,0,0,1,3,0,0,19,3,0,0,37,3,0,0,55,3,0,0,72,3,0,0,
  87,3,0,0,101,3,0,0,116,3,0,0,131,3,0,0,148,3,0,0,166,3,0,0,186,3,0,0,211,3,0,
  0,233,3,0,0,255,3,0,0,15,4,0,0,23,4,0,0,24,4,0,0,25,4,0,0,26,4,0,0,27,4,0,0,
  28,4,0,0,29,4,0,0,30,4,0,0,31,4,0,0,32,4,0,0,33,4,0,0,34,4,0,0,35,4,0,0,36,4,
  0,0,37,4,0,0,38,4,0,0,255,255,255,255,255,255,255,255,255,255,255,255,58,2,
  66,66,69,69,255,58,2,64,64,68,68,255,49,2,0,0,0,0,57,3,64,64,64,237,237,255,
  48,4,110,110,8,0,2,2,57,3,67,67,70,238,238,255,48,5,107,107,108,7,1,238,238,
  56,3,72,72,68,238,238,255,45,8,15,15,110,70,140,140,141,2,73,73,56,3,72,72,
  73,238,238,255,44,9,78,78,237,238,68,69,138,141,105,70,70,56,3,74,74,73,238,
  238,255,39,13,76,76,237,1,1,15,238,76,108,73,61,66,238,70,70,56,3,75,75,73,
  76,76,255,39,13,65,65,69,139,14,15,14,70,238,69,63,69,239,238,238,55,4,14,14,
  237,76,239,239,255,40,12,1,1,238,14,14,141,66,64,67,66,71,237,1,1,53,6,238,
  238,237,75,75,76,2,2,64,2,67,67,73,73,255,42,9,14,14,71,65,65,66,68,148,73,
  238,238,52,6,238,238,237,76,150,76,238,238,63,3,67,67,69,73,73,255,40,17,150,
  150,141,139,66,67,69,68,71,236,238,238,239,238,237,76,76,239,239,62,4,69,69,
  66,67,238,238,255,39,18,63,63,65,65,67,79,238,14,73,149,149,76,239,239,238,
  150,76,239,1,1,61,4,75,75,64,67,76,76,255,44,13,142,142,141,148,148,70,75,1,
  238,15,13,142,239,2,2,60,5,238,238,63,64,72,238,238,255,44,12,147,147,142,
  147,70,149,13,236,75,238,141,138,13,13,59,5,239,239,64,63,66,15,15,255,45,12,
  141,141,148,13,13,13,141,73,15,141,137,142,15,15,59,4,64,64,64,64,72,72,255,
  45,18,141,141,142,108,107,142,70,69,148,149,139,14,238,238,67,64,65,69,15,15,
  255,45,17,108,108,107,108,141,141,69,68,69,69,71,71,67,65,64,67,72,15,15,255,
  45,16,108,108,143,107,141,148,69,68,69,71,67,66,66,66,67,69,74,74,255,44,17,
  108,108,106,142,142,142,140,69,66,68,64,65,69,68,69,71,237,238,238,255,44,16,
  108,108,107,142,141,140,140,138,66,65,66,68,69,71,73,237,238,238,255,44,15,
  108,108,106,107,141,141,105,139,67,64,69,69,70,72,76,238,238,255,44,13,108,
  108,107,107,105,105,107,107,70,70,70,72,76,238,238,255,44,13,108,108,108,108,
  111,109,108,109,13,74,76,238,238,239,239,255,44,13,108,108,107,108,108,109,
  109,78,79,79,239,239,239,239,239,255,44,12,108,108,108,107,108,108,13,109,78,
  79,239,239,239,239,255,44,10,108,108,108,108,108,110,109,110,79,238,239,239,
  255,44,9,107,107,107,107,108,108,143,143,238,1,1,255,44,10,108,108,106,106,
  107,108,108,149,73,238,1,1,255,45,10,107,107,108,108,107,108,141,69,71,238,
  239,239,255,44,12,109,109,110,108,108,108,13,147,70,69,72,76,238,238,255,43,
  13,108,108,107,107,109,108,109,13,142,148,141,236,74,237,237,255,42,15,109,
  109,107,107,140,1,110,109,108,13,13,142,149,142,14,1,1,255,42,5,107,107,105,
  141,13,239,239,48,11,238,238,111,15,73,73,74,139,14,1,239,2,2,255,42,4,142,
  142,141,140,15,15,54,9,67,67,147,14,14,74,72,71,68,67,67,255,41,5,143,143,
  141,108,109,1,1,55,8,71,71,71,70,70,66,64,71,238,238,255,41,4,142,142,138,
  141,1,1,58,3,72,72,71,238,238,255,41,3,147,147,68,79,79,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsf8[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,41,1,0,0,50,1,0,0,59,1,0,0,69,1,0,0,79,1,0,0,89,
  1,0,0,109,1,0,0,131,1,0,0,153,1,0,0,176,1,0,0,197,1,0,0,214,1,0,0,232,1,0,0,
  253,1,0,0,20,2,0,0,44,2,0,0,69,2,0,0,94,2,0,0,114,2,0,0,134,2,0,0,154,2,0,0,
  176,2,0,0,206,2,0,0,239,2,0,0,15,3,0,0,46,3,0,0,80,3,0,0,116,3,0,0,151,3,0,0,
  185,3,0,0,213,3,0,0,241,3,0,0,12,4,0,0,31,4,0,0,53,4,0,0,68,4,0,0,76,4,0,0,
  83,4,0,0,90,4,0,0,97,4,0,0,104,4,0,0,111,4,0,0,112,4,0,0,113,4,0,0,114,4,0,0,
  115,4,0,0,116,4,0,0,117,4,0,0,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,49,3,139,139,68,77,77,255,49,4,141,141,67,72,1,1,255,
  49,4,143,143,140,139,239,239,255,49,5,13,13,141,13,15,1,1,255,49,5,109,109,
  13,108,15,1,1,255,49,5,107,107,141,141,14,1,1,255,49,5,109,109,107,13,236,1,
  1,58,6,73,73,73,73,73,73,238,238,255,49,4,108,108,107,143,238,238,56,9,66,66,
  65,70,69,69,68,65,68,238,238,255,50,4,108,108,5,6,79,79,56,9,69,69,71,14,13,
  70,67,65,63,66,66,255,49,12,110,110,110,109,14,13,70,72,72,142,15,239,239,
  239,63,2,67,67,65,65,255,48,11,108,108,107,109,107,13,13,142,143,148,142,15,
  15,64,1,67,67,67,255,47,12,109,109,109,108,107,13,140,139,140,71,72,149,79,
  79,255,46,13,110,110,107,107,109,107,106,147,69,69,70,73,238,238,238,255,46,
  16,109,109,107,107,107,108,13,147,68,69,73,239,2,238,239,77,238,238,255,45,
  18,111,111,109,108,107,109,109,14,73,71,238,1,238,238,237,237,238,237,239,
  239,255,44,19,111,111,111,110,109,108,14,143,237,1,239,239,1,239,239,238,237,
  15,150,14,14,255,43,14,5,5,6,6,79,111,110,15,75,238,1,1,239,1,239,239,61,2,
  73,73,70,70,255,42,15,6,6,110,15,75,15,238,239,239,238,238,238,1,1,1,239,239,
  62,1,74,74,74,255,41,15,106,106,107,108,71,70,70,73,150,238,238,238,238,239,
  239,238,238,255,40,15,109,109,13,106,141,67,65,68,69,70,73,237,238,238,239,
  239,239,255,38,15,14,14,14,13,142,106,139,66,65,65,67,70,69,72,150,239,239,
  255,37,17,14,14,110,109,14,13,141,147,67,69,67,67,66,68,68,70,73,238,238,255,
  28,2,64,64,15,15,35,19,15,15,14,14,109,14,13,149,148,69,68,71,72,71,67,65,66,
  65,69,150,150,255,28,3,68,68,64,67,67,34,21,15,15,14,149,14,14,13,72,69,68,
  71,70,70,73,73,238,66,65,65,68,71,239,239,255,29,27,71,71,137,68,1,238,15,14,
  149,74,14,13,142,71,69,69,72,72,148,148,238,13,13,67,64,66,69,238,238,255,30,
  26,140,140,146,66,148,239,14,150,73,74,150,149,70,149,140,69,70,149,14,13,13,
  13,1,66,63,66,238,238,255,30,21,13,13,72,66,68,69,73,14,150,238,238,76,73,
  150,139,135,137,141,142,13,14,239,239,52,4,76,76,63,66,73,73,255,27,23,74,74,
  148,236,1,14,68,65,66,148,236,238,239,238,238,237,141,137,138,140,13,143,15,
  2,2,53,4,65,65,65,71,77,77,255,27,23,67,67,137,138,14,15,143,69,64,69,69,237,
  239,239,238,14,13,143,143,14,14,237,239,1,1,54,3,66,66,68,76,76,255,29,12,13,
  13,13,15,71,1,150,63,70,76,72,1,238,238,44,6,238,238,73,73,73,77,1,1,54,3,66,
  66,66,71,71,255,31,9,78,78,150,79,13,64,67,238,239,238,238,46,4,72,72,72,74,
  239,239,55,2,69,69,69,69,255,31,9,14,14,150,14,68,66,66,77,2,31,31,46,4,73,
  73,74,74,239,239,55,2,66,66,72,72,255,32,9,79,79,109,147,156,140,238,2,38,73,
  73,46,4,73,73,74,73,238,238,56,1,239,239,239,255,34,7,142,142,141,14,79,2,44,
  73,73,47,3,73,73,73,238,238,255,34,4,141,141,6,8,2,2,39,2,15,15,237,237,47,3,
  72,72,73,238,238,255,35,3,7,7,8,2,2,47,3,71,71,77,239,239,255,47,3,70,70,238,
  1,1,255,47,2,69,69,238,238,255,47,2,65,65,74,74,255,47,2,65,65,70,70,255,47,
  2,65,65,70,70,255,47,2,237,237,239,239,255,255,255,255,255,255,255,255,
};
static const char dogsh1[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,37,1,0,0,38,1,0,0,39,
  1,0,0,40,1,0,0,41,1,0,0,42,1,0,0,50,1,0,0,77,1,0,0,103,1,0,0,130,1,0,0,155,1,
  0,0,175,1,0,0,191,1,0,0,209,1,0,0,228,1,0,0,249,1,0,0,14,2,0,0,36,2,0,0,56,2,
  0,0,85,2,0,0,117,2,0,0,142,2,0,0,162,2,0,0,163,2,0,0,164,2,0,0,165,2,0,0,166,
  2,0,0,167,2,0,0,168,2,0,0,169,2,0,0,170,2,0,0,171,2,0,0,172,2,0,0,173,2,0,0,
  174,2,0,0,175,2,0,0,176,2,0,0,177,2,0,0,178,2,0,0,179,2,0,0,180,2,0,0,181,2,
  0,0,182,2,0,0,183,2,0,0,184,2,0,0,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,57,3,238,238,
  238,239,239,255,49,16,111,111,110,13,72,72,74,150,237,1,1,238,72,70,69,67,71,
  71,67,2,72,72,70,70,255,48,21,108,108,107,143,148,69,69,72,73,150,15,1,239,
  239,237,72,70,68,67,67,65,67,67,255,47,22,108,108,108,108,5,74,70,69,69,72,
  71,71,68,68,67,68,65,66,66,66,67,66,70,70,255,47,20,109,109,109,108,142,8,0,
  7,68,70,69,69,68,69,71,72,73,70,74,238,239,239,255,46,15,111,111,107,143,1,
  143,8,0,8,73,73,71,70,71,239,1,1,255,46,11,109,109,110,238,110,5,6,8,2,8,150,
  15,15,255,45,13,238,238,15,238,237,238,106,6,2,239,31,15,150,238,238,255,44,
  14,75,75,139,79,76,14,74,156,13,14,239,61,238,14,15,15,255,45,16,141,141,79,
  14,108,70,137,135,142,2,68,142,143,13,239,1,1,1,255,46,16,14,14,150,5,67,65,
  65,239,67,238,142,141,13,15,238,76,75,75,255,45,17,15,15,141,68,63,61,64,70,
  3,76,14,141,139,14,15,15,237,15,15,255,44,15,14,14,65,67,65,65,66,67,73,74,
  239,143,140,141,238,1,1,255,43,17,141,141,137,65,68,69,68,72,143,237,239,238,
  143,149,76,238,239,1,1,66,3,66,66,238,2,2,255,42,27,69,69,64,65,238,0,143,
  150,237,79,238,1,1,150,70,72,72,72,73,75,72,69,70,70,68,66,238,2,2,255,49,20,
  111,111,78,238,239,73,72,68,66,65,67,68,69,65,66,67,66,67,65,74,239,239,255,
  52,15,73,73,74,71,71,70,71,69,68,68,68,70,71,75,75,70,70,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsa4a6[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,40,1,0,0,47,1,0,0,61,1,0,0,79,1,0,0,98,1,0,0,
  122,1,0,0,146,1,0,0,170,1,0,0,193,1,0,0,217,1,0,0,243,1,0,0,6,2,0,0,24,2,0,0,
  42,2,0,0,60,2,0,0,78,2,0,0,96,2,0,0,115,2,0,0,135,2,0,0,156,2,0,0,178,2,0,0,
  202,2,0,0,237,2,0,0,11,3,0,0,38,3,0,0,60,3,0,0,73,3,0,0,86,3,0,0,100,3,0,0,
  114,3,0,0,129,3,0,0,144,3,0,0,160,3,0,0,175,3,0,0,190,3,0,0,208,3,0,0,224,3,
  0,0,241,3,0,0,0,4,0,0,9,4,0,0,17,4,0,0,24,4,0,0,25,4,0,0,26,4,0,0,27,4,0,0,
  28,4,0,0,29,4,0,0,30,4,0,0,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,58,2,15,15,239,239,255,58,2,73,73,238,238,255,46,2,73,73,
  238,238,57,3,68,68,72,238,238,255,43,6,75,75,237,73,72,239,238,238,57,3,69,
  69,72,238,238,255,42,7,72,72,72,69,72,150,238,238,238,56,3,73,73,73,239,239,
  255,38,11,63,63,64,68,70,237,70,149,147,73,238,239,239,55,4,73,73,72,73,238,
  238,255,39,10,76,76,75,237,69,139,140,72,73,15,239,239,54,5,74,74,74,74,74,
  238,238,255,41,9,143,143,147,140,141,141,109,15,238,239,239,52,6,73,73,73,73,
  73,73,77,77,255,40,18,14,14,139,141,142,141,13,14,13,14,75,73,71,70,71,71,70,
  76,238,238,255,38,19,66,66,69,69,150,142,107,142,143,108,13,109,109,13,141,
  68,67,70,76,238,238,255,38,3,69,69,238,238,238,43,14,107,107,107,108,108,107,
  106,106,143,140,68,70,76,238,239,239,255,43,14,109,109,108,108,108,143,142,
  143,142,148,72,74,238,238,239,239,255,44,13,108,108,108,108,107,13,107,13,
  142,73,73,238,238,239,239,255,44,13,109,109,107,109,108,107,13,143,13,13,74,
  14,238,238,238,255,44,13,109,109,107,108,107,109,109,109,14,14,76,77,239,239,
  239,255,44,13,109,109,107,108,109,109,6,6,2,239,238,238,1,238,238,255,44,13,
  108,108,107,107,5,111,109,110,14,237,238,239,239,238,238,255,44,14,109,109,
  109,110,109,108,108,108,108,13,13,14,239,239,1,1,255,44,15,109,109,107,107,
  106,108,108,106,142,141,139,69,71,73,238,1,1,255,44,16,109,109,108,107,106,
  106,107,106,142,141,68,65,66,68,147,79,2,2,255,44,17,109,109,107,107,105,106,
  106,106,141,140,70,65,64,65,62,69,15,1,1,255,44,19,109,109,106,106,106,107,
  106,142,140,68,70,67,68,66,64,63,67,72,15,238,238,255,44,11,108,108,106,106,
  142,106,108,107,140,69,68,239,239,57,8,69,69,62,62,63,65,70,75,237,237,66,3,
  238,238,238,239,239,255,45,10,107,107,106,142,142,107,13,72,238,239,2,2,58,
  11,65,65,62,62,63,62,65,68,70,72,76,1,1,255,45,9,108,108,107,13,108,109,14,
  150,239,2,2,59,9,238,238,71,65,63,62,61,65,72,238,238,255,45,9,110,110,106,
  108,79,79,79,239,239,1,1,63,4,238,238,65,65,238,238,255,46,8,109,109,110,15,
  1,239,238,239,1,1,255,46,8,111,111,79,239,238,238,239,239,1,1,255,45,9,109,
  109,109,15,1,237,238,238,1,239,239,255,45,9,109,109,14,79,1,237,238,238,239,
  1,1,255,44,10,143,143,142,14,15,1,73,239,239,1,1,1,255,44,10,141,141,147,142,
  5,1,75,239,238,239,239,239,255,43,11,109,109,69,147,141,15,239,1,238,15,239,
  1,1,255,43,10,108,108,65,69,143,237,239,238,238,76,238,238,255,43,10,13,13,
  65,69,150,238,239,74,75,14,238,238,255,43,5,147,147,65,68,238,238,238,49,4,
  73,73,75,237,79,79,255,43,4,67,67,68,70,238,238,50,3,72,72,14,79,79,255,43,4,
  67,67,65,77,239,239,50,4,72,72,72,14,1,1,255,43,3,65,65,67,238,238,51,3,64,
  64,147,238,238,255,51,4,67,67,62,70,238,238,255,52,3,66,66,63,68,68,255,53,2,
  70,70,68,68,255,255,255,255,255,255,255,255,
};
static const char dogsb4b6[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,49,1,0,0,67,1,0,0,90,1,0,0,
  112,1,0,0,134,1,0,0,160,1,0,0,187,1,0,0,214,1,0,0,240,1,0,0,4,2,0,0,24,2,0,0,
  51,2,0,0,82,2,0,0,110,2,0,0,138,2,0,0,167,2,0,0,183,2,0,0,199,2,0,0,215,2,0,
  0,230,2,0,0,245,2,0,0,4,3,0,0,19,3,0,0,35,3,0,0,51,3,0,0,66,3,0,0,82,3,0,0,
  97,3,0,0,116,3,0,0,136,3,0,0,160,3,0,0,183,3,0,0,203,3,0,0,214,3,0,0,224,3,0,
  0,233,3,0,0,241,3,0,0,247,3,0,0,248,3,0,0,249,3,0,0,250,3,0,0,251,3,0,0,252,
  3,0,0,253,3,0,0,254,3,0,0,255,3,0,0,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,44,2,72,72,238,238,58,3,239,239,76,238,
  238,255,41,6,71,71,70,73,73,76,238,238,58,3,70,70,72,15,15,255,36,11,72,72,
  62,66,73,75,69,72,70,13,237,238,238,58,3,66,66,69,238,238,255,37,10,67,67,67,
  69,70,139,147,139,72,73,239,239,58,3,68,68,71,238,238,255,39,9,79,79,148,147,
  142,140,71,143,78,1,1,57,4,237,237,72,237,239,239,255,40,10,69,69,139,141,
  141,142,109,14,236,74,238,238,53,7,238,238,237,74,74,74,73,237,237,255,38,22,
  14,14,139,70,141,106,140,143,108,13,14,237,15,73,72,70,72,72,71,72,70,73,238,
  238,255,37,22,64,64,69,73,239,141,106,142,142,13,143,109,108,106,140,67,68,
  67,67,68,70,74,238,238,255,37,2,75,75,238,238,42,15,106,106,107,107,108,107,
  106,105,141,69,65,67,68,71,73,76,76,255,42,15,110,110,107,107,108,142,142,
  106,142,69,70,71,74,238,238,238,238,255,43,15,108,108,108,107,107,106,107,13,
  73,72,74,15,239,1,239,1,1,255,43,16,6,6,108,107,108,107,143,13,14,74,76,15,
  239,76,14,1,1,1,65,2,76,76,239,239,255,44,19,108,108,106,107,107,108,13,148,
  139,13,13,142,71,70,69,143,78,1,239,76,76,64,3,239,239,71,238,238,255,44,23,
  109,109,107,107,109,109,108,13,141,13,13,141,69,237,67,63,65,67,68,67,68,68,
  68,74,74,255,44,23,109,109,108,108,109,108,108,108,107,107,143,13,73,238,71,
  65,63,64,65,65,65,64,64,69,69,255,45,11,109,109,107,108,108,108,109,108,107,
  108,14,15,15,58,9,76,76,67,66,70,72,76,238,238,238,238,255,45,11,108,108,107,
  107,107,108,107,107,109,109,109,79,79,255,45,11,109,109,107,107,106,105,106,
  108,108,109,111,239,239,255,45,11,109,109,107,106,107,106,107,107,108,110,6,
  239,239,255,45,10,109,109,107,107,107,106,106,107,109,111,2,2,255,45,10,108,
  108,106,105,106,106,106,107,109,5,239,239,255,45,10,108,108,107,106,142,107,
  14,110,110,1,239,239,255,46,10,107,107,107,142,14,110,79,1,239,1,2,2,255,46,
  11,110,110,108,142,14,15,1,238,239,239,1,1,1,255,47,11,110,110,13,13,79,239,
  239,238,238,239,1,2,2,255,48,10,13,13,15,239,1,238,15,237,238,1,1,1,255,47,
  11,110,110,13,14,1,238,237,15,76,237,239,2,2,255,47,10,109,109,109,110,79,1,
  239,238,15,237,239,239,255,47,5,13,13,108,13,15,1,1,53,5,238,238,238,75,75,
  238,238,255,47,5,143,143,148,143,78,1,1,53,6,76,76,75,74,73,14,15,15,255,47,
  5,141,141,147,71,111,1,1,54,10,73,73,72,68,72,149,14,239,239,239,238,238,255,
  46,6,108,108,147,68,148,150,239,239,56,8,73,73,68,66,67,68,68,72,238,238,255,
  46,6,14,14,67,67,73,76,239,239,58,5,237,237,65,69,75,239,239,255,46,6,14,14,
  66,68,74,238,1,1,255,46,5,147,147,64,66,15,238,238,255,46,4,67,67,68,69,238,
  238,255,46,3,65,65,64,15,15,255,47,1,67,67,67,255,255,255,255,255,255,255,
  255,255,255,
};
static const char dogsh6[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,47,1,0,0,63,1,0,0,80,1,0,0,97,1,0,0,
  115,1,0,0,134,1,0,0,152,1,0,0,171,1,0,0,189,1,0,0,205,1,0,0,221,1,0,0,236,1,
  0,0,251,1,0,0,11,2,0,0,40,2,0,0,68,2,0,0,96,2,0,0,126,2,0,0,153,2,0,0,173,2,
  0,0,192,2,0,0,217,2,0,0,243,2,0,0,10,3,0,0,32,3,0,0,55,3,0,0,74,3,0,0,93,3,0,
  0,110,3,0,0,126,3,0,0,141,3,0,0,149,3,0,0,150,3,0,0,151,3,0,0,152,3,0,0,153,
  3,0,0,154,3,0,0,155,3,0,0,156,3,0,0,157,3,0,0,158,3,0,0,159,3,0,0,160,3,0,0,
  161,3,0,0,162,3,0,0,163,3,0,0,164,3,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,59,8,76,76,71,63,64,67,74,238,1,1,255,56,
  11,67,67,61,64,64,64,66,68,68,70,71,238,238,255,55,12,72,72,66,65,68,70,74,
  14,1,1,1,239,239,239,255,51,12,239,239,76,71,67,66,79,14,142,139,65,70,239,
  239,255,50,13,66,66,67,68,111,107,108,107,141,141,147,146,70,238,238,255,48,
  14,6,6,107,69,111,107,108,108,107,142,141,140,145,69,72,72,255,48,13,108,108,
  111,109,108,107,107,105,107,106,139,138,65,71,71,255,47,14,109,109,109,107,
  106,107,106,107,3,141,139,139,69,74,1,1,255,47,13,109,109,108,106,107,106,
  105,143,143,141,147,148,238,1,1,255,47,11,109,109,108,106,106,106,143,142,
  108,14,14,2,2,255,46,11,109,109,107,141,142,142,14,15,79,1,2,1,1,255,46,10,
  107,107,107,108,14,110,15,15,238,239,2,2,255,46,10,107,107,106,106,143,13,
  108,109,237,1,239,239,255,46,11,108,108,106,106,106,108,108,108,109,150,239,
  1,1,255,45,12,108,108,108,107,107,106,106,107,108,13,73,73,239,239,59,8,72,
  72,71,72,74,76,238,238,239,239,255,45,23,108,108,107,107,106,106,105,106,107,
  142,148,73,76,1,69,63,64,64,64,65,65,64,68,15,15,255,45,23,108,108,109,109,
  106,106,108,107,107,107,143,73,70,66,64,62,64,67,67,67,68,68,70,15,15,255,44,
  19,109,109,107,107,109,109,109,107,108,108,13,13,69,67,70,67,72,79,1,238,238,
  65,2,74,74,238,238,255,44,16,108,108,106,107,109,110,110,109,109,108,14,13,
  71,73,74,79,2,2,65,2,238,238,239,239,255,44,15,108,108,107,107,107,109,6,5,
  79,238,238,79,239,239,1,2,2,255,44,14,108,108,107,107,108,13,13,13,150,239,
  238,238,239,238,238,238,255,40,1,64,64,64,43,15,143,143,107,108,142,143,13,
  141,69,69,71,75,238,15,15,15,15,255,41,1,72,72,72,43,16,109,109,107,108,107,
  142,3,140,67,66,68,71,72,76,75,76,237,237,255,41,18,239,239,236,72,143,108,
  14,111,108,141,68,69,69,67,69,69,74,73,237,237,255,42,17,147,147,140,142,13,
  9,15,1,76,71,71,72,72,71,71,69,72,15,15,255,42,7,150,150,14,69,14,73,75,239,
  239,53,7,15,15,74,73,72,70,73,238,238,255,43,6,15,15,68,2,149,71,76,76,57,4,
  72,72,73,74,238,238,255,43,6,79,79,73,67,68,148,237,237,58,4,70,70,68,75,238,
  238,255,44,5,15,15,69,70,70,15,15,59,3,68,68,70,237,237,255,45,4,75,75,68,68,
  238,238,59,3,69,69,73,238,238,255,45,4,141,141,68,68,2,2,59,2,148,148,74,74,
  255,45,3,2,2,143,108,108,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,
};
static const char dogsb2b8[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,37,1,0,0,51,1,0,0,69,1,0,0,89,1,0,0,109,1,0,0,130,1,0,0,154,1,0,0,179,1,0,
  0,202,1,0,0,231,1,0,0,7,2,0,0,38,2,0,0,64,2,0,0,89,2,0,0,114,2,0,0,139,2,0,0,
  164,2,0,0,188,2,0,0,212,2,0,0,235,2,0,0,0,3,0,0,19,3,0,0,34,3,0,0,49,3,0,0,
  70,3,0,0,95,3,0,0,120,3,0,0,144,3,0,0,169,3,0,0,183,3,0,0,192,3,0,0,201,3,0,
  0,210,3,0,0,219,3,0,0,229,3,0,0,238,3,0,0,244,3,0,0,245,3,0,0,246,3,0,0,247,
  3,0,0,248,3,0,0,249,3,0,0,250,3,0,0,251,3,0,0,252,3,0,0,253,3,0,0,254,3,0,0,
  255,3,0,0,0,4,0,0,1,4,0,0,2,4,0,0,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,46,2,8,8,0,0,255,46,3,8,8,7,0,0,59,2,66,66,75,75,255,45,6,
  141,141,109,7,8,1,237,237,58,3,67,67,64,74,74,255,43,8,2,2,6,140,156,142,78,
  2,72,72,58,3,65,65,66,238,238,255,42,9,15,15,75,15,69,134,135,238,2,70,70,58,
  2,67,67,73,73,255,41,9,238,238,237,75,106,67,64,66,239,30,30,57,3,72,72,69,
  238,238,255,38,12,141,141,107,107,14,14,74,5,69,64,69,106,238,238,57,3,73,73,
  72,238,238,255,37,13,68,68,147,147,141,15,236,67,66,64,67,69,74,1,1,57,3,74,
  74,73,76,76,255,41,9,150,150,71,65,64,67,71,73,238,239,239,55,5,239,239,73,
  73,73,237,237,255,38,13,239,239,14,142,69,64,67,67,71,148,76,238,239,238,238,
  53,7,237,237,73,71,73,73,74,239,239,255,37,22,63,63,67,139,147,68,68,148,149,
  71,74,73,238,238,237,14,239,15,74,75,74,76,239,239,64,1,71,71,71,255,38,20,
  74,74,15,238,15,15,149,143,141,71,71,75,238,15,15,13,13,238,238,239,239,239,
  63,2,146,146,65,65,255,43,14,13,13,141,71,71,143,143,74,73,238,13,143,13,239,
  1,1,62,3,239,239,67,69,69,255,43,13,141,141,148,140,143,13,142,148,72,150,
  143,143,143,13,13,62,3,66,66,67,72,72,255,44,12,142,142,141,108,142,142,71,
  68,139,141,150,237,14,14,61,4,65,65,65,70,77,77,255,44,12,143,143,14,13,143,
  141,71,68,69,148,73,14,15,15,60,4,72,72,63,64,15,15,255,44,12,109,109,143,
  107,142,143,148,69,69,72,71,74,238,238,60,4,66,66,64,67,238,238,255,44,19,
  109,109,143,142,142,13,141,147,65,71,73,73,68,69,66,67,64,65,66,71,71,255,44,
  19,109,109,13,143,140,141,140,147,66,69,67,68,67,68,67,67,66,69,74,239,239,
  255,44,18,109,109,108,105,141,105,105,139,66,64,64,68,69,69,69,69,69,73,239,
  239,255,44,16,109,109,108,107,3,107,106,139,67,65,68,68,70,71,73,76,239,239,
  255,44,14,14,14,108,109,14,107,108,142,72,71,70,73,73,74,239,239,255,45,10,
  108,108,109,109,110,79,15,74,75,238,238,238,255,45,10,108,108,108,109,15,5,2,
  1,1,2,1,1,255,45,11,108,108,108,110,110,79,2,238,15,75,74,150,150,64,1,77,77,
  77,255,45,14,109,109,107,107,236,68,66,69,68,69,70,143,143,75,1,1,63,2,66,66,
  64,64,255,45,20,108,108,109,107,147,65,64,65,68,71,69,141,142,143,14,73,69,
  69,148,64,69,69,255,46,19,108,108,106,142,146,67,147,140,141,74,148,143,13,
  143,141,68,69,65,66,15,15,255,45,20,5,5,110,109,108,143,142,13,14,74,15,238,
  72,71,71,72,74,76,238,238,239,239,255,45,9,108,108,143,237,7,111,111,5,1,1,1,
  255,44,4,13,13,107,140,13,13,255,44,4,108,108,107,142,15,15,255,44,4,107,107,
  141,143,238,238,255,44,4,143,143,108,109,1,1,255,43,5,142,142,140,141,15,1,1,
  255,43,4,147,147,66,75,239,239,255,44,1,238,238,238,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,
};
static const char dogsc2c8[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,38,1,0,0,47,1,0,0,58,1,0,0,72,1,0,0,86,1,0,0,103,1,0,0,121,1,0,0,
  138,1,0,0,153,1,0,0,180,1,0,0,211,1,0,0,237,1,0,0,3,2,0,0,25,2,0,0,45,2,0,0,
  62,2,0,0,86,2,0,0,111,2,0,0,136,2,0,0,158,2,0,0,180,2,0,0,206,2,0,0,231,2,0,
  0,1,3,0,0,18,3,0,0,35,3,0,0,53,3,0,0,71,3,0,0,93,3,0,0,116,3,0,0,135,3,0,0,
  152,3,0,0,161,3,0,0,169,3,0,0,176,3,0,0,177,3,0,0,178,3,0,0,179,3,0,0,180,3,
  0,0,181,3,0,0,182,3,0,0,183,3,0,0,184,3,0,0,185,3,0,0,186,3,0,0,187,3,0,0,
  188,3,0,0,189,3,0,0,190,3,0,0,191,3,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,43,2,8,8,0,0,255,42,4,110,110,6,8,2,2,255,42,6,105,
  105,105,79,1,239,237,237,255,39,9,150,150,79,72,140,140,140,1,44,237,237,255,
  38,9,238,238,14,150,238,67,65,69,100,64,64,255,35,12,238,238,79,1,149,237,1,
  107,64,64,74,239,238,238,255,34,13,66,66,146,140,14,15,149,15,64,62,67,73,71,
  1,1,255,35,12,238,238,149,14,143,141,64,66,66,67,73,238,1,1,255,37,10,14,14,
  71,66,65,67,147,73,15,238,1,1,255,34,14,72,72,143,13,140,67,67,68,68,148,150,
  15,238,239,15,15,57,4,237,237,68,67,71,71,255,33,16,66,66,62,65,67,68,238,
  237,14,74,73,73,75,79,238,15,14,14,56,6,69,69,67,65,67,69,238,238,255,39,11,
  143,143,143,140,148,70,72,15,15,238,13,143,143,54,6,74,74,72,72,2,239,239,
  239,255,40,17,141,141,139,72,72,13,149,72,15,14,141,14,238,76,150,76,74,239,
  239,255,40,17,142,142,141,143,13,142,148,71,149,142,142,143,143,238,15,15,
  237,238,238,255,40,15,143,143,141,108,106,142,70,67,68,147,71,14,15,237,239,
  239,239,255,41,12,13,13,108,143,142,148,69,69,70,72,73,149,238,238,255,41,13,
  13,13,13,13,142,141,147,68,68,72,74,75,74,238,238,60,2,238,238,1,1,255,42,20,
  13,13,143,142,142,140,68,67,69,69,69,73,239,2,1,239,239,15,75,71,75,75,255,
  42,20,108,108,13,141,141,141,68,66,66,69,67,69,74,239,238,15,238,78,150,15,
  238,238,255,43,17,107,107,105,106,105,139,67,65,68,69,71,74,238,238,238,79,
  15,238,238,255,43,17,109,109,106,106,108,13,70,69,69,71,71,72,73,75,239,73,
  238,238,238,255,43,21,109,109,109,111,6,1,15,76,1,72,73,74,73,75,69,66,65,64,
  67,71,237,1,1,255,44,20,110,110,110,111,79,79,239,239,2,75,76,72,70,70,65,65,
  74,72,69,72,79,79,255,44,14,109,109,109,111,110,238,75,1,1,1,1,1,2,239,238,
  238,61,3,238,238,238,1,1,255,44,12,109,109,108,107,109,148,69,69,73,73,73,73,
  76,76,255,45,12,109,109,107,13,140,67,65,68,67,70,70,13,1,1,255,45,13,110,
  110,107,106,141,146,67,69,139,148,72,142,14,1,1,255,46,13,109,109,108,108,
  142,141,143,143,149,74,141,14,5,1,1,255,46,8,108,108,5,110,110,109,78,79,238,
  238,55,5,68,68,71,14,238,238,238,255,46,6,109,109,141,74,7,7,7,7,56,8,71,71,
  70,71,73,73,238,238,238,238,255,47,4,142,142,142,238,239,239,58,6,69,69,69,
  68,69,71,15,15,255,47,4,142,142,141,14,239,239,59,4,237,237,239,239,239,239,
  255,47,4,109,109,109,15,1,1,255,48,3,149,149,150,79,79,255,49,2,73,73,238,
  238,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsg6[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,46,1,0,0,59,1,0,0,73,1,0,0,85,1,0,0,
  106,1,0,0,130,1,0,0,153,1,0,0,173,1,0,0,191,1,0,0,207,1,0,0,223,1,0,0,239,1,
  0,0,1,2,0,0,20,2,0,0,40,2,0,0,61,2,0,0,83,2,0,0,106,2,0,0,136,2,0,0,167,2,0,
  0,202,2,0,0,245,2,0,0,19,3,0,0,49,3,0,0,77,3,0,0,105,3,0,0,132,3,0,0,150,3,0,
  0,168,3,0,0,182,3,0,0,190,3,0,0,191,3,0,0,192,3,0,0,193,3,0,0,194,3,0,0,195,
  3,0,0,196,3,0,0,197,3,0,0,198,3,0,0,199,3,0,0,200,3,0,0,201,3,0,0,202,3,0,0,
  203,3,0,0,204,3,0,0,205,3,0,0,206,3,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,43,7,63,63,69,74,77,238,238,239,239,255,
  43,8,64,64,67,68,68,74,75,76,239,239,255,43,9,70,70,69,65,69,70,73,14,1,1,1,
  255,45,7,14,14,70,69,148,73,1,1,1,255,46,7,5,5,15,141,13,15,1,1,1,59,5,77,77,
  72,64,65,66,66,255,47,6,111,111,15,14,110,1,1,1,56,9,73,73,68,68,66,66,66,66,
  73,238,238,255,47,18,5,5,110,110,111,1,1,76,238,73,73,70,68,70,74,15,1,239,
  239,239,255,46,15,110,110,109,110,109,15,1,239,15,238,74,73,73,76,238,239,
  239,255,45,13,111,111,109,109,110,110,14,1,239,239,238,76,76,239,239,255,45,
  11,109,109,109,109,108,110,109,238,1,239,2,2,2,255,44,11,110,110,107,108,108,
  108,109,108,238,1,1,2,2,255,43,11,109,109,108,107,107,107,106,108,108,1,1,
  239,239,255,41,13,110,110,109,108,107,106,105,106,142,142,107,110,2,239,239,
  255,40,14,109,109,108,107,106,107,106,106,106,107,108,108,109,2,1,1,255,39,
  15,108,108,108,107,107,107,106,107,107,106,107,106,108,109,5,239,239,255,38,
  16,108,108,107,106,106,107,108,107,107,106,106,107,106,107,109,111,1,1,255,
  37,17,107,107,107,107,107,106,107,106,107,107,106,106,106,107,107,108,109,79,
  79,255,36,18,109,109,107,106,106,106,106,106,107,107,111,108,108,106,108,109,
  108,108,15,15,255,35,25,110,110,108,106,107,106,106,106,107,107,109,109,109,
  109,109,109,108,14,14,237,237,76,68,68,71,238,238,255,35,26,108,108,3,107,
  106,105,107,108,106,108,110,5,109,109,14,14,108,13,142,69,75,65,62,62,65,71,
  76,76,255,34,30,108,108,142,107,106,142,107,107,108,108,14,75,77,1,1,236,73,
  14,13,148,70,71,69,67,67,69,68,69,69,238,239,239,255,31,1,237,237,237,34,23,
  142,142,107,106,106,142,107,142,107,141,69,69,72,76,1,1,238,238,238,14,14,15,
  1,6,6,58,6,237,237,72,69,69,67,238,238,255,31,19,148,148,2,109,106,106,142,
  142,143,108,105,106,142,67,65,68,70,73,237,239,239,61,2,69,69,238,238,255,32,
  19,1,1,140,107,142,142,143,108,13,110,108,149,72,71,71,73,73,74,238,239,239,
  61,2,74,74,238,238,255,31,11,75,75,238,141,142,141,140,109,109,14,238,79,79,
  44,8,73,73,238,76,75,75,237,238,1,1,255,31,9,238,238,139,14,142,140,141,78,
  79,1,1,44,2,72,72,15,15,48,4,73,73,72,237,239,239,255,31,8,63,63,238,1,15,
  142,149,238,1,1,44,2,74,74,238,238,49,4,72,72,74,15,1,1,255,32,6,72,72,238,
  239,1,237,238,238,50,3,69,69,73,238,238,255,32,6,14,14,69,73,76,239,239,239,
  50,3,72,72,74,15,15,255,32,4,108,108,73,237,238,238,37,1,238,238,238,255,32,
  3,8,8,140,71,71,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,
};
static const char dogsc4c6[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,41,1,0,0,50,1,0,0,62,1,0,0,78,1,0,0,93,
  1,0,0,106,1,0,0,127,1,0,0,151,1,0,0,180,1,0,0,209,1,0,0,239,1,0,0,16,2,0,0,
  37,2,0,0,59,2,0,0,82,2,0,0,105,2,0,0,129,2,0,0,157,2,0,0,185,2,0,0,214,2,0,0,
  244,2,0,0,19,3,0,0,41,3,0,0,59,3,0,0,71,3,0,0,81,3,0,0,91,3,0,0,102,3,0,0,
  113,3,0,0,125,3,0,0,137,3,0,0,148,3,0,0,160,3,0,0,171,3,0,0,181,3,0,0,190,3,
  0,0,199,3,0,0,207,3,0,0,208,3,0,0,209,3,0,0,210,3,0,0,211,3,0,0,212,3,0,0,
  213,3,0,0,214,3,0,0,215,3,0,0,216,3,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,41,2,110,110,15,15,255,41,4,71,71,74,8,
  238,238,255,38,7,70,70,70,70,149,75,238,238,238,255,34,11,61,61,64,67,69,72,
  148,72,73,74,238,239,239,255,35,10,71,71,71,73,147,147,139,148,73,239,239,
  239,255,37,8,149,149,139,141,147,141,14,15,79,79,255,37,16,141,141,140,141,
  140,143,109,143,236,73,76,15,73,74,74,73,238,238,255,35,19,150,150,147,68,
  142,106,141,143,108,142,14,110,109,142,68,65,67,68,72,238,238,255,34,4,65,65,
  69,74,239,239,39,16,107,107,143,142,143,142,108,107,106,140,65,67,68,65,65,
  72,238,238,255,39,24,141,141,107,143,107,107,142,105,141,141,69,69,76,68,64,
  66,69,72,72,67,66,66,64,68,79,79,255,39,25,1,1,107,106,108,106,142,143,108,
  143,72,74,1,2,69,66,68,67,69,68,70,74,66,64,64,71,71,255,40,12,109,109,107,
  106,107,106,107,143,143,14,74,238,1,1,53,5,2,2,1,1,239,239,239,61,3,76,76,75,
  238,238,255,41,16,107,107,106,106,108,107,13,143,13,14,76,76,238,239,1,239,1,
  1,255,42,17,107,107,106,107,106,108,109,109,14,149,72,13,236,14,149,143,1,
  238,238,255,42,18,108,108,107,107,107,108,109,108,109,143,142,13,13,69,70,69,
  148,6,239,239,255,42,18,109,109,107,107,107,108,108,109,108,108,107,143,142,
  148,69,67,62,67,15,15,255,43,19,107,107,108,107,108,107,108,108,107,107,107,
  13,15,237,71,63,62,67,73,239,239,255,43,13,143,143,109,107,107,106,107,106,
  107,108,108,109,15,239,239,57,6,69,69,62,61,64,70,15,15,255,43,12,110,110,
  108,107,107,106,106,105,107,108,108,109,238,238,58,7,68,68,63,62,62,66,70,
  238,238,255,44,11,109,109,107,107,106,106,107,107,107,109,110,239,239,59,9,
  239,239,68,62,63,64,66,68,70,76,76,255,44,11,111,111,107,107,106,106,106,107,
  109,110,2,238,238,59,10,1,1,238,15,68,63,66,69,69,238,239,239,255,45,17,108,
  108,106,105,106,106,107,108,110,1,239,239,238,15,15,14,150,74,74,63,5,69,69,
  238,238,239,239,239,255,45,17,107,107,107,105,105,108,14,110,79,239,1,239,
  237,74,74,72,69,238,238,255,46,13,107,107,107,143,109,79,1,239,239,238,76,75,
  74,76,76,255,46,7,5,5,108,141,142,15,1,1,1,255,47,5,5,5,108,140,236,239,239,
  255,48,5,107,107,141,13,15,239,239,255,48,6,108,108,107,13,142,78,1,1,255,49,
  6,108,108,141,108,15,1,1,1,255,49,7,107,107,107,107,13,14,1,1,1,255,49,7,108,
  108,142,141,140,141,79,2,2,255,50,6,106,106,142,138,146,71,79,79,255,50,7,
  107,107,107,140,66,69,238,1,1,255,51,6,142,142,141,146,64,73,239,239,255,52,
  5,142,142,138,67,69,238,238,255,53,4,141,141,146,68,75,75,255,53,4,139,139,
  146,68,72,72,255,54,3,67,67,67,72,72,255,255,255,255,255,255,255,255,255,255,
  255,
};
static const char dogsg8[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,35,1,0,0,36,1,0,0,47,1,0,0,60,1,0,0,73,
  1,0,0,85,1,0,0,96,1,0,0,106,1,0,0,118,1,0,0,130,1,0,0,153,1,0,0,175,1,0,0,
  201,1,0,0,228,1,0,0,0,2,0,0,31,2,0,0,59,2,0,0,81,2,0,0,103,2,0,0,124,2,0,0,
  151,2,0,0,181,2,0,0,210,2,0,0,242,2,0,0,18,3,0,0,50,3,0,0,81,3,0,0,111,3,0,0,
  140,3,0,0,169,3,0,0,190,3,0,0,210,3,0,0,228,3,0,0,247,3,0,0,13,4,0,0,33,4,0,
  0,41,4,0,0,48,4,0,0,55,4,0,0,62,4,0,0,63,4,0,0,64,4,0,0,65,4,0,0,66,4,0,0,67,
  4,0,0,68,4,0,0,69,4,0,0,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,41,6,65,65,147,146,69,148,79,79,255,41,8,69,69,
  68,141,141,141,141,110,2,2,255,42,8,138,138,142,142,107,106,108,15,2,2,255,
  43,7,141,141,107,107,106,105,13,1,1,255,44,6,143,143,107,107,108,107,13,13,
  255,45,5,142,142,107,107,108,13,13,255,45,7,111,111,108,108,7,6,7,7,7,255,47,
  7,110,110,110,111,110,78,79,239,239,255,45,9,111,111,109,108,108,143,143,13,
  72,71,71,55,5,72,72,67,67,70,73,73,255,45,17,108,108,106,107,141,147,140,140,
  141,72,73,69,142,141,139,66,69,74,74,255,44,21,109,109,109,107,142,146,66,67,
  68,147,70,140,140,142,143,139,70,69,67,69,74,238,238,255,43,22,110,110,108,
  107,141,13,67,64,65,66,67,67,140,140,141,1,238,69,67,65,63,67,239,239,255,42,
  23,110,110,110,111,110,14,73,72,69,70,69,69,69,69,236,237,76,237,15,78,141,
  64,62,73,73,255,41,18,111,111,111,5,5,79,239,15,239,1,239,239,238,239,239,79,
  238,15,239,239,61,4,76,76,68,62,69,69,255,40,17,110,110,108,109,142,75,15,
  239,239,239,1,79,239,239,239,1,238,238,238,63,2,63,63,75,75,255,39,17,108,
  108,105,108,141,69,67,71,70,72,74,239,238,239,239,239,79,239,239,255,38,17,
  108,108,13,143,105,141,68,64,66,69,71,73,74,239,238,239,238,239,239,255,37,
  16,15,15,109,108,142,141,140,66,69,66,66,67,69,71,73,77,239,239,255,31,1,65,
  65,65,36,17,14,14,110,110,14,14,142,147,66,70,70,69,67,67,67,66,68,150,150,
  255,31,2,66,66,73,73,35,19,15,15,13,110,108,143,13,71,70,70,71,72,73,239,65,
  65,64,69,72,238,238,255,31,24,139,139,64,239,5,236,13,15,15,14,73,69,69,72,
  72,73,149,238,14,14,72,62,66,69,15,15,255,31,19,236,236,146,239,2,15,74,74,
  14,14,236,73,13,142,74,236,238,14,13,14,14,51,4,66,66,64,64,72,72,255,30,19,
  67,67,2,67,71,239,239,237,150,237,238,238,15,238,142,141,143,13,14,14,14,52,
  4,65,65,62,68,238,238,255,30,18,139,139,79,66,68,71,237,15,15,238,239,238,15,
  13,141,143,14,143,15,15,52,5,239,239,62,66,73,238,238,255,30,18,141,141,15,
  139,67,68,73,238,238,239,239,239,15,13,13,13,14,238,239,239,53,4,67,67,65,69,
  238,238,255,31,18,236,236,67,64,68,149,237,238,238,239,238,238,239,237,74,75,
  237,239,2,2,54,3,65,65,66,72,72,255,31,9,74,74,1,66,67,69,75,238,238,1,1,44,
  5,72,72,73,74,239,1,1,55,2,69,69,69,69,255,31,9,74,74,107,72,65,70,76,238,
  239,1,1,44,5,238,238,74,74,237,239,239,55,2,66,66,71,71,255,31,8,14,14,14,64,
  67,72,2,15,1,1,45,4,72,72,75,75,239,239,255,31,8,110,110,140,146,67,238,239,
  74,2,2,46,3,73,73,74,239,239,255,32,6,156,156,139,147,104,238,69,69,46,3,73,
  73,73,238,238,255,31,7,111,111,111,14,14,1,15,30,30,46,3,71,71,74,239,239,
  255,31,4,8,8,7,1,1,1,36,2,14,14,30,30,46,3,70,70,75,239,239,255,32,2,8,8,1,1,
  36,2,74,74,35,35,46,3,68,68,70,238,238,255,46,3,68,68,67,15,15,255,47,2,65,
  65,71,71,255,47,2,65,65,68,68,255,47,2,72,72,238,238,255,255,255,255,255,255,
  255,255,255,
};
static const char dogsf6[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,33,1,0,0,34,1,0,0,42,1,0,0,51,1,0,0,61,1,0,0,76,1,0,0,94,
  1,0,0,114,1,0,0,135,1,0,0,155,1,0,0,171,1,0,0,185,1,0,0,199,1,0,0,213,1,0,0,
  229,1,0,0,3,2,0,0,38,2,0,0,71,2,0,0,104,2,0,0,132,2,0,0,160,2,0,0,187,2,0,0,
  209,2,0,0,238,2,0,0,8,3,0,0,32,3,0,0,57,3,0,0,88,3,0,0,117,3,0,0,136,3,0,0,
  151,3,0,0,165,3,0,0,172,3,0,0,179,3,0,0,180,3,0,0,181,3,0,0,182,3,0,0,183,3,
  0,0,184,3,0,0,185,3,0,0,186,3,0,0,187,3,0,0,188,3,0,0,189,3,0,0,190,3,0,0,
  191,3,0,0,192,3,0,0,193,3,0,0,194,3,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,53,3,147,147,63,68,68,255,53,4,66,66,65,
  66,239,239,255,52,5,14,14,147,68,69,238,238,255,52,10,108,108,141,68,71,238,
  72,73,75,238,77,77,255,52,13,13,13,143,68,73,239,73,72,73,72,70,70,69,70,70,
  255,50,15,6,6,14,109,141,69,150,239,1,238,76,75,150,75,74,73,73,255,49,16,
  108,108,108,109,107,142,140,237,2,1,238,239,239,1,1,239,239,239,255,48,10,
  107,107,109,109,109,108,143,13,1,1,1,1,64,1,239,239,239,255,47,11,109,109,
  109,108,108,108,109,108,111,2,1,1,1,255,47,9,108,108,109,108,109,108,108,110,
  1,2,2,255,46,9,110,110,107,107,107,108,107,13,1,239,239,255,46,9,107,107,107,
  107,106,107,108,5,2,239,239,255,43,11,5,5,109,109,107,106,106,3,142,13,109,
  79,79,255,42,12,109,109,109,107,107,107,106,3,107,106,108,108,236,236,58,9,
  71,71,69,69,68,66,66,65,66,69,69,255,40,16,111,111,108,108,107,106,106,107,
  106,106,106,106,107,107,141,70,238,238,57,10,73,73,63,62,63,64,64,64,64,66,
  71,71,255,39,28,111,111,108,108,107,107,107,106,106,106,107,106,105,106,142,
  139,72,68,237,65,62,64,68,69,72,73,73,74,238,238,255,38,23,111,111,107,106,
  107,107,106,106,107,107,106,106,106,107,107,106,142,148,67,65,64,67,150,1,1,
  66,1,238,238,238,255,37,23,109,109,108,106,106,106,106,106,107,107,107,109,
  108,107,109,108,13,143,147,66,69,71,78,2,2,255,36,23,108,108,108,141,106,106,
  106,106,106,107,107,108,109,109,109,109,109,108,13,13,149,15,79,1,1,255,35,
  22,108,108,107,142,107,106,106,105,106,108,108,107,109,109,110,5,111,111,15,
  238,79,239,239,239,255,34,17,107,107,142,106,106,106,141,107,107,108,107,13,
  13,13,14,238,239,239,239,255,28,4,63,63,68,238,239,239,33,16,142,142,142,106,
  106,142,106,142,107,142,142,13,142,71,72,76,239,239,255,29,21,149,149,71,70,
  72,140,142,106,141,141,142,142,142,106,3,141,70,67,70,73,239,239,239,255,31,
  19,141,141,67,147,141,142,141,13,108,13,14,109,108,141,67,67,68,69,72,238,
  238,255,31,20,140,140,148,140,140,140,140,14,109,15,76,237,79,73,70,70,72,72,
  72,72,238,238,255,29,11,63,63,69,238,239,147,147,140,73,76,239,239,239,44,2,
  70,70,238,238,47,5,74,74,74,74,76,238,238,255,30,10,70,70,74,238,148,149,149,
  74,238,239,2,2,44,2,70,70,238,238,48,4,73,73,72,73,238,238,255,33,7,72,72,71,
  149,15,238,238,239,239,49,3,72,72,73,238,238,255,34,3,238,238,74,75,75,50,3,
  69,69,75,1,1,255,35,2,142,142,15,15,50,3,68,68,71,238,238,255,50,2,69,69,15,
  15,255,50,2,75,75,238,238,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,
};
static const char dogsn0[] = {
  66,0,71,0,33,0,69,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,26,1,0,0,32,
  1,0,0,39,1,0,0,46,1,0,0,55,1,0,0,64,1,0,0,74,1,0,0,84,1,0,0,102,1,0,0,122,1,
  0,0,145,1,0,0,165,1,0,0,186,1,0,0,206,1,0,0,226,1,0,0,250,1,0,0,15,2,0,0,37,
  2,0,0,62,2,0,0,85,2,0,0,111,2,0,0,134,2,0,0,157,2,0,0,180,2,0,0,203,2,0,0,
  226,2,0,0,249,2,0,0,16,3,0,0,39,3,0,0,62,3,0,0,86,3,0,0,110,3,0,0,133,3,0,0,
  156,3,0,0,179,3,0,0,202,3,0,0,226,3,0,0,250,3,0,0,17,4,0,0,40,4,0,0,63,4,0,0,
  86,4,0,0,109,4,0,0,130,4,0,0,150,4,0,0,170,4,0,0,186,4,0,0,202,4,0,0,217,4,0,
  0,232,4,0,0,247,4,0,0,5,5,0,0,17,5,0,0,28,5,0,0,39,5,0,0,49,5,0,0,56,5,0,0,
  62,5,0,0,68,5,0,0,74,5,0,0,80,5,0,0,255,255,255,255,69,1,180,180,180,255,69,
  1,180,180,180,255,68,2,182,182,180,180,255,68,2,184,184,181,181,255,66,4,68,
  68,44,185,182,182,255,66,4,66,66,72,184,181,181,255,65,5,46,46,66,70,238,181,
  181,255,65,5,46,46,66,68,73,181,181,255,56,1,238,238,238,62,8,7,7,8,184,46,
  67,66,68,184,184,255,55,2,65,65,238,238,61,9,141,141,238,184,184,8,46,65,69,
  184,184,255,55,2,62,62,1,1,58,12,1,1,8,66,140,46,46,184,47,76,65,69,184,184,
  255,55,15,63,63,71,65,66,63,63,46,47,47,184,47,237,68,71,184,184,255,55,16,
  66,66,67,64,64,64,68,47,44,44,44,44,75,68,68,184,184,184,255,56,15,67,67,65,
  63,184,184,46,40,44,44,238,76,68,72,184,184,184,255,56,15,184,184,66,67,65,
  67,46,46,238,239,238,70,70,73,184,182,182,255,55,9,14,14,46,184,67,66,46,72,
  46,238,238,65,6,74,74,71,69,237,185,184,184,255,55,16,141,141,184,63,66,66,
  67,70,76,1,8,71,68,70,44,186,184,184,255,54,17,13,13,147,67,67,68,68,68,73,
  239,8,72,68,68,70,44,186,184,184,255,54,9,141,141,142,139,139,146,66,69,237,
  239,239,64,7,68,68,68,66,238,74,184,183,183,255,53,18,13,13,142,142,139,138,
  66,64,69,15,1,76,68,68,66,70,34,184,183,183,255,53,9,14,14,143,142,146,69,67,
  68,149,238,238,63,8,70,70,65,65,66,69,72,184,183,183,255,53,18,108,108,142,
  138,64,64,146,138,147,14,2,69,65,64,70,65,71,238,182,182,255,53,18,108,108,
  141,147,64,64,146,144,135,14,15,144,67,65,66,66,70,238,182,182,255,53,18,14,
  14,141,67,65,66,66,131,133,142,146,129,144,64,64,67,70,238,181,181,255,53,18,
  13,13,143,146,67,69,68,64,144,130,131,131,70,63,64,70,184,184,181,181,255,53,
  18,13,13,142,146,67,68,71,67,65,144,130,68,70,66,68,47,184,184,181,181,255,
  53,18,143,143,143,68,71,73,74,72,238,69,69,68,71,70,71,47,188,185,181,181,
  255,53,18,108,108,140,66,66,71,71,70,184,67,69,72,38,72,44,47,188,185,181,
  181,255,53,18,108,108,69,65,68,73,38,73,72,70,72,74,38,238,47,47,187,185,181,
  181,255,53,18,108,108,71,184,72,40,44,43,38,38,36,74,238,47,41,186,186,183,
  180,180,255,52,19,46,46,111,74,184,184,43,44,44,40,39,40,237,47,184,184,183,
  183,181,180,180,255,52,19,46,46,5,76,184,180,44,44,184,39,47,44,44,184,41,
  184,184,184,183,181,181,255,53,18,5,5,238,44,47,46,184,180,47,184,44,44,47,
  41,184,184,185,184,182,182,255,53,18,111,111,180,180,184,46,173,180,184,44,
  46,46,47,239,184,185,186,185,183,183,255,53,18,109,109,184,184,47,47,47,178,
  180,184,184,47,44,44,239,47,186,185,183,183,255,53,18,46,46,47,184,47,184,44,
  47,180,178,178,185,185,185,185,180,184,184,183,183,255,52,19,46,46,14,184,
  184,184,47,47,180,178,173,173,173,178,178,180,182,184,181,182,182,255,52,19,
  108,108,237,40,47,184,184,47,47,173,185,185,185,185,40,185,185,180,47,180,
  180,255,52,18,143,143,71,73,40,47,47,44,178,46,44,44,44,44,74,74,1,184,180,
  180,255,52,18,109,109,147,64,70,40,44,44,47,46,44,44,44,44,72,71,238,180,180,
  180,255,53,18,143,143,66,66,69,72,74,239,47,238,237,237,238,184,70,150,2,180,
  180,180,255,53,18,7,7,142,68,67,65,64,70,239,238,237,239,239,74,70,73,73,238,
  180,180,255,54,9,7,7,142,139,65,146,146,69,239,239,239,66,5,74,74,74,65,68,
  180,180,255,56,7,73,73,148,139,147,147,69,1,1,66,5,76,76,34,66,68,180,180,
  255,57,7,238,238,150,140,139,139,149,1,1,66,4,238,238,34,69,73,73,255,58,7,
  71,71,66,141,140,140,73,1,1,66,4,183,183,72,70,238,238,255,59,11,66,66,67,
  147,147,69,72,1,182,72,74,239,239,255,59,11,141,141,70,67,67,69,67,71,1,184,
  184,181,181,255,60,10,139,139,64,68,67,67,65,72,1,182,180,180,255,60,10,13,
  13,66,65,66,70,64,67,74,74,239,239,255,61,10,140,140,138,67,64,15,72,36,69,
  65,70,70,255,62,9,68,68,63,64,68,44,40,72,62,72,72,255,62,7,69,69,65,66,70,
  40,187,185,185,255,63,6,68,68,66,65,73,185,183,183,255,63,6,66,66,64,68,75,
  181,181,181,255,64,5,67,67,71,238,180,180,180,255,67,2,180,180,180,180,255,
  68,1,180,180,180,255,68,1,180,180,180,255,68,1,180,180,180,255,68,1,180,180,
  180,255,255,
};
static const char dogse6[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,37,1,0,0,47,1,0,0,58,1,0,0,70,1,0,0,83,1,0,0,97,1,0,0,113,1,0,0,132,1,0,0,
  155,1,0,0,178,1,0,0,201,1,0,0,224,1,0,0,242,1,0,0,4,2,0,0,23,2,0,0,51,2,0,0,
  83,2,0,0,116,2,0,0,150,2,0,0,185,2,0,0,225,2,0,0,3,3,0,0,34,3,0,0,61,3,0,0,
  88,3,0,0,122,3,0,0,150,3,0,0,172,3,0,0,190,3,0,0,199,3,0,0,206,3,0,0,207,3,0,
  0,208,3,0,0,209,3,0,0,210,3,0,0,211,3,0,0,212,3,0,0,213,3,0,0,214,3,0,0,215,
  3,0,0,216,3,0,0,217,3,0,0,218,3,0,0,219,3,0,0,220,3,0,0,221,3,0,0,222,3,0,0,
  223,3,0,0,224,3,0,0,225,3,0,0,226,3,0,0,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,60,2,148,148,238,238,255,57,5,14,14,141,146,65,64,64,255,
  56,6,14,14,143,141,146,67,67,67,255,55,7,13,13,107,141,141,146,69,73,73,255,
  54,8,110,110,107,106,142,141,69,68,15,15,255,53,9,111,111,108,106,108,141,
  146,67,237,1,1,255,50,11,5,5,7,7,108,107,106,142,140,139,73,239,239,255,49,
  14,109,109,111,5,108,107,107,13,142,142,236,1,2,74,75,75,255,47,18,110,110,
  109,110,110,108,106,108,108,142,109,109,1,1,76,74,74,72,72,72,255,47,18,109,
  109,108,107,107,108,108,107,142,110,1,237,15,238,237,15,14,14,74,74,255,46,
  12,109,109,107,106,107,107,105,140,13,1,1,1,1,1,63,2,238,238,239,239,255,45,
  12,109,109,107,107,106,106,106,107,109,6,2,1,2,2,63,2,15,15,1,1,255,43,13,5,
  5,109,107,107,106,106,106,107,143,108,111,1,1,1,255,42,13,6,6,108,107,108,
  106,106,107,106,107,108,109,110,2,2,255,42,14,108,108,107,107,106,106,106,
  106,106,105,107,108,109,110,238,238,255,41,15,108,108,107,107,107,107,107,
  106,106,105,106,106,107,108,14,76,76,59,4,72,72,69,73,238,238,255,40,27,109,
  109,107,107,106,106,107,108,106,106,106,107,108,108,108,13,236,74,1,68,62,64,
  65,67,68,70,74,239,239,255,39,28,5,5,106,107,106,106,106,107,108,107,107,108,
  107,107,108,13,13,149,70,67,62,61,64,67,67,67,65,65,67,67,255,38,29,5,5,142,
  107,105,105,107,106,108,107,109,109,108,109,108,107,143,141,68,67,70,68,148,
  237,14,72,72,69,69,73,73,255,37,23,109,109,108,107,106,106,107,107,106,109,
  108,109,109,109,109,143,13,14,13,148,73,149,78,2,2,64,3,239,239,71,238,238,
  255,30,2,68,68,238,238,35,23,1,1,108,106,107,106,106,106,106,106,142,141,143,
  14,15,239,111,14,76,14,15,15,239,1,1,65,2,74,74,239,239,255,31,2,72,72,1,1,
  34,16,1,1,108,142,107,106,142,108,106,142,13,142,147,71,73,238,239,239,51,3,
  238,238,238,238,238,255,31,20,148,148,75,150,141,106,106,105,105,142,108,142,
  156,140,69,65,66,69,74,76,1,1,52,2,76,76,237,237,255,32,22,141,141,147,140,
  106,106,140,141,142,142,107,107,141,67,64,63,67,68,69,73,238,73,238,238,255,
  32,22,148,148,147,139,140,141,139,13,108,143,13,111,143,71,70,69,71,71,71,69,
  73,238,238,238,255,29,17,64,64,70,238,239,146,138,141,147,140,14,108,143,149,
  73,76,239,239,239,47,8,239,239,75,74,73,73,73,238,238,238,255,30,12,67,67,72,
  237,70,69,147,148,73,237,15,79,238,238,50,7,76,76,73,72,74,76,238,239,239,
  255,32,7,73,73,69,71,71,73,75,239,239,52,6,76,76,73,72,71,73,15,15,255,33,6,
  70,70,70,72,74,238,238,238,55,3,73,73,74,76,76,255,35,4,72,72,73,238,238,238,
  255,36,2,15,15,239,239,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,
};
static const char dogsd3d7[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,37,1,0,0,47,1,0,
  0,57,1,0,0,67,1,0,0,79,1,0,0,91,1,0,0,104,1,0,0,117,1,0,0,132,1,0,0,150,1,0,
  0,176,1,0,0,197,1,0,0,220,1,0,0,242,1,0,0,17,2,0,0,49,2,0,0,78,2,0,0,103,2,0,
  0,127,2,0,0,150,2,0,0,170,2,0,0,192,2,0,0,212,2,0,0,235,2,0,0,4,3,0,0,31,3,0,
  0,53,3,0,0,73,3,0,0,92,3,0,0,110,3,0,0,128,3,0,0,146,3,0,0,164,3,0,0,184,3,0,
  0,200,3,0,0,216,3,0,0,232,3,0,0,249,3,0,0,11,4,0,0,30,4,0,0,48,4,0,0,64,4,0,
  0,74,4,0,0,83,4,0,0,92,4,0,0,101,4,0,0,109,4,0,0,117,4,0,0,124,4,0,0,125,4,0,
  0,126,4,0,0,127,4,0,0,128,4,0,0,255,255,255,255,255,255,255,255,255,255,255,
  255,40,4,5,5,105,143,2,2,255,40,5,141,141,66,68,7,68,68,255,40,5,67,67,63,68,
  8,76,76,255,40,5,70,70,63,66,34,238,238,255,38,7,148,148,64,66,64,68,73,239,
  239,255,38,7,65,65,64,65,146,70,76,239,239,255,37,8,73,73,64,66,64,147,139,
  74,238,238,255,37,8,68,68,65,139,69,69,69,72,238,238,255,35,10,65,65,66,75,
  14,146,147,146,65,71,15,15,255,33,13,64,64,65,73,239,238,142,139,139,69,69,
  71,149,14,14,255,34,3,65,65,70,239,239,38,9,143,143,141,147,146,147,140,141,
  148,74,74,58,1,68,68,68,255,39,10,141,141,139,147,143,142,142,147,69,14,15,
  15,57,2,65,65,68,68,255,39,11,108,108,140,147,143,106,140,139,67,68,149,15,
  15,56,3,71,71,66,73,73,255,40,11,142,142,141,106,142,141,141,69,69,72,239,
  238,238,56,2,69,69,72,72,255,40,12,108,108,141,142,106,140,143,141,148,72,76,
  79,238,238,55,3,73,73,72,238,238,62,3,69,69,68,75,75,255,41,11,107,107,141,
  141,141,142,110,13,72,79,239,239,239,55,3,75,75,73,238,238,60,5,66,66,65,67,
  69,73,73,255,41,12,109,109,143,142,143,108,107,147,68,67,70,239,239,239,55,8,
  75,75,74,239,70,65,63,66,68,68,255,42,20,107,107,106,105,105,105,69,65,67,69,
  71,73,72,73,72,72,70,67,65,66,15,15,255,42,19,108,108,107,141,104,140,67,66,
  66,69,69,70,71,70,69,67,69,71,68,237,237,255,42,18,108,108,107,106,13,140,64,
  67,68,69,68,69,71,70,71,69,71,73,239,239,255,43,15,108,108,143,107,107,108,
  70,67,64,67,68,69,68,69,71,239,239,255,43,11,109,109,107,108,107,143,141,72,
  239,72,73,238,238,62,2,72,72,238,238,255,44,9,107,107,107,107,142,141,142,74,
  238,239,239,62,2,69,69,76,76,255,44,11,108,108,107,106,107,108,108,143,73,
  237,239,239,239,61,3,14,14,148,237,237,255,44,12,109,109,106,106,108,107,143,
  13,76,238,1,1,1,1,60,4,149,149,143,72,74,74,255,45,12,109,109,106,106,107,
  107,14,149,239,1,239,1,238,238,58,6,238,238,71,149,141,70,75,75,255,45,17,13,
  13,107,107,108,110,110,110,78,239,239,239,79,15,75,74,150,74,74,255,46,15,
  107,107,107,106,107,107,107,107,148,69,237,238,76,76,75,74,74,255,46,14,107,
  107,107,106,107,108,107,141,69,69,71,76,238,238,238,238,255,46,13,107,107,
  107,106,107,107,107,140,140,139,148,149,237,239,239,255,46,13,106,106,107,
  106,106,108,107,143,142,142,141,143,237,239,239,255,46,13,106,106,107,107,
  107,107,107,142,141,141,147,141,72,14,14,255,46,13,108,108,106,107,107,108,
  107,143,143,67,67,69,140,15,15,255,47,7,108,108,13,110,110,109,14,13,13,55,4,
  66,66,63,13,1,1,255,47,4,107,107,143,111,2,2,55,3,63,63,66,13,13,255,47,3,
  109,109,140,14,14,55,4,62,62,65,143,2,2,255,47,3,108,108,139,149,149,55,4,60,
  60,64,71,150,150,255,47,3,108,108,142,149,149,55,5,70,70,63,65,70,238,238,
  255,47,4,106,106,141,142,238,238,56,5,67,67,64,67,69,1,1,255,47,4,107,107,
  142,149,1,1,57,6,65,65,64,66,67,70,74,74,255,47,4,141,141,142,13,15,15,58,5,
  63,63,66,69,72,238,238,255,47,5,13,13,107,13,15,1,1,58,2,237,237,71,71,255,
  47,5,141,141,143,142,236,1,1,255,47,4,141,141,146,141,5,5,255,47,4,140,140,
  146,148,79,79,255,47,4,139,139,66,149,239,239,255,47,3,67,67,66,150,150,255,
  47,3,146,146,67,238,238,255,47,2,70,70,70,70,255,255,255,255,255,255,
};
static const char dogsh7[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,37,1,0,0,45,1,0,0,56,1,0,0,69,1,0,0,86,1,0,0,104,1,0,0,118,1,0,0,
  131,1,0,0,146,1,0,0,160,1,0,0,174,1,0,0,188,1,0,0,214,1,0,0,242,1,0,0,12,2,0,
  0,38,2,0,0,63,2,0,0,82,2,0,0,100,2,0,0,117,2,0,0,134,2,0,0,151,2,0,0,168,2,0,
  0,185,2,0,0,202,2,0,0,219,2,0,0,236,2,0,0,255,2,0,0,19,3,0,0,40,3,0,0,62,3,0,
  0,88,3,0,0,112,3,0,0,137,3,0,0,167,3,0,0,193,3,0,0,221,3,0,0,247,3,0,0,15,4,
  0,0,38,4,0,0,50,4,0,0,60,4,0,0,70,4,0,0,79,4,0,0,87,4,0,0,88,4,0,0,89,4,0,0,
  90,4,0,0,91,4,0,0,92,4,0,0,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,58,1,237,237,237,255,57,3,147,147,69,72,72,255,55,6,14,14,142,65,
  69,238,74,74,255,54,8,109,109,141,68,69,76,73,74,73,73,255,52,12,109,109,107,
  140,146,69,149,1,238,14,150,237,15,15,255,51,13,109,109,108,108,143,141,142,
  15,239,79,79,2,239,238,238,255,50,9,109,109,142,140,143,142,13,111,15,238,
  238,255,49,8,109,109,140,142,140,13,14,237,1,1,255,47,10,109,109,106,139,141,
  141,236,237,239,238,239,239,255,47,9,107,107,140,141,148,72,2,239,239,1,1,
  255,47,9,107,107,140,143,79,2,2,239,239,1,1,255,47,9,106,106,106,13,108,111,
  2,1,1,1,1,255,46,9,108,108,105,105,108,109,109,110,238,238,238,59,8,63,63,64,
  64,67,67,68,72,15,15,255,46,10,106,106,106,107,107,107,107,108,107,13,76,76,
  58,9,63,63,61,65,66,65,65,66,65,70,70,255,46,21,107,107,106,106,106,108,108,
  107,13,143,69,73,66,61,67,148,148,69,69,64,64,72,72,255,46,15,107,107,107,
  106,106,106,107,107,142,140,66,66,68,65,143,14,14,65,2,64,64,72,72,255,46,14,
  107,107,107,106,108,108,107,105,142,141,69,67,69,140,14,14,65,2,65,65,73,73,
  255,46,14,107,107,108,107,108,107,143,140,141,142,140,148,69,142,238,238,255,
  46,13,107,107,107,108,107,108,107,148,147,139,139,72,72,15,15,255,45,12,109,
  109,106,106,107,110,108,108,148,70,71,73,76,76,255,45,12,108,108,106,107,107,
  107,13,14,239,239,2,238,238,238,255,45,12,108,108,106,106,107,108,108,14,74,
  237,239,239,14,14,255,45,12,108,108,106,106,106,142,142,143,73,74,15,239,14,
  14,255,45,12,107,107,108,107,13,142,141,72,74,15,238,238,15,15,255,45,12,107,
  107,106,107,142,106,13,148,73,238,239,239,15,15,255,45,12,106,106,106,143,
  143,143,142,72,76,238,239,239,79,79,255,45,12,107,107,106,105,104,105,141,66,
  65,68,70,15,238,238,255,44,14,108,108,143,106,105,106,106,140,68,68,69,70,69,
  70,15,15,255,44,15,108,108,142,142,143,106,141,69,66,66,69,69,69,69,69,76,76,
  255,44,16,109,109,141,142,142,142,142,68,65,68,68,69,69,71,69,70,73,73,255,
  44,17,109,109,141,143,141,148,148,74,74,69,70,72,69,68,70,69,69,74,74,255,43,
  9,74,74,111,141,107,142,141,69,70,79,79,54,8,15,15,74,70,68,66,67,72,74,74,
  255,43,8,65,65,72,139,141,107,141,148,74,74,56,7,73,73,72,69,69,67,67,73,73,
  255,43,8,149,149,65,140,147,142,141,141,238,238,56,8,72,72,70,69,68,65,64,67,
  239,239,255,41,2,66,66,2,2,44,7,154,154,69,147,67,71,148,1,1,57,8,69,69,70,
  70,68,61,64,70,239,239,255,41,10,139,139,146,2,150,65,2,147,68,70,1,1,58,7,
  67,67,67,70,65,64,67,75,75,255,42,9,140,140,141,15,68,67,70,71,147,239,239,
  58,3,67,67,65,69,69,62,3,66,66,67,73,73,255,43,8,237,237,15,72,67,68,72,149,
  239,239,59,2,64,64,71,71,62,3,68,68,64,15,15,255,43,8,150,150,14,150,146,65,
  68,73,239,239,59,1,64,64,64,62,2,147,147,67,67,255,44,7,238,238,150,238,237,
  67,70,15,15,59,1,70,70,70,62,2,72,72,70,70,255,44,7,149,149,150,79,76,65,68,
  2,2,255,46,5,14,14,68,64,67,2,2,255,46,5,70,70,139,137,138,1,1,255,47,4,108,
  108,105,14,109,109,255,47,3,7,7,8,8,8,255,255,255,255,255,255,255,
};
static const char dogsi0[] = {
  66,0,71,0,33,0,67,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,30,1,0,
  0,31,1,0,0,32,1,0,0,39,1,0,0,48,1,0,0,58,1,0,0,75,1,0,0,100,1,0,0,128,1,0,0,
  163,1,0,0,201,1,0,0,237,1,0,0,16,2,0,0,54,2,0,0,87,2,0,0,115,2,0,0,141,2,0,0,
  162,2,0,0,182,2,0,0,201,2,0,0,224,2,0,0,243,2,0,0,12,3,0,0,33,3,0,0,54,3,0,0,
  82,3,0,0,115,3,0,0,150,3,0,0,184,3,0,0,211,3,0,0,228,3,0,0,250,3,0,0,15,4,0,
  0,31,4,0,0,50,4,0,0,76,4,0,0,99,4,0,0,120,4,0,0,137,4,0,0,148,4,0,0,159,4,0,
  0,169,4,0,0,179,4,0,0,189,4,0,0,197,4,0,0,205,4,0,0,213,4,0,0,214,4,0,0,215,
  4,0,0,216,4,0,0,217,4,0,0,218,4,0,0,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,35,2,64,64,66,66,255,35,4,66,66,102,140,76,76,255,36,
  5,146,146,8,79,79,238,238,255,36,7,8,8,7,2,14,2,237,239,239,48,1,183,183,183,
  255,36,8,7,7,7,14,239,8,67,239,239,239,47,2,183,183,179,179,58,2,68,68,68,68,
  255,36,10,7,7,6,74,2,31,69,239,238,238,239,239,58,3,69,69,66,69,69,65,2,69,
  69,74,74,255,32,17,65,65,68,71,238,107,15,76,8,61,237,239,238,76,75,75,238,
  239,239,58,3,68,68,64,69,69,65,2,66,66,70,70,255,32,19,63,63,138,140,14,79,
  141,1,0,70,238,238,15,73,146,138,140,141,14,238,238,58,4,67,67,66,71,239,239,
  65,2,66,66,71,71,255,35,18,69,69,75,72,238,38,238,238,74,71,67,144,144,144,
  144,134,139,13,238,238,58,9,64,64,66,68,15,76,69,66,66,72,72,255,37,30,179,
  179,189,238,15,72,70,68,64,144,183,129,130,129,135,136,142,15,239,238,74,64,
  62,66,75,70,69,69,69,69,74,74,255,37,2,183,183,179,179,40,27,74,74,13,149,69,
  63,67,183,183,144,128,130,135,140,14,13,239,75,63,62,66,239,70,71,72,72,72,
  237,237,255,38,1,189,189,189,41,23,79,79,13,142,148,67,68,146,144,61,145,144,
  136,14,143,238,64,63,64,71,74,74,74,239,239,255,36,1,179,179,179,42,18,109,
  109,13,142,143,148,69,183,71,69,147,148,13,142,67,63,66,68,76,76,255,34,1,
  189,189,189,43,16,143,143,142,142,142,141,69,70,69,73,73,150,66,64,67,71,14,
  14,255,43,16,142,142,142,141,14,141,70,71,74,76,237,67,67,65,67,73,239,239,
  255,43,15,108,108,107,108,105,146,66,66,65,69,68,68,68,69,70,238,238,255,43,
  14,108,108,142,107,106,69,64,66,66,69,69,68,69,71,15,15,255,41,1,189,189,189,
  43,13,108,108,143,106,142,69,64,65,66,67,68,70,72,74,74,255,41,14,183,183,
  189,109,108,179,183,69,183,68,68,69,69,70,74,74,255,41,2,183,183,189,189,44,
  14,108,108,183,187,183,179,72,70,72,72,74,15,238,238,239,239,255,43,16,189,
  189,108,183,47,187,183,183,73,74,238,238,15,239,239,238,239,239,255,43,16,
  183,183,108,108,183,47,187,149,72,73,75,15,239,239,238,238,238,238,255,42,17,
  183,183,189,108,108,179,183,183,183,71,72,237,238,238,238,239,79,74,74,64,2,
  70,70,239,239,255,41,2,183,183,183,183,45,16,183,183,107,106,183,148,73,239,
  2,79,239,239,237,15,14,239,239,239,64,2,68,68,73,73,255,41,2,183,183,179,179,
  45,10,189,189,109,109,14,149,71,71,239,239,239,239,56,10,238,238,237,75,73,
  72,74,13,149,148,148,148,255,41,1,183,183,183,45,11,108,108,107,106,106,140,
  67,64,68,71,239,1,1,57,9,238,238,76,75,75,149,13,148,148,148,148,255,45,12,
  109,109,107,108,107,138,66,67,67,69,70,72,238,238,59,6,76,76,76,74,73,69,70,
  70,255,46,12,108,108,107,106,139,139,138,146,69,69,69,141,1,1,255,39,1,189,
  189,189,46,12,108,108,108,107,142,141,141,147,139,147,147,141,79,79,255,39,1,
  183,183,183,47,11,108,108,107,107,141,140,68,67,69,138,140,14,14,255,48,11,
  109,109,108,13,141,67,66,68,65,141,143,1,1,255,48,5,13,13,141,14,14,150,150,
  55,5,64,64,145,141,13,239,239,255,48,5,143,143,147,70,238,2,2,56,6,62,62,66,
  139,69,69,75,75,64,2,66,66,70,70,255,49,4,142,142,71,237,1,1,56,10,64,64,64,
  63,65,67,65,64,63,66,72,72,255,49,5,141,141,148,74,239,239,239,58,7,64,64,64,
  63,61,61,63,68,68,255,49,5,140,140,138,141,74,239,239,61,3,64,64,62,68,68,
  255,49,6,106,106,142,141,148,238,239,239,255,49,6,108,108,142,141,140,150,
  239,239,255,50,5,139,139,141,142,14,239,239,255,50,5,146,146,147,140,74,239,
  239,255,50,5,146,146,67,147,75,1,1,255,51,3,65,65,69,238,238,255,51,3,64,64,
  68,238,238,255,51,3,68,68,67,238,238,255,255,255,255,255,255,255,
};
static const char dogsa3a7[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,33,1,0,0,41,1,0,0,63,1,0,
  0,85,1,0,0,110,1,0,0,135,1,0,0,162,1,0,0,190,1,0,0,221,1,0,0,253,1,0,0,21,2,
  0,0,46,2,0,0,69,2,0,0,90,2,0,0,111,2,0,0,131,2,0,0,149,2,0,0,167,2,0,0,185,2,
  0,0,203,2,0,0,221,2,0,0,239,2,0,0,1,3,0,0,19,3,0,0,37,3,0,0,54,3,0,0,70,3,0,
  0,85,3,0,0,100,3,0,0,117,3,0,0,135,3,0,0,154,3,0,0,172,3,0,0,190,3,0,0,208,3,
  0,0,229,3,0,0,0,4,0,0,26,4,0,0,49,4,0,0,69,4,0,0,82,4,0,0,95,4,0,0,109,4,0,0,
  123,4,0,0,140,4,0,0,156,4,0,0,171,4,0,0,184,4,0,0,191,4,0,0,192,4,0,0,193,4,
  0,0,194,4,0,0,195,4,0,0,196,4,0,0,255,255,255,255,255,255,255,255,255,255,
  255,47,1,7,7,7,255,47,3,106,106,141,2,2,255,46,5,148,148,66,67,105,38,38,55,
  2,73,73,71,71,61,2,71,71,239,239,255,46,5,67,67,64,69,8,76,76,55,2,66,66,70,
  70,61,2,69,69,75,75,255,44,7,72,72,68,65,66,68,68,239,239,55,2,67,67,74,74,
  60,3,66,66,66,76,76,255,43,7,72,72,65,65,66,66,69,238,238,54,3,73,73,68,76,
  76,60,2,64,64,69,69,255,42,8,238,238,65,67,65,146,70,73,239,239,54,3,71,71,
  70,239,239,59,3,64,64,65,68,68,255,41,9,70,70,69,65,147,67,146,147,73,238,
  238,54,3,73,73,238,239,239,59,3,62,62,63,76,76,255,39,11,62,62,65,70,76,149,
  70,147,147,69,72,238,238,54,3,74,74,70,75,75,58,4,68,68,63,65,238,238,255,38,
  12,65,65,70,70,239,76,147,147,147,146,67,73,239,239,54,3,72,72,72,74,74,58,4,
  65,65,65,67,239,239,255,43,8,140,140,141,139,67,70,142,150,238,238,54,7,74,
  74,72,15,70,68,70,75,75,255,43,8,13,13,139,138,139,141,141,141,75,75,53,8,76,
  76,74,74,238,66,66,70,237,237,255,43,18,108,108,141,139,142,142,142,139,72,
  239,238,15,76,71,69,68,68,70,239,239,255,44,16,107,107,139,143,143,140,69,67,
  72,15,238,72,69,68,69,72,74,74,255,44,16,108,108,141,107,142,142,69,68,71,76,
  68,70,69,70,72,72,239,239,255,44,15,108,108,141,106,142,142,140,70,70,64,67,
  68,68,69,72,15,15,255,45,13,107,107,141,141,143,142,146,65,64,68,69,69,70,71,
  71,255,45,13,142,142,142,140,142,106,140,69,64,68,69,71,73,238,238,255,44,13,
  108,108,143,106,141,106,107,106,140,68,69,69,75,239,239,255,44,13,108,108,
  106,107,3,3,105,107,141,147,67,238,239,239,239,255,44,13,108,108,106,106,142,
  143,106,142,143,74,238,238,238,239,239,255,44,13,108,108,106,106,106,107,107,
  14,13,149,73,76,238,238,238,255,44,13,108,108,107,107,106,107,142,142,141,72,
  72,76,238,238,238,255,44,13,108,108,106,106,105,142,141,141,143,142,73,74,15,
  238,238,255,44,13,107,107,106,105,106,108,107,142,141,149,72,75,239,239,239,
  255,44,12,108,108,107,106,107,108,143,14,13,74,15,239,1,1,255,44,11,108,108,
  106,106,107,107,108,14,237,239,239,2,2,255,44,10,108,108,106,109,109,108,109,
  109,15,239,2,2,255,44,10,109,109,106,107,109,108,108,107,141,71,72,72,255,44,
  12,108,108,106,106,106,107,107,142,139,147,71,74,239,239,255,44,13,107,107,
  107,105,107,108,107,142,141,141,148,73,150,15,15,255,44,14,107,107,107,106,
  106,106,108,106,142,142,140,141,73,14,1,1,255,45,13,107,107,107,107,108,106,
  142,140,140,147,147,70,141,238,238,255,45,13,108,108,108,107,107,106,142,141,
  67,65,67,69,140,14,14,255,46,13,108,108,108,109,108,142,143,69,65,65,66,64,
  143,111,111,255,46,7,107,107,15,111,2,237,74,1,1,55,5,65,65,61,146,143,1,1,
  255,46,6,109,109,141,2,238,79,1,1,56,6,60,60,63,68,72,70,239,239,64,2,238,
  238,76,76,255,45,7,109,109,141,15,238,238,15,239,239,56,10,64,64,62,64,64,68,
  69,65,70,68,237,237,255,44,7,109,109,143,13,75,238,15,239,239,58,7,65,65,64,
  64,65,64,65,71,71,255,44,7,108,108,143,13,15,239,76,239,239,60,4,68,68,64,64,
  70,70,255,43,8,109,109,141,140,149,239,238,76,238,238,255,43,8,143,143,108,
  107,13,239,239,237,237,237,255,42,9,109,109,141,143,143,14,239,238,237,238,
  238,255,42,9,107,107,145,139,13,15,239,76,236,111,111,255,42,5,107,107,147,
  69,109,79,79,48,3,75,75,70,14,14,255,42,4,143,143,65,68,236,236,49,3,69,69,
  149,1,1,255,42,4,147,147,63,71,238,238,50,2,71,71,237,237,255,42,3,65,65,66,
  150,150,51,1,72,72,72,255,42,2,68,68,68,68,255,255,255,255,255,255,255,
};
static const char dogsb3b7[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,34,1,0,0,42,1,0,0,55,1,0,
  0,65,1,0,0,77,1,0,0,96,1,0,0,121,1,0,0,150,1,0,0,179,1,0,0,208,1,0,0,235,1,0,
  0,6,2,0,0,35,2,0,0,66,2,0,0,90,2,0,0,114,2,0,0,136,2,0,0,157,2,0,0,177,2,0,0,
  196,2,0,0,213,2,0,0,230,2,0,0,247,2,0,0,8,3,0,0,31,3,0,0,55,3,0,0,81,3,0,0,
  107,3,0,0,133,3,0,0,161,3,0,0,177,3,0,0,193,3,0,0,209,3,0,0,224,3,0,0,238,3,
  0,0,251,3,0,0,11,4,0,0,28,4,0,0,49,4,0,0,68,4,0,0,84,4,0,0,94,4,0,0,104,4,0,
  0,114,4,0,0,124,4,0,0,133,4,0,0,142,4,0,0,151,4,0,0,159,4,0,0,165,4,0,0,166,
  4,0,0,167,4,0,0,168,4,0,0,169,4,0,0,255,255,255,255,255,255,255,255,255,255,
  255,45,2,111,111,8,8,255,44,3,15,15,140,139,139,255,44,3,147,147,65,67,67,48,
  1,68,68,68,255,44,5,66,66,64,69,34,15,15,255,42,7,74,74,67,64,65,67,70,238,
  238,255,41,8,14,14,65,64,66,69,69,15,1,1,56,2,66,66,72,72,255,41,8,69,69,65,
  65,67,146,71,238,1,1,56,2,63,63,70,70,61,2,69,69,72,72,255,37,12,65,65,68,70,
  69,69,67,147,67,69,71,76,1,1,56,2,65,65,73,73,61,2,65,65,69,69,255,37,12,67,
  67,67,69,71,238,147,139,69,67,67,76,1,1,56,2,66,66,75,75,61,2,67,67,72,72,
  255,38,11,70,70,71,74,141,137,140,146,67,69,73,79,79,56,2,69,69,239,239,60,3,
  65,65,65,70,70,255,41,9,141,141,140,141,146,67,147,142,74,15,15,56,2,71,71,
  74,74,60,3,62,62,66,15,15,255,42,9,142,142,139,138,141,142,141,140,73,79,79,
  56,2,70,70,70,70,60,3,62,62,65,239,239,255,42,10,108,108,141,139,142,106,142,
  69,67,73,238,238,56,2,72,72,73,73,59,4,67,67,63,67,239,239,255,43,10,143,143,
  141,142,142,140,140,67,68,73,239,239,54,4,76,76,75,74,73,73,59,4,67,67,67,69,
  239,239,255,43,19,108,108,141,142,142,141,139,69,69,71,74,1,238,76,76,69,67,
  66,71,76,76,255,43,19,108,108,106,141,142,141,143,140,72,75,150,73,72,71,69,
  69,68,67,71,1,1,255,44,17,108,108,141,141,141,142,143,139,67,65,69,69,69,71,
  70,69,72,74,74,255,44,16,108,108,142,142,142,107,3,147,66,65,69,69,69,69,72,
  72,75,75,255,44,15,108,108,107,106,3,106,106,140,70,67,67,70,69,69,71,15,15,
  255,44,14,108,108,107,106,105,3,106,141,67,67,69,68,69,70,238,238,255,44,12,
  108,108,106,142,142,142,141,140,66,69,72,238,239,239,255,44,12,109,109,107,
  107,106,107,143,150,150,237,239,1,15,15,255,44,12,109,109,107,107,106,142,13,
  149,72,76,238,239,15,15,255,44,12,108,108,106,106,107,142,143,13,74,150,238,
  1,14,14,255,44,13,13,13,106,106,107,142,13,14,76,2,74,71,13,238,238,65,1,237,
  237,237,255,45,13,107,107,107,106,107,14,69,66,69,71,148,147,141,143,143,64,
  2,75,75,66,66,255,45,15,108,108,107,107,14,147,67,67,146,139,141,69,139,142,
  14,1,1,64,2,67,67,66,66,255,45,21,108,108,107,107,107,141,145,145,140,139,
  139,69,67,69,139,141,141,69,69,64,62,68,68,255,45,21,108,108,107,107,106,142,
  140,140,141,142,67,67,71,63,62,66,65,65,67,65,65,68,68,255,45,11,107,107,107,
  107,107,107,106,143,142,141,69,70,70,58,8,64,64,65,65,66,66,66,69,71,71,255,
  45,11,107,107,107,107,108,108,108,108,143,13,14,239,239,255,45,11,107,107,
  107,107,108,107,108,108,108,109,79,1,1,255,45,11,108,108,107,108,108,108,109,
  110,111,6,238,1,1,255,46,10,106,106,106,108,109,111,6,2,15,79,1,1,255,46,9,
  108,108,107,108,109,110,1,237,75,237,237,255,47,8,141,141,13,79,1,15,75,73,
  237,237,255,47,2,106,106,140,140,51,5,15,15,150,71,72,74,74,255,46,3,108,108,
  140,141,141,52,5,74,74,71,70,73,74,74,255,46,4,108,108,142,72,15,15,53,8,74,
  74,71,147,141,13,150,74,238,238,255,46,4,108,108,141,141,238,238,55,6,67,67,
  68,70,148,74,238,238,255,45,5,108,108,106,141,149,1,1,56,2,67,67,69,69,255,
  45,5,108,108,143,142,143,79,79,255,45,5,107,107,108,107,13,238,238,255,45,5,
  108,108,141,141,13,238,238,255,45,5,142,142,146,139,15,1,1,255,45,4,141,141,
  145,69,15,15,255,45,4,146,146,65,148,238,238,255,44,4,148,148,66,67,76,76,
  255,44,3,67,67,62,68,68,255,45,1,71,71,71,255,255,255,255,255,255,
};
static const char dogsf7[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,34,1,0,0,42,1,0,0,50,1,0,
  0,59,1,0,0,68,1,0,0,78,1,0,0,88,1,0,0,99,1,0,0,110,1,0,0,120,1,0,0,130,1,0,0,
  139,1,0,0,148,1,0,0,167,1,0,0,190,1,0,0,212,1,0,0,237,1,0,0,10,2,0,0,35,2,0,
  0,55,2,0,0,74,2,0,0,94,2,0,0,114,2,0,0,133,2,0,0,150,2,0,0,167,2,0,0,185,2,0,
  0,203,2,0,0,222,2,0,0,241,2,0,0,5,3,0,0,26,3,0,0,49,3,0,0,74,3,0,0,106,3,0,0,
  137,3,0,0,168,3,0,0,198,3,0,0,226,3,0,0,252,3,0,0,22,4,0,0,47,4,0,0,73,4,0,0,
  93,4,0,0,112,4,0,0,119,4,0,0,125,4,0,0,126,4,0,0,127,4,0,0,128,4,0,0,129,4,0,
  0,130,4,0,0,131,4,0,0,132,4,0,0,255,255,255,255,255,255,255,255,255,255,255,
  52,2,68,68,15,15,255,51,3,67,67,68,71,71,255,51,3,147,147,65,69,69,255,51,4,
  141,141,64,69,239,239,255,51,4,142,142,65,69,238,238,255,51,5,142,142,147,
  147,14,239,239,255,51,5,107,107,146,139,78,1,1,255,50,6,107,107,142,142,141,
  14,1,1,255,50,6,108,108,142,108,13,78,239,239,255,50,5,109,109,142,142,14,79,
  79,255,50,5,109,109,107,143,13,2,2,255,50,4,108,108,107,142,14,14,255,50,4,
  108,108,142,143,239,239,255,50,3,108,108,141,143,143,56,7,74,74,74,73,71,73,
  70,69,69,255,50,3,109,109,142,1,1,55,11,76,76,74,72,73,68,67,65,64,64,65,71,
  71,255,49,17,109,109,109,111,2,1,238,76,75,65,62,64,65,65,65,65,65,71,71,255,
  47,20,109,109,108,108,110,111,111,14,239,238,238,61,62,67,69,68,70,65,64,68,
  238,238,255,46,15,109,109,107,108,108,107,108,107,13,73,74,70,62,67,143,238,
  238,62,5,238,238,75,238,66,15,15,255,46,14,108,108,107,107,107,108,107,107,
  142,68,64,65,66,142,109,109,65,2,72,72,74,74,255,45,15,109,109,107,107,106,
  106,107,107,142,140,147,65,67,70,142,79,79,255,45,14,108,108,107,106,107,107,
  108,107,142,142,141,140,148,147,13,13,255,44,15,109,109,107,107,107,108,107,
  107,107,141,141,141,141,148,73,79,79,255,43,15,109,109,108,107,107,109,108,
  108,108,108,148,69,71,73,150,238,238,255,42,14,109,109,108,107,107,107,107,
  107,108,108,108,143,72,72,238,238,255,41,12,109,109,108,107,108,108,107,108,
  108,13,14,239,1,1,255,41,12,108,108,108,109,108,107,108,13,108,13,76,238,2,2,
  255,40,13,109,109,107,108,108,108,14,143,107,143,73,74,238,239,239,255,39,13,
  109,109,107,107,108,109,109,109,14,73,74,238,238,238,238,255,38,14,109,109,
  107,108,106,105,142,140,70,15,238,238,238,239,15,15,255,37,14,110,110,108,
  142,143,104,105,105,69,67,69,69,76,239,14,14,255,36,15,109,109,107,142,142,
  141,107,106,141,69,67,68,69,72,72,238,238,255,35,16,109,109,142,142,107,107,
  142,142,143,137,66,67,68,68,69,71,74,74,255,34,18,109,109,13,140,142,143,141,
  142,142,141,147,67,66,69,70,69,72,73,15,15,255,33,20,14,14,142,139,140,142,
  107,142,141,69,69,71,73,15,239,71,69,69,70,74,239,239,255,28,3,64,64,69,1,1,
  33,20,142,142,142,140,147,142,142,142,71,67,68,148,15,238,76,239,70,68,68,68,
  75,75,255,29,14,147,147,66,238,238,141,139,139,70,70,142,141,148,74,238,238,
  45,8,75,75,72,74,238,67,65,67,74,74,255,31,10,67,67,237,73,73,148,147,69,73,
  237,238,238,45,4,74,74,72,72,238,238,50,4,68,68,69,72,238,238,255,31,9,14,14,
  65,67,139,69,71,70,237,239,239,45,4,74,74,72,72,239,239,50,4,67,67,64,66,73,
  73,255,31,8,14,14,66,67,66,67,148,74,238,238,45,4,70,70,69,73,1,1,51,3,63,63,
  64,70,70,255,32,7,149,149,65,66,68,140,74,239,239,45,3,71,71,68,69,69,51,3,
  64,64,62,68,68,255,32,7,74,74,68,67,68,66,76,15,15,45,3,70,70,69,74,74,52,3,
  64,64,68,15,15,255,34,6,14,14,63,67,239,70,238,238,45,3,71,71,69,1,1,52,3,65,
  65,69,73,73,255,34,4,71,71,60,66,1,1,39,1,72,72,72,45,2,67,67,70,70,52,2,64,
  64,68,68,255,35,3,139,139,146,239,239,45,2,65,65,71,71,52,2,69,69,64,64,255,
  35,2,143,143,13,13,45,2,62,62,68,68,52,2,71,71,67,67,255,45,2,64,64,72,72,
  255,46,1,238,238,238,255,255,255,255,255,255,255,255,255,
};
static const char dogse7[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,31,1,0,0,39,1,0,0,48,1,0,0,57,1,0,0,67,1,0,0,78,1,0,0,88,1,0,
  0,99,1,0,0,110,1,0,0,120,1,0,0,130,1,0,0,139,1,0,0,148,1,0,0,157,1,0,0,169,1,
  0,0,183,1,0,0,209,1,0,0,235,1,0,0,6,2,0,0,34,2,0,0,65,2,0,0,94,2,0,0,122,2,0,
  0,138,2,0,0,155,2,0,0,171,2,0,0,188,2,0,0,206,2,0,0,226,2,0,0,249,2,0,0,18,3,
  0,0,45,3,0,0,76,3,0,0,110,3,0,0,140,3,0,0,169,3,0,0,195,3,0,0,220,3,0,0,238,
  3,0,0,253,3,0,0,11,4,0,0,20,4,0,0,28,4,0,0,29,4,0,0,30,4,0,0,31,4,0,0,32,4,0,
  0,33,4,0,0,34,4,0,0,35,4,0,0,36,4,0,0,37,4,0,0,38,4,0,0,39,4,0,0,40,4,0,0,41,
  4,0,0,42,4,0,0,43,4,0,0,255,255,255,255,255,255,255,56,3,69,69,69,15,15,255,
  56,3,146,146,68,238,238,255,55,4,141,141,66,67,15,15,255,55,4,141,141,62,71,
  239,239,255,54,5,106,106,147,67,149,239,239,255,53,6,109,109,142,146,69,236,
  239,239,255,53,5,106,106,140,140,148,79,79,255,52,6,107,107,143,108,142,13,1,
  1,255,52,6,107,107,140,106,14,15,239,239,255,52,5,108,108,143,141,14,239,239,
  255,51,5,108,108,105,142,143,1,1,255,51,4,108,108,140,143,238,238,255,50,4,
  109,109,108,143,239,239,255,49,4,110,110,110,79,1,1,255,47,7,109,109,109,110,
  5,7,7,1,1,255,46,9,109,109,108,109,108,108,109,110,5,239,239,255,45,10,109,
  109,107,108,106,107,108,108,108,14,14,14,56,7,237,237,237,63,64,66,69,15,15,
  255,45,21,108,108,108,106,108,107,107,107,143,142,147,68,77,64,62,64,66,65,
  65,65,67,15,15,255,44,22,109,109,107,106,106,107,108,107,106,142,141,67,66,
  69,66,147,142,141,68,67,64,65,68,68,255,43,23,109,109,108,107,108,106,108,
  107,143,140,141,142,139,69,69,139,143,14,238,74,68,63,62,68,68,255,42,17,109,
  109,108,107,107,107,110,106,107,142,147,147,140,140,140,69,141,14,14,61,5,
  238,238,150,239,64,66,66,255,42,16,108,108,108,107,108,106,107,110,107,142,
  69,69,69,71,71,141,13,13,62,4,73,73,238,69,67,67,255,41,17,110,110,107,107,
  107,109,107,13,143,239,239,71,68,71,73,74,150,1,1,64,2,71,71,75,75,255,41,11,
  108,108,108,108,107,142,142,143,73,75,1,239,239,255,40,12,109,109,107,108,
  107,109,13,73,75,238,239,239,15,15,255,40,11,108,108,107,13,142,67,66,69,72,
  237,238,15,15,255,39,12,109,109,108,105,105,141,69,67,69,69,70,70,69,69,255,
  39,13,106,106,142,105,106,142,69,67,66,69,69,69,70,70,70,255,38,15,107,107,
  142,142,141,107,141,68,64,68,68,70,68,71,70,71,71,255,37,18,107,107,141,142,
  105,141,141,142,69,67,67,70,72,70,69,70,71,71,1,1,255,36,20,107,107,141,106,
  107,142,142,142,139,71,72,237,239,239,72,70,68,66,69,73,239,239,255,35,22,
  143,143,140,139,142,141,140,141,69,69,69,149,15,79,15,238,76,68,66,70,71,70,
  239,239,255,34,14,141,141,141,147,139,142,140,140,69,66,67,71,76,79,14,14,50,
  8,74,74,73,70,67,66,66,69,239,239,255,29,17,62,62,66,238,238,149,147,141,146,
  67,141,141,140,69,67,139,141,15,15,51,8,71,71,70,72,68,64,64,67,73,73,255,31,
  13,65,65,74,147,147,147,147,69,70,141,139,70,74,15,15,51,8,70,70,69,71,239,
  68,65,67,70,70,255,32,10,68,68,68,147,69,69,65,71,143,76,15,15,52,3,69,69,68,
  238,238,57,3,65,65,68,75,75,255,32,8,66,66,66,66,66,147,147,73,238,238,52,3,
  66,66,65,71,71,58,2,68,68,69,69,255,32,8,143,143,65,65,67,69,71,76,239,239,
  53,2,64,64,72,72,58,2,238,238,73,73,255,33,7,68,68,65,65,65,69,238,1,1,53,2,
  63,63,70,70,255,35,4,64,64,67,70,76,76,53,2,64,64,72,72,255,35,4,65,65,64,
  237,69,69,54,1,238,238,238,255,35,4,139,139,65,73,40,40,255,35,3,109,109,141,
  78,78,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
};
static const char dogsc3c7[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,34,1,0,0,43,1,0,0,56,1,0,
  0,66,1,0,0,78,1,0,0,91,1,0,0,104,1,0,0,120,1,0,0,137,1,0,0,154,1,0,0,167,1,0,
  0,180,1,0,0,194,1,0,0,210,1,0,0,227,1,0,0,252,1,0,0,22,2,0,0,47,2,0,0,68,2,0,
  0,87,2,0,0,113,2,0,0,141,2,0,0,170,2,0,0,185,2,0,0,207,2,0,0,233,2,0,0,255,2,
  0,0,21,3,0,0,42,3,0,0,63,3,0,0,87,3,0,0,112,3,0,0,136,3,0,0,158,3,0,0,178,3,
  0,0,195,3,0,0,203,3,0,0,211,3,0,0,220,3,0,0,230,3,0,0,241,3,0,0,251,3,0,0,5,
  4,0,0,15,4,0,0,25,4,0,0,34,4,0,0,43,4,0,0,51,4,0,0,59,4,0,0,60,4,0,0,61,4,0,
  0,62,4,0,0,63,4,0,0,64,4,0,0,255,255,255,255,255,255,255,255,255,255,255,42,
  2,110,110,8,8,255,41,4,13,13,140,140,111,111,255,41,3,70,70,65,68,68,45,1,68,
  68,68,255,41,5,65,65,63,70,34,15,15,255,39,7,73,73,68,65,65,67,69,238,238,
  255,38,8,150,150,65,64,66,67,69,15,1,1,255,38,8,67,67,66,64,67,69,72,238,1,1,
  255,35,11,69,69,70,69,69,67,68,67,139,71,238,1,1,255,34,12,65,65,66,69,70,
  239,70,147,69,68,69,237,239,239,255,34,12,64,64,70,147,73,148,138,147,146,68,
  67,14,238,238,255,38,8,141,141,140,140,146,68,148,142,237,237,255,39,8,140,
  140,138,146,141,141,140,73,15,15,255,39,9,106,106,140,147,142,141,141,69,70,
  15,15,255,39,11,108,108,140,140,142,142,147,67,67,70,76,238,238,255,40,12,
  141,141,143,107,141,141,69,68,70,73,76,1,239,239,255,40,13,108,108,141,105,
  142,142,141,147,70,71,149,73,74,238,238,57,3,238,238,237,237,237,255,41,13,
  106,106,142,141,141,141,142,69,65,65,69,69,72,75,75,56,4,72,72,72,71,74,74,
  255,41,13,108,108,3,141,140,105,3,139,69,65,66,69,69,72,72,55,3,237,237,238,
  238,238,255,41,16,108,108,142,142,105,106,106,3,69,68,66,65,68,72,73,15,239,
  239,255,42,14,106,106,106,3,3,106,142,69,68,238,68,66,69,72,238,238,255,42,
  21,107,107,106,142,143,143,14,238,239,2,2,65,69,70,70,76,67,66,67,69,70,238,
  238,255,42,23,108,108,107,107,107,107,108,236,73,15,1,237,68,68,69,71,64,66,
  67,67,66,67,14,1,1,255,42,17,108,108,107,106,107,13,143,14,149,75,238,1,1,74,
  72,72,70,238,238,62,3,66,66,72,239,239,255,43,10,107,107,108,108,13,141,13,
  149,73,238,1,1,255,43,11,108,108,107,107,107,108,143,149,71,2,1,239,239,60,2,
  238,238,76,76,255,43,14,109,109,107,107,107,143,13,72,74,70,71,72,73,72,14,
  14,59,3,239,239,150,15,15,255,44,17,107,107,107,106,106,13,148,69,67,68,71,
  149,70,142,1,79,15,15,15,255,44,17,109,109,106,107,108,106,142,146,67,147,
  140,141,146,140,14,15,237,76,76,255,44,16,109,109,106,108,106,106,143,139,
  139,141,141,69,69,146,143,111,238,238,255,44,16,109,109,107,107,107,107,106,
  142,141,141,141,66,237,63,139,109,2,2,255,45,10,107,107,107,107,107,108,107,
  107,142,143,69,69,56,5,65,65,61,69,14,1,1,255,45,10,107,107,107,107,108,107,
  107,107,108,110,239,239,56,6,73,73,62,65,69,73,1,1,255,45,9,107,107,107,108,
  108,108,109,109,110,1,1,57,6,68,68,62,63,68,72,239,239,255,46,7,107,107,108,
  109,110,111,7,1,1,58,6,72,72,62,64,65,68,239,239,255,46,5,109,109,108,109,
  110,2,2,60,6,63,63,64,65,66,69,74,74,255,47,3,141,141,13,1,1,61,5,63,63,67,
  69,70,73,73,255,47,3,107,107,139,236,236,255,47,3,109,109,139,72,72,255,47,4,
  108,108,142,142,75,75,255,47,5,109,109,141,141,148,1,1,255,47,6,109,109,105,
  139,141,237,1,1,255,48,5,143,143,105,13,14,238,238,255,48,5,107,107,107,13,
  143,238,238,255,49,5,141,141,142,142,111,239,239,255,49,5,108,108,146,147,
  236,239,239,255,50,4,139,139,67,149,239,239,255,50,4,13,13,64,72,239,239,255,
  51,3,69,69,67,238,238,255,51,3,147,147,70,238,238,255,255,255,255,255,255,
  255,
};
static const char dogsg7[] = {
  66,0,71,0,33,0,66,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,26,1,0,0,27,1,0,0,28,1,0,0,29,1,0,0,35,1,0,
  0,45,1,0,0,57,1,0,0,69,1,0,0,82,1,0,0,94,1,0,0,105,1,0,0,115,1,0,0,124,1,0,0,
  133,1,0,0,141,1,0,0,155,1,0,0,174,1,0,0,197,1,0,0,226,1,0,0,245,1,0,0,7,2,0,
  0,25,2,0,0,42,2,0,0,68,2,0,0,93,2,0,0,120,2,0,0,150,2,0,0,182,2,0,0,213,2,0,
  0,237,2,0,0,4,3,0,0,29,3,0,0,50,3,0,0,72,3,0,0,100,3,0,0,130,3,0,0,166,3,0,0,
  198,3,0,0,228,3,0,0,0,4,0,0,28,4,0,0,55,4,0,0,81,4,0,0,105,4,0,0,128,4,0,0,
  145,4,0,0,162,4,0,0,180,4,0,0,186,4,0,0,187,4,0,0,188,4,0,0,189,4,0,0,190,4,
  0,0,191,4,0,0,192,4,0,0,193,4,0,0,255,255,255,255,255,255,255,255,255,255,
  255,255,255,43,1,72,72,72,255,42,5,65,65,68,65,73,239,239,255,42,7,67,67,147,
  147,66,69,74,1,1,255,43,7,139,139,142,141,146,147,236,1,1,255,43,8,237,237,
  143,143,141,147,143,15,1,1,255,44,7,238,238,143,108,143,107,109,1,1,255,45,6,
  15,15,108,107,107,107,1,1,255,46,5,109,109,108,105,13,79,79,255,47,4,108,108,
  108,140,15,15,255,47,4,111,111,107,106,142,142,255,48,3,107,107,108,141,141,
  255,48,3,5,5,108,13,13,60,2,68,68,66,66,255,48,4,5,5,14,5,2,2,57,6,72,72,65,
  72,147,70,73,73,255,47,5,109,109,108,109,111,2,2,55,9,74,74,72,70,69,71,143,
  149,149,238,238,255,45,8,109,109,107,108,109,110,111,7,2,2,54,6,76,76,72,71,
  72,72,75,75,62,2,237,237,237,237,255,44,14,109,109,107,108,109,109,109,110,6,
  2,238,237,75,75,75,75,255,44,13,108,108,107,107,107,107,109,109,109,6,1,239,
  79,239,239,255,43,13,109,109,108,108,107,108,106,107,109,108,108,238,1,1,1,
  255,43,12,109,109,107,107,107,107,108,107,107,13,13,143,1,1,255,42,13,109,
  109,108,107,108,107,107,107,3,106,142,140,67,74,74,56,4,74,74,65,66,238,238,
  255,41,20,109,109,107,107,106,107,107,107,142,139,141,142,147,65,67,67,62,63,
  64,66,73,73,255,40,22,109,109,108,107,107,108,107,13,107,139,69,139,141,139,
  69,67,66,67,147,69,64,65,67,67,255,39,25,71,71,108,108,107,107,109,107,143,
  237,147,68,146,147,148,147,147,147,142,143,141,68,66,65,67,238,238,255,39,18,
  108,108,107,107,108,105,143,13,14,238,239,66,68,69,70,67,139,142,79,79,59,5,
  239,239,69,65,64,68,68,255,38,12,108,108,107,107,108,108,13,13,74,239,1,1,
  239,239,52,3,238,238,72,13,13,61,3,66,66,62,69,69,255,37,13,107,107,107,108,
  107,108,13,73,15,239,1,2,239,15,15,62,2,62,62,67,67,255,37,12,108,108,142,
  107,105,105,142,147,68,71,70,74,238,238,62,2,65,65,67,67,255,36,14,108,108,
  107,142,142,105,107,142,70,68,69,71,72,73,238,238,62,2,68,68,72,72,255,35,16,
  108,108,107,141,106,141,143,106,139,67,66,69,69,69,71,74,239,239,255,35,17,
  108,108,142,142,106,142,143,108,139,67,66,68,71,71,72,73,75,1,1,255,30,1,67,
  67,67,34,18,108,108,142,140,107,13,141,141,140,73,75,239,239,238,69,69,71,72,
  76,76,255,30,2,74,74,14,14,34,19,108,108,141,139,107,108,140,71,69,71,237,
  238,238,239,1,69,67,68,72,239,239,255,31,1,63,63,63,33,10,14,14,13,140,147,
  13,143,143,72,71,15,15,44,4,77,77,74,237,239,239,49,4,66,66,67,70,237,237,
  255,31,11,15,15,71,79,15,141,147,149,143,142,237,239,239,44,4,75,75,75,74,
  238,238,50,4,65,65,67,70,238,238,255,31,10,72,72,64,238,238,142,70,68,236,15,
  238,238,45,3,76,76,73,238,238,50,4,68,68,64,64,15,15,255,31,9,14,14,139,74,
  237,143,148,69,14,239,239,45,3,72,72,73,238,238,51,3,64,64,64,72,72,255,31,8,
  13,13,69,66,236,236,148,72,239,239,45,3,71,71,72,238,238,51,4,66,66,64,68,15,
  15,255,31,8,13,13,71,65,67,69,143,75,238,238,45,3,70,70,73,1,1,52,3,65,65,69,
  76,76,255,31,8,15,15,143,65,66,147,236,238,1,1,45,2,69,69,73,73,52,3,65,65,
  67,238,238,255,32,6,72,72,65,67,71,73,238,238,45,2,67,67,74,74,52,3,71,71,66,
  15,15,255,32,6,74,74,15,65,66,74,239,239,45,2,64,64,72,72,53,2,69,69,2,2,255,
  32,6,70,70,65,64,70,2,238,238,45,2,64,64,70,70,255,32,6,140,140,64,65,238,
  239,15,15,45,2,64,64,70,70,255,32,3,106,106,140,146,146,37,1,40,40,40,46,1,
  238,238,238,255,33,1,5,5,5,255,255,255,255,255,255,255,255,255,
};
static const char dogsj0[] = {
  66,0,71,0,33,0,67,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,32,1,0,0,39,1,0,0,52,1,0,0,60,1,0,0,69,1,0,
  0,77,1,0,0,92,1,0,0,112,1,0,0,132,1,0,0,155,1,0,0,173,1,0,0,197,1,0,0,224,1,
  0,0,0,2,0,0,28,2,0,0,60,2,0,0,95,2,0,0,131,2,0,0,161,2,0,0,198,2,0,0,233,2,0,
  0,4,3,0,0,24,3,0,0,49,3,0,0,79,3,0,0,99,3,0,0,122,3,0,0,148,3,0,0,174,3,0,0,
  202,3,0,0,244,3,0,0,26,4,0,0,65,4,0,0,98,4,0,0,115,4,0,0,130,4,0,0,151,4,0,0,
  170,4,0,0,185,4,0,0,200,4,0,0,227,4,0,0,249,4,0,0,13,5,0,0,29,5,0,0,40,5,0,0,
  51,5,0,0,62,5,0,0,73,5,0,0,83,5,0,0,92,5,0,0,98,5,0,0,99,5,0,0,100,5,0,0,101,
  5,0,0,102,5,0,0,103,5,0,0,255,255,255,255,255,255,255,255,255,50,2,64,64,237,
  237,255,50,2,61,61,65,65,255,50,3,62,62,64,76,76,57,1,183,183,183,255,50,3,
  63,63,63,68,68,255,50,4,65,65,62,67,238,238,255,50,3,67,67,64,69,69,255,47,2,
  183,183,183,183,50,4,65,65,65,71,239,239,255,47,2,41,41,37,37,50,4,67,67,67,
  68,238,238,59,1,183,183,183,255,31,1,67,67,67,50,4,66,66,65,68,237,237,58,2,
  37,37,183,183,255,31,3,66,66,138,73,73,50,4,66,66,68,72,239,239,58,3,183,183,
  183,183,183,255,31,4,1,1,15,238,238,238,49,5,71,71,69,68,72,239,239,255,29,8,
  2,2,141,149,239,239,238,239,1,1,47,7,238,238,239,67,68,70,73,1,1,255,28,10,7,
  7,2,71,99,2,239,40,238,239,238,238,45,8,41,41,70,183,73,68,68,69,238,238,255,
  28,11,7,7,236,36,2,239,73,187,187,238,187,239,239,46,7,41,41,179,69,238,68,
  73,239,239,57,1,183,183,183,255,28,13,7,7,148,239,2,76,187,187,70,238,237,74,
  150,1,1,46,6,73,73,65,65,237,1,239,239,255,28,16,15,15,139,239,2,150,187,32,
  15,239,237,187,187,148,13,142,13,13,45,7,73,73,68,66,66,71,238,238,238,255,
  29,23,139,139,239,2,238,32,66,179,189,74,187,131,128,129,131,129,145,134,134,
  65,66,68,76,239,239,58,3,183,183,183,37,37,255,32,20,8,8,238,74,183,179,38,
  67,58,54,55,54,128,130,128,67,64,67,73,14,239,239,54,1,183,183,183,59,2,37,
  37,183,183,255,34,18,150,150,72,189,70,66,57,55,52,52,41,183,130,65,64,65,76,
  13,236,236,54,3,183,183,183,183,183,255,28,1,183,183,183,34,21,179,179,45,71,
  67,146,144,129,54,55,54,183,183,67,65,68,15,142,13,1,183,183,183,56,2,183,
  183,183,183,255,27,1,183,183,183,31,1,189,189,189,36,20,2,2,148,67,66,144,67,
  66,64,64,65,65,67,70,238,142,183,37,37,37,183,183,255,37,16,109,109,13,148,
  69,146,66,146,66,67,183,72,75,1,142,183,237,237,55,2,183,183,183,183,255,38,
  15,14,14,13,140,67,69,71,73,68,67,71,15,143,13,13,237,237,255,29,1,183,183,
  183,38,15,14,14,141,142,69,74,71,38,69,69,73,238,143,13,13,13,13,255,29,2,
  183,183,183,183,39,14,143,143,142,40,36,38,38,71,70,74,238,143,143,143,13,13,
  59,1,183,183,183,255,38,15,189,189,107,14,41,36,36,38,74,73,75,238,236,13,
  143,13,13,255,38,18,189,189,14,238,43,36,39,38,73,74,15,239,150,236,13,15,
  239,238,238,238,255,38,21,189,189,45,238,179,183,41,183,44,44,237,238,150,15,
  14,239,238,238,74,72,237,1,1,255,39,21,189,189,110,183,187,183,179,41,40,73,
  73,76,238,14,1,238,237,71,71,72,149,1,1,255,39,23,183,183,45,183,47,187,183,
  183,39,40,43,44,238,15,239,239,237,236,236,71,69,71,236,1,1,255,38,2,183,183,
  189,189,41,15,109,109,183,47,187,39,41,38,74,238,15,15,239,238,238,237,237,
  57,7,15,15,73,71,68,68,70,239,239,66,1,238,238,238,255,36,2,183,183,183,183,
  41,15,45,45,179,183,183,183,36,73,239,239,237,237,239,238,239,1,1,59,8,238,
  238,70,69,69,139,71,62,72,72,255,36,2,183,183,179,179,40,1,183,183,183,42,13,
  110,110,183,38,38,36,74,239,239,15,237,15,238,1,1,61,6,238,238,73,143,69,65,
  239,239,255,36,1,183,183,183,40,1,189,189,189,42,12,109,109,13,73,72,70,69,
  72,239,238,15,238,239,239,64,2,73,73,74,74,255,42,12,107,107,142,68,66,70,69,
  71,73,238,238,238,239,239,255,43,10,142,142,66,62,62,67,66,69,237,238,239,
  239,255,39,2,41,41,41,41,43,10,110,110,140,66,64,65,65,64,68,73,239,239,255,
  40,1,41,41,41,45,9,13,13,147,69,68,67,66,147,71,238,238,255,46,10,7,7,13,143,
  139,140,140,139,148,238,1,1,255,47,10,147,147,66,74,238,142,141,141,147,73,
  239,239,255,36,1,41,41,41,47,12,142,142,69,72,239,239,73,141,147,68,66,72,7,
  7,60,1,74,74,74,255,47,6,142,142,146,72,239,239,2,2,54,7,69,69,69,69,66,65,
  62,65,65,255,47,6,141,141,147,148,237,238,238,238,56,5,74,74,66,64,64,76,76,
  255,47,6,109,109,141,147,74,15,238,238,59,1,71,71,71,255,47,6,141,141,139,
  148,71,238,238,238,255,47,6,67,67,141,140,73,238,238,238,255,46,6,66,66,146,
  147,70,74,15,15,255,45,6,69,69,65,68,66,73,15,15,255,45,5,65,65,63,69,73,238,
  238,255,45,4,64,64,66,74,15,15,255,46,1,15,15,15,255,255,255,255,255,255,255,
};
static const char dogsl0[] = {
  66,0,71,0,33,0,67,0,16,1,0,0,17,1,0,0,18,1,0,0,19,1,0,0,20,1,0,0,21,1,0,0,22,
  1,0,0,23,1,0,0,24,1,0,0,25,1,0,0,31,1,0,0,43,1,0,0,56,1,0,0,77,1,0,0,103,1,0,
  0,128,1,0,0,151,1,0,0,174,1,0,0,200,1,0,0,224,1,0,0,246,1,0,0,15,2,0,0,41,2,
  0,0,68,2,0,0,93,2,0,0,119,2,0,0,143,2,0,0,170,2,0,0,197,2,0,0,223,2,0,0,248,
  2,0,0,22,3,0,0,52,3,0,0,78,3,0,0,103,3,0,0,128,3,0,0,153,3,0,0,177,3,0,0,200,
  3,0,0,226,3,0,0,248,3,0,0,14,4,0,0,37,4,0,0,60,4,0,0,86,4,0,0,111,4,0,0,134,
  4,0,0,155,4,0,0,178,4,0,0,201,4,0,0,223,4,0,0,245,4,0,0,5,5,0,0,27,5,0,0,49,
  5,0,0,68,5,0,0,84,5,0,0,94,5,0,0,104,5,0,0,113,5,0,0,121,5,0,0,127,5,0,0,128,
  5,0,0,129,5,0,0,130,5,0,0,131,5,0,0,255,255,255,255,255,255,255,255,255,66,1,
  180,180,180,255,61,2,189,189,189,189,66,1,180,180,180,255,60,3,180,180,185,
  189,189,66,1,180,180,180,255,46,4,8,8,8,8,2,2,60,3,180,180,185,189,189,66,1,
  184,184,184,255,45,7,8,8,7,15,15,141,238,237,237,59,5,180,180,180,65,185,68,
  68,66,1,184,184,184,255,45,8,1,1,14,38,44,47,15,239,238,238,62,3,63,63,65,72,
  72,66,1,184,184,184,255,44,9,180,180,137,38,46,44,40,238,239,239,239,62,5,64,
  64,64,72,180,184,184,255,44,9,64,64,65,47,44,44,44,47,239,239,239,62,5,65,65,
  65,15,180,184,184,255,45,9,67,67,44,44,46,46,47,47,238,1,1,62,3,67,67,67,74,
  74,66,1,180,180,180,255,45,9,239,239,2,73,44,46,46,47,239,239,239,61,6,180,
  180,65,67,69,1,180,180,255,46,8,179,179,189,73,44,44,44,238,238,238,62,5,67,
  67,67,69,239,180,180,255,47,8,179,179,68,71,38,40,76,237,239,239,62,3,75,75,
  69,73,73,66,1,180,180,180,255,47,8,147,147,68,66,71,74,74,72,238,238,61,4,69,
  69,70,1,72,72,66,1,180,180,180,255,47,9,15,15,68,69,68,38,73,70,148,239,239,
  61,4,67,67,64,238,73,73,66,1,184,184,184,255,47,9,1,1,140,147,71,73,73,72,
  183,238,238,60,7,75,75,64,64,237,15,180,184,184,255,48,9,143,143,139,67,71,
  71,144,131,183,15,15,59,8,47,47,68,68,68,71,2,180,184,184,255,48,19,108,108,
  139,67,70,146,144,129,129,136,14,239,75,67,69,68,67,238,184,180,180,255,48,
  16,13,13,140,66,66,72,144,129,128,134,139,14,14,62,66,67,70,70,65,2,180,180,
  180,180,255,48,16,108,108,141,139,70,73,69,64,128,131,133,140,67,61,63,68,
  239,239,65,2,180,180,184,184,255,48,16,108,108,143,13,73,73,71,66,66,144,133,
  67,64,63,185,180,239,239,66,1,180,180,180,255,48,15,108,108,141,142,150,40,
  72,71,147,68,239,64,65,65,67,238,238,66,1,185,185,185,255,46,1,189,189,189,
  48,15,13,13,13,14,40,40,76,73,72,185,185,66,66,70,74,14,14,66,1,180,180,180,
  255,46,1,189,189,189,48,15,13,13,15,44,44,38,72,71,68,185,185,66,68,73,142,
  14,14,66,1,180,180,180,255,47,16,189,189,13,15,44,44,45,183,72,185,185,69,68,
  69,239,143,239,239,66,1,184,184,184,255,47,15,109,109,14,238,44,179,45,183,
  180,73,70,69,70,238,143,143,143,66,1,180,180,180,255,47,15,110,110,108,238,
  45,2,183,45,179,40,73,71,238,143,13,14,14,66,1,184,184,184,255,47,15,189,189,
  109,183,183,179,2,45,183,44,40,238,236,142,13,239,239,66,1,184,184,184,255,
  47,14,189,189,108,183,183,179,2,45,183,46,46,44,14,143,14,14,66,1,184,184,
  184,255,47,13,180,180,183,180,179,179,2,2,45,185,185,46,238,14,14,66,1,184,
  184,184,255,46,16,183,183,189,180,183,179,45,45,2,45,179,46,47,47,238,237,
  239,239,66,1,180,180,180,255,46,17,47,47,111,183,179,179,183,179,2,44,183,47,
  47,47,238,73,72,75,75,255,47,17,109,109,183,179,179,183,179,2,47,183,47,47,
  47,44,73,69,69,1,1,255,46,18,47,47,108,180,44,183,45,45,2,180,180,238,239,76,
  76,74,148,70,14,14,255,47,18,143,143,183,44,183,183,180,45,47,180,237,239,
  237,237,237,143,69,69,2,2,255,47,13,108,108,148,38,44,183,180,46,47,47,238,
  238,239,239,239,61,4,14,14,70,68,1,1,255,47,12,110,110,141,72,183,44,180,44,
  47,239,239,239,239,239,62,4,73,73,68,148,2,2,255,48,10,110,110,142,183,44,
  180,44,40,238,239,239,239,62,4,75,75,68,67,238,238,255,49,8,109,109,142,74,
  74,150,73,73,239,239,63,4,70,70,70,69,1,1,255,49,8,78,78,79,70,72,139,139,
  139,73,73,63,6,75,75,148,141,1,71,15,15,255,50,8,139,139,71,73,72,139,139,
  140,239,239,64,6,14,14,142,147,62,67,239,239,255,50,9,69,69,68,70,15,180,141,
  140,141,238,238,65,4,74,74,147,68,238,238,255,50,10,141,141,65,68,238,238,68,
  140,147,68,239,239,65,3,75,75,69,238,238,255,50,11,13,13,147,141,72,76,150,
  66,68,66,67,239,239,255,50,12,108,108,140,147,69,75,238,238,69,67,65,147,238,
  238,63,1,76,76,76,255,51,6,142,142,140,69,70,76,238,238,58,7,73,73,64,62,63,
  62,64,238,238,255,51,6,142,142,139,139,72,75,238,238,60,4,65,65,62,65,67,67,
  255,51,6,142,142,139,141,148,75,238,238,61,1,71,71,71,255,52,5,66,66,146,66,
  74,238,238,255,52,5,70,70,67,68,76,238,238,255,53,4,65,65,70,74,238,238,255,
  53,3,68,68,63,76,76,255,55,1,238,238,238,255,255,255,255,255,255,
};
// replaced sounds with CC-BY 3.0 and CC0 1.0 sounds from GZDoom
static const char dsdgsit[] = {
  3,0,34,86,154,51,0,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  127,127,127,127,127,127,127,127,127,127,127,127,126,126,126,126,126,126,127,
  127,127,127,127,127,127,127,127,128,128,128,128,128,128,128,128,128,129,129,
  129,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,127,127,127,
  126,126,126,126,126,126,126,127,127,127,127,127,127,127,127,127,127,127,127,
  127,127,127,127,127,127,127,127,127,127,127,127,126,126,126,125,125,125,124,
  124,124,124,124,124,124,124,124,124,123,123,123,123,123,123,123,123,123,123,
  123,123,122,122,122,122,122,122,122,122,122,122,122,121,121,120,120,119,118,
  117,116,115,113,111,109,107,106,104,103,102,103,105,107,110,114,119,125,130,
  135,139,143,146,148,150,151,151,151,151,149,147,144,141,137,133,130,127,124,
  123,122,121,121,121,121,121,121,121,121,121,120,120,119,119,118,117,117,116,
  117,117,119,121,123,127,129,133,137,140,144,147,149,150,151,151,151,149,147,
  145,142,139,135,131,128,126,123,120,118,117,115,114,114,114,114,113,114,114,
  114,114,115,115,117,118,120,122,125,127,128,130,131,132,134,135,136,138,139,
  140,141,142,143,143,142,142,142,141,141,141,140,140,139,137,136,135,133,132,
  132,131,131,131,131,131,131,132,133,133,134,135,136,137,138,139,140,140,141,
  141,141,140,140,140,141,141,141,142,142,143,143,143,143,143,143,143,142,141,
  140,138,136,134,132,129,127,124,122,120,117,115,113,110,108,106,104,101,98,
  95,91,87,83,79,74,70,66,63,62,62,64,69,75,83,94,103,113,123,130,138,145,153,
  161,172,181,192,199,205,204,200,194,183,173,165,157,152,151,148,145,140,133,
  125,116,108,102,100,99,100,100,101,100,96,95,93,92,94,98,102,106,112,115,120,
  125,129,137,144,154,162,169,174,177,179,178,179,178,178,178,177,174,170,162,
  154,144,135,127,120,115,109,104,97,91,83,77,72,70,71,75,80,87,95,102,109,116,
  124,130,138,145,154,162,170,177,184,188,190,190,189,184,180,174,170,165,163,
  162,159,158,155,152,146,142,136,131,128,126,122,121,118,114,110,107,104,101,
  101,101,101,100,100,96,92,89,84,76,70,63,51,41,27,24,24,24,23,23,23,24,23,28,
  67,114,158,211,231,230,232,233,234,234,235,235,236,236,236,235,235,233,230,
  193,143,103,77,66,56,55,49,43,34,24,23,24,24,24,24,24,23,26,34,37,39,43,49,
  52,62,68,67,64,56,50,43,49,62,93,137,188,231,235,236,236,236,236,235,234,234,
  234,235,235,235,233,233,215,155,94,41,23,25,25,26,26,26,25,26,25,26,25,35,59,
  88,111,121,125,117,103,92,86,86,98,122,141,156,163,162,144,127,109,88,75,74,
  73,71,81,88,92,105,127,139,157,179,193,200,209,213,212,223,232,231,232,231,
  229,227,201,145,86,36,22,24,23,26,25,44,78,101,116,124,123,118,114,114,121,
  130,136,145,150,145,139,128,111,101,90,86,84,88,93,102,120,135,157,184,220,
  233,232,233,233,235,236,236,234,233,231,227,202,178,155,134,110,77,41,26,27,
  24,25,24,25,25,26,26,26,25,26,24,37,70,95,117,134,154,176,208,229,231,235,
  236,238,238,237,236,236,235,234,232,231,230,224,182,129,80,35,23,24,22,22,20,
  19,17,18,17,17,17,17,18,20,22,24,24,25,24,24,27,42,64,98,137,189,232,233,235,
  236,238,240,242,243,242,242,241,239,236,234,232,231,219,164,112,56,21,23,22,
  23,22,23,23,24,24,24,25,25,26,26,26,25,24,24,24,23,22,22,23,23,24,26,25,70,
  175,232,232,236,236,238,236,236,234,235,234,235,234,238,236,239,232,235,138,
  28,26,21,24,23,25,23,23,22,22,21,23,23,25,26,26,30,51,72,67,78,90,112,143,
  200,233,231,235,233,236,236,238,236,237,235,236,235,235,209,117,32,23,21,18,
  17,17,19,20,22,22,25,26,27,45,86,127,158,195,219,221,214,198,190,166,149,149,
  161,174,191,207,215,234,236,235,234,234,235,234,235,234,235,231,234,205,146,
  113,70,35,26,26,23,23,22,24,23,25,25,27,25,50,97,142,209,234,231,233,233,232,
  231,232,229,230,216,145,66,24,24,23,25,24,27,26,27,26,28,25,41,94,146,212,
  234,232,233,232,232,232,232,231,228,229,184,96,37,20,25,23,39,96,157,216,240,
  237,240,237,237,233,234,231,232,233,233,227,168,95,37,22,22,17,18,18,22,22,
  24,23,25,22,23,21,23,22,26,24,27,27,35,31,86,209,244,241,241,237,235,232,231,
  227,228,229,232,230,231,228,234,185,52,29,29,30,25,57,140,212,238,235,214,
  157,99,40,28,26,25,23,21,20,20,19,20,22,24,24,25,25,23,24,29,56,112,174,226,
  235,237,237,238,239,238,243,245,241,234,229,226,228,228,228,226,231,233,234,
  234,202,87,22,18,9,11,13,17,15,12,13,20,32,80,118,117,101,90,105,133,180,232,
  242,240,240,239,240,237,232,220,192,183,210,228,229,229,228,191,141,72,35,41,
  35,34,28,20,11,18,14,43,123,188,220,220,212,183,144,102,85,96,109,123,122,
  121,113,84,54,43,48,82,146,209,220,227,237,243,248,247,242,242,245,247,244,
  224,122,31,19,12,5,4,12,8,11,18,23,27,31,32,33,29,30,24,23,19,40,105,179,226,
  241,244,242,205,163,140,156,210,234,243,253,249,245,232,238,230,239,169,44,
  37,31,30,32,30,82,155,201,221,203,161,92,44,23,21,24,44,94,135,171,192,202,
  191,163,114,69,38,19,21,20,22,22,21,27,38,38,29,36,77,128,171,218,230,233,
  234,229,223,222,235,234,240,241,236,227,231,230,130,26,30,25,41,18,79,220,
  229,222,231,235,248,184,52,31,21,17,12,10,10,15,20,20,22,25,22,13,17,30,35,
  29,27,30,28,31,20,69,198,239,242,241,243,239,239,239,232,228,227,226,227,230,
  234,235,241,241,241,215,84,20,24,16,19,21,25,21,89,188,223,217,232,212,117,
  29,26,19,15,20,30,35,37,27,25,18,18,16,18,17,22,17,22,22,51,202,243,248,0,0,
  254,239,239,239,241,239,239,206,204,232,224,233,237,241,153,30,14,14,8,6,1,2,
  2,9,30,175,236,234,236,240,130,32,35,26,23,23,19,64,221,224,235,233,243,230,
  241,117,22,33,20,22,19,23,112,226,227,238,244,247,240,241,236,174,104,97,138,
  210,226,224,226,228,233,200,35,19,17,23,18,32,18,103,239,226,242,237,238,226,
  215,62,19,30,19,16,5,12,22,121,218,232,228,232,134,27,33,21,24,13,31,20,143,
  249,232,240,222,230,225,233,197,144,157,226,235,235,238,246,245,252,235,243,
  116,1,25,13,22,13,36,16,105,236,226,238,226,242,112,11,33,15,18,9,18,16,51,
  129,167,148,60,32,36,32,19,21,4,95,225,215,234,237,252,254,0,254,243,236,236,
  204,132,103,140,188,229,233,231,226,135,22,19,13,17,11,23,18,34,14,93,208,
  215,226,237,182,44,32,29,32,34,34,40,173,231,221,228,223,201,74,16,17,23,32,
  68,192,249,240,244,241,242,236,235,230,223,230,213,177,155,152,162,136,106,
  57,19,27,23,20,10,19,14,26,8,124,232,217,238,238,253,238,237,122,30,34,24,26,
  24,27,20,35,31,19,18,20,27,31,26,28,17,72,203,237,245,251,244,238,235,228,
  228,186,92,63,70,144,217,234,232,243,237,250,173,48,30,25,12,11,9,43,182,228,
  219,225,216,231,173,38,29,39,36,28,22,21,25,28,21,12,24,36,20,26,31,31,54,
  179,241,243,243,239,237,233,236,231,238,241,183,141,159,215,220,216,218,228,
  228,241,190,64,37,29,19,15,16,18,23,33,97,138,152,121,46,24,22,15,13,15,17,
  14,14,15,17,26,83,159,213,225,227,232,231,239,238,244,251,0,255,253,244,244,
  234,237,159,66,40,71,145,188,218,209,136,38,24,22,20,16,13,13,17,22,34,105,
  164,178,151,85,30,25,29,25,20,12,9,11,12,4,23,90,120,135,154,165,206,241,239,
  244,253,0,0,253,249,245,237,232,216,103,22,23,41,109,172,217,226,201,140,58,
  31,25,18,21,16,13,17,30,76,111,93,59,28,22,23,16,16,19,17,16,13,15,17,36,129,
  214,230,225,224,231,230,233,233,235,239,247,249,0,254,0,245,246,141,13,18,17,
  31,31,56,113,137,124,92,37,19,20,18,18,16,24,65,107,108,99,74,34,28,35,36,31,
  22,19,15,10,12,10,52,141,201,226,205,162,135,102,91,107,192,241,229,236,235,
  240,238,240,241,239,230,170,81,37,24,22,15,20,29,51,62,68,79,83,81,94,95,119,
  133,133,128,113,65,24,20,11,7,5,9,15,21,26,27,47,124,178,211,203,165,143,142,
  145,204,226,223,231,231,235,237,238,239,240,240,234,203,171,138,93,66,65,71,
  66,57,82,106,107,128,175,186,196,210,215,208,172,118,93,59,16,13,18,26,30,33,
  37,34,29,28,72,137,140,84,31,27,31,26,25,33,171,243,232,233,235,244,244,242,
  236,221,155,113,102,121,168,220,237,213,164,124,84,71,101,134,192,236,241,
  251,250,249,224,124,41,31,23,12,1,1,1,1,4,8,21,21,98,212,237,228,220,215,224,
  231,232,233,228,230,232,241,245,248,244,246,192,116,87,82,96,113,138,162,187,
  173,127,70,35,26,24,19,17,20,26,98,197,229,232,201,101,34,24,25,23,33,26,146,
  245,227,233,229,236,229,236,210,71,15,22,15,23,20,23,16,56,145,195,210,190,
  159,123,85,84,151,221,229,233,238,231,239,180,58,28,32,30,31,22,53,127,198,
  227,227,234,238,236,235,230,244,207,117,54,24,22,25,73,87,91,66,24,23,25,23,
  18,17,18,18,26,85,105,112,102,47,51,81,96,166,230,236,239,238,234,229,228,
  225,226,231,219,181,169,189,218,234,234,237,235,212,164,121,95,70,46,36,59,
  80,72,63,41,22,21,24,24,25,24,23,25,32,30,36,117,199,220,217,222,232,234,236,
  237,240,229,198,172,186,217,228,232,234,234,234,233,229,229,231,231,227,231,
  227,217,182,126,54,25,29,20,22,15,12,17,20,24,24,73,167,212,223,202,155,95,
  42,28,35,76,177,243,242,246,236,239,226,227,171,39,19,23,25,30,34,117,202,
  227,221,228,227,240,239,218,187,180,182,173,173,168,160,140,95,37,31,30,23,
  19,16,13,15,32,178,246,236,239,233,233,232,165,90,89,81,120,183,214,209,216,
  225,208,185,137,76,61,41,20,21,21,31,52,67,84,114,131,128,145,161,159,189,
  226,230,235,237,230,226,227,202,119,38,19,23,20,24,27,35,39,49,120,172,199,
  209,189,138,109,83,62,62,90,120,164,223,235,242,245,231,201,192,183,190,224,
  226,218,220,218,228,237,245,244,240,149,46,27,25,23,19,16,20,18,21,21,18,15,
  17,9,7,7,21,20,103,222,240,249,245,244,243,247,235,191,169,159,191,227,220,
  226,229,234,236,241,235,238,183,77,32,27,30,76,130,164,168,147,64,5,15,20,28,
  23,22,20,21,82,183,229,240,200,143,77,28,20,23,28,78,138,195,228,227,232,231,
  229,233,235,226,201,184,177,177,189,186,164,154,140,126,118,108,80,54,30,20,
  19,43,101,167,228,226,227,224,229,201,125,52,31,29,25,33,28,49,99,146,175,
  172,154,136,112,101,108,131,163,214,233,236,241,248,224,158,101,55,29,18,19,
  18,24,49,73,90,109,127,123,121,120,103,92,76,50,49,74,108,143,188,224,224,
  225,232,231,205,168,137,121,123,133,130,129,140,141,103,64,32,21,20,20,25,39,
  85,123,154,175,167,141,119,76,34,28,26,22,51,125,209,234,236,239,236,229,226,
  204,153,137,152,199,231,235,233,234,216,135,51,17,20,24,32,32,32,25,24,26,28,
  25,26,67,121,174,207,223,222,206,189,195,201,213,221,219,224,228,238,220,158,
  113,78,80,109,159,220,230,232,233,227,154,57,13,15,19,24,26,29,29,28,30,34,
  39,53,92,121,140,153,156,152,137,111,104,113,134,174,219,230,232,234,230,227,
  225,225,225,230,237,241,240,226,180,135,104,61,47,59,80,119,145,156,166,147,
  98,55,34,32,28,27,26,34,60,62,74,60,25,25,25,23,23,26,29,61,114,163,206,227,
  219,219,226,215,216,221,221,225,232,235,235,233,233,228,231,195,129,95,50,28,
  32,34,35,30,33,72,110,120,116,121,114,80,38,28,30,29,30,73,132,178,204,204,
  202,174,143,131,132,145,179,221,232,230,229,231,223,170,116,76,50,47,83,118,
  152,192,218,231,229,222,198,177,145,118,99,78,55,33,31,37,57,89,118,133,133,
  110,85,41,18,21,21,23,43,102,154,207,228,212,187,136,75,36,25,27,51,84,110,
  135,148,151,153,151,151,157,174,193,203,200,186,174,170,171,179,194,216,227,
  212,164,105,40,23,26,26,27,28,95,175,222,228,227,202,160,114,85,92,117,154,
  210,238,234,234,228,199,160,118,90,86,83,87,88,86,78,74,78,92,124,147,179,
  207,214,206,190,162,128,110,105,120,144,170,180,177,149,95,34,23,23,22,21,22,
  19,46,98,136,166,161,148,135,117,107,113,124,149,178,194,206,214,209,204,198,
  177,153,126,92,77,69,70,72,90,115,148,177,189,194,191,175,142,110,84,62,59,
  65,73,85,103,118,128,128,129,131,141,158,174,190,206,219,215,202,182,165,153,
  142,128,124,129,125,116,107,102,98,95,104,121,140,150,164,177,173,164,151,
  145,144,141,150,157,159,153,137,119,99,77,62,60,65,69,82,96,99,96,88,89,95,
  103,123,144,152,155,145,119,89,55,29,28,25,27,51,95,140,169,187,197,194,182,
  177,178,182,196,212,226,229,218,205,190,172,148,129,118,106,103,114,122,125,
  112,93,74,60,54,49,46,48,56,68,79,90,106,135,161,176,198,220,230,229,215,195,
  172,140,111,94,82,83,94,112,130,148,159,162,164,159,157,171,186,200,220,226,
  225,212,191,163,137,110,82,64,58,54,44,46,49,48,56,72,80,95,108,115,115,118,
  125,132,145,152,157,168,173,167,163,151,127,96,74,65,73,87,104,131,149,157,
  162,157,143,137,132,134,140,153,167,174,164,143,117,83,57,41,30,30,33,40,57,
  75,93,110,124,141,161,182,195,200,196,184,168,144,117,99,101,106,117,139,160,
  167,166,160,145,127,106,85,80,88,91,102,111,117,116,113,108,106,113,118,129,
  146,162,166,167,166,153,138,127,124,131,138,142,143,137,125,116,112,110,118,
  142,176,211,229,229,229,218,182,139,97,65,46,40,49,69,91,106,118,120,114,107,
  97,81,79,88,95,106,127,140,150,163,161,151,143,127,111,102,93,90,100,114,134,
  159,176,183,187,183,164,138,116,102,89,78,75,84,108,131,148,158,158,150,137,
  122,110,105,108,124,144,158,168,177,180,176,163,143,123,109,99,95,91,93,96,
  100,104,103,105,113,117,119,119,115,119,128,133,127,122,124,125,127,124,119,
  114,113,115,119,127,131,139,149,157,157,156,158,159,155,152,149,142,138,139,
  137,133,130,127,127,127,118,110,114,118,120,127,133,136,143,152,150,147,145,
  137,130,122,104,84,79,74,77,89,108,125,140,148,147,143,134,119,117,124,132,
  140,154,165,168,165,152,139,130,125,123,130,137,142,147,146,142,140,136,130,
  124,126,125,121,118,118,115,114,117,127,141,150,155,159,161,152,139,127,114,
  102,102,108,113,120,133,145,147,141,136,129,117,102,96,94,91,92,96,106,115,
  122,129,136,137,134,135,134,132,134,136,135,138,140,140,141,140,138,139,142,
  142,142,143,144,140,141,139,139,145,151,158,160,158,149,138,124,106,84,73,71,
  71,82,99,109,117,121,118,112,106,100,100,106,110,119,133,142,149,158,158,156,
  158,159,160,157,148,139,132,124,121,123,128,135,139,141,142,134,122,112,106,
  102,107,124,141,150,152,153,148,137,122,110,110,114,121,127,132,133,128,115,
  101,91,84,87,99,113,128,143,155,160,159,155,154,151,148,147,144,140,138,136,
  130,118,112,116,118,120,128,131,128,123,113,105,104,106,117,131,141,147,150,
  149,139,123,110,105,107,116,128,135,136,138,139,134,122,110,106,108,114,119,
  126,134,137,140,144,146,143,143,143,136,128,122,114,106,107,109,114,121,127,
  133,136,128,113,103,94,86,86,98,117,140,165,190,204,207,204,196,184,165,151,
  143,141,140,141,141,140,143,141,135,131,126,122,119,116,112,113,113,119,128,
  136,139,140,139,134,129,125,120,116,113,113,117,125,135,146,161,171,172,169,
  164,153,141,128,119,117,120,128,138,149,154,156,152,144,137,125,115,110,105,
  103,108,112,118,122,123,121,120,116,114,115,117,119,121,124,129,133,137,141,
  146,151,155,154,151,147,138,124,114,113,115,120,128,136,144,147,145,141,135,
  125,117,115,117,119,123,128,131,133,131,129,130,128,122,117,114,109,102,98,
  101,104,108,114,121,126,129,129,127,123,118,118,122,128,136,143,149,154,153,
  147,141,133,125,120,117,118,124,131,134,132,131,126,118,111,108,106,104,105,
  108,113,117,118,120,122,123,125,126,127,128,129,129,131,129,124,119,113,107,
  102,102,103,109,119,128,138,145,148,147,147,146,143,144,148,149,147,142,135,
  124,116,110,106,107,111,117,124,127,126,127,128,129,132,139,147,152,153,152,
  148,139,129,123,119,117,116,120,126,129,130,130,129,126,123,123,126,128,134,
  142,147,152,155,155,154,153,149,144,136,128,123,119,115,115,119,124,131,139,
  145,147,146,141,135,130,126,124,126,127,128,131,135,134,130,127,124,119,117,
  117,118,120,124,127,130,133,135,137,139,137,134,131,129,129,131,133,135,138,
  143,144,143,142,143,141,135,129,123,119,114,112,114,120,126,133,140,145,145,
  142,137,133,128,126,125,128,133,136,136,135,130,124,121,117,116,120,126,132,
  141,146,143,140,135,129,121,114,115,120,126,132,139,145,147,148,146,140,134,
  127,120,114,108,104,105,106,107,112,117,121,127,132,135,135,134,134,133,134,
  134,134,135,138,141,139,135,129,122,118,114,111,112,115,119,123,124,125,128,
  129,128,127,125,125,125,125,125,125,127,128,132,136,139,142,144,143,139,133,
  130,127,122,118,117,118,121,125,130,133,134,133,134,133,131,132,134,134,135,
  136,135,135,133,129,128,126,121,119,120,121,122,124,126,128,129,131,132,131,
  131,129,127,124,118,113,113,116,119,125,131,136,140,142,140,135,132,128,124,
  121,120,120,124,129,131,133,133,132,134,132,128,128,127,127,128,127,126,126,
  126,125,124,121,120,120,120,120,117,115,115,115,118,122,126,129,134,136,136,
  136,135,134,135,139,142,144,146,148,148,145,140,136,135,135,133,132,134,134,
  135,136,134,131,128,126,124,121,120,120,123,128,129,131,134,136,137,138,136,
  135,136,136,136,137,137,136,135,133,131,131,131,131,129,128,127,126,126,125,
  128,131,132,133,135,138,139,139,138,136,133,132,132,133,132,131,128,126,123,
  121,119,118,118,120,124,128,131,134,137,136,133,131,131,133,136,137,138,140,
  142,144,141,139,136,134,131,128,126,123,122,122,122,124,126,129,136,141,145,
  147,148,148,146,142,139,137,135,134,134,134,134,134,134,132,130,128,127,126,
  125,125,124,124,125,127,128,130,134,138,141,143,144,143,140,136,131,126,123,
  119,116,116,115,116,118,121,124,128,131,136,140,143,144,143,143,142,138,134,
  129,127,126,125,124,125,124,122,119,118,115,112,112,113,115,117,120,124,127,
  129,130,128,126,123,120,118,119,120,125,131,135,138,139,136,132,129,126,123,
  123,124,126,127,128,128,129,129,130,130,130,129,128,127,124,120,117,115,115,
  116,120,123,127,129,131,129,124,122,120,119,120,124,129,134,137,136,135,132,
  129,127,124,124,125,126,126,127,126,125,125,125,127,130,132,134,136,135,132,
  129,127,123,120,119,119,121,123,126,128,129,129,130,129,127,124,120,116,113,
  111,112,114,115,118,121,122,124,125,126,128,129,131,134,137,139,139,139,139,
  138,138,137,135,134,132,130,128,127,125,125,127,129,131,133,133,131,129,125,
  121,120,119,120,124,128,131,135,136,136,135,134,132,132,133,135,137,139,139,
  139,136,134,132,132,132,135,136,136,135,132,127,122,119,117,117,120,123,128,
  132,134,135,135,134,133,134,136,137,139,140,140,141,139,136,135,134,132,130,
  128,126,125,125,126,125,126,128,129,132,135,137,138,138,136,133,129,126,125,
  124,124,125,126,128,130,129,128,127,127,127,128,130,132,137,141,144,144,142,
  140,138,134,129,126,125,125,127,128,130,132,133,134,135,134,132,131,129,126,
  124,122,121,123,126,128,129,130,130,130,128,123,121,120,120,122,125,126,127,
  129,130,130,128,127,127,128,127,126,127,129,131,133,135,137,137,138,137,134,
  130,127,124,122,120,119,119,121,123,123,122,122,123,124,124,123,121,120,119,
  118,118,119,122,126,129,129,129,129,129,130,129,128,129,128,130,132,133,134,
  136,136,134,134,132,129,128,127,126,125,124,126,128,128,128,128,126,124,122,
  122,123,124,126,127,128,129,129,129,128,127,126,126,126,126,125,126,126,125,
  124,124,124,126,128,130,131,130,129,128,127,126,126,126,127,128,128,127,128,
  129,128,128,128,128,129,129,128,126,126,126,126,125,126,127,127,126,125,122,
  120,118,119,119,119,120,123,127,129,131,132,134,136,136,135,134,133,132,131,
  130,129,128,129,130,129,129,129,130,130,131,131,131,132,133,133,133,135,135,
  136,135,134,132,131,129,127,127,126,126,125,125,125,125,124,123,122,123,125,
  128,131,135,139,142,143,141,140,137,134,131,128,125,125,124,124,125,127,128,
  129,129,128,128,128,129,129,131,132,134,135,136,137,136,137,139,140,142,142,
  140,137,134,129,124,120,118,118,121,123,125,126,128,128,126,124,123,123,124,
  127,132,136,139,140,141,140,139,137,136,134,132,131,130,129,128,128,127,127,
  127,127,127,128,128,129,129,129,129,130,130,130,131,132,132,132,132,133,134,
  134,133,131,129,127,126,124,123,123,123,124,125,126,127,128,128,128,128,128,
  128,128,129,130,132,132,133,133,134,134,132,131,130,128,126,124,122,120,119,
  119,120,120,122,123,126,127,129,130,132,134,135,136,136,136,134,133,132,130,
  129,128,127,126,124,122,120,119,117,116,116,117,118,122,126,128,130,132,132,
  132,132,132,131,130,130,130,129,128,129,128,126,125,124,124,124,124,125,125,
  125,124,123,124,125,127,128,131,133,134,134,131,129,128,125,124,124,125,127,
  128,129,130,129,128,126,124,122,121,121,123,124,125,126,127,127,127,128,128,
  129,131,134,135,135,134,133,131,130,128,127,126,125,125,125,123,122,122,123,
  124,125,127,130,132,134,134,133,132,129,128,126,125,126,126,127,127,127,127,
  127,127,126,126,127,129,130,132,132,132,131,131,130,129,128,128,129,130,131,
  130,130,130,129,129,128,127,126,125,124,124,125,126,128,129,129,129,130,130,
  130,130,130,130,130,130,130,129,128,128,129,129,129,128,128,127,126,125,125,
  125,126,128,130,132,132,131,129,128,126,125,124,124,124,127,129,131,133,134,
  133,133,132,130,129,128,128,128,128,127,128,128,128,128,128,128,128,130,132,
  132,132,131,130,130,129,129,129,128,128,127,126,125,125,125,126,127,128,130,
  132,133,132,131,128,126,124,124,125,126,127,129,130,131,131,130,129,128,128,
  128,129,129,130,130,129,128,127,125,124,124,124,125,126,127,128,129,130,130,
  130,130,130,131,130,129,129,127,126,126,125,124,125,127,128,128,128,128,129,
  128,128,128,128,127,127,127,127,128,128,128,128,128,128,127,126,126,125,125,
  126,126,126,127,128,128,130,130,129,129,129,129,129,129,129,128,128,127,127,
  127,127,127,126,126,124,123,122,121,121,121,122,125,128,131,134,136,137,136,
  135,133,130,128,128,127,127,127,128,127,126,125,123,121,120,120,121,122,123,
  125,127,128,130,132,133,133,132,131,130,129,128,127,126,126,126,126,126,127,
  127,126,126,125,125,125,125,125,126,127,129,131,132,133,132,132,131,130,128,
  128,127,127,128,129,130,131,131,130,129,128,126,126,125,126,128,129,130,132,
  134,134,134,134,132,131,129,128,127,126,126,127,127,128,128,129,129,129,128,
  128,128,127,127,128,128,129,130,131,132,132,130,129,129,128,128,128,128,128,
  128,128,128,128,127,128,129,130,131,132,132,131,131,130,130,130,129,128,128,
  128,128,128,128,128,128,127,128,128,128,128,129,130,131,133,135,136,136,136,
  135,134,132,129,127,126,125,123,123,124,125,127,128,129,130,131,132,133,133,
  132,131,130,129,128,128,128,129,129,129,129,128,128,128,128,127,128,128,128,
  129,129,129,130,129,129,130,130,130,130,130,129,128,128,127,127,126,127,127,
  128,128,128,128,128,128,128,128,128,128,128,128,129,130,131,132,132,132,131,
  130,129,128,126,125,125,125,125,126,128,129,130,130,130,130,130,130,129,129,
  129,129,130,130,129,129,129,128,128,127,126,125,125,125,125,126,127,128,128,
  129,129,129,129,129,129,128,128,128,128,127,127,127,127,127,126,125,125,125,
  126,127,127,128,128,128,128,128,128,129,129,130,130,130,129,128,128,127,126,
  126,125,125,125,125,125,126,125,126,126,126,127,128,128,128,128,128,129,128,
  128,128,129,128,128,127,126,126,126,126,126,127,127,128,128,128,128,128,128,
  129,129,130,130,130,130,129,129,128,127,128,128,128,128,128,127,127,127,127,
  127,128,128,128,129,129,129,129,129,128,128,128,128,127,126,125,124,124,125,
  125,125,126,127,128,128,130,131,131,132,132,131,131,132,132,131,131,130,128,
  128,128,127,127,127,127,127,127,128,128,128,129,129,129,129,128,128,128,128,
  128,128,128,128,128,128,128,127,127,126,126,126,127,128,128,129,129,130,130,
  130,130,130,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,129,131,132,133,133,132,131,130,129,128,127,126,126,
  126,126,127,127,127,127,128,127,127,128,128,129,130,130,131,131,130,129,129,
  128,128,128,128,127,127,128,128,128,127,127,127,128,128,129,129,130,130,131,
  131,131,131,132,133,133,133,132,131,129,128,126,125,124,124,125,126,126,127,
  127,127,126,126,126,127,128,128,129,130,131,131,131,130,128,127,126,125,125,
  126,126,127,128,128,128,129,129,129,129,129,129,129,130,130,130,130,129,129,
  128,128,128,128,128,128,128,127,127,127,127,127,127,127,127,127,127,127,127,
  127,127,127,127,128,129,130,130,130,129,129,128,128,128,129,129,129,128,128,
  128,127,126,126,127,127,128,128,128,128,127,127,126,126,126,127,127,128,128,
  128,129,129,128,128,127,127,126,126,126,126,126,126,126,126,126,127,127,128,
  128,128,128,128,128,128,128,128,128,128,128,128,127,127,126,126,125,125,125,
  126,127,127,128,128,128,128,128,128,128,128,129,130,130,130,130,129,129,128,
  127,127,126,126,127,127,127,127,127,127,126,126,126,127,127,128,128,128,129,
  129,129,129,129,129,128,128,128,128,128,127,127,127,128,128,128,128,128,128,
  128,128,128,128,128,128,129,129,129,129,129,128,128,128,127,127,127,127,128,
  128,128,128,129,129,128,128,128,128,128,128,128,129,129,129,129,130,131,131,
  131,131,131,130,129,128,128,127,127,127,127,128,128,129,130,131,131,131,131,
  131,131,131,131,131,130,130,129,128,128,127,127,127,127,127,127,128,129,130,
  131,132,132,131,131,131,130,130,130,129,129,128,128,127,127,127,127,127,128,
  128,128,129,129,129,130,129,129,129,129,129,129,129,129,129,128,128,128,128,
  128,128,128,128,128,128,128,128,127,127,127,127,127,128,129,130,130,131,131,
  130,130,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,129,
  129,130,130,130,130,129,129,129,128,128,128,128,128,127,127,127,127,127,127,
  127,127,127,127,127,127,127,128,128,128,129,129,129,129,128,128,128,128,128,
  129,129,129,129,129,129,128,128,127,126,125,125,125,126,126,127,127,128,128,
  129,129,129,129,129,128,128,128,128,128,128,128,127,127,127,127,127,127,127,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,128,
  128,128,128,128,128,128,127,127,127,127,127,127,127,127,127,127,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,128,128,128,
  128,128,128,128,128,128,129,130,130,130,130,129,129,128,128,128,128,128,128,
  128,128,128,128,128,127,127,126,127,127,127,127,128,128,129,129,129,129,129,
  129,129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,
  128,128,128,127,127,127,127,127,127,127,127,127,127,127,127,128,128,128,128,
  128,128,128,128,128,128,129,129,129,129,129,129,128,128,127,127,127,127,127,
  128,128,128,129,129,129,129,128,128,128,128,128,129,129,129,129,128,128,128,
  128,128,128,128,128,129,129,129,129,129,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,127,127,127,127,127,127,128,128,128,128,129,
  129,129,129,129,129,129,129,129,128,128,128,128,128,129,129,129,129,128,128,
  128,128,127,127,128,128,128,128,128,128,128,127,127,128,128,128,128,128,128,
  128,128,127,127,127,128,128,128,129,129,129,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,129,129,129,128,128,128,128,127,127,
  126,126,126,126,126,127,127,127,127,127,127,127,128,128,128,128,129,129,129,
  129,129,129,129,128,128,128,128,128,128,128,128,128,128,127,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,129,128,128,128,128,128,127,127,127,
  127,127,127,127,128,128,128,128,128,128,128,128,129,129,129,129,129,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,127,127,127,127,127,127,128,128,
  128,128,128,128,128,128,129,129,130,130,130,130,129,129,129,128,128,128,128,
  128,128,128,127,127,127,128,128,128,128,129,129,129,129,129,129,129,128,128,
  128,128,128,128,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,
  128,128,129,129,129,129,129,129,128,128,128,128,128,128,128,129,129,129,129,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,129,129,129,129,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,127,127,127,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,127,127,127,127,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  127,127,128,128,128,128,128,128,128,128,128,128,127,127,127,127,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,
  128,128,128,128,128,128,127,127,127,128,128,128,128,128,128,128,128,128,128,
  128,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,127,127,127,
  127,128,128,128,128,128,128,128,128,128,128,128,127,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,127,127,127,127,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,129,129,129,129,129,129,128,128,128,128,128,128,128,128,129,128,128,128,
  128,128,128,128,128,128,128,129,129,129,129,129,129,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,
  127,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,
  129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,129,129,129,129,129,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,
  129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,
  129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,129,129,129,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,
  129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,
  129,129,129,129,129,128,128,128,128,128,128,128,128,129,129,129,128,129,128,
  128,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,129,
  129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,129,129,129,129,129,129,129,129,129,129,129,129,129,129,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,
  129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,
};

static const char dsdgatk[] = {
  3,0,34,86,161,44,0,0,120,120,120,120,120,120,120,120,120,120,120,120,120,120,
  120,120,120,113,113,110,110,107,106,104,103,102,102,103,105,107,110,113,118,
  122,127,131,136,140,143,146,148,149,150,151,151,151,150,149,147,144,141,138,
  134,131,128,125,123,122,121,121,121,121,121,121,121,121,121,121,121,121,120,
  120,119,119,118,117,117,117,117,118,119,121,123,126,128,131,134,137,140,143,
  146,148,149,150,151,151,150,148,146,144,141,138,135,132,129,127,124,121,119,
  117,116,114,114,113,113,113,113,113,113,113,114,114,114,115,116,117,119,121,
  123,125,127,128,130,131,132,133,135,136,137,138,140,141,142,143,143,143,143,
  143,142,142,142,142,142,141,141,140,139,137,136,135,134,133,132,132,131,130,
  130,130,130,130,131,132,133,134,136,137,138,139,140,140,141,141,141,141,142,
  142,142,142,142,142,142,142,142,142,142,143,143,143,143,143,143,142,142,141,
  140,138,136,134,132,129,128,126,123,121,119,116,114,112,111,109,107,105,104,
  101,99,97,93,90,86,82,78,74,70,67,63,62,62,62,66,70,76,84,93,102,110,120,128,
  134,142,149,157,164,174,182,191,198,203,206,202,198,192,181,172,165,158,152,
  150,149,146,144,139,133,126,119,111,105,101,99,99,100,100,101,101,98,96,95,
  93,93,95,98,102,106,111,115,118,122,126,131,138,144,153,160,166,172,174,176,
  176,175,175,174,174,175,174,172,169,164,156,149,141,133,127,121,117,112,107,
  102,97,91,84,80,75,73,74,76,80,86,92,100,107,113,120,127,132,139,144,152,158,
  165,172,178,184,188,190,191,191,188,183,179,174,169,164,162,160,158,156,154,
  151,147,142,138,133,128,125,124,121,119,117,115,113,110,108,105,102,101,100,
  101,101,101,100,97,95,93,89,84,80,76,67,56,44,32,26,25,23,22,21,21,22,24,23,
  32,70,111,152,200,231,234,234,234,233,233,234,235,236,237,237,237,237,236,
  235,234,230,198,153,115,86,73,63,58,57,52,46,38,28,26,26,26,26,26,25,26,25,
  27,34,38,40,42,47,51,53,62,66,65,63,58,52,45,44,51,66,96,135,180,222,231,231,
  232,233,234,233,234,234,234,234,235,234,234,233,232,232,209,155,101,52,27,26,
  26,26,27,27,27,26,25,25,25,25,28,43,67,92,110,120,123,118,105,95,87,84,86,
  100,120,136,151,159,160,150,135,122,107,90,83,81,81,77,83,90,92,97,112,129,
  139,157,177,193,199,207,213,213,215,226,232,232,233,233,232,232,222,185,132,
  81,40,28,30,29,30,29,35,61,88,105,117,124,124,120,115,114,117,126,132,138,
  147,152,149,144,137,123,111,103,94,91,88,91,96,102,118,130,146,166,193,221,
  230,230,231,231,232,232,233,232,232,232,231,227,205,182,161,139,119,92,60,30,
  21,21,20,20,19,18,17,18,19,20,22,23,25,27,34,62,91,114,133,148,166,184,210,
  230,231,231,231,231,231,232,233,232,234,234,235,234,235,232,233,219,177,129,
  86,44,25,25,22,22,19,19,18,17,17,17,17,17,17,18,19,20,22,23,25,26,27,27,29,
  35,50,70,99,135,179,224,233,233,234,235,237,238,239,239,239,238,237,236,235,
  233,233,232,233,222,175,128,80,37,27,26,24,22,21,21,20,19,18,17,16,16,16,16,
  16,16,15,16,18,20,22,24,28,34,49,72,101,135,176,214,236,240,240,240,241,240,
  239,238,237,236,236,236,236,233,234,219,176,127,82,40,20,22,21,22,22,23,23,
  23,22,21,21,21,22,23,24,24,24,24,24,24,26,31,37,47,68,95,128,167,205,229,232,
  234,235,236,236,236,236,236,234,234,232,231,228,230,217,176,129,87,44,24,24,
  21,22,21,22,21,23,22,22,22,22,22,23,23,25,30,37,46,56,68,83,97,104,116,132,
  152,170,185,203,208,222,235,234,235,235,237,236,236,234,233,231,232,230,230,
  196,145,113,76,40,25,25,25,24,23,22,22,22,23,24,26,26,29,63,103,146,205,231,
  231,232,233,233,234,235,235,234,233,222,163,89,36,23,24,22,23,23,23,22,21,21,
  22,24,24,54,104,154,214,236,234,235,234,234,233,233,230,230,226,228,196,120,
  56,23,22,22,23,41,94,146,201,231,231,232,231,231,230,231,231,233,233,234,232,
  233,206,139,78,31,24,24,23,22,20,20,19,18,18,19,19,20,21,22,21,22,22,24,25,
  27,29,32,37,50,71,99,132,171,207,228,232,234,234,235,234,233,233,233,232,233,
  232,233,231,232,218,175,128,85,44,26,28,27,28,26,26,24,23,23,23,21,20,18,17,
  16,16,15,16,16,17,19,22,25,29,32,34,34,35,45,79,129,182,228,236,237,238,237,
  236,233,233,232,232,231,231,230,225,174,110,53,26,27,24,24,23,24,24,24,23,23,
  22,21,20,20,19,20,20,21,23,24,26,30,35,45,63,88,116,150,186,217,232,232,233,
  232,233,232,232,230,230,220,200,195,210,229,232,232,233,217,172,116,48,21,19,
  17,16,15,16,19,21,24,24,58,131,189,230,234,233,211,180,138,104,89,95,98,109,
  115,115,112,90,60,36,39,48,90,149,216,235,237,242,243,245,245,246,243,242,
  239,236,232,232,191,90,25,18,15,13,13,13,14,17,20,22,24,25,26,26,25,24,24,24,
  24,26,28,53,114,174,222,231,230,233,231,230,231,234,234,235,234,234,231,225,
  171,105,48,22,24,23,27,29,31,32,31,27,24,22,21,20,21,21,21,21,23,24,26,26,29,
  34,43,66,96,127,157,184,199,195,176,144,107,73,38,25,23,22,24,23,24,23,23,23,
  25,24,40,78,126,166,212,227,227,222,204,178,140,111,98,98,98,110,110,115,104,
  81,50,43,48,74,126,193,238,236,239,237,236,233,233,232,233,232,233,234,236,
  226,149,55,27,25,22,21,20,19,20,22,25,26,26,25,25,22,22,22,22,22,26,30,98,
  205,236,237,240,241,239,236,233,232,233,233,233,233,234,234,234,236,237,236,
  236,234,215,109,31,30,27,28,28,28,28,32,84,176,227,236,234,234,192,101,31,26,
  23,20,18,21,24,27,27,28,26,26,25,26,25,22,19,16,14,15,16,46,189,233,230,239,
  240,246,248,250,248,246,241,236,231,216,196,222,235,233,234,231,234,181,60,
  24,22,20,18,16,14,16,16,21,22,107,220,229,232,230,221,119,23,21,11,15,11,22,
  20,109,234,240,246,244,242,238,235,214,68,12,20,14,18,17,24,63,189,239,237,
  239,240,235,234,228,228,190,122,87,101,157,220,238,242,239,242,234,240,136,
  16,25,11,11,7,12,13,33,176,246,235,250,241,247,234,239,116,21,23,16,16,14,15,
  17,23,100,201,234,236,238,216,102,25,22,18,13,12,16,22,36,188,243,237,245,
  238,240,233,236,222,169,134,168,222,233,232,239,239,246,242,247,237,234,104,
  16,26,13,14,9,11,18,24,136,238,229,244,234,245,186,43,18,17,10,12,11,20,23,
  37,102,159,171,123,39,21,20,21,21,24,25,45,179,240,234,241,238,244,243,246,
  244,242,238,235,233,198,134,103,131,171,213,234,231,231,214,122,28,21,19,17,
  15,15,18,21,26,22,102,223,230,234,229,224,117,20,24,15,19,14,22,21,114,223,
  235,239,240,240,209,98,27,27,21,21,22,83,195,230,230,237,241,242,243,245,243,
  241,236,236,213,180,153,147,154,143,120,93,46,22,23,18,16,14,10,12,13,20,23,
  152,243,232,245,240,244,237,237,210,94,22,23,19,23,23,27,26,31,46,37,24,26,
  21,21,19,22,21,25,28,108,216,236,238,242,245,246,242,239,232,229,171,98,68,
  71,118,195,234,233,238,239,237,233,173,49,20,20,17,17,20,23,64,195,235,233,
  234,234,230,230,133,26,23,17,16,14,16,15,16,17,19,21,22,34,42,27,28,25,25,26,
  88,202,238,239,244,247,250,249,245,242,235,234,217,161,128,143,203,232,234,
  239,239,239,235,236,192,72,22,18,14,15,15,18,21,27,28,81,132,153,148,91,33,
  21,17,13,12,12,13,13,14,14,15,17,27,80,156,218,236,233,234,231,234,235,238,
  240,244,246,248,247,247,244,245,239,234,162,75,46,57,119,168,202,219,197,122,
  38,21,20,14,14,11,14,17,23,25,35,97,149,167,156,112,52,19,22,19,19,17,20,19,
  22,23,27,28,59,108,122,130,143,151,173,219,237,238,242,243,245,245,245,245,
  243,242,238,239,209,108,42,40,49,97,153,195,218,206,171,110,42,22,24,23,24,
  26,26,27,25,37,70,95,86,57,34,20,22,22,21,18,17,15,15,17,20,22,26,73,156,221,
  242,239,237,235,236,237,241,242,244,246,249,248,245,239,236,232,228,155,69,
  42,58,123,173,208,225,202,128,43,24,17,9,14,17,24,26,25,19,25,84,130,138,128,
  94,41,18,21,18,19,21,24,22,23,25,28,29,37,100,171,212,223,199,161,136,111,86,
  88,128,214,234,227,227,227,234,241,243,244,244,238,226,159,85,39,27,25,26,30,
  32,38,54,59,62,70,75,77,89,95,98,119,122,122,121,110,75,32,27,26,24,22,20,20,
  17,13,11,15,22,83,149,190,213,199,157,134,138,143,185,230,234,234,233,232,
  235,237,237,232,230,229,233,219,195,173,143,105,81,75,69,72,65,72,91,105,102,
  117,160,178,182,194,202,205,204,173,136,111,81,37,20,18,14,12,10,16,17,18,20,
  25,27,80,146,188,208,191,153,134,141,146,187,237,245,241,239,236,236,234,237,
  236,233,227,227,215,190,168,139,101,81,84,91,93,80,76,79,85,100,125,176,223,
  237,244,241,234,235,203,117,29,8,10,7,5,2,3,12,24,36,44,48,42,58,142,201,206,
  200,202,196,200,214,228,234,234,235,235,238,242,245,245,243,237,210,147,104,
  82,88,109,133,160,180,200,195,159,95,45,22,15,15,17,18,17,17,47,127,215,241,
  237,210,147,83,35,23,19,14,14,26,115,181,213,234,230,231,229,230,210,146,50,
  24,24,27,33,64,91,97,101,134,180,203,211,198,177,158,142,128,136,173,205,208,
  207,195,186,171,150,89,34,20,21,22,22,18,30,61,122,213,228,230,233,232,232,
  231,236,204,151,101,61,38,30,29,33,55,73,78,67,42,18,14,9,12,16,24,24,39,96,
  143,170,176,178,175,150,144,155,164,170,160,148,142,154,188,229,240,242,239,
  233,226,214,150,95,94,91,103,112,119,118,173,226,217,204,188,178,165,132,48,
  28,40,39,38,34,32,25,26,25,26,23,20,27,26,23,24,22,24,62,165,233,234,231,232,
  235,240,243,243,241,233,223,187,146,143,188,229,234,232,232,233,235,233,224,
  157,86,86,83,80,76,75,65,53,32,37,84,109,113,86,36,16,15,13,16,20,21,31,49,
  60,63,57,46,52,97,151,191,201,210,222,235,238,245,250,248,246,249,249,240,
  224,221,216,206,172,86,42,44,91,152,189,222,226,179,90,38,35,28,27,27,26,27,
  29,38,39,71,133,162,165,133,77,26,18,20,21,22,21,22,22,25,21,16,25,79,115,
  122,136,150,158,197,234,236,241,243,244,247,249,254,0,0,254,244,231,161,61,
  33,34,59,117,166,203,214,189,142,74,33,32,30,24,16,12,10,13,20,48,82,91,71,
  47,27,20,19,17,17,13,15,22,31,34,35,33,48,103,135,150,160,162,180,228,245,
  245,252,248,243,240,240,241,239,235,233,233,202,100,40,41,48,96,154,195,223,
  216,176,108,34,16,25,28,32,36,36,31,20,29,65,91,87,72,47,17,14,17,23,25,24,
  22,22,24,21,23,24,66,144,195,221,215,179,139,113,78,70,92,173,242,242,250,
  250,246,238,239,239,242,237,232,193,116,52,20,18,19,25,31,38,56,68,61,58,63,
  68,78,93,95,119,139,137,132,120,95,49,28,27,22,22,17,16,15,20,22,28,23,43,
  107,159,199,205,177,141,140,143,161,215,240,237,237,234,238,238,235,233,233,
  234,234,233,206,180,154,117,75,63,62,72,75,72,89,108,116,114,146,180,186,187,
  195,199,200,177,137,117,102,60,25,19,16,22,24,19,18,16,16,15,31,86,154,167,
  124,59,29,27,17,16,18,36,159,237,229,237,235,240,241,245,239,238,229,164,114,
  99,111,142,188,224,227,211,175,141,99,78,87,117,151,204,227,231,239,240,239,
  228,164,63,15,15,12,10,7,8,10,17,13,12,13,18,24,104,210,243,234,227,225,220,
  232,229,231,234,241,243,242,241,239,238,237,232,226,170,98,71,68,84,102,129,
  156,182,192,171,126,74,37,29,26,21,19,24,26,38,103,193,234,236,216,150,63,23,
  22,20,18,24,23,88,219,237,235,235,241,247,245,234,219,116,25,25,22,29,30,32,
  26,23,43,114,165,195,201,177,144,119,89,87,130,206,232,232,233,232,231,235,
  184,69,25,23,23,23,27,28,67,133,201,230,228,234,232,232,233,231,223,226,196,
  122,70,36,33,31,59,94,102,105,81,32,14,11,6,7,8,11,15,23,33,87,120,117,119,
  87,49,66,84,102,155,213,217,212,216,218,220,223,229,236,242,244,237,207,182,
  189,207,228,232,231,232,232,214,179,134,97,73,46,40,43,73,95,90,69,50,27,16,
  20,19,18,15,18,20,25,20,21,19,44,134,203,228,229,233,234,234,234,231,228,229,
  220,189,166,177,211,231,237,240,244,243,241,234,232,230,231,226,224,220,225,
  218,212,185,133,70,23,23,16,17,18,19,19,19,16,17,19,42,120,183,209,211,184,
  150,118,87,82,125,200,227,228,234,237,234,230,179,71,34,34,30,27,29,28,64,
  128,193,227,226,232,233,235,231,216,165,140,109,63,37,27,28,28,32,78,143,181,
  198,184,157,135,108,90,110,176,232,236,236,232,233,236,226,131,41,21,15,13,
  19,22,36,97,164,219,228,230,235,239,235,227,212,183,149,100,57,37,31,28,29,
  45,55,66,90,118,133,132,151,167,166,177,210,225,221,222,224,226,229,239,232,
  173,89,35,28,24,24,25,28,23,18,12,59,131,172,201,206,194,153,128,106,81,68,
  79,108,131,178,217,226,225,225,221,205,193,189,182,195,227,234,234,232,232,
  233,232,230,229,232,222,146,55,32,31,23,21,22,24,25,27,24,21,17,13,10,14,16,
  19,22,29,30,89,203,230,231,238,237,235,233,236,227,191,168,155,164,209,231,
  232,239,241,239,232,227,225,229,224,143,64,40,35,31,65,123,159,175,165,129,
  49,18,16,10,13,16,20,24,26,41,118,195,234,242,214,160,105,50,28,26,27,35,79,
  135,191,230,230,225,218,215,214,217,219,221,212,202,194,186,189,194,178,153,
  141,131,115,107,103,92,68,51,34,30,35,52,99,150,208,228,226,232,233,234,194,
  121,48,23,24,22,19,17,19,45,93,141,180,185,173,161,145,131,125,128,139,159,
  195,218,223,231,231,234,202,144,92,53,30,20,25,29,32,41,58,74,85,100,116,120,
  115,115,110,96,84,63,44,44,72,106,141,177,220,240,239,242,237,232,201,166,
  129,111,105,118,129,129,136,149,148,116,82,52,32,28,27,29,28,36,70,105,133,
  163,175,163,141,118,74,35,32,35,34,41,83,153,218,229,229,228,230,233,234,227,
  188,151,141,161,196,224,223,225,225,228,184,107,42,23,27,24,26,28,30,28,28,
  27,27,25,23,26,65,112,159,199,225,236,227,209,191,192,194,208,222,226,229,
  232,235,237,207,147,98,65,59,86,125,184,230,232,235,236,235,183,95,33,21,23,
  23,27,27,28,27,25,23,23,26,35,59,97,124,146,156,157,151,141,120,104,111,127,
  148,180,216,227,225,225,223,222,224,227,230,233,230,229,229,230,227,204,162,
  133,111,82,69,76,89,115,143,157,166,175,162,123,77,40,27,20,16,14,15,31,52,
  55,59,52,25,14,17,23,28,30,32,36,60,108,164,198,218,223,216,223,225,221,220,
  217,213,214,217,227,232,233,237,238,236,236,222,178,146,118,65,31,25,24,25,
  36,54,82,118,135,131,107,100,90,67,36,23,24,26,25,35,92,151,194,210,202,184,
  159,121,101,97,98,108,144,188,219,235,238,235,233,217,183,154,132,118,109,
  119,136,150,173,193,200,197,192,187,176,164,149,124,98,75,53,38,26,30,54,89,
  131,159,172,179,177,167,137,85,37,26,25,24,24,39,66,102,149,187,200,192,168,
  133,101,78,71,74,92,121,156,181,190,193,192,195,174,138,114,90,75,67,70,71,
  72,79,91,99,106,114,122,122,119,119,113,101,91,80,71,74,94,114,138,163,190,
  201,198,197,195,196,179,161,139,128,124,132,139,137,137,142,141,117,90,65,49,
  51,52,54,53,60,87,116,137,159,166,155,138,119,85,52,46,46,43,48,85,145,199,
  210,213,217,221,224,223,216,179,143,132,147,177,202,205,209,209,211,177,115,
  62,43,45,43,42,41,42,41,40,39,40,38,37,40,77,116,159,192,210,217,212,200,195,
  203,209,214,217,216,220,227,227,229,230,230,230,230,200,157,134,94,47,32,27,
  27,32,46,63,97,124,128,109,92,92,76,51,28,28,29,32,33,68,131,176,202,202,188,
  169,142,115,104,107,118,141,180,220,235,234,233,231,230,229,228,228,228,229,
  229,228,228,226,206,164,132,105,68,51,60,76,105,136,150,159,165,146,101,58,
  32,27,26,24,23,22,36,58,62,72,67,34,23,25,25,25,25,26,27,57,106,151,190,218,
  225,216,221,223,219,219,229,229,230,231,233,235,237,236,236,234,234,216,157,
  113,80,39,26,27,26,26,25,26,48,91,117,123,120,124,114,85,46,29,30,31,30,42,
  90,140,179,200,202,203,190,159,137,129,131,145,177,215,232,231,231,229,229,
  206,154,109,75,54,46,69,104,132,168,199,220,229,228,224,211,199,187,158,122,
  96,62,33,27,25,25,26,27,36,74,111,127,127,128,127,106,67,32,26,25,25,25,57,
  111,156,188,198,199,195,172,144,132,130,138,160,199,229,234,233,233,231,226,
  185,134,91,58,40,43,56,65,66,91,149,203,231,231,227,202,165,123,90,84,98,124,
  162,210,231,231,231,230,212,179,142,105,89,87,87,91,93,93,89,82,82,84,98,123,
  143,174,204,216,211,199,178,146,121,108,106,117,134,157,170,175,164,132,80,
  32,25,24,22,22,23,23,28,68,109,144,167,163,150,138,124,109,108,114,128,153,
  181,198,211,220,217,207,199,185,163,142,120,92,80,75,76,78,84,105,127,157,
  179,189,189,186,175,152,120,94,70,52,51,59,69,81,98,114,127,133,133,135,139,
  149,163,175,189,197,209,211,204,190,175,163,158,152,140,129,128,128,118,109,
  101,98,95,92,96,108,126,141,151,168,180,181,176,163,152,149,143,142,149,155,
  155,152,139,123,103,85,67,60,63,70,75,86,97,100,97,89,85,91,99,108,125,144,
  150,152,146,129,104,78,48,29,29,27,30,47,84,126,162,187,203,208,202,189,181,
  178,176,181,193,207,219,220,212,203,193,179,157,136,124,117,109,110,117,120,
  118,111,96,82,68,59,53,50,47,47,53,63,75,86,99,117,145,164,178,196,217,227,
  227,221,207,191,169,141,116,100,85,80,84,97,113,128,145,156,160,164,163,161,
  170,184,196,212,229,233,231,218,201,176,148,123,92,66,52,47,43,39,41,47,49,
  53,66,77,85,99,108,115,117,120,125,130,139,147,153,161,169,172,166,161,151,
  131,105,83,69,65,76,89,104,128,149,159,166,167,156,142,135,131,132,136,146,
  159,168,169,155,134,109,78,54,40,33,32,35,39,48,63,78,93,107,119,133,151,170,
  188,197,200,196,184,169,147,125,104,98,104,112,124,142,159,166,165,161,151,
  134,119,102,87,85,92,96,103,113,118,119,115,110,103,102,109,114,123,139,156,
  165,166,169,168,158,145,132,125,127,132,136,141,140,133,123,117,113,111,114,
  129,158,191,218,229,228,230,224,193,154,116,81,58,44,41,49,66,88,103,117,121,
  118,111,104,93,78,76,83,91,98,115,132,145,156,168,165,155,148,133,117,106,96,
  89,91,100,114,133,155,171,179,184,185,176,156,133,114,102,92,82,78,81,97,119,
  137,150,156,156,150,139,126,114,105,102,108,124,141,155,167,178,185,185,179,
  166,145,125,109,98,92,87,87,91,95,100,104,103,104,110,115,117,118,115,115,
  121,129,133,128,125,126,127,127,127,123,118,115,116,117,121,127,131,135,143,
  151,156,154,154,157,159,157,153,150,145,138,135,137,135,133,131,130,130,131,
  127,116,113,117,118,119,125,130,133,138,147,152,149,148,144,137,131,125,110,
  90,81,77,73,78,90,106,121,133,141,141,138,134,123,113,116,126,136,146,160,
  172,175,175,166,152,139,128,122,119,124,131,136,140,142,140,137,136,133,130,
  127,128,131,130,128,126,124,120,117,118,123,135,145,152,156,159,157,147,133,
  123,112,102,98,103,109,114,124,137,147,149,144,138,133,124,110,98,95,92,89,
  90,95,103,112,119,125,131,137,138,136,137,139,138,139,140,139,138,139,139,
  138,137,135,134,135,139,141,140,140,142,143,141,140,141,141,145,152,158,163,
  162,157,149,138,124,108,90,75,71,71,74,85,101,112,119,124,123,117,109,103,97,
  97,104,109,116,127,137,144,151,157,156,155,156,159,161,160,155,146,139,132,
  125,121,121,125,130,136,139,141,141,134,123,114,109,103,103,113,129,142,150,
  153,153,151,143,130,116,111,113,117,121,126,130,132,128,119,107,97,88,85,90,
  101,113,127,140,152,159,161,159,157,156,153,151,148,145,140,137,134,129,121,
  111,108,112,116,119,126,130,130,128,121,112,107,105,109,119,132,142,147,150,
  150,145,133,119,108,103,105,113,123,133,137,139,140,140,135,125,113,107,106,
  108,112,118,125,132,137,140,145,149,149,146,147,145,136,128,121,113,105,104,
  106,109,115,123,128,134,137,129,116,106,99,92,87,92,104,120,140,162,184,200,
  206,206,202,194,182,165,152,144,140,139,139,138,137,138,141,139,135,132,128,
  123,119,117,113,111,112,113,119,128,136,140,141,142,140,134,128,124,119,115,
  113,113,115,121,128,137,148,161,169,170,168,164,157,147,136,126,119,119,122,
  129,138,148,154,156,155,149,142,134,123,115,109,105,104,106,111,115,120,123,
  123,121,120,117,114,113,114,116,119,121,124,129,134,137,141,144,148,151,152,
  150,148,146,139,128,119,114,115,118,123,131,139,146,149,148,144,140,131,121,
  114,113,115,117,121,127,131,134,135,132,131,131,128,123,118,114,110,104,99,
  97,100,104,108,114,120,125,128,129,128,126,122,119,119,122,127,134,142,148,
  153,156,153,147,141,134,126,121,117,116,120,127,132,134,133,131,127,120,112,
  108,106,104,103,105,109,114,117,119,120,122,123,123,124,125,125,126,128,129,
  130,130,126,122,117,112,106,101,101,102,105,114,122,131,140,147,150,150,150,
  151,149,146,147,149,148,143,137,130,121,113,108,105,106,109,115,122,129,132,
  130,129,129,128,128,131,137,145,150,152,152,151,144,135,128,122,119,117,116,
  118,124,128,129,130,130,129,125,122,123,125,127,131,139,146,150,154,156,155,
  154,152,149,145,139,132,126,122,118,115,115,118,123,128,135,141,145,146,144,
  140,135,131,127,124,124,125,126,128,132,136,135,131,129,126,121,117,116,116,
  116,119,123,128,131,134,137,139,140,140,136,133,130,128,129,131,133,136,139,
  142,145,144,142,140,140,139,134,129,124,120,118,115,114,116,121,126,131,137,
  142,144,144,141,137,133,130,128,126,127,130,134,136,136,135,131,126,122,120,
  116,117,121,126,132,139,144,143,140,137,132,125,118,113,115,120,125,130,136,
  143,147,149,150,147,141,135,128,121,115,109,104,102,104,105,107,112,117,121,
  125,130,134,135,135,135,135,134,135,135,135,135,136,139,140,138,134,129,123,
  118,114,110,110,112,116,120,124,126,126,128,130,128,127,124,123,123,123,124,
  124,125,127,129,132,136,139,141,143,144,144,140,134,131,128,124,119,116,115,
  116,118,122,127,130,132,132,133,134,133,132,133,135,136,136,137,137,137,136,
  133,130,128,126,122,119,119,120,120,121,123,126,128,129,131,132,133,133,133,
  130,128,125,120,114,113,114,117,120,126,131,135,139,141,140,136,133,130,127,
  123,121,120,120,123,128,130,133,134,134,135,136,134,130,128,128,127,127,127,
  125,125,125,125,124,124,122,120,120,120,120,119,116,114,115,115,117,121,125,
  128,132,136,138,138,137,136,134,134,136,139,141,143,145,148,148,145,141,137,
  135,135,134,132,132,133,133,135,136,135,132,129,127,125,123,121,120,120,123,
  127,128,129,131,133,134,134,136,136,135,136,138,139,140,140,140,138,136,133,
  131,129,129,128,128,127,126,126,126,126,127,128,132,134,135,136,138,139,140,
  139,138,136,133,131,129,131,132,132,130,129,126,124,121,120,119,118,118,120,
  124,128,131,134,138,139,137,134,132,131,133,134,135,137,138,140,143,143,141,
  139,137,134,132,129,127,125,123,122,122,122,123,125,128,133,138,142,145,147,
  148,148,146,143,140,137,135,134,135,135,135,135,135,135,134,132,129,127,126,
  124,123,123,122,122,123,125,127,129,132,136,139,141,143,144,144,141,138,134,
  128,124,120,117,115,115,115,115,117,120,123,126,129,131,135,139,142,144,145,
  144,144,142,138,134,130,128,126,125,124,125,125,123,121,119,117,115,113,112,
  112,113,115,117,120,124,127,129,130,129,127,124,122,119,118,119,120,124,129,
  133,137,139,138,135,131,128,126,124,123,124,125,127,128,128,129,129,129,129,
  129,129,129,128,128,127,123,120,118,116,115,116,118,122,124,127,130,131,129,
  126,123,121,120,119,121,125,129,133,135,135,134,132,129,127,125,123,124,126,
  127,128,128,128,128,127,126,126,128,129,131,133,135,135,133,130,128,126,123,
  120,118,119,120,122,125,127,128,129,129,129,129,127,124,120,117,113,111,111,
  112,114,116,119,121,123,123,125,126,126,127,128,130,133,135,138,139,139,139,
  139,139,138,137,137,135,133,131,130,128,127,125,125,127,128,130,132,133,132,
  131,128,125,122,121,119,119,122,125,128,132,135,136,136,136,134,133,132,133,
  134,136,138,140,140,139,137,134,132,131,131,132,135,136,136,135,133,128,125,
  121,118,117,118,120,124,128,132,135,136,136,135,134,133,134,135,136,137,138,
  139,140,140,138,136,135,134,133,131,129,127,126,126,125,126,126,127,128,129,
  131,134,136,137,138,137,135,132,129,126,124,124,124,124,126,127,129,130,130,
  128,127,126,126,126,128,130,133,137,141,144,144,143,142,140,137,133,129,127,
  126,126,127,128,129,130,131,132,133,134,134,133,132,130,128,126,124,122,121,
  123,125,127,128,129,130,130,129,126,122,120,119,119,121,123,124,126,128,130,
  131,131,129,128,128,129,128,127,126,127,129,131,133,134,136,137,137,137,136,
  132,129,126,124,122,120,119,119,121,123,125,125,124,124,124,124,124,123,121,
  119,118,118,117,117,117,120,124,128,129,129,130,130,130,130,130,129,129,129,
  129,131,133,134,136,136,136,135,134,133,130,128,127,126,124,124,124,125,127,
  128,129,129,128,127,125,123,122,123,124,125,126,127,128,128,128,128,128,127,
  125,125,126,126,126,126,126,127,125,124,124,124,125,127,128,130,131,130,129,
  128,127,126,126,126,126,127,128,128,128,128,129,129,129,129,129,128,129,129,
  128,126,126,126,126,126,125,126,127,127,126,124,122,120,118,117,118,118,119,
  120,123,127,130,131,132,133,135,136,135,135,134,133,132,131,130,130,128,128,
  129,130,129,129,129,129,130,131,131,132,132,132,133,133,133,134,134,135,135,
  134,133,132,131,130,128,128,128,127,126,126,125,125,125,124,123,122,122,123,
  126,128,131,135,138,141,142,141,140,138,136,134,130,128,126,125,124,124,125,
  125,126,128,128,128,128,128,128,129,129,130,131,132,133,135,136,137,136,136,
  138,139,140,141,141,140,137,134,130,126,122,119,118,118,120,123,124,126,128,
  128,128,126,125,123,123,124,126,130,134,137,139,140,140,140,138,137,135,134,
  132,131,130,129,128,128,128,128,128,128,128,128,128,128,129,129,129,129,128,
  128,129,129,129,130,131,131,132,132,133,134,135,134,133,131,129,128,126,124,
  123,123,122,123,123,124,126,127,127,128,128,128,128,128,129,129,130,131,131,
  132,132,133,133,134,133,133,132,130,128,126,124,122,121,119,118,119,120,120,
  122,123,125,127,129,130,131,133,134,135,136,136,136,134,133,132,131,129,129,
  129,128,127,125,123,122,120,118,117,116,116,116,118,121,125,127,129,131,132,
  132,132,133,132,131,130,130,130,130,129,129,129,128,127,125,125,124,124,124,
  124,125,125,125,124,124,124,125,126,128,129,131,133,134,133,131,130,128,126,
  125,125,125,125,127,128,129,129,129,128,126,124,123,122,121,122,123,124,126,
  127,128,128,128,128,128,128,129,130,132,134,134,134,133,132,131,129,128,127,
  126,126,126,126,125,124,123,123,123,123,125,126,128,131,133,134,134,133,132,
  130,128,127,125,125,126,126,127,127,128,128,128,128,127,127,127,128,129,130,
  131,131,131,130,130,129,129,128,128,128,129,130,131,131,131,130,130,129,128,
  128,127,126,125,125,124,124,125,126,128,128,129,129,130,131,131,130,131,131,
  130,130,130,130,129,128,128,128,128,128,128,128,128,127,127,126,126,125,126,
  127,128,130,131,132,131,130,128,126,125,124,123,124,125,127,129,131,133,134,
  134,134,133,132,130,129,128,128,128,128,127,127,127,128,128,128,128,128,129,
  129,131,132,132,131,131,130,129,129,129,129,129,128,128,127,126,126,126,126,
  126,127,128,129,131,132,132,132,131,128,127,125,124,124,125,126,127,129,130,
  131,132,131,130,129,128,128,128,128,129,129,130,130,129,128,127,126,125,125,
  125,125,126,126,126,128,129,129,130,130,130,130,131,131,130,130,129,128,126,
  126,126,125,125,126,126,127,128,128,128,128,129,129,129,128,128,128,127,127,
  127,128,128,128,128,128,128,127,127,126,126,126,126,126,126,126,127,128,128,
  128,129,130,129,129,129,129,129,129,129,128,128,128,127,127,127,127,127,127,
  127,126,125,124,123,123,122,122,122,124,126,128,131,133,135,136,135,134,133,
  131,129,128,127,127,127,127,128,128,128,127,125,124,123,122,121,121,122,123,
  124,126,127,128,130,131,132,133,133,132,131,130,129,128,127,127,126,126,126,
  126,127,127,127,127,127,126,126,126,126,126,126,127,128,129,130,131,131,132,
  131,131,130,129,128,128,128,127,128,128,129,130,130,130,130,129,128,127,126,
  126,126,127,128,129,130,132,133,133,133,133,132,131,130,129,128,127,127,126,
  127,128,128,128,129,129,128,128,128,128,128,127,127,127,128,128,129,130,131,
  131,131,131,130,129,129,128,128,128,128,128,128,128,128,128,127,128,128,129,
  130,131,131,131,131,131,130,129,129,129,129,128,128,128,128,128,128,128,128,
  127,127,128,128,128,128,128,129,130,131,133,134,135,135,135,134,133,132,131,
  129,127,126,125,124,124,124,125,126,127,128,129,130,131,131,132,132,132,131,
  130,130,128,128,128,128,128,129,129,129,129,129,128,128,128,128,128,128,128,
  128,129,129,129,129,129,129,129,129,129,129,129,129,129,128,128,128,128,127,
  127,127,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,130,130,
  131,131,131,131,131,130,129,128,127,126,126,125,125,126,127,128,128,129,130,
  130,130,129,129,129,129,129,129,129,129,129,129,129,129,129,128,128,128,127,
  126,126,125,125,126,126,127,128,128,129,129,129,129,129,129,129,129,128,128,
  128,128,127,127,127,127,127,126,126,125,126,126,127,127,128,128,128,128,128,
  128,128,129,129,129,130,130,129,129,128,128,128,127,127,126,126,126,126,126,
  126,126,126,126,126,126,127,127,128,128,128,128,128,129,129,128,128,128,128,
  128,128,128,127,126,126,126,126,127,127,127,128,128,128,128,128,128,128,129,
  129,129,129,130,129,129,128,128,128,127,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,129,129,129,129,128,128,128,128,128,127,126,126,125,
  125,125,126,126,126,126,127,128,128,129,130,130,131,131,131,131,131,131,131,
  131,130,130,129,128,128,128,127,127,127,127,127,128,128,128,128,128,128,129,
  129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,
  127,127,128,128,128,129,129,129,129,130,130,129,129,129,129,129,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,
  130,131,132,132,132,131,130,130,129,128,127,127,127,126,127,127,127,127,128,
  128,128,128,128,128,128,128,129,130,130,130,130,130,129,129,128,128,128,128,
  128,127,128,128,128,128,128,128,127,128,128,128,129,129,129,130,130,130,130,
  130,130,131,131,132,132,131,131,129,128,127,126,125,125,125,126,127,127,127,
  127,127,127,127,126,126,127,127,128,128,129,130,131,131,131,130,129,128,127,
  127,126,126,126,127,127,128,128,128,128,128,129,129,129,129,129,129,129,130,
  130,130,130,129,129,128,128,128,128,128,128,128,128,127,127,127,127,128,128,
  128,127,127,127,127,128,128,127,127,127,127,128,128,129,129,129,129,129,129,
  128,128,128,128,129,129,128,128,128,128,128,127,127,127,127,127,128,128,128,
  128,128,127,127,127,127,127,127,127,128,128,128,128,129,129,128,128,128,128,
  127,127,127,126,126,126,126,126,127,127,127,127,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,127,127,126,126,126,126,126,127,127,128,
  128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,128,128,127,
  127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,128,128,128,128,
  129,129,129,129,129,129,129,128,128,128,128,128,128,127,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,129,129,129,129,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,129,128,128,128,128,128,128,128,128,128,
  128,129,129,129,129,130,130,130,130,130,130,129,129,128,128,128,127,127,128,
  128,128,129,129,130,130,130,130,130,130,130,130,130,130,130,129,129,129,128,
  128,128,127,127,127,127,128,128,128,129,130,130,131,131,130,130,130,130,129,
  129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,
  129,129,129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,127,127,127,127,128,128,129,129,129,130,130,130,130,
  129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  129,129,129,130,130,129,129,129,129,128,128,128,128,128,128,128,127,127,127,
  128,128,128,128,127,127,127,127,127,128,128,128,128,128,128,129,129,128,128,
  128,128,128,128,128,128,129,129,129,129,129,128,128,128,127,127,126,126,126,
  127,127,127,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,127,127,127,127,127,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,129,128,128,128,128,128,128,128,128,128,
  128,128,128,128,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,127,127,127,127,127,127,128,128,128,129,129,129,129,
  129,129,128,128,128,128,129,129,129,129,129,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,127,127,127,127,127,127,127,127,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,127,127,127,127,128,128,128,128,128,129,129,128,128,128,128,128,128,
  128,128,129,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,127,127,127,127,128,128,128,128,128,128,128,128,128,129,129,129,
  129,129,128,128,128,128,128,128,128,128,129,129,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,128,128,128,128,
  128,127,127,127,127,127,127,127,127,128,128,128,128,128,128,128,128,128,128,
  128,128,128,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,127,128,128,128,128,128,128,128,128,128,
  128,128,128,128,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,129,129,129,129,129,128,128,128,128,128,128,128,
  128,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,
  129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,
};

static const char dsdgact[] = {
  3,0,34,86,63,43,0,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,127,
  127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,
  126,126,126,126,126,126,126,126,126,126,127,127,128,128,128,129,131,132,133,
  134,134,134,133,131,130,129,128,128,129,130,130,129,128,128,126,125,125,125,
  126,127,128,128,129,128,128,128,128,127,127,128,128,128,128,129,129,130,130,
  131,132,132,133,133,133,132,132,131,130,130,130,130,130,130,129,129,128,127,
  127,127,127,127,127,127,126,125,124,123,122,122,121,121,120,120,120,120,120,
  120,120,119,119,118,118,118,119,120,123,126,130,135,141,147,151,153,153,150,
  145,139,133,130,128,128,129,130,130,128,124,120,116,114,114,117,121,125,127,
  128,128,126,124,122,120,119,119,119,118,117,117,117,119,122,125,129,133,136,
  138,139,139,139,139,139,139,139,140,140,139,139,138,137,135,132,131,129,128,
  129,130,131,132,132,132,130,128,126,124,123,123,123,124,126,127,128,128,128,
  129,130,131,132,133,133,133,133,133,132,132,131,131,131,131,131,130,130,130,
  130,130,130,131,131,131,130,129,127,125,122,120,117,114,110,106,102,98,94,91,
  90,92,96,105,118,133,150,165,176,180,178,169,158,146,138,135,138,143,147,147,
  143,134,121,110,103,102,108,118,129,137,140,139,134,127,122,118,117,118,117,
  114,110,104,99,97,100,107,116,127,135,141,143,142,141,140,141,145,150,153,
  155,154,150,146,141,136,134,134,136,138,142,143,142,140,138,135,132,131,129,
  129,128,126,124,122,119,117,117,117,119,120,121,121,120,117,115,114,112,110,
  108,105,102,98,94,90,86,84,86,91,99,111,128,149,170,187,199,204,201,191,176,
  161,150,143,141,144,148,146,138,128,112,97,87,82,88,103,117,128,136,141,138,
  132,126,120,118,118,116,113,109,101,95,92,93,99,109,120,129,136,137,136,135,
  136,139,146,155,163,169,172,168,160,153,144,139,137,137,138,141,141,141,140,
  138,137,136,136,136,136,135,133,130,127,123,121,119,117,115,113,108,101,93,
  82,70,57,47,40,37,40,55,80,113,149,186,215,228,226,213,191,165,148,142,150,
  169,185,195,195,176,145,113,87,77,80,98,128,157,171,176,169,148,123,102,87,
  82,83,82,80,76,67,61,62,68,82,99,116,133,142,142,140,137,135,140,152,167,180,
  188,189,182,170,158,148,143,145,151,160,167,170,166,157,146,132,122,114,111,
  110,109,109,108,102,94,87,80,71,61,49,38,27,23,23,31,58,98,145,188,223,232,
  232,226,199,169,150,150,162,184,211,224,220,201,171,138,112,96,95,110,129,
  144,157,159,148,131,115,98,89,83,77,74,70,64,60,60,67,79,93,109,122,128,131,
  126,118,116,118,126,140,154,164,172,175,174,175,176,177,183,186,186,180,170,
  155,142,132,126,122,119,111,100,86,68,52,41,35,39,51,66,83,98,106,112,117,
  120,125,131,139,149,156,159,161,165,170,171,174,180,184,186,188,183,177,170,
  159,153,159,157,154,155,152,146,138,122,107,97,83,76,76,77,82,88,92,98,102,
  102,103,105,106,110,115,120,125,128,129,128,120,109,95,79,64,57,59,75,101,
  138,176,213,231,230,229,211,182,152,130,128,141,156,173,182,175,156,124,87,
  61,42,35,48,74,105,137,161,177,189,189,183,178,166,150,135,120,108,101,98,
  100,105,109,110,111,111,105,99,99,104,115,133,150,164,173,166,153,138,120,99,
  86,77,67,64,61,57,60,69,81,108,140,171,206,230,233,234,233,219,208,203,195,
  196,192,177,161,136,104,82,68,60,65,77,92,105,110,111,105,97,96,96,98,107,
  109,108,105,94,87,86,85,93,106,112,114,104,87,73,61,57,71,103,145,193,230,
  233,232,231,230,229,218,199,188,188,183,180,175,158,134,114,91,75,65,60,65,
  76,82,86,92,95,100,103,109,116,120,121,115,105,93,78,67,60,57,52,46,37,32,32,
  30,33,61,113,178,224,228,227,229,228,230,207,186,181,195,225,232,232,229,183,
  128,80,32,28,27,44,83,121,145,154,148,129,120,119,122,128,134,130,116,93,67,
  49,45,54,74,102,128,141,146,138,126,118,123,141,167,193,210,215,204,175,139,
  109,81,59,48,34,25,25,25,25,42,90,145,216,239,236,235,233,221,171,126,97,106,
  132,162,184,185,154,115,61,26,26,29,60,117,179,226,234,235,220,180,145,121,
  113,109,106,97,82,66,56,59,74,99,133,168,191,200,191,173,152,131,118,119,126,
  118,103,61,30,29,28,26,27,23,56,147,220,231,227,230,227,230,202,183,176,187,
  219,233,228,213,161,105,58,25,28,28,38,82,144,186,211,232,218,192,176,145,
  121,101,74,54,39,25,24,25,41,74,117,152,183,210,217,210,201,183,157,134,111,
  76,38,22,24,22,25,23,38,101,181,233,234,235,234,234,232,230,229,229,228,229,
  221,183,135,75,27,26,25,27,25,40,75,102,118,108,92,77,72,74,93,129,160,190,
  191,174,155,124,92,71,70,79,103,131,161,195,217,214,212,204,191,179,162,147,
  135,130,126,122,127,129,130,144,146,155,162,157,150,144,135,123,127,128,127,
  130,131,113,104,84,62,39,25,26,26,25,24,22,24,23,33,102,196,236,233,235,234,
  231,231,208,147,104,101,121,155,189,195,164,125,70,26,28,27,39,95,167,218,
  235,232,233,216,187,166,163,159,154,134,95,51,26,26,28,27,63,118,155,188,182,
  161,126,88,49,27,26,25,22,20,18,31,107,202,234,233,239,235,240,214,114,43,36,
  79,188,239,236,234,231,133,23,22,17,21,22,37,165,242,233,240,225,157,78,59,
  76,119,173,203,192,171,120,66,45,45,71,123,179,223,232,233,216,184,159,142,
  151,156,150,124,72,29,27,22,21,17,20,15,70,183,228,228,233,233,235,231,192,
  149,134,146,169,202,215,198,160,108,37,19,20,20,20,60,170,234,237,240,239,
  239,231,170,94,41,25,24,23,22,21,40,82,112,143,174,181,178,166,131,89,39,25,
  23,20,19,21,22,45,124,206,235,233,238,236,239,229,172,118,84,96,139,204,234,
  230,234,210,132,42,21,24,24,45,143,218,234,232,234,222,149,74,31,26,26,24,25,
  25,27,29,60,99,130,148,169,187,174,143,88,34,23,23,22,24,26,27,47,152,233,
  237,240,240,240,237,234,226,167,134,148,196,229,231,233,217,119,30,22,23,21,
  25,21,72,162,195,202,161,88,29,24,23,22,24,25,28,28,31,30,63,135,213,234,232,
  233,232,229,230,216,176,160,167,197,228,232,233,230,205,182,172,172,176,184,
  182,176,153,112,75,41,24,24,24,26,25,26,26,26,26,26,25,24,24,23,23,23,68,171,
  230,230,235,235,235,235,236,235,234,217,228,238,237,235,232,188,118,50,23,25,
  37,90,140,182,192,173,134,79,36,27,28,41,55,46,29,27,24,23,22,22,32,123,215,
  232,232,234,234,233,232,218,195,164,146,131,104,82,62,53,62,73,67,64,69,68,
  64,81,123,178,227,234,236,233,235,187,79,26,25,23,24,25,31,80,120,142,159,
  174,196,219,228,231,235,236,234,236,211,144,93,63,43,32,32,28,26,24,20,31,96,
  159,213,232,233,234,189,87,26,25,20,19,18,17,50,125,183,213,211,198,173,171,
  185,213,228,231,234,234,233,235,229,163,76,31,29,27,28,26,26,47,81,104,119,
  127,137,154,164,164,153,130,106,74,40,35,35,47,83,142,205,221,222,227,228,
  235,200,151,128,121,126,135,142,148,127,93,44,19,23,23,20,17,16,58,134,196,
  234,239,241,239,231,175,118,80,54,43,38,46,71,87,89,88,100,115,141,172,210,
  228,227,231,233,237,238,225,189,150,113,87,64,51,42,34,44,58,82,110,141,158,
  154,117,66,25,12,16,25,31,36,36,53,115,185,232,230,232,232,233,230,231,209,
  186,173,163,196,222,226,222,165,100,46,31,24,23,24,52,110,141,168,170,145,
  105,106,118,134,154,154,137,109,51,32,35,36,76,158,224,224,221,223,219,157,
  90,44,36,39,33,40,44,43,52,80,127,196,231,229,235,235,233,219,128,44,22,25,
  24,45,93,128,123,90,40,20,19,22,29,80,152,202,229,225,223,219,221,209,211,
  211,179,146,112,76,46,41,59,106,158,213,235,228,224,192,136,93,47,31,23,18,
  15,19,15,16,12,70,195,235,238,241,238,239,237,160,69,42,78,158,218,221,216,
  224,179,62,29,33,29,27,54,152,221,224,222,213,153,94,65,62,71,90,81,69,55,38,
  30,46,87,145,211,235,231,233,214,154,103,60,37,27,23,32,38,31,33,45,78,135,
  202,238,233,232,236,236,236,205,153,117,101,93,84,76,50,23,16,19,21,21,29,76,
  125,162,189,204,205,203,202,196,194,190,181,167,144,119,105,99,102,116,131,
  145,162,164,156,137,108,77,46,23,22,18,20,22,27,22,42,96,161,224,230,229,234,
  236,236,233,230,190,143,99,69,64,62,79,98,112,126,125,108,93,81,75,76,84,102,
  118,134,142,147,143,137,130,119,109,99,90,86,90,98,113,138,172,204,218,217,
  200,171,131,97,76,65,58,67,97,122,130,131,138,147,148,126,102,80,57,34,17,29,
  71,116,178,230,237,243,241,241,228,180,129,89,76,84,93,100,100,88,77,58,42,
  45,62,84,111,142,157,156,151,144,140,143,145,154,162,159,146,123,96,75,70,75,
  91,118,143,172,191,193,187,170,149,131,106,90,78,63,50,36,24,23,29,55,100,
  158,215,231,227,231,232,237,214,163,123,95,85,101,124,139,142,130,104,74,46,
  29,28,40,70,102,133,153,149,140,137,133,131,133,143,157,159,149,134,105,79,
  64,65,80,117,160,202,228,232,223,198,166,130,99,82,88,101,105,106,95,70,41,
  26,25,27,26,60,129,194,232,232,234,232,233,218,183,158,139,132,128,124,121,
  115,103,93,84,80,87,93,90,93,96,92,88,88,97,111,123,129,132,133,132,124,113,
  109,110,116,125,135,148,164,182,191,191,190,186,175,154,125,100,88,86,94,109,
  125,137,134,114,96,68,46,34,39,71,113,158,208,232,231,232,232,223,192,161,
  131,108,95,82,68,59,50,47,46,46,55,68,82,97,118,141,157,165,170,171,168,161,
  144,128,120,116,115,118,128,140,152,159,161,162,162,167,176,184,189,192,192,
  178,152,119,87,60,38,28,33,49,63,72,76,70,62,50,41,47,71,102,137,180,220,231,
  231,232,232,224,197,173,155,146,133,126,126,126,126,124,125,130,128,126,124,
  118,109,93,78,71,62,54,51,56,63,63,72,88,105,125,145,167,192,206,211,211,205,
  194,179,165,161,157,149,141,133,125,113,102,93,90,94,104,117,129,134,136,137,
  129,119,108,101,97,92,93,96,104,109,113,122,132,144,153,157,160,166,167,167,
  167,164,157,148,134,112,90,67,47,39,40,49,69,96,120,134,139,142,147,148,148,
  155,172,194,206,207,203,195,179,161,144,132,130,130,127,125,124,117,108,104,
  104,108,112,116,119,116,109,98,81,68,63,63,67,78,91,103,115,121,126,135,146,
  156,164,175,185,186,177,171,165,162,165,168,174,182,176,162,145,122,99,80,66,
  66,76,87,96,104,110,112,109,107,110,113,116,120,119,116,111,108,110,121,138,
  158,179,196,206,206,193,172,148,125,113,107,104,108,114,118,118,114,110,107,
  103,101,101,103,105,106,108,111,115,122,128,133,137,141,140,136,134,132,133,
  136,141,149,155,154,152,150,145,139,135,131,134,135,130,125,117,108,101,96,
  93,98,109,120,125,124,122,116,110,106,107,116,129,150,168,182,188,188,180,
  165,148,136,129,124,118,113,109,107,101,96,93,95,101,111,123,130,133,132,128,
  119,112,107,109,119,128,136,143,147,146,141,133,128,128,131,137,142,149,156,
  155,149,144,141,136,131,128,125,119,111,102,95,95,94,97,107,122,133,138,141,
  145,143,135,128,122,117,114,113,112,113,113,117,122,126,129,132,136,140,142,
  145,147,153,160,163,166,169,167,160,148,135,123,111,102,94,90,89,93,105,118,
  131,146,160,173,182,177,167,153,135,118,102,91,89,92,97,106,117,127,133,135,
  134,133,132,129,127,123,122,124,127,129,133,134,132,128,120,109,99,91,87,86,
  89,100,114,127,138,149,158,164,168,167,168,168,163,155,147,137,125,116,113,
  114,120,128,138,145,150,151,145,135,121,108,101,98,100,104,109,114,118,122,
  125,124,121,118,114,113,112,111,116,127,139,154,170,182,187,182,173,160,141,
  123,111,105,105,106,109,113,115,115,114,113,113,114,116,120,123,123,123,125,
  128,131,134,139,143,145,145,143,140,138,138,141,146,151,154,155,152,144,136,
  129,127,126,125,122,115,108,99,87,76,74,82,94,109,125,137,145,144,135,128,
  125,124,130,141,154,162,162,158,147,136,127,119,115,118,122,126,128,126,120,
  115,110,108,111,118,126,131,137,137,135,132,128,125,122,120,123,126,129,130,
  132,135,137,139,139,139,138,136,134,133,131,131,130,129,129,129,128,125,122,
  122,125,131,140,148,155,160,162,158,150,141,135,130,124,118,116,115,112,108,
  101,96,95,97,102,109,116,123,128,128,125,122,121,123,127,133,141,150,155,156,
  153,145,136,129,123,120,120,123,128,131,132,135,136,136,137,138,139,137,133,
  129,124,118,112,108,108,110,114,118,123,127,128,130,134,140,145,151,156,159,
  158,152,142,131,120,111,103,99,98,99,99,100,101,102,104,109,116,123,131,138,
  144,148,148,145,144,142,139,137,134,131,129,129,130,129,130,132,135,136,138,
  138,136,133,131,128,127,125,125,125,125,124,122,122,123,124,125,127,129,132,
  135,137,137,138,139,141,140,139,138,136,135,132,128,124,122,121,120,121,123,
  126,129,131,128,125,122,119,118,119,121,125,128,131,132,130,128,125,120,118,
  119,121,125,127,129,130,131,131,132,133,136,138,139,141,141,140,138,136,134,
  132,131,129,128,128,128,127,129,132,135,137,136,134,130,122,112,106,103,104,
  107,112,118,125,129,128,125,124,124,124,126,131,139,143,144,143,140,134,127,
  120,117,116,116,119,121,124,128,131,135,139,143,149,150,150,149,144,136,128,
  123,119,115,115,119,122,125,128,128,128,127,126,125,126,128,131,134,134,135,
  135,134,132,131,129,128,130,131,130,131,131,130,127,126,125,124,123,123,124,
  123,123,123,121,120,118,114,112,111,111,113,117,124,130,136,143,147,149,148,
  147,145,142,143,145,146,147,147,146,143,137,131,127,123,120,120,119,120,119,
  118,117,116,115,117,121,125,128,129,128,128,126,121,120,122,126,132,139,145,
  150,152,151,148,142,137,133,130,129,130,131,132,133,132,132,131,128,127,124,
  122,120,117,116,116,116,118,120,122,124,125,126,127,128,129,131,133,137,140,
  142,145,146,145,143,142,141,137,135,133,131,129,126,124,123,123,125,126,126,
  127,127,125,122,119,119,119,119,120,122,125,127,128,128,128,127,126,126,128,
  129,132,134,137,139,139,138,138,138,136,137,139,139,141,141,139,135,129,124,
  119,114,112,113,115,118,120,121,122,121,119,117,118,120,125,129,134,139,141,
  140,137,135,134,132,133,135,136,136,135,132,128,122,119,117,118,122,126,130,
  135,137,138,136,133,131,130,130,128,128,130,130,129,127,124,122,119,115,113,
  111,111,113,118,122,128,134,142,147,149,149,149,147,143,140,138,137,136,135,
  135,136,135,135,133,132,130,127,124,120,115,111,107,105,107,110,114,117,121,
  123,122,119,118,116,116,118,124,131,140,147,153,155,157,156,154,150,148,147,
  145,142,140,136,132,127,123,121,120,121,122,123,123,122,118,114,111,108,107,
  107,109,114,118,122,124,127,129,132,135,138,142,147,148,148,147,146,144,140,
  138,137,136,134,133,130,128,124,121,119,118,118,119,120,122,124,125,125,124,
  123,122,120,120,120,122,125,128,132,137,141,144,145,144,142,138,134,130,126,
  125,124,124,124,124,124,124,123,121,120,120,119,120,121,121,123,123,125,127,
  128,130,132,134,136,136,136,135,133,132,132,131,131,131,132,132,133,133,133,
  133,132,131,129,128,126,124,122,121,121,121,122,123,124,124,125,124,123,122,
  123,123,124,126,128,129,131,132,133,133,134,136,136,136,136,135,133,130,127,
  124,122,120,118,118,120,122,123,124,125,126,127,127,128,128,128,128,128,128,
  127,128,128,127,127,128,128,128,130,131,132,132,133,133,133,132,132,131,132,
  131,131,131,129,128,128,126,126,125,124,125,124,124,124,123,123,123,123,124,
  126,128,130,131,132,133,133,132,131,132,133,132,132,132,132,130,128,128,127,
  127,128,128,129,130,130,130,129,129,128,127,128,128,128,128,128,129,128,127,
  126,125,125,127,128,129,132,134,135,135,134,133,132,131,130,130,131,131,130,
  130,129,128,126,125,125,125,126,128,128,129,129,128,127,125,124,123,122,123,
  124,125,126,127,128,128,128,128,128,129,131,133,135,138,140,141,142,142,139,
  135,132,128,126,124,123,125,127,130,131,132,131,129,127,124,122,122,123,125,
  127,129,131,132,131,129,128,127,127,127,127,128,129,130,132,133,134,135,135,
  134,131,129,128,126,125,125,127,129,132,133,134,134,132,130,128,126,124,124,
  124,124,124,125,125,125,125,126,128,130,131,132,133,133,133,132,132,133,133,
  134,135,135,135,133,131,129,128,126,125,124,123,122,122,123,124,125,126,128,
  128,128,128,126,125,124,123,124,124,126,128,130,132,133,134,135,136,137,138,
  138,138,136,134,131,128,125,123,122,121,122,123,125,126,127,127,127,127,126,
  127,127,127,127,128,128,128,128,128,129,130,131,133,133,134,134,134,134,133,
  131,130,128,127,126,124,123,122,121,121,121,122,123,124,126,127,128,129,131,
  132,132,132,133,133,132,130,128,128,128,128,128,130,132,133,133,131,129,127,
  124,122,122,123,124,125,126,126,126,125,124,124,125,127,129,131,132,132,131,
  130,129,128,129,131,133,134,134,132,130,128,124,122,121,122,124,125,127,128,
  128,128,128,127,127,126,126,125,125,124,123,122,123,125,128,130,132,134,134,
  134,132,131,129,128,129,129,129,129,128,127,125,123,123,124,126,128,131,134,
  135,136,134,133,131,129,128,126,125,124,122,120,120,119,120,121,124,127,129,
  131,133,134,134,134,133,132,131,130,129,128,126,125,124,123,123,123,125,126,
  128,129,130,130,131,131,131,132,133,134,133,132,130,128,125,122,120,120,121,
  122,125,127,128,130,131,131,132,132,132,132,131,130,129,127,126,124,125,126,
  128,129,131,131,131,131,129,128,128,128,128,129,130,130,129,129,128,127,126,
  125,125,124,124,124,124,124,125,127,128,129,130,130,129,129,128,129,129,130,
  132,133,134,135,134,133,132,132,131,131,130,128,126,124,121,119,118,118,119,
  121,123,125,126,127,127,127,127,127,128,129,130,131,131,132,132,131,130,130,
  129,130,130,131,132,133,134,134,134,133,131,129,127,124,122,120,119,118,118,
  119,119,120,121,123,125,127,128,130,131,132,132,132,132,131,130,130,129,128,
  128,128,128,129,129,130,131,131,131,130,129,128,128,127,126,126,126,126,125,
  125,124,124,124,124,125,126,126,126,127,127,127,127,128,128,129,129,129,128,
  127,126,124,124,123,124,125,126,128,129,129,130,130,129,129,128,128,128,127,
  127,126,126,126,126,127,127,128,128,128,127,128,128,128,129,129,129,128,127,
  126,126,126,126,128,129,130,130,129,129,128,127,127,128,128,130,130,130,129,
  128,126,124,124,123,124,125,127,128,129,129,130,131,132,132,132,132,132,130,
  128,127,126,126,126,126,127,127,127,127,126,125,125,126,127,128,129,131,131,
  131,130,129,128,127,127,128,129,131,131,132,132,131,130,129,128,128,128,128,
  128,128,128,127,126,125,125,125,126,126,128,128,129,130,131,131,131,131,130,
  130,130,129,128,128,127,127,127,128,128,129,129,129,128,127,126,125,125,126,
  126,127,128,128,128,128,128,129,129,131,132,133,133,132,132,132,131,131,132,
  132,132,131,130,129,128,126,125,125,125,126,126,127,126,125,124,123,122,122,
  122,122,123,123,124,125,126,127,129,131,134,135,137,137,137,136,134,133,131,
  130,130,129,129,128,127,126,126,125,125,126,126,126,126,126,126,126,126,126,
  126,126,125,125,124,124,124,125,126,127,128,128,128,128,128,128,128,128,129,
  129,129,129,128,128,128,127,127,126,126,125,125,125,125,125,125,126,126,126,
  125,124,123,122,121,120,121,121,123,124,126,128,129,131,133,134,135,135,135,
  134,133,131,129,128,127,127,128,128,129,129,128,127,126,125,125,124,125,125,
  125,126,125,126,127,126,127,126,125,125,125,125,127,128,129,130,130,131,130,
  130,129,128,128,128,128,128,127,128,128,128,128,128,130,131,131,132,132,132,
  132,131,131,130,129,129,129,128,128,127,127,126,126,125,125,125,125,126,127,
  128,129,129,129,129,130,130,129,129,128,127,126,126,126,126,128,129,131,132,
  132,133,133,133,132,132,131,131,130,129,128,127,127,126,125,125,125,126,126,
  127,128,128,128,128,128,127,127,128,128,129,130,131,132,132,131,130,129,129,
  129,129,129,130,131,131,131,131,131,130,129,128,127,125,125,125,126,126,127,
  127,127,127,127,126,126,127,127,128,128,128,129,129,130,130,131,132,132,133,
  133,132,132,131,130,128,127,126,125,124,125,125,126,127,127,126,126,125,125,
  125,125,126,127,128,128,128,129,130,130,130,130,129,129,128,128,128,128,128,
  128,127,127,126,126,127,128,129,129,129,129,127,126,124,123,123,124,126,127,
  128,129,130,130,130,130,129,129,129,129,129,129,129,128,128,127,127,126,126,
  126,126,126,126,126,125,125,124,125,126,127,128,128,128,128,127,125,124,123,
  123,124,124,125,125,126,127,127,127,127,127,126,125,124,123,123,123,123,124,
  124,124,124,123,122,122,122,122,123,125,127,129,131,132,133,133,133,133,133,
  133,133,133,132,130,128,127,125,124,124,124,125,125,125,125,124,123,123,123,
  124,124,125,126,127,127,127,126,125,125,125,126,127,127,128,128,129,130,130,
  130,130,131,132,132,133,133,133,133,133,133,132,131,130,129,128,127,127,126,
  127,128,128,129,129,129,128,128,128,127,127,127,127,128,128,129,129,129,128,
  128,128,129,129,130,130,131,131,131,131,131,131,132,134,136,137,139,140,139,
  138,136,135,134,134,135,136,137,138,137,135,132,130,128,127,127,128,129,130,
  130,130,129,128,127,127,127,127,128,128,128,128,128,127,127,127,127,128,128,
  127,126,125,123,121,120,118,117,115,112,108,103,99,96,95,98,107,121,138,154,
  168,177,178,173,161,148,138,133,135,141,147,151,150,142,130,117,106,102,104,
  112,122,130,138,141,140,137,133,133,133,132,130,124,116,106,97,92,92,97,107,
  117,125,130,130,129,127,127,130,136,144,152,158,160,158,153,147,140,134,132,
  130,131,131,132,132,131,130,129,130,130,130,130,128,127,125,121,118,116,115,
  114,113,112,111,109,105,101,98,95,93,93,97,103,111,119,126,132,138,138,139,
  140,141,145,149,155,162,167,169,169,164,158,149,142,138,135,135,135,133,133,
  130,125,120,116,113,113,111,113,112,111,108,104,102,102,101,105,109,112,117,
  118,120,121,121,123,128,132,139,145,150,154,155,155,154,152,153,152,152,153,
  153,154,153,149,145,142,137,134,131,128,128,126,125,125,124,123,122,122,123,
  124,125,124,124,122,119,115,112,108,104,100,95,89,80,72,68,67,72,82,101,124,
  148,172,190,200,202,198,188,178,168,162,162,163,164,163,156,145,133,118,107,
  101,102,111,121,132,142,145,145,140,131,126,120,114,110,104,97,90,82,77,77,
  81,91,103,116,125,130,132,131,129,131,134,143,153,160,165,167,164,160,155,
  152,152,154,157,159,159,156,153,149,146,142,138,138,135,131,126,120,114,110,
  106,104,103,102,99,96,90,82,71,60,52,51,50,54,69,90,118,145,168,188,203,201,
  195,182,169,164,162,167,181,187,189,187,173,156,137,120,115,117,122,133,140,
  146,146,138,127,118,109,102,96,92,89,84,77,73,73,76,82,93,101,110,116,113,
  111,109,105,108,116,126,141,153,165,176,183,186,187,187,188,187,184,181,176,
  169,161,151,141,129,119,109,102,97,91,90,89,91,94,94,94,93,89,83,78,73,70,68,
  67,70,79,91,107,125,144,160,173,179,180,179,173,169,173,179,188,198,207,211,
  205,190,173,157,142,127,121,121,122,123,123,121,116,105,95,91,86,84,83,84,88,
  89,90,94,97,103,107,111,117,120,121,122,123,126,131,138,145,150,152,151,144,
  134,125,116,111,109,114,128,141,150,163,175,176,173,164,155,148,131,118,116,
  111,103,105,105,106,105,102,103,105,103,102,113,129,139,155,173,183,192,192,
  184,176,162,139,128,117,107,105,105,107,112,115,121,128,131,134,135,138,139,
  134,133,133,128,126,124,119,112,105,95,84,71,56,45,37,32,34,49,71,101,135,
  167,199,223,234,235,233,228,223,218,218,219,212,203,185,156,130,105,86,79,73,
  74,84,92,93,95,92,91,95,95,103,113,109,106,103,93,83,77,75,79,84,79,76,75,67,
  60,64,80,113,153,191,224,231,230,232,230,212,191,180,180,189,200,207,203,195,
  177,152,128,115,105,100,101,103,107,109,106,100,94,87,83,82,80,80,79,80,81,
  82,88,95,100,101,97,88,73,49,30,28,27,35,65,107,153,198,230,232,232,231,230,
  229,228,228,227,228,229,229,211,179,150,120,91,70,58,54,62,74,81,90,99,100,
  104,110,110,110,112,104,97,91,81,80,82,86,96,109,118,125,129,126,121,122,128,
  138,153,169,187,202,208,209,209,203,193,184,177,168,158,147,134,120,102,79,
  60,38,29,29,27,26,37,76,120,163,197,213,210,188,153,120,95,87,101,128,156,
  173,177,163,134,101,80,78,94,129,170,214,229,228,230,211,179,156,149,147,151,
  150,136,117,92,67,54,54,66,89,117,138,146,139,123,102,88,84,92,109,119,121,
  114,95,70,40,26,27,25,27,53,95,139,187,223,229,229,231,231,230,203,179,165,
  163,175,193,204,196,175,141,107,70,46,42,58,83,113,138,148,139,120,106,100,
  106,118,128,133,128,111,91,72,59,64,82,112,148,173,181,174,154,136,126,124,
  137,160,181,187,180,164,140,113,85,69,60,53,44,43,44,55,76,107,153,195,227,
  229,228,199,153,113,86,86,108,151,192,218,218,196,154,111,81,72,80,102,136,
  174,199,195,173,145,114,90,74,67,67,67,64,71,78,89,109,132,158,179,186,178,
  157,120,79,45,22,23,24,29,56,96,136,173,204,218,221,214,191,165,144,143,158,
  190,222,231,230,232,222,167,108,60,31,32,51,82,117,146,161,158,137,116,101,
  96,101,110,113,110,98,80,65,55,51,53,52,41,28,22,22,20,22,23,69,149,219,234,
  233,236,235,234,232,227,223,232,234,234,234,227,189,141,100,69,64,76,91,109,
  121,124,113,104,92,91,91,99,106,103,86,61,36,24,24,24,26,29,46,59,67,74,80,
  94,123,161,212,234,232,234,232,232,228,195,155,126,114,117,123,120,113,108,
  103,94,86,91,96,97,101,117,131,142,151,163,179,188,177,157,124,78,34,23,24,
  25,27,62,115,154,182,200,212,210,201,196,196,201,202,192,180,157,127,105,95,
  94,108,122,136,146,146,143,139,130,126,127,130,131,119,93,65,33,24,25,24,25,
  26,28,46,80,112,138,162,186,211,231,232,232,231,230,230,226,198,175,153,135,
  129,131,134,140,146,145,148,144,140,137,143,150,150,144,130,115,91,61,33,26,
  26,26,26,37,57,66,67,63,57,36,23,22,28,51,87,132,188,228,232,233,233,232,233,
  232,232,232,233,233,233,232,210,179,145,117,107,97,100,101,96,92,88,72,59,49,
  46,54,56,54,51,42,30,29,27,26,25,25,28,65,107,142,170,185,187,182,179,178,
  186,200,223,231,231,232,232,232,232,232,230,231,224,207,191,177,161,141,128,
  112,103,90,71,55,44,27,21,22,23,23,24,36,55,61,59,47,28,26,26,25,25,24,35,89,
  150,209,232,229,230,229,231,230,231,231,231,230,231,229,230,219,195,181,173,
  175,183,185,174,152,121,83,51,31,35,51,64,81,86,75,52,29,25,27,26,38,65,78,
  80,56,28,29,27,28,26,39,121,212,232,230,232,232,231,231,230,230,231,232,232,
  233,231,232,217,168,125,93,78,69,64,72,76,63,43,26,25,24,28,46,67,66,49,31,
  27,27,26,25,24,25,28,73,134,175,207,225,227,226,224,215,208,201,207,225,230,
  229,230,230,230,230,230,228,203,173,148,131,125,122,116,106,88,71,59,51,51,
  45,35,26,25,25,26,26,29,48,67,75,69,43,24,26,22,25,22,46,128,209,233,230,234,
  232,234,225,202,201,225,233,232,233,231,232,222,184,139,116,108,105,111,109,
  98,90,72,42,37,44,69,105,116,106,86,48,24,27,24,26,26,28,47,74,82,85,89,87,
  95,124,155,197,228,232,232,233,232,232,232,232,231,232,230,232,224,174,122,
  78,51,46,57,68,83,88,84,74,61,56,64,69,73,86,85,74,60,43,29,27,28,27,29,40,
  62,90,117,142,171,199,221,231,231,226,217,213,217,229,230,230,230,231,227,
  186,134,92,68,66,73,86,108,119,113,96,77,69,66,71,84,107,116,111,93,69,48,34,
  31,33,42,51,56,51,43,48,65,100,142,202,233,232,234,233,232,232,223,196,195,
  207,223,222,203,177,144,103,69,48,45,51,57,70,80,76,60,43,36,43,49,62,68,67,
  69,59,47,47,51,67,96,115,137,162,178,189,193,195,206,221,228,230,229,230,214,
  205,212,213,210,196,188,189,166,124,106,91,75,64,58,68,78,72,67,65,61,61,62,
  65,71,73,74,70,67,63,63,70,74,85,102,114,133,159,180,214,229,228,229,229,228,
  201,154,126,101,77,79,94,107,134,155,159,165,163,156,152,149,133,119,107,79,
  60,54,49,60,77,87,101,105,84,67,54,39,37,53,88,131,167,191,210,222,223,220,
  220,219,214,202,191,178,158,134,119,121,132,155,173,186,186,168,141,113,87,
  66,53,50,60,73,83,92,94,89,86,83,80,81,82,68,48,28,27,25,39,66,113,169,218,
  234,234,234,234,221,181,162,153,159,168,179,188,187,175,166,160,154,142,135,
  135,133,126,107,81,65,58,64,78,93,110,120,121,112,91,78,77,72,83,106,126,145,
  151,144,138,128,120,121,127,130,134,137,135,128,120,117,125,139,161,187,201,
  205,200,184,165,144,128,120,107,92,76,71,72,67,66,79,99,128,149,159,160,150,
  134,114,103,105,109,119,130,139,141,134,125,124,123,119,115,116,123,123,121,
  117,118,132,152,166,180,193,196,189,173,157,151,146,143,141,142,143,130,116,
  106,99,95,93,97,115,129,132,128,122,119,111,103,102,103,104,93,79,75,66,59,
  59,60,77,94,99,112,125,132,146,155,170,192,206,212,213,210,203,187,171,162,
  159,157,146,140,141,138,130,123,122,130,134,129,125,121,117,105,95,90,88,90,
  92,92,97,98,93,88,85,83,88,94,103,118,128,135,145,156,163,169,170,169,172,
  170,157,149,144,140,136,132,132,133,134,132,126,118,115,121,123,119,120,128,
  142,150,146,145,147,140,129,113,96,92,89,86,87,96,107,117,126,130,129,125,
  119,122,122,113,108,115,128,138,141,149,166,175,174,169,166,165,161,151,148,
  153,156,152,147,143,142,132,117,111,107,99,92,89,93,97,92,88,87,87,84,86,90,
  90,94,99,106,118,128,132,143,149,150,149,149,155,160,159,164,173,179,180,175,
  172,170,160,152,148,141,133,122,111,102,93,92,96,101,107,111,113,112,107,99,
  91,87,89,99,108,117,127,137,141,140,142,143,143,142,141,144,149,149,146,147,
  150,151,151,155,160,157,151,146,146,142,131,124,124,124,120,117,118,121,118,
  117,117,117,118,114,105,99,99,101,100,101,104,106,104,102,101,102,108,111,
  115,121,128,133,136,135,134,134,135,140,153,168,171,169,169,166,158,146,137,
  137,139,137,140,148,154,154,151,149,149,149,143,137,132,124,113,105,98,91,90,
  90,92,97,100,99,99,95,95,98,101,107,113,120,127,131,134,140,147,151,152,155,
  160,161,159,155,149,144,138,135,138,141,142,144,142,141,139,138,136,130,126,
  123,120,118,114,109,105,97,91,87,88,89,89,96,104,114,118,120,124,127,128,130,
  133,135,141,146,148,146,147,150,149,148,148,153,156,155,153,150,147,143,140,
  139,138,138,138,137,138,132,122,114,110,106,101,96,94,98,100,99,99,102,105,
  106,104,104,109,113,116,121,128,131,133,136,136,134,134,137,141,143,147,152,
  158,161,157,150,145,142,142,140,143,151,158,163,169,168,163,157,148,138,128,
  118,110,104,97,91,88,89,95,101,105,108,112,111,107,105,105,106,111,116,123,
  131,139,142,142,142,143,142,140,140,140,138,133,130,128,126,126,130,137,142,
  149,156,161,160,153,146,142,138,132,128,123,118,113,106,97,92,89,87,85,83,88,
  93,94,97,103,111,117,124,135,143,146,150,154,154,149,142,139,137,132,129,129,
  135,141,145,150,155,160,163,163,164,163,159,155,151,143,135,130,127,121,116,
  113,109,106,102,98,92,88,86,86,88,92,101,110,117,122,125,128,131,133,135,136,
  135,136,139,140,142,139,137,139,143,144,146,151,155,158,158,157,159,158,155,
  153,150,147,142,136,130,125,118,110,105,101,97,96,98,102,104,106,111,114,115,
  116,115,116,118,118,121,125,127,128,129,129,131,132,132,134,138,140,142,147,
  147,146,148,150,152,154,156,157,154,147,138,131,123,117,114,112,113,116,117,
  115,112,108,104,101,101,103,105,107,112,116,121,124,126,130,135,139,142,144,
  145,147,147,147,148,150,154,157,160,161,161,159,158,155,151,146,142,138,133,
  127,121,117,111,106,104,104,105,106,105,105,106,105,104,103,104,105,107,109,
  112,116,121,125,130,136,140,144,147,149,148,147,148,148,148,148,151,155,156,
  156,155,153,150,148,145,141,139,135,131,127,122,118,115,112,109,107,108,107,
  103,98,95,94,95,96,101,109,114,122,128,131,134,135,136,138,138,138,140,141,
  143,144,145,147,150,151,154,155,154,153,150,147,144,139,136,133,129,125,123,
  121,117,115,113,112,112,110,109,106,102,100,101,102,105,111,120,127,131,136,
  138,139,139,139,138,141,143,144,144,143,139,135,132,129,128,130,135,140,144,
  147,148,147,146,143,140,136,133,129,126,120,114,108,104,100,96,95,97,99,101,
  104,109,116,121,126,133,137,140,142,142,140,138,138,140,140,140,142,144,143,
  142,141,141,139,138,140,141,139,138,138,138,136,136,137,137,135,131,125,119,
  113,108,105,101,103,107,107,108,108,106,107,107,108,113,122,130,139,144,147,
  149,149,147,146,145,141,141,141,139,137,135,134,135,135,135,139,141,143,143,
  140,138,137,134,133,133,129,129,129,126,121,115,111,109,107,105,105,107,109,
  111,115,116,119,123,124,126,129,130,133,135,136,139,139,139,142,143,143,143,
  142,141,138,136,136,135,134,136,135,134,134,132,130,128,127,128,128,128,128,
  125,121,118,114,110,109,110,111,114,118,124,126,127,129,131,133,134,137,139,
  138,139,139,136,136,134,133,133,132,132,131,130,130,129,128,130,130,132,135,
  136,138,137,134,132,130,128,126,124,125,125,123,123,121,119,121,123,123,122,
  121,121,120,118,117,119,122,126,129,134,138,140,141,140,138,136,133,131,130,
  129,129,127,126,127,127,127,129,131,134,136,139,140,139,135,132,129,127,123,
  120,118,119,119,116,113,115,117,117,118,117,118,121,124,128,133,136,140,145,
  148,149,148,145,142,138,133,131,130,128,128,128,128,128,129,128,128,128,127,
  127,126,125,124,125,127,128,129,131,129,128,128,125,123,122,120,119,119,118,
  120,120,121,124,128,132,136,139,143,145,145,143,141,139,135,130,128,127,125,
  126,126,124,124,123,122,123,123,125,127,127,127,128,128,128,129,129,128,129,
  128,128,128,125,122,120,119,119,120,120,121,124,126,128,128,130,133,135,135,
  135,137,138,138,137,137,136,133,131,129,128,127,127,128,129,128,129,131,131,
  131,132,133,134,135,134,132,129,127,126,124,122,122,122,122,121,120,121,121,
  122,122,124,127,128,129,132,133,135,137,139,138,138,138,138,137,136,134,133,
  132,131,131,130,130,130,131,131,131,130,130,129,128,125,124,122,122,122,123,
  123,125,127,127,125,124,124,123,124,124,125,127,128,130,132,132,133,136,138,
  139,141,141,140,139,136,132,129,127,124,123,123,123,123,124,124,124,124,126,
  127,128,130,131,131,130,129,129,129,129,129,130,130,129,129,128,127,126,125,
  125,126,127,128,129,129,130,130,130,131,132,133,134,134,133,133,130,129,128,
  128,128,128,129,130,130,130,129,128,126,126,127,128,129,131,133,133,133,133,
  132,129,127,125,123,122,123,124,125,126,127,127,128,128,127,127,127,128,129,
  130,131,132,131,131,131,130,128,128,128,128,128,128,127,126,125,126,128,128,
  129,131,131,131,130,128,126,126,126,126,128,129,130,130,130,130,129,129,129,
  128,128,128,129,129,129,128,128,128,128,130,131,133,134,134,133,131,128,127,
  126,126,126,126,125,124,124,124,125,126,126,127,128,129,130,131,132,133,133,
  132,131,130,129,129,128,128,128,128,128,128,127,127,128,129,131,132,133,134,
  132,131,130,128,128,127,126,125,125,124,125,125,126,127,129,130,131,132,130,
  128,127,125,124,125,126,128,130,131,132,133,133,131,130,129,128,128,127,127,
  126,127,127,128,128,127,128,129,129,130,130,130,131,131,130,129,128,129,128,
  127,126,126,126,127,126,127,127,126,125,126,127,128,128,129,130,130,131,132,
  131,130,129,130,130,128,126,125,124,124,125,125,127,129,131,132,132,134,133,
  132,130,130,130,130,129,129,128,127,126,125,125,125,125,126,127,126,127,126,
  126,126,128,129,131,131,131,131,130,129,128,127,126,124,124,123,122,122,122,
  123,124,125,128,130,132,134,134,134,135,134,134,134,133,131,130,128,127,125,
  125,125,125,126,126,126,126,126,126,126,127,128,130,130,131,131,131,130,128,
  128,127,126,126,125,125,124,124,125,126,126,127,129,130,130,131,132,132,133,
  133,133,133,134,133,132,131,131,130,129,128,128,127,127,126,125,125,125,126,
  127,128,128,128,129,129,129,128,128,127,127,126,125,125,126,126,126,126,126,
  127,128,128,129,130,131,133,132,132,132,131,130,130,129,128,128,128,128,128,
  127,127,126,126,126,126,127,127,128,128,128,128,128,128,128,128,128,128,128,
  127,127,127,127,128,128,128,129,130,130,130,129,129,128,128,128,128,129,130,
  131,131,130,129,128,128,127,127,127,127,127,126,126,126,126,127,128,128,129,
  129,129,130,129,128,128,128,127,127,127,128,128,129,130,130,130,130,129,128,
  128,128,128,128,128,128,127,127,127,127,127,127,126,127,127,127,127,127,128,
  128,128,128,128,128,129,129,128,128,128,128,127,128,128,129,129,130,130,130,
  130,130,130,130,129,129,129,128,128,127,126,126,125,124,124,125,126,127,128,
  128,128,128,128,128,127,128,128,128,128,128,128,128,128,128,128,127,127,128,
  128,128,128,129,130,130,130,131,131,131,131,130,130,130,130,129,129,129,128,
  128,128,127,127,126,125,124,124,124,124,125,126,126,126,127,127,127,127,127,
  126,126,126,126,126,126,127,127,128,128,129,130,130,130,132,133,133,133,133,
  132,131,130,129,129,128,128,128,127,127,127,126,126,126,126,127,127,127,127,
  127,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,128,128,
  128,127,128,127,127,128,128,128,128,128,128,128,128,128,127,128,128,128,128,
  128,128,128,128,129,129,130,130,130,130,129,128,128,127,127,127,127,127,128,
  128,129,128,128,128,128,127,126,126,126,125,125,126,126,126,127,128,128,129,
  129,130,130,130,130,131,131,130,131,131,130,130,130,129,129,128,128,128,128,
  128,128,127,127,127,127,127,127,128,128,128,128,128,128,128,128,128,128,127,
  126,126,126,126,127,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  129,129,129,130,130,129,129,129,129,129,129,128,128,128,128,128,128,129,129,
  129,130,129,128,127,127,126,125,125,126,127,127,128,127,127,127,126,126,126,
  126,126,127,126,126,126,127,128,128,128,128,129,129,130,130,131,131,132,132,
  131,131,131,130,130,129,129,128,128,126,125,125,126,126,126,126,126,125,125,
  126,126,127,127,127,127,127,127,127,128,128,128,128,127,127,127,127,127,128,
  128,128,128,129,130,130,131,130,130,130,130,129,129,129,128,128,128,128,128,
  128,128,128,128,128,127,127,126,126,126,127,128,128,128,128,129,129,129,129,
  129,128,128,128,128,128,127,127,127,127,127,127,128,128,129,129,129,130,129,
  129,129,129,129,129,128,128,129,128,128,128,128,128,127,127,127,127,127,127,
  127,127,127,126,126,126,127,128,128,128,128,128,128,128,128,128,128,128,128,
  128,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,127,128,128,
  128,128,128,127,127,127,127,128,128,127,127,127,127,127,127,128,128,128,128,
  128,128,128,128,128,129,129,129,129,129,129,129,129,129,129,129,129,128,128,
  128,128,128,127,127,128,128,128,129,129,129,128,128,128,128,128,128,128,128,
  127,127,127,127,127,127,128,128,128,128,128,127,127,127,128,128,128,128,129,
  129,129,128,128,128,128,128,128,128,128,129,129,129,130,129,129,129,129,129,
  129,128,128,128,127,127,127,126,126,126,126,126,126,126,126,126,126,127,127,
  128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,130,129,129,
  129,129,129,129,129,129,129,129,128,128,128,128,127,127,126,126,125,124,124,
  124,125,125,126,126,127,128,128,128,128,128,128,128,128,129,129,129,129,129,
  129,128,128,128,129,129,129,129,129,129,129,130,130,130,130,130,129,129,129,
  128,128,128,128,127,127,126,126,126,126,127,127,127,127,127,127,127,127,127,
  126,126,127,127,127,127,127,128,128,128,129,129,129,130,130,130,131,131,131,
  131,131,131,131,130,130,130,129,129,129,128,128,128,127,127,127,126,125,125,
  125,126,126,126,126,126,126,127,126,126,126,127,127,128,128,128,128,128,128,
  128,129,129,130,130,130,130,130,130,130,130,130,130,130,130,129,129,129,129,
  128,128,128,127,127,127,126,126,126,126,126,126,126,127,127,127,127,127,127,
  127,128,128,128,129,129,129,129,129,129,129,129,129,129,129,129,129,129,129,
  129,129,129,129,129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,
  128,128,128,128,128,128,127,127,127,128,128,128,128,128,128,128,128,129,129,
  129,129,129,129,129,129,129,129,128,128,129,129,129,130,129,129,128,128,128,
  128,128,129,129,128,128,128,128,127,127,127,127,127,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,129,129,129,129,129,130,130,129,129,129,
  129,129,129,129,129,128,128,128,128,128,127,127,127,126,127,127,127,127,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,
  129,129,129,129,129,129,129,129,129,129,129,129,128,128,128,128,127,127,127,
  127,128,128,128,128,128,127,127,127,127,128,128,128,128,128,128,128,128,129,
  128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,127,127,127,127,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,
  129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,127,127,127,128,128,128,128,128,128,128,128,128,128,
  128,128,128,129,129,129,130,130,129,129,129,130,129,129,129,128,128,128,128,
  128,128,128,128,128,128,128,128,127,127,127,127,127,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,
  129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,127,127,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,129,129,129,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,
  129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,129,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,
};

static const char dsdgdth[] = {
  3,0,34,86,67,85,0,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,129,128,128,128,128,128,128,128,129,129,128,129,
  129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,129,129,129,129,129,128,128,129,129,129,129,129,128,129,129,129,129,
  129,129,129,128,129,128,129,129,128,129,129,129,129,129,128,128,129,129,129,
  129,129,128,128,129,129,129,128,128,128,128,129,128,128,129,129,128,128,129,
  129,129,128,129,128,128,129,129,129,129,128,128,129,128,128,129,129,128,128,
  129,129,128,128,128,129,128,128,128,128,129,128,128,128,129,128,128,128,128,
  128,128,128,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,129,129,129,128,128,128,128,128,128,128,128,128,128,128,129,129,128,128,
  129,129,128,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,129,129,129,129,129,129,129,129,129,128,128,128,129,129,129,129,
  129,129,129,129,129,129,129,129,129,130,129,129,129,129,128,128,128,128,128,
  128,128,128,128,128,128,128,129,129,129,129,130,130,130,131,131,131,131,131,
  131,130,130,130,129,129,129,128,128,129,128,128,128,128,129,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,129,129,130,131,131,131,131,132,
  131,130,130,129,129,129,129,129,129,128,128,128,128,128,128,128,127,127,127,
  128,128,128,128,128,127,127,127,128,128,128,129,130,131,132,133,133,133,132,
  131,130,129,128,128,128,127,127,128,128,128,128,128,128,127,127,127,128,128,
  129,129,129,129,128,128,128,128,128,128,128,129,131,131,131,130,130,130,129,
  128,128,127,127,127,128,128,128,128,127,126,125,125,125,125,127,128,129,129,
  129,129,129,129,128,128,129,129,130,132,132,130,129,128,127,126,125,125,125,
  125,127,128,128,128,128,128,127,126,125,125,126,127,128,128,128,128,128,129,
  130,130,130,131,132,134,135,134,133,132,130,128,126,124,122,122,123,125,126,
  127,127,127,126,124,123,123,124,126,128,129,129,129,130,129,129,128,127,128,
  130,134,137,137,136,136,135,133,130,128,125,124,126,128,128,126,125,124,121,
  119,116,117,118,121,124,127,128,129,131,133,133,131,130,131,134,138,139,137,
  136,135,133,130,127,124,123,123,126,128,128,128,127,126,123,120,118,118,120,
  123,125,126,126,126,127,128,128,127,128,131,135,139,140,139,138,136,134,131,
  129,127,126,128,129,129,127,124,122,120,116,113,113,116,121,126,128,128,129,
  131,133,132,129,128,130,135,139,139,137,134,132,129,126,123,120,121,126,129,
  131,129,128,128,125,121,117,116,118,122,125,126,126,126,128,129,128,126,127,
  132,138,141,140,139,138,136,133,129,126,123,123,125,127,125,123,121,119,117,
  114,113,116,122,127,129,130,131,134,136,135,132,131,132,136,139,138,135,132,
  130,128,125,121,120,122,126,127,126,124,124,123,120,117,115,117,121,125,128,
  129,130,132,134,133,131,133,137,139,140,139,138,137,136,133,129,125,122,123,
  123,122,120,118,116,114,111,110,113,118,123,127,131,136,140,142,142,140,140,
  140,141,140,139,135,132,131,129,125,123,122,123,122,120,120,121,119,116,114,
  114,115,118,121,126,130,134,137,139,141,142,142,143,143,141,138,136,135,133,
  130,127,126,125,123,120,120,119,116,112,110,110,110,113,117,121,126,131,134,
  138,143,146,148,149,150,149,145,143,141,137,131,127,127,125,120,117,116,115,
  110,105,103,104,105,108,115,122,127,131,136,141,144,146,148,151,152,150,147,
  144,142,137,132,129,128,125,123,123,123,120,114,109,104,102,102,104,109,113,
  117,124,131,134,136,139,143,146,149,152,155,155,153,151,148,142,135,129,125,
  120,116,113,110,107,104,103,103,101,101,105,113,121,127,134,145,153,154,154,
  154,153,147,140,137,138,137,136,137,138,136,130,123,114,109,103,100,101,105,
  111,114,116,120,126,128,127,126,129,136,142,146,152,157,155,147,143,144,146,
  141,133,132,132,126,115,111,111,105,94,88,92,103,111,114,120,132,141,140,140,
  148,158,157,146,144,150,149,141,137,140,139,128,120,117,118,115,106,95,94,
  102,105,103,111,127,136,133,129,137,147,147,143,144,147,142,135,136,141,146,
  142,132,124,124,125,117,108,108,111,105,98,103,116,123,118,118,126,135,138,
  142,146,148,150,147,136,130,135,139,133,123,124,130,128,120,116,118,113,107,
  105,108,118,128,129,122,123,135,139,134,135,144,147,136,125,124,131,130,122,
  120,127,135,138,133,128,130,130,115,98,104,123,128,117,117,132,139,128,119,
  124,131,125,120,124,134,145,147,141,135,140,144,130,118,125,137,134,121,110,
  109,109,99,89,93,113,128,135,135,145,165,164,143,128,138,151,141,124,128,147,
  144,120,108,108,108,96,85,89,108,128,137,136,134,146,157,146,123,119,135,140,
  131,124,131,146,136,110,94,105,118,106,101,110,139,158,151,140,139,148,142,
  123,110,115,135,135,121,114,119,123,103,82,80,104,128,130,128,140,169,182,
  163,137,135,147,133,110,108,127,144,130,109,99,102,102,84,75,80,111,143,149,
  146,151,173,177,150,123,117,134,129,107,107,126,142,128,106,103,114,122,110,
  98,101,127,152,152,138,135,150,156,134,115,115,129,126,107,104,115,130,125,
  115,111,116,134,143,138,128,137,161,155,127,111,123,135,115,91,93,114,126,
  113,105,114,133,148,140,128,131,152,166,150,131,126,136,135,112,98,105,122,
  115,90,82,98,116,121,118,123,141,165,170,153,137,143,158,146,119,115,134,146,
  126,103,99,106,108,86,78,90,118,142,143,139,143,166,174,146,118,114,135,137,
  113,109,128,150,136,106,95,104,120,115,102,103,131,165,166,141,129,143,150,
  123,94,91,116,130,113,102,116,146,152,129,108,119,154,159,131,111,133,165,
  151,109,95,118,129,100,76,86,114,128,116,113,127,159,176,158,136,143,174,176,
  133,99,105,132,120,80,70,97,119,94,74,89,126,148,137,128,145,192,210,177,140,
  144,170,159,109,79,94,118,100,65,61,94,115,92,72,99,157,180,155,139,169,208,
  194,139,110,129,149,120,82,85,118,126,102,83,96,133,140,114,106,144,189,179,
  134,119,144,156,118,78,81,116,124,99,92,123,158,143,111,118,161,183,155,132,
  151,184,166,119,98,107,106,79,58,69,96,110,109,106,128,166,172,149,146,174,
  196,175,142,142,163,150,108,79,88,93,66,53,68,110,126,111,118,153,186,171,
  142,146,174,177,147,128,140,149,126,97,87,101,106,83,71,98,142,144,122,128,
  162,179,148,117,130,159,150,115,108,131,136,115,100,109,137,136,112,111,139,
  164,143,116,118,142,145,118,107,127,138,122,110,118,137,134,110,112,138,155,
  133,117,133,151,139,113,114,137,138,117,109,130,145,133,110,113,143,150,121,
  106,130,155,142,109,105,133,141,111,91,120,147,138,123,130,157,169,141,119,
  128,145,134,106,107,125,131,113,96,110,122,110,102,121,144,147,141,151,171,
  173,150,132,142,145,122,95,93,110,107,85,85,113,129,121,115,131,167,175,151,
  138,156,171,150,119,110,120,116,88,78,100,120,119,120,134,154,156,142,143,
  156,159,139,124,127,124,105,85,91,109,108,97,109,145,163,152,138,156,108,103,
  88,67,78,116,140,136,126,141,164,175,160,137,144,159,148,108,81,83,97,93,77,
  92,131,163,155,137,147,171,189,173,142,133,131,115,77,52,65,97,119,114,121,
  146,174,170,150,149,168,187,169,130,109,98,85,59,45,72,121,153,143,132,153,
  179,180,160,153,159,165,132,87,64,62,75,71,83,121,170,182,149,136,157,187,
  183,164,151,145,125,72,37,25,58,87,102,126,160,187,173,157,159,189,197,177,
  150,125,97,51,28,32,75,112,132,141,152,167,158,152,159,187,195,176,141,107,
  71,35,20,42,97,137,151,149,153,166,168,170,180,197,192,159,111,64,38,26,35,
  75,132,165,161,141,136,150,159,173,189,202,180,132,81,40,29,39,71,120,163,
  173,148,131,139,162,180,191,195,186,145,81,27,13,38,74,113,156,180,164,128,
  115,137,175,201,208,198,167,109,49,15,23,65,108,148,167,159,133,118,131,163,
  199,219,210,174,117,52,15,20,59,105,147,172,162,130,108,114,147,186,219,222,
  192,135,72,31,22,53,101,145,168,157,125,104,108,135,173,212,226,197,139,74,
  32,21,51,100,146,177,169,131,99,102,134,171,208,225,204,148,80,35,19,43,98,
  149,177,164,124,95,96,127,163,206,234,217,155,79,32,21,46,92,142,179,176,131,
  87,86,126,165,201,223,216,164,88,35,19,44,98,150,177,167,128,89,78,108,154,
  202,231,223,170,97,44,24,47,95,150,181,170,128,86,75,103,146,191,222,220,171,
  101,49,32,53,100,149,177,171,131,86,70,97,146,196,227,221,177,108,47,24,46,
  102,152,172,166,134,92,66,87,141,197,230,222,174,111,58,32,47,101,158,181,
  166,127,85,64,85,139,194,231,228,179,108,50,33,52,104,155,180,170,131,83,57,
  79,139,199,231,226,185,120,58,32,48,101,157,181,167,128,85,60,77,133,194,233,
  230,182,113,54,34,50,98,153,183,176,135,85,57,75,132,192,230,231,190,123,59,
  33,51,101,154,179,169,132,85,54,68,125,190,230,233,191,125,61,34,49,97,151,
  178,169,135,92,62,72,125,191,233,236,193,126,65,35,47,90,146,174,162,128,87,
  61,72,123,188,230,238,201,132,69,42,59,97,146,174,165,129,87,57,64,114,181,
  226,233,199,138,80,46,57,92,141,173,164,129,92,65,67,107,171,222,236,208,144,
  86,53,62,89,130,166,165,134,97,70,69,103,160,214,237,218,157,95,55,55,76,114,
  154,161,133,98,77,73,96,149,210,241,230,175,112,67,58,75,105,143,158,139,103,
  73,64,89,139,195,231,234,191,127,76,60,76,104,140,158,142,107,78,61,79,128,
  189,228,236,198,136,85,63,73,97,138,160,151,115,80,57,69,115,174,218,233,204,
  143,88,60,70,97,135,160,157,126,92,64,68,109,171,221,239,210,145,87,55,59,84,
  126,159,160,127,91,66,69,104,163,220,246,223,156,96,60,64,85,123,153,159,128,
  90,60,58,97,159,217,243,224,160,97,60,63,84,120,152,163,138,100,70,61,93,150,
  209,239,228,166,103,64,64,86,118,145,154,134,98,69,59,91,144,204,237,232,173,
  110,69,66,81,113,148,163,147,108,76,63,91,140,194,229,227,176,109,67,62,77,
  106,136,155,148,115,83,62,85,139,198,235,235,186,121,75,67,79,107,137,154,
  147,113,81,59,77,127,187,228,233,188,122,74,64,79,110,140,157,153,121,86,60,
  77,129,187,228,234,189,122,73,63,73,102,141,159,156,125,90,60,71,121,182,227,
  234,191,122,68,58,72,105,142,164,164,132,98,65,68,118,180,224,232,193,124,68,
  53,66,99,141,166,166,135,101,67,71,118,175,221,234,195,123,69,57,69,99,138,
  161,164,141,103,62,65,119,180,223,230,193,126,70,54,67,103,143,167,165,139,
  103,65,61,111,175,221,232,196,126,63,50,65,100,143,171,171,148,110,67,62,111,
  177,221,232,198,128,64,42,46,77,128,164,167,147,118,74,61,106,173,219,236,
  206,135,72,45,46,80,131,168,175,159,128,78,62,107,176,226,241,208,138,70,35,
  33,70,129,176,186,165,128,80,60,93,165,222,241,210,141,74,42,42,78,134,180,
  188,161,119,70,55,96,168,217,227,196,130,59,28,36,75,138,187,191,162,122,73,
  61,109,184,224,222,190,126,63,36,44,83,140,183,183,152,110,65,61,117,188,221,
  215,176,111,60,41,50,90,145,182,180,162,124,76,71,125,187,210,202,170,116,70,
  50,58,98,144,174,172,157,121,76,70,119,185,206,194,164,118,77,49,54,95,147,
  181,183,172,131,79,71,121,180,197,187,154,107,74,46,50,93,149,183,185,173,
  130,84,82,129,181,191,180,146,102,68,42,51,96,153,181,184,173,128,85,94,143,
  183,183,167,129,95,74,55,65,107,158,176,175,164,124,97,113,156,179,164,142,
  103,82,78,76,91,124,154,157,156,155,134,129,153,184,173,134,103,77,80,90,93,
  102,125,143,146,148,146,134,144,171,185,154,116,87,65,80,103,115,121,132,138,
  134,140,144,142,164,197,198,146,98,63,51,72,99,117,132,146,148,141,138,131,
  139,170,202,196,144,96,58,47,67,97,118,135,148,151,145,141,138,153,184,207,
  178,123,76,40,36,65,98,121,142,158,159,150,140,146,168,193,201,163,110,63,37,
  44,76,102,123,141,152,151,145,139,151,183,213,203,140,78,41,36,55,86,112,131,
  145,150,145,140,140,159,189,213,197,135,74,39,37,61,89,111,130,144,149,146,
  144,153,178,206,219,178,108,49,28,38,65,88,112,136,149,151,150,151,163,190,
  217,213,160,85,31,22,44,76,96,121,140,149,147,146,151,174,210,235,211,135,57,
  21,23,48,79,106,128,136,140,143,150,160,187,222,236,195,118,47,22,29,56,79,
  102,122,132,139,145,154,174,204,230,227,176,95,36,27,43,64,79,103,123,132,
  137,147,164,189,222,239,212,145,69,24,23,45,66,84,111,131,142,150,160,172,
  198,229,232,186,114,51,26,31,49,69,92,116,130,139,149,161,181,214,243,230,
  169,95,45,30,39,53,69,92,115,127,131,147,169,195,230,249,218,143,70,37,38,47,
  57,74,97,117,125,135,158,185,215,243,241,189,110,52,34,41,50,61,78,102,119,
  125,140,174,205,233,254,229,157,77,34,30,43,56,70,88,112,124,128,151,192,228,
  255,253,199,112,45,29,39,52,66,78,93,110,119,133,175,219,248,0,234,158,74,28,
  29,45,60,74,84,100,113,120,143,196,244,0,255,198,110,41,31,45,56,73,83,86,97,
  108,126,169,227,0,0,227,152,73,37,45,50,58,70,79,92,105,112,137,192,253,0,
  252,193,115,57,50,59,57,60,66,70,83,104,127,164,216,0,0,235,175,106,61,52,49,
  47,48,58,77,97,116,146,192,245,0,0,216,143,83,55,42,30,32,45,65,87,108,132,
  168,215,0,0,254,194,121,64,36,14,7,27,59,92,117,130,149,186,241,0,0,236,163,
  90,47,24,1,3,32,67,99,125,141,166,209,0,0,0,210,136,68,33,17,6,17,50,85,110,
  131,159,194,239,0,0,224,165,102,47,16,5,8,26,60,98,124,150,184,225,0,0,252,
  192,122,65,32,10,3,24,54,81,107,126,153,199,254,0,0,217,139,70,37,26,17,29,
  53,75,100,120,142,180,230,0,0,242,167,87,34,26,32,42,54,63,72,90,113,154,211,
  0,0,0,211,124,52,27,30,38,47,52,60,77,102,140,195,251,0,0,238,156,76,27,20,
  29,39,47,57,69,92,128,180,235,0,0,0,195,116,46,19,20,27,37,47,63,87,115,153,
  202,251,0,0,233,155,76,31,24,29,37,45,61,79,101,132,182,234,0,0,0,191,115,50,
  20,19,30,37,49,68,89,117,161,211,0,0,0,220,150,81,40,27,28,34,41,56,78,103,
  136,186,242,0,0,246,179,109,58,29,21,26,31,43,63,94,128,172,223,0,0,0,210,
  138,78,41,26,25,24,29,45,71,109,155,203,254,0,0,239,173,107,58,30,26,26,21,
  34,60,94,138,188,237,0,0,0,195,128,79,44,22,22,20,26,50,86,126,169,216,0,0,0,
  216,150,90,52,29,26,27,28,47,74,110,150,189,233,0,0,248,184,113,67,41,26,29,
  27,36,63,97,133,176,222,0,0,0,205,138,83,41,20,22,25,35,63,88,120,163,201,
  241,0,0,238,166,98,55,29,21,28,36,52,81,116,151,183,226,0,0,249,190,112,52,
  24,15,23,42,64,86,109,138,168,204,255,0,0,215,142,75,39,19,15,27,47,76,105,
  132,161,192,236,0,0,227,167,92,43,24,22,27,38,69,102,128,155,180,220,0,0,242,
  192,123,60,29,13,14,24,51,94,125,148,173,199,248,0,253,206,150,82,39,24,17,
  20,42,85,125,155,176,187,221,0,0,223,168,95,39,22,21,23,30,60,105,137,164,
  186,213,0,0,246,193,133,67,30,23,22,21,40,84,124,152,175,190,232,0,0,208,150,
  90,44,28,23,22,33,72,116,144,170,184,212,0,0,235,176,110,51,26,26,24,23,49,
  100,134,160,181,197,240,0,0,198,131,73,35,26,34,35,39,75,121,147,170,186,220,
  0,0,228,164,97,43,24,21,23,34,64,109,143,168,180,198,249,0,254,188,117,56,24,
  18,28,34,49,95,135,153,171,181,216,0,0,223,151,79,30,18,21,28,41,72,121,152,
  172,179,197,248,0,248,179,104,44,15,16,26,33,56,105,144,166,178,183,219,0,0,
  215,142,73,33,11,13,29,42,82,135,163,175,176,195,246,0,246,171,95,43,21,16,
  21,32,64,115,156,173,172,173,210,0,0,216,136,66,30,19,27,35,47,85,137,167,
  175,169,181,226,0,250,182,98,38,17,18,29,45,72,120,165,183,177,177,203,244,0,
  215,132,57,21,11,23,38,61,99,147,173,177,173,190,226,0,242,171,87,38,24,26,
  32,46,81,133,167,172,166,172,203,244,254,205,126,60,33,30,32,29,49,97,149,
  176,179,172,182,217,253,245,179,98,47,35,43,32,24,52,110,158,179,173,168,191,
  237,0,232,155,78,35,28,27,18,34,81,142,178,185,171,176,210,252,253,197,110,
  41,16,20,21,28,60,119,170,191,184,171,181,223,0,240,157,66,17,14,27,27,44,92,
  151,178,183,169,163,189,241,255,209,129,54,15,20,37,38,56,108,157,182,183,
  159,159,207,255,244,178,93,35,20,32,30,42,86,141,175,186,172,151,167,219,248,
  215,146,70,23,23,43,51,62,105,157,187,189,163,153,186,236,229,166,96,42,20,
  30,46,63,102,150,175,180,171,151,158,209,244,214,148,80,36,27,29,26,50,106,
  156,179,180,162,158,190,233,232,185,127,70,35,31,38,42,72,127,167,186,176,
  148,151,204,252,228,160,90,41,24,30,35,51,105,165,195,195,165,140,156,218,
  237,194,125,58,19,21,45,59,78,131,177,194,182,153,140,188,248,226,163,91,32,
  5,20,50,72,113,166,190,187,162,137,148,214,239,202,139,71,21,12,31,53,81,137,
  180,189,165,137,134,183,241,229,180,106,40,9,18,53,74,108,160,193,184,148,
  123,139,211,240,205,143,67,15,1,31,68,95,146,192,198,165,134,126,170,231,225,
  177,105,35,1,8,50,84,126,179,202,180,141,119,137,205,232,203,149,72,13,1,26,
  70,110,160,197,197,157,119,113,164,225,225,189,119,40,1,3,43,90,137,184,207,
  183,132,102,123,196,232,215,164,84,23,2,23,65,113,163,198,199,156,108,95,146,
  217,232,203,132,52,10,8,40,87,134,181,206,182,129,97,110,177,224,218,177,102,
  36,4,18,58,102,149,189,197,157,114,96,133,206,237,211,145,69,25,13,32,74,127,
  170,197,178,130,101,101,150,212,228,194,121,55,22,23,51,93,138,179,197,163,
  115,88,109,177,230,226,172,99,48,22,31,62,109,153,185,182,139,102,90,126,192,
  234,215,156,91,45,32,46,78,122,159,183,168,127,98,98,136,199,223,195,139,82,
  48,42,60,90,127,155,168,152,123,101,104,149,210,220,183,128,84,56,57,72,98,
  129,150,155,140,115,97,108,157,208,207,171,128,90,67,69,80,103,125,139,147,
  136,114,104,119,170,210,201,167,132,97,76,77,84,103,118,129,134,127,108,105,
  124,171,206,202,173,142,106,85,84,86,99,113,121,126,119,106,112,130,169,201,
  199,179,149,113,94,90,89,96,106,109,110,111,108,117,133,168,200,202,184,157,
  128,109,96,85,88,95,94,97,104,112,126,133,159,196,205,190,165,136,114,103,91,
  85,81,82,86,99,112,129,139,162,196,203,191,173,147,121,106,94,82,70,71,77,91,
  108,128,145,160,191,206,198,186,164,132,112,99,78,61,60,71,81,100,122,149,
  163,182,202,198,190,174,146,115,94,77,64,58,68,74,88,120,147,167,172,191,203,
  193,186,164,128,92,73,62,61,63,71,82,112,147,166,178,179,192,195,186,176,143,
  97,72,68,65,70,68,81,107,131,153,172,185,182,189,184,164,153,123,89,75,70,69,
  77,81,99,121,145,168,179,184,180,169,158,139,136,119,91,80,74,77,93,98,104,
  124,150,170,180,177,169,153,146,130,128,116,95,90,90,97,108,111,119,133,152,
  160,163,159,155,146,136,128,122,116,106,112,113,111,114,119,122,128,135,145,
  145,149,145,145,136,118,116,122,127,130,129,124,125,125,118,116,123,128,135,
  135,143,135,132,121,122,130,137,139,142,138,134,125,116,109,108,113,124,133,
  126,128,124,130,130,138,146,154,155,150,134,126,116,105,101,109,112,116,123,
  118,129,130,139,144,151,154,153,142,137,128,126,125,121,108,98,98,104,123,
  125,138,143,144,147,144,140,139,135,139,146,144,131,114,100,91,98,105,125,
  136,142,140,131,132,131,130,143,152,152,144,137,129,119,119,115,117,119,119,
  124,120,119,119,124,141,144,141,132,128,129,132,139,142,142,142,136,127,114,
  106,110,120,138,140,135,128,119,115,115,121,131,137,144,147,141,138,135,135,
  138,139,135,126,122,118,114,114,110,115,123,122,123,120,122,131,148,165,173,
  168,157,141,129,119,113,109,108,113,106,97,99,105,119,134,147,152,153,153,
  156,156,157,151,142,132,125,116,98,94,97,106,121,132,136,133,128,129,129,136,
  137,134,138,143,139,136,134,136,136,135,134,128,127,126,123,121,114,109,112,
  117,117,112,114,127,142,156,164,163,161,157,150,140,130,121,110,109,107,96,
  92,98,110,124,134,140,142,146,153,159,162,160,148,137,131,122,113,104,107,
  113,119,126,127,124,123,125,132,139,136,130,127,125,125,126,132,139,144,148,
  148,144,137,129,129,130,130,119,108,103,99,98,102,112,126,138,147,149,150,
  152,157,159,157,145,132,124,114,105,97,96,107,120,127,125,121,123,130,140,
  148,147,139,135,135,132,128,125,125,129,135,134,130,125,124,129,133,131,123,
  115,110,111,108,109,117,128,138,146,151,152,147,144,142,141,139,131,122,113,
  105,101,103,110,117,120,125,131,135,136,140,141,141,140,136,133,128,125,118,
  115,122,129,135,136,135,132,129,132,130,124,117,111,108,110,113,113,113,123,
  137,147,152,149,144,146,149,149,140,128,120,114,111,111,108,104,105,113,122,
  127,127,131,137,140,139,136,134,135,133,128,123,127,135,140,142,137,129,126,
  130,133,128,114,101,98,102,108,108,109,115,128,141,149,152,151,152,156,156,
  151,140,128,120,112,106,103,103,106,109,111,111,113,121,133,139,138,134,129,
  131,133,133,131,132,138,143,144,143,138,130,128,129,126,116,105,100,99,100,
  100,101,110,126,136,143,147,149,151,156,155,151,142,133,128,121,111,100,98,
  106,117,122,122,120,123,129,135,136,133,131,130,128,124,120,118,121,125,128,
  133,140,143,144,144,141,134,128,128,126,121,111,102,101,108,116,118,121,127,
  133,140,146,147,143,137,130,124,122,124,127,128,128,128,127,128,131,135,137,
  134,128,122,122,121,114,107,105,111,119,126,129,132,135,138,142,147,152,150,
  144,137,128,120,114,113,117,122,123,120,119,123,129,132,131,127,126,128,128,
  123,117,113,115,121,129,135,142,149,153,153,151,145,138,131,127,120,111,103,
  102,105,112,114,114,119,128,136,142,143,140,136,133,129,127,124,123,127,132,
  137,136,136,138,139,138,136,131,125,121,117,114,110,105,104,108,117,124,128,
  133,139,141,143,145,143,143,142,139,136,131,127,123,120,118,116,118,123,126,
  128,130,129,126,123,120,117,118,121,122,124,129,130,130,134,139,142,144,144,
  141,139,137,129,117,111,108,107,110,114,118,125,134,139,141,141,139,135,132,
  128,123,119,117,114,112,113,116,120,128,139,148,152,151,145,138,132,126,119,
  115,115,116,118,121,123,125,130,135,135,136,136,135,132,126,118,112,109,109,
  110,117,126,134,141,146,149,148,143,137,132,130,130,128,123,118,115,115,117,
  117,116,120,125,128,131,132,129,127,125,123,123,124,126,128,134,139,140,141,
  140,138,138,137,134,131,128,127,125,120,113,108,108,111,114,117,124,128,130,
  132,133,132,131,131,133,135,135,135,136,136,135,131,132,133,134,136,136,134,
  130,125,119,115,109,104,103,106,112,117,122,128,132,133,135,137,139,139,141,
  143,143,139,133,128,125,123,123,125,129,133,133,132,130,126,119,113,108,104,
  103,105,109,113,117,121,127,133,140,145,148,151,153,153,152,146,136,126,119,
  115,113,113,116,120,125,126,126,123,119,116,114,113,112,111,111,114,118,124,
  128,135,143,149,154,157,159,158,154,146,136,125,115,107,103,103,105,109,112,
  116,117,119,120,121,123,125,127,128,128,127,126,126,129,133,137,142,145,149,
  151,149,145,140,133,125,113,103,99,98,99,104,110,116,121,124,127,130,133,136,
  140,142,140,136,132,128,124,123,125,125,126,129,134,138,138,135,134,131,125,
  119,115,111,110,112,117,121,123,126,129,131,133,135,136,138,140,138,134,126,
  119,114,113,116,118,120,125,132,136,138,139,139,138,135,131,127,122,120,118,
  120,122,122,122,122,124,125,127,129,133,135,135,129,124,120,116,116,118,121,
  125,129,133,136,138,139,139,140,138,134,128,123,120,119,120,120,118,118,118,
  120,123,126,128,130,134,135,135,130,126,122,121,123,123,124,126,127,128,130,
  133,133,135,136,135,132,130,128,127,129,129,128,127,126,124,123,123,124,124,
  125,124,123,121,120,118,118,119,120,123,127,130,133,137,141,144,146,144,140,
  137,135,132,130,128,126,124,122,121,119,119,119,119,119,120,120,120,122,123,
  123,124,126,127,128,129,132,135,139,142,145,145,142,139,135,133,131,130,128,
  127,125,123,123,122,120,120,120,120,121,120,119,121,123,126,127,128,130,132,
  134,136,137,138,139,139,138,136,133,130,129,129,128,128,128,128,128,128,128,
  127,129,131,129,126,121,116,114,115,118,121,123,126,129,132,133,133,134,138,
  141,141,138,134,131,130,131,131,130,129,129,130,129,128,127,128,130,130,127,
  121,116,116,118,121,123,125,128,131,134,135,133,132,135,137,138,135,129,126,
  125,126,127,127,128,130,134,135,134,132,133,135,135,132,127,121,117,118,119,
  120,121,123,127,129,130,129,130,133,136,137,136,132,129,129,131,131,130,128,
  128,131,134,133,131,130,130,130,128,124,120,119,122,124,124,123,124,128,132,
  135,134,134,135,137,136,133,128,125,124,126,127,126,125,127,130,134,135,134,
  135,136,136,135,132,128,126,126,126,125,124,122,123,125,127,127,128,129,129,
  128,126,125,125,128,131,132,132,132,133,136,140,140,138,136,134,132,129,127,
  125,125,126,127,126,125,124,126,129,132,132,132,132,130,128,127,125,125,127,
  131,132,130,128,129,131,134,135,133,133,134,133,130,127,125,126,128,131,131,
  129,127,128,131,132,131,128,128,127,126,124,124,124,127,132,135,136,136,135,
  137,139,139,138,136,135,131,128,123,120,119,121,125,127,127,128,130,133,135,
  136,135,134,133,129,126,124,123,125,128,130,130,130,131,135,138,140,139,137,
  136,133,130,128,126,125,127,128,128,127,126,128,131,134,134,132,131,130,127,
  126,124,124,126,128,129,129,128,127,130,134,137,136,133,133,133,131,128,128,
  127,128,130,131,130,129,129,131,132,131,129,128,127,125,124,123,124,126,129,
  131,132,132,132,135,138,139,137,133,129,127,125,123,123,123,125,127,128,128,
  128,129,130,132,132,131,131,131,129,127,127,128,131,135,137,137,135,134,135,
  136,137,135,132,128,123,119,117,119,122,125,128,128,128,129,132,136,138,138,
  134,130,127,126,126,126,126,127,129,132,136,137,137,138,139,139,139,135,131,
  128,125,123,121,121,122,124,127,128,128,128,130,133,135,134,131,128,125,124,
  124,126,127,128,130,132,134,135,137,140,140,138,136,132,129,127,126,123,121,
  121,124,127,130,132,133,135,135,134,132,131,129,127,125,125,124,125,127,131,
  134,134,133,132,135,138,138,135,132,129,127,127,128,127,126,126,128,131,134,
  135,136,137,137,133,129,125,123,121,119,116,114,115,120,126,129,130,131,134,
  137,139,141,141,139,135,132,131,131,131,132,132,130,128,128,132,136,136,130,
  125,121,121,124,126,125,121,118,121,127,133,138,139,138,137,134,133,133,133,
  131,127,123,119,118,122,126,128,126,124,124,128,134,137,136,131,127,124,126,
  128,129,128,126,125,125,127,132,137,138,135,129,126,125,128,130,129,126,120,
  117,120,127,132,133,132,129,129,130,133,135,132,128,122,118,119,121,124,125,
  124,121,121,125,130,133,133,129,126,126,129,131,131,130,128,127,128,130,134,
  135,134,131,128,126,126,125,125,124,121,117,115,117,121,125,127,127,126,128,
  130,133,135,134,131,129,127,126,125,124,123,122,122,123,124,125,126,126,126,
  125,125,126,128,128,128,126,125,127,129,132,133,133,132,131,131,131,130,128,
  128,126,123,121,120,119,120,122,123,123,123,125,126,128,128,128,128,129,130,
  129,128,126,124,125,127,126,125,124,123,122,124,124,124,126,128,129,129,129,
  129,130,133,134,133,131,130,130,129,129,128,127,128,128,127,125,122,121,122,
  125,127,127,124,122,121,121,122,122,123,125,127,127,126,125,126,128,128,127,
  124,122,122,123,124,124,123,125,128,132,134,133,133,134,135,134,133,131,130,
  128,126,125,123,123,125,127,128,127,125,124,126,128,127,124,121,118,117,117,
  118,119,121,124,127,128,128,128,128,128,126,123,122,122,122,121,120,119,120,
  124,130,135,136,136,136,137,137,136,133,131,128,125,122,120,119,120,124,126,
  126,124,123,124,126,126,124,120,118,118,118,118,118,119,123,128,132,132,130,
  130,131,132,132,129,127,125,122,120,119,119,121,126,130,132,130,129,130,133,
  134,132,128,125,124,123,123,122,122,124,128,129,128,127,127,128,129,127,123,
  120,119,119,119,118,117,119,123,127,128,127,126,126,127,128,127,126,125,125,
  127,128,129,130,133,136,137,135,133,132,132,133,131,128,124,123,123,124,124,
  123,122,123,125,125,124,122,120,120,118,117,117,119,123,126,128,128,128,130,
  133,136,136,134,132,130,128,127,124,123,125,127,128,129,130,131,133,135,136,
  134,131,128,126,124,123,122,122,123,123,123,122,121,122,124,126,125,122,121,
  121,123,124,125,125,125,126,125,124,123,125,128,131,131,130,129,130,133,135,
  135,134,133,133,134,133,131,130,129,129,128,127,124,123,122,122,121,120,119,
  119,120,120,118,117,118,121,125,127,127,127,128,129,129,129,129,129,129,128,
  128,127,127,128,132,135,137,136,135,135,134,132,131,129,128,127,126,123,121,
  121,123,125,126,125,124,123,123,123,122,121,120,119,119,118,117,118,120,124,
  127,129,129,130,132,134,135,136,135,135,136,136,134,132,132,131,131,130,130,
  128,128,127,126,125,124,124,124,124,123,122,121,122,123,124,123,121,119,119,
  120,122,122,122,123,125,127,127,127,127,128,130,132,133,134,134,135,137,138,
  138,138,138,138,138,136,133,131,129,128,128,126,123,121,119,119,119,119,119,
  118,118,117,116,115,117,119,119,119,120,122,124,126,127,128,130,132,135,137,
  137,138,140,141,142,141,139,135,133,131,130,128,127,126,127,127,127,126,126,
  127,128,128,126,124,121,119,117,115,113,112,113,114,116,118,121,123,126,129,
  131,131,131,131,131,132,132,132,133,135,136,137,137,138,139,140,140,138,136,
  134,133,132,130,128,126,125,124,123,121,119,118,118,117,116,114,113,112,113,
  114,114,115,117,120,124,126,127,130,135,139,142,143,142,142,142,143,141,139,
  137,136,136,135,133,131,132,132,132,131,130,128,125,124,124,121,118,115,113,
  112,112,112,113,116,118,120,121,121,121,123,126,129,131,132,133,134,136,138,
  139,141,142,143,143,141,138,135,134,134,134,134,133,133,132,130,130,129,128,
  127,124,121,117,114,111,110,109,109,110,112,114,117,119,121,123,126,129,132,
  134,134,134,135,137,139,141,141,141,139,138,137,137,138,137,136,136,136,134,
  132,130,128,126,124,122,119,116,111,108,107,107,109,110,111,113,114,117,121,
  125,128,131,133,135,135,137,138,139,141,142,143,143,142,140,137,136,137,137,
  137,137,135,133,131,131,129,128,126,124,121,118,116,113,111,109,109,110,113,
  115,118,119,121,123,126,129,133,134,134,136,138,139,141,141,141,140,139,140,
  139,138,137,135,135,136,138,138,136,133,132,131,130,129,126,122,116,112,111,
  110,109,108,108,109,113,118,121,123,126,129,133,138,140,140,139,140,141,141,
  140,138,136,135,135,137,138,137,136,137,139,141,140,138,135,129,126,123,118,
  116,114,110,107,110,114,117,118,120,119,122,128,130,130,129,129,132,134,137,
  137,134,133,136,140,138,138,136,134,137,141,141,139,137,137,135,136,136,133,
  126,122,122,119,117,117,113,111,110,114,119,120,120,123,125,127,131,135,134,
  135,136,136,137,138,136,131,130,131,133,133,132,133,131,133,138,141,139,136,
  135,133,131,129,124,119,115,112,113,114,116,115,116,118,123,128,131,132,130,
  130,135,136,133,132,131,130,130,131,131,128,126,127,131,134,136,135,135,135,
  138,139,139,135,128,125,122,120,117,113,110,110,114,118,121,123,124,123,127,
  133,138,137,133,130,131,135,135,132,128,129,130,133,134,133,132,132,136,141,
  143,140,137,134,133,132,128,121,114,109,108,109,110,110,109,111,117,124,130,
  132,132,133,137,142,145,142,137,134,133,133,131,129,126,125,128,134,137,136,
  135,136,141,146,147,143,135,129,124,120,116,109,103,101,103,107,111,113,115,
  120,127,133,137,137,136,136,138,141,139,135,131,128,129,131,131,130,129,132,
  137,143,145,144,142,141,141,139,134,126,117,110,108,108,106,105,106,109,115,
  121,125,126,127,129,133,137,136,132,130,131,134,136,135,132,129,131,136,140,
  139,138,138,140,143,144,141,134,128,124,121,118,112,107,105,107,112,116,119,
  120,122,126,130,133,132,130,129,129,129,128,126,125,125,128,131,131,132,133,
  138,144,148,150,149,146,143,140,136,130,122,114,110,109,109,110,111,114,120,
  127,132,134,134,133,133,132,129,125,120,118,118,120,121,122,122,127,135,144,
  149,151,151,151,152,151,147,139,129,121,115,111,108,108,108,112,120,127,133,
  136,137,137,138,137,132,124,116,112,111,112,111,111,113,119,128,137,142,145,
  151,157,161,161,154,146,137,129,122,117,112,109,110,115,123,130,136,141,144,
  147,146,141,132,123,115,110,107,104,100,99,103,113,121,129,134,140,148,156,
  159,157,151,143,137,132,127,122,118,118,122,128,137,144,148,150,150,149,146,
  139,128,118,111,107,104,101,99,101,107,115,124,130,134,139,145,148,147,140,
  131,125,121,119,117,116,118,125,136,150,159,163,164,163,161,156,146,133,121,
  112,107,104,102,101,104,112,122,131,137,139,139,140,141,137,128,118,107,100,
  100,102,107,114,124,138,154,168,176,176,172,167,160,150,138,124,111,104,103,
  105,107,111,117,128,139,147,151,149,145,140,133,125,113,99,89,89,95,106,116,
  126,137,150,162,168,167,160,151,142,136,130,121,112,106,110,119,127,132,136,
  141,149,156,157,151,140,128,119,113,107,101,99,102,112,125,136,144,147,149,
  149,148,144,134,122,112,107,107,110,112,115,123,135,148,158,162,162,159,154,
  148,138,125,111,101,98,102,110,117,126,135,146,156,160,158,151,139,128,117,
  107,98,91,88,93,103,116,128,138,147,158,168,173,169,158,144,132,122,113,105,
  100,101,108,120,135,148,155,159,159,157,152,141,128,112,100,93,91,92,94,99,
  109,123,138,150,156,158,158,157,154,147,137,127,120,117,120,124,129,134,140,
  147,151,152,148,141,132,123,116,109,102,96,93,95,103,112,121,128,134,138,142,
  145,145,142,136,130,128,129,133,138,141,146,152,158,162,161,155,145,134,125,
  117,110,102,95,92,96,102,109,116,121,126,131,137,139,136,129,123,118,116,119,
  121,123,129,139,151,164,173,176,173,168,163,156,146,133,119,108,101,98,97,97,
  99,103,111,121,128,133,132,130,128,124,120,117,114,111,110,114,122,132,142,
  151,158,164,169,170,167,160,148,137,129,123,117,111,104,103,106,113,121,126,
  129,132,135,138,137,133,127,119,113,110,111,114,118,122,126,130,137,144,150,
  149,144,139,136,134,132,128,120,113,110,112,118,124,128,133,139,147,152,153,
  149,142,135,129,126,125,124,123,122,121,123,126,129,131,130,128,128,127,128,
  125,119,113,108,106,109,113,118,121,124,130,139,148,153,152,147,143,142,143,
  146,145,140,132,127,127,130,131,129,126,124,124,127,129,127,119,111,106,106,
  110,115,117,117,118,121,127,133,137,137,134,132,135,141,146,147,142,135,132,
  134,137,138,136,133,130,129,130,130,127,121,117,116,117,121,127,128,127,123,
  120,120,123,127,127,124,123,126,131,136,138,135,130,128,128,132,135,136,133,
  130,128,128,129,128,127,126,126,128,133,137,139,135,129,124,122,123,126,127,
  125,121,120,123,127,128,127,123,121,123,127,131,132,128,123,121,121,124,128,
  130,132,133,135,136,139,141,140,136,130,126,125,128,131,131,126,119,116,118,
  123,126,125,123,120,120,123,126,126,125,123,122,125,130,136,140,139,135,130,
  128,129,132,134,132,129,129,131,134,136,134,129,125,122,122,124,125,124,120,
  117,117,119,123,128,130,130,129,130,132,136,137,134,128,122,120,124,130,136,
  135,130,125,125,130,134,134,128,123,121,122,125,126,124,121,120,122,127,133,
  137,138,136,132,130,130,133,135,133,128,123,124,129,136,138,136,129,123,120,
  120,120,118,113,107,106,110,117,125,129,131,131,133,137,143,148,148,143,135,
  129,129,133,138,138,133,127,125,128,134,137,132,124,117,115,115,116,115,112,
  109,108,111,117,124,130,135,137,136,133,133,137,141,141,135,128,122,122,128,
  133,134,132,129,128,131,135,137,133,125,116,111,111,115,117,117,114,113,116,
  122,133,143,146,143,138,135,136,138,139,136,129,123,119,121,126,131,132,129,
  125,122,121,122,122,119,112,104,102,104,112,122,128,128,128,129,135,144,152,
  154,149,141,134,131,133,135,132,126,120,118,122,128,133,134,128,121,116,114,
  114,115,114,109,104,102,105,111,121,131,136,137,137,138,141,146,149,147,140,
  131,125,124,127,130,128,123,118,119,124,129,132,130,124,118,116,117,120,122,
  122,119,115,114,117,124,134,140,141,137,133,131,134,138,140,137,130,124,122,
  124,127,128,125,119,116,117,121,125,128,126,123,120,120,123,127,131,134,135,
  133,131,130,133,139,142,138,130,123,120,123,128,129,124,115,110,111,117,123,
  126,124,120,118,119,124,129,134,136,134,131,129,131,136,142,146,143,136,130,
  128,128,130,130,127,121,116,116,118,122,123,120,114,108,108,111,117,122,124,
  124,124,126,130,136,144,148,148,143,137,134,135,138,140,137,130,123,121,125,
  131,135,132,124,117,114,115,117,119,116,109,104,105,110,118,127,133,136,135,
  134,135,139,145,148,145,137,127,122,122,127,131,130,125,120,121,127,134,139,
  136,127,118,114,114,117,121,122,119,116,115,118,126,137,144,145,140,134,130,
  133,137,137,132,123,114,111,114,121,127,127,122,118,118,122,130,137,138,133,
  125,120,121,128,138,143,140,131,124,124,131,141,144,138,126,116,114,119,125,
  125,118,109,104,106,113,122,127,127,123,122,125,132,141,149,151,146,138,133,
  135,143,150,150,141,127,117,115,120,128,128,120,108,101,103,111,120,121,114,
  105,103,109,119,130,136,136,132,130,134,144,157,165,163,153,141,136,138,144,
  147,141,128,115,108,110,116,121,118,107,98,95,101,111,118,117,111,105,107,
  119,135,148,151,145,139,138,145,156,164,162,149,133,123,123,130,135,132,122,
  111,108,112,120,126,124,117,110,108,112,121,130,131,126,120,120,128,142,154,
  156,147,134,127,129,136,141,137,126,113,107,109,117,124,126,121,115,114,121,
  131,139,138,130,122,121,130,143,152,151,142,135,135,142,151,153,145,131,120,
  115,117,118,116,108,99,96,99,107,115,120,121,120,120,124,133,143,149,147,142,
  138,141,149,157,158,153,144,137,136,141,144,139,128,115,107,106,109,110,104,
  97,94,97,107,118,126,126,123,123,129,140,151,155,150,141,135,138,147,154,153,
  145,136,132,133,137,137,132,123,115,111,111,114,116,114,108,103,106,116,130,
  138,137,131,128,131,140,149,151,144,132,126,128,135,141,141,133,125,121,124,
  130,133,129,122,117,117,122,129,134,132,125,121,123,131,140,144,139,130,124,
  124,128,135,137,131,123,119,122,128,132,130,125,120,121,125,131,134,130,123,
  120,124,132,140,142,136,128,126,130,137,141,138,128,120,120,126,132,133,128,
  122,120,125,131,137,136,131,125,122,125,133,140,138,129,120,119,127,138,142,
  137,127,119,118,124,131,133,126,116,111,117,129,138,139,134,130,130,136,145,
  151,148,137,128,125,129,135,137,131,121,114,115,122,128,129,124,116,111,115,
  123,128,126,119,117,121,130,140,144,142,139,139,145,152,156,152,144,134,129,
  130,133,131,123,112,105,106,112,118,118,113,109,111,118,127,132,131,129,128,
  131,138,148,154,151,143,137,139,145,150,148,138,128,121,120,123,125,122,113,
  106,106,113,122,126,124,120,121,128,137,145,147,140,133,130,136,143,148,145,
  136,127,125,128,131,130,124,114,107,108,115,121,121,115,110,114,125,138,145,
  144,138,135,138,146,151,150,143,133,127,128,134,138,133,123,114,111,115,121,
  123,119,112,108,111,118,126,128,127,125,126,132,141,146,143,136,130,130,136,
  142,142,134,126,121,122,126,129,128,123,116,116,122,129,131,128,123,122,125,
  132,139,141,137,130,128,131,137,140,136,127,119,118,122,125,125,120,112,105,
  109,120,128,129,125,123,128,137,147,151,147,139,133,132,136,141,140,132,121,
  115,117,125,129,126,118,111,109,115,123,126,122,116,115,121,132,141,144,139,
  133,133,140,147,149,144,135,127,125,127,130,128,121,112,108,110,117,123,124,
  120,115,116,122,130,135,135,129,126,129,136,142,144,139,132,129,131,137,140,
  135,126,117,114,117,123,126,122,115,112,118,126,132,134,131,128,128,134,141,
  145,142,133,124,122,128,135,137,130,120,114,116,123,128,128,122,114,111,117,
  126,130,128,125,124,128,138,147,151,149,142,136,137,142,144,140,130,120,114,
  113,117,124,126,117,105,105,117,127,128,125,118,118,124,130,135,135,131,127,
  127,131,139,144,144,138,131,129,132,137,136,130,121,113,112,120,129,126,115,
  111,115,124,133,138,137,132,128,128,131,135,134,128,122,120,121,125,130,130,
  125,119,117,123,130,132,128,123,119,120,128,134,135,130,125,126,132,139,141,
  137,130,127,126,129,133,132,126,118,112,114,121,128,128,121,114,113,119,126,
  130,129,123,119,120,128,136,139,138,133,129,130,136,142,142,136,128,124,124,
  128,130,129,122,114,112,117,126,130,128,121,116,118,127,134,137,132,125,120,
  123,130,135,134,128,121,118,122,128,134,134,128,122,120,126,133,136,133,126,
  121,123,131,138,140,136,128,124,126,134,140,139,130,119,114,118,127,132,129,
  120,113,111,117,125,128,124,116,110,112,124,135,138,132,125,124,130,140,147,
  147,140,131,128,132,139,142,138,128,118,116,123,131,134,128,118,112,115,123,
  131,131,126,117,113,116,124,130,132,127,120,118,123,130,135,134,128,123,122,
  128,136,141,138,130,124,124,129,135,136,131,122,117,121,129,137,139,135,128,
  125,127,131,134,131,124,116,112,116,125,130,128,122,116,116,124,132,136,135,
  129,125,126,132,138,141,137,128,122,122,129,136,137,130,122,117,120,127,134,
  135,129,122,118,123,131,138,138,131,124,124,129,138,144,143,136,130,128,132,
  137,139,134,124,115,113,117,123,125,122,116,113,116,125,132,134,131,126,124,
  127,135,143,145,141,134,130,133,140,145,144,135,126,120,121,125,128,127,122,
  116,114,119,127,133,132,125,118,118,125,134,139,137,131,125,126,132,141,145,
  140,130,123,121,125,130,133,128,119,113,115,123,130,133,130,122,118,122,131,
  141,145,140,130,126,130,140,148,150,142,131,123,122,127,132,133,126,116,110,
  112,121,129,129,121,114,111,117,128,137,140,134,127,124,128,139,149,151,143,
  130,121,122,128,136,137,129,118,112,115,123,131,133,127,117,112,115,127,138,
  140,132,123,121,127,138,148,150,143,132,124,125,130,135,134,126,114,107,110,
  119,128,129,125,119,117,123,132,141,143,137,128,123,126,135,145,148,141,131,
  125,125,129,134,135,129,122,116,115,121,128,131,129,123,117,118,125,132,135,
  133,128,124,126,132,141,146,144,138,130,127,128,131,133,131,124,117,114,117,
  124,129,130,126,120,116,119,127,135,139,136,131,128,131,139,147,150,145,137,
  129,126,127,130,132,128,121,114,112,117,125,128,126,120,116,115,119,127,133,
  134,131,127,126,131,140,148,149,144,136,130,128,129,132,132,128,122,118,117,
  122,128,132,131,126,122,121,125,131,135,134,129,124,122,124,129,134,135,132,
  126,122,121,123,127,128,127,124,121,121,124,129,133,133,130,126,124,126,131,
  138,142,142,139,136,135,136,139,139,136,129,122,116,114,117,121,123,122,117,
  115,116,122,128,131,129,126,121,120,125,133,142,144,141,136,133,134,138,141,
  139,134,127,120,118,121,126,128,127,122,118,118,123,129,134,135,132,127,124,
  125,130,136,139,136,130,125,123,125,128,130,128,125,121,120,123,129,133,134,
  131,126,122,122,126,131,133,131,127,123,123,127,134,139,139,135,129,124,122,
  123,125,125,122,118,117,120,127,133,137,137,134,130,128,129,133,137,139,137,
  132,128,125,126,127,128,126,121,115,111,111,115,120,122,123,122,121,123,128,
  135,140,141,139,135,133,134,138,144,146,143,137,130,126,125,126,125,120,113,
  106,105,108,114,119,122,121,119,121,126,135,144,147,144,139,135,134,136,140,
  142,139,132,126,121,119,118,118,116,111,108,107,111,118,127,133,136,135,133,
  133,135,138,141,144,144,140,136,133,132,132,131,128,123,116,110,107,108,111,
  115,117,116,115,118,125,133,141,144,143,140,138,138,142,148,151,149,141,131,
  123,119,117,115,112,107,102,100,105,115,125,131,132,131,129,132,138,145,149,
  149,145,139,135,136,140,142,139,130,119,108,105,106,110,111,109,106,105,111,
  122,134,142,144,140,135,134,139,146,153,156,152,143,136,132,131,128,124,115,
  103,95,92,97,107,116,122,123,123,125,132,141,148,150,147,142,138,140,146,152,
  155,151,140,130,122,117,113,108,102,95,90,91,101,115,128,137,138,134,132,135,
  143,150,153,152,147,143,142,143,145,143,135,122,109,100,96,96,98,101,102,104,
  111,123,137,149,153,148,140,135,137,144,152,155,153,148,144,140,137,132,124,
  111,96,84,79,82,90,100,109,115,122,130,140,149,154,152,147,143,144,149,157,
  161,159,153,145,137,128,119,108,95,84,79,81,90,103,116,125,130,134,137,140,
  143,144,143,141,141,145,151,158,161,157,149,138,127,116,104,93,84,79,80,88,
  102,118,130,137,138,139,139,141,144,146,147,149,152,156,160,161,158,149,135,
  120,105,93,85,81,80,84,94,109,125,137,144,144,140,134,131,133,138,142,146,
  150,155,160,163,159,149,133,118,102,89,82,80,84,91,102,117,131,143,148,146,
  140,133,130,133,139,146,152,156,159,161,159,153,142,125,104,86,75,74,80,90,
  101,114,126,138,147,150,147,139,129,124,126,133,145,155,163,166,166,162,153,
  139,122,104,89,79,76,80,92,106,119,131,141,145,144,139,132,127,127,130,139,
  149,159,167,171,169,161,148,131,112,94,81,75,78,87,99,113,127,138,146,147,
  141,132,125,121,122,128,138,150,162,170,173,170,161,147,128,109,93,82,78,82,
  92,107,122,134,142,145,142,134,125,118,117,120,128,140,152,164,172,174,168,
  154,135,115,96,83,77,78,88,102,117,130,142,149,149,142,131,120,114,116,124,
  134,146,158,168,173,171,162,146,126,106,90,80,77,81,93,109,126,139,148,149,
  142,131,121,115,113,118,127,138,151,164,173,176,171,157,137,115,96,83,78,80,
  88,101,117,132,144,151,150,141,128,116,109,111,120,133,145,156,166,172,172,
  165,149,128,107,90,81,78,84,96,113,128,141,148,148,142,131,120,111,108,112,
  122,134,149,163,174,178,173,159,140,120,101,87,81,82,90,104,120,135,148,152,
  148,137,124,113,106,107,114,125,138,152,164,172,174,167,152,131,110,94,85,84,
  89,100,114,129,142,150,150,144,131,118,108,104,107,118,133,148,161,171,175,
  171,160,142,122,102,88,81,83,93,108,125,138,145,147,144,137,126,113,105,102,
  108,122,138,154,168,176,176,167,152,134,115,98,86,81,86,98,115,132,145,151,
  150,142,130,116,105,100,102,111,126,144,160,172,178,175,164,146,126,105,90,
  84,85,94,107,123,137,148,151,148,138,126,113,101,96,101,115,133,151,166,175,
  177,171,159,142,122,103,89,84,87,97,113,130,143,150,150,143,132,120,107,99,
  96,104,120,138,155,170,178,177,168,151,130,111,95,86,84,91,103,119,135,146,
  152,151,143,130,116,104,97,99,110,127,145,161,173,178,175,164,147,128,108,92,
  84,86,95,110,126,139,147,149,145,135,124,111,101,97,100,113,131,150,165,175,
  177,171,158,140,120,101,89,84,88,98,114,131,144,150,150,143,133,122,110,100,
  98,104,119,137,156,171,179,179,171,154,134,115,99,88,85,89,101,117,132,143,
  148,145,138,127,115,104,98,99,107,122,140,158,172,179,178,169,153,134,115,99,
  90,89,97,109,122,135,145,148,145,136,125,112,102,96,97,105,121,141,160,173,
  178,176,166,151,132,113,99,90,90,97,108,123,136,146,148,143,133,123,112,104,
  99,100,108,124,142,159,172,178,175,165,148,129,113,100,94,94,100,110,122,134,
  143,146,143,135,124,113,103,98,100,108,122,140,158,172,179,178,168,151,132,
  114,100,92,92,98,107,118,129,139,143,139,131,120,111,104,102,103,111,124,140,
  158,172,180,180,172,156,137,119,105,96,96,99,106,116,126,135,139,138,131,121,
  111,102,97,98,107,121,137,154,169,178,180,174,161,142,123,108,99,97,100,107,
  117,127,135,140,138,132,123,112,104,98,99,105,118,134,152,167,177,179,174,
  162,144,126,110,101,97,100,106,115,125,135,141,141,135,125,113,103,97,96,101,
  114,130,148,162,173,177,175,164,146,127,111,101,97,99,105,115,126,135,141,
  142,137,128,119,109,101,99,103,114,129,147,163,175,180,177,164,145,126,110,
  99,94,95,101,111,122,133,141,143,138,128,117,107,101,100,104,113,127,144,162,
  175,182,180,169,151,132,116,104,97,97,103,110,120,130,139,141,138,128,117,
  107,99,96,98,108,124,142,160,174,181,179,170,153,133,115,103,97,97,102,110,
  120,130,138,142,140,132,123,112,104,99,99,107,121,137,154,169,178,179,171,
  156,137,120,107,99,97,101,109,118,127,134,138,136,131,122,112,103,97,97,103,
  117,134,152,168,178,181,174,161,142,124,109,100,98,102,108,116,126,135,141,
  142,137,128,117,107,101,99,105,116,131,147,161,172,177,174,163,145,125,108,
  98,95,98,103,111,121,130,137,140,139,132,122,111,102,98,101,113,128,145,162,
  174,181,179,169,151,131,115,104,99,100,104,112,121,129,136,139,139,133,123,
  112,103,98,99,106,120,137,154,170,179,179,170,155,136,118,105,98,98,103,111,
  120,129,137,141,141,136,127,116,106,100,99,105,118,135,153,168,176,178,172,
  158,139,120,105,97,96,100,105,115,127,136,141,142,138,130,121,110,101,97,100,
  112,128,146,162,174,178,174,161,142,123,108,98,95,98,105,115,126,135,142,144,
  142,136,126,115,105,98,98,108,124,141,159,172,178,175,164,147,128,111,99,94,
  95,101,110,121,131,140,144,144,139,130,119,109,102,100,107,119,136,153,167,
  176,178,171,155,134,114,101,95,96,100,107,117,128,138,144,145,141,134,124,
  112,102,98,102,114,132,150,164,173,176,171,158,140,122,106,97,94,97,104,115,
  127,137,143,145,143,137,128,117,106,98,97,106,122,140,158,170,176,174,164,
  147,127,109,96,91,93,100,110,122,133,143,148,149,144,135,125,113,102,97,103,
  117,135,153,167,175,175,168,155,135,115,99,89,86,89,96,108,122,135,144,148,
  145,139,130,120,108,99,98,108,125,145,162,174,180,177,166,147,125,107,95,89,
  89,94,104,118,132,143,149,149,146,139,128,116,105,100,105,120,138,157,171,
  180,181,172,156,135,115,97,85,80,82,92,108,125,140,151,155,153,146,135,122,
  109,100,100,108,124,144,162,174,179,175,163,145,124,106,93,87,88,95,106,121,
  136,149,156,155,148,136,124,111,100,96,102,116,135,154,167,174,173,164,149,
  128,108,92,83,82,87,98,113,131,148,157,158,152,143,131,119,106,99,101,114,
  134,154,167,173,172,165,152,135,117,100,89,85,87,95,109,125,139,150,154,152,
  144,133,120,106,96,96,107,125,144,161,171,172,167,156,139,120,104,93,87,87,
  93,105,120,134,147,154,154,150,142,132,120,109,103,107,120,137,154,165,168,
  164,156,145,131,116,104,96,92,94,102,113,126,138,147,150,149,144,136,126,114,
  104,102,108,125,144,158,164,163,157,149,137,124,111,100,92,90,95,105,120,134,
  146,153,155,154,149,140,126,112,104,105,115,130,145,155,159,157,152,143,129,
  117,107,98,92,91,96,109,124,138,149,154,154,151,144,133,121,111,108,114,126,
  140,152,158,158,153,145,134,122,109,99,90,86,89,100,115,129,143,150,154,153,
  149,140,128,117,111,114,124,137,149,155,155,151,143,133,123,112,105,98,94,93,
  99,109,124,137,147,151,152,149,144,136,126,119,119,125,135,145,152,151,145,
  137,128,118,110,105,103,102,102,105,112,123,132,139,142,142,141,141,139,135,
  132,132,137,145,152,154,148,139,128,118,110,105,103,105,108,109,110,113,118,
  125,131,136,138,138,139,138,136,134,134,139,145,151,152,147,137,126,116,108,
  102,100,103,110,116,120,123,125,128,131,132,132,132,133,134,135,136,137,142,
  151,159,162,157,145,130,117,105,97,92,93,99,108,115,120,125,129,133,136,137,
  136,135,134,134,133,134,139,147,155,160,160,154,141,126,111,100,92,89,93,99,
  107,116,122,128,133,137,139,140,138,136,134,134,135,139,146,154,160,162,156,
  144,128,113,100,90,85,85,91,101,111,119,126,132,137,141,143,142,139,136,134,
  135,138,144,152,158,162,160,151,137,121,106,94,87,85,88,95,105,113,120,126,
  131,135,138,139,138,137,135,135,138,145,153,162,168,167,156,139,120,103,92,
  86,85,87,94,103,112,119,125,130,134,136,137,137,136,135,135,137,141,147,155,
  162,166,163,152,135,118,101,90,84,84,89,97,105,112,119,125,130,134,137,138,
  138,138,138,139,143,149,157,164,169,168,158,141,122,103,89,81,80,84,91,99,
  106,113,120,127,132,136,138,139,139,139,140,143,149,156,164,170,171,166,152,
  133,111,92,81,77,79,86,94,102,110,117,123,129,134,137,139,139,138,139,141,
  146,153,163,172,177,175,163,144,121,100,85,77,76,80,87,96,105,113,120,126,
  130,133,135,136,137,138,140,143,149,157,166,174,177,173,159,138,115,95,83,79,
  80,84,90,98,105,111,118,123,127,130,132,134,136,138,142,148,156,165,173,177,
  176,167,151,129,109,92,82,80,83,88,94,99,104,111,118,124,128,129,131,133,136,
  141,147,155,163,172,179,181,175,160,140,119,99,85,77,76,80,87,94,100,106,114,
  121,128,132,135,137,139,142,146,151,158,167,175,179,177,166,149,128,108,91,
  81,77,79,84,90,96,103,109,116,122,127,130,132,136,139,143,148,155,163,172,
  181,184,177,163,144,123,104,90,82,80,82,86,90,95,101,108,114,120,125,127,129,
  133,138,144,151,159,168,177,184,184,176,159,137,115,98,88,84,84,86,89,92,96,
  101,107,114,119,123,126,129,134,141,149,158,167,176,183,185,180,167,149,127,
  106,91,84,82,83,86,89,93,97,102,109,116,122,125,128,131,137,146,155,165,173,
  181,186,186,177,161,139,117,98,86,80,81,84,87,91,96,100,106,112,119,124,127,
  129,133,140,150,162,173,181,187,188,183,170,149,125,103,88,81,81,84,87,91,95,
  99,103,108,113,117,121,124,128,135,145,158,170,181,188,191,189,180,163,139,
  115,95,83,79,81,84,88,93,97,101,105,109,114,118,122,125,128,136,149,164,178,
  189,194,193,186,171,149,125,104,89,83,83,86,89,93,97,101,105,107,109,112,115,
  119,124,131,143,158,173,184,190,191,186,176,159,136,114,97,87,85,87,90,92,94,
  97,100,103,106,109,113,116,119,124,131,145,161,178,191,197,195,187,172,152,
  129,109,94,89,90,92,93,94,94,96,97,99,101,104,109,115,121,129,139,151,166,
  180,190,194,191,182,168,150,131,115,102,95,92,91,91,90,89,90,92,95,99,105,
  111,116,123,132,144,158,172,186,195,198,195,184,168,148,129,113,101,93,88,85,
  81,79,80,83,88,93,100,107,114,121,128,138,150,164,177,190,198,200,195,184,
  167,148,127,109,96,86,79,74,70,70,75,83,92,101,110,117,123,128,134,141,151,
  165,180,194,201,201,195,183,166,145,124,106,93,83,76,70,66,66,72,80,90,100,
  109,118,125,130,136,144,154,167,181,193,200,200,193,181,164,144,123,105,91,
  82,76,72,69,70,75,84,93,103,112,119,126,133,141,150,161,173,185,195,199,196,
  184,171,154,135,118,101,88,78,72,69,68,71,76,84,93,104,114,122,129,137,146,
  157,168,180,192,199,200,193,181,165,146,127,110,96,86,79,74,70,70,75,83,90,
  98,105,112,118,125,132,141,155,170,185,197,203,201,192,178,158,136,114,98,87,
  81,78,76,75,78,83,90,96,103,110,117,124,130,138,148,160,175,190,201,205,200,
  187,169,147,124,104,89,81,79,80,83,85,89,92,95,98,100,104,109,115,123,134,
  148,164,181,196,207,209,202,187,167,143,120,99,85,79,80,82,85,87,90,92,93,95,
  97,101,108,116,127,140,155,172,188,202,211,210,199,182,160,136,114,96,84,78,
  77,79,82,85,87,89,92,95,99,103,110,120,132,147,163,179,194,206,211,209,198,
  180,159,136,114,95,82,75,74,75,77,79,82,85,89,93,99,107,115,124,135,148,162,
  177,191,203,210,208,197,180,158,134,113,95,84,78,76,77,79,82,84,87,92,96,102,
  108,115,124,136,150,166,181,194,204,207,204,193,177,156,133,113,96,84,76,74,
  74,77,79,81,84,88,94,101,108,117,127,139,153,169,184,197,207,209,203,191,174,
  154,133,113,97,86,79,76,75,76,78,80,83,89,95,102,110,119,129,140,154,170,185,
  198,206,208,201,188,171,151,130,112,97,86,78,74,73,74,76,78,82,87,93,101,111,
  122,133,145,159,173,188,202,210,210,203,189,170,149,128,110,96,85,79,76,76,
  77,78,78,81,85,91,98,108,120,132,145,158,172,186,198,206,207,201,188,171,151,
  132,114,100,89,81,77,76,77,77,77,79,83,90,98,109,120,133,147,162,176,189,200,
  205,205,197,184,167,148,130,114,101,91,83,77,75,75,76,77,79,83,90,100,111,
  123,135,148,161,175,189,201,207,206,198,183,164,145,128,111,98,88,82,78,77,
  78,79,80,82,87,94,103,112,124,135,147,158,170,182,193,201,202,196,184,167,
  148,129,113,100,92,85,81,80,79,79,80,82,86,93,102,112,123,134,146,158,171,
  185,197,204,204,197,184,167,149,131,114,100,89,80,76,75,76,78,80,84,90,98,
  106,115,124,135,146,157,168,178,190,199,203,200,189,172,152,132,114,99,89,81,
  77,75,77,79,83,87,92,98,107,117,127,136,146,156,167,180,192,200,202,196,185,
  169,149,129,109,92,82,75,72,73,76,80,86,92,99,105,112,120,128,136,146,155,
  166,177,190,199,201,196,186,170,151,131,112,97,86,79,75,74,74,77,82,88,95,
  103,111,119,128,136,146,155,165,177,187,195,198,194,185,170,153,134,114,97,
  86,79,76,76,77,79,82,87,95,104,113,122,130,138,146,154,163,174,186,194,197,
  194,185,172,156,138,120,102,89,81,76,73,73,74,77,84,92,103,113,122,129,136,
  143,150,158,168,181,192,197,195,186,173,158,142,125,108,94,85,79,76,75,74,76,
  83,92,103,114,125,134,141,148,153,158,165,176,187,195,197,189,176,160,142,
  124,105,91,81,77,76,76,76,78,81,88,97,108,119,129,137,145,153,159,167,177,
  188,197,199,194,181,166,150,133,116,99,87,79,76,76,76,76,78,84,92,103,115,
  126,134,142,149,154,161,170,181,192,197,194,183,168,152,135,119,104,91,83,79,
  78,77,77,79,84,91,102,113,124,132,140,147,153,158,165,176,189,200,201,191,
  176,159,142,125,108,93,82,78,77,77,77,76,79,85,95,108,121,130,138,144,151,
  157,163,171,183,195,200,195,183,166,148,129,112,97,86,80,78,79,81,82,83,86,
  91,102,114,125,134,141,147,153,158,166,176,188,196,197,191,177,159,141,122,
  105,91,82,78,76,76,76,77,81,88,98,110,122,133,141,147,152,155,161,170,183,
  194,200,197,186,168,148,128,109,94,84,77,75,76,77,79,82,86,93,105,118,128,
  136,141,146,150,154,162,173,185,195,199,193,179,161,141,121,103,88,78,75,75,
  75,77,79,84,90,99,110,123,133,141,147,151,154,159,167,179,191,199,196,186,
  169,149,129,110,93,80,74,73,74,77,80,83,88,95,106,119,130,138,144,149,152,
  154,158,167,180,191,195,191,179,163,143,123,105,91,81,75,73,73,76,81,85,91,
  100,112,125,135,142,147,149,151,153,160,171,184,194,196,190,176,156,135,115,
  99,87,79,76,74,74,75,79,84,92,104,117,130,140,146,149,149,149,152,158,169,
  183,193,197,190,175,155,133,113,96,85,78,75,76,77,80,83,88,95,106,119,132,
  142,147,150,150,150,152,157,168,180,189,191,186,172,152,129,108,91,80,75,74,
  75,77,81,86,92,101,113,126,137,146,152,153,153,153,155,160,169,179,186,188,
  181,167,146,124,104,89,79,74,72,73,77,83,88,95,105,117,129,140,147,151,151,
  151,152,156,162,171,180,187,189,181,164,141,120,101,88,80,77,76,78,81,84,89,
  96,108,121,133,143,147,148,148,147,149,153,160,170,181,188,187,177,160,139,
  119,102,90,84,81,80,80,80,80,81,88,101,115,129,140,147,151,152,152,152,154,
  161,170,180,188,189,182,165,143,122,105,93,86,83,83,83,82,78,77,83,94,108,
  123,136,145,150,151,151,150,153,159,169,180,191,194,187,171,151,129,110,94,
  83,79,78,77,76,75,76,83,94,108,124,138,148,154,155,154,152,152,157,167,179,
  188,191,185,171,151,128,107,90,79,73,72,74,75,75,78,86,98,113,128,142,152,
  158,159,156,152,151,154,161,172,184,191,187,174,154,130,108,89,77,72,73,76,
  77,78,80,87,98,113,128,141,150,154,155,153,150,147,148,155,167,179,186,185,
  176,160,140,118,99,84,76,74,77,81,83,85,87,94,106,120,133,144,152,155,154,
  149,144,144,150,162,175,185,188,181,165,145,123,103,87,78,76,79,80,80,81,86,
  94,106,121,135,146,153,155,154,151,146,144,146,155,167,178,182,179,169,152,
  131,109,92,80,74,75,79,84,88,91,95,104,117,130,143,153,158,159,155,149,145,
  146,153,163,173,180,177,165,146,126,107,92,81,76,76,78,83,88,93,101,111,123,
  136,146,152,154,154,152,149,145,144,147,156,167,176,177,171,157,139,120,103,
  91,83,79,78,78,78,80,85,95,110,125,139,148,153,154,153,149,146,145,149,157,
  167,176,181,176,163,147,130,116,102,90,83,80,81,82,85,88,93,104,118,132,143,
  150,153,153,150,144,140,139,146,160,175,184,184,173,156,137,118,102,89,81,78,
  77,79,82,84,91,103,117,133,146,156,160,160,156,149,140,136,138,147,161,174,
  178,172,159,142,125,107,91,80,75,76,81,87,92,97,103,112,125,138,149,156,158,
  156,151,144,138,137,145,158,172,181,180,168,150,132,114,95,81,73,70,73,81,89,
  96,104,113,125,137,148,155,157,155,150,143,137,135,139,149,163,175,178,173,
  161,145,127,108,91,79,73,73,77,83,89,97,105,117,130,142,151,154,154,149,141,
  135,133,136,146,159,174,182,180,169,154,136,117,99,84,75,71,74,81,89,97,104,
  113,125,137,149,155,156,152,145,137,130,128,134,148,164,175,178,172,160,143,
  124,105,88,76,70,70,76,85,95,104,112,123,135,147,156,160,158,150,140,132,129,
  132,140,154,168,175,174,166,153,135,116,95,78,70,67,70,78,88,99,110,120,132,
  145,156,161,160,153,142,133,128,128,134,146,161,173,177,173,162,147,128,107,
  89,76,68,67,72,81,93,105,117,129,142,154,161,161,156,146,136,128,124,126,135,
  150,166,175,176,170,158,141,121,100,83,72,67,67,73,85,99,112,125,137,149,158,
  163,162,153,141,129,121,119,125,138,155,169,176,174,166,153,135,115,96,80,72,
  69,72,80,91,103,116,129,142,153,161,162,157,147,135,124,117,118,126,140,157,
  172,179,177,166,150,131,111,92,78,71,69,73,82,94,107,120,133,146,157,163,162,
  155,143,130,120,114,116,125,141,159,172,176,173,163,148,129,110,92,78,70,68,
  73,84,96,108,121,133,145,154,159,158,151,140,128,119,115,118,128,146,163,174,
  178,174,162,146,127,107,90,79,74,74,77,85,96,109,124,137,148,156,160,157,148,
  135,124,116,114,118,128,143,160,172,176,172,162,146,127,107,92,82,77,77,81,
  89,99,111,124,136,146,154,157,154,146,135,124,116,112,116,128,144,160,171,
  176,172,161,146,128,110,95,84,78,78,82,88,98,111,123,135,145,152,155,153,146,
  134,124,116,114,118,127,140,155,168,175,174,166,151,133,116,102,91,84,82,85,
  90,99,109,120,131,141,149,152,151,145,136,127,119,115,115,122,133,147,161,
  169,170,164,152,137,122,108,97,89,86,86,91,98,107,116,125,133,141,146,147,
  144,137,129,123,118,116,119,128,142,156,167,171,167,157,143,130,118,107,100,
  95,93,95,99,105,112,120,128,134,138,140,139,135,129,124,118,115,116,122,134,
  149,161,167,166,159,149,138,127,117,108,101,98,99,102,105,110,116,123,128,
  132,134,136,135,131,127,121,118,118,122,130,143,155,164,168,165,157,147,137,
  127,117,110,105,102,102,104,107,112,117,122,126,128,129,130,128,125,121,117,
  116,118,124,134,146,157,164,167,164,158,150,140,131,123,114,109,106,106,106,
  108,111,115,119,122,124,125,126,125,122,120,119,120,123,128,136,146,156,163,
  164,162,157,150,142,133,125,118,113,111,110,110,110,111,113,115,117,118,118,
  119,119,120,119,120,122,126,131,138,147,156,162,165,163,158,152,145,136,129,
  123,118,115,112,109,108,108,110,111,111,111,111,112,114,117,120,124,127,131,
  134,138,145,153,160,163,163,159,154,147,140,133,127,122,119,116,113,110,107,
  105,104,103,104,104,106,110,114,119,124,128,133,136,141,147,154,160,164,164,
  161,156,151,145,138,131,125,120,117,114,110,106,102,99,97,98,100,103,107,111,
  116,122,128,133,137,142,147,153,159,163,165,163,161,157,152,146,139,132,125,
  120,116,111,106,101,96,93,94,96,99,102,106,111,117,123,129,137,143,147,151,
  156,160,163,164,162,159,156,151,145,138,129,122,115,109,105,101,97,95,94,95,
  97,100,103,107,112,120,128,135,141,146,149,152,154,159,163,163,162,159,156,
  152,146,138,128,119,110,103,99,97,96,95,95,96,98,100,103,108,115,124,132,139,
  145,149,152,154,155,157,160,161,161,158,155,151,145,136,126,115,107,102,99,
  98,97,97,97,98,99,101,106,113,119,126,132,138,143,148,152,154,156,156,156,
  156,156,153,149,144,140,135,128,120,111,105,102,100,99,98,99,101,104,106,110,
  114,119,126,131,138,144,149,152,154,155,155,153,151,149,145,142,138,134,131,
  128,124,118,112,107,104,103,102,103,105,109,112,114,115,117,121,127,133,139,
  145,149,152,153,153,152,149,146,142,139,135,132,129,127,125,123,118,113,110,
  109,109,110,112,114,117,119,120,121,123,126,130,134,139,143,145,146,146,145,
  144,142,140,138,135,131,129,127,125,124,122,121,119,119,120,120,121,120,120,
  120,122,123,124,125,126,128,130,132,135,137,138,138,138,138,137,136,136,135,
  132,128,125,123,123,123,125,126,128,128,129,128,128,127,127,127,127,126,125,
  124,123,123,124,126,128,130,131,132,132,133,134,133,132,130,128,127,125,125,
  127,129,130,132,133,134,134,134,133,132,130,128,127,125,122,121,120,120,120,
  121,122,125,128,129,129,128,128,127,127,127,127,128,129,130,131,134,137,139,
  140,141,141,140,137,134,131,128,126,124,121,118,116,116,117,118,119,120,121,
  122,124,125,125,126,128,129,130,131,133,135,136,137,139,140,140,140,138,136,
  134,132,131,129,128,128,128,126,125,122,119,116,114,113,114,116,118,122,125,
  127,128,131,133,135,136,136,137,138,137,136,135,134,134,133,133,133,134,135,
  136,136,135,132,128,123,119,116,113,111,111,112,115,118,121,126,130,133,134,
  135,135,134,132,131,130,130,130,130,131,133,136,139,140,139,138,136,134,132,
  130,128,126,124,123,122,122,122,122,123,124,125,125,126,126,125,125,124,124,
  124,125,127,129,133,136,137,136,134,132,130,129,128,128,129,130,131,132,134,
  134,134,134,134,134,132,130,128,126,123,120,118,118,119,122,126,129,133,135,
  135,134,132,130,128,125,123,122,122,122,123,125,127,129,131,133,135,136,136,
  136,134,133,132,132,132,132,132,133,134,134,133,132,131,128,127,125,124,123,
  122,121,121,121,121,121,121,123,125,126,126,126,125,125,125,125,126,128,131,
  135,139,143,147,149,149,148,146,143,139,135,131,128,126,123,121,120,119,119,
  119,119,120,120,118,115,114,114,115,116,118,121,125,129,133,136,138,140,141,
  141,141,141,141,142,142,142,142,141,140,137,135,132,130,128,126,124,121,117,
  114,112,112,113,114,116,119,123,127,129,131,131,131,131,130,129,129,129,129,
  130,131,132,132,132,132,132,133,134,134,134,133,132,131,131,131,132,132,132,
  132,132,131,130,129,128,128,127,127,127,126,126,125,123,122,120,120,120,121,
  122,123,123,122,121,121,122,124,127,131,135,139,142,144,146,146,145,144,143,
  142,140,139,137,135,132,129,127,125,122,120,119,119,118,117,114,112,111,112,
  113,116,120,123,127,129,132,133,134,134,135,136,137,139,141,143,144,145,144,
  143,141,138,134,132,130,128,127,124,122,120,119,118,118,120,121,122,124,125,
  126,127,127,127,126,126,127,127,128,130,132,133,133,132,130,129,128,127,127,
  126,126,126,127,128,130,132,134,135,136,137,138,138,138,137,135,134,132,130,
  128,128,129,129,129,128,126,124,120,118,116,115,115,114,114,114,115,117,120,
  123,127,130,133,135,136,137,138,138,139,140,141,142,143,144,144,143,141,138,
  134,130,128,125,123,120,118,116,114,112,112,113,116,120,123,125,127,128,127,
  126,126,126,127,128,130,133,135,137,137,137,136,134,132,131,131,131,130,129,
  129,128,128,128,128,128,129,130,131,131,131,130,128,128,127,126,127,128,129,
  130,130,129,128,126,124,122,121,120,120,120,119,119,119,120,121,124,126,129,
  131,134,136,138,139,140,140,139,138,137,136,136,135,135,134,134,132,131,128,
  126,123,121,119,117,116,115,115,116,118,120,121,123,124,125,127,128,129,131,
  132,132,132,133,133,134,134,134,134,134,134,133,132,131,130,128,128,127,125,
  124,123,123,124,125,127,129,130,131,132,132,131,131,130,130,129,129,130,130,
  130,128,127,125,123,121,120,119,118,118,119,120,120,120,121,122,123,126,128,
  132,135,137,139,139,139,138,137,136,136,137,138,139,138,137,135,132,129,127,
  125,123,122,121,120,119,119,119,118,117,116,117,118,120,122,124,126,127,128,
  128,128,129,131,133,134,134,134,134,133,132,131,131,131,131,130,129,128,127,
  126,126,128,129,131,133,134,134,134,133,132,130,129,128,128,129,130,130,130,
  128,125,122,119,115,114,113,114,115,117,118,118,118,118,119,121,124,127,130,
  134,136,138,139,140,140,140,140,141,141,142,142,141,140,138,136,132,129,127,
  125,123,121,119,117,116,115,115,115,116,117,118,119,120,121,121,121,121,122,
  124,127,129,131,133,133,133,132,131,130,130,130,130,130,130,130,130,130,130,
  130,131,132,133,135,135,136,135,135,134,132,131,130,129,129,129,129,128,127,
  124,121,119,117,115,114,114,114,114,114,114,114,115,117,120,123,127,130,132,
  134,136,137,138,139,139,140,141,141,142,141,140,139,137,135,133,131,129,128,
  126,124,122,119,117,115,114,114,117,119,122,123,125,125,125,125,125,125,126,
  127,128,130,131,132,132,131,130,130,130,129,129,128,127,126,125,124,124,124,
  124,125,126,127,128,129,130,131,133,134,135,135,136,136,136,136,135,133,132,
  130,129,128,128,127,127,126,124,122,120,117,116,115,116,117,119,121,123,123,
  124,125,126,127,129,130,132,134,135,136,137,137,136,135,133,132,130,129,128,
  127,126,126,126,127,128,128,128,128,128,128,128,128,128,128,128,129,130,131,
  132,132,132,131,130,129,128,126,125,125,125,124,123,122,120,118,117,117,118,
  120,122,124,126,127,128,129,129,130,131,132,133,134,135,136,137,138,139,139,
  139,138,136,134,133,130,128,126,124,122,122,121,121,122,123,125,126,126,126,
  125,124,124,125,126,127,128,128,129,129,129,128,128,127,127,128,128,128,128,
  127,126,124,123,122,122,122,123,125,127,128,130,131,133,134,136,138,139,139,
  140,140,139,139,138,136,134,132,130,129,128,127,125,123,121,119,117,116,116,
  116,117,119,120,121,122,122,122,123,124,126,128,130,132,133,134,135,135,134,
  133,132,131,130,129,129,128,128,127,127,126,126,127,128,129,130,131,132,133,
  133,133,133,133,133,133,133,133,132,132,131,130,128,127,126,125,124,123,122,
  121,120,119,119,118,117,117,118,120,121,123,125,127,128,129,130,131,132,133,
  134,134,135,135,135,135,135,135,135,135,135,134,134,133,132,131,130,128,128,
  127,126,125,125,124,123,123,122,123,124,125,126,127,128,128,129,129,129,128,
  128,127,127,126,125,124,124,123,123,123,124,124,125,125,126,126,128,128,129,
  129,129,129,130,131,132,133,134,135,135,136,136,135,135,134,133,133,132,131,
  129,127,124,122,120,119,119,119,119,119,120,120,121,122,123,125,127,129,131,
  133,134,134,135,135,134,134,133,132,131,130,129,128,127,126,125,124,123,122,
  122,121,120,120,120,121,122,123,124,126,128,130,132,135,137,138,139,140,139,
  138,137,136,134,132,131,129,128,126,124,123,122,121,121,122,122,122,123,124,
  124,125,126,126,127,128,129,131,131,132,132,132,133,133,132,132,132,131,129,
  128,126,124,122,121,119,119,118,119,119,119,121,123,125,127,129,131,133,134,
  136,137,138,138,138,137,136,135,134,132,131,130,130,129,129,129,128,128,127,
  125,124,123,122,122,122,122,122,122,122,122,122,123,124,125,127,128,128,129,
  129,130,129,129,128,128,127,127,127,126,126,126,126,126,127,127,128,128,130,
  131,132,133,134,134,134,134,134,134,133,133,133,132,132,132,131,131,130,129,
  128,128,128,127,127,126,125,124,122,120,119,119,118,119,120,120,121,122,123,
  124,125,127,128,129,129,130,130,130,130,130,130,130,130,129,130,130,130,131,
  131,131,131,132,132,132,132,132,132,132,131,131,130,130,130,130,131,131,131,
  131,131,131,131,131,130,129,128,127,126,124,123,122,120,119,118,117,117,116,
  117,118,119,120,121,123,124,125,126,128,129,130,131,131,132,132,132,133,133,
  133,134,134,134,135,135,136,135,135,134,133,131,130,129,128,127,126,126,126,
  126,126,127,128,128,129,130,130,130,130,130,129,129,128,127,126,124,122,121,
  120,119,118,117,116,116,116,117,118,119,120,121,122,123,124,126,127,128,130,
  132,133,135,136,137,138,139,139,140,140,140,140,140,139,138,136,134,132,130,
  128,126,125,123,123,122,122,121,121,122,123,124,125,126,126,127,127,127,127,
  127,126,125,124,123,122,121,121,121,121,120,120,120,120,120,120,121,122,123,
  124,125,126,128,129,130,132,134,136,137,139,140,141,142,143,143,143,143,142,
  141,140,138,136,133,130,128,126,123,121,119,117,116,116,116,116,117,118,118,
  119,120,121,122,122,123,123,124,124,124,124,125,125,126,126,127,127,128,128,
  128,128,128,128,128,128,128,127,128,128,129,130,131,132,133,134,135,136,137,
  138,139,139,139,139,138,137,136,134,133,131,129,127,125,122,119,117,115,114,
  114,114,114,114,115,116,117,119,120,122,123,124,125,126,126,127,128,129,129,
  130,131,132,133,134,135,135,135,134,134,132,131,130,129,128,127,126,126,126,
  126,126,127,127,127,127,128,129,130,131,132,133,133,133,132,132,131,131,130,
  128,128,126,124,123,122,121,120,120,119,120,120,121,122,123,124,125,126,126,
  127,127,128,128,129,130,130,131,132,132,132,133,133,134,134,134,133,132,131,
  129,128,126,124,122,121,121,121,121,122,123,123,124,125,125,127,128,129,131,
  132,133,133,134,134,134,134,134,133,132,131,130,129,128,128,127,126,125,124,
  124,124,124,124,125,125,125,125,125,125,125,125,126,126,126,127,127,128,128,
  129,130,130,131,132,132,131,130,128,127,126,125,124,123,122,122,123,123,124,
  124,125,126,127,128,129,130,131,132,132,133,133,133,134,134,134,134,133,133,
  132,131,130,129,128,126,125,125,124,124,124,124,124,124,124,124,123,123,123,
  123,124,124,125,125,126,127,128,128,129,129,130,130,131,131,131,131,130,129,
  128,127,126,126,125,125,125,126,126,126,126,126,127,127,127,128,128,128,128,
  129,130,130,131,131,131,132,132,132,132,132,131,131,130,129,129,128,128,128,
  128,128,128,128,128,128,128,128,128,128,127,127,126,126,126,126,126,126,126,
  126,126,126,126,126,126,126,125,125,125,124,124,124,124,124,124,124,124,124,
  125,125,126,127,128,129,129,130,131,132,133,134,135,136,137,137,137,136,135,
  134,133,133,132,131,131,130,129,129,128,128,128,128,127,127,126,126,125,125,
  125,124,124,124,124,124,124,124,124,124,124,124,124,124,124,124,125,125,126,
  126,126,126,126,126,126,127,127,127,128,128,128,129,129,130,131,131,132,133,
  134,135,136,136,137,137,136,135,134,134,133,132,131,131,130,130,129,129,128,
  128,128,128,127,127,126,126,126,126,126,125,125,125,124,124,124,124,124,124,
  125,125,124,124,124,124,124,124,125,126,126,127,127,128,128,128,128,129,129,
  130,130,131,131,132,132,132,133,133,133,133,134,134,133,133,133,132,132,131,
  130,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,129,129,129,128,128,127,126,124,123,122,122,121,122,
  122,123,123,124,125,125,126,127,128,128,129,129,130,130,131,131,131,131,132,
  132,133,134,134,134,134,134,133,132,131,130,129,129,129,129,129,129,129,129,
  129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,129,129,129,
  128,128,127,125,124,123,122,122,123,123,124,124,125,125,126,127,127,128,128,
  129,130,131,131,131,131,131,131,130,131,131,132,132,133,133,132,132,131,130,
  128,128,127,127,127,127,127,128,128,128,128,128,128,129,129,130,131,131,131,
  131,131,131,130,130,131,131,131,132,131,131,130,129,127,126,125,124,123,123,
  123,123,124,124,124,125,125,125,126,127,127,128,128,129,129,129,129,129,129,
  130,130,131,132,132,132,132,132,131,130,130,129,129,129,129,129,129,129,129,
  129,129,128,128,128,129,129,130,131,131,131,130,130,129,129,129,129,129,129,
  128,128,128,127,126,125,124,124,124,125,125,126,126,126,126,126,126,126,127,
  127,128,129,130,131,131,131,131,131,131,131,132,132,132,132,131,131,130,129,
  128,128,127,127,127,127,127,128,128,128,128,127,127,127,128,128,129,130,131,
  131,131,131,131,131,131,131,132,132,132,132,131,131,130,129,129,128,128,127,
  127,127,127,127,127,127,126,126,126,125,126,126,126,127,127,128,128,128,128,
  128,128,128,129,129,128,128,128,128,127,127,127,127,127,128,128,129,129,130,
  130,130,130,130,130,130,131,132,132,133,134,134,134,133,133,132,132,131,131,
  131,130,129,129,128,128,127,127,127,127,127,127,128,128,128,127,127,127,127,
  126,127,127,128,128,129,130,130,130,130,130,130,130,129,129,129,128,128,128,
  127,127,127,127,127,127,128,129,129,130,130,129,129,129,128,128,129,129,130,
  131,131,131,131,131,131,131,131,131,131,131,130,130,129,128,128,127,127,127,
  127,128,128,128,129,129,129,129,128,128,128,128,128,128,129,129,130,130,129,
  129,129,128,128,128,128,128,128,127,127,127,126,126,126,126,127,128,128,130,
  131,131,131,132,132,132,132,132,132,132,133,133,134,134,133,133,133,132,132,
  131,131,130,129,128,128,126,126,125,124,124,124,125,125,126,127,127,128,128,
  128,128,128,128,129,130,130,131,132,132,132,132,132,131,131,131,130,130,129,
  128,128,127,127,126,126,126,127,127,128,128,129,129,129,129,129,129,130,130,
  131,132,133,133,134,134,134,133,133,133,132,132,131,131,130,129,129,128,128,
  127,127,127,127,127,128,128,128,128,128,128,128,128,128,128,128,129,130,131,
  131,131,131,130,130,130,129,129,129,128,128,127,127,127,126,126,126,127,127,
  128,128,128,128,129,129,128,128,128,128,128,129,130,131,132,132,132,132,132,
  131,131,130,130,130,130,130,129,129,128,128,128,128,128,128,128,129,129,129,
  129,129,129,129,128,128,128,128,129,129,130,130,130,129,129,128,128,128,128,
  128,127,127,127,126,126,126,126,127,127,128,128,129,129,130,130,130,130,130,
  130,130,131,131,132,132,133,133,133,132,131,131,130,129,128,128,128,127,127,
  126,126,126,126,126,126,126,127,127,127,128,128,128,128,128,128,128,128,129,
  129,129,130,130,130,130,130,130,130,130,129,129,129,128,128,128,128,128,128,
  128,129,129,130,131,132,132,132,132,132,132,131,131,131,131,131,132,132,132,
  132,132,131,130,129,128,128,127,126,125,124,124,123,124,124,124,125,126,127,
  128,128,128,128,128,128,128,129,129,130,130,131,132,132,133,133,132,132,131,
  130,129,128,128,128,127,127,127,127,127,127,127,127,128,128,128,129,130,131,
  132,132,132,132,132,132,133,133,133,133,134,133,133,133,132,131,130,130,129,
  128,128,127,126,126,125,125,125,125,125,125,125,126,127,127,128,128,128,128,
  128,128,128,129,129,130,131,131,131,131,131,130,130,129,128,128,127,126,126,
  126,126,126,127,127,128,128,128,128,129,129,130,130,130,131,131,131,132,132,
  133,134,134,134,134,133,133,132,131,131,130,129,128,128,128,128,127,127,126,
  126,125,125,125,125,126,127,127,128,128,129,130,130,131,131,131,131,132,131,
  131,131,130,130,130,129,129,128,128,128,127,127,127,127,126,126,126,126,126,
  127,127,128,128,129,129,130,130,129,129,129,130,130,131,132,132,132,132,132,
  131,131,130,130,129,128,128,128,128,128,128,128,128,127,127,127,127,128,128,
  128,129,130,130,131,131,132,132,132,132,133,133,133,132,131,130,129,128,128,
  127,127,126,126,125,125,124,124,123,123,122,122,123,124,125,126,127,128,129,
  129,130,130,130,131,131,132,133,133,134,134,135,135,135,135,134,133,133,132,
  131,130,130,129,129,129,129,129,129,129,129,128,128,128,127,127,127,127,127,
  128,128,129,130,131,130,130,129,128,127,126,125,125,124,125,125,126,127,127,
  127,127,126,125,125,124,125,125,127,128,129,131,132,133,134,134,134,134,133,
  133,133,132,132,131,131,131,131,131,131,131,130,129,128,127,126,125,124,124,
  124,124,125,126,127,127,128,128,127,127,126,126,126,126,127,128,129,130,131,
  132,132,132,131,130,129,128,128,127,127,127,128,128,128,128,129,128,128,128,
  127,127,126,126,126,126,127,128,128,129,131,131,132,132,131,130,129,128,127,
  127,126,126,127,128,128,129,129,129,128,128,127,125,124,124,123,124,125,127,
  128,129,130,130,131,131,130,130,129,129,129,129,129,130,131,131,132,132,132,
  131,130,129,128,127,126,125,125,125,125,126,126,127,127,127,127,126,126,125,
  125,125,125,126,127,128,128,129,130,130,130,129,129,128,128,128,127,128,128,
  129,129,129,130,130,129,129,128,128,128,127,127,127,128,128,129,129,130,130,
  130,130,130,129,129,128,128,127,127,127,127,127,128,128,128,128,127,127,126,
  126,125,124,124,123,124,124,125,126,127,127,128,128,128,128,128,128,128,128,
  128,128,129,130,130,131,131,131,131,130,130,129,129,129,128,128,128,128,128,
  128,127,127,127,127,126,126,126,126,126,126,126,127,127,127,127,127,127,127,
  127,127,127,127,127,127,127,127,127,127,128,128,128,128,128,127,127,126,126,
  126,126,126,127,128,128,128,128,128,129,129,129,129,129,129,129,129,129,129,
  129,129,129,128,128,128,128,128,128,128,128,127,127,126,126,126,126,125,126,
  126,126,126,126,126,126,126,126,126,127,127,127,128,128,128,128,129,129,129,
  129,129,129,130,130,130,130,130,130,130,129,129,128,128,127,127,127,127,127,
  127,127,127,127,126,126,126,126,125,125,125,125,126,126,126,126,126,126,127,
  127,128,128,128,129,129,129,128,128,128,128,128,128,128,128,128,129,129,129,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,127,127,126,126,126,126,126,127,127,128,128,128,128,128,
  127,127,127,127,126,126,127,127,127,127,128,128,128,128,128,129,129,130,130,
  130,129,129,129,128,128,128,128,128,128,128,128,128,128,128,127,126,126,125,
  125,125,125,125,125,125,125,125,126,126,126,127,127,128,128,129,129,129,129,
  129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,
  127,127,127,126,126,126,127,127,127,128,128,129,129,129,129,129,128,128,128,
  127,127,127,128,128,128,128,128,128,127,127,126,125,125,125,125,125,125,125,
  125,126,126,126,127,127,128,128,128,129,129,129,129,129,129,128,128,128,128,
  128,128,128,128,128,127,127,126,126,126,126,126,126,126,125,125,125,125,125,
  125,125,125,126,127,128,128,129,129,129,129,129,129,129,129,129,129,130,130,
  130,130,129,129,128,128,128,128,127,127,127,126,126,125,125,125,125,125,125,
  126,126,127,127,127,127,127,127,127,127,126,127,127,127,128,128,128,128,128,
  128,127,127,126,126,126,127,127,127,127,127,127,127,127,127,127,128,128,128,
  129,130,130,130,130,130,129,129,129,129,129,129,129,129,129,129,128,128,128,
  127,127,126,126,126,125,125,125,125,124,124,124,124,124,125,125,126,127,128,
  128,128,128,128,128,128,128,128,128,129,129,130,130,129,129,129,128,128,128,
  128,127,127,128,128,128,127,127,127,127,127,127,127,127,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,127,127,127,
  127,127,127,127,126,126,125,125,126,126,127,127,128,128,128,129,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,127,
  127,127,127,127,127,127,127,128,128,128,129,129,129,129,128,128,128,128,128,
  128,129,129,129,129,129,129,128,128,128,128,127,127,127,127,127,127,127,127,
  127,127,127,127,127,127,127,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,127,127,127,127,127,127,127,128,128,128,128,128,128,128,128,128,128,
  128,128,128,129,129,130,130,130,130,130,129,129,128,128,128,128,128,128,128,
  127,127,127,126,126,126,126,126,126,126,126,127,127,127,127,128,128,128,128,
  129,129,130,130,130,130,130,130,130,130,129,129,129,129,129,128,128,128,128,
  127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,127,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,127,127,127,127,127,126,126,126,126,126,127,127,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,127,127,127,127,127,127,127,127,127,127,127,127,127,127,
  127,127,128,128,128,128,128,128,128,128,128,129,129,129,130,130,130,130,129,
  129,129,129,128,128,128,128,128,128,128,128,127,127,127,127,127,126,126,126,
  126,126,126,126,126,126,126,126,127,127,128,128,128,129,129,129,129,129,129,
  129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,
  128,127,127,127,127,127,127,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,127,
  127,127,127,127,127,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,127,127,127,127,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,127,127,
  127,127,128,128,128,128,128,127,127,127,127,127,128,128,128,128,128,128,128,
  128,128,128,128,128,129,129,129,129,129,129,129,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,127,127,127,127,127,127,127,127,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,128,
  128,128,128,128,128,128,127,127,127,127,127,127,127,127,127,127,127,127,127,
  128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,128,128,128,
  128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,129,128,
  128,128,128,128,128,128,128,128,128,127,127,127,127,127,127,127,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,129,129,129,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,129,129,129,129,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,127,127,127,127,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,129,129,129,129,129,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,129,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,
  129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
};

static const char dsdgpain[] = {
  3,0,34,86,196,33,0,0,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,130,132,132,132,134,137,141,144,142,138,136,133,128,124,
  119,116,114,117,121,123,124,125,124,123,120,117,116,118,124,128,130,130,131,
  131,131,130,128,125,127,132,140,145,146,144,144,142,137,132,127,121,120,123,
  127,127,124,121,118,114,109,104,104,108,114,121,125,128,131,134,137,137,133,
  132,134,141,147,150,147,144,142,138,132,126,120,117,118,123,128,128,127,125,
  123,118,112,108,108,112,118,122,123,123,123,125,127,127,125,127,133,143,151,
  152,150,148,145,141,135,130,125,124,127,129,129,125,120,116,111,103,98,97,
  103,114,123,127,129,130,134,138,137,131,129,133,143,149,149,146,140,136,130,
  124,117,112,114,123,131,133,131,129,127,122,113,106,103,107,116,121,123,123,
  123,126,128,127,124,126,136,148,155,153,150,148,144,139,131,123,118,118,123,
  125,122,117,113,110,105,99,97,103,116,126,130,131,134,139,144,143,137,133,
  136,144,149,148,142,136,132,129,122,114,111,116,123,125,123,120,120,118,112,
  105,101,105,114,122,126,129,132,136,139,138,135,137,145,151,152,150,148,146,
  144,139,129,120,116,118,119,115,111,108,104,99,93,93,98,108,118,126,133,143,
  151,155,155,152,152,153,154,153,150,143,137,135,131,123,117,116,118,115,111,
  112,112,109,103,100,100,102,107,115,124,133,141,146,149,154,156,156,157,156,
  153,147,143,142,139,131,126,124,122,117,112,112,111,104,96,92,92,93,97,105,
  114,124,133,140,148,157,164,168,170,172,169,163,157,153,145,133,127,125,121,
  112,105,104,101,92,83,78,79,83,89,101,115,127,135,143,153,160,164,167,173,
  176,172,164,160,156,147,136,131,128,122,116,117,117,111,100,89,82,79,78,83,
  92,99,106,119,134,142,147,154,163,169,172,176,179,175,168,163,159,152,143,
  135,128,122,113,102,91,80,71,66,66,68,74,87,103,118,129,143,160,171,175,179,
  184,183,175,168,167,164,155,148,142,134,126,118,110,97,85,72,66,69,74,81,90,
  99,107,120,136,146,151,158,169,178,181,181,176,167,158,153,151,152,151,148,
  144,135,120,102,87,74,62,56,56,65,84,103,111,119,134,148,153,158,168,180,183,
  177,173,176,174,163,158,159,154,143,134,125,110,90,66,48,45,51,58,64,81,106,
  128,139,147,161,175,177,176,181,185,175,163,164,167,164,155,147,143,135,120,
  101,85,73,60,49,48,60,79,92,100,119,143,154,152,160,176,178,169,164,168,167,
  156,151,159,166,159,148,143,138,125,101,75,62,59,54,46,58,87,102,107,127,152,
  165,164,164,173,175,159,149,161,173,168,166,177,181,166,141,119,97,76,52,40,
  50,67,72,76,100,126,133,131,142,163,165,155,160,172,167,152,155,171,177,166,
  159,161,153,128,98,82,76,68,55,58,83,101,98,105,135,155,147,144,160,165,148,
  137,148,160,149,142,161,181,175,157,148,138,114,80,61,66,76,73,74,105,136,
  133,127,145,165,153,137,143,155,140,120,136,165,162,147,158,178,163,126,100,
  92,77,53,47,69,99,109,115,144,172,165,147,155,165,142,120,128,145,139,130,
  152,184,180,156,150,148,115,66,45,53,57,51,65,105,137,137,144,174,190,167,
  146,153,153,128,114,138,165,159,154,178,198,175,133,108,91,56,19,24,57,75,82,
  110,153,168,158,160,175,168,137,126,139,139,125,139,180,199,187,178,178,157,
  106,57,44,43,33,35,76,119,124,126,154,175,153,128,139,152,134,119,143,174,
  172,169,194,215,188,141,116,95,59,29,34,62,82,93,113,142,149,133,127,132,127,
  111,119,141,151,156,182,217,220,200,180,163,126,69,43,52,60,65,95,138,151,
  139,137,135,111,79,75,98,108,110,145,198,221,216,223,228,193,132,85,61,48,41,
  61,110,150,158,160,164,142,92,59,60,61,58,77,129,173,194,224,255,252,203,149,
  115,79,46,46,89,136,160,177,194,180,128,78,54,34,13,19,66,114,138,177,228,
  247,214,172,147,116,84,81,117,166,194,209,218,207,156,91,53,36,11,2,36,89,
  116,143,186,209,177,124,100,92,77,86,145,221,255,0,0,237,165,83,37,26,11,15,
  63,74,124,175,221,252,243,176,74,6,1,18,90,161,196,180,127,74,60,102,163,228,
  0,0,183,84,13,1,17,83,156,199,183,125,67,54,93,156,220,0,0,190,90,17,1,20,86,
  154,195,189,134,72,51,88,155,223,0,0,195,99,14,1,14,92,161,188,179,136,79,45,
  74,147,224,0,255,190,103,30,1,17,91,170,202,181,127,69,40,67,139,213,0,0,199,
  104,24,1,22,92,163,198,186,135,71,35,63,142,221,0,0,202,117,36,4,25,94,168,
  198,180,128,68,33,55,130,211,0,0,204,112,34,6,27,88,160,198,188,136,74,37,60,
  134,210,0,0,206,120,39,7,31,95,163,194,181,133,73,35,54,126,206,255,0,203,
  119,42,10,30,91,159,191,178,136,84,47,60,125,206,0,0,207,125,51,16,31,84,150,
  182,168,127,79,48,61,121,198,248,0,214,134,59,28,48,93,150,183,171,129,80,46,
  53,109,187,241,251,213,141,72,32,43,85,144,181,170,128,85,54,57,104,179,239,
  255,220,145,77,41,53,84,131,171,170,135,93,62,61,99,165,226,250,227,159,89,
  47,48,71,111,156,165,135,97,73,70,94,150,216,250,240,181,113,62,51,68,99,141,
  159,141,103,71,59,85,138,199,238,243,198,128,73,55,69,99,139,160,146,109,75,
  56,73,126,192,236,246,205,137,80,56,68,95,138,163,152,114,77,53,67,115,178,
  223,239,208,145,87,56,66,94,134,160,156,125,91,64,68,110,174,225,241,210,143,
  85,54,59,85,127,159,159,125,89,65,70,106,163,220,247,225,158,96,60,63,84,123,
  154,159,127,88,59,58,98,160,218,243,224,161,99,62,64,84,119,150,162,137,100,
  70,61,93,150,209,239,228,166,104,64,64,85,117,144,154,134,98,68,58,91,145,
  206,239,233,174,109,67,64,82,115,149,164,148,108,74,60,90,140,195,231,230,
  177,108,64,59,77,109,140,159,150,114,78,57,84,140,199,236,235,185,119,75,69,
  82,109,137,154,147,113,80,59,76,126,186,228,234,189,122,75,65,79,109,140,157,
  153,120,85,59,77,128,186,227,234,191,124,73,62,72,103,140,159,157,126,90,59,
  70,119,180,226,235,193,124,70,57,71,105,144,165,163,132,97,62,66,117,181,225,
  232,193,125,70,55,67,98,140,165,166,134,99,65,69,116,174,221,236,198,126,71,
  57,68,98,139,162,165,140,102,62,65,117,179,222,230,194,128,72,55,67,101,141,
  166,166,140,104,64,60,110,175,221,234,199,128,63,49,63,98,143,172,172,148,
  110,65,60,109,176,221,234,200,130,66,42,44,75,127,164,168,148,119,73,59,104,
  172,220,238,207,135,73,45,46,79,130,167,176,161,129,79,62,106,174,225,241,
  209,139,71,34,32,68,128,176,187,166,129,82,61,95,165,221,238,209,141,74,42,
  41,76,131,179,189,164,122,72,55,94,166,216,227,196,129,58,28,36,77,139,187,
  190,162,124,75,61,108,182,221,220,191,128,65,37,44,82,140,184,184,154,110,64,
  60,115,186,221,216,177,111,59,41,50,91,146,183,181,163,125,76,71,124,186,208,
  200,169,116,70,51,59,97,142,174,173,159,123,77,71,120,184,204,191,162,117,77,
  49,54,96,147,181,183,172,133,81,72,121,178,195,187,155,109,75,47,50,92,148,
  182,186,174,131,85,82,128,180,191,180,146,102,69,42,51,95,151,181,183,173,
  128,87,97,145,183,183,166,127,93,72,55,65,107,158,177,175,164,124,97,115,160,
  182,165,141,101,80,76,74,90,124,153,156,155,155,137,132,156,185,174,134,103,
  76,79,89,91,99,123,142,147,150,148,134,142,169,185,155,119,89,66,79,101,113,
  121,134,140,135,140,142,141,164,196,197,147,98,64,51,71,98,117,133,146,148,
  141,138,132,140,171,202,195,143,96,58,46,67,99,120,137,150,149,142,137,137,
  155,186,209,179,122,73,38,35,66,100,124,145,158,157,146,138,144,168,196,205,
  165,110,62,35,42,76,104,124,142,152,149,142,137,151,183,213,203,140,78,41,36,
  56,89,114,132,145,149,144,140,140,158,187,211,195,135,74,40,38,63,91,112,131,
  145,149,146,145,153,177,205,217,176,107,49,28,39,66,89,113,136,149,151,150,
  151,164,191,217,213,160,83,30,20,42,76,98,123,141,151,149,147,151,173,208,
  232,208,135,59,23,24,49,80,105,126,135,139,144,151,161,186,220,234,196,120,
  50,25,31,56,78,100,119,129,137,144,155,174,205,231,228,177,96,37,27,43,64,79,
  103,122,131,137,147,163,187,221,240,214,147,70,25,24,45,65,83,110,131,142,
  149,158,171,198,231,234,187,114,49,24,30,49,70,95,118,131,139,148,160,179,
  214,243,231,170,96,45,29,37,52,69,92,115,126,131,148,170,195,230,250,218,143,
  71,38,39,48,57,73,95,114,123,133,158,187,218,246,243,189,110,51,33,40,48,60,
  79,103,121,127,141,174,206,234,253,228,155,76,33,29,42,56,72,91,114,125,128,
  150,191,228,255,252,198,112,45,29,39,51,65,78,94,112,120,134,175,218,246,0,
  233,159,76,30,30,43,57,72,83,101,115,121,145,197,245,0,254,197,109,41,31,44,
  55,71,83,87,98,109,128,171,228,0,0,227,152,72,35,44,50,58,70,79,91,105,112,
  137,193,254,0,255,195,114,54,47,58,58,61,66,70,83,105,128,164,217,0,0,235,
  174,104,58,52,51,49,49,57,74,94,115,146,193,246,0,0,217,145,86,57,43,29,30,
  42,63,85,106,131,169,215,0,0,0,197,124,67,36,13,6,27,58,91,115,128,147,185,
  239,0,0,239,166,92,47,23,1,3,32,66,99,125,141,165,207,0,0,0,210,136,69,33,18,
  8,19,52,87,112,132,158,192,236,0,0,222,164,101,48,18,8,11,29,62,100,124,149,
  182,221,0,0,251,191,122,66,34,12,5,26,55,81,105,123,150,197,252,0,0,220,142,
  72,39,26,16,28,52,74,98,118,139,178,228,0,0,246,171,88,34,25,32,41,53,63,73,
  90,112,153,210,0,0,0,210,123,52,26,30,39,48,55,61,75,99,138,194,254,0,0,238,
  154,74,26,20,29,39,47,56,68,90,126,180,237,0,0,0,197,115,43,16,19,25,35,45,
  61,86,115,153,203,254,0,0,234,154,76,32,24,29,36,44,59,77,100,132,182,234,0,
  0,0,192,116,53,23,20,30,35,46,65,88,117,161,212,0,0,0,223,152,82,39,25,25,29,
  37,53,77,105,140,191,246,0,0,247,180,108,56,27,18,22,27,39,62,96,132,176,227,
  0,0,0,210,137,75,37,22,22,21,28,46,74,113,160,206,0,0,0,237,170,103,56,28,24,
  25,23,37,64,97,140,190,237,0,0,252,193,127,77,44,24,25,23,27,50,87,128,171,
  216,0,0,0,213,147,88,50,28,26,29,31,48,76,112,152,190,234,0,0,246,181,111,66,
  41,26,29,28,38,64,98,134,176,223,0,0,0,203,138,84,42,21,23,26,36,63,87,120,
  163,200,240,0,0,239,167,97,54,28,21,29,37,53,81,115,149,183,227,0,0,251,192,
  112,51,22,14,23,42,64,85,107,138,171,207,0,0,0,214,142,74,39,21,18,28,46,73,
  102,131,162,194,237,0,0,226,166,94,45,25,24,28,38,67,100,127,157,182,221,0,0,
  239,189,123,62,32,16,16,24,50,92,124,148,173,199,247,0,252,206,150,83,39,25,
  18,20,41,84,123,154,177,188,221,0,0,223,168,95,39,22,20,23,31,60,104,136,165,
  188,215,0,0,244,191,131,65,29,22,22,23,42,86,125,153,175,192,234,0,0,207,150,
  90,43,26,22,21,33,73,118,147,173,187,212,0,0,233,174,109,50,24,24,22,23,53,
  104,137,162,183,198,239,0,0,198,131,73,34,25,32,35,40,76,122,148,170,186,218,
  0,0,230,166,99,44,23,20,19,31,63,111,146,170,180,197,248,0,254,190,119,58,26,
  19,27,32,46,92,133,152,170,182,217,0,0,225,154,82,30,16,18,25,39,72,122,152,
  172,179,197,250,0,249,179,104,42,13,13,26,36,61,110,146,166,176,182,218,0,0,
  210,136,68,30,12,18,36,49,86,136,162,173,174,194,244,0,242,168,93,43,25,21,
  25,35,64,115,154,171,169,171,209,0,0,216,136,66,31,20,29,39,50,85,135,166,
  175,169,181,224,0,247,181,97,37,18,22,32,47,74,121,164,181,176,176,200,241,0,
  215,133,60,22,11,23,38,62,102,150,174,176,171,187,222,255,243,172,89,39,24,
  25,31,46,82,134,169,172,164,170,201,243,254,207,128,63,35,29,30,27,47,96,150,
  176,178,171,181,218,255,247,182,100,46,33,39,30,24,54,112,159,179,171,166,
  190,237,0,234,157,79,36,28,25,16,33,79,141,178,185,172,177,211,253,254,199,
  112,41,15,17,17,25,59,118,171,193,187,172,182,223,0,241,159,67,14,11,26,25,
  42,91,150,179,185,172,163,188,241,0,210,130,54,15,20,36,37,56,106,156,182,
  182,158,158,207,0,247,181,96,36,19,29,27,39,82,138,174,186,173,152,169,222,
  252,219,150,72,23,19,38,46,57,102,156,187,189,165,155,190,240,232,167,94,40,
  19,27,43,59,100,150,176,184,176,155,161,209,243,211,145,78,35,25,27,23,48,
  107,161,184,184,164,158,190,232,231,182,124,69,35,32,37,42,74,130,170,186,
  173,147,152,206,253,227,158,87,38,24,30,35,53,107,167,195,193,163,139,156,
  218,238,193,124,58,19,22,47,61,80,133,177,193,181,152,139,186,245,225,162,90,
  30,5,23,55,75,116,166,188,185,161,136,146,212,239,201,137,69,22,14,35,56,85,
  139,181,188,163,133,131,180,240,228,179,107,42,10,20,54,75,110,162,192,184,
  148,122,138,206,237,205,144,69,17,4,34,71,96,145,188,195,164,133,124,166,227,
  223,180,109,39,3,9,52,86,126,176,198,177,139,116,134,204,234,208,154,77,18,1,
  27,69,108,155,192,193,154,117,111,162,225,227,193,122,44,5,7,45,90,134,179,
  202,179,129,102,124,197,233,216,166,86,25,3,21,64,112,161,197,197,155,110,98,
  147,217,232,204,133,53,9,6,40,87,134,181,205,180,127,96,111,177,227,221,178,
  102,35,3,17,58,104,150,190,198,155,111,94,132,208,238,213,147,71,26,14,32,73,
  126,169,197,179,129,99,99,150,213,229,196,123,56,23,23,49,91,138,179,198,163,
  116,90,110,175,228,226,172,100,48,23,31,62,109,153,184,182,142,106,94,127,
  191,231,213,156,90,44,31,45,78,122,159,183,171,131,102,99,136,198,222,193,
  138,81,47,41,59,90,127,155,169,153,124,103,105,148,208,218,183,130,85,57,58,
  72,98,129,148,154,139,115,97,107,156,208,208,172,129,91,68,71,80,103,124,138,
  147,135,113,102,118,169,211,203,170,135,100,78,77,82,100,117,127,132,125,106,
  104,125,172,207,203,175,146,109,88,86,87,98,110,118,123,118,105,112,130,170,
  202,200,179,148,113,95,91,90,98,107,109,110,110,107,116,131,168,202,205,184,
  155,126,107,95,85,89,95,94,95,103,113,129,137,162,196,204,188,163,134,114,
  103,90,82,79,81,86,99,113,132,141,161,194,202,190,173,148,121,107,94,81,68,
  70,78,92,109,129,147,159,188,202,196,185,164,134,114,101,78,59,58,69,81,102,
  125,151,166,183,202,196,188,174,147,114,93,75,63,57,66,75,89,122,150,169,174,
  192,203,192,184,162,126,91,72,62,61,63,71,83,112,147,167,180,181,194,194,183,
  174,143,99,74,67,63,67,67,82,109,133,155,172,185,180,188,184,165,153,123,89,
  75,69,67,77,84,99,121,144,169,179,184,179,169,157,140,135,119,92,81,75,79,93,
  98,104,124,149,170,179,177,170,153,145,130,127,116,95,90,90,99,108,111,118,
  133,152,161,163,159,155,146,136,126,121,115,106,112,112,110,113,119,123,130,
  137,147,148,151,145,144,133,118,117,122,126,128,127,122,123,125,120,118,124,
  129,136,135,143,135,132,121,123,132,136,138,141,137,133,125,115,110,110,114,
  125,134,127,128,123,128,128,136,145,154,154,149,134,127,118,107,103,111,113,
  117,124,118,128,127,135,141,149,152,152,142,138,131,128,127,121,109,100,100,
  104,121,124,137,142,144,146,143,139,139,135,140,147,145,131,113,100,92,99,
  104,124,137,143,141,132,133,131,130,143,151,151,143,136,128,119,119,114,117,
  119,121,125,121,119,119,125,141,145,142,133,129,128,132,138,141,140,141,135,
  126,112,105,110,122,140,143,137,129,120,115,115,120,130,136,143,147,140,136,
  134,135,137,139,136,127,122,119,115,115,111,115,122,120,122,120,123,132,147,
  164,171,167,156,141,131,120,114,109,108,115,107,97,100,104,117,132,146,151,
  152,154,156,157,156,151,141,131,125,117,99,95,97,105,121,133,136,134,129,129,
  130,135,136,133,137,142,138,135,133,135,136,135,134,129,128,126,124,122,116,
  110,113,117,117,113,115,128,142,155,162,161,158,154,148,139,130,121,111,112,
  110,99,94,100,111,125,134,139,140,144,151,157,161,158,146,136,130,124,115,
  107,108,113,119,126,128,126,124,125,131,138,135,128,125,125,125,126,132,139,
  144,148,147,144,137,130,129,130,129,118,109,105,102,101,103,112,125,137,144,
  146,148,153,157,159,156,144,130,123,114,106,98,97,108,120,128,127,122,123,
  130,141,148,146,139,134,133,130,127,124,125,129,134,134,130,125,125,129,134,
  131,123,115,112,112,109,111,118,127,136,144,150,151,145,143,140,140,139,132,
  123,114,107,104,105,111,117,121,126,132,135,135,138,139,139,139,136,132,128,
  125,119,117,123,130,135,137,135,131,129,132,131,124,117,111,108,110,114,114,
  115,124,137,147,151,149,143,144,148,147,139,128,121,115,113,112,109,106,106,
  114,122,127,127,131,137,140,139,135,134,135,133,128,123,127,134,140,142,137,
  128,125,129,132,127,115,104,100,104,109,110,111,117,128,140,148,150,149,150,
  154,154,149,140,128,120,112,106,103,104,108,111,114,114,115,122,133,138,137,
  134,130,131,132,131,129,130,137,142,144,142,137,130,128,130,128,118,108,102,
  101,101,101,103,112,126,136,142,145,147,149,153,154,150,141,133,128,122,113,
  103,100,107,118,123,123,121,122,128,134,136,133,131,131,128,124,120,118,121,
  125,128,133,138,141,143,143,140,134,128,128,126,122,113,105,105,111,118,120,
  122,127,133,139,143,143,141,136,130,125,123,125,127,127,128,128,127,129,132,
  135,137,134,128,122,122,122,116,109,107,112,120,126,129,132,135,137,140,146,
  150,148,143,136,128,120,116,115,119,122,123,121,121,124,128,130,129,126,126,
  128,128,125,120,117,118,123,129,134,140,146,148,148,147,143,136,130,127,122,
  113,107,106,109,114,117,116,121,128,135,140,141,139,135,131,128,126,124,124,
  128,133,137,136,136,137,137,137,134,130,125,121,118,115,112,108,107,112,120,
  127,129,133,138,140,140,141,139,139,139,137,135,131,127,124,122,120,118,121,
  125,127,128,129,128,126,124,121,118,120,122,123,125,128,129,129,132,137,140,
  142,142,140,138,135,128,119,114,112,112,113,116,119,125,132,137,139,138,137,
  134,132,128,124,120,118,116,113,115,118,123,129,139,146,149,148,143,136,131,
  125,119,116,116,117,119,122,124,126,130,134,135,135,135,134,132,127,120,114,
  112,111,113,118,126,132,138,143,146,145,142,136,132,131,130,128,124,119,116,
  116,118,117,117,121,125,129,131,131,129,127,126,124,124,125,126,129,134,138,
  139,139,138,136,135,134,132,129,127,127,126,122,116,111,112,114,117,120,125,
  129,131,132,132,131,130,129,131,133,134,134,135,136,134,132,132,133,134,134,
  133,132,129,125,120,116,112,108,108,110,115,119,123,128,132,133,134,136,137,
  138,139,141,141,137,132,128,125,124,123,125,129,132,132,131,130,126,121,116,
  112,108,107,109,112,116,120,123,128,134,139,143,145,147,148,148,146,141,133,
  126,120,117,116,116,119,123,126,128,128,125,121,118,116,115,114,113,114,116,
  120,124,128,133,140,144,148,152,153,152,149,143,135,126,118,112,109,108,110,
  112,115,118,119,120,121,122,123,126,128,128,129,128,127,127,129,132,135,139,
  142,145,147,145,142,137,131,125,116,108,105,104,105,109,114,118,122,124,127,
  129,132,135,138,140,138,135,132,128,124,123,125,125,126,129,133,136,136,134,
  133,130,125,121,117,114,113,115,119,122,124,126,128,130,133,135,136,137,138,
  136,133,126,120,116,116,118,120,122,125,130,134,136,137,137,136,134,130,127,
  124,122,121,122,123,123,124,123,124,125,127,129,132,134,133,129,125,121,119,
  119,121,123,126,129,132,135,136,137,137,137,136,133,128,124,122,121,122,122,
  120,120,120,122,124,127,128,130,132,133,133,129,126,124,123,124,124,125,126,
  127,128,130,132,133,134,135,134,131,129,128,127,129,129,128,128,127,125,124,
  124,124,125,125,125,124,123,122,121,121,122,122,124,127,129,132,135,138,141,
  142,141,138,135,133,131,129,128,127,125,124,123,122,121,121,121,121,122,121,
  121,123,124,124,125,127,127,128,129,132,134,137,140,142,142,140,137,134,132,
  131,129,127,126,125,124,124,123,122,123,122,123,123,122,121,123,124,126,127,
  128,129,131,133,134,135,136,137,137,136,134,132,130,129,129,128,128,128,128,
  128,128,128,127,128,130,129,126,122,119,117,118,120,122,125,127,129,131,132,
  132,133,136,138,138,135,132,130,130,130,130,129,128,129,130,129,128,127,128,
  130,130,127,123,119,119,121,123,124,126,128,130,132,133,132,131,133,135,136,
  133,129,127,126,127,127,127,128,129,132,133,133,131,132,133,134,131,127,123,
  120,121,121,122,123,124,127,129,129,129,129,132,134,135,134,131,129,129,130,
  130,129,128,129,131,133,132,130,129,130,129,128,125,122,122,124,125,125,125,
  125,128,131,133,133,132,133,134,133,131,128,126,126,127,128,127,127,128,130,
  133,133,133,133,134,134,133,131,128,126,126,126,126,125,124,125,126,128,128,
  128,128,129,128,127,126,126,128,130,131,131,131,131,134,137,137,135,134,133,
  131,129,127,126,125,127,127,126,125,125,127,129,131,131,131,131,129,128,127,
  126,126,128,130,131,130,128,129,131,133,133,132,132,132,131,129,128,126,126,
  128,130,130,128,127,128,130,131,130,128,128,128,127,126,125,125,127,130,133,
  133,133,133,134,136,136,135,134,132,130,128,125,122,122,123,126,127,128,128,
  129,132,134,134,133,132,131,129,127,125,125,126,128,129,129,129,130,133,136,
  137,136,135,134,132,130,128,127,126,127,128,128,127,127,128,130,132,132,131,
  130,129,127,126,126,125,127,128,129,129,128,128,129,133,135,134,132,132,132,
  130,129,128,127,128,130,130,129,128,128,129,131,130,129,128,128,127,126,125,
  126,127,129,131,131,131,131,133,135,135,134,131,128,127,125,125,125,125,126,
  127,128,128,128,129,130,131,131,130,130,130,129,128,128,128,130,132,134,134,
  133,132,132,133,134,133,131,128,125,122,121,122,124,126,128,128,128,129,131,
  133,135,135,132,129,128,127,127,127,127,127,129,131,133,134,134,135,135,136,
  135,133,130,128,126,124,123,123,124,126,127,128,128,128,130,132,133,132,130,
  128,126,125,125,126,127,128,129,131,132,132,134,136,137,135,133,131,129,128,
  127,125,124,123,125,127,129,131,132,133,133,132,131,130,129,127,126,126,125,
  124,126,128,130,130,130,130,132,134,134,133,131,128,127,127,128,127,127,127,
  128,130,132,133,133,134,134,132,129,127,126,125,123,121,120,120,124,127,129,
  130,131,133,135,136,137,137,135,133,131,129,129,129,130,129,128,126,126,128,
  132,132,129,125,123,123,125,127,126,124,122,124,128,132,135,136,136,135,133,
  132,132,132,131,128,125,123,122,124,127,128,126,125,125,128,132,134,133,130,
  128,126,127,129,129,128,127,126,125,127,129,133,134,132,128,125,125,128,129,
  129,127,123,121,124,128,131,132,131,130,129,130,132,134,132,130,126,123,123,
  125,126,127,125,124,123,126,129,131,131,129,127,127,128,130,130,129,128,127,
  127,128,130,131,131,129,128,127,127,127,127,127,125,123,122,123,126,128,128,
  128,127,128,129,131,132,132,130,129,128,128,128,127,126,125,125,126,126,127,
  127,127,127,126,126,126,128,128,127,126,125,125,127,128,128,128,128,128,129,
  129,129,128,129,128,127,126,125,125,125,125,126,125,125,126,127,128,129,129,
  129,131,131,131,130,128,127,127,127,127,126,125,124,124,125,125,125,127,128,
  129,129,128,128,128,129,129,129,128,127,127,127,127,127,127,128,128,129,128,
  126,125,125,127,128,128,127,126,126,126,127,128,128,129,130,130,129,128,128,
  128,128,128,126,124,124,124,124,124,125,126,128,130,130,129,128,128,129,129,
  128,128,127,126,126,126,125,126,128,129,130,128,127,127,127,128,128,127,125,
  123,123,124,124,125,127,128,130,130,129,128,129,128,128,126,125,125,124,124,
  124,123,124,126,128,130,130,129,129,130,131,130,129,128,127,126,125,123,123,
  124,126,127,127,125,125,126,127,128,127,126,125,126,127,127,127,127,128,130,
  131,131,129,129,129,130,129,128,126,125,124,123,123,122,123,125,127,128,127,
  127,128,129,130,130,128,127,127,127,127,126,125,126,127,128,128,128,128,128,
  129,128,127,126,126,127,126,125,124,124,126,128,129,128,127,127,128,128,128,
  127,127,127,127,127,126,126,127,129,130,129,128,128,129,130,129,128,127,126,
  126,127,127,126,125,126,128,128,128,127,127,127,127,126,126,127,128,128,128,
  128,128,128,130,131,131,129,128,128,128,127,125,124,124,125,126,126,127,128,
  129,131,132,131,130,129,129,128,127,126,126,126,126,126,125,126,126,128,129,
  129,128,128,128,128,128,128,127,127,127,126,125,124,125,127,128,129,128,128,
  128,128,129,129,128,128,128,129,128,128,128,128,128,129,128,128,127,126,126,
  126,126,125,126,127,127,127,127,128,129,131,131,131,129,129,129,128,128,127,
  126,125,125,124,124,124,125,127,129,129,129,129,129,129,129,129,128,128,128,
  128,127,126,126,127,128,128,128,128,128,128,128,128,128,127,127,127,126,126,
  125,126,127,128,128,128,128,128,128,129,129,128,128,129,128,128,127,127,127,
  127,128,127,127,127,127,128,127,127,127,127,128,128,127,127,127,128,129,129,
  128,128,128,128,129,129,128,128,128,128,128,127,126,126,127,127,127,127,126,
  127,127,128,128,128,129,129,129,129,128,128,128,128,128,128,126,126,126,127,
  127,127,127,128,128,128,128,128,128,129,129,128,127,127,127,127,127,127,127,
  128,128,128,128,128,129,129,129,129,128,128,127,126,126,126,126,126,127,128,
  128,128,128,128,129,129,129,128,128,128,128,127,127,126,127,127,128,128,128,
  129,129,130,130,129,128,128,128,127,127,126,126,127,127,127,127,127,128,129,
  129,129,128,128,128,128,128,128,127,127,128,128,128,128,128,129,129,129,129,
  128,128,128,128,128,127,126,126,127,127,127,127,128,129,129,129,129,128,128,
  129,128,128,128,128,127,127,127,127,128,128,129,129,129,128,128,128,129,129,
  128,128,128,128,128,128,128,128,128,128,128,128,127,127,127,128,128,128,128,
  128,128,128,128,128,129,129,130,129,128,128,128,128,128,128,128,128,128,127,
  127,128,128,128,129,129,128,128,128,128,128,128,128,128,128,127,127,127,127,
  128,128,129,129,128,128,128,128,128,128,128,128,128,128,127,127,128,128,128,
  128,128,128,128,127,127,127,127,127,127,127,127,127,128,128,128,128,128,129,
  128,128,128,128,128,128,129,129,128,128,127,128,128,129,129,129,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,
  129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,129,129,
  128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,130,129,129,129,
  129,129,129,129,128,129,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,129,129,128,129,130,130,130,130,129,129,129,129,128,128,128,
  128,128,128,128,128,128,128,129,128,128,129,128,128,129,129,128,128,128,128,
  128,128,129,128,128,129,129,129,129,128,128,128,129,129,128,128,129,129,128,
  129,129,128,129,130,129,128,129,129,128,128,128,128,127,127,128,128,127,128,
  128,128,128,128,128,128,129,128,128,128,128,128,128,128,128,128,128,128,129,
  129,129,129,128,129,128,127,128,128,128,127,127,127,127,128,128,128,128,128,
  128,128,128,128,128,128,129,129,129,129,129,129,129,128,129,129,129,128,128,
  128,127,127,128,128,128,128,127,127,128,128,128,128,129,129,129,129,129,129,
  129,128,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,129,129,128,128,128,129,129,129,129,128,128,128,128,128,128,128,128,
  128,128,128,128,128,129,129,129,130,129,129,129,128,128,128,128,128,128,128,
  128,127,128,128,128,128,128,128,128,128,128,129,129,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,130,130,129,
  129,129,129,128,128,128,127,127,127,127,127,127,127,128,128,128,128,129,129,
  129,129,129,128,128,128,128,128,128,128,128,129,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,129,128,128,129,128,129,129,128,128,128,
  128,128,128,128,128,128,129,129,129,129,129,129,129,129,128,128,128,128,127,
  127,128,128,128,128,128,128,128,128,129,129,129,129,128,128,128,128,128,128,
  128,128,128,128,128,129,129,129,129,129,128,128,128,128,128,128,128,128,128,
  128,128,128,128,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,
  129,129,130,130,129,129,129,129,128,128,128,128,128,128,128,128,128,128,129,
  129,129,129,129,129,129,128,128,128,128,128,129,129,128,128,129,129,129,129,
  129,128,128,128,128,128,128,128,128,128,128,128,129,129,129,130,130,129,129,
  129,129,128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,
  129,129,129,129,129,129,129,129,129,129,129,130,129,129,129,129,128,128,128,
  128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,129,129,129,129,
  129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,
  129,129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,129,129,
  129,129,129,129,129,129,129,129,128,128,129,129,128,128,128,129,129,129,129,
  128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,
  129,129,129,128,128,128,128,128,128,128,128,129,129,129,129,129,128,128,128,
  128,128,128,128,128,128,129,129,129,129,129,129,129,129,129,128,128,128,128,
  128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,128,128,128,128,
  128,128,128,129,128,129,129,129,129,129,129,128,128,128,128,128,128,129,129,
  129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,129,
  129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,129,129,
  129,129,129,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,129,
  129,129,128,128,128,128,128,129,129,129,128,129,129,129,129,129,129,129,128,
  128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,129,129,128,128,
  128,128,128,128,128,128,128,129,129,129,129,129,129,129,128,128,128,128,128,
  128,128,128,128,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,
  128,129,129,129,129,129,129,128,129,128,128,128,128,128,128,128,128,128,128,
  129,129,129,129,129,129,129,129,129,128,128,128,128,128,128,128,128,129,129,
  129,128,128,128,128,128,128,128,128,128,128,128,128,128,129,129,129,129,129,
  129,129,129,129,128,128,128,129,129,129,128,129,129,129,129,128,128,128,128,
  128,128,128,128,128,128,128,128,128,129,129,129,129,129,129,129,129,129,128,
  128,128,128,128,128,128,128,129,129,129,128,128,128,128,128,128,128,128,128,
  128,128,128,128,129,129,129,129,129,129,129,129,129,128,128,128,129,129,129,
  129,129,129,129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,129,
  129,129,129,129,128,129,129,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,129,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,129,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,129,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,129,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  129,129,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,128,
  128,128,128,128,128,128,128,128,128,128,128,
};

static const char dogovrly[] = {
  64,1,200,0,0,0,0,0,8,5,0,0,51,5,0,0,94,5,0,0,137,5,0,0,180,5,0,0,223,5,0,0,
  10,6,0,0,53,6,0,0,96,6,0,0,139,6,0,0,182,6,0,0,225,6,0,0,12,7,0,0,55,7,0,0,
  98,7,0,0,141,7,0,0,184,7,0,0,227,7,0,0,14,8,0,0,57,8,0,0,100,8,0,0,143,8,0,0,
  186,8,0,0,229,8,0,0,16,9,0,0,59,9,0,0,102,9,0,0,145,9,0,0,188,9,0,0,231,9,0,
  0,18,10,0,0,61,10,0,0,104,10,0,0,147,10,0,0,190,10,0,0,233,10,0,0,20,11,0,0,
  63,11,0,0,106,11,0,0,149,11,0,0,192,11,0,0,235,11,0,0,22,12,0,0,65,12,0,0,
  108,12,0,0,151,12,0,0,194,12,0,0,237,12,0,0,24,13,0,0,67,13,0,0,110,13,0,0,
  153,13,0,0,196,13,0,0,239,13,0,0,26,14,0,0,69,14,0,0,112,14,0,0,155,14,0,0,
  198,14,0,0,241,14,0,0,28,15,0,0,71,15,0,0,114,15,0,0,157,15,0,0,200,15,0,0,
  243,15,0,0,30,16,0,0,73,16,0,0,116,16,0,0,159,16,0,0,202,16,0,0,245,16,0,0,
  32,17,0,0,75,17,0,0,118,17,0,0,161,17,0,0,204,17,0,0,205,17,0,0,206,17,0,0,
  207,17,0,0,208,17,0,0,209,17,0,0,210,17,0,0,211,17,0,0,212,17,0,0,213,17,0,0,
  214,17,0,0,215,17,0,0,216,17,0,0,217,17,0,0,218,17,0,0,219,17,0,0,220,17,0,0,
  221,17,0,0,222,17,0,0,223,17,0,0,224,17,0,0,225,17,0,0,226,17,0,0,227,17,0,0,
  228,17,0,0,229,17,0,0,230,17,0,0,231,17,0,0,232,17,0,0,233,17,0,0,234,17,0,0,
  235,17,0,0,236,17,0,0,237,17,0,0,238,17,0,0,239,17,0,0,240,17,0,0,241,17,0,0,
  242,17,0,0,243,17,0,0,244,17,0,0,245,17,0,0,246,17,0,0,247,17,0,0,248,17,0,0,
  249,17,0,0,250,17,0,0,251,17,0,0,252,17,0,0,253,17,0,0,254,17,0,0,255,17,0,0,
  0,18,0,0,1,18,0,0,2,18,0,0,3,18,0,0,4,18,0,0,5,18,0,0,6,18,0,0,7,18,0,0,8,18,
  0,0,9,18,0,0,10,18,0,0,11,18,0,0,12,18,0,0,13,18,0,0,14,18,0,0,15,18,0,0,16,
  18,0,0,17,18,0,0,18,18,0,0,19,18,0,0,20,18,0,0,21,18,0,0,22,18,0,0,23,18,0,0,
  24,18,0,0,25,18,0,0,26,18,0,0,27,18,0,0,28,18,0,0,29,18,0,0,30,18,0,0,31,18,
  0,0,32,18,0,0,33,18,0,0,34,18,0,0,35,18,0,0,36,18,0,0,37,18,0,0,38,18,0,0,39,
  18,0,0,40,18,0,0,41,18,0,0,42,18,0,0,43,18,0,0,44,18,0,0,45,18,0,0,46,18,0,0,
  47,18,0,0,48,18,0,0,49,18,0,0,50,18,0,0,51,18,0,0,52,18,0,0,53,18,0,0,54,18,
  0,0,55,18,0,0,56,18,0,0,57,18,0,0,58,18,0,0,59,18,0,0,60,18,0,0,61,18,0,0,62,
  18,0,0,63,18,0,0,64,18,0,0,120,18,0,0,176,18,0,0,232,18,0,0,32,19,0,0,88,19,
  0,0,144,19,0,0,200,19,0,0,0,20,0,0,56,20,0,0,112,20,0,0,168,20,0,0,224,20,0,
  0,24,21,0,0,80,21,0,0,136,21,0,0,192,21,0,0,248,21,0,0,48,22,0,0,104,22,0,0,
  160,22,0,0,216,22,0,0,16,23,0,0,72,23,0,0,128,23,0,0,184,23,0,0,240,23,0,0,
  40,24,0,0,96,24,0,0,152,24,0,0,208,24,0,0,8,25,0,0,64,25,0,0,120,25,0,0,176,
  25,0,0,232,25,0,0,32,26,0,0,88,26,0,0,144,26,0,0,200,26,0,0,0,27,0,0,56,27,0,
  0,112,27,0,0,168,27,0,0,224,27,0,0,24,28,0,0,80,28,0,0,136,28,0,0,192,28,0,0,
  248,28,0,0,48,29,0,0,104,29,0,0,160,29,0,0,216,29,0,0,16,30,0,0,72,30,0,0,
  128,30,0,0,184,30,0,0,240,30,0,0,40,31,0,0,96,31,0,0,152,31,0,0,208,31,0,0,8,
  32,0,0,64,32,0,0,120,32,0,0,176,32,0,0,232,32,0,0,32,33,0,0,88,33,0,0,144,33,
  0,0,200,33,0,0,0,34,0,0,56,34,0,0,112,34,0,0,168,34,0,0,224,34,0,0,24,35,0,0,
  80,35,0,0,136,35,0,0,192,35,0,0,248,35,0,0,48,36,0,0,63,36,0,0,78,36,0,0,93,
  36,0,0,108,36,0,0,123,36,0,0,138,36,0,0,153,36,0,0,168,36,0,0,183,36,0,0,198,
  36,0,0,213,36,0,0,228,36,0,0,243,36,0,0,2,37,0,0,17,37,0,0,32,37,0,0,47,37,0,
  0,62,37,0,0,77,37,0,0,92,37,0,0,107,37,0,0,122,37,0,0,137,37,0,0,152,37,0,0,
  167,37,0,0,182,37,0,0,197,37,0,0,212,37,0,0,227,37,0,0,242,37,0,0,1,38,0,0,
  16,38,0,0,31,38,0,0,46,38,0,0,61,38,0,0,76,38,0,0,91,38,0,0,106,38,0,0,121,
  38,0,0,136,38,0,0,151,38,0,0,166,38,0,0,181,38,0,0,196,38,0,0,211,38,0,0,162,
  38,177,177,179,177,179,182,181,182,183,183,183,183,183,184,183,183,185,185,
  185,186,183,178,179,183,186,185,187,2,189,189,2,47,47,47,47,47,47,47,0,0,255,
  162,38,45,45,47,47,2,45,181,182,183,183,183,184,183,183,183,183,183,185,185,
  186,188,183,183,185,188,0,2,2,2,2,186,0,47,46,47,2,47,47,0,0,255,162,38,184,
  184,177,181,181,177,181,181,183,183,183,183,183,183,183,183,184,182,183,186,
  187,188,187,45,2,2,0,2,47,188,47,47,46,47,47,2,47,47,0,0,255,162,38,183,183,
  187,47,183,183,181,181,183,183,183,183,183,184,47,187,182,178,180,185,185,
  187,189,2,2,188,188,45,47,188,0,47,189,47,2,47,47,47,2,2,255,162,38,177,177,
  45,181,177,177,181,181,183,183,183,183,184,182,188,187,182,181,183,185,188,
  188,188,45,188,45,189,47,45,47,47,46,46,47,2,47,47,45,187,187,255,162,38,179,
  179,183,177,183,184,181,181,183,183,183,183,182,181,187,189,187,185,185,188,
  188,45,45,45,189,188,188,47,47,47,47,189,47,2,47,47,188,47,45,45,255,162,38,
  183,183,184,179,177,177,177,181,183,183,183,184,181,184,188,45,47,47,47,188,
  188,186,47,47,187,186,188,47,47,0,46,189,47,2,47,47,0,0,0,0,255,162,38,179,
  179,181,181,181,179,179,181,183,183,183,184,181,186,182,186,45,47,187,187,
  184,45,47,189,187,188,45,47,47,47,46,46,2,47,47,2,0,0,0,0,255,162,38,181,181,
  183,181,181,181,181,183,183,183,183,184,184,181,182,184,45,188,186,184,187,
  187,47,188,45,188,47,47,0,47,189,47,2,47,47,0,0,2,47,47,255,162,38,179,179,
  177,179,181,179,181,181,182,182,182,184,182,181,181,184,188,187,184,182,188,
  188,184,187,185,185,47,47,0,46,46,47,47,47,47,47,47,187,185,185,255,162,38,
  181,181,187,187,187,188,188,186,47,183,183,184,180,180,181,184,187,182,187,
  181,45,188,45,185,185,185,47,47,0,189,47,47,47,47,47,188,188,187,186,186,255,
  162,38,187,187,179,182,185,188,184,186,187,45,45,186,178,180,180,184,184,180,
  184,180,45,188,185,185,185,185,47,0,0,46,47,47,47,47,189,188,187,187,187,187,
  255,162,38,178,178,182,185,189,184,182,187,47,47,186,45,178,181,181,181,180,
  180,177,180,45,188,185,185,188,185,47,0,47,46,47,47,47,47,189,188,188,188,
  186,186,255,162,38,178,178,179,185,47,182,182,186,188,47,187,45,184,180,181,
  184,182,177,184,180,45,188,184,185,45,185,47,47,45,47,47,47,47,188,47,188,45,
  189,189,189,255,162,38,181,181,181,47,189,182,181,183,189,47,188,47,186,181,
  181,181,187,178,181,180,45,188,189,185,47,187,47,46,189,47,47,47,188,47,47,
  184,47,184,187,187,255,162,38,177,177,184,47,186,181,182,184,189,47,45,45,
  182,178,178,178,45,178,184,180,188,187,187,187,45,189,47,189,189,47,47,188,
  186,184,184,187,184,187,45,45,255,162,38,177,177,183,188,183,181,182,184,187,
  189,47,189,180,176,177,178,45,182,184,181,187,47,46,47,188,47,45,188,45,47,
  47,186,186,186,186,184,184,188,188,188,255,162,38,183,183,182,45,186,186,182,
  184,182,45,47,189,180,176,177,178,45,186,181,187,184,45,47,189,189,188,189,
  189,47,47,188,186,186,186,186,186,187,187,187,187,255,162,38,183,183,183,47,
  187,188,182,182,183,188,45,47,180,176,178,180,188,187,184,45,181,188,47,45,
  45,46,188,45,47,47,45,188,186,184,186,187,187,187,187,187,255,162,38,185,185,
  184,47,45,188,186,187,183,47,47,188,180,178,178,181,45,187,186,184,186,182,
  188,47,47,45,189,47,47,187,45,45,188,186,184,186,187,187,188,188,255,162,38,
  185,185,189,187,47,47,182,182,184,189,47,187,182,180,178,186,47,45,187,188,
  188,186,186,47,47,45,47,47,47,47,47,47,47,45,186,184,184,188,45,45,255,162,
  38,186,186,189,188,186,184,182,183,189,189,47,188,178,182,184,47,2,2,2,47,45,
  188,188,47,47,47,2,47,47,47,47,47,47,47,47,188,188,184,188,188,255,162,38,
  188,188,189,188,182,184,186,186,186,187,189,47,180,178,178,47,47,47,47,2,47,
  47,45,47,46,47,0,47,47,47,47,47,47,47,47,47,47,188,47,47,255,162,38,188,188,
  189,188,47,2,47,181,183,186,2,2,181,178,180,45,45,188,45,47,47,47,2,2,47,47,
  2,45,187,184,182,180,178,180,47,47,47,188,47,47,255,162,38,47,47,188,47,2,47,
  47,47,186,2,0,2,182,180,180,188,188,185,185,188,45,45,2,47,47,0,47,185,182,
  179,179,178,178,180,180,180,178,184,45,45,255,162,38,189,189,188,47,47,47,47,
  47,47,2,2,2,186,181,181,187,186,181,182,183,187,189,185,47,47,2,183,185,180,
  179,179,180,180,180,182,182,179,178,45,45,255,162,38,189,189,47,47,47,47,47,
  2,47,47,47,47,45,184,182,184,188,182,177,179,182,184,186,47,0,47,189,188,179,
  180,180,183,187,182,185,185,180,179,186,186,255,162,38,189,189,47,47,47,47,
  46,45,46,45,47,47,46,187,188,187,188,183,177,180,183,185,185,2,2,185,186,189,
  188,185,184,187,180,184,188,189,186,181,186,186,255,162,38,45,45,189,47,47,
  47,46,47,2,2,47,47,45,46,184,188,45,185,183,183,184,184,186,183,179,183,185,
  187,184,187,184,180,182,187,189,45,45,189,185,185,255,162,38,187,187,45,47,
  47,189,45,47,47,46,45,189,185,184,180,180,180,176,185,185,185,185,183,179,
  177,183,185,183,182,187,185,187,189,186,45,45,47,47,45,45,255,162,38,186,186,
  188,45,187,45,47,47,47,47,188,188,183,184,180,180,180,176,179,182,187,186,
  185,183,183,184,187,185,180,185,184,180,181,187,45,189,189,47,47,47,255,162,
  38,186,186,186,189,189,47,47,189,187,188,184,184,182,180,180,180,176,176,181,
  184,187,184,189,185,185,186,189,189,183,183,184,187,189,189,47,189,47,47,189,
  189,255,162,38,186,186,186,189,45,47,188,187,188,187,186,186,181,180,178,176,
  176,177,182,181,184,184,184,185,186,186,187,189,180,186,186,181,185,47,189,
  47,47,187,187,187,255,162,38,185,185,189,188,189,47,47,189,47,47,189,187,186,
  182,181,178,176,177,185,185,188,184,186,186,186,187,188,189,182,189,47,47,
  189,47,47,187,188,187,186,186,255,162,38,189,189,189,189,47,187,45,47,188,
  188,47,187,189,187,186,182,176,180,185,185,187,184,187,187,186,188,45,47,47,
  47,47,47,2,47,189,187,189,187,187,187,255,162,38,45,45,189,189,188,186,47,47,
  189,186,188,187,185,189,188,186,177,181,188,189,186,187,45,45,188,45,45,47,
  47,47,47,47,47,2,189,187,188,186,187,187,255,162,38,45,45,45,187,185,185,47,
  47,189,187,184,189,187,185,189,188,177,181,188,45,186,184,184,188,47,47,47,
  47,47,47,47,2,2,47,47,189,189,187,186,186,255,162,38,45,45,45,189,188,186,47,
  47,47,188,185,186,189,185,185,180,177,182,188,45,184,181,181,187,47,47,47,47,
  47,47,47,2,47,47,47,189,188,188,186,186,255,162,38,45,45,189,188,189,187,46,
  47,47,188,187,185,189,187,183,177,177,183,188,187,184,181,184,187,47,47,47,
  47,47,47,47,47,2,47,189,2,189,188,187,187,255,162,38,187,187,187,188,189,189,
  189,46,187,184,189,188,47,184,180,177,178,184,47,186,184,187,188,188,47,47,
  46,47,47,47,47,47,47,47,47,2,47,189,186,186,255,162,38,184,184,186,188,189,
  45,188,47,187,183,187,47,47,181,177,177,179,185,187,188,186,184,186,186,188,
  47,47,47,47,47,47,47,47,47,46,45,189,187,186,186,255,162,38,184,184,186,188,
  189,47,47,189,188,185,189,47,183,179,177,177,180,185,187,47,187,182,186,182,
  188,189,184,47,47,47,47,47,47,47,47,47,47,189,186,186,255,162,38,184,184,186,
  184,184,47,47,47,189,188,47,47,180,177,177,177,180,185,187,47,188,187,182,
  182,188,45,189,47,47,45,47,46,47,47,47,2,189,189,186,186,255,162,38,185,185,
  184,181,186,185,184,184,188,47,47,47,0,44,44,178,180,185,187,45,45,184,184,
  186,45,46,47,47,47,46,47,47,47,47,47,47,47,188,187,187,255,162,38,183,183,
  185,185,184,181,179,178,178,178,181,185,0,40,184,40,181,186,187,188,45,178,
  184,188,47,47,189,45,47,47,47,47,47,47,2,47,47,189,186,186,255,162,38,178,
  178,180,184,186,181,177,177,178,178,179,44,184,177,181,40,40,186,187,188,45,
  187,187,47,47,47,45,47,47,47,46,47,2,2,2,47,189,188,187,187,255,162,38,179,
  179,180,182,184,185,178,177,178,179,44,185,177,177,181,185,40,186,189,188,47,
  186,188,47,45,47,47,46,46,45,47,47,47,47,2,47,189,188,189,189,255,162,38,182,
  182,183,184,186,183,182,178,178,44,185,0,177,177,181,185,40,187,188,188,45,
  47,47,47,47,47,47,46,47,47,47,47,47,47,47,47,189,189,187,187,255,162,38,184,
  184,184,185,184,184,184,182,180,44,184,177,177,177,184,181,184,184,184,184,
  184,181,181,177,177,184,47,47,47,177,177,177,2,2,0,2,47,189,186,186,255,162,
  38,177,177,178,180,182,184,181,181,181,184,44,184,177,177,177,184,177,177,
  177,177,181,177,177,177,177,177,184,47,177,177,177,185,181,47,2,47,189,187,
  186,186,255,162,38,179,179,179,183,183,184,184,184,184,184,44,184,177,177,
  177,177,177,177,177,177,184,177,177,177,177,177,177,177,177,177,185,181,181,
  0,0,189,2,187,186,186,255,162,38,180,180,180,183,183,186,179,177,184,184,184,
  184,177,181,177,181,177,177,177,185,177,177,177,177,177,182,177,177,177,185,
  181,181,39,47,47,47,47,187,0,0,255,162,38,182,182,183,184,186,183,178,178,
  179,44,184,184,181,181,181,177,177,181,185,185,39,185,181,177,177,184,177,
  177,185,185,45,39,39,47,47,47,47,47,47,47,255,162,38,184,184,184,186,186,186,
  186,183,183,184,44,184,181,181,181,181,181,40,40,40,40,39,185,181,177,177,
  184,184,185,189,189,189,45,47,45,47,47,47,47,47,255,162,38,186,186,186,189,
  47,47,2,189,182,184,44,184,181,181,181,181,44,41,41,40,40,40,39,181,177,181,
  184,184,184,186,188,188,189,45,188,45,47,47,181,181,255,162,38,2,2,2,2,2,47,
  47,181,182,44,185,181,181,184,181,181,44,44,41,41,40,40,39,185,177,177,177,
  177,177,181,185,185,188,45,188,189,47,181,181,181,255,162,38,47,47,2,2,2,2,
  181,182,184,184,44,184,184,184,181,181,44,44,41,41,40,39,185,177,177,177,177,
  177,177,177,181,185,185,189,184,177,177,181,39,39,255,162,38,47,47,47,47,2,2,
  184,184,47,2,47,47,189,184,184,184,44,44,41,41,40,185,177,177,177,177,177,
  177,177,177,177,40,181,177,177,177,181,39,39,39,255,162,38,47,47,47,47,2,2,2,
  2,47,47,2,2,45,45,46,189,189,44,41,41,39,181,177,177,177,177,177,177,177,177,
  177,177,177,177,181,181,39,44,44,44,255,162,38,47,47,2,2,2,2,2,2,2,47,47,2,
  47,45,45,45,47,47,44,41,181,177,177,177,177,177,177,180,180,189,189,189,45,
  47,47,47,189,187,186,186,255,162,38,2,2,2,2,2,2,2,47,47,2,47,47,47,47,47,47,
  47,46,47,185,177,177,177,177,180,180,2,236,140,137,138,142,236,237,238,238,
  79,14,142,142,255,162,38,47,47,2,2,2,2,2,2,2,47,47,47,47,47,47,45,47,47,47,
  185,177,177,180,181,182,183,236,140,138,139,141,143,236,237,238,238,238,238,
  79,79,255,162,38,47,47,47,2,2,2,2,2,2,47,47,47,47,45,45,45,46,47,47,185,177,
  177,180,182,183,14,140,138,139,141,142,143,236,237,238,238,238,79,79,79,255,
  162,38,2,2,47,2,2,2,2,2,2,2,47,47,47,188,189,45,47,47,47,185,181,177,180,185,
  37,140,139,139,140,142,143,236,237,15,15,15,15,15,79,79,255,162,38,2,2,2,2,2,
  2,47,2,2,2,2,47,47,189,45,47,2,47,47,47,182,177,180,40,40,140,140,140,142,14,
  14,237,15,14,15,15,15,15,15,15,255,162,38,189,189,45,47,47,47,47,0,47,47,47,
  47,47,47,47,2,0,45,188,188,182,182,39,137,134,132,132,135,136,138,142,14,14,
  15,15,15,79,15,15,15,255,162,38,45,45,47,47,47,47,47,47,47,47,47,47,47,47,47,
  2,0,45,188,187,182,183,137,121,121,123,124,125,124,123,138,138,15,79,79,15,
  15,79,79,79,255,162,38,45,45,47,47,0,47,47,47,47,47,47,47,47,47,0,2,0,45,188,
  39,183,122,120,119,121,121,123,124,125,127,125,124,139,14,79,79,79,79,79,79,
  255,162,38,45,45,47,0,47,47,47,47,47,47,47,47,47,47,2,2,45,45,187,39,121,119,
  118,120,121,123,123,124,124,125,127,126,125,141,14,5,5,79,79,79,255,162,38,
  45,45,47,47,47,47,47,47,47,0,47,47,47,47,187,45,45,188,187,121,118,119,120,
  121,123,123,124,125,125,125,126,127,127,125,142,14,79,5,5,5,255,162,38,45,45,
  45,47,47,47,47,0,0,0,0,47,187,181,187,45,45,189,121,118,119,120,121,122,123,
  124,125,125,125,126,126,127,127,127,5,141,142,15,79,79,255,162,38,47,47,2,
  187,188,188,188,186,184,182,184,187,181,183,179,177,185,188,116,119,120,121,
  122,123,124,125,125,126,126,126,127,127,126,126,127,5,14,14,15,15,255,162,38,
  187,187,187,185,189,0,0,2,0,47,187,187,184,179,181,180,185,121,118,120,121,
  122,123,123,124,125,125,125,126,127,127,127,126,126,126,6,6,5,79,79,255,162,
  38,187,187,179,179,185,181,181,182,182,184,45,181,181,183,179,179,180,118,
  120,121,122,123,124,124,125,125,125,127,127,127,127,127,126,126,126,126,6,6,
  6,6,255,162,38,181,181,188,187,47,47,2,187,183,185,179,179,179,177,177,185,
  118,119,120,122,123,124,124,124,125,125,124,126,127,127,126,127,126,127,126,
  126,126,126,126,126,255,162,38,188,188,183,179,179,179,181,187,184,184,177,
  185,184,185,45,120,119,119,121,122,124,124,124,124,124,124,124,123,126,127,
  127,127,127,126,126,127,126,125,125,125,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
  255,255,255,255,255,255,255,255,255,255,255,255,255,149,51,14,14,9,9,9,9,9,
  125,5,5,11,8,8,8,2,185,181,181,182,179,179,181,183,185,189,188,185,185,185,
  182,35,73,72,69,67,163,60,210,209,208,160,227,227,226,226,226,226,226,227,
  160,226,226,226,255,149,51,9,9,9,9,9,9,9,110,126,5,8,1,8,8,2,180,181,182,181,
  179,179,181,183,186,45,187,184,184,187,41,15,73,69,67,164,163,162,211,210,
  209,209,210,210,210,210,209,209,160,160,160,227,226,226,255,149,51,9,9,125,
  125,125,125,125,79,5,5,8,8,8,2,185,180,179,182,179,178,179,181,183,186,45,
  188,185,183,187,43,76,165,164,164,164,163,163,61,58,57,58,56,54,53,210,210,
  209,160,160,160,160,160,160,255,149,51,125,125,125,125,125,126,11,1,10,5,8,8,
  8,2,181,181,181,182,179,178,180,182,184,189,45,189,186,184,185,185,166,165,
  165,165,68,66,64,131,131,60,130,129,128,53,53,210,210,209,160,227,160,160,
  160,255,149,51,125,125,125,125,126,110,10,5,5,8,1,8,2,2,181,181,182,181,180,
  178,180,183,185,45,189,189,186,183,182,183,165,165,166,166,71,69,65,144,131,
  144,132,130,128,54,53,53,210,209,160,227,160,160,160,255,149,51,110,110,126,
  126,110,79,141,6,5,8,8,8,47,186,182,182,182,182,179,179,182,184,45,187,188,
  187,187,185,187,185,166,167,36,36,73,72,69,145,144,67,144,131,129,128,54,55,
  54,210,160,160,160,160,160,255,149,51,110,110,10,10,110,126,142,10,8,5,8,8,
  181,182,183,184,184,184,183,182,183,45,187,185,186,187,186,187,185,185,183,
  182,182,186,36,36,72,69,66,67,144,132,130,129,129,129,128,52,160,160,227,160,
  160,255,149,51,10,10,110,10,126,79,15,126,8,8,8,181,181,181,184,184,184,185,
  185,185,45,187,185,182,183,187,185,186,183,185,185,186,183,182,34,36,73,71,
  71,69,64,144,131,131,132,144,129,54,160,160,227,128,128,255,149,51,110,110,
  10,5,110,5,79,15,8,8,179,179,180,182,182,183,184,185,187,185,188,185,183,185,
  185,185,185,185,186,184,185,186,186,183,31,36,36,74,15,71,67,67,144,132,134,
  144,129,55,161,160,211,66,66,255,149,51,10,10,110,5,126,5,5,6,5,79,178,179,
  179,181,180,182,183,185,187,186,185,184,184,185,186,186,185,185,186,185,184,
  186,185,184,184,186,188,77,45,70,69,69,65,145,136,144,128,55,161,160,56,60,
  60,255,149,51,111,111,110,110,126,8,1,6,5,79,15,179,181,179,180,180,183,185,
  187,187,184,184,184,185,185,187,187,185,185,186,185,186,185,184,186,188,188,
  189,188,70,69,70,69,146,147,145,58,57,161,161,57,59,59,255,149,51,10,10,111,
  5,5,5,1,5,5,79,15,178,178,178,179,180,183,185,187,186,184,183,183,185,185,
  186,187,187,186,185,185,186,185,186,189,47,47,47,45,72,71,71,70,148,148,69,
  64,62,162,161,59,25,25,255,149,51,5,5,111,126,9,7,5,5,110,110,15,178,177,178,
  178,180,183,186,188,185,183,183,183,184,185,186,187,188,187,185,185,186,186,
  189,189,47,47,47,47,74,72,73,72,73,73,71,66,65,162,161,61,68,68,255,149,51,
  10,10,5,1,8,8,8,8,110,110,15,14,177,178,179,181,183,188,187,183,182,182,182,
  183,185,185,186,187,188,186,186,186,189,45,45,188,188,47,187,41,43,77,77,75,
  73,71,68,65,164,62,61,26,26,255,149,51,5,5,1,10,8,8,8,8,78,15,238,15,178,179,
  180,181,185,188,186,182,180,180,181,183,185,185,185,186,188,188,187,189,45,
  45,45,188,187,47,186,188,45,45,45,77,75,71,68,65,164,164,61,26,26,255,149,51,
  10,10,5,10,107,8,8,8,15,15,15,14,178,179,180,181,185,188,186,182,180,180,181,
  183,185,185,185,186,188,188,187,45,46,45,188,187,187,187,186,188,45,45,45,45,
  77,72,70,66,165,165,62,26,26,255,149,51,5,5,79,107,159,157,108,108,15,79,15,
  74,180,180,180,182,187,187,186,186,182,181,181,183,185,185,185,185,188,188,
  189,45,45,47,47,47,47,45,187,188,45,45,45,45,45,77,76,68,165,165,64,43,43,
  255,149,51,1,1,3,94,3,10,5,5,78,238,142,179,180,180,182,183,186,186,186,185,
  185,185,184,185,186,187,186,185,188,189,45,45,189,47,189,189,45,189,188,187,
  185,45,189,189,189,42,188,186,234,232,184,30,30,255,149,51,107,107,91,106,78,
  6,110,110,110,15,47,178,180,182,183,185,186,184,181,181,183,185,185,187,187,
  189,187,187,188,189,188,189,189,189,45,47,47,47,45,46,47,45,45,45,189,188,
  188,186,233,232,184,184,184,255,149,51,91,91,78,0,7,5,78,5,78,238,180,179,
  180,182,184,187,184,183,182,182,181,182,182,184,185,187,189,189,189,189,188,
  189,189,45,189,189,45,189,187,186,186,45,45,189,189,188,186,186,222,221,184,
  185,185,255,149,51,138,138,0,7,110,78,78,15,79,182,181,179,180,182,185,187,
  186,185,185,184,183,182,182,182,182,185,187,186,186,189,186,44,189,45,45,47,
  189,187,187,187,189,45,45,45,45,188,186,189,182,180,184,185,185,255,149,51,0,
  0,0,0,5,15,9,79,180,183,180,180,181,182,185,187,187,186,185,185,185,183,183,
  183,184,184,184,186,188,186,189,186,47,189,189,189,189,189,187,47,187,187,45,
  45,47,188,186,189,105,110,184,185,185,255,149,51,5,5,101,101,108,110,9,15,
  179,183,180,179,180,180,184,184,184,183,183,183,182,182,183,184,185,186,186,
  185,186,188,44,47,47,189,189,47,189,189,45,45,189,45,45,189,189,189,189,188,
  188,46,47,183,183,255,149,51,101,101,91,82,98,106,143,180,178,181,181,179,
  180,182,184,187,188,47,185,184,183,182,182,182,185,185,186,187,185,184,185,
  184,184,187,187,47,47,45,45,45,45,47,45,45,189,189,189,188,189,184,183,183,
  183,255,149,51,82,82,134,102,153,153,141,180,178,180,180,180,181,184,185,184,
  184,47,47,185,185,183,183,183,184,184,185,185,186,182,184,185,186,184,186,
  187,45,45,45,47,47,47,45,45,189,188,188,189,187,184,183,183,183,255,149,51,9,
  9,78,2,2,78,108,37,178,179,182,180,181,183,182,180,187,45,47,185,185,184,182,
  182,183,184,185,185,185,182,185,184,185,183,185,186,187,45,47,47,47,47,47,
  189,189,186,186,186,184,184,182,182,182,255,149,51,2,2,0,0,0,2,143,2,178,178,
  181,179,181,180,179,179,189,188,45,184,185,184,181,181,182,184,185,184,182,
  184,182,184,184,185,183,185,186,187,47,47,188,188,188,189,187,185,185,184,
  182,180,181,182,182,255,149,51,15,15,2,2,108,108,79,2,187,178,181,179,180,
  179,179,178,45,184,47,183,183,183,182,182,183,183,184,182,183,183,183,182,
  183,183,184,184,184,184,187,188,186,186,186,188,185,185,184,182,180,179,181,
  182,182,255,149,51,110,110,238,238,79,238,2,2,2,178,180,179,179,178,178,178,
  182,187,182,182,180,181,182,182,183,184,184,183,184,184,183,183,182,183,183,
  183,181,182,184,184,188,185,188,185,185,184,182,180,178,178,180,181,181,255,
  149,51,106,106,79,0,2,2,2,2,2,187,179,179,178,178,179,179,179,181,182,182,
  181,182,181,182,184,186,186,186,185,184,184,183,182,182,183,183,182,181,182,
  183,185,185,185,184,184,182,182,180,178,175,178,182,182,255,149,51,106,106,
  79,79,2,2,2,2,2,2,179,180,179,178,179,179,182,183,184,183,181,183,181,185,
  186,186,186,186,186,185,185,185,185,185,185,185,183,182,181,182,184,185,185,
  182,182,182,180,177,175,174,178,183,183,255,149,51,78,78,79,79,2,2,2,2,2,2,
  178,180,179,179,179,181,184,45,186,183,181,183,183,186,186,187,187,187,185,
  185,184,185,184,185,184,184,184,183,181,181,183,184,185,182,182,180,179,175,
  174,174,178,183,183,255,149,51,0,0,79,79,2,2,2,2,2,2,178,179,180,179,180,181,
  189,45,182,179,182,185,186,186,185,183,182,182,182,182,182,182,182,182,182,
  181,181,180,180,180,182,184,185,182,180,179,178,175,174,174,181,188,188,255,
  149,51,108,108,79,2,2,2,2,2,2,2,180,178,180,180,181,183,46,45,177,185,186,
  185,186,186,186,185,184,184,183,184,184,187,187,186,184,184,186,181,182,179,
  181,185,185,182,180,179,177,176,174,182,187,187,187,255,149,51,79,79,2,2,2,2,
  2,2,2,2,181,178,179,180,181,184,188,46,182,45,46,186,186,187,187,185,182,186,
  184,183,187,187,188,189,186,186,189,181,182,179,180,185,186,182,180,179,177,
  174,182,186,187,187,187,255,149,51,2,2,2,2,2,2,2,2,2,2,181,180,178,179,181,
  183,185,188,45,46,189,185,185,187,186,180,184,186,187,188,189,186,184,183,
  184,179,189,185,182,179,182,186,186,182,180,180,180,182,185,187,186,187,187,
  255,149,51,2,2,2,2,2,2,2,2,2,2,187,180,180,178,180,181,183,187,46,189,187,
  185,185,186,185,180,185,185,187,189,189,186,184,182,182,184,187,189,187,182,
  184,187,186,184,182,180,183,185,186,185,185,186,186,255,149,51,2,2,2,2,2,2,2,
  2,2,2,2,181,181,181,182,183,186,189,46,47,185,185,186,189,181,185,185,186,
  186,189,189,187,184,182,181,183,183,186,47,183,185,187,186,184,184,183,185,
  184,185,184,185,187,187,255,149,51,2,2,2,2,2,2,2,2,2,2,2,188,182,182,183,185,
  185,189,46,2,189,189,186,184,186,187,189,186,186,189,189,187,186,182,183,183,
  183,182,47,184,187,187,188,182,184,183,185,185,185,185,187,187,187,255,149,
  51,2,2,2,47,2,2,2,2,2,2,2,2,188,183,184,184,182,182,47,0,181,185,189,184,188,
  189,187,187,186,189,187,186,186,184,183,181,183,181,177,177,177,179,180,181,
  182,183,185,184,185,185,187,188,188,255,149,51,47,47,2,2,47,47,2,2,2,2,0,2,0,
  47,47,47,47,47,0,2,185,181,184,183,187,187,187,187,186,187,186,185,186,184,0,
  0,40,177,177,177,179,179,180,181,181,186,184,184,186,185,187,188,188,255,149,
  51,47,47,47,47,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0,0,0,189,190,187,189,188,189,
  187,189,184,182,183,185,0,0,185,40,176,176,177,179,179,180,183,186,184,184,
  186,185,187,188,188,255,149,51,2,2,2,47,2,2,2,2,2,2,2,0,2,0,0,0,0,0,0,0,0,
  190,190,190,189,188,186,186,189,188,184,183,183,184,44,44,184,40,176,177,177,
  177,177,180,183,185,185,185,186,185,185,187,187,255,149,51,47,47,47,2,2,2,2,
  2,2,2,2,2,2,2,188,0,0,0,0,0,0,190,189,190,190,187,186,186,189,188,185,184,
  184,184,44,40,184,185,40,177,177,177,177,181,183,184,184,185,186,185,182,186,
  186,255,149,51,2,2,2,47,47,0,0,2,0,0,0,2,2,2,2,0,0,0,0,0,0,190,190,190,190,
  187,187,186,186,187,0,0,0,44,44,184,177,185,40,177,177,177,179,181,184,184,
  185,185,186,185,177,185,185,255,149,51,2,2,2,47,47,2,0,0,0,0,2,185,2,2,188,0,
  0,0,0,0,0,190,189,190,47,190,47,186,188,185,176,183,45,44,40,177,177,184,40,
  177,177,177,180,180,183,182,184,185,185,185,177,185,185,255,149,51,2,2,2,0,
  47,47,2,0,0,0,2,2,2,180,187,0,0,0,0,0,190,47,190,190,47,190,190,186,176,181,
  187,188,44,40,0,177,177,184,39,178,177,179,179,180,180,182,184,185,185,182,
  177,186,186,255,149,51,2,2,2,2,2,2,2,0,0,0,185,185,182,2,0,0,0,190,190,187,
  190,190,190,190,190,190,186,176,181,183,179,177,44,184,181,177,177,184,177,
  39,40,178,179,179,180,182,183,184,185,177,177,186,186,255,149,51,2,2,2,2,0,2,
  2,2,2,0,185,45,0,185,0,0,190,187,184,185,189,190,190,190,190,47,181,183,184,
  179,180,44,44,184,177,177,177,181,177,184,39,40,40,40,180,182,184,184,185,
  177,181,186,186,255,149,51,2,2,2,2,2,2,2,2,2,2,2,0,2,0,181,190,188,184,179,
  179,185,181,188,190,45,184,45,181,181,180,40,44,40,184,177,177,177,177,184,
  177,184,185,39,40,40,40,40,185,182,177,181,185,185,255,149,51,2,2,2,2,2,2,0,
  2,0,2,2,2,45,45,182,181,186,181,188,187,47,47,0,187,183,185,179,179,179,180,
  184,180,40,185,184,181,177,177,177,177,181,184,184,184,185,39,40,40,177,177,
  181,185,185,255,149,51,0,0,2,2,2,2,2,2,2,2,2,187,45,178,45,184,181,188,183,
  179,179,179,181,187,184,184,177,185,184,185,180,184,40,37,184,181,177,177,
  177,177,177,177,181,181,184,184,185,40,177,177,181,185,185,255,149,51,0,0,0,
  2,2,47,0,2,187,187,182,185,187,45,184,186,188,179,181,183,183,189,47,179,181,
  189,183,177,177,179,180,40,181,184,182,180,177,177,177,177,177,177,177,177,
  177,181,184,183,177,177,181,186,186,255,149,51,2,2,2,2,0,2,0,2,47,45,185,45,
  185,186,180,45,186,183,183,185,183,184,183,181,181,189,180,179,184,189,183,
  187,47,184,182,180,180,177,177,177,177,177,177,177,181,181,177,177,177,179,
  186,186,186,255,149,51,2,2,187,187,187,183,183,45,45,188,179,185,183,45,186,
  186,183,181,183,177,176,176,177,181,183,185,179,189,179,181,177,45,181,181,
  184,182,182,180,177,177,177,177,177,181,184,177,214,177,177,214,186,186,186,
  255,149,51,47,47,0,2,47,187,184,2,185,2,179,181,184,186,45,186,181,179,177,
  184,183,181,179,179,177,177,45,181,179,179,179,183,177,185,185,185,184,182,
  182,177,177,177,181,184,231,177,231,231,231,231,186,185,185,255,149,51,185,
  185,47,47,47,183,2,188,184,184,182,183,183,177,181,185,181,181,184,187,184,
  188,184,185,187,184,45,181,179,181,183,184,179,181,181,181,182,40,184,184,
  184,184,184,184,177,177,177,177,181,214,186,185,185,255,149,51,183,183,2,2,
  188,185,2,182,47,189,185,183,181,185,187,45,45,183,177,179,179,176,176,179,
  177,179,181,45,177,179,179,181,181,183,182,182,181,181,40,39,185,185,184,177,
  177,177,177,183,183,186,186,184,184,255,149,51,186,186,0,182,187,185,187,185,
  47,182,177,181,181,179,187,45,45,179,179,179,179,179,179,179,176,176,177,47,
  176,181,181,183,181,183,183,182,181,181,40,40,39,185,181,177,177,177,214,231,
  231,214,40,184,184,255,149,51,47,47,47,180,47,183,47,181,47,182,183,183,181,
  181,181,187,183,179,183,184,183,184,183,181,185,181,47,188,179,179,179,177,
  179,183,182,184,184,183,184,40,40,39,184,177,177,179,231,179,184,231,40,184,
  184,255,149,51,187,187,185,180,187,187,188,181,45,179,183,181,181,183,183,
  184,181,181,181,177,179,177,177,181,177,188,185,176,179,179,181,181,180,180,
  182,184,186,183,186,40,40,40,185,181,214,179,231,179,184,231,40,185,185,255,
  149,51,47,47,183,183,185,188,178,188,188,183,184,183,183,184,181,179,179,177,
  177,176,177,176,177,176,188,0,181,183,179,177,181,179,180,180,184,184,186,
  184,185,40,40,40,39,184,231,231,231,231,231,231,40,183,183,255,149,51,181,
  181,185,188,185,184,183,45,188,181,183,181,181,179,179,184,184,188,45,188,45,
  47,45,47,0,0,177,181,181,181,179,177,180,182,184,184,185,186,186,40,40,40,39,
  184,177,180,180,181,184,214,40,182,182,255,149,51,185,185,184,189,0,187,184,
  45,182,181,181,183,184,184,184,183,181,179,177,177,179,179,179,181,183,185,
  177,181,181,179,179,179,182,185,187,47,183,185,186,40,40,40,39,184,177,180,
  180,181,37,40,186,180,180,255,149,51,187,187,184,189,2,184,183,189,178,184,
  185,180,183,183,181,181,185,185,177,181,181,181,186,179,181,187,177,177,179,
  179,179,181,184,45,185,47,184,184,185,40,40,40,39,185,177,180,180,181,39,40,
  185,177,177,255,149,51,183,183,184,45,47,181,183,186,183,183,184,183,181,181,
  185,189,188,187,45,185,183,183,179,179,183,185,177,181,183,179,179,187,45,
  187,189,189,180,181,181,40,40,40,39,185,177,213,231,213,39,231,185,179,179,
  255,149,51,186,186,185,181,45,181,186,47,183,184,184,186,181,189,188,177,177,
  176,176,176,179,183,181,183,185,188,177,179,179,179,185,45,187,185,187,185,
  183,181,182,40,40,40,39,185,177,231,182,231,40,231,185,179,179,255,149,51,
  186,186,47,188,2,183,47,178,179,179,181,183,177,177,177,188,187,188,181,45,
  179,183,181,181,181,182,179,177,179,179,47,187,184,185,184,185,179,179,180,
  40,40,40,39,185,184,231,183,213,231,215,39,177,177,255,149,51,189,189,185,
  184,186,45,45,184,184,183,181,185,188,188,183,181,188,178,188,188,183,184,
  183,183,184,183,45,177,179,47,2,179,179,45,179,185,178,179,181,40,40,39,185,
  184,181,181,184,37,41,44,39,176,176,255,149,51,2,2,187,187,188,47,187,185,
  181,183,184,189,185,181,188,184,184,183,45,188,181,183,181,181,177,181,179,
  181,189,2,47,177,181,187,183,189,180,181,181,40,39,185,185,177,177,218,231,
  231,231,214,40,176,176,255,149,51,45,45,179,186,188,2,187,183,184,186,185,
  189,189,184,186,183,187,184,45,182,181,181,183,184,184,184,179,177,176,177,
  47,178,184,189,182,188,181,184,181,185,184,181,177,177,177,231,175,180,37,
  231,44,177,177,255,149,51,45,45,185,186,187,2,185,181,183,184,186,47,189,181,
  181,183,184,183,189,178,184,185,180,183,177,177,179,185,45,2,47,177,184,183,
  180,185,183,181,181,185,181,177,177,177,177,214,231,231,231,214,41,177,177,
  255,149,51,45,45,181,188,189,2,184,181,183,183,186,47,181,183,181,183,181,
  183,186,183,183,184,183,181,179,179,179,177,177,179,47,183,183,186,183,189,
  186,181,178,181,177,177,177,177,177,177,177,177,177,182,40,176,176,255,149,
  51,2,2,183,189,188,187,188,184,183,184,188,185,181,183,184,186,181,186,47,
  183,184,184,186,181,181,179,183,189,45,47,45,184,183,186,181,189,186,183,180,
  181,180,177,177,177,177,231,231,231,231,231,37,176,176,255,149,51,189,189,
  181,183,189,187,188,183,184,183,186,185,184,186,183,181,183,47,178,179,179,
  181,183,177,179,181,181,181,181,176,2,184,184,183,180,187,184,181,178,181,
  182,177,177,177,177,231,177,231,177,177,37,177,177,255,149,51,185,185,189,
  188,187,187,189,183,184,184,185,185,184,181,183,184,45,45,184,184,183,181,
  183,188,177,177,179,183,184,181,47,184,181,188,180,179,187,179,177,179,180,
  180,177,184,180,180,177,177,177,177,181,177,177,255,149,51,188,188,189,47,
  182,187,45,179,184,186,189,189,187,184,184,183,185,188,185,180,181,179,177,
  177,179,181,183,184,185,47,0,185,181,45,183,181,184,183,178,179,179,182,177,
  181,184,231,177,177,177,177,177,176,176,255,149,51,184,184,186,47,185,189,47,
  181,186,184,45,185,183,183,185,185,183,187,189,47,179,177,179,179,179,177,
  177,177,176,176,47,187,180,189,179,181,183,183,179,179,179,39,181,39,181,231,
  231,231,231,231,177,176,176,255,149,51,187,187,188,2,185,184,45,188,185,184,
  186,187,186,185,183,184,186,183,182,183,181,183,181,183,181,183,185,189,47,2,
  187,45,184,183,183,181,183,189,178,177,182,39,181,39,181,231,184,184,181,177,
  177,176,176,255,149,51,185,185,188,47,185,188,189,187,189,184,45,188,183,184,
  186,181,181,184,47,187,183,181,181,183,184,181,181,179,179,181,2,47,181,185,
  183,181,181,183,179,182,182,39,181,39,181,181,181,181,181,181,177,176,176,
  255,149,51,189,189,185,47,185,187,186,45,181,188,177,183,183,181,181,183,183,
  189,179,179,179,185,179,177,179,179,179,177,177,179,184,47,184,181,188,183,
  185,183,181,182,182,39,183,39,181,161,231,231,231,218,184,176,176,255,149,51,
  47,47,47,2,184,185,188,47,179,185,180,183,183,188,189,181,177,183,181,181,
  184,177,179,179,179,185,181,187,45,181,179,45,187,181,47,180,181,181,177,182,
  182,39,183,43,181,177,176,175,177,231,178,177,177,255,190,10,205,205,43,181,
  176,176,161,231,177,178,177,177,255,190,10,205,205,43,181,177,176,175,177,
  231,176,176,176,255,190,10,205,205,0,181,161,231,231,231,218,177,176,176,255,
  190,10,205,205,183,181,176,175,176,176,176,177,178,178,255,190,10,205,205,
  180,178,214,231,231,231,231,176,178,178,255,190,10,205,205,176,177,161,177,
  231,176,176,177,178,178,255,190,10,205,205,177,177,214,231,231,231,231,177,
  178,178,255,190,10,205,205,176,177,176,176,176,176,176,176,178,178,255,190,
  10,205,205,177,177,231,231,231,231,231,176,178,178,255,190,10,205,205,176,
  177,231,176,231,176,176,176,176,176,255,190,10,205,205,176,177,214,231,176,
  231,231,176,176,176,255,190,10,205,205,176,176,176,176,176,176,176,176,176,
  176,255,190,10,205,205,176,176,231,231,231,231,231,176,176,176,255,190,10,
  205,205,176,176,161,176,161,176,161,176,176,176,255,190,10,205,205,176,176,
  176,176,176,176,176,176,176,176,255,190,10,205,205,176,176,176,176,176,176,
  231,231,176,176,255,190,10,205,205,176,176,176,176,176,176,176,176,176,176,
  255,190,10,205,205,178,176,176,181,176,176,176,176,176,176,255,190,10,205,
  205,178,178,179,176,176,176,176,176,177,177,255,190,10,205,205,185,218,231,
  231,231,231,231,218,177,177,255,190,10,205,205,185,231,178,176,176,176,176,
  231,177,177,255,190,10,205,205,186,231,177,231,231,231,177,231,180,180,255,
  190,10,205,205,183,231,176,231,176,231,176,231,177,177,255,190,10,205,205,
  182,231,177,231,177,231,177,231,177,177,255,190,10,205,205,178,231,182,178,
  178,178,178,231,177,177,255,190,10,205,205,182,223,231,231,231,231,231,223,
  178,178,255,190,10,205,205,187,189,188,182,182,180,180,180,184,184,255,190,
  10,205,205,188,189,189,231,185,182,231,182,185,185,255,190,10,205,205,2,188,
  231,231,231,231,231,185,187,187,255,190,10,205,205,188,2,2,190,189,189,231,
  186,188,188,255,190,10,205,205,2,2,212,231,212,190,189,189,189,189,255,190,
  10,205,205,189,2,231,233,231,2,190,2,2,2,255,190,10,205,205,0,2,231,2,231,0,
  0,0,2,2,255,190,10,205,205,189,2,231,231,231,231,231,2,2,2,255,190,10,205,
  205,2,2,2,0,190,2,190,2,2,2,255,190,10,205,205,0,2,214,231,214,2,0,2,2,2,255,
  190,10,205,205,2,2,231,233,231,0,0,0,2,2,255,190,10,205,205,2,2,231,2,231,2,
  2,0,0,0,255,190,10,205,205,2,2,231,231,231,231,231,2,2,2,255,190,10,205,205,
  0,0,0,2,0,0,0,0,2,2,255,190,10,205,205,0,0,219,0,2,2,219,0,2,2,255,190,10,
  205,205,0,0,231,2,2,0,231,0,0,0,255,190,10,205,205,0,0,231,0,231,0,231,2,2,2,
  255,190,10,205,205,0,0,231,231,2,231,231,2,2,2,255,190,10,2,2,0,2,47,47,47,
  47,2,2,2,2,255,190,10,2,2,2,47,2,47,47,47,47,2,2,2,255,
};

#endif /* DOGS */
// 2023/12/02 replace beta art w/ sprites from Freedoom (copied from REKKR code)
#ifdef BETA
static const char pls1a0[] = { // 2022/07/05 Freedoom lump
  15,0,15,0,8,0,12,0,68,0,0,0,78,0,0,0,92,0,0,0,108,0,0,0,126,0,0,0,144,0,0,0,
  164,0,0,0,184,0,0,0,204,0,0,0,224,0,0,0,244,0,0,0,6,1,0,0,24,1,0,0,40,1,0,0,
  54,1,0,0,5,5,164,164,165,164,165,164,164,255,3,9,164,164,164,162,164,164,162,
  164,164,164,164,255,2,11,164,164,164,163,162,161,160,161,162,162,163,165,165,
  255,1,13,165,165,162,162,161,161,160,227,161,160,161,162,163,165,165,255,1,
  13,165,165,164,161,161,161,227,225,226,161,227,162,164,166,166,255,0,15,164,
  164,163,162,162,161,226,225,4,4,226,161,161,162,163,73,73,255,0,15,164,164,
  162,161,160,161,225,4,4,4,225,161,161,163,163,164,164,255,0,15,164,164,163,
  161,162,161,225,4,4,4,225,227,160,163,163,165,165,255,0,15,164,164,163,162,
  161,161,227,225,225,4,226,161,162,162,162,164,164,255,0,15,164,164,163,162,
  161,227,227,227,226,227,161,162,162,163,163,164,164,255,1,13,164,164,164,163,
  162,161,161,160,161,160,160,163,163,164,164,255,1,13,164,164,163,163,162,163,
  161,161,161,163,161,163,163,73,73,255,2,11,164,164,164,162,163,161,163,162,
  162,162,162,164,164,255,3,9,163,163,165,162,163,162,162,162,165,163,163,255,
  5,5,166,166,166,164,166,165,165,255,
};
static const char pls1b0[] = { // 2022/07/05 Freedoom lump
  15,0,15,0,8,0,12,0,68,0,0,0,78,0,0,0,92,0,0,0,108,0,0,0,126,0,0,0,144,0,0,0,
  164,0,0,0,184,0,0,0,204,0,0,0,224,0,0,0,244,0,0,0,6,1,0,0,24,1,0,0,40,1,0,0,
  54,1,0,0,5,5,163,163,165,164,72,73,73,255,3,9,164,164,165,163,162,163,163,
  163,164,164,164,255,2,11,165,165,162,162,161,161,161,162,162,162,162,164,164,
  255,1,13,166,166,163,163,161,160,227,227,226,162,161,163,162,163,163,255,1,
  13,164,164,163,162,161,226,225,225,226,226,227,162,163,165,165,255,0,15,164,
  164,164,163,162,161,225,225,4,4,226,227,161,162,162,165,165,255,0,15,163,163,
  162,162,161,227,225,4,4,4,225,226,227,163,163,164,164,255,0,15,165,165,164,
  161,227,227,225,4,4,4,226,226,161,162,163,164,164,255,0,15,163,163,162,161,
  161,226,226,225,4,4,226,226,227,161,163,164,164,255,0,15,165,165,162,162,162,
  161,226,226,226,226,227,227,160,163,162,163,163,255,1,13,165,165,163,161,160,
  227,227,227,161,161,227,161,164,164,164,255,1,13,165,165,163,162,162,161,161,
  160,162,160,162,163,162,164,164,255,2,11,165,165,163,163,161,162,161,162,162,
  164,162,164,164,255,3,9,73,73,164,163,164,164,163,162,165,164,164,255,5,5,
  166,166,165,164,165,165,165,255,
};
static const char pls1c0[] = { // 2022/07/05 Freedoom lump
  15,0,15,0,8,0,12,0,68,0,0,0,78,0,0,0,92,0,0,0,108,0,0,0,126,0,0,0,144,0,0,0,
  164,0,0,0,184,0,0,0,204,0,0,0,224,0,0,0,244,0,0,0,6,1,0,0,24,1,0,0,40,1,0,0,
  54,1,0,0,5,5,164,164,165,164,165,164,164,255,3,9,164,164,164,162,164,164,162,
  164,164,164,164,255,2,11,164,164,164,163,162,161,160,161,162,162,163,165,165,
  255,1,13,165,165,162,162,161,161,160,227,161,160,161,162,163,165,165,255,1,
  13,165,165,164,161,161,161,227,225,226,161,227,162,164,166,166,255,0,15,164,
  164,163,162,162,161,226,225,4,4,226,161,161,162,163,73,73,255,0,15,164,164,
  162,161,160,161,225,4,4,4,225,161,161,163,163,164,164,255,0,15,164,164,163,
  161,162,161,225,4,4,4,225,227,160,163,163,165,165,255,0,15,164,164,163,162,
  161,161,227,225,225,4,226,161,162,162,162,164,164,255,0,15,164,164,163,162,
  161,227,227,227,226,227,161,162,162,163,163,164,164,255,1,13,164,164,164,163,
  162,161,161,160,161,160,160,163,163,164,164,255,1,13,164,164,163,163,162,163,
  161,161,161,163,161,163,163,73,73,255,2,11,164,164,164,162,163,161,163,162,
  162,162,162,164,164,255,3,9,163,163,165,162,163,162,162,162,165,163,163,255,
  5,5,166,166,166,164,166,165,165,255,
};
static const char pls1d0[] = { // 2022/07/05 Freedoom lump
  15,0,15,0,8,0,12,0,68,0,0,0,78,0,0,0,92,0,0,0,108,0,0,0,126,0,0,0,144,0,0,0,
  164,0,0,0,184,0,0,0,204,0,0,0,224,0,0,0,244,0,0,0,6,1,0,0,24,1,0,0,40,1,0,0,
  54,1,0,0,5,5,163,163,165,164,72,73,73,255,3,9,164,164,165,163,162,163,163,
  163,164,164,164,255,2,11,165,165,162,162,161,161,161,162,162,162,162,164,164,
  255,1,13,166,166,163,163,161,160,227,227,226,162,161,163,162,163,163,255,1,
  13,164,164,163,162,161,226,225,225,226,226,227,162,163,165,165,255,0,15,164,
  164,164,163,162,161,225,225,4,4,226,227,161,162,162,165,165,255,0,15,163,163,
  162,162,161,227,225,4,4,4,225,226,227,163,163,164,164,255,0,15,165,165,164,
  161,227,227,225,4,4,4,226,226,161,162,163,164,164,255,0,15,163,163,162,161,
  161,226,226,225,4,4,226,226,227,161,163,164,164,255,0,15,165,165,162,162,162,
  161,226,226,226,226,227,227,160,163,162,163,163,255,1,13,165,165,163,161,160,
  227,227,227,161,161,227,161,164,164,164,255,1,13,165,165,163,162,162,161,161,
  160,162,160,162,163,162,164,164,255,2,11,165,165,163,163,161,162,161,162,162,
  164,162,164,164,255,3,9,73,73,164,163,164,164,163,162,165,164,164,255,5,5,
  166,166,165,164,165,165,165,255,
};
static const char pls1e0[] = { // 2022/07/05 Freedoom lump
  34,0,33,0,17,0,21,0,144,0,0,0,150,0,0,0,157,0,0,0,180,0,0,0,195,0,0,0,216,0,
  0,0,246,0,0,0,20,1,0,0,50,1,0,0,87,1,0,0,129,1,0,0,166,1,0,0,202,1,0,0,234,1,
  0,0,13,2,0,0,47,2,0,0,80,2,0,0,119,2,0,0,155,2,0,0,189,2,0,0,223,2,0,0,0,3,0,
  0,32,3,0,0,62,3,0,0,96,3,0,0,132,3,0,0,167,3,0,0,196,3,0,0,229,3,0,0,3,4,0,0,
  28,4,0,0,53,4,0,0,76,4,0,0,83,4,0,0,19,1,166,166,166,255,7,2,166,166,166,166,
  255,7,2,166,166,166,166,13,7,166,166,166,165,166,165,166,166,166,26,1,166,
  166,166,255,11,10,166,166,165,165,164,164,164,164,164,165,166,166,255,6,16,
  165,165,165,165,165,164,164,164,164,163,164,163,164,164,164,164,166,166,255,
  5,18,165,165,164,164,163,164,164,163,163,160,160,161,161,163,161,163,164,165,
  166,166,25,3,166,166,164,164,164,255,4,25,165,165,165,163,164,163,162,162,
  163,160,163,160,164,160,161,164,163,163,165,164,165,164,164,164,164,166,166,
  255,4,25,165,165,164,164,164,162,160,165,163,162,163,164,227,160,160,160,164,
  163,163,165,164,164,165,164,164,166,166,255,3,27,166,166,164,164,164,163,227,
  227,164,161,161,227,160,160,163,227,227,160,163,162,161,163,164,164,163,165,
  164,166,166,32,1,166,166,166,255,1,1,166,166,166,3,27,166,166,164,163,163,
  160,227,227,162,160,160,160,227,160,227,227,160,160,160,160,163,164,164,160,
  164,164,164,164,164,32,1,166,166,166,255,0,2,166,166,166,166,4,26,165,165,
  165,163,164,160,161,160,160,160,163,227,160,162,161,161,160,161,227,160,160,
  164,164,162,165,164,165,165,255,0,1,166,166,166,4,26,165,165,164,164,161,160,
  162,161,227,161,163,161,162,163,162,164,164,163,162,227,226,161,161,162,164,
  164,165,165,255,4,27,165,165,164,164,160,161,163,227,161,162,164,160,165,164,
  164,165,165,164,163,160,160,227,160,163,163,164,165,165,165,255,3,14,166,166,
  165,164,164,163,163,160,160,163,164,163,164,165,166,166,19,12,166,166,165,
  161,160,227,227,160,161,163,164,165,166,166,255,2,14,164,164,164,164,165,163,
  163,161,161,163,163,164,165,164,166,166,20,11,166,166,163,161,227,161,227,
  162,162,164,165,165,165,255,2,13,166,166,164,164,161,163,161,227,160,227,162,
  165,165,165,165,20,11,165,165,165,162,160,227,160,160,163,164,165,164,164,
  255,0,1,166,166,166,2,13,165,165,164,164,162,160,227,227,227,161,162,164,164,
  165,165,19,12,166,166,166,165,163,160,227,160,160,163,163,166,165,165,255,2,
  14,165,165,164,164,162,160,227,227,160,163,164,165,164,164,166,166,18,13,165,
  165,166,165,164,162,160,227,227,162,162,164,165,165,165,255,2,29,165,165,164,
  164,162,161,227,227,227,161,163,164,165,164,165,166,165,166,167,165,161,160,
  227,227,160,161,162,164,166,166,166,255,2,29,164,164,164,163,161,162,161,161,
  160,227,161,161,164,164,165,165,164,165,165,164,162,162,227,160,161,161,163,
  164,165,165,165,255,2,28,166,166,164,164,164,163,160,162,162,227,160,160,163,
  163,164,162,162,164,165,162,162,227,160,162,162,162,163,165,166,166,255,3,27,
  166,166,164,164,164,164,162,161,160,227,226,160,163,161,227,161,163,162,161,
  227,164,162,163,163,163,164,166,166,166,255,4,25,165,165,165,164,164,164,162,
  161,227,161,163,160,160,227,227,161,160,227,165,164,163,164,162,163,165,166,
  166,255,4,24,165,165,164,163,165,161,161,162,161,160,161,160,227,160,160,160,
  160,162,164,164,164,164,163,164,166,166,31,1,165,165,165,255,3,24,166,166,
  164,164,164,162,163,227,161,163,161,227,227,227,227,227,160,162,162,164,163,
  164,162,164,166,166,30,3,166,166,166,166,166,255,2,25,166,166,165,165,164,
  164,164,162,162,163,161,162,163,162,227,160,162,163,164,164,163,164,164,164,
  165,166,166,31,1,166,166,166,255,2,24,165,165,164,164,164,163,164,164,162,
  162,227,163,163,163,162,163,163,164,163,162,161,165,165,166,166,166,255,2,23,
  164,164,164,164,164,164,164,164,164,163,164,161,161,163,162,227,164,164,164,
  163,164,165,165,165,165,26,1,165,165,165,255,2,3,165,165,165,164,164,6,18,
  164,164,164,164,164,164,164,163,164,163,164,164,164,164,164,164,165,164,166,
  166,255,3,20,166,166,164,164,165,164,165,164,165,164,165,164,164,164,164,164,
  164,166,165,165,166,166,255,5,5,166,166,164,165,165,165,165,11,11,164,164,
  165,165,164,164,164,165,164,165,164,165,165,255,3,1,166,166,166,6,3,165,165,
  166,165,165,14,6,165,165,164,165,165,164,165,165,255,11,2,167,167,166,166,
  255,12,3,166,166,166,166,166,255,
};
static const char pls1f0[] = { // 2022/07/05 Freedoom lump
  40,0,35,0,19,0,25,0,168,0,0,0,181,0,0,0,197,0,0,0,217,0,0,0,242,0,0,0,14,1,0,
  0,45,1,0,0,77,1,0,0,111,1,0,0,145,1,0,0,180,1,0,0,218,1,0,0,0,2,0,0,38,2,0,0,
  76,2,0,0,115,2,0,0,155,2,0,0,195,2,0,0,229,2,0,0,8,3,0,0,42,3,0,0,77,3,0,0,
  113,3,0,0,148,3,0,0,190,3,0,0,236,3,0,0,16,4,0,0,51,4,0,0,85,4,0,0,119,4,0,0,
  152,4,0,0,185,4,0,0,219,4,0,0,254,4,0,0,30,5,0,0,59,5,0,0,86,5,0,0,115,5,0,0,
  141,5,0,0,168,5,0,0,14,8,165,165,166,166,166,166,166,165,165,165,255,12,11,
  165,165,164,165,164,164,164,164,164,164,164,166,166,255,9,15,165,165,165,164,
  166,164,165,164,163,164,163,163,164,165,165,165,165,255,6,20,164,164,165,165,
  164,164,164,165,164,164,164,163,163,163,161,164,164,165,165,165,165,165,255,
  5,23,165,165,164,165,164,165,165,165,164,164,163,161,162,162,163,161,161,164,
  165,164,164,165,165,165,165,255,4,26,164,164,165,164,164,166,164,165,164,164,
  164,162,163,163,162,162,163,160,164,164,164,164,161,163,164,164,165,165,255,
  3,27,164,164,165,164,165,164,165,164,162,164,164,163,162,227,164,162,161,161,
  163,165,164,165,164,165,164,164,164,165,165,255,2,29,166,166,165,164,166,164,
  163,164,161,163,161,162,162,163,164,162,163,162,162,227,162,164,164,165,165,
  165,166,165,164,165,165,255,2,29,166,166,164,165,164,164,163,163,163,160,164,
  162,163,160,162,162,162,161,163,163,161,162,164,163,164,165,164,164,164,165,
  165,255,2,30,165,165,165,164,164,164,161,163,160,160,161,163,163,162,160,160,
  162,161,227,160,163,160,162,163,163,164,164,164,164,164,165,165,255,1,33,166,
  166,165,165,164,164,163,163,162,226,160,162,163,163,160,227,227,162,227,226,
  227,227,162,162,162,163,163,164,164,164,164,164,165,165,165,255,1,33,165,165,
  164,164,164,163,163,163,162,227,161,161,162,160,227,160,160,160,160,227,227,
  227,160,161,161,163,162,163,163,164,164,165,164,165,165,255,1,33,165,165,164,
  164,163,165,164,163,163,162,162,162,162,161,162,227,227,160,163,161,227,226,
  226,227,160,227,162,163,164,164,164,164,165,165,165,255,1,33,164,164,164,165,
  164,164,163,161,163,163,163,163,163,160,227,227,161,162,163,164,164,164,163,
  160,227,160,227,160,161,163,164,164,165,165,165,255,1,34,165,165,164,164,164,
  164,161,227,160,162,161,160,227,227,227,161,164,164,165,164,165,164,164,161,
  160,227,160,227,160,162,162,164,164,165,165,165,255,0,17,165,165,165,166,165,
  165,164,161,160,163,162,161,160,227,160,162,164,164,164,21,14,164,164,162,
  164,163,227,160,160,162,163,162,162,164,165,166,166,255,0,15,166,166,165,164,
  164,164,164,162,161,163,160,227,160,227,160,164,164,22,2,164,164,163,163,25,
  10,164,164,163,160,227,160,162,161,163,165,165,165,255,0,14,164,164,164,165,
  164,164,163,163,162,160,227,160,227,162,164,164,24,11,164,164,161,160,227,
  227,160,163,162,164,166,165,165,255,0,14,165,165,164,165,164,162,161,162,161,
  160,160,227,160,161,165,165,23,12,165,165,162,162,160,227,160,162,161,162,
  164,164,166,166,255,0,13,165,165,164,164,164,162,161,160,161,160,227,160,162,
  164,164,23,12,164,164,164,161,163,163,162,164,163,164,164,164,165,165,255,0,
  13,165,165,164,164,163,160,162,160,163,227,160,227,161,165,165,22,13,165,165,
  165,163,161,160,160,161,163,164,164,163,164,165,165,255,0,14,165,165,165,165,
  164,163,161,162,163,227,160,160,160,163,165,165,22,13,165,165,164,160,160,
  226,227,227,163,161,164,164,163,165,165,255,1,13,165,165,165,164,164,162,160,
  163,161,226,160,227,163,164,164,22,13,164,164,162,227,227,160,160,161,161,
  162,164,164,164,165,165,255,1,13,165,165,164,164,164,163,163,163,160,160,160,
  227,161,164,164,17,4,164,164,164,164,163,163,23,12,162,162,227,227,227,163,
  163,162,162,164,164,165,165,165,255,1,14,166,166,164,164,164,164,163,161,161,
  161,161,227,227,163,164,164,16,2,227,227,161,161,20,2,161,161,160,160,23,11,
  161,161,160,161,162,163,161,161,165,164,165,166,166,255,2,15,166,166,164,164,
  164,164,163,161,161,163,160,227,160,161,163,164,164,22,12,162,162,227,161,
  163,162,162,161,164,164,164,166,165,165,255,3,30,165,165,164,164,163,164,162,
  163,161,160,227,160,161,162,163,164,164,164,164,164,227,161,161,163,164,164,
  161,163,164,165,166,166,255,3,29,164,164,165,164,164,164,163,161,162,163,160,
  160,160,227,162,227,162,160,160,227,160,163,163,162,164,164,163,164,166,165,
  165,255,3,29,165,165,165,164,164,164,162,160,163,163,162,160,160,226,227,227,
  160,227,160,160,162,160,164,164,164,164,164,164,166,166,166,255,3,28,164,164,
  164,164,164,164,160,160,161,163,162,160,160,163,227,160,162,162,161,162,162,
  163,164,161,164,163,164,164,166,166,255,2,28,166,166,164,164,164,163,164,164,
  163,163,162,161,163,160,227,162,162,162,161,162,161,162,164,164,164,164,164,
  164,166,166,255,1,29,164,164,165,165,165,164,164,163,164,161,163,161,161,162,
  162,160,162,161,160,162,161,164,164,162,164,165,165,166,165,165,165,255,1,4,
  164,164,165,164,164,164,7,22,163,163,164,164,162,162,160,163,162,161,160,163,
  163,161,164,164,164,164,164,164,164,164,166,166,255,1,3,165,165,164,165,165,
  8,20,164,164,164,163,164,164,163,162,161,163,227,227,164,164,164,163,164,165,
  165,165,164,164,255,1,3,165,165,164,165,165,9,17,164,164,164,164,164,164,164,
  163,163,161,163,164,165,164,164,165,165,165,165,255,2,3,166,166,164,165,165,
  10,15,164,164,164,164,164,164,164,164,163,162,163,164,164,165,166,166,166,
  255,2,4,166,166,164,164,165,165,9,16,165,165,164,164,165,164,165,164,164,164,
  164,164,164,166,165,165,165,165,255,3,21,165,165,165,164,165,165,166,166,164,
  165,164,164,165,164,165,164,164,164,165,166,165,166,166,255,4,7,165,165,165,
  165,164,165,164,165,165,12,11,165,165,166,164,165,165,164,164,165,164,165,
  165,165,255,6,4,165,165,165,166,165,165,15,6,165,165,165,165,164,165,164,164,
  255,
};
static const char pls1g0[] = { // 2022/07/05 Freedoom lump
  46,0,39,0,22,0,28,0,192,0,0,0,201,0,0,0,214,0,0,0,234,0,0,0,5,1,0,0,38,1,0,0,
  76,1,0,0,113,1,0,0,150,1,0,0,187,1,0,0,226,1,0,0,9,2,0,0,55,2,0,0,98,2,0,0,
  135,2,0,0,171,2,0,0,205,2,0,0,236,2,0,0,10,3,0,0,40,3,0,0,67,3,0,0,100,3,0,0,
  128,3,0,0,156,3,0,0,184,3,0,0,212,3,0,0,244,3,0,0,17,4,0,0,48,4,0,0,79,4,0,0,
  120,4,0,0,162,4,0,0,196,4,0,0,227,4,0,0,10,5,0,0,48,5,0,0,82,5,0,0,108,5,0,0,
  140,5,0,0,174,5,0,0,209,5,0,0,242,5,0,0,14,6,0,0,45,6,0,0,74,6,0,0,104,6,0,0,
  20,4,166,166,166,166,166,166,255,17,8,166,166,164,164,164,164,164,164,167,
  167,255,10,2,166,166,166,166,16,9,166,166,164,163,163,163,164,163,164,167,
  167,255,6,8,165,165,166,166,166,164,164,164,166,166,15,10,164,164,164,162,
  162,161,161,161,164,163,165,165,255,4,20,165,165,166,166,164,164,164,166,165,
  166,164,164,164,162,163,161,161,162,161,162,163,163,27,4,164,164,165,166,165,
  165,255,3,22,166,166,166,166,164,166,166,166,164,164,164,163,164,162,161,160,
  161,162,161,161,161,163,164,164,26,7,163,163,160,162,162,163,166,166,166,255,
  2,32,166,166,166,166,164,165,164,164,163,163,161,163,164,162,161,163,162,161,
  160,160,161,161,161,164,166,166,166,164,164,166,166,165,166,166,255,2,32,166,
  166,166,164,164,166,164,164,164,163,161,161,160,162,162,161,163,161,161,161,
  161,161,162,164,164,164,165,164,163,163,164,166,165,165,255,2,32,166,166,166,
  164,164,165,164,164,164,163,162,162,162,162,161,164,163,160,161,162,160,163,
  162,161,164,164,166,166,164,164,164,166,165,165,255,1,34,166,166,166,164,166,
  165,164,164,164,163,161,163,162,161,161,162,164,161,162,164,163,162,164,162,
  161,161,164,164,166,166,167,166,164,165,166,166,255,1,34,167,167,164,166,164,
  164,163,163,162,161,160,161,164,161,163,163,162,161,164,164,164,161,162,161,
  161,164,164,163,164,164,163,164,164,164,166,166,255,1,17,166,166,166,164,164,
  163,163,161,162,161,160,163,162,162,164,161,163,160,160,19,6,165,165,164,164,
  163,161,163,163,26,10,162,162,162,163,164,164,164,163,164,164,166,166,255,0,
  17,165,165,166,164,163,163,164,162,161,162,161,163,163,162,163,164,164,164,
  164,22,2,164,164,164,164,27,11,164,164,162,163,163,164,163,163,164,164,166,
  166,166,255,0,15,166,166,164,163,163,163,163,161,161,162,162,164,161,161,163,
  164,164,25,13,160,160,162,162,160,162,163,164,164,163,163,164,164,166,166,
  255,0,14,166,166,163,164,164,166,166,163,160,163,164,161,162,163,164,164,25,
  13,164,164,164,163,161,161,161,163,163,163,163,164,164,166,166,255,0,13,166,
  166,163,166,164,163,163,162,161,162,161,162,162,164,164,26,12,163,163,163,
  163,163,160,161,162,161,163,164,164,166,166,255,0,12,166,166,164,165,165,163,
  163,161,164,163,164,163,164,164,28,10,164,164,163,163,161,161,162,162,163,
  164,166,166,255,1,11,167,167,166,166,163,163,161,161,164,164,163,164,164,26,
  10,164,164,164,163,164,163,163,161,161,164,164,164,255,2,9,164,164,163,165,
  164,161,164,163,163,164,164,27,12,164,164,163,162,164,163,164,164,161,161,
  163,167,164,164,255,4,7,162,162,160,161,160,166,163,164,164,28,11,164,164,
  164,164,163,161,160,162,162,163,166,165,165,255,2,9,167,167,162,161,162,160,
  161,162,164,164,164,13,1,162,162,162,29,10,160,160,164,160,162,162,161,162,
  164,166,166,166,255,1,9,164,164,164,161,162,161,160,161,163,161,161,29,10,
  164,164,161,162,163,163,161,163,164,164,166,166,255,0,10,165,165,164,164,161,
  162,162,160,161,163,164,164,30,9,164,164,160,163,161,163,163,163,164,166,166,
  255,0,10,166,166,164,163,162,161,161,160,161,163,164,164,30,9,164,164,161,
  161,161,164,163,164,164,166,166,255,0,10,165,165,164,164,162,161,162,161,162,
  163,164,164,30,9,164,164,161,163,162,163,163,163,164,166,166,255,1,9,164,164,
  163,161,162,161,161,161,163,164,164,28,1,160,160,160,30,9,163,163,163,163,
  162,161,163,163,164,166,166,255,0,11,166,166,164,163,163,161,162,160,161,163,
  161,164,164,30,9,164,164,160,163,162,161,163,164,164,166,166,255,0,11,166,
  166,164,164,164,163,162,160,161,161,161,164,164,28,11,164,164,163,161,162,
  161,163,162,163,164,167,166,166,255,0,12,167,167,164,164,163,163,163,161,162,
  162,161,162,164,164,28,10,164,164,164,164,163,162,162,164,163,166,167,167,
  255,1,10,167,167,164,164,163,164,163,160,161,163,164,164,13,5,164,164,164,
  164,161,164,164,25,2,164,164,164,164,30,7,164,164,164,163,163,164,164,167,
  167,255,2,9,166,166,164,164,164,163,164,164,164,166,166,12,7,164,164,163,163,
  163,160,161,163,163,22,8,163,163,164,164,160,162,164,164,163,163,34,1,164,
  164,164,255,4,5,166,166,166,164,165,165,165,11,20,164,164,162,161,163,163,
  163,161,161,161,160,163,163,163,160,161,162,162,164,164,164,164,255,11,20,
  160,160,163,162,160,163,163,163,164,163,164,163,161,161,162,161,163,164,164,
  164,164,164,34,2,166,166,166,166,255,3,3,165,165,165,165,165,10,22,164,164,
  163,161,161,160,160,161,163,163,164,160,160,162,162,162,163,163,162,164,164,
  162,164,164,34,1,166,166,166,255,3,4,167,167,167,167,166,166,9,9,164,164,163,
  163,162,161,162,163,163,164,164,20,12,164,164,161,162,161,161,164,164,163,
  164,164,163,164,164,255,4,2,166,166,166,166,9,8,164,164,161,161,162,162,161,
  163,164,164,21,11,164,164,162,164,163,161,163,164,164,166,166,165,165,255,9,
  7,163,163,164,161,161,164,161,164,164,22,10,163,163,164,163,163,164,164,165,
  166,167,167,167,255,3,2,165,165,166,166,10,7,163,163,163,164,163,161,161,164,
  164,21,10,163,163,163,164,163,164,164,164,166,166,166,166,255,2,4,166,166,
  166,166,166,166,10,7,163,163,164,164,164,162,161,164,164,21,10,163,163,164,
  164,163,164,164,165,166,166,166,166,255,2,4,166,166,166,166,166,166,10,8,163,
  163,163,163,163,161,162,164,164,164,20,10,163,163,164,164,164,163,164,165,
  166,166,167,167,255,1,6,163,163,166,164,165,166,165,165,10,18,164,164,164,
  164,163,163,163,163,162,162,162,164,164,164,163,164,166,166,166,166,255,1,3,
  167,167,163,166,166,11,16,164,164,164,163,163,164,163,163,160,162,162,163,
  164,164,167,166,167,167,255,1,5,167,167,164,164,166,166,166,10,17,165,165,
  164,164,164,164,164,164,163,163,164,163,163,164,167,166,165,166,166,255,2,24,
  166,166,164,163,163,164,166,166,166,166,166,166,164,164,164,164,164,163,163,
  164,165,166,166,165,166,166,255,3,9,166,166,166,166,166,166,166,164,164,166,
  166,13,12,166,166,166,166,165,164,164,164,164,164,166,165,165,165,255,6,5,
  166,166,166,166,166,166,166,16,7,165,165,166,165,166,166,166,166,166,255,
};
static const char pls2a0[] = { // 2022/07/05 Freedoom lump
  15,0,15,0,8,0,14,0,68,0,0,0,81,0,0,0,93,0,0,0,108,0,0,0,125,0,0,0,143,0,0,0,
  163,0,0,0,182,0,0,0,201,0,0,0,220,0,0,0,240,0,0,0,2,1,0,0,20,1,0,0,36,1,0,0,
  51,1,0,0,5,1,254,254,254,7,3,254,254,253,253,253,255,4,7,254,254,201,252,203,
  253,205,253,253,255,2,10,254,254,253,203,204,201,201,200,200,203,254,254,255,
  1,12,254,254,203,204,201,201,201,198,201,201,201,203,254,254,255,1,13,253,
  253,203,201,201,200,197,197,198,198,201,201,254,253,253,255,0,15,253,253,201,
  203,200,198,196,196,196,196,197,198,201,200,201,253,253,255,0,14,253,253,203,
  200,201,197,196,194,196,193,196,197,201,201,253,253,255,1,14,254,254,201,201,
  196,195,195,194,196,195,196,198,201,253,253,253,255,1,14,253,253,201,198,196,
  195,195,193,195,195,196,201,201,203,254,254,255,0,15,254,254,203,201,201,197,
  196,194,195,195,195,197,201,204,203,253,253,255,1,13,254,254,203,198,198,196,
  196,195,195,196,198,201,203,254,254,255,1,13,253,253,252,203,203,198,197,196,
  197,200,201,203,203,253,253,255,2,11,253,253,203,204,204,201,201,201,203,203,
  203,253,253,255,3,10,254,254,254,203,201,253,253,201,253,254,254,254,255,5,2,
  253,253,253,253,9,1,254,254,254,255,
};
static const char pls2b0[] = { // 2022/07/05 Freedoom lump
  16,0,16,0,8,0,15,0,72,0,0,0,84,0,0,0,96,0,0,0,110,0,0,0,126,0,0,0,144,0,0,0,
  163,0,0,0,183,0,0,0,202,0,0,0,223,0,0,0,244,0,0,0,7,1,0,0,25,1,0,0,42,1,0,0,
  57,1,0,0,70,1,0,0,6,1,254,254,254,8,2,253,253,253,253,255,4,7,253,253,253,
  203,253,201,203,254,254,255,3,9,253,253,253,203,201,201,201,201,253,254,254,
  255,2,11,254,254,253,201,203,200,198,198,203,203,203,253,253,255,1,13,253,
  253,253,201,200,201,200,200,201,200,200,198,203,254,254,255,1,14,254,254,204,
  201,200,197,197,196,196,198,201,201,200,254,254,254,255,0,15,254,254,203,201,
  201,198,196,196,193,195,196,198,201,201,201,253,253,255,1,14,253,253,201,201,
  197,195,195,194,195,195,197,201,201,203,253,253,255,0,16,253,253,203,201,200,
  197,195,193,195,193,195,197,200,201,201,205,253,253,255,0,16,253,253,201,201,
  201,198,196,195,194,195,195,198,201,201,203,204,254,254,255,1,14,253,253,204,
  201,198,196,195,195,195,197,201,201,201,204,254,254,255,1,13,253,253,253,203,
  201,200,198,197,197,201,201,204,203,254,254,255,2,12,254,254,201,203,201,201,
  201,201,204,203,201,253,254,254,255,3,10,254,254,205,205,204,203,201,201,205,
  254,253,253,255,4,8,253,253,253,205,205,254,253,253,253,253,255,6,2,253,253,
  253,253,255,
};
static const char pls2c0[] = { // 2022/07/05 Freedoom lump
  32,0,30,0,16,0,23,0,136,0,0,0,142,0,0,0,155,0,0,0,172,0,0,0,192,0,0,0,217,0,
  0,0,238,0,0,0,6,1,0,0,37,1,0,0,65,1,0,0,94,1,0,0,123,1,0,0,160,1,0,0,193,1,0,
  0,229,1,0,0,8,2,0,0,44,2,0,0,76,2,0,0,104,2,0,0,134,2,0,0,163,2,0,0,193,2,0,
  0,224,2,0,0,251,2,0,0,23,3,0,0,48,3,0,0,72,3,0,0,95,3,0,0,122,3,0,0,148,3,0,
  0,178,3,0,0,198,3,0,0,17,1,207,207,207,255,3,1,243,243,243,16,3,205,205,241,
  242,242,255,10,4,205,205,206,207,206,206,15,4,204,204,204,207,241,241,255,8,
  2,241,241,207,207,12,9,207,207,206,241,205,203,204,243,244,242,242,255,4,1,
  206,206,206,7,15,241,241,207,205,206,241,207,207,241,241,202,204,243,244,243,
  241,241,255,6,16,205,205,207,207,205,205,205,205,206,203,205,204,204,244,244,
  242,241,241,255,4,1,242,242,242,7,14,205,205,205,204,204,203,204,206,203,204,
  203,202,241,241,207,207,255,3,21,243,243,242,241,241,206,204,204,204,204,204,
  205,198,198,198,198,203,206,205,206,207,241,241,26,1,241,241,241,255,2,23,
  242,242,242,207,207,207,206,205,204,203,204,203,204,199,198,198,198,204,202,
  204,205,206,242,242,242,255,2,24,243,243,242,205,205,204,206,205,203,202,202,
  202,204,198,197,198,199,202,202,204,207,241,243,242,241,241,255,3,24,242,242,
  207,207,207,241,207,203,199,199,198,198,198,197,197,198,197,202,202,207,207,
  242,241,242,241,241,255,1,27,243,243,244,245,244,243,242,241,242,207,198,197,
  196,197,196,195,196,196,196,197,199,203,205,207,206,207,241,242,242,29,1,243,
  243,243,255,0,28,241,241,242,244,244,243,242,242,243,242,206,205,196,195,195,
  195,193,194,194,195,196,197,203,203,204,206,241,242,242,242,255,2,23,242,242,
  207,206,204,206,206,241,207,198,196,194,193,193,193,193,193,194,195,197,202,
  205,205,206,206,26,4,242,242,207,242,207,207,255,2,24,205,205,205,205,204,
  203,202,198,198,197,195,193,193,193,193,192,194,193,195,196,203,206,207,241,
  207,207,28,2,241,241,242,242,255,1,26,206,206,205,204,205,205,204,199,198,
  197,195,194,193,193,192,192,193,193,193,194,196,198,203,206,205,205,205,205,
  28,1,207,207,207,255,1,27,205,205,204,205,205,207,205,203,198,197,195,193,
  192,192,82,192,82,192,192,194,195,203,203,205,206,241,242,242,242,255,3,23,
  203,203,203,205,205,204,203,203,196,193,193,82,82,192,82,192,193,194,196,198,
  205,204,205,206,206,255,1,25,242,242,207,202,202,203,207,206,203,198,197,195,
  194,193,82,82,193,192,194,196,197,205,206,206,207,207,207,255,3,24,204,204,
  205,206,206,206,203,198,196,195,194,194,193,193,193,194,195,197,206,207,206,
  206,207,206,206,206,255,2,25,205,205,207,242,241,207,204,203,198,197,195,194,
  195,194,194,194,195,196,197,203,207,207,205,207,206,206,206,255,3,21,243,243,
  242,242,207,207,206,204,198,197,197,196,195,195,195,196,197,198,203,206,207,
  206,206,25,1,207,207,207,255,4,22,243,243,242,243,242,207,207,203,203,202,
  198,197,197,196,203,203,203,204,205,206,242,243,207,207,255,3,23,204,204,206,
  241,207,207,242,205,204,202,205,203,202,199,199,205,205,204,206,206,206,207,
  243,207,207,255,5,20,241,241,206,242,241,206,204,203,204,205,204,199,198,202,
  207,205,203,207,206,241,241,241,255,5,19,241,241,242,242,242,241,206,205,207,
  207,203,202,202,202,206,205,206,207,207,241,241,255,6,18,242,242,242,241,241,
  207,205,241,241,203,203,204,204,204,205,205,204,207,205,205,255,5,17,243,243,
  242,241,204,206,206,205,241,241,202,206,203,204,206,207,243,205,205,26,1,241,
  241,241,255,7,1,207,207,207,9,1,241,241,241,11,11,241,241,241,242,241,207,
  204,206,207,207,241,207,207,255,8,2,207,207,242,242,11,3,243,243,243,242,242,
  15,1,207,207,207,18,2,241,241,241,241,21,1,207,207,207,255,8,1,207,207,207,
  11,3,241,241,243,241,241,19,3,206,206,206,207,207,255,11,2,241,241,241,241,
  19,1,206,206,206,255,
};
static const char pls2d0[] = { // 2022/07/05 Freedoom lump
  40,0,35,0,20,0,27,0,168,0,0,0,189,0,0,0,227,0,0,0,10,1,0,0,40,1,0,0,71,1,0,0,
  109,1,0,0,138,1,0,0,190,1,0,0,235,1,0,0,18,2,0,0,59,2,0,0,98,2,0,0,141,2,0,0,
  182,2,0,0,229,2,0,0,16,3,0,0,59,3,0,0,100,3,0,0,140,3,0,0,187,3,0,0,234,3,0,
  0,23,4,0,0,70,4,0,0,116,4,0,0,162,4,0,0,201,4,0,0,238,4,0,0,22,5,0,0,68,5,0,
  0,118,5,0,0,157,5,0,0,201,5,0,0,243,5,0,0,29,6,0,0,72,6,0,0,110,6,0,0,138,6,
  0,0,166,6,0,0,188,6,0,0,15,3,207,207,206,207,207,19,1,241,241,241,25,4,207,
  207,207,207,206,206,255,13,1,207,207,207,15,1,206,206,206,19,1,207,207,207,
  21,1,241,241,241,24,3,207,207,207,206,206,28,1,207,207,207,30,1,241,241,241,
  255,12,4,207,207,241,207,206,206,17,2,205,205,206,206,20,3,206,206,241,205,
  205,24,1,241,241,241,27,3,241,241,206,206,206,31,1,207,207,207,255,12,1,242,
  242,242,14,1,206,206,206,17,2,205,205,205,205,20,2,206,206,206,206,29,3,207,
  207,206,206,206,255,10,1,242,242,242,13,3,206,206,206,205,205,17,8,203,203,
  205,205,206,205,205,205,206,206,30,2,241,241,205,205,255,9,1,206,206,206,12,
  2,205,205,205,205,15,3,206,206,203,202,202,19,1,205,205,205,21,5,205,205,205,
  205,206,206,206,31,1,207,207,207,255,8,3,207,207,241,241,241,14,12,205,205,
  206,201,199,204,202,203,202,205,205,206,204,204,31,1,206,206,206,255,6,1,241,
  241,241,8,1,207,207,207,11,3,205,205,206,205,205,15,1,203,203,203,17,4,202,
  202,203,203,204,204,22,4,199,199,203,203,205,205,27,1,205,205,205,30,4,206,
  206,206,241,207,207,255,7,1,207,207,207,10,2,205,205,203,203,13,11,205,205,
  203,203,202,203,198,203,199,203,203,202,202,25,3,203,203,205,205,205,29,2,
  206,206,241,241,32,1,207,207,207,255,5,1,241,241,241,7,23,205,205,207,205,
  206,205,202,202,204,203,203,202,203,199,199,202,202,202,203,204,205,206,205,
  205,205,31,2,206,206,242,242,255,4,7,243,243,205,205,203,206,203,205,205,12,
  15,203,203,203,202,201,202,201,199,202,202,199,203,202,203,203,202,202,28,1,
  205,205,205,30,1,206,206,206,255,3,7,244,244,242,205,205,206,206,205,205,11,
  17,204,204,204,202,202,201,199,199,199,199,201,202,204,199,199,202,203,205,
  205,30,2,207,207,241,241,255,3,4,242,242,243,205,205,205,8,5,205,205,203,203,
  203,201,201,14,12,202,202,199,202,199,202,205,199,199,202,203,203,202,202,27,
  5,205,205,205,206,242,207,207,255,2,6,241,241,241,207,204,203,205,205,9,19,
  203,203,202,202,201,202,203,203,204,203,203,206,206,199,199,202,203,203,204,
  206,206,29,3,206,206,206,206,206,255,1,2,242,242,241,241,4,1,205,205,205,7,3,
  204,204,203,203,203,11,18,201,201,203,203,206,206,206,202,205,203,199,199,
  199,203,203,202,203,202,206,206,31,2,206,206,243,243,255,1,1,207,207,207,3,3,
  206,206,205,203,203,7,21,203,203,203,201,199,203,203,202,206,206,206,202,206,
  203,202,203,205,199,202,199,202,203,203,30,1,206,206,206,255,1,13,242,242,
  205,206,202,202,202,203,199,201,206,206,203,204,204,15,2,202,202,203,203,18,
  15,203,203,203,203,204,205,206,199,199,202,204,203,204,205,205,206,206,255,0,
  12,207,207,206,207,205,203,203,202,199,199,206,206,203,203,19,1,203,203,203,
  21,10,199,199,204,206,202,199,202,203,202,203,205,205,33,1,241,241,241,255,0,
  1,207,207,207,2,2,205,205,205,205,5,6,203,203,201,199,201,206,202,202,22,6,
  207,207,202,199,199,199,203,203,30,4,203,203,206,205,207,207,255,1,9,205,205,
  204,204,206,203,199,199,202,203,203,11,3,203,203,203,205,205,17,1,206,206,
  206,21,12,203,203,203,206,199,199,201,203,199,203,203,206,206,206,34,1,207,
  207,207,255,0,10,207,207,206,206,205,203,205,203,202,203,206,206,12,2,206,
  206,203,203,16,3,206,206,203,207,207,20,9,203,203,202,205,203,201,201,202,
  203,202,202,31,2,205,205,205,205,255,0,1,206,206,206,2,9,206,206,203,202,203,
  202,199,201,206,203,203,16,17,203,203,203,203,205,206,203,203,202,199,199,
  202,202,202,203,203,206,206,206,34,1,206,206,206,255,0,11,207,207,241,204,
  203,203,202,199,199,201,205,205,205,13,1,202,202,202,15,17,205,205,206,203,
  207,204,203,203,202,201,201,199,203,202,204,202,204,205,205,34,1,207,207,207,
  255,1,4,241,241,206,203,202,202,6,1,201,201,201,8,4,203,203,204,205,203,203,
  15,14,203,203,205,203,203,206,205,206,201,201,198,202,204,202,203,203,33,2,
  207,207,207,207,255,0,1,243,243,243,2,12,206,206,203,202,202,202,199,201,199,
  203,206,202,202,202,15,20,203,203,203,203,203,206,203,205,199,199,203,203,
  204,202,204,205,241,207,242,206,207,207,255,0,7,243,243,241,206,205,203,202,
  202,202,8,23,201,201,199,201,206,203,203,206,204,206,206,206,203,199,199,201,
  202,203,203,201,199,203,205,207,207,255,2,3,207,207,206,205,205,7,25,202,202,
  199,199,199,198,202,207,205,205,205,199,198,199,201,201,203,203,203,203,203,
  203,204,206,205,207,207,255,4,25,204,204,205,206,202,203,201,199,201,202,199,
  199,203,203,202,201,199,202,199,203,203,203,202,206,203,205,205,30,1,205,205,
  205,33,1,242,242,242,255,1,5,241,241,205,206,205,206,206,7,21,205,205,203,
  199,203,203,201,201,201,201,203,199,201,201,199,202,203,202,202,201,203,203,
  203,29,2,205,205,206,206,32,1,207,207,207,255,1,2,241,241,207,207,4,2,205,
  205,205,205,7,22,205,205,204,203,203,203,202,201,199,201,198,203,199,201,199,
  203,203,203,201,199,199,203,203,203,30,1,205,205,205,32,2,207,207,241,241,
  255,4,24,206,206,205,206,205,205,206,202,203,201,202,199,199,202,202,203,199,
  203,202,203,203,201,202,204,203,203,29,1,205,205,205,32,1,241,241,241,255,4,
  4,206,206,205,205,206,206,9,2,205,205,205,205,12,14,203,203,203,203,204,203,
  204,203,204,201,203,202,206,201,202,202,28,2,204,204,206,206,32,1,243,243,
  243,255,4,20,206,206,206,241,241,207,205,207,206,205,203,199,203,203,202,202,
  206,203,202,203,202,202,25,2,203,203,205,205,28,2,205,205,207,207,31,1,207,
  207,207,255,5,12,207,207,206,206,205,206,205,207,205,205,203,202,202,202,18,
  5,205,205,201,203,203,205,205,24,1,203,203,203,27,2,206,206,207,207,31,1,207,
  207,207,255,5,3,206,206,207,206,206,9,5,202,202,205,206,206,206,206,15,11,
  204,204,203,203,203,203,203,205,205,206,206,206,206,27,1,207,207,207,29,2,
  207,207,241,241,255,7,1,207,207,207,9,2,206,206,207,207,12,4,207,207,204,203,
  203,203,17,8,203,203,202,202,205,205,206,206,207,207,26,2,206,206,206,206,
  255,9,6,241,241,207,206,242,205,205,205,16,8,203,203,203,202,205,205,206,207,
  206,206,28,1,207,207,207,255,11,1,207,207,207,13,7,205,205,205,205,205,204,
  206,206,206,21,1,207,207,207,23,2,207,207,241,241,255,12,1,242,242,242,14,1,
  206,206,206,16,7,206,206,206,206,206,206,241,207,207,255,13,6,241,241,207,
  241,207,241,241,241,20,1,242,242,242,255,
};

static const char pls2e0[] = { // 2022/07/05 Freedoom lump
  49,0,44,0,25,0,34,0,204,0,0,0,215,0,0,0,237,0,0,0,33,1,0,0,84,1,0,0,137,1,0,
  0,192,1,0,0,234,1,0,0,37,2,0,0,69,2,0,0,116,2,0,0,156,2,0,0,208,2,0,0,1,3,0,
  0,47,3,0,0,88,3,0,0,137,3,0,0,177,3,0,0,240,3,0,0,28,4,0,0,65,4,0,0,111,4,0,
  0,151,4,0,0,184,4,0,0,221,4,0,0,11,5,0,0,49,5,0,0,94,5,0,0,127,5,0,0,174,5,0,
  0,231,5,0,0,27,6,0,0,75,6,0,0,119,6,0,0,179,6,0,0,234,6,0,0,28,7,0,0,77,7,0,
  0,142,7,0,0,202,7,0,0,11,8,0,0,73,8,0,0,142,8,0,0,196,8,0,0,251,8,0,0,48,9,0,
  0,77,9,0,0,119,9,0,0,141,9,0,0,27,1,207,207,207,30,1,207,207,207,255,6,2,241,
  241,206,206,13,1,206,206,206,28,1,241,241,241,36,1,242,242,242,255,7,1,207,
  207,207,10,1,241,241,241,12,1,207,207,207,18,2,241,241,207,207,21,4,206,206,
  207,241,241,241,26,1,207,207,207,28,1,242,242,242,30,3,241,241,207,242,242,
  39,1,207,207,207,255,9,1,241,241,241,13,1,206,206,206,16,4,241,241,242,241,
  205,205,21,1,206,206,206,23,4,206,206,207,207,206,206,29,3,206,206,206,206,
  206,33,2,207,207,207,207,36,2,242,242,207,207,255,3,1,207,207,207,7,2,206,
  206,207,207,12,1,207,207,207,16,4,206,206,241,206,207,207,21,11,205,205,205,
  205,206,206,206,206,206,207,206,207,207,33,1,206,206,206,35,4,205,205,204,
  206,241,241,255,5,1,207,207,207,10,1,207,207,207,12,1,206,206,206,14,4,242,
  242,241,241,242,242,19,5,207,207,205,207,205,206,206,25,1,207,207,207,27,4,
  206,206,205,206,241,241,35,5,207,207,241,206,206,241,241,255,12,1,207,207,
  207,14,8,207,207,241,241,206,207,205,207,203,203,23,1,241,241,241,26,3,206,
  206,204,206,206,32,1,207,207,207,37,3,241,241,205,242,242,255,0,1,206,206,
  206,9,2,241,241,241,241,12,7,242,242,241,207,207,206,205,206,206,21,1,204,
  204,204,23,1,207,207,207,25,6,205,205,205,204,205,206,205,205,32,1,207,207,
  207,34,7,207,207,206,207,205,243,205,207,207,255,10,9,206,206,242,241,207,
  207,206,205,206,206,206,24,1,205,205,205,26,1,202,202,202,35,4,207,207,207,
  207,241,241,255,10,3,207,207,207,241,241,14,1,207,207,207,16,2,204,204,206,
  206,19,1,205,205,205,24,3,206,206,203,202,202,29,1,205,205,205,35,2,207,207,
  206,206,38,1,206,206,206,255,3,1,242,242,242,5,2,242,242,207,207,10,5,207,
  207,242,241,207,205,205,17,4,205,205,205,207,241,241,31,2,207,207,207,207,36,
  1,241,241,241,255,9,1,243,243,243,12,2,207,207,206,206,15,3,206,206,206,241,
  241,21,1,207,207,207,23,1,206,206,206,25,3,206,206,203,204,204,30,2,205,205,
  206,206,34,1,207,207,207,36,1,241,241,241,255,9,8,243,243,241,207,205,205,
  206,203,205,205,18,2,204,204,206,206,26,2,204,204,207,207,30,2,203,203,241,
  241,33,1,207,207,207,35,4,207,207,242,207,206,206,40,1,206,206,206,255,10,1,
  206,206,206,13,1,205,205,205,18,1,207,207,207,23,1,205,205,205,25,1,205,205,
  205,27,4,202,202,206,207,205,205,34,8,241,241,207,207,241,207,207,206,241,
  241,255,12,1,207,207,207,14,1,207,207,207,18,1,241,241,241,22,2,205,205,205,
  205,26,2,207,207,202,202,29,1,207,207,207,34,4,241,241,242,207,241,241,255,5,
  2,207,207,241,241,10,2,206,206,206,206,13,2,241,241,205,205,17,5,241,241,241,
  206,207,207,207,25,1,204,204,204,28,2,241,241,199,199,36,1,207,207,207,42,1,
  206,206,206,255,5,4,241,241,207,206,206,206,10,1,207,207,207,17,1,205,205,
  205,21,1,199,199,199,27,1,206,206,206,32,1,207,207,207,35,2,207,207,207,207,
  255,5,2,242,242,207,207,10,1,207,207,207,12,1,206,206,206,14,1,203,203,203,
  16,1,207,207,207,18,1,207,207,207,22,1,204,204,204,24,2,202,202,207,207,27,2,
  205,205,241,241,32,2,207,207,206,206,35,4,207,207,206,207,206,206,255,4,1,
  242,242,242,7,1,205,205,205,12,1,241,241,241,15,1,206,206,206,25,2,206,206,
  206,206,31,1,242,242,242,33,1,203,203,203,37,3,205,205,206,242,242,255,3,7,
  207,207,241,205,205,202,203,205,205,29,2,207,207,203,203,32,1,203,203,203,34,
  1,203,203,203,36,5,205,205,205,206,207,243,243,255,2,2,207,207,242,242,5,7,
  205,205,204,203,207,202,206,207,207,26,1,206,206,206,32,1,207,207,207,35,1,
  203,203,203,37,4,205,205,207,206,241,241,42,1,242,242,242,255,1,7,241,241,
  207,207,206,206,203,241,241,9,1,201,201,201,28,1,207,207,207,31,1,202,202,
  202,36,2,207,207,206,206,39,3,205,205,207,207,207,255,2,3,207,207,241,204,
  204,7,4,203,203,207,207,206,206,33,2,204,204,202,202,36,2,202,202,203,203,39,
  1,206,206,206,255,2,3,241,241,206,206,206,8,3,205,205,202,206,206,19,1,241,
  241,241,23,1,207,207,207,32,8,206,206,203,201,203,202,204,205,206,206,255,1,
  3,206,206,206,206,206,6,1,206,206,206,8,1,203,203,203,13,1,207,207,207,19,1,
  206,206,206,31,2,206,206,205,205,35,1,207,207,207,38,3,205,205,206,207,207,
  255,2,6,207,207,205,206,205,205,203,203,10,1,205,205,205,26,1,206,206,206,32,
  3,205,205,204,202,202,38,1,204,204,204,40,1,207,207,207,255,2,7,207,207,206,
  206,205,199,205,204,204,10,1,203,203,203,14,1,207,207,207,31,2,202,202,205,
  205,35,3,203,203,202,203,203,39,1,207,207,207,41,1,241,241,241,255,4,5,206,
  206,203,202,207,207,207,10,1,205,205,205,15,1,241,241,241,33,3,241,241,199,
  203,203,37,2,207,207,241,241,255,0,1,207,207,207,2,1,207,207,207,4,4,205,205,
  203,203,202,202,9,2,206,206,205,205,12,1,206,206,206,20,1,207,207,207,30,1,
  205,205,205,33,3,202,202,241,202,202,255,2,2,206,206,207,207,6,1,203,203,203,
  9,1,205,205,205,11,3,242,242,204,206,206,19,1,241,241,241,28,1,207,207,207,
  32,3,241,241,242,203,203,36,2,207,207,207,207,39,1,207,207,207,41,1,207,207,
  207,255,0,1,241,241,241,6,2,205,205,241,241,9,1,206,206,206,12,1,206,206,206,
  18,4,241,241,207,242,242,242,30,2,205,205,205,205,33,2,202,202,206,206,36,1,
  205,205,205,39,1,242,242,242,255,2,3,241,241,207,206,206,6,1,203,203,203,8,2,
  202,202,242,242,14,2,242,242,207,207,32,3,204,204,205,203,203,36,1,205,205,
  205,39,2,206,206,242,242,42,1,206,206,206,255,4,1,206,206,206,6,1,207,207,
  207,8,2,203,203,207,207,12,1,205,205,205,27,2,207,207,206,206,30,7,203,203,
  204,207,204,207,204,207,207,40,1,207,207,207,255,3,1,207,207,207,5,1,241,241,
  241,7,3,241,241,205,241,241,11,1,207,207,207,13,2,241,241,206,206,17,1,207,
  207,207,19,1,207,207,207,29,1,207,207,207,32,6,203,203,203,205,241,207,204,
  204,39,2,205,205,207,207,255,3,1,206,206,206,8,4,205,205,205,205,241,241,13,
  1,207,207,207,15,1,203,203,203,20,2,242,242,207,207,29,3,203,203,242,202,202,
  35,3,205,205,205,205,205,39,2,205,205,206,206,42,1,207,207,207,255,3,8,207,
  207,207,205,205,206,205,207,205,205,13,1,207,207,207,17,1,241,241,241,22,2,
  207,207,207,207,25,2,205,205,207,207,30,11,206,206,241,202,203,203,206,241,
  205,207,207,207,207,255,5,1,241,241,241,7,5,206,206,205,206,206,241,241,16,4,
  204,204,207,204,205,205,25,1,202,202,202,28,2,202,202,205,205,32,1,241,241,
  241,34,1,207,207,207,36,1,205,205,205,255,6,2,207,207,206,206,10,4,204,204,
  206,206,205,205,16,2,206,206,203,203,19,1,243,243,243,21,3,205,205,206,203,
  203,25,1,205,205,205,27,3,207,207,206,207,207,31,1,242,242,242,33,1,206,206,
  206,38,1,207,207,207,40,1,207,207,207,255,6,9,206,206,207,205,241,241,242,
  207,206,207,207,16,1,202,202,202,20,1,203,203,203,23,1,203,203,203,25,2,241,
  241,204,204,28,1,202,202,202,30,1,203,203,203,32,6,205,205,205,205,207,207,
  241,241,39,1,241,241,241,255,7,4,207,207,241,206,206,206,12,2,242,242,241,
  241,15,2,205,205,203,203,18,1,203,203,203,20,1,202,202,202,22,1,202,202,202,
  24,2,204,204,202,202,28,1,202,202,202,31,1,205,205,205,33,3,206,206,207,207,
  207,42,2,207,207,207,207,255,4,2,207,207,206,206,7,5,207,207,207,206,205,205,
  205,14,4,205,205,206,206,241,241,19,7,241,241,203,202,202,204,203,203,203,27,
  1,203,203,203,29,1,207,207,207,32,3,205,205,206,241,241,37,1,241,241,241,40,
  1,207,207,207,255,5,3,207,207,207,207,207,9,1,206,206,206,11,3,206,206,206,
  207,207,15,1,207,207,207,17,1,202,202,202,19,1,202,202,202,21,2,202,202,202,
  202,24,4,203,203,203,204,205,205,30,5,205,205,205,207,206,207,207,36,2,207,
  207,206,206,41,1,207,207,207,255,2,1,207,207,207,9,1,241,241,241,11,2,204,
  204,203,203,14,3,205,205,206,206,206,18,2,202,202,202,202,24,10,203,203,206,
  207,242,206,207,206,207,207,206,206,37,1,207,207,207,39,1,206,206,206,255,4,
  2,206,206,205,205,8,1,207,207,207,10,3,241,241,207,206,206,17,2,241,241,206,
  206,20,2,202,202,202,202,23,1,203,203,203,25,10,207,207,206,206,241,241,207,
  205,206,241,242,242,39,1,241,241,241,255,2,1,207,207,207,4,1,207,207,207,13,
  1,206,206,206,16,1,207,207,207,19,3,204,204,241,202,202,23,1,207,207,207,27,
  6,207,207,241,241,242,207,241,241,34,1,241,241,241,37,1,241,241,241,255,14,1,
  207,207,207,16,9,242,242,205,205,205,205,206,205,207,241,241,29,1,207,207,
  207,37,1,241,241,241,255,12,1,207,207,207,15,5,207,207,242,206,206,206,206,
  21,3,206,206,206,207,207,28,1,207,207,207,31,1,241,241,241,39,1,243,243,243,
  42,1,206,206,206,255,0,1,206,206,206,7,1,207,207,207,16,1,207,207,207,18,2,
  207,207,243,243,255,22,1,241,241,241,32,2,207,207,207,207,255,
};
#endif /* BETA */

// killough 1/31/98: 
// Table of predefined lumps, used to make up for any wad deficiencies.
// Always overriden by iwad or pwads.

// see also the lump definitions above for additional comments

const lumpinfo_t predefined_lumps[]={
  { "M_HORSEN", sizeof m_horsen, m_horsen },
  { "M_VERSEN", sizeof m_versen, m_versen },
  { "M_COMPAT", sizeof m_compat, m_compat },   // killough 10/98
  { "M_GENERL", sizeof m_generl, m_generl },   // killough 10/98

  { "STBR123",  sizeof stbr123, stbr123 }, //jff 02/16/98 bargraph patches
  { "STBR124",  sizeof stbr124, stbr124 },
  { "STBR125",  sizeof stbr125, stbr125 },
  { "STBR126",  sizeof stbr126, stbr126 },
  { "STBR127",  sizeof stbr127, stbr127 },
  { "DIG0",     sizeof dig0, dig0 }, //jff 02/20/98 smaller digit patches
  { "DIG1",     sizeof dig1, dig1 },
  { "DIG2",     sizeof dig2, dig2 },
  { "DIG3",     sizeof dig3, dig3 },
  { "DIG4",     sizeof dig4, dig4 },
  { "DIG5",     sizeof dig5, dig5 },
  { "DIG6",     sizeof dig6, dig6 },
  { "DIG7",     sizeof dig7, dig7 },
  { "DIG8",     sizeof dig8, dig8 },
  { "DIG9",     sizeof dig9, dig9 },
  { "DIGA",     sizeof diga, diga }, //jff 2/22/98 full new smaller hud font
  { "DIGB",     sizeof digb, digb },
  { "DIGC",     sizeof digc, digc },
  { "DIGD",     sizeof digd, digd },
  { "DIGE",     sizeof dige, dige },
  { "DIGF",     sizeof digf, digf },
  { "DIGG",     sizeof digg, digg },
  { "DIGH",     sizeof digh, digh },
  { "DIGI",     sizeof digi, digi },
  { "DIGJ",     sizeof digj, digj },
  { "DIGK",     sizeof digk, digk },
  { "DIGL",     sizeof digl, digl },
  { "DIGM",     sizeof digm, digm },
  { "DIGN",     sizeof dign, dign },
  { "DIGO",     sizeof digo, digo },
  { "DIGP",     sizeof digp, digp },
  { "DIGQ",     sizeof digq, digq },
  { "DIGR",     sizeof digr, digr },
  { "DIGS",     sizeof digs, digs },
  { "DIGT",     sizeof digt, digt },
  { "DIGU",     sizeof digu, digu },
  { "DIGV",     sizeof digv, digv },
  { "DIGW",     sizeof digw, digw },
  { "DIGX",     sizeof digx, digx },
  { "DIGY",     sizeof digy, digy },
  { "DIGZ",     sizeof digz, digz },
  { "DIG45",    sizeof dig45, dig45 },        //jff 3/3/98 add minus sign
  { "DIG47",    sizeof dig47, dig47 },
  { "DIG58",    sizeof dig58, dig58 },
  { "DIG91",    sizeof dig91, dig91 },
  { "DIG93",    sizeof dig93, dig93 },
  { "STKEYS6",  sizeof stkeys6, stkeys6 }, //jff 2/24/98 double key graphics
  { "STKEYS7",  sizeof stkeys7, stkeys7 },
  { "STKEYS8",  sizeof stkeys8, stkeys8 },
  { "BOXUL",    sizeof boxul, boxul }, //jff 2/26/98 background window graphics
  { "BOXUC",    sizeof boxuc, boxuc },
  { "BOXUR",    sizeof boxur, boxur },
  { "BOXCL",    sizeof boxcl, boxcl },
  { "BOXCC",    sizeof boxcc, boxcc },
  { "BOXCR",    sizeof boxcr, boxcr },
  { "BOXLL",    sizeof boxll, boxll },
  { "BOXLC",    sizeof boxlc, boxlc },
  { "BOXLR",    sizeof boxlr, boxlr },
  { "STTMINUS", sizeof sttminus, sttminus }, //jff 2/27/98 -- for v1.2
  { "WIMINUS",  sizeof wiminus,  wiminus},   // killough 4/25/98 -- for v1.1
  { "M_NMARE",  sizeof m_nmare,  m_nmare},   // killough 4/25/98 -- for v1.1
  { "STBAR",    sizeof stbar,    stbar},     // killough 4/25/98 -- for v1.1
  { "DSGETPOW", sizeof dsgetpow, dsgetpow},  // killough 4/25/98 -- for v1.1

  // killough 8/98: remove ENDBOOM lump, use id's ENDOOM instead

  { "S_START"},                              // killough 3/21/98
  { "TNT1A0",   sizeof invis_sprite, invis_sprite },            // invis sprite

#ifdef DOGS
  // killough 7/19/98: Marine's best friend :)
  {"DOGSD5", sizeof dogsd5, dogsd5},
  {"DOGSH5", sizeof dogsd5, dogsd5},
  {"DOGSC5", sizeof dogsc5, dogsc5},
  {"DOGSG5", sizeof dogsc5, dogsc5},
  {"DOGSB1", sizeof dogsb1, dogsb1},
  {"DOGSA5", sizeof dogsa5, dogsa5},
  {"DOGSE5", sizeof dogsa5, dogsa5},
  {"DOGSC1", sizeof dogsc1, dogsc1},
  {"DOGSD1", sizeof dogsd1, dogsd1},
  {"DOGSB5", sizeof dogsb5, dogsb5},
  {"DOGSF5", sizeof dogsb5, dogsb5},
  {"DOGSA1", sizeof dogsa1, dogsa1},
  {"DOGSE1", sizeof dogsa1, dogsa1},
  {"DOGSF1", sizeof dogsf1, dogsf1},
  {"DOGSD2D8", sizeof dogsd2d8, dogsd2d8},
  {"DOGSH2", sizeof dogsd2d8, dogsd2d8},
  {"DOGSG1", sizeof dogsg1, dogsg1},
  {"DOGSH8", sizeof dogsh8, dogsh8},
  {"DOGSE8", sizeof dogse8, dogse8},
  {"DOGSD4D6", sizeof dogsd4d6, dogsd4d6},
  {"DOGSH4", sizeof dogsd4d6, dogsd4d6},
  {"DOGSA2A8", sizeof dogsa2a8, dogsa2a8},
  {"DOGSE2", sizeof dogsa2a8, dogsa2a8},
  {"DOGSF8", sizeof dogsf8, dogsf8},
  {"DOGSH1", sizeof dogsh1, dogsh1},
  {"DOGSA4A6", sizeof dogsa4a6, dogsa4a6},
  {"DOGSE4", sizeof dogsa4a6, dogsa4a6},
  {"DOGSB4B6", sizeof dogsb4b6, dogsb4b6},
  {"DOGSF4", sizeof dogsb4b6, dogsb4b6},
  {"DOGSH6", sizeof dogsh6, dogsh6},
  {"DOGSB2B8", sizeof dogsb2b8, dogsb2b8},
  {"DOGSF2", sizeof dogsb2b8, dogsb2b8},
  {"DOGSC2C8", sizeof dogsc2c8, dogsc2c8},
  {"DOGSG2", sizeof dogsc2c8, dogsc2c8},
  {"DOGSG6", sizeof dogsg6, dogsg6},
  {"DOGSC4C6", sizeof dogsc4c6, dogsc4c6},
  {"DOGSG4", sizeof dogsc4c6, dogsc4c6},
  {"DOGSG8", sizeof dogsg8, dogsg8},
  {"DOGSF6", sizeof dogsf6, dogsf6},
  {"DOGSN0", sizeof dogsn0, dogsn0},
  {"DOGSE6", sizeof dogse6, dogse6},
  {"DOGSD3D7", sizeof dogsd3d7, dogsd3d7},
  {"DOGSH3", sizeof dogsd3d7, dogsd3d7},
  {"DOGSH7", sizeof dogsh7, dogsh7},
  {"DOGSI0", sizeof dogsi0, dogsi0},
  {"DOGSA3A7", sizeof dogsa3a7, dogsa3a7},
  {"DOGSE3", sizeof dogsa3a7, dogsa3a7},
  {"DOGSB3B7", sizeof dogsb3b7, dogsb3b7},
  {"DOGSF3", sizeof dogsb3b7, dogsb3b7},
  {"DOGSF7", sizeof dogsf7, dogsf7},
  {"DOGSE7", sizeof dogse7, dogse7},
  {"DOGSC3C7", sizeof dogsc3c7, dogsc3c7},
  {"DOGSG3", sizeof dogsc3c7, dogsc3c7},
  {"DOGSG7", sizeof dogsg7, dogsg7},
  {"DOGSJ0", sizeof dogsj0, dogsj0},
  {"DOGSK0", sizeof dogsj0, dogsj0},
  {"DOGSL0", sizeof dogsl0, dogsl0},
  {"DOGSM0", sizeof dogsl0, dogsl0},
#endif

#ifdef BETA
  // killough 7/19/98: Pre-Beta BFG fireballs
  {"PLS1A0", sizeof pls1a0, pls1a0},
  {"PLS1B0", sizeof pls1b0, pls1b0},
  {"PLS1C0", sizeof pls1c0, pls1c0},
  {"PLS1D0", sizeof pls1d0, pls1d0},
  {"PLS1E0", sizeof pls1e0, pls1e0},
  {"PLS1F0", sizeof pls1f0, pls1f0},
  {"PLS1G0", sizeof pls1g0, pls1g0},
  {"PLS2A0", sizeof pls2a0, pls2a0},
  {"PLS2B0", sizeof pls2b0, pls2b0},
  {"PLS2C0", sizeof pls2c0, pls2c0},
  {"PLS2D0", sizeof pls2d0, pls2d0},
  {"PLS2E0", sizeof pls2e0, pls2e0},
#endif

  // end of sprites
  { "S_END"},                          // phares 3/9/98    killough 3/21/98

#ifdef DOGS
  {"DSDGSIT", sizeof dsdgsit, dsdgsit},
  {"DSDGATK", sizeof dsdgatk, dsdgatk},
  {"DSDGACT", sizeof dsdgact, dsdgact},
  {"DSDGDTH", sizeof dsdgdth, dsdgdth},
  {"DSDGPAIN", sizeof dsdgpain, dsdgpain},
  {"DOGOVRLY", sizeof dogovrly, dogovrly},
#endif

  { "SWITCHES", sizeof switches, switches},   // jff 3/23/98 switch list
  { "ANIMATED", sizeof animated, animated},   // jff 3/23/98 animation list
  { "M_SETUP" , sizeof m_setup,  m_setup },   // phares 3/24/98 menu item
  { "M_KEYBND", sizeof m_keybnd, m_keybnd},   // phares 3/24/98 menu item
  { "M_AUTO"  , sizeof m_auto,   m_auto  },   // phares 3/30/98 menu item
  { "M_CHAT"  , sizeof m_chat,   m_chat  },   // phares 4/10/98 menu item
  { "M_ENEM",   sizeof m_enem,   m_enem  },   // phares 4/08/98 menu item
  { "M_STAT",   sizeof m_stat,   m_stat  },   // phares 3/29/98 menu item
  { "M_WEAP",   sizeof m_weap,   m_weap  },   // phares 3/28/98 menu item
  { "M_MESS"  , sizeof m_mess,   m_mess  },   // phares 4/08/98 menu item
  { "M_COLORS", sizeof m_colors, m_colors},   // phares 4/01/98 menu item
  { "M_PALNO",  sizeof m_palno,  m_palno},    // phares 4/01/98 menu item
  { "C_START" },  // killough 4/4/98: colormap markers
  { "WATERMAP", sizeof watermap, watermap},   // killough 3/21/98, 4/4/98
  { "C_END" },    // killough 4/4/98: end colormap markers
  { "STCFN096", sizeof stcfn096, stcfn096},   // jff 4/5/98 missing char
  { "M_BUTT1",  sizeof m_butt1,  m_butt1},    // phares 4/16/98 reset button
  { "M_BUTT2",  sizeof m_butt2,  m_butt2},    // phares 4/16/98 reset button
  { "CRBRICK",  sizeof cr_brick, cr_brick},   // jff 4/24/98 brick color range
  { "CRTAN",    sizeof cr_tan,   cr_tan},     // jff 4/24/98 tan color range
  { "CRGRAY",   sizeof cr_gray,  cr_gray},    // jff 4/24/98 gray color range
  { "CRGREEN",  sizeof cr_green, cr_green},   // jff 4/24/98 green color range
  { "CRBROWN",  sizeof cr_brown, cr_brown},   // jff 4/24/98 brown color range
  { "CRGOLD",   sizeof cr_gold,  cr_gold},    // jff 4/24/98 gold color range
  { "CRRED",    sizeof cr_red,   cr_red},     // jff 4/24/98 red color range
  { "CRBLUE",   sizeof cr_blue,  cr_blue},    // jff 4/24/98 blue color range
  { "CRBLUE2",  sizeof cr_blue2, cr_blue2},   // jff 4/24/98 blue2 color range
  { "CRORANGE", sizeof cr_orange,cr_orange},  // jff 4/24/98 orange color range
  { "CRYELLOW", sizeof cr_yellow,cr_yellow},  // jff 4/24/98 yellow color range
  { "MBFTEXT",  sizeof mbftext, mbftext},     // killough 11/98: mbf title text
};

// Calculate how many of them there are
const size_t num_predefined_lumps =
  sizeof predefined_lumps / sizeof *predefined_lumps;

//----------------------------------------------------------------------------
//
// $Log: info.c,v $
// Revision 1.2  2000-08-12 21:29:25  fraggle
// change license header
//
// Revision 1.1.1.1  2000/07/29 13:20:39  fraggle
// imported sources
//
// Revision 1.44  1998/05/12  12:46:36  phares
// Removed OVER_UNDER code
//
// Revision 1.43  1998/05/12  09:35:07  phares
// Corrected 4001->5001 and 4002->5002 in OVER/UNDER table
//
// Revision 1.42  1998/05/12  08:41:13  jim
// fix decl of endboom
//
// Revision 1.40  1998/05/11  12:21:20  jim
// 4001/2 deconflicted with DosDOOM
//
// Revision 1.39  1998/05/06  11:30:54  jim
// Moved predefined lump writer info->w_wad
//
// Revision 1.38  1998/05/04  21:34:49  thldrmn
// commenting and reformatting
//
// Revision 1.37  1998/05/03  23:23:50  killough
// Fix #includes at the top, nothing else
//
// Revision 1.36  1998/04/29  09:20:37  jim
// New ENDBOOM
//
// Revision 1.35  1998/04/27  02:15:10  killough
// Fix cr_gold declaration, add missing v1.1 lumps
//
// Revision 1.34  1998/04/24  08:08:36  jim
// Make text translate tables lumps
//
// Revision 1.33  1998/04/22  13:45:37  phares
// Added Setup screen Reset to Defaults
//
// Revision 1.32  1998/04/22  06:34:43  killough
// Make WritePredefinedLumpWad endian-independent, remove tabs
//
// Revision 1.31  1998/04/21  23:46:21  jim
// Predefined lump dumper option
//
// Revision 1.30  1998/04/17  00:04:11  jim
// text file changes and new ENDBOOM
//
// Revision 1.29  1998/04/12  22:54:55  phares
// Remaining 3 Setup screens
//
// Revision 1.28  1998/04/06  04:36:51  killough
// Change WATERMAP, add C_START/C_END
//
// Revision 1.27  1998/04/05  10:10:13  jim
// added STCFN096 lump
//
// Revision 1.26  1998/04/03  19:18:46  phares
// Automap Palette work, slot 0 = disable, 247 = BLACK
//
// Revision 1.25  1998/04/02  05:01:49  jim
// Added ENDOOM, BOOM.TXT mods
//
// Revision 1.24  1998/04/01  15:34:30  phares
// Added Automap Setup Screen, fixed Seg Viol in Setup Menus
//
// Revision 1.23  1998/03/31  01:08:26  phares
// Initial Setup screens and Extended HELP screens
//
// Revision 1.22  1998/03/23  18:39:10  jim
// Switch and animation tables now lumps
//
// Revision 1.21  1998/03/23  15:23:54  phares
// Changed pushers to linedef control
//
// Revision 1.20  1998/03/23  03:18:09  killough
// Add WATERMAP colormap lump for underwater viewing
//
// Revision 1.19  1998/03/09  18:30:28  phares
// Added invisible sprite for MT_PUSH
//
// Revision 1.18  1998/03/09  07:15:14  killough
// Remove unnecessary translucency lumps
//
// Revision 1.17  1998/03/04  22:23:04  phares
// Removed BOOMHELP predefined lump
//
// Revision 1.16  1998/03/04  11:52:43  jim
// Add TRAN50 TRAN66 predefined lumps
//
// Revision 1.15  1998/03/03  00:21:54  jim
// Added predefined ENDBETA lump for beta test
//
// Revision 1.13  1998/02/27  11:51:50  jim
// Add predefined lump STTMINUS
//
// Revision 1.11  1998/02/24  08:45:44  phares
// Pushers, recoil, new friction, and over/under work
//
// Revision 1.10  1998/02/24  04:13:37  jim
// Added double keys to status
//
// Revision 1.8  1998/02/23  04:31:13  killough
// Make tranlucency apply realistically
//
// Revision 1.7  1998/02/22  12:51:38  jim
// HUD control on F5, z coord, spacing change
//
// Revision 1.5  1998/02/20  21:56:49  phares
// Preliminarey sprite translucency
//
// Revision 1.4  1998/02/18  00:59:44  jim
// Addition of HUD
//
// Revision 1.3  1998/02/02  13:36:12  killough
// Add predefined lumps
//
// Revision 1.2  1998/01/26  19:23:34  phares
// First rev with no ^Ms
//
// Revision 1.1.1.1  1998/01/19  14:02:56  rand
// Lee's Jan 19 sources
//
//----------------------------------------------------------------------------
